
(* ::Section:: *)
(* Public Declaration Part: executes in Theorema`Provers` *)

With[ {lang = "English"},

MessageName[ complexityGBBivariate, "usage", lang] = "Rules for proving statements about the complexity of Buchberger's algorithm (especially in the bivariate case)";

MessageName[ failedInteractiveRule, "usage", lang] = "Interactive rule could not be applied";

MessageName[ CGB2notGoal, "usage", lang] = "Prove negation by contradiction";
MessageName[ CGB2andKB, "usage", lang] = "Split conjunctions in the knowledge base";
MessageName[ CGB2andGoal, "usage", lang] = "Split a conjunction in the goal";
MessageName[ CGB2orGoal, "usage", lang] = "Prove disjunction by proving last disjunct";
MessageName[ CGB2orGoalInteractive, "usage", lang] = "Interactively select disjunct to prove";
MessageName[ CGB2orKB, "usage", lang] = "Case distinction based on a disjunction in the knowledge base";
MessageName[ CGB2implGoalDirect, "usage", lang] = "Prove implication directly";
MessageName[ CGB2equivKB, "usage", lang] = "Transform equivalences in the knowledge base into implications";
MessageName[ CGB2exchangeGoalKBInteractive, "usage", lang] = "Interactively exchange the goal and a formula in the knowledge base";
MessageName[ CGB2expandNegationKB, "usage", lang] = "Simplify negated assumptions";
MessageName[ CGB2expandNegationGoal, "usage", lang] = "Simplify negated goal";

MessageName[ CGB2integerIntervalKB, "usage", lang] = "Handle integer intervals in the knowledge base";
MessageName[ CGB2integerIntervalGoal, "usage", lang] = "Handle integer intervals in the goal";

MessageName[ CGB2subscriptTupleOf, "usage", lang] = "Handle a.b.f. subscripts of abstraction tuples";
MessageName[ CGB2elementTupleOf, "usage", lang] = "Handle subscripts of abstraction tuples with trivial conditions";
MessageName[ CGB2lengthTupleOf, "usage", lang] = "Handle lengths of abstraction tuples with trivial condition";
MessageName[ CGB2nonemptyTupleOf, "usage", lang] = "Prove non-emptiness of abstraction tuples";
MessageName[ CGB2isTupleTupleOf, "usage", lang] = "Prove that abstraction tuples are tuples";
MessageName[ CGB2appendElement, "usage", lang] = "Handle appending elements to tuples by case distinctions";
MessageName[ CGB2appendTupleOf, "usage", lang] = "Handle abstraction tuples constructed over existing tuples with some element appended";

MessageName[ CGB2cancellationGoal, "usage", lang] = "Simplify the goal by using cancellation property in TOMs";
MessageName[ CGB2cancellationKB, "usage", lang] = "Simplify assumptions by using cancellation property in TOMs";

MessageName[ CGB2contradictionOrdering, "usage", lang] = "Knowledge base contains contradicting formulae in theory of totally-ordered domains";
MessageName[ CGB2notLessEqual, "usage", lang] = "Handle negated order relations in totally-ordered domains";
MessageName[ CGB2reflexiveClosure, "usage", lang] = "Infer reflexive relations from strict ones in totally-ordered domains";
MessageName[ CGB2transitivityKB, "usage", lang] = "Infer new formulae by transitivity in totally-ordered domains";
MessageName[ CGB2transitivityGoal, "usage", lang] = "Use transitivity in totally-ordered domains for changing proof goal";
MessageName[ CGB2transitivityGoalAddition, "usage", lang] = "Use transitivity/cancellation properties in TOMs on both sides of inequalities";
MessageName[ CGB2transitivityGoalAdditionLeft, "usage", lang] = "Use transitivity/cancellation properties in TOMs only on the LHS of inequalities";

MessageName[ CGB2ArgMin, "usage", lang] = "Handle argmin-quantifier in totally-ordered domains";
MessageName[ CGB2ArgMax, "usage", lang] = "Handle argmax-quantifier in totally-ordered domains";
MessageName[ CGB2tupleAdditionGoal, "usage", lang] = "Move addition into tuples";
MessageName[ CGB2minGoalLeft, "usage", lang] = "'min[S] < x' in the goal";
MessageName[ CGB2maxGoalRight, "usage", lang] = "'x < max[S]' in the goal";
MessageName[ CGB2minKBRight, "usage", lang] = "'x < min[S]' among the assumptions";
MessageName[ CGB2maxKBLeft, "usage", lang] = "'max[S] < x' among the assumptions";

MessageName[ CGB2forallGoal, "usage", lang] = "Prove universally quantified goal";
MessageName[ CGB2existsGoal, "usage", lang] = "Prove existentially quantified goal by matching assumptions";
MessageName[ CGB2existsKB, "usage", lang] = "Instantiate existentially quantified formula";
MessageName[ CGB2forallKBInteractive, "usage", lang] = "Interactively instantiate a universally quantified formula in the knowledge base";
MessageName[ CGB2existsGoalInteractive, "usage", lang] = "Interactively instantiate existentially quantified goal";
MessageName[ CGB2existsTupleOfGoal, "usage", lang] = "Handle existential quantifiers ranging over abstraction tuples in the goal";
MessageName[ CGB2forallTupleOfKB, "usage", lang] = "Handle universal quantifiers ranging over abstraction tuples in the knowledge base";
MessageName[ CGB2existsAppendGoal, "usage", lang] = "Handle existential quantifiers ranging over appended-tuples in the goal";
MessageName[ CGB2forallAppendKB, "usage", lang] = "Handle universal quantifiers ranging over appended-tuples in the knowledge base";

MessageName[ CGB2modusPonensTollens, "usage", lang] = "Modus ponens and modus tollens";
MessageName[ CGB2selectImplicationInteractive, "usage", lang] = "Interactively select an implication in the knowledge base for proving its premise and assuming its conclusion";
MessageName[ CGB2generateRewriteRules, "usage", lang] = "Generate rewrite rules from initial proof situation";
MessageName[ CGB2goalRewriting, "usage", lang] = "Goal rewriting based on formulae in the knowledge base";
MessageName[ CGB2elementarySubstitution, "usage", lang] = "Substitution based on quantified equalities and equivalences in the knowledge base";
MessageName[ CGB2groundSubstitution, "usage", lang] = "Substitution based on ground equalities and equivalences in the knowledge base";
MessageName[ CGB2expandDef, "usage", lang] = "Expand explicit definitions";

] (* With *)


(* ::Section:: *)
(* Private Implementation Part: executes in Theorema`Provers`Private` *)

Begin["`Private`"]

With[ {lang = "English"},

translate[ "Complexity of bivariate Buchberger's algorithm", lang] = "Rules in Theory of Complexity Analysis of bivariate Buchberger's Algorithm";
translate[ "Propositional Logic Rules", lang] = "Propositional Logic";
translate[ "Miscellaneous Rules", lang] = "Miscellaneous";
translate[ "Integer Rules", lang] = "Integers";
translate[ "Tuple Rules", lang] = "Tuples";
translate[ "Addition Rules", lang] = "Addition in TOMs";
translate[ "Ordering Rules", lang] = "Order Relation in TOMs";
translate[ "min/max Rules", lang] = "Minimum and Maximum";
translate[ "Quantifier Rules", lang] = "Logical Quantifiers";
translate[ "Implication Rules", lang] = "Implications";
translate[ "Rules for Rewriting", lang] = "Rewriting";

translate[ "currentGoal", lang] = "Current Goal";
translate[ "currentKnowledge", lang] = "Current Knowledge";

translate[ "goalHeader", lang] = "Select goal to prove";
translate[ "!goal", lang] = "Abort";

translate[ "exchangeHeader", lang] = "Exchange Goal and Assumption Formula";
translate[ "!exchange", lang] = "Do not exchange";

translate[ "selectImplicationHeader", lang] = "Select Implication";
translate[ "!selectImplication", lang] = "Do not select an implication";

translate[ "selectForallHeader", lang] = "Select Universally Quantified Formula";
translate[ "!selectForall", lang] = "Do not select a formula";

translate[ "instantiateForallHeader", lang] = "Instantiate Universally Quantified Formula";
translate[ "instantiateExistsHeader", lang] = "Instantiate Existentially Quantified Formula";

proofStepText[ failedInteractiveRule, lang, ___] := {};

proofStepText[ CGB2andGoal, lang, {{g_}}, _, ___] :=
	{textCell[ "For proving ", formulaReference[ g],
		" we prove the individual conjuncts one after the other, immediately adding the ones already proved to our knowledge base."]
	};
subProofHeader[ CGB2andGoal, lang, _, {generated_List}, ___, pVal_, {p_}] := {textCell[ "Proof of ", formulaReference[ Part[ generated, p]], ":"],
	textCell[ "We need to prove"],
	goalCell[ Part[ generated, p], "."]
	};

proofStepText[ CGB2andKB, lang, {used_List}, {generated_List}, ___] :=
	Module[ {suffix = If[ Length[ used] === 1, "", "s"]},
		{textCell[ "We split the conjunction" <> suffix <> " ", formulaReferenceSequence[ used, lang], " and add"],
			assumptionListCells[ generated, ",", ""],
			textCell[ "to the knowledge base."]
		}
	];
	
proofStepText[ CGB2orKB, lang, {{g_}}, {generated_List}, ___] :=
	{textCell[ "Based on the assumption ", formulaReference[ g], " we distinguish " <>
		Switch[ Length[ generated],
			2, "two",
			3, "three",
			4, "four",
			5, "five",
			6, "six",
			7, "seven",
			8, "eight",
			9, "nine",
			10, "ten",
			11, "eleven",
			12, "twelve",
			_, ToString[ Length[ generated]]
		]
		<> " cases:"]};

subProofHeader[ CGB2orKB, lang, _, {generated_List}, ___, pVal_, {p_}] := {textCell[ "Case ", ToString[p], ": we assume"],
	assumptionCell[ Part[ generated, p], "."]
	};
	
proofStepText[ CGB2equivKB, lang, {equiv_List}, gen:{{_, _}..}, ___] /; Length[ equiv] === Length[ gen] :=
	Module[ {},
		Flatten[ MapThread[(
				{
					textCell[ "From equivalence ", formulaReference[ #1], " we obtain the implications"],
					assumptionListCells[ #2, ",", "."]
				}
				)&,
				{equiv, gen}
			],
			1
		]
	];
	
proofStepText[ CGB2integerIntervalKB, lang, {{k_}}, {}, proved] :=
	{ 
		textCell[ "The proof is finished, because ", formulaReference[ k],
			" clearly is a contradiction in the knowledge base (empty interval does not contain elements)."]
	};
proofStepText[ CGB2integerIntervalKB, lang, {{k_}}, {gen_List}, ___] :=
	{ 
		textCell[ "We replace formula ", formulaReference[ k], " in the knowledge base by"],
		assumptionListCells[ gen, ",", "."]
	};
proofStepText[ CGB2integerIntervalGoal, lang, {{g_}}, {{gen_}}, ___] :=
	{
		textCell[ "In order to show ", formulaReference[ g], " we show"],
		goalCell[ gen], textCell["instead."]
	};
	
proofStepText[ CGB2subscriptTupleOf, lang, {{T_, j_}}, {{abf_, _, t_, _, _, _, _}}, ___] :=
	{
		textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ Subscript[ T, j]],
			" by ", inlineTheoremaExpression[ t], ", where ", inlineTheoremaExpression[ abf], " is some fresh a.b.f. constant."]
	};
subProofHeader[ CGB2subscriptTupleOf, lang, _, {{_, cond_, _, _, _, _, _}}, ___, {1}] :=
	{
		textCell[ "Before that, however, we first have to show"], goalCell[ cond, "."]
	};
subProofHeader[ CGB2subscriptTupleOf, lang, _, {{_, cond_, _, phi_, RC_, newG_, newK_}}, ___, {2}] :=
	{
		textCell[ "Now we continue showing"], goalCell[ newG],
		textCell[ "having"], assumptionListCells[ Join[ {cond, phi, RC}, newK], ",", ""],
		textCell[ "among our assumptions."]
	};
proofStepText[ CGB2elementTupleOf, lang, {{S_, j_}}, {gen_List, {t_}}, ___] :=
	{
		textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ Subscript[ S, j]],
			" by ", inlineTheoremaExpression[ t], ", meaning that we now have"],
		assumptionListCells[ gen, ",", ""],
		textCell[ "among our assumptions. The validity of this replacement has already been checked in a previous step."]
	};
proofStepText[ CGB2elementTupleOf, lang, {{S_, j_}}, {{g_}, gen_List, {t_}}, ___] :=
	If[ Length[ gen] > 0,
		{textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ Subscript[ S, j]],
				" by ", inlineTheoremaExpression[ t], ", meaning that we now have to prove"],
			goalCell[ g], textCell[ "under the additional assumptions"],
			assumptionListCells[ gen, ",", ""],
			textCell[ "The validity of this replacement has already been checked in a previous step."]},
		{textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ Subscript[ S, j]],
				" by ", inlineTheoremaExpression[ t], ", meaning that we now have to prove"],
			goalCell[ g, "."], textCell[ "The validity of this replacement has already been checked in a previous step."]}
	];
proofStepText[ CGB2elementTupleOf, lang, {{S_, j_}}, {{_, _}, _List, {t_}}, ___] :=
	{
		textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ Subscript[ S, j]],
			" by ", inlineTheoremaExpression[ t], "."]
	};
subProofHeader[ CGB2elementTupleOf|CGB2lengthTupleOf, lang, _, {{cond_, _}, _List, _}, ___, {1}] :=
	{
		textCell[ "Before that, however, we first have to prove"], goalCell[ cond, "."]
	};
subProofHeader[ CGB2elementTupleOf|CGB2lengthTupleOf, lang, _, {{cond_, g_}, gen_List, _}, ___, {2}] :=
	{
		textCell[ "Now we continue proving"], goalCell[ g],
		textCell[ "having"], assumptionListCells[ Prepend[ gen, cond], ",", ""],
		textCell[ "among our assumptions."]
	};
proofStepText[ CGB2lengthTupleOf, lang, {{S_}}, {gen_List, {u_}}, ___] :=
	{
		textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ BracketingBar[ S]],
			" by ", inlineTheoremaExpression[ u], ", meaning that we now have"],
		assumptionListCells[ gen, ",", ""],
		textCell[ "among our assumptions. The validity of this replacement has already been checked in a previous step."]
	};
proofStepText[ CGB2lengthTupleOf, lang, {{S_}}, {{g_}, gen_List, {u_}}, ___] :=
	If[ Length[ gen] > 0,
		{textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ BracketingBar[ S]],
				" by ", inlineTheoremaExpression[ u], ", meaning that we now have to prove"],
			goalCell[ g], textCell[ "under the additional assumptions"],
			assumptionListCells[ gen, ",", ""],
			textCell[ "The validity of this replacement has already been checked in a previous step."]},
		{textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ BracketingBar[ S]],
				" by ", inlineTheoremaExpression[ u], ", meaning that we now have to prove"],
			goalCell[ g, "."], textCell[ "The validity of this replacement has already been checked in a previous step."]}
	];
proofStepText[ CGB2lengthTupleOf, lang, {{S_}}, {{_, _}, _List, {u_}}, ___] :=
	{
		textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ BracketingBar[ S]],
		" by ", inlineTheoremaExpression[ u], "."]
	};
proofStepText[ CGB2nonemptyTupleOf, lang, {{g_}}, {{gen_}}, ___] :=
	{
		textCell[ "If ", formulaReference[ g], " has to be shown, we show"],
		goalCell[ gen], textCell["instead."]
	};
proofStepText[ CGB2isTupleTupleOf, lang, {{g_}}, {{gen_}}, ___] :=
	{
		textCell[ "If ", formulaReference[ g], " has to be shown, we show"],
		goalCell[ gen], textCell[ "instead."]
	};
proofStepText[ CGB2appendElement, lang, {{A_, x_, t_}}, {{first_, rest___}, {}}, ___] :=
	If[ ListQ[ first] && Length[ first] === 1,
		Append[
			If[ Length[ {rest}] > 0,
				{textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ Subscript[ Theorema`Language`appendElem$TM[ A, x], t]],
					" by ", inlineTheoremaExpression[ Subscript[ A, t]], ", meaning that we now have to prove"],
				goalCell[ First[ first]], textCell[ " under the additional assumptions"],
				assumptionListCells[ {rest}, ",", "."]
				},
				{textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ Subscript[ Theorema`Language`appendElem$TM[ A, x], t]],
					" by ", inlineTheoremaExpression[ Subscript[ A, t]], ", meaning that we now have to prove"],
				goalCell[ First[ first], "."]
				}
			],
			textCell[ "This replacement is performed according to the case assumption made in a previous step."]
		],
		{textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ Subscript[ Theorema`Language`appendElem$TM[ A, x], t]],
			" by ", inlineTheoremaExpression[ Subscript[ A, t]], ", meaning that we now have"],
		assumptionListCells[ {first, rest}, ",", ""],
		textCell[ "among our assumptions. This replacement is performed according to the case assumption made in a previous step."]
		}
	];
proofStepText[ CGB2appendElement, lang, {{A_, x_, t_}}, {{}, {first_, rest___}}, ___] :=
	If[ ListQ[ first] && Length[ first] === 1,
		Append[
			If[ Length[ {rest}] > 0,
				{textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ Subscript[ Theorema`Language`appendElem$TM[ A, x], t]],
					" by ", inlineTheoremaExpression[ x], ", meaning that we now have to prove"],
				goalCell[ First[ first]], textCell[ " under the additional assumptions"],
				assumptionListCells[ {rest}, ",", "."]
				},
				{textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ Subscript[ Theorema`Language`appendElem$TM[ A, x], t]],
					" by ", inlineTheoremaExpression[ x], ", meaning that we now have to prove"],
				goalCell[ First[ first], "."]
				}
			],
			textCell[ "This replacement is performed according to the case assumption made in a previous step."]
		],
		{textCell[ "We replace all occurrences of ", inlineTheoremaExpression[ Subscript[ Theorema`Language`appendElem$TM[ A, x], t]],
			" by ", inlineTheoremaExpression[ x], ", meaning that we now have"],
		assumptionListCells[ {first, rest}, ",", ""],
		textCell[ "among our assumptions. This replacement is performed according to the case assumption made in a previous step."]
		}
	];
proofStepText[ CGB2appendElement, lang, {{A_, x_, t_}}, {{__}, {__}, {_, {_, _}, {_, _, _}}}, ___] :=
	{
		textCell[ "We make a case distinction depending on whether index ", inlineTheoremaExpression[ t], " of tuple ",
			inlineTheoremaExpression[ Theorema`Language`appendElem$TM[ A, x]], " is an index of ", inlineTheoremaExpression[ A],
			" or not."
		]
	};
subProofHeader[ CGB2appendElement, lang, {{A_, x_, t_}}, {_List, _List, {cond_, _List, _List}}, ___, {1}] :=
	{
		textCell[ "Before that, however, we first have to verify that the term ",
			inlineTheoremaExpression[ Subscript[ Theorema`Language`appendElem$TM[ A, x], t]],
			" is well-defined, i.e. we have to prove"],
		goalCell[ cond, "."]
	};
subProofHeader[ CGB2appendElement, lang, {{A_, x_, t_}}, gen:{{__}, {__}, {_, {__}, {__}}}, ___, {i:(2|3)}] :=
Module[ {new, assm, out},
	new = gen[[i - 1]];
	assm = gen[[3, i]];
	out = {
		textCell[ "Case " <> ToString[ i - 1] <> ": We assume"],
		assumptionListCells[ assm, ",", "."],
		textCell[ "This means that we can replace every occurrence of ", inlineTheoremaExpression[ Subscript[ Theorema`Language`appendElem$TM[ A, x], t]],
			" by ", Switch[ i, 2, inlineTheoremaExpression[ Subscript[ A, t]], 3, inlineTheoremaExpression[ x]], "."]
		
	};
	If[ ListQ[ First[ new]] && Length[ First[ new]] === 1,
		If[ Length[ new] > 1,
			Join[ out, {textCell[ "Hence, we now have to prove"],
						goalCell[ First[ First[ new]]],
						textCell[ "under the assumptions"],
						assumptionListCells[ Rest[ new], ",", "."]}],
			Join[ out, {textCell[ "Hence, we now have to prove"],
						goalCell[ First[ First[ new]], "."]}]
		],
		Join[ out, {textCell[ "Hence, we now have"],
					assumptionListCells[ new, ",", ""],
					textCell[ "among our assumptions."]}]
	]
];
proofStepText[ CGB2appendTupleOf, lang, {{t_, _, _, _}}, {{__}, {__}, {_, {_, _}}}, ___] :=
	{
		textCell[ "We simplify the abstraction tuple ", inlineTheoremaExpression[ t], " by moving the 'append'-operator out of it."]
	};
subProofHeader[ CGB2appendTupleOf, lang, {{_, A_, _, _}}, {_List, _List, {cond_, _List}}, ___, {1}] :=
	{
		textCell[ "Before that, however, we first have to verify that ",
			inlineTheoremaExpression[ A],
			" is a tuple, i.e. we have to prove"],
		goalCell[ cond, "."]
	};
subProofHeader[ CGB2appendTupleOf, lang, {used:{t_, _, _, _}}, gen:{{__}, {__}, {cond_, phi_List}}, ___, {i:(2|3)}] :=
Module[ {new, assm, out},
	new = gen[[i - 1]];
	assm = {cond, gen[[3, 2, i - 1]]};
	out = {
		textCell[ "Case " <> ToString[ i - 1] <> ": We assume"],
		assumptionListCells[ assm, ",", "."],
		textCell[ "This means that we can replace every occurrence of ", inlineTheoremaExpression[ t],
			" by ", inlineTheoremaExpression[ used[[i + 1]]], "."]
		
	};
	If[ ListQ[ First[ new]] && Length[ First[ new]] === 1,
		If[ Length[ new] > 1,
			Join[ out, {textCell[ "Hence, we now have to prove"],
						goalCell[ First[ First[ new]]],
						textCell[ "under the assumptions"],
						assumptionListCells[ Rest[ new], ",", "."]}],
			Join[ out, {textCell[ "Hence, we now have to prove"],
						goalCell[ First[ First[ new]], "."]}]
		],
		Join[ out, {textCell[ "Hence, we now have"],
					assumptionListCells[ new, ",", ""],
					textCell[ "among our assumptions."]}]
	]
];
	
proofStepText[ CGB2cancellationGoal, lang, {{old_}}, _, _] :=
	{textCell[ "We can use the cancellation property to simplify our current goal ", formulaReference[ old], ":"]};
subProofHeader[ CGB2cancellationGoal|CGB2cancellationKB, lang, _, {{cond_, _}}, ___, {1}] :=
	{textCell[ "In order to use the cancellation property, we first have to show"], goalCell[ cond, "."]};
subProofHeader[ CGB2cancellationGoal, lang, _, {{cond_, g_}}, ___, {2}] :=
	{textCell[ "Now we continue proving"], goalCell[ g], textCell["having"], assumptionCell[ cond], textCell["among our assumptions."]};
proofStepText[ CGB2cancellationKB, lang, {{_, old_}}, _, _] :=
	{textCell[ "We can use the cancellation property to simplify formula ", formulaReference[ old], " in the knowledge base:"]};
subProofHeader[ CGB2cancellationKB, lang, {{g_, _}}, {l:{_, _}}, ___, {2}] :=
	{textCell[ "Now we continue proving our original goal ", formulaReference[ g], ", having"], assumptionListCells[ l, ",", ""], textCell["among our assumptions."]};

proofStepText[ CGB2contradictionOrdering, lang, {{k_}}, {{g_}}, ___] := {textCell[ "Formula ", formulaReference[ k], 
	" seems to be a contradiction in the theory of totally-ordered domains: We only need to verify "],
	goalCell[ g, "."]
    };
proofStepText[ CGB2notLessEqual, lang, {{form_}}, {{newForm_, __}}, ___] :=
	{textCell[ "Since formula ", inlineTheoremaExpression[ form], " is in our proof situation, we replace it by ",
		inlineTheoremaExpression[ newForm], "."
	]};
subProofHeader[ CGB2notLessEqual, lang, _, {{_, newg_, ___}}, ___, {1}] :=
	{textCell[ "Before that, however, we first have to show"], goalCell[ newg, "."]};
subProofHeader[ CGB2notLessEqual, lang, _, {{_, _, newg_}}, ___, {2}] :=
	{textCell[ "Now we continue proving"], goalCell[ newg, "."]};
subProofHeader[ CGB2notLessEqual, lang, _, {{_, _, newg_, kb__}}, ___, {2}] :=
	{
		textCell[ "Now we continue proving"], goalCell[ newg], textCell[ "having"],
		assumptionListCells[ {kb}, ",", ""], textCell[ "in our knowledge base."]
	};
proofStepText[ CGB2reflexiveClosure, lang, {{k_, newg_, dom_, x_, y_, g_}}, {gen_}, ___] :=
	{
	 	textCell[ "Since ", formulaReference[ k], " is known, but ", inlineTheoremaExpression[ DomainOperation$TM[ dom, LessEqual$TM][ x, y]], 
	 	" is not, we must first show "], goalCell[ newg], textCell[ " without ", formulaReference[ k], 
	 	" in our knowledge base. Then we continue with proving "], goalCell[ g], textCell[ " with "],
	 	assumptionListCells[ gen, ",", ""], textCell[ " in our knowledge base."]
	};
subProofHeader[ CGB2reflexiveClosure, lang, {{_, newg_, __}}, __, {1}] :=
	{textCell[ "First part: We show"], goalCell[ newg, "."]};
subProofHeader[ CGB2reflexiveClosure, lang, {{__, g_}}, {gen_List}, ___, {2}] :=
	{
		textCell[ "Second part: We continue proving"], goalCell[ g], textCell[ "having"],
		assumptionListCells[ gen, ",", ""], textCell[ "in our knowledge base."]
	};
proofStepText[ CGB2transitivityKB, lang, {{k_, c_}}, {{g_, f_}}, ___] := {textCell[ "We take ", formulaReference[ k], 
	" and ", formulaReference[ c], " for making use of transitivity in totally-ordered domains. Before, however, we need to verify "],
	goalCell[ g, "."]
    };
subProofHeader[ CGB2transitivityKB, lang, _, {{g_, _}}, ___, {1}] := {textCell[ "First part of transitivity: We prove"],
	goalCell[ g, "."]
	};
subProofHeader[ CGB2transitivityKB, lang, _, {{_, f_}}, ___, {2}] := {textCell[ "Second part of transitivity: We continue with the main branch, having"],
	assumptionCell[ f], textCell[ " among our assumptions."]
	};
proofStepText[ CGB2transitivityGoal, lang, {{g_, k_}}, {{newG_}}, ___] :=
	{
		textCell[ "Rather than proving ", formulaReference[ g], " we use assumption ", formulaReference[ k],
		" and prove"], goalCell[ newG],
		textCell[ "instead. This step is justified by transitivity of the order relation in totally-ordered domains."]
    };
proofStepText[ CGB2transitivityGoalAddition|CGB2transitivityGoalAdditionLeft, lang, {{g_, f_}}, _, ___, "side" -> s_, ___] :=
	{
		textCell[ "Rather than proving ", formulaReference[ g], " we make use of formula ",
		formulaReference[ f], " and transitivity in totally-ordered Abelian monoids to rewrite its ",
		Switch[ s, Left, "left", _, "right"], "-hand-side."]
    };
subProofHeader[ CGB2transitivityGoalAddition|CGB2transitivityGoalAdditionLeft, lang, _, {{cond_, _}}, ___, {1}] :=
	{textCell[ "Before that, however, we first have to show"], goalCell[ cond, "."]};
subProofHeader[ CGB2transitivityGoalAddition|CGB2transitivityGoalAdditionLeft, lang, _, {{cond_, g_}}, ___, {2}] :=
	{textCell[ "Now we continue proving"], goalCell[ g], textCell["having"], assumptionCell[ cond], textCell["among our assumptions."]};
    
proofStepText[ CGB2ArgMin|CGB2ArgMax, lang, {{term_, abf_}}, _List, ___] :=
	{
		textCell[ "We replace every occurrence of ", inlineTheoremaExpression[ term],
		" by the fresh arbitrary but fixed constant ", inlineTheoremaExpression[ abf],
		". Before, we still have to check that ", inlineTheoremaExpression[ term], " is well-defined:"]
	};
subProofHeader[ CGB2ArgMin|CGB2ArgMax, lang, {{term_, _}}, l:{{cond1_, __}}, ___, {1}] :=
	{
		textCell[ "We prove that the range of ", inlineTheoremaExpression[ term], " is non-empty:"],
		goalCell[ cond1, "."]
	};
subProofHeader[ CGB2ArgMin|CGB2ArgMax, lang, {{term_, _}}, {{_, {cond2_, _, _, _}, __}}, ___, {2}] :=
	{
		textCell[ "We prove that the terms in ", inlineTheoremaExpression[ term], " are domain elements:"],
		goalCell[ cond2, "."]
	};
subProofHeader[ CGB2ArgMin|CGB2ArgMax, lang, _List, {{_, {cond2_, assm1_, assm2_, assm3_}, {goal_}, rest___}}, ___, {3}] :=
	{
		textCell[ "We now continue with the main branch, meaning that we prove"], goalCell[ goal],
		textCell[ "under the additional assumptions"], assumptionListCells[ {cond2, assm1, assm2, assm3, rest}, ",", "."]
	};
proofStepText[ CGB2tupleAdditionGoal, lang, {{g_}}, {{newG_}}, ___] :=
	{
		textCell[ "We can move addition into a tuple in formula ", formulaReference[ g],
		" and obtain"], goalCell[ newG], textCell["as our new goal."]
	};
proofStepText[ CGB2minGoalLeft|CGB2maxGoalRight, lang, {{g_}}, {{newG_}}, ___] :=
	{
		textCell[ "Rather than showing ", formulaReference[ g],
		" we may show"], goalCell[ newG], textCell["instead."]
	};
proofStepText[ CGB2minKBRight|CGB2maxKBLeft, lang, {{f_, _}}, {{_, newF_}}, ___] :=
	{
		textCell[ "Since we know ", formulaReference[ f],
		" we can add"], assumptionCell[ newF], textCell[ "to our assumptions."]
	};
subProofHeader[ CGB2minKBRight|CGB2maxKBLeft, lang, _, {{cond_, _}}, ___, {1}] :=
	{textCell[ "Before that, however, we first have to show"], goalCell[ cond, "."]};
subProofHeader[ CGB2minKBRight|CGB2maxKBLeft, lang, {{_, g_}}, {{cond_, newF_}}, ___, {2}] :=
	{textCell[ "Now we continue proving"], goalCell[ g], textCell["having"], assumptionCell[ cond],
	textCell[ "and"], assumptionCell[ newF], textCell["among our assumptions."]};
	
proofStepText[ CGB2existsGoal, lang, {{g_}}, {{newG:FML$[ _, False, __]}}, ___] :=
	{
		textCell[ "There is a contradiction in our existentially quantified goal ", formulaReference[ g],
		". Hence, we continue by deriving a contradiction in the knowledge base, i.e. we prove"],
		goalCell[ newG, "."]
	};
proofStepText[ CGB2existsGoal, lang, {{g_}}, {{newG:FML$[ _, True, __]}}, ___, "instantiation" -> {}, ___] :=
	{
		textCell[ "All conditions in our existentially quantified goal ", formulaReference[ g],
		" are vacuously satisfied, meaning that we have to prove"],
		goalCell[ newG, "."]
	};
proofStepText[ CGB2existsGoal, lang, {{g_}}, {{newG_}}, ___, "instantiation" -> inst_List, ___] :=
	{
		textCell[ "We instantiate our existentially quantified goal ", formulaReference[ g],
		" by ", inlineTheoremaExpressionSeq[ inst, lang], " and drop all conditions that are already among our assumptions. Therefore, we now have to prove"],
		goalCell[ newG, "."]
	};
proofStepText[ CGB2existsTupleOfGoal, lang, {{g_}}, {{new_}}, ___] :=
	{
		textCell[ "Since our current goal ", formulaReference[ g], " ranges over all elements of an abstraction tuple, we can simplify it to"],
		goalCell[ new, "."]
	};
proofStepText[ CGB2forallTupleOfKB, lang, {{k_, _}}, {{_, newk_}}, ___] :=
	{
		textCell[ "Since ", formulaReference[ k], " is known and the quantifier ranges over all elements of an abstraction tuple, we know "],
		assumptionCell[ newk], textCell[ "as well."]
	};
subProofHeader[ CGB2forallTupleOfKB, lang, _, {{newg_, _}}, ___, {1}] :=
	{textCell[ "Still, we first have to check"], goalCell[ newg, "."]};
subProofHeader[ CGB2forallTupleOfKB, lang, {{_, g_}}, {{_, newk_}}, ___, {2}] :=
	{
		textCell[ "Now we can continue proving our original goal ", formulaReference[ g], ", additionally knowing"],
		assumptionCell[ newk, "."]
	};
proofStepText[ CGB2existsAppendGoal, lang, {{g_}}, {{gen_}}, ___] :=
	{
		textCell[ "Since our current goal ", formulaReference[ g], " ranges over all elements of a tuple with some element appended, we can simplify it to"],
		goalCell[ gen, "."]
	};
proofStepText[ CGB2forallAppendKB, lang, {{k_, _}}, {{_, newk_}}, ___] :=
	{
		textCell[ "Since ", formulaReference[ k], " is known and the quantifier ranges over all elements of a tuple with some element appended, we know "],
		assumptionCell[ newk], textCell[ "as well."]
	};
subProofHeader[ CGB2forallAppendKB, lang, _, {{newg_, _}}, ___, {1}] :=
	{textCell[ "Still, we first have to check"], goalCell[ newg, "."]};
subProofHeader[ CGB2forallAppendKB, lang, {{_, g_}}, {{_, newk_}}, ___, {2}] :=
	{
		textCell[ "Now we can continue proving our original goal ", formulaReference[ g], ", additionally knowing"],
		assumptionCell[ newk, "."]
	};
    
proofStepText[ CGB2exchangeGoalKBInteractive, lang, {{g_FML$}}, {{neg_FML$}}, ___] := 
	{ textCell[ "We assume the negation of our current goal ", formulaReference[ g], ":"],
				assumptionCell[ neg], textCell[ "and try to derive a contradiction."]};
proofStepText[ CGB2exchangeGoalKBInteractive, lang, {{g_FML$, fml_FML$}}, {{neg_FML$, negFml_FML$}}, ___] := 
	{ textCell[ "We exchange our current goal ", formulaReference[ g], " and formula ", formulaReference[ fml], ", meaning that we prove"],
				goalCell[ negFml], textCell[ "under the additional assumption"], assumptionCell[ neg, "."]};

proofStepText[ CGB2expandNegationKB, lang, {{used_FML$}}, {{gen_FML$}}, ___] := 
	{ textCell[ "We simplify the negated formula ", formulaReference[ used], " in our current knowledge base to"],
				assumptionCell[ gen, "."]};
				
proofStepText[ CGB2expandNegationGoal, lang, {{used_FML$}}, {{gen_FML$}}, ___] := 
	{ textCell[ "We simplify the negated goal ", formulaReference[ used], " to"],
				goalCell[ gen, "."]};
    
proofStepText[ CGB2modusPonensTollens, lang, {{u_FML$}}, {{g_FML$}}, ___, "usedForms" -> refs_List, "action" -> "modusPonens", ___] := 
	Module[ {suffix = If[ Length[ refs] === 1, "", "s"]},
		If[ refs === {},
			{ textCell[ "From formula ", formulaReference[ u], " we can infer"],
						assumptionCell[ g], textCell[ "by modus ponens."]},
			{ textCell[ "From formula ", formulaReference[ u], ", together with formula" <> suffix <> " ", formulaReferenceSequence[ refs, lang], ", we can infer"],
						assumptionCell[ g], textCell[ "by modus ponens."]}
		]
	];
proofStepText[ CGB2modusPonensTollens, lang, {{u_FML$}}, {{g_FML$}}, ___, "usedForms" -> refs_List, "action" -> "simplification", ___] := 
	{ textCell[ "We can simplify formula ", formulaReference[ u], " to"],
				assumptionCell[ g], textCell[ "by dropping the premises in the implication that follow immediately from what we already know, in particular from ", formulaReferenceSequence[ refs, lang], "."]
	};
proofStepText[ CGB2modusPonensTollens, lang, {{u_FML$}}, {{g_FML$}}, ___, "usedForms" -> refs_List, "action" -> "modusTollens", ___] := 
	Module[ {suffix = If[ Length[ refs] === 1, "", "s"]},
		If[ refs === {},
			{ textCell[ "From formula ", formulaReference[ u], " we can infer"],
						assumptionCell[ g], textCell[ "by modus tollens."]},
			{ textCell[ "From formula ", formulaReference[ u], ", together with formula" <> suffix <> " ", formulaReferenceSequence[ refs, lang], ", we can infer"],
						assumptionCell[ g], textCell[ "by modus tollens."]}
		]
	];
	
proofStepText[ CGB2selectImplicationInteractive, lang, {{f_FML$, g_FML$}}, ___] := 
	{ textCell[ "We prove the premise of implication ", formulaReference[ f], " as well as our current goal ", formulaReference[ g], " assuming its conclusion:"]};
subProofHeader[ CGB2selectImplicationInteractive, lang, _, {{goal_FML$, assm_FML$}, _}, ___, {1}] := {textCell[ "First part: We prove"],
	goalCell[ goal], textCell[ "under the additional assumption"], assumptionCell[ assm, "."]
	};
subProofHeader[ CGB2selectImplicationInteractive, lang, {{_, g_FML$}}, {_, assm:{_, _}}, ___, {2}] := {textCell[ "Second part: We prove our current goal ", formulaReference[ g], " under the additional assumptions"],
	assumptionListCells[ assm, ",", "."]
	};
    
proofStepText[ CGB2generateRewriteRules, lang, {}, {}, pVal_] := {textCell[ "We generate rewrite rules and try to simplify the current goal and the current knowledge base by computation."]};
	
proofStepText[ CGB2elementarySubstitution|CGB2groundSubstitution, lang, u_List, g_List, ___, "usedForms" -> subst_List, ___] := 
	Module[ {stepText = {}, j, repl},
		If[ u[[1]] =!= g[[1]],
			repl = subst[[1]];
			stepText = { textCell[ "In order to prove ", formulaReference[ u[[1, 1]]], ", using formula ", 
				formulaReference[ repl[[1]]], ", we now have to show"],
				goalCell[ g[[1, 1]], "."]};
			If[ repl[[2]] =!= {},
				AppendTo[ stepText, textCell[ "The applicability of this substitution follows immediately from ", formulaReferenceSequence[ repl[[2]], lang], "."]]
			]
		];
		PrependTo[ stepText, textCell[ "We apply substitutions:"]];
		Do[
			repl = subst[[j]];
			stepText = Join[ stepText, 
				{textCell[ "From ", formulaReference[ u[[j, 1]]], " we know, by formula ", formulaReference[ repl[[1]]], ","], 
				assumptionCell[ g[[j, 1]]]}, If[ repl[[2]] === {}, {}, {textCell[ "The applicability of this substitution follows immediately from ",
				formulaReferenceSequence[ repl[[2]], lang], "."]}]],
			{j, 2, Length[ g]}
		];
		stepText
	];
	
proofStepText[ CGB2expandDef, lang, u_List, g_List, ___, "usedForms" -> defs_List, ___] := 
	Module[ {stepText = {}, j, repl},
		If[ u[[1]] =!= g[[1]],
			repl = defs[[1]];
			stepText = { textCell[ "In order to prove ", formulaReference[ u[[1, 1]]], ", using definition ", 
				formulaReference[ repl[[1]]], ", we now have to show"],
				goalCell[ g[[1, 1]], "."]};
			If[ repl[[2]] =!= {},
				AppendTo[ stepText, textCell[ "The applicability of this definition follows immediately from ", formulaReferenceSequence[ repl[[2]], lang], "."]]
			]
		];
		PrependTo[ stepText, textCell[ "We expand definitions:"]];
		Do[
			repl = defs[[j]];
			stepText = Join[ stepText, 
				{textCell[ "From ", formulaReference[ u[[j, 1]]], " we know, by definition ", formulaReference[ repl[[1]]], ","], 
				assumptionCell[ g[[j, 1]]]}, If[ repl[[2]] === {}, {}, {textCell[ "The applicability of this definition follows immediately from ",
				formulaReferenceSequence[ repl[[2]], lang], "."]}]],
			{j, 2, Length[ g]}
		];
		stepText
	];
	
] (* With *)

End[]