(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[   3073397,      84553]
NotebookOptionsPosition[   2487753,      68516]
NotebookOutlinePosition[   2867551,      78411]
CellTagsIndexPosition[   2823532,      77293]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell[TextData[{
 "Complexity Analysis of the Bivariate Buchberger Algorithm in ",
 StyleBox["Theorema",
  FontSlant->"Italic"]
}], "Title",
 CellChangeTimes->{{3.601197904263093*^9, 3.601197926158814*^9}, {
   3.601198383174067*^9, 3.601198391837915*^9}, {3.601199303731331*^9, 
   3.601199335935206*^9}, {3.603455772551382*^9, 3.603455772822835*^9}, {
   3.604127879726345*^9, 3.604127879893604*^9}, {3.615023795087659*^9, 
   3.615023812837507*^9}, 3.620471738943336*^9},
 TextAlignment->Center,
 CellID->1633829306],

Cell[TextData[{
 "\n",
 StyleBox["Alexander Maletzky\tBruno Buchberger",
  FontSize->24],
 "\n",
 StyleBox["\nDoktoratskolleg Computational Mathematics\nand\nResearch \
Institute for Symbolic Computation\n\nJohannes Kepler University Linz, \
Austria\n\n",
  FontSize->14],
 StyleBox["\[LeftAngleBracket]firstname.lastname\[RightAngleBracket]@risc.jku.\
at",
  FontFamily->"Courier 10 Pitch",
  FontSize->14]
}], "Subsubtitle",
 CellChangeTimes->{{3.603455603105163*^9, 3.603455700763731*^9}, {
  3.621222459543142*^9, 3.621222497470026*^9}},
 TextAlignment->Center,
 CellID->777484366],

Cell[TextData[StyleBox["Abstract",
 FontWeight->"Bold",
 FontSlant->"Italic"]], "Text",
 CellFrame->{{0, 0}, {0, 2}},
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.604134818862079*^9, 3.604134838112925*^9}},
 CellID->446898962],

Cell["\<\
The present report describes the formalization and formal verification of the \
complexity analysis of Buchberger\[CloseCurlyQuote]s algorithm for computing \
Groebner bases in the bivariate case, carried out in the Theorema system. \
Hence, it provides a major case study in computer-supported mathematical \
theory exploration, one of the long-term goals of the Theorema project.
The formalized theory was not developed from scratch, but rather is the one \
already described in papers by Buchberger dating back to various papers and \
reports around 1980.\
\>", "Text",
 CellFrame->{{0, 0}, {2, 0}},
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.604134831529103*^9, 3.604134838113017*^9}, {
  3.604134971610347*^9, 3.604135164612481*^9}, {3.60413524875372*^9, 
  3.604135271826778*^9}, {3.604137361832655*^9, 3.604137365735666*^9}, {
  3.604227639322711*^9, 3.604227657092874*^9}, {3.613885887371916*^9, 
  3.613885894153351*^9}, {3.615016577449774*^9, 3.615016587283442*^9}},
 CellID->644531705],

Cell[TextData[{
 StyleBox["Acknowledgement:",
  FontWeight->"Bold"],
 " The research was funded by the Austrian Science Fund (FWF): grant no. \
W1214-N15, project DK1"
}], "Text",
 CellChangeTimes->{{3.601198235144733*^9, 3.601198271539991*^9}, {
  3.603455710416767*^9, 3.603455714092092*^9}, {3.621061307819605*^9, 
  3.62106130904952*^9}, {3.621222513973757*^9, 3.621222515422334*^9}},
 CellID->95312927],

Cell[CellGroupData[{

Cell["How to Use this File", "Section",
 CellChangeTimes->{{3.621222571240571*^9, 3.621222574769414*^9}},
 CellID->730227987],

Cell[TextData[{
 "This file is not only a ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "-, but also a ",
 StyleBox["Theorema 2.0",
  FontSlant->"Italic"],
 " notebook. If you are only interested in viewing the notebook contents, you \
do not necessarily need ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ", but if you also want to view proof notebooks, carry out computations, or \
even do some proofs yourself, you first need to get ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " installed on your computer. It can be downloaded from the ",
 ButtonBox["official Theorema webpage",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["http://www.risc.jku.at/research/theorema/software/"], None},
  ButtonNote->"http://www.risc.jku.at/research/theorema/software/"],
 ", which also contains a detailed description of how to install it and how \
to use it properly. Note that ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " 8 or newer is needed for working with ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 "!"
}], "Text",
 CellChangeTimes->{{3.621222598834323*^9, 3.621222757352361*^9}, {
  3.621222802367572*^9, 3.621222802396062*^9}, {3.621222861912182*^9, 
  3.621222891336843*^9}, {3.62122292325233*^9, 3.621222923270889*^9}, {
  3.621223030875063*^9, 3.621223033064469*^9}, {3.621223068254281*^9, 
  3.621223085190089*^9}, {3.62122402401262*^9, 3.621224024605097*^9}, {
  3.62156853161953*^9, 3.621568532167565*^9}},
 CellID->1185065486],

Cell[TextData[{
 "This file comes with an additional folder (with the same name as the file; \
only needed when using ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 "). Just store the folder in the same directory as the file (or at least \
link it to that directory), and make sure that file and folder have the same \
name \[Dash] The rest is taken care of by ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellChangeTimes->{{3.621223194171996*^9, 3.621223309739046*^9}, {
  3.621223849197594*^9, 3.621223849382446*^9}},
 CellID->1524961043],

Cell[TextData[{
 "Assuming you opened this file with ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ", the following is a list of things you can do now:"
}], "Text",
 CellChangeTimes->{{3.621223052327041*^9, 3.621223063843111*^9}, {
  3.621223098613001*^9, 3.621223119803325*^9}},
 CellID->1630144601],

Cell[CellGroupData[{

Cell["\<\
View the notebook contents, i.e. informal explanatory text, formal \
definitions, lemmas, and theorems.\
\>", "Item",
 CellChangeTimes->{{3.621223129698906*^9, 3.621223168138499*^9}, {
  3.621224235247897*^9, 3.621224236328622*^9}},
 CellID->2091470436],

Cell[TextData[{
 "View automatically generated proof notebooks. Just click on any of the \
\[OpenCurlyDoubleQuote]",
 StyleBox["Show proof",
  FontSlant->"Italic"],
 "\[CloseCurlyDoubleQuote] - buttons."
}], "Item",
 CellChangeTimes->{{3.621223129698906*^9, 3.621223183767608*^9}, {
  3.621223334053509*^9, 3.621223398683729*^9}},
 CellID->73938940],

Cell[TextData[{
 "Carry out computations. For detailed information on how this works, we \
refer to the online documentation of ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " (see webpage). Only note that at the moment it is not possible to \
automatically restore the original computation-settings (knowledge, \
built-ins) for the first time, by clicking \[OpenCurlyDoubleQuote]OK\
\[CloseCurlyDoubleQuote] in the \[OpenCurlyDoubleQuote]Restore Environment\
\[CloseCurlyDoubleQuote]-tabs. This needs to be done manually."
}], "Item",
 CellChangeTimes->{{3.621223129698906*^9, 3.621223183767608*^9}, {
  3.621223334053509*^9, 3.621223555242498*^9}},
 CellID->1984099723],

Cell[TextData[{
 "Carry out proofs. Again, for a detailed information on this we refer to the \
online documentation, and just as with computations it is not possible to \
automatically restore the original proof-settings. However, unfortunately the \
special prover that was developed for verifying the theory formalized in this \
notebook is not yet integrated into the system, and therefore most formulas \
will ",
 StyleBox["not",
  FontSlant->"Italic"],
 " be provable at the moment (with the default \
\[OpenCurlyDoubleQuote]Theorema Basic Language \
Prover\[CloseCurlyDoubleQuote])."
}], "Item",
 CellChangeTimes->{{3.621223129698906*^9, 3.621223183767608*^9}, {
  3.621223334053509*^9, 3.621223709362123*^9}, {3.621223759959319*^9, 
  3.621223771012647*^9}, {3.62122390936831*^9, 3.621223912010409*^9}},
 CellID->1184009133]
}, Open  ]],

Cell[TextData[{
 "In case there are any problems in using and working with this notebook as \
described above, please contact ",
 ButtonBox["alexander.maletzky@risc.jku.at",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["mailto:alexander.maletzky@risc.jku.at"], None},
  ButtonNote->"mailto:alexander.maletzky@risc.jku.at"],
 "."
}], "Text",
 CellChangeTimes->{{3.621224072098754*^9, 3.621224165224739*^9}},
 CellID->2119521717]
}, Open  ]],

Cell[CellGroupData[{

Cell["Introduction", "Section",
 CellChangeTimes->{{3.603455826785983*^9, 3.603455828777928*^9}},
 CellID->590276341],

Cell[TextData[{
 "In 1965, Bruno Buchberger (",
 ButtonBox["[Buchberger1965]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1965"],
 ") introduced his famous algorithm for computing Groebner bases of \
polynomial ideals over fields. Since then lots of improvements have been \
achieved in terms of efficiency and applicability, and although the algorithm \
is known to be intrinsically complex (double-exponential time complexity in \
the number of indeterminates, c.f. ",
 ButtonBox["[Mayr-Meyer1982]",
  BaseStyle->"Hyperlink",
  ButtonData->"Mayr-Meyer1982"],
 "), there has ever since been an interest in deriving tight bounds for the \
number of steps in some special cases (Buchberger\[CloseCurlyQuote]s 1965 PhD \
thesis already contained a rough analysis).\nAround 1980, in various papers \
and reports, Buchberger indeed managed to derive such bounds in the ",
 StyleBox["bivariate",
  FontSlant->"Italic"],
 " case, i.e. in the case of only two indeterminates (",
 ButtonBox["[Buchberger1979]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979"],
 ", ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 ", ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 "). He proceeded by first giving a bound on the degrees of the polynomials \
in the Groebner basis resulting from his algorithm if a ",
 StyleBox["graded",
  FontSlant->"Italic"],
 " order relation on the set of monomials is used. Afterwards, knowing such a \
bound, it is not so hard any more to obtain also a bound for the number of \
steps in the algorithm.\nThis is now exactly what the present report aims at: \
First ",
 StyleBox["formalizing",
  FontSlant->"Italic"],
 " the complexity analysis in the computer system ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " and then also ",
 StyleBox["verifying",
  FontSlant->"Italic"],
 " it, making use of the automated reasoning capabilities of ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ". The overall strategy that is pursued for achieving this aim is the same \
as in the literature. However, a couple of simplifications could be applied \
in order to make some proofs shorter. But, apart from that, no essentially \
new insights and results could be expected to be obtained."
}], "Text",
 CellChangeTimes->{{3.604127992036448*^9, 3.604128128883038*^9}, {
  3.604130454880871*^9, 3.60413049275107*^9}, {3.604130529509482*^9, 
  3.604130733612069*^9}, {3.604130785360899*^9, 3.60413082695703*^9}, {
  3.604130863913688*^9, 3.604131134098975*^9}, {3.604131240953703*^9, 
  3.604131241736775*^9}, {3.604131279861179*^9, 3.604131318026426*^9}, {
  3.604134245854246*^9, 3.604134309313644*^9}, {3.604134389998632*^9, 
  3.604134413601263*^9}, {3.615017384770809*^9, 3.615017432988751*^9}, {
  3.615017472595033*^9, 3.615017473366621*^9}, {3.615017519525649*^9, 
  3.615017544016998*^9}, {3.619842353130185*^9, 3.619842415394799*^9}},
 CellID->831646513],

Cell[TextData[{
 "The ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " system (",
 ButtonBox["[Theorema]",
  BaseStyle->"Hyperlink",
  ButtonData->"Theorema"],
 ", ",
 ButtonBox["[Buchberger2006]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger2006"],
 ") is a system for supporting automated theory exploration, conceived and \
developed by Bruno Buchberger in the mid-nineties and now developed at RISC. \
Since there exist at least two different versions of ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " at the moment and this exploration was carried out in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " 2.0, the name ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " in the sequel always refers to version 2.0 (unless explicitly stated \
otherwise)."
}], "Text",
 CellChangeTimes->{{3.604131329221801*^9, 3.604131361342357*^9}, {
   3.604131395301233*^9, 3.604131430267208*^9}, {3.604131463088212*^9, 
   3.604131486177091*^9}, {3.604131535500493*^9, 3.604131600774751*^9}, 
   3.604131642150648*^9, {3.604132131430087*^9, 3.604132134117458*^9}, {
   3.604132169114063*^9, 3.604132169119728*^9}, {3.60413295551942*^9, 
   3.604133063802392*^9}, {3.604133105325056*^9, 3.604133126303266*^9}, {
   3.615017587877882*^9, 3.615017726854432*^9}},
 CellID->1530781958],

Cell[TextData[{
 "The structure of this report is as follows:\nSection ",
 ButtonBox["Formalizing Mathematics",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::FormalMathematics"],
 " expresses some general views on the way how mathematics could/should be \
formalized in computer systems such as ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ", and also describes the overall strategy that was pursued in the \
particular case of the formalization of the complexity analysis of Buchberger\
\[CloseCurlyQuote]s algorithm. Section ",
 ButtonBox["Problem Setting",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::ProblemSetting"],
 " states Buchberger\[CloseCurlyQuote]s algorithm with chain criterion and \
repeats the original ideas of the complexity analysis. Section ",
 ButtonBox["Basic Notions",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::BasicNotions"],
 " introduces all the basic notions (i.e. functions and predicates) that \
serve as the foundations for defining the more important notions directly \
related to the complexity analysis. All this, of course, is done in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " syntax, such that it is possible to already use these notions in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " proofs and computations. In the next section, ",
 ButtonBox["Lemmata on Basic Notions",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataBasic"],
 ", some simple, \[OpenCurlyDoubleQuote]rewrite-kind\[CloseCurlyDoubleQuote] \
properties of the basic notions just introduced are proved. For an informal \
description of the term \[OpenCurlyDoubleQuote]rewrite-kind\
\[CloseCurlyDoubleQuote] please see section ",
 ButtonBox["Formalizing Mathematics",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::FormalMathematics"],
 ". Section ",
 ButtonBox["Specific Notions for Complexity Analysis",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::SpecificNotions"],
 " eventually introduces all the relevant notions needed for carrying out the \
complexity analysis, and section ",
 ButtonBox["Computations",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Computations"],
 " contains some computations involving these new notions. The computations \
are just included for illustrating ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 "\[CloseCurlyQuote]s capabilities of proving, solving, and ",
 StyleBox["computing",
  FontSlant->"Italic"],
 " within one coherent framework, but they are of no relevance for the formal \
verification of the theory and hence might be skipped if one is only \
interested in that. In section ",
 ButtonBox["Lemmata on Specific Notions",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataSpecific"],
 " again some basic properties and \[OpenCurlyDoubleQuote]rewrite-kind\
\[CloseCurlyDoubleQuote] lemmata are stated and proved, but this time about \
the new specific notions needed for the complexity analysis.\nThe most \
important section of this report, or at least one of them, is section ",
 ButtonBox["Theorems",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Theorems"],
 ": This section contains all the non-trivial results that are later needed \
for proving the main theorems the present report is all about (in sections ",
 ButtonBox["M[A]+W[A] Does Not Increase",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Invariant"],
 ", ",
 ButtonBox["M[A]+W[A]\[LessEqual]2maxdeg[A]",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Bound"],
 " and ",
 ButtonBox["maxdeg[G]\[LessEqual]2maxdeg[F]",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::MainResult"],
 "); However, most of these main theorems are in fact just corollaries of \
theorems in section ",
 ButtonBox["Theorems",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Theorems"],
 ".\nSection ",
 ButtonBox["Prover",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Prover"],
 " briefly describes the special prover that was implemented for the purpose \
of verifying the theory, section ",
 ButtonBox["Conclusions",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Conclusions"],
 " contains a short summary and outlook, and section ",
 ButtonBox["References",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::References"],
 " contains the bibliography."
}], "Text",
 CellChangeTimes->{{3.603520999771938*^9, 3.603521008315002*^9}, {
   3.603529287550589*^9, 3.603529288023959*^9}, {3.604061879942833*^9, 
   3.604061889348177*^9}, {3.604127956615023*^9, 3.604127968605057*^9}, {
   3.604131140220954*^9, 3.604131140368254*^9}, {3.604131492993467*^9, 
   3.604131494027521*^9}, 3.604131616664446*^9, {3.604132193796439*^9, 
   3.604132253618449*^9}, {3.604132287156878*^9, 3.604132607634394*^9}, {
   3.60413313452271*^9, 3.604133512473703*^9}, {3.615017750262473*^9, 
   3.615017790892693*^9}, {3.615017836117099*^9, 3.61501798185263*^9}, {
   3.615018029040384*^9, 3.615018029060772*^9}, {3.615018083128519*^9, 
   3.615018084079678*^9}, {3.615018115102977*^9, 3.615018372660049*^9}, {
   3.615018421295702*^9, 3.615018564458012*^9}, {3.61501862628675*^9, 
   3.615018759198139*^9}, {3.615018927282597*^9, 3.61501907182816*^9}, {
   3.615019117925114*^9, 3.615019388721506*^9}, {3.615019421874639*^9, 
   3.615019784166373*^9}},
 CellID->1271320345],

Cell[CellGroupData[{

Cell["Formula Labelling Conventions", "Subsection",
 CellChangeTimes->{{3.615518862748078*^9, 3.615518870134691*^9}, 
   3.615518911088446*^9},
 CellID->448627247],

Cell[TextData[{
 "Each formula in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " comes with a unique label, which can be found on the right margin next to \
the formula. Labels in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " have nothing to do with the semantics of formulas, but they are solely \
used for referencing them, e.g. in proofs."
}], "Text",
 CellChangeTimes->{{3.61551892324346*^9, 3.615519050038628*^9}, {
  3.615519116214226*^9, 3.615519117303578*^9}},
 CellID->728060281],

Cell[TextData[{
 "Usually, formulas are labelled by numbers starting from 1. However, many \
formulas in this document have \[OpenCurlyDoubleQuote]speaking\
\[CloseCurlyDoubleQuote] labels that indicate, to some extent, the statement \
of the corresponding formula. For instance, consider formula ",
 ButtonBox["(min\[LessEqual])",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:48233088"],
 " in section ",
 ButtonBox["Lemmata on Basic Notions / min, max",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Lemmataminmax"],
 ": The label \
\[OpenCurlyDoubleQuote](min\[LessEqual])\[CloseCurlyDoubleQuote] clearly \
indicates that the formula is concerned with relating the min-function and \
the binary relation \[LessEqual]."
}], "Text",
 CellChangeTimes->{{3.615519072135783*^9, 3.61551928671752*^9}, {
  3.615519316801146*^9, 3.615519427493998*^9}, {3.615519463967765*^9, 
  3.615519464893306*^9}, {3.615519722664515*^9, 3.615519752290249*^9}},
 CellID->1524435427],

Cell[TextData[{
 "In addition, some formulas have \[OpenCurlyDoubleQuote]speaking\
\[CloseCurlyDoubleQuote] labels followed by a number, i.e. something like \
\[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"name", " ", "n"}], ")"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote]., where the number ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " can either be arabic or roman. In the first case, if ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " is an arabic number, the meaning is that formula \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"name", " ", "n"}], ")"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] is an ausiliary result needed for proving the more \
\[OpenCurlyDoubleQuote]interesting\[CloseCurlyDoubleQuote] formula \
\[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", "name", ")"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote]. In the second case, if ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " is a roman number, the meaning is that formula \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"name", " ", "n"}], ")"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] is an immediate consequence of the more \
\[OpenCurlyDoubleQuote]interesting\[CloseCurlyDoubleQuote] formula \
\[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", "name", ")"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote]. For instance, consider formulas ",
 ButtonBox["(M\:293a 1)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1399613937"],
 ", ",
 ButtonBox["(M\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1628665526"],
 " and ",
 ButtonBox["(M\:293a I)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1351181837"],
 ", which can all be found in section ",
 ButtonBox["Lemmata on Specific Notions / General Case / M",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataSpecificGeneralM"],
 ": ",
 ButtonBox["(M\:293a 1)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1399613937"],
 " is an auxiliary result needed for proving ",
 ButtonBox["(M\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1628665526"],
 ", and ",
 ButtonBox["(M\:293a I)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1351181837"],
 " is a corollary of ",
 ButtonBox["(M\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1628665526"],
 ", which is the most \[OpenCurlyDoubleQuote]interesting\
\[CloseCurlyDoubleQuote] formula among the three."
}], "Text",
 CellChangeTimes->{
  3.615519466515061*^9, {3.61551950174136*^9, 3.615519719087575*^9}, {
   3.615519760529029*^9, 3.615519763978734*^9}, {3.615519831186984*^9, 
   3.615520082556594*^9}},
 CellID->380145561],

Cell["\<\
All the remaining formulas, which are either of little importance/interest or \
for which no good and concise \[OpenCurlyDoubleQuote]speaking\
\[CloseCurlyDoubleQuote] label could be found, are labelled by numbers from 1 \
to 100 in increasing order.\
\>", "Text",
 CellChangeTimes->{{3.615520097454044*^9, 3.615520518094843*^9}, {
  3.615520562668316*^9, 3.61552072640022*^9}, {3.615787287088304*^9, 
  3.615787303409327*^9}},
 CellID->83311740]
}, Closed]],

Cell[CellGroupData[{

Cell["Summary of the Formalization", "Subsection",
 CellChangeTimes->{{3.615782344833342*^9, 3.615782350023059*^9}},
 CellID->1413123984],

Cell["The formalization consists in total of", "Text",
 CellChangeTimes->{{3.615782355555688*^9, 3.615782380380535*^9}},
 CellID->1937745430],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"f", "=", "FormulaCount"}]], "Input",
 CellChangeTimes->{{3.615781109873648*^9, 3.615781117049998*^9}, {
   3.615781156492828*^9, 3.615781162762444*^9}, {3.615781215538244*^9, 
   3.615781228458643*^9}, {3.615781302628788*^9, 3.615781343486476*^9}, {
   3.615781375995301*^9, 3.615781376472145*^9}, {3.615781434919674*^9, 
   3.615781435827705*^9}, {3.615781523665247*^9, 3.615781587242445*^9}, {
   3.615781843387318*^9, 3.615781867467427*^9}, {3.615781996381721*^9, 
   3.615782111923597*^9}, {3.615782183570863*^9, 3.615782184482756*^9}, {
   3.615782394579885*^9, 3.615782395429224*^9}, {3.615782450391953*^9, 
   3.615782451511121*^9}, {3.61578362594566*^9, 3.615783628815537*^9}, 
   3.615783660958988*^9, {3.615784022805615*^9, 3.615784033465119*^9}},
 CellID->1715249823],

Cell[BoxData["292"], "Output",
 CellChangeTimes->{{3.615781853243182*^9, 3.615781869992593*^9}, {
   3.615782039787351*^9, 3.615782113316703*^9}, 3.615782185944128*^9, 
   3.615782399679261*^9, 3.615782453241866*^9, 3.615783154110421*^9, 
   3.615783630786662*^9, 3.615784035028261*^9},
 CellID->1104335559]
}, Open  ]],

Cell["formulas (in definitions, lemmas, theorems), where in total", "Text",
 CellChangeTimes->{{3.615782374292534*^9, 3.615782375425417*^9}, {
  3.615782405725677*^9, 3.615782430296701*^9}, {3.615783891842891*^9, 
  3.615783895076734*^9}},
 CellID->2105302388],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"p", "=", "ProofCount"}]], "Input",
 CellChangeTimes->{{3.615782461779112*^9, 3.615782462227479*^9}, {
  3.615783223931775*^9, 3.615783234473727*^9}, {3.615783306342968*^9, 
  3.615783313562054*^9}, {3.615783383334351*^9, 3.615783441351515*^9}, {
  3.615783560117806*^9, 3.615783716400274*^9}, {3.615783761363383*^9, 
  3.615783828917578*^9}, {3.615784090635146*^9, 3.615784116127954*^9}},
 CellID->503539654],

Cell[BoxData["230"], "Output",
 CellChangeTimes->{
  3.615782463215173*^9, 3.615783234996918*^9, {3.615783401193785*^9, 
   3.615783442061923*^9}, {3.6157836453343*^9, 3.615783664353944*^9}, {
   3.61578370729054*^9, 3.615783719539997*^9}, 3.615783773268969*^9, 
   3.615783829707437*^9, 3.615784126343007*^9, 3.615784251731622*^9, {
   3.615784390351136*^9, 3.615784416436465*^9}, {3.61578462569611*^9, 
   3.615784639081554*^9}, {3.615784695715148*^9, 3.615784714670206*^9}, 
   3.61578482594376*^9},
 CellID->1004345167]
}, Open  ]],

Cell["of them have been proved. This means that", "Text",
 CellChangeTimes->{{3.61578243792795*^9, 3.615782442304937*^9}, {
  3.61578389957577*^9, 3.615783903525194*^9}},
 CellID->675589372],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"f", "-", "p"}]], "Input",
 CellChangeTimes->{{3.615782458810436*^9, 3.615782460103645*^9}},
 CellID->834113638],

Cell[BoxData["62"], "Output",
 CellChangeTimes->{3.615782465802219*^9, 3.615784645041352*^9},
 CellID->1599837690]
}, Open  ]],

Cell["formulas are given without proof (i.e. in definitions).", "Text",
 CellChangeTimes->{{3.615782470010717*^9, 3.615782486226281*^9}, {
  3.615783909095957*^9, 3.615783919594819*^9}},
 CellID->352834493],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " Functions for Notebook Data"
}], "Subsubsection",
 CellChangeTimes->{{3.61578397728518*^9, 3.615783983333089*^9}, {
  3.615784149019362*^9, 3.615784157794169*^9}},
 CellID->872582327],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "FormulaCount", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"FormulaCount", ":=", 
  RowBox[{"Length", "[", 
   RowBox[{"Cases", "[", 
    RowBox[{
     RowBox[{"NotebookGet", "[", "]"}], ",", 
     RowBox[{
      RowBox[{"Cell", "[", 
       RowBox[{"___", ",", 
        RowBox[{"CellTags", "\[Rule]", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"_", "?", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"StringMatchQ", "[", 
               RowBox[{"#", ",", 
                RowBox[{"\<\"ID:\"\>", "~~", "__"}]}], "]"}], "&"}], ")"}]}], 
           ",", "_", ",", "l_"}], "}"}]}], ",", "___"}], "]"}], 
      "\[RuleDelayed]", "l"}], ",", "Infinity"}], "]"}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.615783992664973*^9, 3.615784025501526*^9}},
 CellID->458961238],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "ProofCount", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"ProofCount", ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"nm", "=", 
      RowBox[{"CurrentValue", "[", "\<\"NotebookFullFileName\"\>", "]"}]}], 
     "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{"Length", "[", 
     RowBox[{"DeleteDuplicates", "[", 
      RowBox[{"Map", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"getID", "[", 
           RowBox[{"FileBaseName", "[", "#", "]"}], "]"}], ")"}], "&"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"FileNames", "[", 
         RowBox[{
          RowBox[{"\<\"p\"\>", "~~", "__", "~~", "\<\"po.m\"\>"}], ",", 
          RowBox[{"{", 
           RowBox[{"FileNameJoin", "[", 
            RowBox[{"{", 
             RowBox[{
              RowBox[{"DirectoryName", "[", "nm", "]"}], ",", 
              RowBox[{"FileBaseName", "[", "nm", "]"}]}], "}"}], "]"}], 
           "}"}]}], "]"}]}], "]"}], "]"}], "]"}]}], "\[IndentingNewLine]", 
   "]"}]}]}], "Input",
 CellChangeTimes->{{3.615784069317074*^9, 3.615784120594578*^9}, {
  3.615784217207121*^9, 3.615784246410157*^9}, {3.615784300854988*^9, 
  3.615784314215343*^9}, {3.615784373801131*^9, 3.615784382938681*^9}, {
  3.615784611330179*^9, 3.615784634480102*^9}, {3.615784688274191*^9, 
  3.615784710056905*^9}, {3.615784807191823*^9, 3.615784815356164*^9}},
 CellID->1459432484],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "getID", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"getID", "[", "s_String", "]"}], ":=", 
  RowBox[{"StringTake", "[", 
   RowBox[{
    RowBox[{"StringDrop", "[", 
     RowBox[{"s", ",", "1"}], "]"}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"StringPosition", "[", 
       RowBox[{"s", ",", "\<\"-\"\>", ",", "1"}], "]"}], "[", 
      RowBox[{"[", 
       RowBox[{"1", ",", "1"}], "]"}], "]"}], "-", "2"}]}], "]"}]}]}], "Input",\

 CellChangeTimes->{{3.615784333569135*^9, 3.615784412097048*^9}, {
  3.615784457621711*^9, 3.61578446179638*^9}, {3.615784494365744*^9, 
  3.615784519379766*^9}, {3.615784575958083*^9, 3.615784591030394*^9}},
 CellID->251027589]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Formalizing Mathematics", "Section",
 CellChangeTimes->{{3.607316481526497*^9, 3.60731649576992*^9}, {
  3.607316943835559*^9, 3.60731694406101*^9}, {3.615017918422279*^9, 
  3.615017920901125*^9}},
 CellTags->"sect::FormalMathematics",
 CellID->1882786356],

Cell[TextData[{
 "In our perception ",
 ButtonBox["[Buchberger2014]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger2014"],
 ", systematic ",
 StyleBox["formal",
  FontSlant->"Italic"],
 " development of mathematical theories ideally proceeds in the following \
way, following the concept of the ",
 StyleBox["Theory Exploration Cycle",
  FontSlant->"Italic"],
 ":"
}], "Text",
 CellChangeTimes->{{3.607316605271416*^9, 3.607316688446549*^9}, {
  3.607316845741827*^9, 3.607316846788432*^9}, {3.607316898098146*^9, 
  3.607316899391269*^9}, {3.61527975248268*^9, 3.61527976644289*^9}},
 CellID->723171646],

Cell[TextData[{
 "Start by introducing new notions, define them in terms of known ones.\nIn \
our case, this is done in section ",
 ButtonBox["Basic Notions",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::BasicNotions"],
 "."
}], "Text",
 CellDingbat->"1.",
 CellChangeTimes->{{3.607316701812803*^9, 3.607316795458193*^9}, {
  3.607320607269223*^9, 3.607320613740559*^9}, {3.615279490158959*^9, 
  3.615279512115713*^9}, {3.615283709843012*^9, 3.61528371253633*^9}},
 CellID->228411370],

Cell[TextData[{
 "Prove simple lemmata about these notions using an existing prover. This \
prover should be general in the sense that it only uses knowledge about known \
notions, i.e. notions the new notions are defined in terms of.\nIn the \
present theory exploration, section ",
 ButtonBox["Lemmata on Basic Notions",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataBasic"],
 " contains exactly this kind of lemmata. The prover we used for proving \
them, however, is not the very general \
\[OpenCurlyDoubleQuote]BasicTheoremaLanguageProver\[CloseCurlyDoubleQuote], \
because already at this point we needed some fundamental properties of tuples \
(in particular, the tuple quantifier). Hence, already at this stage we used a \
small fragment of our new \[OpenCurlyDoubleQuote]ComplexityProver\
\[CloseCurlyDoubleQuote]."
}], "Text",
 CellDingbat->"2.",
 CellChangeTimes->{{3.607316701812803*^9, 3.607316828889119*^9}, {
  3.607316862155348*^9, 3.607316878403402*^9}, {3.607316910329983*^9, 
  3.607316931796148*^9}, {3.607316965489994*^9, 3.6073171198334*^9}, {
  3.607317198427243*^9, 3.607317356405953*^9}, {3.607317425886106*^9, 
  3.607317429374452*^9}, {3.60732062042978*^9, 3.607320620763526*^9}, {
  3.615279539703507*^9, 3.615279702830462*^9}, {3.615279796037426*^9, 
  3.61527979604505*^9}, {3.615282414850587*^9, 3.615282415191673*^9}, {
  3.615283717694683*^9, 3.615283718175367*^9}},
 CellID->1933341333],

Cell[TextData[{
 "Ideally, the lemmata that are proved in the second step have \
\[OpenCurlyDoubleQuote]rewrite\[CloseCurlyDoubleQuote]-character: They are \
(universally quantified, conditional) equalities/equivalences, giving rise to \
rewrite rules that can later be used for proving more complicated theorems \
(as well as for computing).\nOur theory for the most part is concerned with \
",
 StyleBox["inequalities",
  FontSlant->"Italic"],
 " rather than equalities, so unfortunately there are not many \
\[OpenCurlyDoubleQuote]rewrite-kind\[CloseCurlyDoubleQuote] formulas in it \
(unless one considers inequalities, like ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[LessEqual]", "b"}], TraditionalForm]]],
 ", as \[OpenCurlyDoubleQuote]rewrite-kind\[CloseCurlyDoubleQuote] formulas \
as well; After all, ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 " is still transitive)."
}], "Text",
 CellChangeTimes->{{3.607316701812803*^9, 3.607316828889119*^9}, {
   3.607316862155348*^9, 3.607316878403402*^9}, {3.607316910329983*^9, 
   3.607316931796148*^9}, {3.607316965489994*^9, 3.6073171198334*^9}, {
   3.607317198427243*^9, 3.607317550581166*^9}, {3.607317582302762*^9, 
   3.607317663001215*^9}, {3.607317729460787*^9, 3.607317833896686*^9}, {
   3.607317865725069*^9, 3.607317882715474*^9}, {3.607318099701972*^9, 
   3.607318100072323*^9}, 3.607319310372986*^9, {3.607319344654258*^9, 
   3.607319346693839*^9}, {3.607320628673293*^9, 3.607320628887746*^9}, {
   3.615279817485518*^9, 3.615279853633941*^9}, {3.615279957824947*^9, 
   3.615280031715256*^9}, {3.615283767432452*^9, 3.615283767728716*^9}},
 CellID->547946198],

Cell[TextData[{
 "As soon as a reasonable collection of lemmata is at the user\
\[CloseCurlyQuote]s disposal, move on to more complicated theorems. If the \
lemmata are of the \
\[OpenCurlyDoubleQuote]rewrite-kind\[CloseCurlyDoubleQuote], most probably no \
special prover is needed yet, as long as inference rules capable of dealing \
with equalities/equivalences in an efficient way are available.\nIn our case, \
this is taken care of by the inference rules of our ComplexityProver that are \
described in section ",
 ButtonBox["Prover / Rewriting",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::ProverRewriting"],
 "."
}], "Text",
 CellDingbat->"3.",
 CellChangeTimes->{{3.607316701812803*^9, 3.607316828889119*^9}, {
  3.607316862155348*^9, 3.607316878403402*^9}, {3.607316910329983*^9, 
  3.607316931796148*^9}, {3.607316965489994*^9, 3.6073171198334*^9}, {
  3.607317198427243*^9, 3.607317550581166*^9}, {3.607317582302762*^9, 
  3.607317663001215*^9}, {3.607317729460787*^9, 3.607317833896686*^9}, {
  3.607317865725069*^9, 3.607317923752411*^9}, {3.607317976465846*^9, 
  3.607318054376852*^9}, {3.607318109079034*^9, 3.607318239197025*^9}, {
  3.607319188187269*^9, 3.607319220272683*^9}, {3.607319272764298*^9, 
  3.607319272989687*^9}, {3.607320638708316*^9, 3.607320639335947*^9}, {
  3.615280058905191*^9, 3.615280095512163*^9}, {3.615280127266684*^9, 
  3.615280127272287*^9}, {3.615282446238462*^9, 3.615282452816497*^9}, {
  3.615283773787768*^9, 3.61528377412079*^9}},
 CellID->1462515714],

Cell[TextData[{
 "As a last step, ",
 StyleBox["lift",
  FontSlant->"Italic"],
 " formulas to the level of inference rules: Although in principle reasoning \
can always be done ",
 StyleBox["from first principles, by first principles",
  FontSlant->"Italic"],
 ", mathematical experience shows that in many theory explorations proving by \
first principles all the time leads to unnatural, unnecessarily long, \
incomprehensible (for humans) proofs. Therefore, one should strive to turn \
parts of the theory, i.e. formulas, into inference rules on the meta level. \
These inference rules could have attached all the meta-information human \
mathematicians always have in mind when working in the theory, e.g. always \
instantiating particular quantifiers by certain terms, considering nested \
occurrences of associative-commutative operators in \
\[OpenCurlyDoubleQuote]flattened\[CloseCurlyDoubleQuote] form with \
syntactically-sorted arguments, or checking the conditions of some class of \
formulas rather a-posteriori than a-priori (\[Rule] see subsection ",
 ButtonBox["Conditional Equalities/Equivalences",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Conditional"],
 "). Also, sometimes one wants to lift functions to quantifiers (e.g. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"min", "[", "A", "]"}], TraditionalForm]]],
 " \[Rule] ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["min", "x"], " ", "t"}], TraditionalForm]]],
 "), which can reasonably only be done on the meta level, i.e. on the level \
of inference rules.\nSummarizing, as a last step design and implement a new \
",
 StyleBox["special prover",
  FontSlant->"Italic"],
 " tailored for reasoning ",
 StyleBox["with",
  FontSlant->"Italic"],
 " the notions of the theory, ",
 StyleBox["in",
  FontSlant->"Italic"],
 " the theory.\nThe prover we implemented for carrying out the complexity \
analysis is called \[OpenCurlyDoubleQuote]ComplexityProver\
\[CloseCurlyDoubleQuote] and explained in detail in section ",
 ButtonBox["Prover",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Prover"],
 ". In particular, the inference rules described in subsections ",
 ButtonBox["Tuples",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::ProverTuples"],
 ", ",
 ButtonBox["Addition in TOMs",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::ProverAddition"],
 ", ",
 ButtonBox["Order Relation in TOMs,",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::ProverOrdering"],
 " ",
 ButtonBox["Minimum, Maximum",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::ProverMinMax"],
 " and ",
 ButtonBox["Logical Quantifiers",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::ProverQuantifiers"],
 " serve precisely the aforementioned purpose of ",
 StyleBox["natural",
  FontSlant->"Italic"],
 " and ",
 StyleBox["efficient",
  FontSlant->"Italic"],
 " proving with notions in the complexity analysis of Buchberger\
\[CloseCurlyQuote]s algorithm."
}], "Text",
 CellDingbat->"4.",
 CellChangeTimes->{{3.607316701812803*^9, 3.607316828889119*^9}, {
   3.607316862155348*^9, 3.607316878403402*^9}, {3.607316910329983*^9, 
   3.607316931796148*^9}, {3.607316965489994*^9, 3.6073171198334*^9}, {
   3.607317198427243*^9, 3.607317550581166*^9}, {3.607317582302762*^9, 
   3.607317663001215*^9}, {3.607317729460787*^9, 3.607317833896686*^9}, {
   3.607317865725069*^9, 3.607317923752411*^9}, {3.607317976465846*^9, 
   3.607318054376852*^9}, {3.607318109079034*^9, 3.607318239197025*^9}, {
   3.607319188187269*^9, 3.607319220259437*^9}, {3.607319351355622*^9, 
   3.607319489311834*^9}, {3.607319527133064*^9, 3.607319540381023*^9}, {
   3.607319609187945*^9, 3.60731970344027*^9}, {3.607319783206722*^9, 
   3.607319810130927*^9}, {3.60731985805721*^9, 3.607319898653124*^9}, {
   3.607319966742377*^9, 3.607320017419677*^9}, {3.607320128720407*^9, 
   3.607320134309279*^9}, {3.607320166411299*^9, 3.607320230471255*^9}, {
   3.607320291096372*^9, 3.607320436887998*^9}, {3.6073204923639*^9, 
   3.607320501421818*^9}, {3.607320537023314*^9, 3.60732058965077*^9}, {
   3.607320643548225*^9, 3.607320789215968*^9}, {3.607320825372732*^9, 
   3.607320825815669*^9}, {3.607321012295937*^9, 3.607321013088534*^9}, {
   3.614937376324339*^9, 3.614937380158065*^9}, {3.615280170361881*^9, 
   3.615280221671077*^9}, {3.615280273961692*^9, 3.615280318609252*^9}, {
   3.61528060662453*^9, 3.615280868315487*^9}, {3.615280898331114*^9, 
   3.615280901987673*^9}, {3.615281098622436*^9, 3.615281158313265*^9}, 
   3.615281188918372*^9, {3.615281249462812*^9, 3.615281717349757*^9}, {
   3.615281771744552*^9, 3.615282007715686*^9}, {3.61528204200973*^9, 
   3.61528204282494*^9}, {3.615282228432449*^9, 3.615282276258251*^9}, {
   3.61528246585167*^9, 3.615282473086159*^9}, {3.615283309615147*^9, 
   3.615283377944027*^9}, {3.615283430346894*^9, 3.61528362998822*^9}, {
   3.615283779945492*^9, 3.61528378392452*^9}, {3.615544673131165*^9, 
   3.615544685347753*^9}},
 CellID->1090519763],

Cell[TextData[{
 "Finally, the whole process starts again from the beginning: New notions may \
be introduced and new lemmata and theorems may be proved. All the proving can \
now be done, of course, using the new prover developed in the previous \
step(s).\nIn our case, we introduce notions directly related to the \
complexity analysis (like ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 ") in section ",
 ButtonBox["Specific Notions for Complexity Analysis",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::SpecificNotions"],
 ", defining them in terms of the basic notions introduced in the previous \
round of the Theory Exploration Cycle (min, max, tuples, ...). Then we prove \
simple lemmata about them in section ",
 ButtonBox["Lemmata on Specific Notions",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataSpecific"],
 ", before we move on to the more advanced theorems in section ",
 ButtonBox["Theorems",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Theorems"],
 ". Finally, we turn to the main results the whole complexity analysis is all \
about, proving them in the same (rewriting) spirit as the other theorems."
}], "Text",
 CellChangeTimes->{{3.607321126621396*^9, 3.607321187975821*^9}, {
  3.607321257510742*^9, 3.607321314483194*^9}, {3.615280331530053*^9, 
  3.615280533911258*^9}, {3.615282307909117*^9, 3.615282343932688*^9}, {
  3.615283788492838*^9, 3.615283788733065*^9}},
 CellID->651312377],

Cell["\<\
A graphical visualization of the Theory Exploration cycle can be found in the \
diagram below.\
\>", "Text",
 CellChangeTimes->{{3.615282364154847*^9, 3.615282385493193*^9}},
 CellID->887402679],

Cell[BoxData[
 GraphicsBox[{GeometricTransformationBox[GraphicsGroupBox[{
      {EdgeForm[{RGBColor[0, 0, 0.498039], Thickness[0.0041209751909732995`], 
       JoinForm[{"Miter", 0.}]}], FaceForm[{RGBColor[0, 0.556863, 0.835294], 
       Opacity[1.]}], 
       StyleBox[
        RectangleBox[{0.4663123054051776, 0.9682080924855498}, {1.128869690424766, 0.7138728323699429},
         
         RoundingRadius->0.1],
        RectangleBoxOptions->{RoundingRadius->0}]}, 
      StyleBox[InsetBox[
        StyleBox[Cell[TextData[{
          StyleBox["Introduce",
           FontFamily->"Bitstream Charter",
           FontWeight->"Bold"],
          StyleBox["\nnew notions",
           FontFamily->"Bitstream Charter",
           FontWeight->"Plain"]
         }],
          GeneratedCell->False,
          CellAutoOverwrite->False,
          CellBaseline->Baseline,
          TextAlignment->Center,
          TextJustification->0],
         TextJustification->0,
         FontSize->24,
         FontWeight->Plain], {0.6171286297096231, 0.8664418754014138}, {
        Left, Baseline}, {0.40299944835581974`, 0.21387283236994226`}, {{1., 
        0.}, {0., 1.}},
        Alignment->{Left, Top}],
       FontColor->GrayLevel[1],
       Background->RGBColor[0., 0.6666666666666666, 1., 0]]}], {{
    0.34675511673351855`, -0.02873563218390774}}], 
   GeometricTransformationBox[GraphicsGroupBox[{
      {EdgeForm[{RGBColor[0, 0, 0.498039], Thickness[0.0041209751909732995`], 
       JoinForm[{"Miter", 0.}]}], FaceForm[{RGBColor[0, 0.556863, 0.835294], 
       Opacity[1.]}], 
       StyleBox[
        RectangleBox[{0.993155872205548, 0.6243438974154545}, {1.7560143240488815`, 0.37000863729984756`},
         
         RoundingRadius->0.1],
        RectangleBoxOptions->{RoundingRadius->0}]}, 
      StyleBox[InsetBox[
        StyleBox[Cell[TextData[{
          StyleBox["Prove\n",
           FontFamily->"Bitstream Charter",
           FontWeight->"Bold"],
          StyleBox["rewrite-kind lemmata",
           FontFamily->"Bitstream Charter",
           FontWeight->"Plain"]
         }],
          GeneratedCell->False,
          CellAutoOverwrite->False,
          CellBaseline->Baseline,
          TextAlignment->Center,
          TextJustification->0],
         TextJustification->0,
         FontSize->24,
         FontWeight->Plain], {1.03815620724368, 0.5218224702677561}, {
        Left, Baseline}, {0.6891344796312399, 0.20555555555555557`}, {{1., 
        0.}, {0., 1.}},
        Alignment->{Left, Top}],
       FontColor->GrayLevel[1],
       Background->RGBColor[0., 0.6666666666666666, 1., 0]]}], {{{{
      1.0300527025370756`, 0.}, {0., 1.}}, {0.4687925713519892, 
      0.002873563218391117}}}], GeometricTransformationBox[GraphicsGroupBox[{
      {EdgeForm[{RGBColor[0, 0, 0.498039], Thickness[0.0041209751909732995`], 
       JoinForm[{"Miter", 0.}]}], FaceForm[{RGBColor[0, 0.556863, 0.835294], 
       Opacity[1.]}], 
       StyleBox[
        RectangleBox[{0.4782942039961479, 0.3311906185635507}, {1.1838377605116268`, 0.07685535844794372},
         
         RoundingRadius->0.1],
        RectangleBoxOptions->{RoundingRadius->0}]}, 
      StyleBox[InsetBox[
        StyleBox[Cell[TextData[{
          StyleBox["Prove\n",
           FontFamily->"Bitstream Charter",
           FontWeight->"Bold"],
          StyleBox["advanced theorems",
           FontFamily->"Bitstream Charter",
           FontWeight->"Plain"]
         }],
          GeneratedCell->False,
          CellAutoOverwrite->False,
          CellBaseline->Baseline,
          TextAlignment->Center,
          TextJustification->0],
         TextJustification->0,
         FontSize->24,
         FontWeight->Plain], {0.5316106529659446, 0.23039720505835698`}, {
        Left, Baseline},
        Alignment->{Left, Top}],
       FontColor->GrayLevel[1],
       Background->RGBColor[0., 0.6666666666666666, 1., 0]]}], {{
    0.332426392901555, -0.022988505747126332`}}], 
   GeometricTransformationBox[GraphicsGroupBox[{
      {EdgeForm[{RGBColor[0, 0, 0.498039], Thickness[0.0041209751909732995`], 
       JoinForm[{"Miter", 0.}]}], FaceForm[{RGBColor[0, 0.556863, 0.835294], 
       Opacity[1.]}], 
       StyleBox[
        RectangleBox[{0.09715015006591665, 0.6099262507474585}, {0.8026937065813957, 0.3555909906318517},
         
         RoundingRadius->0.1],
        RectangleBoxOptions->{RoundingRadius->0}]}, 
      StyleBox[InsetBox[
        StyleBox[Cell[TextData[{
          StyleBox["Create",
           FontFamily->"Bitstream Charter",
           FontWeight->"Bold"],
          "\n",
          StyleBox["special prover",
           FontFamily->"Bitstream Charter",
           FontWeight->"Plain"]
         }],
          GeneratedCell->False,
          CellAutoOverwrite->False,
          CellBaseline->Baseline,
          TextAlignment->Center,
          TextJustification->0],
         TextJustification->0,
         FontSize->24,
         FontWeight->Plain], {0.24017617551634837`, 0.5190613026819921}, {
        Left, Baseline},
        Alignment->{Left, Top}],
       FontColor->GrayLevel[1],
       Background->RGBColor[
        0., 0.6666666666666666, 1., 0]]}], {{-0.011462979065570868`, 
    0.017241379310344862`}}], 
   {Thickness[0.0041209751909732995`], 
    JoinForm[{"Miter", 1.2000000000000002`}], StrokeForm[{RGBColor[
    0., 0., 0.4980392156862745], Opacity[1.]}], EdgeForm[{RGBColor[
    0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{1.4844557889914265`, 0.8189655172413793}, {1.8799285667536212`, 
     0.8189655172413793}}]}, 
   {Thickness[0.0041209751909732995`], 
    JoinForm[{"Miter", 1.2000000000000002`}], 
    Arrowheads[{{0.020044720273651614`, 1, {
        GraphicsBox[{
          EdgeForm[None], 
          Dashing[{}], 
          LineBox[{{-1, 0.5}, {0, 0}, {-1, -0.5}}]}], 0}}}], StrokeForm[{
    RGBColor[0., 0., 0.4980392156862745], Opacity[1.]}], EdgeForm[{RGBColor[
    0., 0., 0.4980392156862745], Opacity[1.], Thickness[
    0.0041209751909732995`], JoinForm[{"Miter", 1.2000000000000002`}]}], 
    FaceForm[{RGBColor[0, 0.666667, 1], Opacity[1.]}], 
    ArrowBox[{{1.8799285667536212`, 0.8160919540229885}, {1.8799285667536212`,
      0.6379310344827587}}]}, 
   {Thickness[0.0041209751909732995`], 
    JoinForm[{"Miter", 1.2000000000000002`}], 
    Arrowheads[{{0.020044720273651614`, 1, {
        GraphicsBox[{
          EdgeForm[None], 
          Dashing[{}], 
          LineBox[{{-1, 0.5}, {0, 0}, {-1, -0.5}}]}], 0}}}], StrokeForm[{
    RGBColor[0., 0., 0.4980392156862745], Opacity[1.]}], EdgeForm[{RGBColor[
    0., 0., 0.4980392156862745], Opacity[1.], Thickness[
    0.0041209751909732995`], JoinForm[{"Miter", 1.2000000000000002`}]}], 
    FaceForm[{RGBColor[0, 0.666667, 1], Opacity[1.]}], 
    ArrowBox[{{1.882794311520014, 0.17816091954023}, {1.53030770525371, 
     0.17816091954023}}]}, 
   {Thickness[0.0041209751909732995`], 
    JoinForm[{"Miter", 1.2000000000000002`}], StrokeForm[{RGBColor[
    0., 0., 0.4980392156862745], Opacity[1.]}], EdgeForm[{RGBColor[
    0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{1.8799285667536212`, 0.17816091954022995`}, {
     1.8799285667536212`, 0.3678160919540231}}]}, 
   {Thickness[0.0041209751909732995`], 
    JoinForm[{"Miter", 1.2000000000000002`}], StrokeForm[{RGBColor[
    0., 0., 0.4980392156862745], Opacity[1.]}], EdgeForm[{RGBColor[
    0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{0.8024085345899602, 0.18103448275862077`}, {
     0.44132469402447794`, 0.18103448275862088`}}]}, 
   {Thickness[0.0041209751909732995`], 
    JoinForm[{"Miter", 1.2000000000000002`}], 
    Arrowheads[{{0.020044720273651614`, 1, {
        GraphicsBox[{
          EdgeForm[None], 
          Dashing[{}], 
          LineBox[{{-1, 0.5}, {0, 0}, {-1, -0.5}}]}], 0}}}], StrokeForm[{
    RGBColor[0., 0., 0.4980392156862745], Opacity[1.]}], EdgeForm[{RGBColor[
    0., 0., 0.4980392156862745], Opacity[1.], Thickness[
    0.0041209751909732995`], JoinForm[{"Miter", 1.2000000000000002`}]}], 
    FaceForm[{RGBColor[0, 0.666667, 1], Opacity[1.]}], 
    ArrowBox[{{0.4384589492580853, 0.18091002610104276`}, {0.4384589492580853,
      0.35919540229885083`}}]}, 
   {Thickness[0.0041209751909732995`], 
    JoinForm[{"Miter", 1.2000000000000002`}], StrokeForm[{RGBColor[
    0., 0., 0.4980392156862745], Opacity[1.]}], EdgeForm[{RGBColor[
    0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{0.4355932044916925, 0.6379310344827587}, {0.4355932044916925, 
     0.8045977011494254}}]}, 
   {Thickness[0.0041209751909732995`], 
    JoinForm[{"Miter", 1.2000000000000002`}], 
    Arrowheads[{{0.020044720273651614`, 1, {
        GraphicsBox[{
          EdgeForm[None], 
          Dashing[{}], 
          LineBox[{{-1, 0.5}, {0, 0}, {-1, -0.5}}]}], 0}}}], StrokeForm[{
    RGBColor[0., 0., 0.4980392156862745], Opacity[1.]}], EdgeForm[{RGBColor[
    0., 0., 0.4980392156862745], Opacity[1.], Thickness[
    0.0041209751909732995`], JoinForm[{"Miter", 1.2000000000000002`}]}], 
    FaceForm[{RGBColor[0, 0.666667, 1], Opacity[1.]}], 
    ArrowBox[{{0.432655576126875, 0.8074712643678162}, {0.8052742793563528, 
     0.8074712643678162}}]}},
  AspectRatio->0.4317617866004963,
  ContentSelectable->True,
  ImagePadding->{{0., 0.}, {0., 0.}},
  ImageSize->{806., 348.},
  PlotRange->{{0., 2.3097902817125284`}, {0., 1.}},
  PlotRangePadding->Automatic]], "Input",
 Editable->False,
 Deployed->True,
 Evaluatable->False,
 CellChangeTimes->{{3.61554355153544*^9, 3.615543649083512*^9}, {
   3.615543686948122*^9, 3.615543714994059*^9}, {3.615543833130234*^9, 
   3.615543859385998*^9}, {3.615543902888292*^9, 3.615543903913033*^9}, {
   3.615543940837443*^9, 3.615544049602759*^9}, {3.61554410909378*^9, 
   3.615544167157309*^9}, {3.615544197348542*^9, 3.615544199609373*^9}, {
   3.615544293871739*^9, 3.615544333705148*^9}, {3.615544386062911*^9, 
   3.615544485158099*^9}, 3.615544626521217*^9},
 TextAlignment->Center,
 CellID->985362478],

Cell[CellGroupData[{

Cell["Conditional Equalities/Equivalences", "Subsection",
 CellChangeTimes->{{3.607321031688356*^9, 3.60732104052678*^9}},
 CellTags->"sect::Conditional",
 CellID->1621317523],

Cell["\<\
Assume a conditional equality which is meant to be used as a rewrite rule in \
automated proving:\
\>", "Text",
 CellChangeTimes->{{3.607321603806231*^9, 3.60732163190191*^9}},
 CellID->2117311209],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{"\[Phi]", "\[Implies]", 
   RowBox[{"a", "\[Equal]", "b"}]}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.61528105209452*^9, 3.615281058668355*^9}},
 CellID->1740035160],

Cell[TextData[{
 "In the present consideration it does not matter whether parts of the \
formula are in the scope of a universal quantifier or not, as long as all \
free variables in ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " are also free in ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 ". The rewrite rule originating from the equality would obviously replace \
all occurrences of term ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " by term ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 ", if formula ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " holds (w.l.o.g. we restrict ourselves to rewriting from left to right \
only).\nNow comes the crucial observation:"
}], "Text",
 CellChangeTimes->{{3.607321657390389*^9, 3.607321659708708*^9}, {
  3.607321694127045*^9, 3.607321778653125*^9}, {3.60732184763884*^9, 
  3.607321867381015*^9}, {3.607321919206449*^9, 3.607322071213518*^9}},
 CellID->62928363],

Cell[TextData[{
 StyleBox["In a proof-situation where ",
  FontSlant->"Italic"],
 Cell[BoxData[
  FormBox["a", TraditionalForm]],
  FontSlant->"Italic"],
 StyleBox[" appears somewhere, ",
  FontSlant->"Italic"],
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]],
  FontSlant->"Italic"],
 StyleBox[", although valid, might not be among the current assumptions.",
  FontSlant->"Italic"]
}], "Text",
 CellChangeTimes->{{3.607322080134776*^9, 3.607322137724434*^9}, {
  3.607327090064207*^9, 3.607327090170533*^9}, {3.607401808022174*^9, 
  3.607401808845304*^9}, {3.607401916123202*^9, 3.607401928692954*^9}},
 TextAlignment->Center,
 CellID->1867399679],

Cell["The question is now how to proceed:", "Text",
 CellChangeTimes->{{3.607322189930822*^9, 3.607322208825668*^9}},
 CellID->1242073734],

Cell[CellGroupData[{

Cell[TextData[{
 "Replace ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " anyway, and try to show ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " in a subproof. Of course, this can only be one alternative in the proof \
search, since ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " could be invalid or unprovable, but the main proof could nevertheless work."
}], "Item",
 CellChangeTimes->{{3.60732223716473*^9, 3.607322259038162*^9}, {
   3.607326751712302*^9, 3.607326801067228*^9}, {3.607326854316773*^9, 
   3.607326855515266*^9}, {3.607326999962222*^9, 3.607327026413442*^9}, {
   3.607398385719742*^9, 3.607398387542309*^9}, 3.607401867465799*^9},
 CellID->1092788260],

Cell[TextData[{
 "Wait until ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " is added to the knowledge base, and if it is never added, then just do not \
apply the rewrite rule (even though ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " might be easily provable)."
}], "Item",
 CellChangeTimes->{{3.607326611328108*^9, 3.607326724577948*^9}, {
  3.607398409996702*^9, 3.607398418316459*^9}},
 CellID->1560604785]
}, Open  ]],

Cell[TextData[{
 "In either case, from the purely logical point of view, everything is fine: \
If the replacement of ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " is needed for the proof to go through, then ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " has to be shown as well: Either it is already among the assumptions, or it \
is shown later. From the automated reasoning point of view, on the other \
hand, the two possibilities are essentially different:"
}], "Text",
 CellChangeTimes->{{3.607326842213332*^9, 3.607326973565636*^9}, {
  3.607327032920902*^9, 3.607327054956318*^9}, {3.607327104485379*^9, 
  3.607327106984009*^9}, {3.607401846371473*^9, 3.607401850800154*^9}},
 CellID->1630050989],

Cell[CellGroupData[{

Cell["\<\
The first possibility is more general but at the same time less efficient. \
Creating an alternative branch certainly will never do any harm, but if there \
are many equalities of the above kind, and only the conditions of some of \
them are really satisfied, then trying all of them will in most cases lead to \
a dramatic decrease of efficiency.\
\>", "Item",
 CellChangeTimes->{{3.60732722105348*^9, 3.607327381947737*^9}, {
  3.607327419460567*^9, 3.607327419721604*^9}},
 CellID->2003563700],

Cell["\<\
The second possibility, on the other hand, is more restrictive but also more \
efficient. A rewrite is only performed if it is known that it can be \
performed, meaning that on average fewer branches will be created during a \
proof search.\
\>", "Item",
 CellChangeTimes->{{3.60732722105348*^9, 3.607327407770034*^9}, {
  3.607327506167993*^9, 3.607327571216884*^9}, {3.607333100198001*^9, 
  3.607333124396083*^9}, {3.607333210012436*^9, 3.607333215128554*^9}},
 CellID->556412106]
}, Open  ]],

Cell[TextData[{
 "In the our opinion, the highest impact on which of the two possibilities \
should be preferred has the formula itself (not so much the term that is to \
be rewritten, or the proof situation): If the condition ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " can be considered a ",
 StyleBox["definedness condition",
  FontSlant->"Italic"],
 ", i.e. if it does not hold then the term ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " is not defined at all, then the first alternative seems to be better: It \
is very likely that in a proof everything is defined, hence just replace ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " and show ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " afterwards (or before).\nIf, however, the equality arises from a case \
distinction such that everything would be perfectly well defined even if ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " did not hold, then the second, more \[OpenCurlyDoubleQuote]cautious\
\[CloseCurlyDoubleQuote] alternative should be chosen."
}], "Text",
 CellChangeTimes->{{3.607333289794003*^9, 3.60733329598446*^9}, {
   3.607398083902368*^9, 3.607398084864654*^9}, {3.607398137134306*^9, 
   3.60739814195377*^9}, {3.607398211290026*^9, 3.607398257554815*^9}, 
   3.607398302156456*^9, {3.607398452839109*^9, 3.607398610524803*^9}, {
   3.607398697539144*^9, 3.607398888728125*^9}, {3.615278254278484*^9, 
   3.615278302534067*^9}},
 CellID->1546318573],

Cell[TextData[{
 "It has to be made very clear that all this is just ",
 StyleBox["heuristics",
  FontSlant->"Italic"],
 " for efficient automated reasoning: As mentioned already above, from the \
purely logical point of view, in either case ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " has to be provenly valid."
}], "Text",
 CellChangeTimes->{{3.607399012102162*^9, 3.607399084796302*^9}, {
  3.607399131223093*^9, 3.607399155952111*^9}},
 CellID->2094289943],

Cell[TextData[{
 "As an example consider the equalities/equivalences in section ",
 ButtonBox["Basic Notions / Categories",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::DefinitionCategories"],
 ", in particular the ones in the definition of Abelian monoids: There, every \
equality is subject to the constraint that all variables are elements of the \
given domain ",
 Cell[BoxData[
  FormBox["dom", TraditionalForm]]],
 ". However, if they were ",
 StyleBox["not",
  FontSlant->"Italic"],
 " elements of ",
 Cell[BoxData[
  FormBox["dom", TraditionalForm]]],
 ", then the left-hand-sides (which involve the monoid-operation in ",
 Cell[BoxData[
  FormBox["dom", TraditionalForm]]],
 ") would not be defined anyway, which in turn would be an indication that \
something went completely wrong in a proof.\nFormula ",
 ButtonBox["(W\:293a FarExterior)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1051080023"],
 " on the other hand, serves as an example for the other kind of equalities: \
There, the conditions are that some point has to lie in the far exterior of a \
set of points. If the conditions are not satisfied, then still both sides of \
the equalities are perfectly well-defined, only that the equalities simply do \
not hold."
}], "Text",
 CellChangeTimes->{{3.607399181516753*^9, 3.607399185198323*^9}, {
  3.607399230117474*^9, 3.607399257009001*^9}, {3.607399319463166*^9, 
  3.607399482912143*^9}, {3.607399545379172*^9, 3.607399552357965*^9}, {
  3.607399628403755*^9, 3.607399686703866*^9}, {3.607399779078105*^9, 
  3.607399779373732*^9}, {3.607399814890738*^9, 3.607399852420564*^9}, {
  3.607399903196903*^9, 3.607399915286368*^9}, {3.607400719349569*^9, 
  3.607400787775235*^9}, {3.607400846443076*^9, 3.607400967565685*^9}, {
  3.607401110047787*^9, 3.607401141145047*^9}, {3.607401940384232*^9, 
  3.607401962218021*^9}, {3.615278326744637*^9, 3.615278385932179*^9}, {
  3.615278467993021*^9, 3.615278504446682*^9}},
 CellID->8212415],

Cell[TextData[{
 "The main question is now: Is it possible to automatically decide whether \
the condition of an equality/equivalence should rather be checked a-priori or \
a-posteriori? Domain constraints seem to be a good candidate for the latter \
(especially if the left-hand-side of the equality/equivalence involves a \
domain operation), however there are certainly more: Minimum and maximum of a \
tuple are not only undefined if the tuple is not made up by elements of a \
min-max domain (\[Rule] domain constraint), but also if the tuple is ",
 StyleBox["empty",
  FontSlant->"Italic"],
 ". Hence, this should better also be checked in a subproof rather than being \
required to appear among the assumptions at the time a rewrite is attempted.\n\
One possible solution could be the introduction of a unary \
\[OpenCurlyDoubleQuote]definedness predicate\[CloseCurlyDoubleQuote], like \
the \[DownArrow]-predicate commonly used in computer science to indicate that \
a procedure terminates on some input. Then, one could write something like, \
say,"
}], "Text",
 CellChangeTimes->{{3.607399932901001*^9, 3.607399973254925*^9}, {
  3.607400009399197*^9, 3.607400053196141*^9}, {3.607400083993221*^9, 
  3.607400387854796*^9}, {3.607400834553008*^9, 3.607400836561582*^9}, {
  3.607401548367975*^9, 3.607401637868091*^9}, {3.615278543449618*^9, 
  3.615278654354283*^9}, {3.615278860834609*^9, 3.615278862844121*^9}},
 CellID->1516906979],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"dom", ",", "A"}]], 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         UnderscriptBox["min", "dom"], "[", "A", "]"}], "\[DownArrow]"}], 
       ")"}], 
      TagBox[
       RowBox[{":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
        RowBox[{"\[DoubleLongLeftRightArrow]", 
         RowBox[{
          UnderscriptBox["\[SquareSubsetEqual]", "dom"], 
          RowBox[{"[", "A", "]"}]}]}]}],
       Identity,
       SyntaxForm->"a\[DoubleLongLeftRightArrow]b"]}], "\[And]", 
     RowBox[{"1", "\[LessEqual]", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]}], 
    ")"}]}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.615278673874835*^9, 3.615278822383223*^9}},
 CellID->380971111],

Cell[TextData[{
 "and impose the condition ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["min", "dom"], "[", "A", "]"}], "\[DownArrow]"}], 
   TraditionalForm]]],
 " in formulas like ",
 ButtonBox["(min\[LessEqual])",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:48233088"],
 ", ",
 ButtonBox["(\[LessEqual]min)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1034128987"],
 " and so on. Now, in proofs, conditions of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[DownArrow]"}], TraditionalForm]]],
 " could be subject to a-posteriori checking, whereas all other conditions \
might still be checked a-priori.\n(Of course, \[DownArrow] has to be treated \
differently when expanding definitions in order to make sure that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[DownArrow]"}], TraditionalForm]]],
 " is not just replaced by its definition ",
 StyleBox["before",
  FontSlant->"Italic"],
 " applying the rewrite-rule is attempted.)"
}], "Text",
 CellChangeTimes->{{3.615278834665705*^9, 3.615278836783191*^9}, {
  3.615278868259551*^9, 3.615278902176534*^9}, {3.615278932285091*^9, 
  3.615279118021153*^9}, {3.615279197660686*^9, 3.615279211061471*^9}, {
  3.615279286093716*^9, 3.615279293573588*^9}},
 CellID->2067379196],

Cell[TextData[{
 "The rewriting mechanism of the prover we implemented for our present theory \
exploration, however, still treats all conditions in the same manner: They \
are all checked a-priori. All the formulas/rewrite-rules which should their \
conditions have checked a-posteriori, we simply lifted to the level of \
inference rules. For more details, please see section ",
 ButtonBox["Prover",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Prover"],
 "."
}], "Text",
 CellChangeTimes->{{3.615279228338623*^9, 3.615279468538034*^9}},
 CellID->737651657],

Cell["\<\
For the sake of completeness it has to be mentioned that deciding whether \
conditions of an inference rule should be checked a-priori or a-posteriori is \
a general problem in automated reasoning, not only related to inference rules \
dealing with rewriting.\
\>", "Text",
 CellChangeTimes->{{3.615544726134498*^9, 3.615544819578828*^9}},
 CellID->1982664360]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Problem Setting", "Section",
 CellChangeTimes->{{3.603513563731837*^9, 3.603513566696873*^9}},
 CellTags->"sect::ProblemSetting",
 CellID->264353757],

Cell[TextData[{
 "The problem which is studied in this paper is the problem of analyzing the \
complexity of Buchberger\[CloseCurlyQuote]s algorithm in the bivariate case, \
or, more precisely, the formalization thereof. This section first presents \
the algorithm and then repeats the most important steps of the complexity \
analysis, as already carried out in the late 1970s / early 1980s by \
Buchberger in ",
 ButtonBox["[Buchberger1979]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979"],
 ", ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 ", ",
 ButtonBox["[Buchberger1983a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983a"],
 " and ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 "."
}], "Text",
 CellChangeTimes->{{3.603513828179648*^9, 3.603513867806993*^9}, {
  3.603513903029373*^9, 3.603513914112895*^9}, {3.603513953678226*^9, 
  3.603513961866312*^9}},
 CellID->229041570],

Cell[CellGroupData[{

Cell["The Algorithm", "Subsection",
 CellChangeTimes->{{3.603513977405846*^9, 3.603513980180293*^9}},
 CellID->1990765324],

Cell[TextData[{
 "The algorithm that is dealt with throughout the whole paper is Buchberger\
\[CloseCurlyQuote]s algorithm for computing a Gr\[ODoubleDot]bner basis ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " of an ideal ",
 Cell[BoxData[
  FormBox["I", TraditionalForm]]],
 " over the polynomial ring over some field ",
 Cell[BoxData[
  FormBox["K", TraditionalForm]]],
 ", given any basis ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " of ",
 Cell[BoxData[
  FormBox["I", TraditionalForm]]],
 ". In particular, we concentrate here only on the bivariate case, i. e. our \
polynomial ring will in the sequel always be denoted as ",
 Cell[BoxData[
  FormBox[
   RowBox[{"K", "[", 
    RowBox[{"x", ",", "y"}], "]"}], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " are the two indeterminates.\nThere are two more things to note:"
}], "Text",
 CellChangeTimes->{{3.60351400531627*^9, 3.60351407158612*^9}, {
  3.615019891488608*^9, 3.61501990723894*^9}},
 CellID->453193865],

Cell[CellGroupData[{

Cell[TextData[{
 "We will ",
 StyleBox["not",
  FontSlant->"Italic"],
 " consider the original algorithm invented in ",
 ButtonBox["[Buchberger1965]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1965"],
 ", but an improved version where the so-called ",
 StyleBox["chain criterion",
  FontSlant->"Italic"],
 " is used. This criterion allows to ignore some (in fact many) pairs of \
polynomials, and still get the corrrect result, which, of course, has a \
dramatic influence on the complexity of the algorithm. For more information \
about the chain criterion, including a correctness proof thereof, we refer to \
",
 ButtonBox["[Buchberger1979]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979"],
 "."
}], "Item",
 CellChangeTimes->{{3.603514090207411*^9, 3.603514181014179*^9}},
 CellID->708458737],

Cell[TextData[{
 "We will only consider ",
 StyleBox["graded",
  FontSlant->"Italic"],
 " admissible term orderings, simply because the complexity analysis can only \
be carried out if graded orderings are used. Graded means: First the degrees \
of two monomials are compared, and only if they are equal, other criteria are \
used for deciding which monomial is greater."
}], "Item",
 CellChangeTimes->{{3.603514090207411*^9, 3.603514197291496*^9}, {
  3.615019934260492*^9, 3.615019965522169*^9}},
 CellID->709660607]
}, Open  ]],

Cell[TextData[{
 "We now fix some admissible ordering ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 " on the set of monomials, assume that it is graded, and tacitly assume in \
the rest of the paper that all leading monomials, S-polynomials, reductions, \
Gr\[ODoubleDot]bner bases, etc. are meant w.r.t. ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 ". Also, since we consider the classical setting of polynomials over a \
field, we tacitly assume from now on that the leading coefficient of every \
polynomial w.r.t. ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 " is 1, for otherwise the polynomial can always be made monic by \
multiplication by some suitable non-zero constant."
}], "Text",
 CellChangeTimes->{{3.603514227209507*^9, 3.603514284921867*^9}, {
  3.615019977267692*^9, 3.615019988854021*^9}},
 CellID->808711520],

Cell["Algorithm 1 is the algorithm this paper is all about.", "Text",
 CellChangeTimes->{{3.603514306259753*^9, 3.603514310831261*^9}},
 CellID->1150089532],

Cell[TextData[{
 StyleBox["Algorithm 1",
  FontWeight->"Bold"],
 " ",
 StyleBox["Buchberger\[CloseCurlyQuote]s Algorithm with chain criterion",
  FontSlant->"Italic"]
}], "Text",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.60351440820717*^9, 3.60351444343292*^9}, {
  3.603514835118822*^9, 3.603514838479405*^9}},
 Background->RGBColor[0.94, 0.91, 0.88],
 CellTags->"Alg::BB",
 CellID->1191384728],

Cell[TextData[{
 "Input: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"F", "=", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       SubscriptBox["p", "1"], ",", "...", ",", 
       SubscriptBox["p", "n"]}], "}"}], "\[SubsetEqual]", 
     RowBox[{"K", "[", 
      RowBox[{"x", ",", "y"}], "]"}]}]}], TraditionalForm]]],
 "\nOutput: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"G", "\[SubsetEqual]", 
    RowBox[{"K", "[", 
     RowBox[{"x", ",", "y"}], "]"}]}], TraditionalForm]]],
 " s. t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"ideal", "[", "F", "]"}], "=", 
    RowBox[{"ideal", "[", "G", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " is Gr\[ODoubleDot]bner basis"
}], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{
  3.60351444343301*^9, {3.603514473829169*^9, 3.603514571122951*^9}, {
   3.603514691473249*^9, 3.603514691952671*^9}, {3.603514835118898*^9, 
   3.60351483847951*^9}},
 CellID->103682965],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{"P", "\[LeftArrow]", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       SubscriptBox["p", "i"], ",", 
       SubscriptBox["p", "j"]}], ")"}], "|", 
     RowBox[{"1", "\[LessEqual]", "i", "<", "j", "\[LessEqual]", "n"}]}], 
    "}"}]}], TraditionalForm]]]], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["1", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9},
 CellID->1180081128],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{"G", "\[LeftArrow]", "B"}], TraditionalForm]]]], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["2", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603514968468598*^9, 3.603514972303034*^9}},
 CellID->772744986],

Cell[TextData[{
 StyleBox["while",
  FontWeight->"Bold"],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{"P", "\[NotEqual]", "\[EmptySet]"}], TraditionalForm]]],
 " ",
 StyleBox["do",
  FontWeight->"Bold"]
}], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["3", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603514976432092*^9, 3.603514989419247*^9}},
 CellID->789122655],

Cell[TextData[{
 "\tchoose some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"f", ",", "g"}], ")"}], "\[Element]", "P"}], TraditionalForm]]]
}], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["4", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603514999139785*^9, 3.603515013391641*^9}},
 CellID->1759507030],

Cell[TextData[{
 "\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"P", "\[LeftArrow]", 
    RowBox[{"P", "\[Backslash]", 
     RowBox[{"{", 
      RowBox[{"(", 
       RowBox[{"f", ",", "g"}], ")"}], "}"}]}]}], TraditionalForm]]]
}], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["5", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603515017438977*^9, 3.603515037706655*^9}},
 CellID->2063177316],

Cell[TextData[{
 "\t",
 StyleBox["if",
  FontWeight->"Bold"],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    StyleBox["chainCrit",
     FontFamily->"Courier 10 Pitch"], "[", 
    RowBox[{"f", ",", "g", ",", "G"}], "]"}], TraditionalForm]]],
 " ",
 StyleBox["then",
  FontWeight->"Bold"]
}], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["6", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603515042408664*^9, 3.603515070490339*^9}},
 CellID->2110989409],

Cell[TextData[{
 "\t\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"h", "\[LeftArrow]", 
    RowBox[{
     StyleBox["sPoly",
      FontFamily->"Courier 10 Pitch"], "[", 
     RowBox[{"f", ",", "g"}], "]"}]}], TraditionalForm]]]
}], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["7", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603515079508397*^9, 3.603515093915798*^9}},
 CellID->2023860787],

Cell[TextData[{
 "\t\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"h", "\[LeftArrow]", 
    RowBox[{
     StyleBox["totalReduce",
      FontFamily->"Courier 10 Pitch"], "[", 
     RowBox[{"h", ",", "G"}], "]"}]}], TraditionalForm]]]
}], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["8", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.60351509659434*^9, 3.603515111635871*^9}},
 CellID->144735239],

Cell[TextData[{
 "\t\t",
 StyleBox["if",
  FontWeight->"Bold"],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{"h", "\[NotEqual]", "0"}], TraditionalForm]]],
 " ",
 StyleBox["then",
  FontWeight->"Bold"]
}], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["9", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603515114719383*^9, 3.603515121259113*^9}},
 CellID->1954271085],

Cell[TextData[{
 "\t\t\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"P", "\[LeftArrow]", 
    RowBox[{"P", "\[Union]", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"p", ",", "h"}], ")"}], "|", 
       RowBox[{"p", "\[Element]", "G"}]}], "}"}]}]}], TraditionalForm]]]
}], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["10", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603515136795569*^9, 3.603515157158676*^9}},
 CellID->329571004],

Cell[TextData[{
 "\t\t\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"G", "\[LeftArrow]", 
    RowBox[{"G", "\[Union]", 
     RowBox[{"{", "h", "}"}]}]}], TraditionalForm]]]
}], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["11", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603515159980712*^9, 3.603515168780605*^9}},
 CellID->853193976],

Cell[TextData[{
 "\t\t",
 StyleBox["end if",
  FontWeight->"Bold"]
}], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["12", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603515173374754*^9, 3.603515176530522*^9}},
 CellID->516448044],

Cell[TextData[{
 "\t",
 StyleBox["end if",
  FontWeight->"Bold"]
}], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["13", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603515181426625*^9, 3.603515184289519*^9}},
 CellID->1138282524],

Cell[TextData[StyleBox["end while",
 FontWeight->"Bold"]], "Text",
 CellDingbat->Cell[
   TextData[
    StyleBox["14", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603515190523074*^9, 3.603515194961055*^9}},
 CellID->2003283509],

Cell[TextData[{
 StyleBox["return",
  FontWeight->"Bold"],
 " ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]]
}], "Text",
 CellFrame->{{0, 0}, {0.5, 0}},
 CellDingbat->Cell[
   TextData[
    StyleBox["15", FontSize -> 10]], "Text"],
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603514720578782*^9, 3.603514743448958*^9}, 
   3.603514838479707*^9, {3.603515200385105*^9, 3.603515202400119*^9}},
 CellID->1764664231],

Cell[TextData[{
 "Apparently there are three auxiliary functions used in the algorithm: ",
 StyleBox["chainCrit",
  FontFamily->"Courier 10 Pitch"],
 ", ",
 StyleBox["sPoly",
  FontFamily->"Courier 10 Pitch"],
 " and ",
 StyleBox["totalReduce",
  FontFamily->"Courier 10 Pitch"],
 ". Apart from the first one there should not be any doubts concerning their \
meaning: ",
 StyleBox["sPoly",
  FontFamily->"Courier 10 Pitch"],
 " simply computes the S-polynomial of two given polynomials and ",
 StyleBox["totalReduce",
  FontFamily->"Courier 10 Pitch"],
 " reduces a given polynomial modulo a given set in some way until an \
irreducible normal form is reached. ",
 StyleBox["chainCrit",
  FontFamily->"Courier 10 Pitch"],
 " is the predicate which decides whether a pair of polynomials has to be \
considered (i.e. its S-polynomial has to be reduced) or not. It is defined in \
the following way:"
}], "Text",
 CellChangeTimes->{{3.603515871229201*^9, 3.603515877699088*^9}, {
   3.60351604293495*^9, 3.60351608364731*^9}, {3.603516176881798*^9, 
   3.603516194572595*^9}, 3.615020032981593*^9},
 CellID->764371851],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"f", ",", 
      RowBox[{"g", "\[Element]", 
       RowBox[{"K", "[", 
        RowBox[{"x", ",", "y"}], "]"}]}]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"A", "\[SubsetEqual]", 
       RowBox[{"K", "[", 
        RowBox[{"x", ",", "y"}], "]"}]}]], 
     RowBox[{
      StyleBox["chainCrit",
       FontFamily->"Courier 10 Pitch"], "[", 
      RowBox[{"f", ",", "g", ",", "A"}], "]"}]}]}], ":", 
   RowBox[{
    RowBox[{"\[DoubleLeftRightArrow]", 
     RowBox[{"\[Not]", 
      UnderscriptBox["\[Exists]", 
       RowBox[{"p", "\[Element]", "A"}]]}]}], "\[And]", 
    TagBox[GridBox[{
       {"\[Piecewise]", GridBox[{
          {
           RowBox[{
            OverscriptBox["p", "_"], "|", 
            RowBox[{"lcm", "[", 
             RowBox[{
              OverscriptBox["f", "_"], ",", 
              OverscriptBox["g", "_"]}], "]"}]}]},
          {
           RowBox[{
            RowBox[{"deg", "[", 
             RowBox[{"lcm", "[", 
              RowBox[{
               OverscriptBox["p", "_"], ",", 
               OverscriptBox["f", "_"]}], "]"}], "]"}], "<", 
            RowBox[{"deg", "[", 
             RowBox[{"lcm", "[", 
              RowBox[{
               OverscriptBox["f", "_"], ",", 
               OverscriptBox["g", "_"]}], "]"}], "]"}]}]},
          {
           RowBox[{
            RowBox[{"deg", "[", 
             RowBox[{"lcm", "[", 
              RowBox[{
               OverscriptBox["p", "_"], ",", 
               OverscriptBox["g", "_"]}], "]"}], "]"}], "<", 
            RowBox[{"deg", "[", 
             RowBox[{"lcm", "[", 
              RowBox[{
               OverscriptBox["f", "_"], ",", 
               OverscriptBox["g", "_"]}], "]"}], "]"}]}]}
         },
         AllowedDimensions->{2, Automatic},
         Editable->True,
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
         GridBoxItemSize->{
          "Columns" -> {{Automatic}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{1.}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.84]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}},
         Selectable->True]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{
       "Columns" -> {{Automatic}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.35]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     "Piecewise",
     DeleteWithContents->True,
     Editable->False,
     SelectWithContents->True,
     Selectable->False]}]}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.60351626512132*^9, 3.603516397783175*^9}, {
  3.615020232899932*^9, 3.615020244501825*^9}, {3.615020278928716*^9, 
  3.615020305550719*^9}, {3.615020430580325*^9, 3.615020453954097*^9}},
 CellTags->"Eqn::ChainCrit",
 CellID->1279890854],

Cell[TextData[{
 "meaning that ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 " can be ignored if it ",
 StyleBox["does not",
  FontSlant->"Italic"],
 " hold."
}], "Text",
 CellChangeTimes->{{3.615020084343688*^9, 3.615020118957384*^9}},
 CellID->1752125308],

Cell[TextData[{
 "The meanings of the symbols appearing in this definition are the following, \
for all monomials ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " and all polynomials ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{
  3.603516469257423*^9, {3.615020348298978*^9, 3.615020389471855*^9}},
 CellID->471797467],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   OverscriptBox["c", "_"], TraditionalForm]]],
 " denotes the leading monomial of polynomial ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " (w.r.t. ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 ")"
}], "Item",
 CellChangeTimes->{{3.603516481478761*^9, 3.603516505611879*^9}, {
  3.615020146771598*^9, 3.615020147852486*^9}, {3.615020264839988*^9, 
  3.61502026944788*^9}, {3.61502031227259*^9, 3.615020337597559*^9}, {
  3.615020375666084*^9, 3.615020377323549*^9}},
 CellID->1506065214],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"deg", "[", "a", "]"}], TraditionalForm]]],
 " denotes the degree of monomial ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]]
}], "Item",
 CellChangeTimes->{{3.603516481478761*^9, 3.603516542300097*^9}, {
  3.615020316420436*^9, 3.615020361222721*^9}},
 CellID->684093844],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "|", "b"}], TraditionalForm]]],
 " denotes divisibility of monomial ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " by monomial ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]]
}], "Item",
 CellChangeTimes->{{3.603516481478761*^9, 3.603516560476021*^9}, {
  3.615020461861531*^9, 3.615020461868001*^9}},
 CellID->417957227],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "[", 
    RowBox[{"a", ",", "b"}], "]"}], TraditionalForm]]],
 " denotes the least common multiple of monomials ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]]
}], "Item",
 CellChangeTimes->{{3.603516481478761*^9, 3.603516560476021*^9}, {
  3.615020461861531*^9, 3.6150204792164*^9}},
 CellID->1858807165]
}, Open  ]],

Cell[TextData[{
 "Please note here that the definition of ",
 StyleBox["chainCrit",
  FontFamily->"Courier 10 Pitch"],
 " as it is presented above is only one of many different possibilities. For \
instance, one could also compare ",
 Cell[BoxData[
  FormBox[
   OverscriptBox["p", "_"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "[", 
    RowBox[{
     OverscriptBox["f", "_"], ",", 
     OverscriptBox["g", "_"]}], "]"}], TraditionalForm]]],
 " w.r.t. ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 " rather than only comparing some degrees. However, all this does not really \
make any big difference, so we just concentrate on the variant as defined \
above in this report."
}], "Text",
 CellChangeTimes->{{3.603516577842718*^9, 3.603516662324047*^9}, {
  3.61502016768799*^9, 3.615020215690305*^9}, {3.61502049766072*^9, 
  3.615020564757481*^9}, {3.615020652153076*^9, 3.615020735003706*^9}},
 CellID->1268377962]
}, Closed]],

Cell[CellGroupData[{

Cell["The Complexity Analysis", "Subsection",
 CellChangeTimes->{{3.603516713820491*^9, 3.603516718572327*^9}},
 CellID->1943935944],

Cell[TextData[{
 "Fix some finite non-empty ",
 Cell[BoxData[
  FormBox[
   RowBox[{"F", "\[SubsetEqual]", 
    RowBox[{"K", "[", 
     RowBox[{"x", ",", "y"}], "]"}]}], TraditionalForm]]],
 ".\n\nIn ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 ", ",
 ButtonBox["[Buchberger1983a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983a"],
 " and ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 " the complexity analysis of ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ", when applied on ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 ", is carried out in the following way:"
}], "Text",
 CellChangeTimes->{{3.603516743486818*^9, 3.603516841891178*^9}, {
  3.603516873848898*^9, 3.603516875528761*^9}},
 CellID->2505218],

Cell[TextData[{
 "Derive an upper bound ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["B", "F"], TraditionalForm]]],
 " such that every polynomial ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 " which is constructed in ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 " (and in particular every element of the final Gr\[ODoubleDot]bner basis ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 ") satisfies ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "[", "g", "]"}], "\[LessEqual]", 
    SubscriptBox["B", "F"]}], TraditionalForm]]],
 "."
}], "Text",
 CellDingbat->"1.",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.615020798047349*^9, 3.61502090846506*^9}, 
   3.615283683118134*^9},
 CellID->1802962063],

Cell[TextData[{
 "Argue that as soon as ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["B", "F"], TraditionalForm]]],
 " is known one can also derive a bound ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "F"], TraditionalForm]]],
 " for the number of polynomials in ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 ". This is one of the proofs where ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 " is required to be graded."
}], "Text",
 CellDingbat->"2.",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603516872439702*^9, 3.603516967550631*^9}, 
   3.603517609668004*^9, {3.615020789472474*^9, 3.615020789618496*^9}},
 CellID->1700275030],

Cell[TextData[{
 "As soon as ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "F"], TraditionalForm]]],
 " is known one can estimate the number of elementary operations performed in \
",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 "."
}], "Text",
 CellDingbat->"3.",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603516872439702*^9, 3.603516990621042*^9}, 
   3.603517609668124*^9, {3.615020991927137*^9, 3.61502099194006*^9}},
 CellID->1831213623],

Cell[TextData[{
 "The second and the third part are dealt with in ",
 ButtonBox["[Buchberger1979]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979"],
 " and are, compared to the first part, rather straightforward. The \
complicated task is to derive ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["B", "F"], TraditionalForm]]],
 ", which is done in all detail in ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 " and ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 ", where a suitbale value of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["B", "F"], TraditionalForm]]],
 " is shown to be ",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    RowBox[{"maxdeg", "[", "F", "]"}]}], TraditionalForm]]],
 ". ",
 Cell[BoxData[
  FormBox[
   RowBox[{"maxdeg", "[", "F", "]"}], TraditionalForm]]],
 " is the maximum degree of polynomials in ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 ", formally defined as ",
 Cell[BoxData[
  FormBox[
   RowBox[{"max", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"deg", "[", "f", "]"}], "|", 
      RowBox[{"f", "\[Element]", "F"}]}], "}"}], "]"}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.603517008130733*^9, 3.603517110634276*^9}, {
  3.603517450903783*^9, 3.603517471704964*^9}},
 CellID->1948833634],

Cell["The proof of this bound is again split into two parts:", "Text",
 CellChangeTimes->{
  3.603517138889573*^9, {3.603517612717007*^9, 3.603517616281042*^9}},
 CellID->827429947],

Cell[TextData[{
 "First prove ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "[", "h", "]"}], "\[LessEqual]", 
    RowBox[{
     SubscriptBox["M", "F"], "+", 
     SubscriptBox["W", "F"]}]}], TraditionalForm]]],
 " for all ",
 Cell[BoxData[
  FormBox["h", TraditionalForm]]],
 " added to ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " in line 11 of ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ". This proof can be found in ",
 ButtonBox["[Buchberger1979a].",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"]
}], "Text",
 CellDingbat->"1.1",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603516872439702*^9, 3.603516930647182*^9}, {
  3.603517427458401*^9, 3.603517437758654*^9}, {3.603517484249102*^9, 
  3.603517524748578*^9}, {3.603517583487507*^9, 3.603517620042918*^9}, {
  3.60352950475488*^9, 3.603529519101211*^9}, {3.615021038421776*^9, 
  3.615021038446014*^9}},
 CellID->1480049348],

Cell[TextData[{
 "Afterwards show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      SubscriptBox["M", "F"], "+", 
      SubscriptBox["W", "F"]}], "\[LessEqual]", 
     RowBox[{"2", 
      RowBox[{"maxdeg", "[", "F", "]"}]}]}], "=", 
    SubscriptBox["B", "F"]}], TraditionalForm]]],
 ". This proof can be found in ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 "."
}], "Text",
 CellDingbat->"1.2",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603516872439702*^9, 3.603516967550631*^9}, {
  3.603517532524171*^9, 3.603517620043091*^9}, {3.603529500533645*^9, 
  3.603529521687514*^9}},
 CellID->200043405],

Cell[TextData[{
 "The two symbols ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["W", TraditionalForm]]],
 " appearing above are defined in the following way:"
}], "Text",
 CellChangeTimes->{{3.603517643953584*^9, 3.603517650910419*^9}},
 CellID->1617422914],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   SubscriptBox["M", "A"], ":=", 
   RowBox[{"max", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"deg", "[", 
        RowBox[{"f", ",", "g"}], "]"}], "|", "f"}], ",", 
      RowBox[{
       RowBox[{"g", "\[Element]", "A"}], "\[And]", 
       RowBox[{
        StyleBox["chainCrit",
         FontFamily->"Courier 10 Pitch"], "[", 
        RowBox[{"f", ",", "g", ",", "A"}], "]"}]}]}], "}"}], "]"}]}], 
  TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.603517671071237*^9, 3.603517713645143*^9}, {
  3.615021053743397*^9, 3.615021053743509*^9}},
 CellTags->"Eqn::M",
 CellID->862257598],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   SubscriptBox["W", "A"], ":=", 
   RowBox[{
    RowBox[{"min", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        SubscriptBox["deg", "x"], "[", "f", "]"}], "|", 
       RowBox[{"f", "\[Element]", "A"}]}], "}"}], "]"}], "+", 
    RowBox[{"min", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        SubscriptBox["deg", "y"], "[", "f", "]"}], "|", 
       RowBox[{"f", "\[Element]", "A"}]}], "}"}], "]"}]}]}], 
  TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.603517748447124*^9, 3.603517804058184*^9}, {
  3.615021055067661*^9, 3.615021055067771*^9}},
 CellTags->"Eqn::W",
 CellID->2012096862],

Cell[TextData[{
 "for all sets ",
 Cell[BoxData[
  FormBox[
   RowBox[{"A", "\[SubsetEqual]", 
    RowBox[{"K", "[", 
     RowBox[{"x", ",", "y"}], "]"}]}], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["deg", "x"], "[", "f", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["deg", "y"], "[", "f", "]"}], TraditionalForm]]],
 " denote the exponent of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", respectively, in ",
 Cell[BoxData[
  FormBox[
   OverscriptBox["f", "_"], TraditionalForm]]],
 " (i.e. the leading monomial of ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 "). Note that ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["M", "F"], TraditionalForm]]],
 " is the maximum degree of the least common multiple of all pairs of \
polynomials in ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " that cannot be ruled out by the chain criterion. ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["W", "F"], TraditionalForm]]],
 ", on the other hand, can be regarded as the \[OpenCurlyDoubleQuote]width\
\[CloseCurlyDoubleQuote] of ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 "; A good informal explanation of both ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["M", "F"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["W", "F"], TraditionalForm]]],
 " by means of drawings can be found, for instance, in ",
 ButtonBox["[Buchberger1979a] ",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 "as well as in section ",
 ButtonBox["Specific Notions for Complexity Analysis",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::SpecificNotions"],
 ". Both proofs originally are based on tedious case distinctions. These case \
distinctions can most probably ",
 StyleBox["not",
  FontSlant->"Italic"],
 " be avoided in principle, but, as the next sections show, simplified. \
Especially for the second proof of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["M", "F"], "+", 
     SubscriptBox["W", "F"]}], "\[LessEqual]", 
    RowBox[{"2", 
     RowBox[{"maxdeg", "[", "f", "]"}]}]}], TraditionalForm]]],
 " a drastic reduction of the required effort could be achieved."
}], "Text",
 CellChangeTimes->{{3.603517834991829*^9, 3.603517985432912*^9}, {
  3.615021067369342*^9, 3.615021171171457*^9}, {3.615021219222156*^9, 
  3.615021219784335*^9}},
 CellID->44376836],

Cell[TextData[{
 "This concludes the introductory part of this report, and we proceed with \
the formalization of the whole theory in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellChangeTimes->{{3.603518012890823*^9, 3.603518041357014*^9}},
 CellID->584987536]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Basic Notions", "Section",
 CellChangeTimes->{{3.596370403719267*^9, 3.59637040546714*^9}, {
  3.613971526381298*^9, 3.613971537944887*^9}},
 CellTags->"sect::BasicNotions",
 CellID->2131707921],

Cell[TextData[{
 "In this section all basic notions are introduced completely formally by \
means of ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " formulas. However, we cannot immediately start with the definitions of ",
 Cell[BoxData[
  FormBox["deg", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["M", "F"], TraditionalForm]]],
 " and the like, but first have to give formal definitions of several \
elementary notions (such as ",
 Cell[BoxData[
  FormBox["min", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["max", TraditionalForm]]],
 " and properties of orderings) that are needed as well."
}], "Text",
 CellChangeTimes->{{3.603518082070668*^9, 3.603518254977325*^9}, {
  3.603518297249111*^9, 3.603518310281262*^9}, {3.615021727458152*^9, 
  3.61502173906022*^9}},
 CellID->2115353011],

Cell[CellGroupData[{

Cell["General Notions", "Subsection",
 CellChangeTimes->{{3.598255110729544*^9, 3.598255111846261*^9}, {
  3.605522225588609*^9, 3.6055222271046*^9}},
 CellID->1121114900],

Cell[CellGroupData[{

Cell["Tuples", "Subsubsection",
 CellChangeTimes->{{3.615021532577185*^9, 3.615021542943566*^9}, 
   3.615021597716545*^9},
 CellTags->"sect::DefinitionTuples",
 CellID->25850532],

Cell["\<\
First of all, we introduce some basic knowledge about tuples. Tuples are of \
utmost importance for us, as they are the only building block for grouping \
and structuring collections of objects (e.g. individual components to \
exponent vectors, individual exponent vectors to tuples of exponent vectors, \
etc.). Sets are, for various reasons, not used at all.\
\>", "Text",
 CellChangeTimes->{{3.603600650700655*^9, 3.603600722741473*^9}, {
  3.603601211711103*^9, 3.603601338229597*^9}, {3.603601434495053*^9, 
  3.603601475612896*^9}, {3.615021843756115*^9, 3.615021849413657*^9}, {
  3.6150219154073*^9, 3.615021938185826*^9}, {3.615021982430476*^9, 
  3.615022000357375*^9}, {3.615022064431584*^9, 3.615022091346983*^9}, {
  3.615022200091207*^9, 3.61502233649015*^9}, {3.615022482707465*^9, 
  3.615022511717409*^9}},
 CellID->849860725],

Cell[TextData[{
 "Please note that \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox["isTuple", TraditionalForm]]],
 "\[CloseCurlyDoubleQuote], \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftBracketingBar]", "\[CenterDot]", "\[RightBracketingBar]"}], 
   TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] and tuple-indices are built-in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " symbols that can immediately, without any further effort, be used for \
computations. For reasoning however, special inference rule are needed; They \
are part of the ComplexityProver. Still, the theory of tuples could in \
principle also be built-up from scratch (e.g. from ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["\:293a", TraditionalForm]]],
 "), such that most formulas in the definition below would then rather be a \
lemma. This, however, is certainly ",
 StyleBox["not",
  FontSlant->"Italic"],
 " what we want to do here, since we want to exclusively focus on the \
complexity analysis and not on the theory of tuples. In other words: We \
proceed by proving not from ",
 StyleBox["first",
  FontSlant->"Italic"],
 ", but from ",
 StyleBox["intermediate",
  FontSlant->"Italic"],
 " principles."
}], "Text",
 CellChangeTimes->{{3.615544947214191*^9, 3.615545052109929*^9}, {
  3.615545101254546*^9, 3.615545177507439*^9}},
 CellID->2133492005],

Cell["", "OpenEnvironment",
 CellID->1844672668],

Cell[CellGroupData[{

Cell["Definition: tuples", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.60301456286258*^9, 3.603014571122563*^9}, 
   3.615021585308622*^9},
 CellID->796541026],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"isTuple", "[", "A", "]"}], ",", "C", ",", 
   "x"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.59928339993839*^9, 3.599283404973195*^9}, {
   3.60006569074781*^9, 3.600065691145244*^9}, {3.602918745500581*^9, 
   3.602918745793523*^9}, 3.602919233470488*^9},
 CellID->1752837732],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"[", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}], "]"}]}], 
  "\[And]", 
  RowBox[{
   UnderscriptBox["0", "\[DoubleStruckCapitalZ]"], 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(length is natural)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1969097229", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600149731447885*^9, 3.600149757084257*^9}, {
  3.600150130338942*^9, 3.600150141243081*^9}, {3.601195283810426*^9, 
  3.601195289588218*^9}, {3.603014742897557*^9, 3.60301476452911*^9}, {
  3.603170049900146*^9, 3.603170073838615*^9}, {3.603170397072157*^9, 
  3.603170436131136*^9}, {3.603177600395033*^9, 3.603177606487908*^9}, {
  3.603177754830228*^9, 3.603177774607562*^9}, {3.606119847122059*^9, 
  3.606119864013369*^9}, {3.6098253043214*^9, 3.609825305775691*^9}},
 CellTags->{
  "ID:1969097229", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length is natural"},
 CellID->1969097229],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "B"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[Equal]", "B"}], 
     TagBox[")",
      "AutoParentheses"]}], 
    TagBox["\[DoubleLeftRightArrow]",
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"isTuple", "[", "B", "]"}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], 
        "\[Equal]", 
        RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         SubscriptBox["B", "i"], "\[Equal]", 
         SubscriptBox["A", "i"]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(tuple equality)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1508742051", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.603692778631047*^9, 3.60369286514297*^9}, {
  3.603692941116594*^9, 3.603692971061688*^9}},
 CellTags->{
  "ID:1508742051", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "tuple equality"},
 CellID->1508742051],

Cell[BoxData[
 RowBox[{
  RowBox[{"isTuple", "[", 
   RowBox[{"C", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], "]"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"isTuple", "[", "C", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isTuple\:293a)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1264372773", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600149690432087*^9, 3.60014969649956*^9}, {
  3.600688305026612*^9, 3.600688309830191*^9}, {3.602918752421613*^9, 
  3.602918769809855*^9}, {3.602919235960657*^9, 3.602919249721165*^9}},
 CellTags->{
  "ID:1264372773", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isTuple\:293a"},
 CellID->1264372773],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[LeftBracketingBar]", 
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], "\[RightBracketingBar]"}], "\[Equal]", 
  RowBox[{
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}], 
   UnderscriptBox["+", "\[DoubleStruckCapitalZ]"], 
   "1"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(length\:293a)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2086441677", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600148414461241*^9, 3.600148428924182*^9}, {
  3.600688314137584*^9, 3.600688317926132*^9}, {3.603177622242685*^9, 
  3.603177627170487*^9}},
 CellTags->{
  "ID:2086441677", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a"},
 CellID->2086441677],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    SubscriptBox[
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"A", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "x"}], 
      TagBox[")",
       "AutoParentheses"]}], "i"], "\[Equal]", 
    SubscriptBox["A", "i"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(element\:293a 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:585712945", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.60014948946046*^9, 3.60014961221371*^9}, 
   3.600688330485127*^9, {3.603177631558162*^9, 3.603177633513364*^9}, 
   3.606555288784369*^9, {3.615546521621216*^9, 3.61554655225701*^9}},
 CellTags->{
  "ID:585712945", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "element\:293a 1"},
 CellID->585712945],

Cell[BoxData[
 RowBox[{
  SubscriptBox[
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], 
    TagBox[")",
     "AutoParentheses"]}], 
   RowBox[{"\[LeftBracketingBar]", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "\[RightBracketingBar]"}]], "\[Equal]", 
  "x"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(element\:293a 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1075236060", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.615546457842311*^9, 3.615546485754637*^9}},
 CellTags->{
  "ID:1075236060", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "element\:293a 2"},
 CellID->1075236060],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1505428581]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->32281044],

Cell[TextData[{
 "The most important formulas are certainly ",
 ButtonBox["(element\:293a 1)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:585712945"],
 " and ",
 ButtonBox["(element\:293a 2)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1075236060"],
 ", which provide us with information about the elements of a tuple that is \
obtained by appending a new element to an existing tuple. Note that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], TraditionalForm]]],
 " is the built-in notation for appending ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " to tuple ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ". Of course, we could also invent our own notation, but there is no need to \
do so."
}], "Text",
 CellChangeTimes->{{3.60360153248506*^9, 3.603601756866946*^9}, {
  3.605435797605108*^9, 3.605435918897891*^9}, {3.615022353021536*^9, 
  3.615022399836042*^9}, {3.61502488721291*^9, 3.615024892873213*^9}, {
  3.615546597220676*^9, 3.615546659844847*^9}},
 CellID->1869820165],

Cell[TextData[{
 "We now define an auxiliary unary predicate, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[SquareSubsetEqual]", "dom"], 
    RowBox[{"[", "A", "]"}]}], TraditionalForm]]],
 ", that is parameterized over domains ",
 Cell[BoxData[
  FormBox["dom", TraditionalForm]]],
 " and which holds for all objects that are tuples consisting entirely of \
elements in ",
 Cell[BoxData[
  FormBox["dom", TraditionalForm]]],
 " and fails for all others."
}], "Text",
 CellChangeTimes->{{3.603518320593976*^9, 3.603518427526788*^9}, {
   3.60351856461245*^9, 3.603518665327748*^9}, {3.605522325621463*^9, 
   3.605522329718164*^9}, {3.605522858854578*^9, 3.605522906701475*^9}, {
   3.605524475681323*^9, 3.605524486482547*^9}, {3.605588878744955*^9, 
   3.605588915226089*^9}, {3.605589316952339*^9, 3.605589365061838*^9}, {
   3.613973678703914*^9, 3.613973683296204*^9}, 3.615022460556141*^9},
 CellID->1183041550],

Cell["", "OpenEnvironment",
 CellID->822989180],

Cell[CellGroupData[{

Cell["Definition: Tuples of Domain Elements", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.598255117295602*^9, 3.598255119626807*^9}, {
  3.599285326943449*^9, 3.599285328612805*^9}, {3.600060110065106*^9, 
  3.600060131583479*^9}, {3.600060174227421*^9, 3.600060175643571*^9}, {
  3.602391056959598*^9, 3.602391059046307*^9}, {3.605588784028585*^9, 
  3.605588804911435*^9}, {3.605588868506354*^9, 3.60558886880201*^9}, {
  3.60558991185645*^9, 3.605589921689306*^9}, {3.615021351393429*^9, 
  3.615021355785782*^9}},
 CellID->1879793312],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"dom", ",", "A"}]]], "GlobalDeclaration",
 ShowCellTags->False,
 CellChangeTimes->{{3.598255128541232*^9, 3.598255133298214*^9}, {
   3.605524471954095*^9, 3.605524472221186*^9}, 3.605588812986259*^9, {
   3.605589269503979*^9, 3.60558926975049*^9}},
 CellTags->{
  "ID:1405024037", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->1405024037],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[SquareSubsetEqual]", "dom"], 
   RowBox[{"[", "A", "]"}]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[And]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{"isTuple", "[", "A", "]"}]},
      {
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
        RowBox[{
         UnderscriptBox["\[Element]", "dom"], 
         RowBox[{"[", 
          SubscriptBox["A", "i"], "]"}]}]}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(domain tuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:379346204", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.598255139210982*^9, 3.598255162821805*^9}, {
  3.598329687443484*^9, 3.598329696540591*^9}, {3.60552448935095*^9, 
  3.605524492462293*^9}, {3.605588817229036*^9, 3.605588825575322*^9}, {
  3.605589259316634*^9, 3.605589277453253*^9}, {3.615022534478046*^9, 
  3.615022535447611*^9}},
 CellTags->{
  "ID:379346204", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "domain tuple"},
 CellID->379346204],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1742211894]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->615285686]
}, Closed]],

Cell[CellGroupData[{

Cell["Other Coordinate", "Subsubsection",
 CellChangeTimes->{{3.615021551081392*^9, 3.615021554081986*^9}},
 CellTags->"sect::DefinitionOC",
 CellID->157546407],

Cell[TextData[{
 "The next basic notions we introduce is the notion of \
\[OpenCurlyDoubleQuote]other coordinate\[CloseCurlyDoubleQuote]: When working \
in a two-dimensional space and considering one of the two coordinate axes, \
\[OpenCurlyDoubleQuote]other coordinate\[CloseCurlyDoubleQuote] (denoted as \
",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[EmptySquare]", "-"], TraditionalForm]]],
 ") refers to, as its name suggests, the \[OpenCurlyDoubleQuote]other\
\[CloseCurlyDoubleQuote] coordinate axis."
}], "Text",
 CellChangeTimes->{{3.615022572295966*^9, 3.615022698150144*^9}},
 CellID->2005240008],

Cell["", "OpenEnvironment",
 CellID->205551855],

Cell[CellGroupData[{

Cell["Definition: Other Coordinate", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.612239889850224*^9, 3.612239898107588*^9}, {
  3.612240441029197*^9, 3.612240441545729*^9}},
 CellID->868169721],

Cell[BoxData[
 RowBox[{
  SuperscriptBox["1", "-"], ":=", "2"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(other coordinate 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1473378407", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612239916209355*^9, 3.612239923334791*^9}},
 CellTags->{
  "ID:1473378407", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "other coordinate 1"},
 CellID->1473378407],

Cell[BoxData[
 RowBox[{
  SuperscriptBox["2", "-"], ":=", "1"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(other coordinate 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:406630295", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612239929005285*^9, 3.612239935758437*^9}},
 CellTags->{
  "ID:406630295", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "other coordinate 2"},
 CellID->406630295],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->882180403]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1782370292],

Cell[TextData[{
 "The use of the notion \[OpenCurlyDoubleQuote]other coordinate\
\[CloseCurlyDoubleQuote] proved to be extremely convient when showing various \
lemmas/theorems in the two dimensional setting, when it allowed us to deal \
with two different cases in proofs at once (mostly in sections ",
 ButtonBox["Lemmata on Specific Notions / Special Case n=2 / W",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataSpecificSpecialW"],
 " and ",
 ButtonBox["Theorems / Special Case n=2 / Far Exterior",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::TheoremsFarExterior"],
 ")."
}], "Text",
 CellChangeTimes->{{3.615022751064433*^9, 3.615022891687719*^9}, {
  3.615022964269024*^9, 3.615023025667698*^9}, {3.61554540787541*^9, 
  3.615545464185782*^9}, {3.615545497876906*^9, 3.615545567293234*^9}, {
  3.615545608561418*^9, 3.615545608581055*^9}, {3.615545638936466*^9, 
  3.615545676066564*^9}},
 CellID->388142211]
}, Closed]],

Cell[CellGroupData[{

Cell["\[DoubleStruckCapitalZ]", "Subsubsection",
 CellChangeTimes->{{3.615021560524541*^9, 3.615021561636607*^9}},
 CellTags->"sect::DefinitionZ",
 CellID->1036621313],

Cell[TextData[{
 "The following definition summarizes some well-known properties of the \
domain of integers. In particular, formula ",
 ButtonBox["(Z is TOM)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1091916109"],
 " states that \[DoubleStruckCapitalZ] (together with ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["+", "\[DoubleStruckCapitalZ]"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   TraditionalForm]]],
 ") is a totally-ordered Abelian monoid, and formula ",
 ButtonBox["(some integers)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:492790093"],
 " states that numbers 0, 1 and 2 are integers (knowledge which is also \
available in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ", but nevertheless we state it explicitly)."
}], "Text",
 CellChangeTimes->{{3.603531350610991*^9, 3.603531430932816*^9}, {
   3.603531466266706*^9, 3.603531681388625*^9}, {3.603531755395493*^9, 
   3.603531822720273*^9}, {3.603532116770549*^9, 3.603532181821032*^9}, {
   3.603532261034047*^9, 3.603532262174339*^9}, {3.605435381058295*^9, 
   3.605435383756137*^9}, {3.612767302150604*^9, 3.612767322733393*^9}, 
   3.615021411401457*^9, {3.615021453403584*^9, 3.615021454756815*^9}, {
   3.615023043040411*^9, 3.615023148987479*^9}},
 CellID->512342355],

Cell[TextData[{
 "Please note that ",
 Cell[BoxData[
  FormBox["\[DoubleStruckCapitalZ]", TraditionalForm]]],
 " is a built-in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " domain that can immediately, without any further effort, be used for \
computations (in particular, ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["\[Element]", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " are defined in \[DoubleStruckCapitalZ]). For reasoning however, special \
inference rule are needed; They are part of the ComplexityProver. Still, the \
theory of \[DoubleStruckCapitalZ] could in principle also be built-up from \
scratch (e.g. from the Peano axioms), such that most formulas in the \
definition below would then rather be a lemma. This, however, is certainly ",
 StyleBox["not",
  FontSlant->"Italic"],
 " what we want to do here, since we want to exclusively focus on the \
complexity analysis and not on the theory of integers. In other words: We \
proceed by proving not from ",
 StyleBox["first",
  FontSlant->"Italic"],
 ", but from ",
 StyleBox["intermediate",
  FontSlant->"Italic"],
 " principles, just as what we do with tuples."
}], "Text",
 CellChangeTimes->{{3.615544947214191*^9, 3.615545052109929*^9}, {
  3.615545101254546*^9, 3.615545355877403*^9}},
 CellID->1266703458],

Cell["", "OpenEnvironment",
 CellChangeTimes->{3.615021411401611*^9},
 CellID->51464841],

Cell[CellGroupData[{

Cell["Definition: \[DoubleStruckCapitalZ]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.603168862807876*^9, 3.603168867363455*^9}, {
  3.612066121050234*^9, 3.61206612714745*^9}, {3.615021381519373*^9, 
  3.61502141140173*^9}},
 CellID->2104076581],

Cell[BoxData[
 RowBox[{"isTOM", "[", "\[DoubleStruckCapitalZ]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Z is TOM)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1091916109", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600150032417633*^9, 3.60015005079494*^9}, {
   3.600409114928145*^9, 3.600409120595419*^9}, {3.60040934542344*^9, 
   3.600409347752078*^9}, {3.603177045602664*^9, 3.603177049527965*^9}, 
   3.615021411402028*^9},
 CellTags->{
  "ID:1091916109", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Z is TOM"},
 CellID->1091916109],

Cell[BoxData[
 RowBox[{"isTotallyOrdered", "[", "\[DoubleStruckCapitalZ]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Z is totally ordered)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1845867532", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610960903417123*^9, 3.610960909821716*^9}, 
   3.615021411402328*^9},
 CellTags->{
  "ID:1845867532", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Z is totally ordered"},
 CellID->1845867532],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"[", "0", "]"}]}], "\[And]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"[", "1", "]"}]}], "\[And]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"[", "2", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(some integers)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:492790093", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6001503586599*^9, 3.600150363600155*^9}, {
   3.60040922888906*^9, 3.600409239108806*^9}, {3.600409396988456*^9, 
   3.600409419801036*^9}, {3.603177241784517*^9, 3.603177275755286*^9}, {
   3.603181355777566*^9, 3.603181363623871*^9}, 3.615021411402673*^9},
 CellTags->{
  "ID:492790093", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "some integers"},
 CellID->492790093],

Cell[BoxData[
 RowBox[{"1", 
  UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
  "2"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1\[LessEqual]2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1907318616", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610960081669835*^9, 3.610960090296792*^9}, 
   3.615021411402968*^9},
 CellTags->{
  "ID:1907318616", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1\[LessEqual]2"},
 CellID->1907318616],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y"}], "]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.603168916933564*^9, 3.603168921652525*^9}, {
   3.603168956279516*^9, 3.603168956427934*^9}, {3.603170220526243*^9, 
   3.603170242102509*^9}, {3.603170356643132*^9, 3.60317036380568*^9}, {
   3.603177060398173*^9, 3.603177066880995*^9}, 3.615021411403278*^9},
 CellID->529342921],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"y", 
    UnderscriptBox["+", "\[DoubleStruckCapitalZ]"], "1"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], "y"}], 
    "\[Or]", 
    RowBox[{"x", "\[Equal]", 
     RowBox[{"y", 
      UnderscriptBox["+", "\[DoubleStruckCapitalZ]"], "1"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1708561859", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.60285065792467*^9, 3.602850682393345*^9}, {
   3.603171943237611*^9, 3.603171962737383*^9}, {3.603181564678217*^9, 
   3.603181577981979*^9}, 3.615021411403581*^9},
 CellTags->{
  "ID:1708561859", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1"},
 CellID->1708561859],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"x", 
    UnderscriptBox["+", "\[DoubleStruckCapitalZ]"], "1"}], 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], "y"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"x", 
   UnderscriptBox["<", "\[DoubleStruckCapitalZ]"], 
   "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:49894867", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.603181635536399*^9, 3.603181650164901*^9}, {
   3.603601060843854*^9, 3.603601074655373*^9}, {3.606118714235499*^9, 
   3.606118722416922*^9}, 3.615021411403899*^9},
 CellTags->{
  "ID:49894867", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "2"},
 CellID->49894867],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellChangeTimes->{3.615021411404101*^9},
 CellID->786201735]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellChangeTimes->{3.615021411404178*^9},
 CellID->712329467]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Categories", "Subsection",
 CellChangeTimes->{{3.598252889962826*^9, 3.59825289187517*^9}},
 CellTags->"sect::DefinitionCategories",
 CellID->1469832235],

Cell["\<\
In this subsection we introduce some of the categories of domains we are \
going to use in the rest of this document.\
\>", "Text",
 CellChangeTimes->{{3.615023387338606*^9, 3.615023434225503*^9}, {
  3.615172587329401*^9, 3.615172588271628*^9}},
 CellID->1067846941],

Cell[TextData[{
 "The basic category in the formalized theory is the category of ",
 StyleBox["totally-ordered Abelian monoids",
  FontSlant->"Italic"],
 ". This category plays the role of the category the exponents in the \
polynomials need to belong to. This means, in particular, that we do not \
restrict ourselves to the case where the exponents have to be natural numbers \
(as it ",
 StyleBox["has to",
  FontSlant->"Italic"],
 " be the case in polynomial rings, strictly speaking), but we add some \
degree of freedom and are still able to carry out the complexity analysis \
just as in the referenced literature. Note, however, that for establishing \
the link between ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["B", "F"], TraditionalForm]]],
 " and the number of elementary operations in ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ", additional, much stronger requirements have to be fulfiled by the \
exponents, such that in fact only ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 " makes sense there."
}], "Text",
 CellChangeTimes->{{3.603527321595573*^9, 3.603527431677319*^9}, {
  3.60403693637464*^9, 3.604036936418842*^9}, {3.61502350855917*^9, 
  3.615023541749502*^9}, {3.617952988998849*^9, 3.617952990721267*^9}, {
  3.617953150062771*^9, 3.617953151324632*^9}},
 CellID->1228146070],

Cell[CellGroupData[{

Cell["Abelian Monoid", "Subsubsection",
 CellChangeTimes->{{3.615021764016139*^9, 3.615021768224817*^9}, {
  3.6179529807124*^9, 3.617952982511403*^9}, {3.617953143127069*^9, 
  3.617953144349245*^9}},
 CellTags->"sect::DefinitionAbelianMonoid",
 CellID->1828699765],

Cell[TextData[{
 "Before we can give the definition of totally-ordered Abelian monoids we \
have to define the category of commutative monoids, which are monoids with \
commutative monoid operation. But we require even more: They should also have \
the so-called cancellation property, stated in the last conjunct of formula \
",
 ButtonBox["(Abelian monoid)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1003575752"],
 ":"
}], "Text",
 CellChangeTimes->{{3.603527476309427*^9, 3.603527485065909*^9}, {
  3.603527539448496*^9, 3.603527578554198*^9}, {3.603527614274646*^9, 
  3.60352769857466*^9}, {3.615024680973806*^9, 3.615024712147246*^9}, {
  3.617953003004023*^9, 3.617953009197401*^9}, {3.617953139025039*^9, 
  3.617953140392312*^9}},
 CellID->337701861],

Cell["", "OpenEnvironment",
 CellID->1763006174],

Cell[CellGroupData[{

Cell["Definition: Defining Properties of Abelian Monoids", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.615549483984002*^9, 3.615549498258268*^9}},
 CellID->1874281841],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "dom"]], "GlobalDeclaration",
 CellChangeTimes->{{3.615549508723186*^9, 3.615549509795706*^9}},
 CellID->1405235777],

Cell[BoxData[
 RowBox[{
  RowBox[{"isClosed", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "dom"], 
     RowBox[{"[", 
      RowBox[{"x", ",", "y"}], "]"}]}]], 
   RowBox[{
    UnderscriptBox["\[Element]", "dom"], 
    RowBox[{"[", 
     RowBox[{"x", 
      UnderscriptBox["+", "dom"], "y"}], 
     "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(closure)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:349901806", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61554938049811*^9, 3.615549391020267*^9}},
 CellTags->{
  "ID:349901806", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "closure"},
 CellID->349901806],

Cell[BoxData[
 RowBox[{
  RowBox[{"containsZero", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{"(", 
   RowBox[{
    UnderscriptBox["\[Element]", "dom"], 
    RowBox[{"[", 
     UnderscriptBox["0", "dom"], "]"}]}], ")"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 in domain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1476372605", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61554930660096*^9, 3.615549326330696*^9}, {
  3.615549362624708*^9, 3.615549364249804*^9}},
 CellTags->{
  "ID:1476372605", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "0 in domain"},
 CellID->1476372605],

Cell[BoxData[
 RowBox[{
  RowBox[{"isZeroNeutral", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "dom"], 
     RowBox[{"[", "x", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"x", 
      UnderscriptBox["+", "dom"], 
      UnderscriptBox["0", "dom"]}], "\[Equal]", "x"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 is neutral)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1359514637", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.615549428869174*^9, {3.615549526904043*^9, 3.615549542693474*^9}},
 CellTags->{
  "ID:1359514637", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "0 is neutral"},
 CellID->1359514637],

Cell[BoxData[
 RowBox[{
  RowBox[{"isAssociative", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "dom"], 
     RowBox[{"[", 
      RowBox[{"x", ",", "y", ",", "z"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"x", 
        UnderscriptBox["+", "dom"], "y"}], ")"}], 
      UnderscriptBox["+", "dom"], "z"}], "\[Equal]", 
     RowBox[{"x", 
      UnderscriptBox["+", "dom"], 
      RowBox[{"(", 
       RowBox[{"y", 
        UnderscriptBox["+", "dom"], "z"}], ")"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(associativity)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1215514258", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.602999533965962*^9, 3.602999549756949*^9}, {
  3.605524681452107*^9, 3.605524689260022*^9}},
 CellTags->{
  "ID:1215514258", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "associativity"},
 CellID->1215514258],

Cell[BoxData[
 RowBox[{
  RowBox[{"isCommutative", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "dom"], 
     RowBox[{"[", 
      RowBox[{"x", ",", "y"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"x", 
      UnderscriptBox["+", "dom"], "y"}], "\[Equal]", 
     RowBox[{"y", 
      UnderscriptBox["+", "dom"], "x"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(commutativity)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1531751179", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.602999578359386*^9, 3.6029995903254*^9}, {
  3.605524682302273*^9, 3.605524693655343*^9}},
 CellTags->{
  "ID:1531751179", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "commutativity"},
 CellID->1531751179],

Cell[BoxData[
 RowBox[{
  RowBox[{"isCancellative", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "dom"], 
     RowBox[{"[", 
      RowBox[{"x", ",", "y", ",", "z"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      RowBox[{"z", 
       UnderscriptBox["+", "dom"], "x"}], "\[Equal]", 
      RowBox[{"z", 
       UnderscriptBox["+", "dom"], "y"}]}], 
     TagBox["\[DoubleLeftRightArrow]",
      Identity,
      SyntaxForm->"a\[Implies]b"], 
     RowBox[{"x", "\[Equal]", "y"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cancellativity)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:929558765", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.615549223938428*^9, 3.615549238310483*^9}},
 CellTags->{
  "ID:929558765", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "cancellativity"},
 CellID->929558765],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->42267126]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1814682497],

Cell["", "OpenEnvironment",
 CellID->617751546],

Cell[CellGroupData[{

Cell["Definition: Commutative Monoid", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.598253960663223*^9, 3.598253962232559*^9}, {
  3.598255590908547*^9, 3.598255591800125*^9}, {3.598596991340704*^9, 
  3.598596994269573*^9}, {3.617953071892066*^9, 3.617953074805615*^9}},
 CellID->1751978973],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "dom"]], "GlobalDeclaration",
 CellChangeTimes->{{3.598253970614262*^9, 3.598253971880306*^9}, {
  3.6055246535678368`*^9, 3.605524653800086*^9}},
 CellID->996748630],

Cell[BoxData[
 RowBox[{
  RowBox[{"isAbelianMonoid", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[And]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{"containsZero", "[", "dom", "]"}]},
      {
       RowBox[{"isClosed", "[", "dom", "]"}]},
      {
       RowBox[{"isAssociative", "[", "dom", "]"}]},
      {
       RowBox[{"isCommutative", "[", "dom", "]"}]},
      {
       RowBox[{"isZeroNeutral", "[", "dom", "]"}]},
      {
       RowBox[{"isCancellative", "[", "dom", "]"}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Abelian monoid)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1003575752", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.598253979497776*^9, 3.598254179953555*^9}, {
   3.59825559526801*^9, 3.598255596456258*^9}, {3.598597008705199*^9, 
   3.598597010881333*^9}, {3.598597085235368*^9, 3.598597172640237*^9}, {
   3.598605523010841*^9, 3.598605526189298*^9}, {3.599196902099993*^9, 
   3.599196911389072*^9}, {3.599196966706184*^9, 3.599196966890859*^9}, {
   3.59989989246608*^9, 3.599899948435514*^9}, {3.599904781286143*^9, 
   3.599904781759873*^9}, {3.600408017762585*^9, 3.60040807833292*^9}, {
   3.602933756349681*^9, 3.602933903727903*^9}, {3.602999547727958*^9, 
   3.602999597259768*^9}, {3.605524656045936*^9, 3.605524678687199*^9}, {
   3.605943550473692*^9, 3.605943564049535*^9}, {3.615549236060283*^9, 
   3.615549245490331*^9}, {3.615549330384632*^9, 3.615549333874799*^9}, {
   3.615549393902568*^9, 3.615549396580649*^9}, 3.615549525169647*^9, {
   3.61554955905454*^9, 3.615549563258343*^9}},
 CellTags->{
  "ID:1003575752", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Abelian monoid"},
 CellID->1003575752],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->36024838]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1667893272],

Cell["\<\
In the definition above it can also be seen how categories are defined: They \
are just higher-order predicates which hold iff the argument-domain (together \
with its operations) has some properties. The formulas above should be more \
or less self-explaining.\
\>", "Text",
 CellChangeTimes->{{3.603527762920557*^9, 3.603527933619976*^9}, {
  3.605943639744809*^9, 3.605943644303275*^9}},
 CellID->1636890686],

Cell[TextData[StyleBox["Remark:",
 FontFamily->"Bitstream Charter",
 FontSize->14,
 FontWeight->"Bold"]], "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellChangeTimes->{{3.603522907354292*^9, 3.603522909426996*^9}, {
   3.603522955835897*^9, 3.603522963082172*^9}, {3.603793011527348*^9, 
   3.603793024231905*^9}, 3.60552479158886*^9},
 CellID->1203490576],

Cell["\<\
Monoids are always written additively throughout this report, since they \
reflect the monoid of natural numbers together with addition (coming from the \
 exponent vectors of polynomials).\
\>", "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellChangeTimes->{{3.603522955835983*^9, 3.603523016264877*^9}, {
   3.603523067993002*^9, 3.603523068774317*^9}, {3.603526511396674*^9, 
   3.603526537459005*^9}, {3.603526632531704*^9, 3.603526636839904*^9}, {
   3.603526720792316*^9, 3.60352672617742*^9}, {3.60352675821136*^9, 
   3.603526794949112*^9}, {3.603526838114751*^9, 3.60352684115434*^9}, {
   3.603526877383081*^9, 3.603527143643606*^9}, {3.603527173960989*^9, 
   3.603527187399389*^9}, {3.603793026985894*^9, 3.603793319090968*^9}, {
   3.603793361177027*^9, 3.60379351484412*^9}, 3.604035104537365*^9, {
   3.604035136662751*^9, 3.604035137669769*^9}, {3.604834858331772*^9, 
   3.604834958512149*^9}, {3.605524793469748*^9, 3.605524860399844*^9}, {
   3.61502362714648*^9, 3.615023633337579*^9}},
 CellID->268251286]
}, Closed]],

Cell[CellGroupData[{

Cell["Totally-Ordered Domain", "Subsubsection",
 CellChangeTimes->{{3.615021785084461*^9, 3.615021789362655*^9}},
 CellTags->"sect::DefinitionTotallyOrderedDomain",
 CellID->898592553],

Cell[TextData[{
 "Before we can introduce the category of totally-ordered domains, we first \
need to consider a definition that is concerned with properties of binary \
relations. More precisely, the notions of ",
 StyleBox["total ordering",
  FontSlant->"Italic"],
 ", ",
 StyleBox["irreflexive relation",
  FontSlant->"Italic"],
 " and ",
 StyleBox["strict version of relations",
  FontSlant->"Italic"],
 " are introduced."
}], "Text",
 CellChangeTimes->{{3.603519699150366*^9, 3.603519778054931*^9}, {
  3.60369735353982*^9, 3.603697367046468*^9}, {3.605590576866514*^9, 
  3.605590577364658*^9}, {3.615023650865967*^9, 3.615023709307591*^9}},
 CellID->1503249513],

Cell["", "OpenEnvironment",
 CellID->833748108],

Cell[CellGroupData[{

Cell["Definition: Orderings", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.598695652948873*^9, 3.598695662115955*^9}, {
  3.600060021107348*^9, 3.600060022813954*^9}},
 CellID->567790588],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox["\[Precedes]",
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox["\[PrecedesEqual]",
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "dom"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.598695877092063*^9, 3.598695916512196*^9}, {
  3.605524536640659*^9, 3.605524536865854*^9}},
 CellID->1996046852],

Cell[BoxData[
 RowBox[{
  RowBox[{"isTotalOrdering", "[", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[PrecedesEqual]", "dom"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[And]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "dom"], 
          RowBox[{"[", 
           RowBox[{"x", ",", "y"}], "]"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            RowBox[{"x", 
             UnderscriptBox["\[PrecedesEqual]", "dom"], "y"}], "\[And]", 
            RowBox[{"y", 
             UnderscriptBox["\[PrecedesEqual]", "dom"], "x"}]}], 
           TagBox[")",
            "AutoParentheses"]}], "\[Implies]", 
          RowBox[{"x", "\[Equal]", "y"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}]},
      {
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "dom"], 
          RowBox[{"[", 
           RowBox[{"x", ",", "y"}], "]"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{"x", 
           UnderscriptBox["\[PrecedesEqual]", "dom"], "y"}], "\[Or]", 
          RowBox[{"y", 
           UnderscriptBox["\[PrecedesEqual]", "dom"], "x"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}]},
      {
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "dom"], 
          RowBox[{"[", 
           RowBox[{"x", ",", "y", ",", "z"}], "]"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            RowBox[{"x", 
             UnderscriptBox["\[PrecedesEqual]", "dom"], "y"}], "\[And]", 
            RowBox[{"y", 
             UnderscriptBox["\[PrecedesEqual]", "dom"], "z"}]}], 
           TagBox[")",
            "AutoParentheses"]}], "\[Implies]", 
          RowBox[{"x", 
           UnderscriptBox["\[PrecedesEqual]", "dom"], "z"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(total ordering)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1939013113", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.60006003526745*^9, {3.600406986171307*^9, 3.600407002872435*^9}, {
   3.600407229608044*^9, 3.600407275455475*^9}, {3.602933283899007*^9, 
   3.602933294252884*^9}, {3.602933607974339*^9, 3.60293369324444*^9}, {
   3.602935218568836*^9, 3.602935250100204*^9}, {3.605524565279483*^9, 
   3.605524592272715*^9}},
 CellTags->{
  "ID:1939013113", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "total ordering"},
 CellID->1939013113],

Cell[BoxData[
 RowBox[{
  RowBox[{"isIrreflexive", "[", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Precedes]", "dom"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "dom"], 
     RowBox[{"[", "x", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"\[Not]", 
     RowBox[{"x", 
      UnderscriptBox["\[Precedes]", "dom"], "x"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(irreflexive relation)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1585090210", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600060311367899*^9, 3.600060341900635*^9}, {
   3.600407270939987*^9, 3.600407280540092*^9}, 3.602933298103187*^9, {
   3.602933329979876*^9, 3.602933338952657*^9}, {3.605524594511367*^9, 
   3.60552459973064*^9}},
 CellTags->{
  "ID:1585090210", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "irreflexive relation"},
 CellID->1585090210],

Cell[BoxData[
 RowBox[{
  RowBox[{"isStrictVersionOf", "[", 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      UnderscriptBox["\[Precedes]", "dom"],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}], ",", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      UnderscriptBox["\[PrecedesEqual]", "dom"],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}]}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "dom"], 
     RowBox[{"[", 
      RowBox[{"x", ",", "y"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"x", 
      UnderscriptBox["\[Precedes]", "dom"], "y"}], 
     TagBox["\[DoubleLeftRightArrow]",
      Identity,
      SyntaxForm->"a\[Implies]b"], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"x", 
        UnderscriptBox["\[PrecedesEqual]", "dom"], "y"}], "\[And]", 
       RowBox[{"x", "\[NotEqual]", "y"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(strict version of ordering)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1535164863", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.598695922289647*^9, 3.598696002312046*^9}, {
  3.598696166217766*^9, 3.598696171064273*^9}, {3.599195449319344*^9, 
  3.599195453652513*^9}, {3.600060052699957*^9, 3.600060062245279*^9}, {
  3.600407339760749*^9, 3.600407422109049*^9}, {3.605524601826171*^9, 
  3.605524620814635*^9}},
 CellTags->{
  "ID:1535164863", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "strict version of ordering"},
 CellID->1535164863],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->251656430]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1535644204],

Cell[TextData[{
 "Now comes the definition of the category of totally-ordered domains. \
Totally-ordered domains are domains which provide two binary relations ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "dom"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["<", "dom"], TraditionalForm]]],
 ", which are a reflexive total order relation and the strict version \
thereof, respectively."
}], "Text",
 CellChangeTimes->{{3.603528643634155*^9, 3.603528657190337*^9}, {
  3.603528689936548*^9, 3.603528760424367*^9}, {3.605524761666476*^9, 
  3.605524764470478*^9}},
 CellID->672448588],

Cell["", "OpenEnvironment",
 CellID->2106479689],

Cell[CellGroupData[{

Cell["Definition: Totally-Ordered Domain", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6004085239317*^9, 3.600408530160545*^9}},
 CellID->341504943],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "dom"]], "GlobalDeclaration",
 CellChangeTimes->{{3.600408535766183*^9, 3.60040853664133*^9}, {
  3.605524867370971*^9, 3.605524867653592*^9}},
 CellID->1430533020],

Cell[BoxData[
 RowBox[{
  RowBox[{"isTotallyOrdered", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[And]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{"isTotalOrdering", "[", 
        UnderscriptBox["\[LessEqual]", "dom"], "]"}]},
      {
       RowBox[{"isStrictVersionOf", "[", 
        RowBox[{
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox[
           UnderscriptBox["<", "dom"],
           Identity], 
          TagBox[")",
           "AutoParentheses"]}], ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox[
           UnderscriptBox["\[LessEqual]", "dom"],
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}], "]"}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(totally-ordered domain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:795999034", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600408540761958*^9, 3.600408600099212*^9}, {
  3.605524871837871*^9, 3.60552487573967*^9}},
 CellTags->{
  "ID:795999034", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "totally-ordered domain"},
 CellID->795999034],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->461069387]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1286085183]
}, Closed]],

Cell[CellGroupData[{

Cell["min-max Domain", "Subsubsection",
 CellChangeTimes->{{3.615021800997159*^9, 3.615021803561603*^9}},
 CellTags->"sect::DefinitionminmaxDomain",
 CellID->2071321112],

Cell[TextData[{
 "The next category is the category of ",
 StyleBox["min-max domains",
  FontSlant->"Italic"],
 ". A min-max domain ",
 Cell[BoxData[
  FormBox["dom", TraditionalForm]]],
 " is a domain that provides the two functions ",
 Cell[BoxData[
  FormBox["min", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["max", TraditionalForm]]],
 " which, for non-empty tuples of elements of ",
 Cell[BoxData[
  FormBox["dom", TraditionalForm]]],
 ", return a minimal (resp. maximal) element. The order relation that is used \
to determine minimal/maximal elements is ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "dom"], TraditionalForm]]],
 "; Note we do not require at this point that ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "dom"], TraditionalForm]]],
 " has ",
 StyleBox["any",
  FontSlant->"Italic"],
 " properties (not even transitivity). However, the only case of interest is \
the case where ",
 Cell[BoxData[
  FormBox["dom", TraditionalForm]]],
 " is a totally-ordered domain anyway, meaning that ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "dom"], TraditionalForm]]],
 " then automatically has all the properties one would expect for computing \
minimal/maximal elements."
}], "Text",
 CellChangeTimes->{{3.603519814339715*^9, 3.603519861033067*^9}, {
  3.603519891068969*^9, 3.60352022804147*^9}, {3.603520265593965*^9, 
  3.603520374996526*^9}, {3.603520423868919*^9, 3.603520714616445*^9}, {
  3.603520750876*^9, 3.603520751395768*^9}, {3.604908744054738*^9, 
  3.6049090667048*^9}, {3.605943681526278*^9, 3.605943704509577*^9}, {
  3.614937393563513*^9, 3.61493741342404*^9}, {3.615023873937331*^9, 
  3.615023902893838*^9}},
 CellID->1673907690],

Cell["", "OpenEnvironment",
 CellID->658511437],

Cell[CellGroupData[{

Cell["Definition: min-max Domain", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599193991165801*^9, 3.599193993771375*^9}, {
  3.6004089789869*^9, 3.60040898375914*^9}, {3.600577685817027*^9, 
  3.600577688316454*^9}, {3.604908691901627*^9, 3.604908696586916*^9}},
 CellID->807480567],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "dom"]], "GlobalDeclaration",
 CellChangeTimes->{{3.60040745068486*^9, 3.600407451544043*^9}, {
   3.600408948809799*^9, 3.60040895377191*^9}, 3.604909073515824*^9, 
   3.605524888709034*^9},
 CellID->1209270974],

Cell[BoxData[
 RowBox[{
  RowBox[{"isMinMaxDomain", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"isMinDomain", "[", "dom", "]"}], "\[And]", 
    RowBox[{"isMaxDomain", "[", "dom", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min-max domain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:616702295", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599995913005111*^9, 3.599995915588951*^9}, {
  3.599996431734451*^9, 3.599996493991047*^9}, {3.600060199577071*^9, 
  3.600060215944807*^9}, {3.600406078356189*^9, 3.600406088254645*^9}, {
  3.600407539754052*^9, 3.600407549072247*^9}, {3.600407582989886*^9, 
  3.600407610812663*^9}, {3.600408961262689*^9, 3.60040896777956*^9}, {
  3.60057651054235*^9, 3.600576523833483*^9}, {3.604909256696003*^9, 
  3.604909361671533*^9}, {3.604909398046337*^9, 3.604909550179476*^9}, {
  3.604993077555138*^9, 3.604993080489887*^9}, {3.605342938628413*^9, 
  3.605342939826725*^9}, {3.605524890483132*^9, 3.605524911279392*^9}, {
  3.61088180543761*^9, 3.610881820885409*^9}},
 CellTags->{
  "ID:616702295", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min-max domain"},
 CellID->616702295],

Cell[BoxData[
 RowBox[{
  RowBox[{"isMinDomain", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[SquareSubsetEqual]", "dom"], 
      RowBox[{"[", "A", "]"}]}]], 
    RowBox[{"1", 
     UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["min", "dom"], "[", "A", "]"}], "\[Equal]", 
     SubscriptBox["A", 
      RowBox[{
       UnderscriptBox[
        SubscriptBox["argmin", "dom"], 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
       SubscriptBox["A", "i"]}]]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min domain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1204372660", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610881747033793*^9, 3.610881755884834*^9}, {
  3.610950105637539*^9, 3.61095015712001*^9}, {3.61459169741257*^9, 
  3.614591747563948*^9}, {3.615546697390451*^9, 3.61554674708551*^9}},
 CellTags->{
  "ID:1204372660", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min domain"},
 CellID->1204372660],

Cell[BoxData[
 RowBox[{
  RowBox[{"isMaxDomain", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[SquareSubsetEqual]", "dom"], 
      RowBox[{"[", "A", "]"}]}]], 
    RowBox[{"1", 
     UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["max", "dom"], "[", "A", "]"}], "\[Equal]", 
     SubscriptBox["A", 
      RowBox[{
       UnderscriptBox[
        SubscriptBox["argmax", "dom"], 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
       SubscriptBox["A", "i"]}]]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max domain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1875462394", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610881784666488*^9, 3.610881792963927*^9}, {
  3.610952996315865*^9, 3.610953015459004*^9}, {3.610953054211278*^9, 
  3.610953061925075*^9}, {3.614591766200618*^9, 3.614591779025073*^9}, {
  3.615546710422171*^9, 3.615546761832616*^9}},
 CellTags->{
  "ID:1875462394", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max domain"},
 CellID->1875462394],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->856781492]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1692376529],

Cell[TextData[{
 "Although it might appear a bit strange at this point, but splitting the \
definition of min-max domains into two parts turned out to be very convenient \
when it comes to proving: One does not have to mess around with knowledge \
(either about ",
 Cell[BoxData[
  FormBox["min", TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox["max", TraditionalForm]]],
 ") one does not need."
}], "Text",
 CellChangeTimes->{{3.610881841443992*^9, 3.610881911851236*^9}, {
  3.615023939424932*^9, 3.615023962266149*^9}},
 CellID->413114653],

Cell[TextData[StyleBox["Remark:",
 FontFamily->"Bitstream Charter",
 FontSize->14,
 FontWeight->"Bold"]], "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellChangeTimes->{{3.603522907354292*^9, 3.603522909426996*^9}, {
  3.603522955835897*^9, 3.603522963082172*^9}},
 CellID->1691893048],

Cell[TextData[{
 "Please note that here we use the ",
 Cell[BoxData[
  FormBox["argmin", TraditionalForm]]],
 " (and ",
 Cell[BoxData[
  FormBox["argmax", TraditionalForm]]],
 ") quantifier with a slightly modified syntax compared to ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 "\[CloseCurlyQuote]s built-in ",
 Cell[BoxData[
  FormBox["argmin", TraditionalForm]]],
 "/",
 Cell[BoxData[
  FormBox["argmax", TraditionalForm]]],
 " quantifiers: Instead of passing the order relation, ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["argmin", 
    UnderscriptBox["\[LessEqual]", "dom"]], TraditionalForm]]],
 ", for the sake of brevity we only pass the domain, i.e. ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["argmin", "dom"], TraditionalForm]]],
 ". This does not work in computations with the built-in definitions of ",
 Cell[BoxData[
  FormBox["argmin", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["argmax", TraditionalForm]]],
 ", but the inference rules we created for ",
 Cell[BoxData[
  FormBox["argmin", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["argmax", TraditionalForm]]],
 " (see section ",
 ButtonBox["Prover",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Prover"],
 ") are fine with it."
}], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellChangeTimes->{{3.603522955835983*^9, 3.603523016264877*^9}, {
   3.603523067993002*^9, 3.603523068774317*^9}, {3.603526511396674*^9, 
   3.603526537459005*^9}, {3.603526632531704*^9, 3.603526636839904*^9}, {
   3.603526720792316*^9, 3.60352672617742*^9}, {3.60352675821136*^9, 
   3.603526794949112*^9}, {3.603526838114751*^9, 3.60352684115434*^9}, {
   3.603526877383081*^9, 3.603527143643606*^9}, {3.603527173960989*^9, 
   3.603527187399389*^9}, {3.603527951768331*^9, 3.603528070301583*^9}, {
   3.60352815324365*^9, 3.60352822178133*^9}, {3.603528388840451*^9, 
   3.603528588399781*^9}, {3.603528806263065*^9, 3.603528808684593*^9}, {
   3.604909801742781*^9, 3.604909897099116*^9}, {3.604909928525638*^9, 
   3.60491011042773*^9}, {3.604910143666234*^9, 3.604910202470122*^9}, {
   3.605524915815679*^9, 3.605524936372402*^9}, {3.611034582514166*^9, 
   3.611034986372365*^9}, 3.613963197695136*^9, {3.614591813820818*^9, 
   3.614591848432581*^9}, {3.615023995688081*^9, 3.615024005981436*^9}},
 CellID->1732619610]
}, Closed]],

Cell[CellGroupData[{

Cell["Totally-Ordered Abelian Monoid", "Subsubsection",
 CellChangeTimes->{{3.615021812692854*^9, 3.615021819055861*^9}, {
  3.615105886005133*^9, 3.615105887040868*^9}},
 CellTags->"sect::DefinitionTOM",
 CellID->991805917],

Cell[TextData[{
 "Finally we are able to define the category of totally-ordered Abelian \
monoids. A domain ",
 Cell[BoxData[
  FormBox["dom", TraditionalForm]]],
 " is a totally-ordered Abelian monoid iff it is an Abelian monoid according \
to formula ",
 ButtonBox["(Abelian monoid)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1003575752"],
 ", a totally-ordered domain according to formula ",
 ButtonBox["(totally-ordered domain)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:795999034"],
 ", a min-max domain according to formula ",
 ButtonBox["(min-max domain)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:616702295"],
 ", and the order relation ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "dom"], TraditionalForm]]],
 " is monotonic w.r.t. the monoid operation ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["+", "dom"], TraditionalForm]]],
 " (moreover, in totally-ordered Abelian monoids also the direction from \
right to left holds in formula ",
 ButtonBox["(monotonicity)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1329821571"],
 ", a fact which is stated and proved in formula ",
 ButtonBox["(\[LessEqual]+)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:786186036"],
 "):"
}], "Text",
 CellChangeTimes->{{3.603528829894397*^9, 3.603528990428434*^9}, {
  3.605524940585002*^9, 3.605524950479413*^9}, {3.614591544179404*^9, 
  3.614591544369055*^9}, {3.615024745435039*^9, 3.615024811110687*^9}, {
  3.615610392090702*^9, 3.615610476368237*^9}, {3.615632761576789*^9, 
  3.615632879093374*^9}},
 CellID->1505273851],

Cell["", "OpenEnvironment",
 CellID->2123192141],

Cell[CellGroupData[{

Cell["Definition: Totally-Ordered Abelian Monoid", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.598253860564973*^9, 3.59825387011476*^9}, {
  3.598255370190875*^9, 3.59825537044961*^9}, {3.598255598770993*^9, 
  3.598255599885881*^9}, {3.598597028181129*^9, 3.598597029853375*^9}, {
  3.599906731572495*^9, 3.599906732774659*^9}},
 CellID->1445007926],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "dom"]], "GlobalDeclaration",
 CellChangeTimes->{{3.598253881330652*^9, 3.59825388370418*^9}, 
   3.60552495311556*^9},
 CellID->432239872],

Cell[BoxData[
 RowBox[{
  RowBox[{"isTOM", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[And]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{"isAbelianMonoid", "[", "dom", "]"}]},
      {
       RowBox[{"isTotallyOrdered", "[", "dom", "]"}]},
      {
       RowBox[{"isMinMaxDomain", "[", "dom", "]"}]},
      {
       RowBox[{"hasMonotonicOrdering", "[", "dom", "]"}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(totally-ordered Abelian monoid)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:743220188", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.598253888134595*^9, 3.598253909641531*^9}, {
   3.598253993068697*^9, 3.59825399811398*^9}, {3.598254216541358*^9, 
   3.598254219710068*^9}, {3.598254541717082*^9, 3.598254548809966*^9}, {
   3.598254644227555*^9, 3.598254685164217*^9}, 3.598254777073554*^9, {
   3.598254943378035*^9, 3.598254950161112*^9}, {3.598255375683733*^9, 
   3.598255477562723*^9}, {3.598255602735909*^9, 3.598255611819728*^9}, {
   3.598255681876439*^9, 3.598255686643885*^9}, {3.598255768191167*^9, 
   3.598255905724051*^9}, {3.598256036768957*^9, 3.598256042708593*^9}, 
   3.598329726067989*^9, {3.598593450067157*^9, 3.598593465305093*^9}, {
   3.598593559879511*^9, 3.598593574658772*^9}, {3.598594114230049*^9, 
   3.598594144652056*^9}, {3.598594412035933*^9, 3.598594429617016*^9}, {
   3.598597253804167*^9, 3.598597294929719*^9}, {3.598597379724791*^9, 
   3.598597425633253*^9}, {3.598597575962365*^9, 3.598597633586692*^9}, {
   3.598597731406561*^9, 3.598597817375373*^9}, {3.598598088243889*^9, 
   3.598598148750921*^9}, {3.598598209392595*^9, 3.598598221238372*^9}, {
   3.598598342787772*^9, 3.598598361657108*^9}, {3.598598550025016*^9, 
   3.59859855024072*^9}, {3.598603803682718*^9, 3.598603827992454*^9}, {
   3.59860386376616*^9, 3.598603893120115*^9}, {3.598604044361339*^9, 
   3.598604044632076*^9}, {3.598604332474951*^9, 3.598604336962802*^9}, {
   3.59860439811068*^9, 3.598604449015375*^9}, {3.598605409917865*^9, 
   3.598605430796042*^9}, {3.598605530126302*^9, 3.598605531831988*^9}, {
   3.598696265368626*^9, 3.59869626628768*^9}, 3.598696345058401*^9, {
   3.598696377839555*^9, 3.598696388486182*^9}, {3.59919704213236*^9, 
   3.599197082925092*^9}, {3.599197143083377*^9, 3.599197145106888*^9}, {
   3.599209219722213*^9, 3.599209258279505*^9}, {3.599285425246206*^9, 
   3.599285451043874*^9}, {3.599286560301857*^9, 3.599286568807763*^9}, {
   3.599286727760451*^9, 3.599286731830497*^9}, {3.599906726837318*^9, 
   3.599906743005131*^9}, {3.599906783704684*^9, 3.599906839548179*^9}, {
   3.599906953460352*^9, 3.599906953893817*^9}, {3.599907133157881*^9, 
   3.599907144050489*^9}, {3.600059853161388*^9, 3.600059929201029*^9}, {
   3.600060369633391*^9, 3.600060377976447*^9}, {3.600407781514826*^9, 
   3.60040780921018*^9}, {3.600407876125204*^9, 3.600407880017611*^9}, {
   3.600408609567131*^9, 3.600408618648349*^9}, {3.601182323149107*^9, 
   3.601182355575942*^9}, {3.602999165778915*^9, 3.602999236590525*^9}, {
   3.604909638811403*^9, 3.604909645408158*^9}, {3.605524954775648*^9, 
   3.605524964068521*^9}, {3.615549636661336*^9, 3.615549672503047*^9}},
 CellTags->{
  "ID:743220188", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "totally-ordered Abelian monoid"},
 CellID->743220188],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasMonotonicOrdering", "[", "dom", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "dom"], 
     RowBox[{"[", 
      RowBox[{"x", ",", "y", ",", "z"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"x", 
      UnderscriptBox["\[LessEqual]", "dom"], "y"}], "\[Implies]", 
     RowBox[{
      RowBox[{"z", 
       UnderscriptBox["+", "dom"], "x"}], 
      UnderscriptBox["\[LessEqual]", "dom"], 
      RowBox[{"z", 
       UnderscriptBox["+", "dom"], "y"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(monotonicity)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1329821571", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.615549633358275*^9, 3.615549663295717*^9}, {
  3.615610269992379*^9, 3.615610303664557*^9}},
 CellTags->{
  "ID:1329821571", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "monotonicity"},
 CellID->1329821571],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->270602909]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1275160981],

Cell[TextData[{
 "In the following paragraphs, we summarize some interesting facts about \
totally-ordered Abelian monoids. Since none of these facts plays a role in \
the complexity analysis carried out in this report, all the proofs are given \
\[OpenCurlyDoubleQuote]semi-formally\[CloseCurlyDoubleQuote] without the \
support of the ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " system."
}], "Text",
 CellChangeTimes->{{3.615691326809879*^9, 3.615691442223566*^9}, {
  3.61569820738743*^9, 3.615698232953787*^9}},
 CellID->733533220],

Cell[TextData[{
 "As a starting point, it is straightforward to see that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"\[DoubleStruckCapitalZ]", ",", 
     RowBox[{"+", 
      RowBox[{",", "\[LessEqual]"}]}]}], ")"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"\[DoubleStruckCapitalQ]", ",", 
     RowBox[{"+", 
      RowBox[{",", "\[LessEqual]"}]}]}], ")"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"\[DoubleStruckCapitalR]", ",", 
     RowBox[{"+", 
      RowBox[{",", "\[LessEqual]"}]}]}], ")"}], TraditionalForm]]],
 " are totally-ordered Abelian monoids, if one takes addition as monoid \
operation and the standard ordering as order relation; The proof of this \
statement is omitted here."
}], "Text",
 CellChangeTimes->{{3.615691464542117*^9, 3.615691649794759*^9}},
 CellID->1975846358],

Cell[TextData[{
 "Now we proceed with two so-called ",
 StyleBox["conservation theorems",
  FontSlant->"Italic"],
 ". A conservation theorem is a theorem of the form \[OpenCurlyDoubleQuote]If \
",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " is a domain in some category ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[ScriptCapitalC]", "1"], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is a functor, then ",
 Cell[BoxData[
  FormBox[
   RowBox[{"F", "[", "D", "]"}], TraditionalForm]]],
 " is in category ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[ScriptCapitalC]", "2"], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote]. In our case, the roles of both ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[ScriptCapitalC]", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[ScriptCapitalC]", "2"], TraditionalForm]]],
 " are taken on by totally-ordered Abelian monoids, and the functors we are \
interested in are the Cartesian-product functor and the polynomial functor.\n\
Therefore, assume we are given a totally-ordered Abelian monoid ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"D", ",", 
     RowBox[{"+", 
      RowBox[{",", "\[LessEqual]"}]}]}], ")"}], TraditionalForm]]],
 ". The claim is that also ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox["D", "n"], ",", 
     RowBox[{
      SubscriptBox["+", "c"], 
      RowBox[{",", "\[PrecedesEqual]"}]}]}], ")"}], TraditionalForm]]],
 " is a totally-ordered Abelian monoid, where ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["+", "c"], TraditionalForm]]],
 " is component-wise addition and ",
 Cell[BoxData[
  FormBox["\[PrecedesEqual]", TraditionalForm]]],
 " is any total order relation relation on ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["D", "n"], TraditionalForm]]],
 " which respects monotonicity w.r.t. ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["+", "c"], TraditionalForm]]],
 ". Examples of such order relations are the purely lexicographic as well as \
the degree-lexicographic ordering; Moreover, every order relation that in \
some sense corresponds to an admissible order relation on the set of terms in \
",
 Cell[BoxData[
  FormBox[
   RowBox[{"K", "[", 
    RowBox[{
     SubscriptBox["x", "1"], ",", "\[Ellipsis]", ",", 
     SubscriptBox["x", "n"]}], "]"}], TraditionalForm]]],
 " possesses the required properties.\nFirst of all note that ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["D", "n"], TraditionalForm]]],
 " is indeed an Abelian monoid: The properties of ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " being a (cancellative) Abelian monoid carry over readily to ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["D", "n"], TraditionalForm]]],
 ", as can easily be checked. So, we only show that ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[LessEqual]", "lex"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[LessEqual]", "deglex"], TraditionalForm]]],
 " are monotonic w.r.t. ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["+", "c"], TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.615691664968611*^9, 3.615691934171017*^9}, {
  3.615691972647148*^9, 3.615692174132139*^9}, {3.615692809152296*^9, 
  3.615692815359083*^9}, {3.615707100062073*^9, 3.615707109699952*^9}, {
  3.615707189995691*^9, 3.615707305481092*^9}, {3.615707420697097*^9, 
  3.615707508233678*^9}},
 CellID->1496861707],

Cell[TextData[StyleBox["Proof.",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.615693521567889*^9, 3.615693528982753*^9}},
 CellID->1979602345],

Cell[TextData[{
 "Assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        SubscriptBox["x", "1"], ",", "\[Ellipsis]", ",", 
        SubscriptBox["x", "n"]}], ")"}], 
      SubscriptBox["\[LessEqual]", "lex"], 
      RowBox[{"(", 
       RowBox[{
        SubscriptBox["y", "1"], ",", "\[Ellipsis]", ",", 
        SubscriptBox["y", "n"]}], ")"}]}], "=", "y"}]}], TraditionalForm]]],
 ", and let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"z", "=", 
    RowBox[{"(", 
     RowBox[{
      SubscriptBox["z", "1"], ",", "\[Ellipsis]", ",", 
      SubscriptBox["z", "n"]}], ")"}]}], TraditionalForm]]],
 " be arbitrary. We have to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", 
     SubscriptBox["+", "c"], "z"}], "=", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SubscriptBox["x", "1"], "+", 
         SubscriptBox["z", "1"]}], ",", "\[Ellipsis]", ",", 
        RowBox[{
         SubscriptBox["x", "n"], "+", 
         SubscriptBox["z", "n"]}]}], ")"}], 
      SubscriptBox["\[LessEqual]", "lex"], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SubscriptBox["y", "1"], "+", 
         SubscriptBox["z", "1"]}], ",", "\[Ellipsis]", ",", 
        RowBox[{
         SubscriptBox["y", "n"], "+", 
         SubscriptBox["z", "n"]}]}], ")"}]}], "=", 
     RowBox[{"y", 
      SubscriptBox["+", "c"], "z"}]}]}], TraditionalForm]]],
 ". Since ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    SubscriptBox["\[LessEqual]", "lex"], "y"}], TraditionalForm]]],
 ", we distinguish whether ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    SubscriptBox["<", "lex"], "y"}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", "y"}], TraditionalForm]]],
 "; In the latter case we are done, because ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["D", "n"], TraditionalForm]]],
 " possesses the cancellation property and we can conclude ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "+", "z"}], "=", 
    RowBox[{"y", "+", "z"}]}], TraditionalForm]]],
 ", from which ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "+", "z"}], 
    SubscriptBox["\[LessEqual]", "lex"], 
    RowBox[{"y", "+", "z"}]}], TraditionalForm]]],
 " follows immediately. Hence assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    SubscriptBox["<", "lex"], "y"}], TraditionalForm]]],
 ", which means there must be ",
 Cell[BoxData[
  FormBox[
   RowBox[{"1", "\[LessEqual]", "i", "\[LessEqual]", "n"}], TraditionalForm]]],
 " such that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["x", "1"], "=", 
    SubscriptBox["y", "1"]}], TraditionalForm]]],
 ", ..., ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["x", 
     RowBox[{"i", "-", "1"}]], "=", 
    SubscriptBox["y", 
     RowBox[{"i", "-", "1"}]]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["x", "i"], "<", 
    SubscriptBox["y", "i"]}], TraditionalForm]]],
 ". This, however, implies ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["x", "1"], "+", 
     SubscriptBox["z", "1"]}], "=", 
    RowBox[{
     SubscriptBox["y", "1"], "+", 
     SubscriptBox["z", "1"]}]}], TraditionalForm]]],
 ", ..., ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["x", 
      RowBox[{"i", "-", "1"}]], "+", 
     SubscriptBox["z", 
      RowBox[{"i", "-", "1"}]]}], "=", 
    RowBox[{
     SubscriptBox["y", 
      RowBox[{"i", "-", "1"}]], "+", 
     SubscriptBox["z", 
      RowBox[{"i", "-", "1"}]]}]}], TraditionalForm]]],
 " (using the cancellation property in ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 "). What remains to be shown is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["x", "i"], "+", 
     SubscriptBox["z", "i"]}], "<", 
    RowBox[{
     SubscriptBox["y", "i"], "+", 
     SubscriptBox["z", "i"]}]}], TraditionalForm]]],
 ", from which our proof goal follows. First of all notice that we clearly \
have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["x", "i"], "+", 
     SubscriptBox["z", "i"]}], "\[LessEqual]", 
    RowBox[{
     SubscriptBox["y", "i"], "+", 
     SubscriptBox["z", "i"]}]}], TraditionalForm]]],
 " because of monotonicity of ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ". However, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["x", "i"], "+", 
     SubscriptBox["z", "i"]}], "=", 
    RowBox[{
     SubscriptBox["y", "i"], "+", 
     SubscriptBox["z", "i"]}]}], TraditionalForm]]],
 " is not possible, for otherwise we would also have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["x", "i"], "=", 
    SubscriptBox["y", "i"]}], TraditionalForm]]],
 " by the cancellation property. Therefore, the only possibility is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["x", "i"], "+", 
     SubscriptBox["z", "i"]}], "<", 
    RowBox[{
     SubscriptBox["y", "i"], "+", 
     SubscriptBox["z", "i"]}]}], TraditionalForm]]],
 ", and we are done with showing that ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[LessEqual]", "lex"], TraditionalForm]]],
 " is monotonic w.r.t. ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["+", "c"], TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.615692206372244*^9, 3.615692438198217*^9}, {
   3.615692484165701*^9, 3.615692911459597*^9}, 3.615693513332329*^9, {
   3.62105474264926*^9, 3.621054747701182*^9}},
 CellID->1924228650],

Cell[TextData[{
 "Now assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    SubscriptBox["\[LessEqual]", "deglex"], "y"}], TraditionalForm]]],
 ". Again, using the same argument as above, if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", "y"}], TraditionalForm]]],
 " we are done. Hence, assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    SubscriptBox["<", "deglex"], "y"}], TraditionalForm]]],
 ". We distinguish two cases:\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", "x", ")"}], "<", 
    RowBox[{"deg", "(", "y", ")"}]}], TraditionalForm]]],
 ".\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", "x", ")"}], "=", 
    RowBox[{
     SubscriptBox["x", "1"], "+", "\[Ellipsis]", "+", 
     SubscriptBox["x", "n"]}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", "y", ")"}], "=", 
    RowBox[{
     SubscriptBox["y", "1"], "+", "\[Ellipsis]", "+", 
     SubscriptBox["y", "n"]}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", 
     RowBox[{"x", 
      SubscriptBox["+", "c"], "z"}], ")"}], "=", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        SubscriptBox["x", "1"], "+", 
        SubscriptBox["z", "1"]}], ")"}], "+", "\[Ellipsis]", "+", 
      RowBox[{"(", 
       RowBox[{
        SubscriptBox["x", "n"], "+", 
        SubscriptBox["z", "n"]}], ")"}]}], 
     OverscriptBox["=", 
      RowBox[{"assoc", "."}]], 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         SubscriptBox["x", "1"], "+", "\[Ellipsis]", "+", 
         SubscriptBox["x", "n"]}], ")"}], "+", 
       RowBox[{"(", 
        RowBox[{
         SubscriptBox["z", "1"], "+", "\[Ellipsis]", "+", 
         SubscriptBox["z", "n"]}], ")"}]}], "=", 
      RowBox[{
       RowBox[{"deg", "(", "x", ")"}], "+", 
       RowBox[{"deg", "(", "z", ")"}]}]}]}]}], TraditionalForm]]],
 "; Same for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", 
     RowBox[{"y", 
      SubscriptBox["+", "c"], "z"}], ")"}], "=", 
    RowBox[{
     RowBox[{"deg", "(", "y", ")"}], "+", 
     RowBox[{"deg", "(", "z", ")"}]}]}], TraditionalForm]]],
 ". Summarizing: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", "x", ")"}], "<", 
    RowBox[{"deg", "(", "y", ")"}]}], TraditionalForm]]],
 " implies, using monotonicity of ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", 
     RowBox[{"x", 
      SubscriptBox["+", "c"], "z"}], ")"}], "=", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"deg", "(", "x", ")"}], "+", 
       RowBox[{"deg", "(", "z", ")"}]}], "\[LessEqual]", 
      RowBox[{
       RowBox[{"deg", "(", "y", ")"}], "+", 
       RowBox[{"deg", "(", "z", ")"}]}]}], "=", 
     RowBox[{"deg", "(", 
      RowBox[{"y", 
       SubscriptBox["+", "c"], "z"}], ")"}]}]}], TraditionalForm]]],
 ". Equality cannot hold, for it would imply ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", "x", ")"}], "=", 
    RowBox[{"deg", "(", "y", ")"}]}], TraditionalForm]]],
 " in contradiction to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", "x", ")"}], "<", 
    RowBox[{"deg", "(", "y", ")"}]}], TraditionalForm]]],
 ", so we have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", 
     RowBox[{"x", 
      SubscriptBox["+", "c"], "z"}], ")"}], "<", 
    RowBox[{"deg", "(", 
     RowBox[{"y", 
      SubscriptBox["+", "c"], "z"}], ")"}]}], TraditionalForm]]],
 " and hence also ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", 
     SubscriptBox["+", "c"], "z"}], 
    SubscriptBox["<", "deglex"], 
    RowBox[{"y", 
     SubscriptBox["+", "c"], "z"}]}], TraditionalForm]]],
 ".\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", "x", ")"}], "=", 
    RowBox[{"deg", "(", "y", ")"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    SubscriptBox["\[LessEqual]", "lex"], "y"}], TraditionalForm]]],
 ".\nUsing a similar argument as above we can conclude ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "(", 
     RowBox[{"x", 
      SubscriptBox["+", "c"], "z"}], ")"}], "=", 
    RowBox[{"deg", "(", 
     RowBox[{"y", 
      SubscriptBox["+", "c"], "z"}], ")"}]}], TraditionalForm]]],
 ", and monotonicity of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[LessEqual]", "lex"], TraditionalForm]]],
 " finally yields the desired result."
}], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.615692948502763*^9, 3.61569328352505*^9}, {
   3.615693318953671*^9, 3.615693474878414*^9}, 3.615693513332546*^9},
 CellID->666720646],

Cell[TextData[{
 StyleBox["Summarizing",
  FontWeight->"Bold"],
 ": ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isTOM", "[", "D", "]"}], "\[Implies]", 
    RowBox[{"isTOM", "[", 
     SuperscriptBox["D", "n"], "]"}]}], TraditionalForm]]],
 " for component-wise addition and lexicographic, degree-lexicographic, and \
infinitely other order relations on ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["D", "n"], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.615693547681012*^9, 3.615693624597904*^9}, {
  3.615694712840168*^9, 3.615694740441164*^9}},
 CellID->1515751530],

Cell[TextData[{
 "The next claim we make is that the property of being a totally-ordered \
Abelian monoid carries over from ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"D", ",", 
     RowBox[{"+", 
      RowBox[{",", "\[LessEqual]"}]}]}], ")"}], TraditionalForm]]],
 " to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{"D", "[", "X", "]"}], ",", 
     RowBox[{"+", 
      RowBox[{",", 
       SubscriptBox["\[LessEqual]", "p"]}]}]}], ")"}], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[LessEqual]", "p"], TraditionalForm]]],
 " is the total order relation induced by some total order relation ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[PrecedesEqual]", "t"], TraditionalForm]]],
 " on ",
 Cell[BoxData[
  FormBox[
   RowBox[{"[", "X", "]"}], TraditionalForm]]],
 " (the set of terms) and the total order relation ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 " on ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ", defined in the following way: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    SubscriptBox["<", "p"], "y"}], TraditionalForm]]],
 " iff there exists some term ",
 Cell[BoxData[
  FormBox["\[Tau]", TraditionalForm]]],
 " such that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "(", 
     RowBox[{"x", ",", "\[Tau]"}], ")"}], "<", 
    RowBox[{"C", "(", 
     RowBox[{"y", ",", "\[Tau]"}], ")"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "(", 
     RowBox[{"x", ",", "\[Sigma]"}], ")"}], "=", 
    RowBox[{"C", "(", 
     RowBox[{"y", ",", "\[Sigma]"}], ")"}]}], TraditionalForm]]],
 " for all ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Sigma]", 
    SubscriptBox["\[Succeeds]", "t"], "\[Tau]"}], TraditionalForm]]],
 ". It is straightforward to see that ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[LessEqual]", "p"], TraditionalForm]]],
 " is indeed a total order relation, and proving that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{"C", "[", "X", "]"}], ",", 
     RowBox[{"+", 
      RowBox[{",", 
       SubscriptBox["\[LessEqual]", "p"]}]}]}], ")"}], TraditionalForm]]],
 " is a totally-ordered Abelian monoid can be done in the same spirit as \
proving that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox["D", "n"], ",", 
     RowBox[{
      SubscriptBox["+", "c"], 
      RowBox[{",", "\[LessEqual]"}]}]}], ")"}], TraditionalForm]]],
 " is a totally-ordered Abelian monoid. Hence, this proof is omitted here."
}], "Text",
 CellChangeTimes->{{3.615695177505821*^9, 3.615695322813317*^9}, {
  3.615695370585653*^9, 3.615695582595794*^9}, {3.615695728445045*^9, 
  3.615695731694734*^9}},
 CellID->1432108623],

Cell[TextData[{
 StyleBox["Summarizing",
  FontWeight->"Bold"],
 ": ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isTOM", "[", "D", "]"}], "\[Implies]", 
    RowBox[{"isTOM", "[", 
     RowBox[{"D", "[", "X", "]"}], "]"}]}], TraditionalForm]]],
 " for standard addition in polynomial rings and any order relation induced \
by some total order relation ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[PrecedesEqual]", "t"], TraditionalForm]]],
 " on ",
 Cell[BoxData[
  FormBox[
   RowBox[{"[", "X", "]"}], TraditionalForm]]],
 ". Here, it is important to note that ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[PrecedesEqual]", "t"], TraditionalForm]]],
 " does ",
 StyleBox["not",
  FontSlant->"Italic"],
 " need to be an admissible ordering in the sense that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Tau]", 
     SubscriptBox["\[PrecedesEqual]", "t"], "\[Sigma]"}], "\[Implies]", 
    RowBox[{
     RowBox[{"\[Tau]", "\[CenterDot]", "\[Mu]"}], 
     SubscriptBox["\[PrecedesEqual]", "t"], 
     RowBox[{"\[Sigma]", "\[CenterDot]", "\[Mu]"}]}]}], TraditionalForm]]],
 "!"
}], "Text",
 CellChangeTimes->{{3.615693547681012*^9, 3.615693624597904*^9}, {
  3.615694712840168*^9, 3.615694740441164*^9}, {3.615695637611093*^9, 
  3.615695709447057*^9}, {3.615695743805008*^9, 3.615695834510758*^9}},
 CellID->1887306109],

Cell[TextData[{
 "After we have seen how to \[OpenCurlyDoubleQuote]extended\
\[CloseCurlyDoubleQuote] totally-ordered Abelian monoids, we now show how \
they can be \[OpenCurlyDoubleQuote]restricted\[CloseCurlyDoubleQuote]: For \
this, let as usual ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"D", ",", 
     RowBox[{"+", 
      RowBox[{",", "\[LessEqual]"}]}]}], ")"}], TraditionalForm]]],
 " be a totally-ordered Abelian monoid, and let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"B", "\[SubsetEqual]", "D"}], TraditionalForm]]],
 " be any subset of ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ". Then ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", "B", "\[RightAngleBracket]"}], ",", 
     RowBox[{"+", 
      RowBox[{",", "\[LessEqual]"}]}]}], ")"}], TraditionalForm]]],
 " is a totally-ordered Abelian monoid as well, where ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", "B", "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 " is defined to be the set of all finite sums (in ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ") of elements in ",
 Cell[BoxData[
  FormBox["B", TraditionalForm]]],
 " (where an individual element may be summed up several times). Clearly we \
have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", "B", "\[RightAngleBracket]"}], 
    "\[SubsetEqual]", "D"}], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", "B", "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 " are may be taken as ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.61569591886092*^9, 3.615695968059351*^9}, {
  3.61569713008788*^9, 3.615697397948503*^9}, {3.615697928620805*^9, 
  3.615697938618807*^9}, {3.615702267276722*^9, 3.615702274106883*^9}},
 CellID->1082635799],

Cell[TextData[StyleBox["Proof.",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.615697505590566*^9, 3.615697506413846*^9}, 
   3.615697875320345*^9},
 CellID->593477305],

Cell[TextData[{
 "The proof of the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", "B", "\[RightAngleBracket]"}], ",", 
     RowBox[{"+", 
      RowBox[{",", "\[LessEqual]"}]}]}], ")"}], TraditionalForm]]],
 " is a totally-ordered Abelian monoid is in fact straightforward: Apart from \
",
 ButtonBox["(closure)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:349901806"],
 " and ",
 ButtonBox["(0 in domain)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1476372605"],
 ", all other properties of totally-ordered Abelian monoids carry over from \
",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " to ",
 StyleBox["any",
  FontSlant->"Italic"],
 " of its subsets. Hence, we only need to show that 0 is in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", "B", "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 " and that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "+", "y"}], TraditionalForm]]],
 " is in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", "B", "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 " for all ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", "B", "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 ": 0 in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", "B", "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 " is trivial, since the empty sum by definition is 0, and also the closure \
of ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", "B", "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 " can be easily be shown by arguing that the sum of two ",
 StyleBox["finite",
  FontSlant->"Italic"],
 " sums is again a ",
 StyleBox["finite",
  FontSlant->"Italic"],
 " sum (using associativity)."
}], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.615697513963812*^9, 3.615697542071291*^9}, {
  3.615697579929234*^9, 3.615697719188559*^9}, {3.615697753398569*^9, 
  3.615697875320524*^9}, {3.615702281943151*^9, 3.615702282009961*^9}},
 CellID->909437173],

Cell[TextData[{
 StyleBox["Summarizing",
  FontWeight->"Bold"],
 ": ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isTOM", "[", "D", "]"}], "\[Implies]", 
    RowBox[{"isTOM", "[", 
     RowBox[{"\[LeftAngleBracket]", "B", "\[RightAngleBracket]"}], "]"}]}], 
   TraditionalForm]]],
 " for any ",
 Cell[BoxData[
  FormBox[
   RowBox[{"B", "\[SubsetEqual]", "D"}], TraditionalForm]]],
 ". This, in particular, implies that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     SubscriptBox["\[DoubleStruckCapitalN]", "0"], ",", 
     RowBox[{"+", 
      RowBox[{",", "\[LessEqual]"}]}]}], ")"}], TraditionalForm]]],
 " is a totally-ordered Abelian monoid: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[DoubleStruckCapitalN]", "0"], "=", 
    RowBox[{"\[LeftAngleBracket]", "1", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 " over ",
 Cell[BoxData[
  FormBox["\[DoubleStruckCapitalZ]", TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox["\[DoubleStruckCapitalZ]", TraditionalForm]]],
 " is already known to be a totally-ordered Abelian monoid."
}], "Text",
 CellChangeTimes->{{3.615693547681012*^9, 3.615693624597904*^9}, {
  3.615694712840168*^9, 3.615694740441164*^9}, {3.615695637611093*^9, 
  3.615695709447057*^9}, {3.615695743805008*^9, 3.615695834510758*^9}, {
  3.615697949088842*^9, 3.615698064805829*^9}, {3.615702288031964*^9, 
  3.615702292797128*^9}},
 CellID->423204886],

Cell[TextData[{
 "After illustrating that totally-ordered Abelian monoids are indeed much \
more general than ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 ", it is time to turn to the structures that arise when exponents of \
indeterminates in polynomials are not any more restricted to natural numbers, \
but may be elements of arbitrary totally-ordered Abelian monoids. It is clear \
that these structures should not be referred to as \
\[OpenCurlyDoubleQuote]polynomials\[CloseCurlyDoubleQuote] any more, hence we \
simply call them ",
 StyleBox["generalized polynomials",
  FontSlant->"Italic"],
 " and denote the polynomial ring in the set of indeterminates ",
 Cell[BoxData[
  FormBox["X", TraditionalForm]]],
 " over ",
 Cell[BoxData[
  FormBox["K", TraditionalForm]]],
 " with exponents in the totally-ordered Abelian monoid ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"K", "[", 
    SuperscriptBox["X", "D"], "]"}], TraditionalForm]]],
 "; It is apparent that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"K", "[", 
    SuperscriptBox["X", "D"], "]"}], TraditionalForm]]],
 " really forms a ring, simply because multiplication of terms is still \
realized by adding their exponents and addition in ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " possesses all the necessary properties for giving rise to an associative \
and commutative term multiplication (with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"1", "=", 
    SuperscriptBox["X", "0"]}], TraditionalForm]]],
 " as its neutral element)."
}], "Text",
 CellChangeTimes->{{3.615698417710112*^9, 3.615698602911525*^9}, {
   3.615698659989189*^9, 3.615698711035569*^9}, {3.615698836732784*^9, 
   3.615698902703746*^9}, {3.615698974211526*^9, 3.615699126577088*^9}, 
   3.615699157659932*^9, {3.61569918787852*^9, 3.615699192307223*^9}, {
   3.615699247042878*^9, 3.615699339981831*^9}, {3.615702223560957*^9, 
   3.615702229245687*^9}},
 CellID->174363722],

Cell[TextData[{
 "Now, what does all this mean for our complexity analysis? The answer to \
this question is rather trivial: One can define and carry out Buchberger\
\[CloseCurlyQuote]s algorithm (",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ") also in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"K", "[", 
    SuperscriptBox[
     RowBox[{"{", 
      RowBox[{"x", ",", "y"}], "}"}], "D"], "]"}], TraditionalForm]]],
 ", and ",
 StyleBox["if it terminates",
  FontWeight->"Bold"],
 " on a set ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 ", returning ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 ", we have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"maxdeg", "[", "G", "]"}], "\[LessEqual]", 
    RowBox[{"2", "\[CenterDot]", 
     RowBox[{"maxdeg", "[", "F", "]"}]}]}], TraditionalForm]]],
 ", according to ",
 ButtonBox["(main theorem)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1891911403"],
 "; ",
 Cell[BoxData[
  FormBox[
   RowBox[{"maxdeg", "[", "G", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"maxdeg", "[", "F", "]"}], TraditionalForm]]],
 " are, of course, elements in ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ". As indicated above, termination is not guaranteed in general: Take, for \
instance, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"D", "=", "\[DoubleStruckCapitalZ]"}], TraditionalForm]]],
 ": There, the reduction relation is not Noetherian, meaning that already \
step 8 in ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 " might not terminate. In such case, ",
 ButtonBox["(main theorem)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1891911403"],
 " only ensures that in each iteration of the main loop in the algorithm the \
bound ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"maxdeg", "[", "G", "]"}], "\[LessEqual]", 
    RowBox[{"2", "\[CenterDot]", 
     RowBox[{"maxdeg", "[", "F", "]"}]}]}], TraditionalForm]]],
 " holds."
}], "Text",
 CellChangeTimes->{{3.615699428623523*^9, 3.615699693652565*^9}, {
  3.61569974593474*^9, 3.61569975920116*^9}, {3.615699813896409*^9, 
  3.615699852307092*^9}, {3.615700020738832*^9, 3.615700047238907*^9}, {
  3.615700118780447*^9, 3.61570013743266*^9}, {3.615700168979557*^9, 
  3.615700272389236*^9}, {3.615700319073836*^9, 3.61570035477495*^9}, {
  3.615704675879288*^9, 3.615704676097388*^9}},
 CellID->136010290],

Cell[TextData[{
 "But even under the assumption the algorithm terminates we cannot, in \
general, draw any conclusions concerning its complexity, i.e. the number of \
steps that actually have to be carried out. This is only possible if ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " has a much stronger property, namely the property that for each element in \
",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " there are only ",
 StyleBox["finitely",
  FontSlant->"Italic"],
 " many elements strictly less than it. This is apparently not the case in ",
 Cell[BoxData[
  FormBox["\[DoubleStruckCapitalZ]", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["\[DoubleStruckCapitalQ]", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["\[DoubleStruckCapitalR]", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["\[DoubleStruckCapitalQ]", "0", "+"], 
   TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["\[DoubleStruckCapitalR]", "0", "+"], 
   TraditionalForm]]],
 ", etc., and even though all sets could be well-ordered in principle \
(assuming the axiom of choice), the order relation would still have to be \
monotonic w.r.t. ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.615700358981808*^9, 3.615700499446731*^9}, {
  3.615700530563577*^9, 3.615700568860535*^9}, {3.615700646927235*^9, 
  3.615700793996462*^9}, {3.615701685463113*^9, 3.615701692364734*^9}},
 CellID->1956243226],

Cell[TextData[{
 "However, apart from ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 " there are also other totally-ordered Abelian monoids having this \
additional property; For instance, consider ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["\[DoubleStruckCapitalN]", "0", "n"], 
   TraditionalForm]]],
 " with a graded order relation (like degree-lexicographic). This immediately \
leads to an interesting question: Is it possible to extended the present \
complexity analysis carried out for the ",
 StyleBox["bivariate",
  FontSlant->"Italic"],
 " case to, say, four indeterminates ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["u", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["v", TraditionalForm]]],
 ", by grouping ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["u", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["v", TraditionalForm]]],
 " together and working in the bivariate polynomial ring ",
 Cell[BoxData[
  FormBox[
   RowBox[{"K", "[", 
    SuperscriptBox[
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"x", ",", "y"}], ")"}], ",", 
       RowBox[{"(", 
        RowBox[{"u", ",", "v"}], ")"}]}], "}"}], 
     SubsuperscriptBox["\[DoubleStruckCapitalN]", "0", "2"]], "]"}], 
   TraditionalForm]]],
 "? Unfortunately, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"K", "[", 
     RowBox[{"x", ",", "y", ",", "u", ",", "v"}], "]"}], " "}], 
   TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"K", "[", 
    SuperscriptBox[
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"x", ",", "y"}], ")"}], ",", 
       RowBox[{"(", 
        RowBox[{"u", ",", "v"}], ")"}]}], "}"}], 
     SubsuperscriptBox["\[DoubleStruckCapitalN]", "0", "2"]], "]"}], 
   TraditionalForm]]],
 " are not isomorphic to each other in the sense that in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"K", "[", 
    SuperscriptBox[
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"x", ",", "y"}], ")"}], ",", 
       RowBox[{"(", 
        RowBox[{"u", ",", "v"}], ")"}]}], "}"}], 
     SubsuperscriptBox["\[DoubleStruckCapitalN]", "0", "2"]], "]"}], 
   TraditionalForm]]],
 " the notion of ",
 Cell[BoxData[
  FormBox["lcm", TraditionalForm]]],
 ", as defined below in section ",
 ButtonBox["Specific Notions for Complexity Analysis",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::SpecificNotions"],
 ", is not compatible with divisibility: For instance, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lcm", "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"2", ",", "1"}], ")"}], ",", 
        RowBox[{"(", 
         RowBox[{"0", ",", "0"}], ")"}]}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"0", ",", "2"}], ")"}], ",", 
        RowBox[{"(", 
         RowBox[{"0", ",", "0"}], ")"}]}], "\[RightAngleBracket]"}]}], "]"}], 
    "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"2", ",", "1"}], ")"}], ",", 
      RowBox[{"(", 
       RowBox[{"0", ",", "0"}], ")"}]}], "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 " because ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"2", ",", "1"}], ")"}], TraditionalForm]]],
 " is greater than ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"0", ",", "2"}], ")"}], TraditionalForm]]],
 " w.r.t. any graded ordering and ",
 Cell[BoxData[
  FormBox["lcm", TraditionalForm]]],
 " just takes the maximum exponent, but clearly ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"2", ",", "1"}], ")"}], ",", 
     RowBox[{"(", 
      RowBox[{"0", ",", "0"}], ")"}]}], "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 " is not divisible by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"0", ",", "2"}], ")"}], ",", 
     RowBox[{"(", 
      RowBox[{"0", ",", "0"}], ")"}]}], "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 ". Hence, although ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 " might indeed compute some set ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " for given input ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"K", "[", 
    SuperscriptBox[
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"x", ",", "y"}], ")"}], ",", 
       RowBox[{"(", 
        RowBox[{"u", ",", "v"}], ")"}]}], "}"}], 
     SubsuperscriptBox["\[DoubleStruckCapitalN]", "0", "2"]], "]"}], 
   TraditionalForm]]],
 ", this ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " will have nothing to do with a Gr\[ODoubleDot]bner basis in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"K", "[", 
    RowBox[{"x", ",", "y", ",", "u", ",", "v"}], "]"}], TraditionalForm]]],
 ", and the degree bound for ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " is completely irrelevant."
}], "Text",
 CellChangeTimes->{{3.615701708861246*^9, 3.615701772148031*^9}, {
  3.61570181449256*^9, 3.615701814967877*^9}, {3.615701909056317*^9, 
  3.61570196901041*^9}, {3.61570204932798*^9, 3.615702197354804*^9}, {
  3.615702244666365*^9, 3.615702244771155*^9}, {3.615702367955818*^9, 
  3.615702379868448*^9}, {3.615704314787724*^9, 3.615704320671258*^9}, {
  3.615704397226841*^9, 3.61570445116588*^9}, {3.615704483487788*^9, 
  3.615704556282953*^9}, {3.615704642892355*^9, 3.615704646966105*^9}, {
  3.615704708058551*^9, 3.615705074953187*^9}, {3.615705210756277*^9, 
  3.615705224221843*^9}},
 CellID->72224179],

Cell[TextData[{
 "Still one might wonder whether it is possible to overcome the problem of \
incompatibility of lcm and divisibility. One could, for example, place much \
stronger requirements for the domain of exponents, like requiring ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[LessEqual]", "y"}], TraditionalForm]]],
 " iff ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "+", "z"}], "=", "y"}], TraditionalForm]]],
 " for some ",
 Cell[BoxData[
  FormBox["z", TraditionalForm]]],
 ". This would indeed lead to a more reasonable definition of ",
 Cell[BoxData[
  FormBox["lcm", TraditionalForm]]],
 " which is compatible with divisibility in the sense that for all exponent \
vectors ",
 Cell[BoxData[
  FormBox["u", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["v", TraditionalForm]]],
 " there exists an exponent vector ",
 Cell[BoxData[
  FormBox["w", TraditionalForm]]],
 " such that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"u", "+", "w"}], "=", 
    RowBox[{"lcm", "[", 
     RowBox[{"u", ",", "v"}], "]"}]}], TraditionalForm]]],
 " (and analogous for ",
 Cell[BoxData[
  FormBox["v", TraditionalForm]]],
 "). The problem here however is that any order relation on ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["\[DoubleStruckCapitalN]", "0", "2"], 
   TraditionalForm]]],
 " satisfying the requirement \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[LessEqual]", "y"}], TraditionalForm]]],
 " iff ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "+", "z"}], "=", "y"}], TraditionalForm]]],
 " for some ",
 Cell[BoxData[
  FormBox["z", TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] cannot be total at the same time."
}], "Text",
 CellChangeTimes->{{3.615705132656723*^9, 3.615705173568712*^9}, {
  3.615705258247207*^9, 3.615705286541606*^9}, {3.615705331241316*^9, 
  3.615705454956216*^9}, {3.615705507460997*^9, 3.615705586888229*^9}, {
  3.615706419124704*^9, 3.615706645948245*^9}, {3.615706752053537*^9, 
  3.615706862567049*^9}},
 CellID->2079853062],

Cell[TextData[{
 StyleBox["Summarizing",
  FontWeight->"Bold"],
 ": Drawing any conclusions about the complexity of Buchberger\
\[CloseCurlyQuote]s algorithm in more than two indeterminates by grouping \
several indeterminates together is not possible, at least following the \
approach presented in this document."
}], "Text",
 CellChangeTimes->{{3.615693547681012*^9, 3.615693624597904*^9}, {
  3.615694712840168*^9, 3.615694740441164*^9}, {3.615695637611093*^9, 
  3.615695709447057*^9}, {3.615695743805008*^9, 3.615695834510758*^9}, {
  3.615697949088842*^9, 3.615698064805829*^9}, {3.615702288031964*^9, 
  3.615702292797128*^9}, {3.615706977640377*^9, 3.615707071982011*^9}},
 CellID->181517487]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Domain and Dimension", "Subsection",
 CellChangeTimes->{{3.605525534853522*^9, 3.60552553898357*^9}},
 CellID->362214253],

Cell[TextData[{
 "In the rest of this report (not only in this section) let ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " always denote some totally-ordered Abelian monoid:"
}], "Text",
 CellChangeTimes->{{3.605522955295857*^9, 3.605523160333948*^9}, {
  3.605524015905175*^9, 3.605524033244144*^9}, {3.60552439146962*^9, 
  3.605524395783149*^9}, {3.605524449885599*^9, 3.605524452309062*^9}, {
  3.605525856263125*^9, 3.605525916317402*^9}, {3.60552597034784*^9, 
  3.605525980538209*^9}},
 CellID->1764685072],

Cell["", "OpenEnvironment",
 CellID->1607947619],

Cell[CellGroupData[{

Cell[TextData[{
 "Convention: Domain ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.605525630262746*^9, 3.605525660347683*^9}, {
  3.605525841377797*^9, 3.605525843409443*^9}},
 CellID->1463782225],

Cell[BoxData[
 RowBox[{"isTOM", "[", "D", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(D is TOM)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2137457113", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.605525846232255*^9, 3.605525849112352*^9}},
 CellTags->{
  "ID:2137457113", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is TOM"},
 CellID->2137457113],

Cell[BoxData[
 RowBox[{"isAbelianMonoid", "[", "D", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(D is Abelian monoid)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1975341048", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610944893331213*^9, 3.610944898607773*^9}},
 CellTags->{
  "ID:1975341048", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is Abelian monoid"},
 CellID->1975341048],

Cell[BoxData[
 RowBox[{"isTotallyOrdered", "[", "D", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(D is totally ordered)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:275749992", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610942790788065*^9, 3.610942795420077*^9}},
 CellTags->{
  "ID:275749992", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is totally ordered"},
 CellID->275749992],

Cell[BoxData[
 RowBox[{"isMinMaxDomain", "[", "D", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(D is min-max domain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1904592343", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61094491438172*^9, 3.610944918852517*^9}},
 CellTags->{
  "ID:1904592343", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is min-max domain"},
 CellID->1904592343],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->763770611]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1609998906],

Cell[TextData[{
 "In some proofs we need to know at each step that ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " is a totally-ordered Abelian monoid, but in particular that it is \
totally-ordered / Abelian monoid / min-max domain. This is the reason why we \
state all properties explicitly above (if we only use ",
 Cell[BoxData[
  FormBox["isTOM", TraditionalForm]]],
 " in proofs, after expanding definitions this knowledge is not available any \
more)."
}], "Text",
 CellChangeTimes->{{3.610942820699088*^9, 3.610942869614824*^9}, {
  3.610944976298516*^9, 3.610945014526307*^9}, {3.613965027755604*^9, 
  3.613965028030583*^9}},
 CellID->1977697537],

Cell[TextData[{
 "Also, let ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " always be a natural number (Later, in about half of the rest of this \
report, ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " will be fixed to 2):"
}], "Text",
 CellChangeTimes->{{3.605525985558815*^9, 3.605525992501896*^9}, {
  3.605597644614443*^9, 3.605597662498582*^9}},
 CellID->981023549],

Cell["", "OpenEnvironment",
 CellID->1810710293],

Cell[CellGroupData[{

Cell[TextData[{
 "Convention: Number ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.605526002669922*^9, 3.605526008699401*^9}},
 CellID->1016833517],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
  RowBox[{"[", "n", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(n is integer)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:589296925", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.605526028062*^9, 3.605526030748108*^9}, {
  3.610958455708267*^9, 3.610958463532322*^9}, {3.610958592577602*^9, 
  3.610958593692277*^9}},
 CellTags->{
  "ID:589296925", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n is integer"},
 CellID->589296925],

Cell[BoxData[
 RowBox[{"1", 
  UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
  "n"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(n is positive)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:54512703", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610958613873144*^9, 3.610958619902113*^9}},
 CellTags->{
  "ID:54512703", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n is positive"},
 CellID->54512703],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1567750413]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->307153015]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmata on Basic Notions", "Section",
 CellChangeTimes->{{3.599194523384213*^9, 3.59919452983264*^9}, 
   3.605435269353079*^9, {3.610956225929403*^9, 3.610956226365731*^9}, 
   3.613885821159292*^9, {3.613971542548043*^9, 3.61397154334385*^9}},
 CellTags->"sect::LemmataBasic",
 CellID->1171017268],

Cell[TextData[{
 "After having defined all the necessary basic notions, it is time to start \
with the formal verification of some well-known properties of these notions.\n\
The formal verification was done with the (semi-) automated theorem proving \
mechanisms of ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ", more precisely with a new prover implemented for the present complexity \
analysis, in the sequel called ",
 StyleBox["ComplexityProver",
  FontSlant->"Italic"],
 " (note that its real name which is displayed in the dropdown-menu of ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 "\[CloseCurlyQuote]s prover setup is \[OpenCurlyDoubleQuote]Rules in Theory \
of Complexity Analysis of bivariate Buchberger\[CloseCurlyQuote]s Algorithm\
\[CloseCurlyDoubleQuote]). More information on it can be found in section ",
 ButtonBox["Prover",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Prover"],
 "."
}], "Text",
 CellChangeTimes->{{3.603530690940708*^9, 3.603530750146392*^9}, {
  3.603530780961436*^9, 3.603530851738582*^9}, {3.603531011389613*^9, 
  3.603531017752887*^9}, {3.603531062192716*^9, 3.603531177819123*^9}, {
  3.603531230797807*^9, 3.603531260641389*^9}, {3.603600634352187*^9, 
  3.6036006345338*^9}, {3.605435324608294*^9, 3.605435335070067*^9}, {
  3.605436244600801*^9, 3.6054362457048*^9}, {3.615024932169779*^9, 
  3.615024937904924*^9}, {3.615025035103827*^9, 3.615025049015312*^9}},
 CellID->615481853],

Cell[TextData[{
 "The crucial observation regarding the lemmata below is the following:\nMost \
of them will not be used later as they are, but as ",
 StyleBox["inference rules",
  FontSlant->"Italic"],
 " of the ComplexityProver, as already explained in section ",
 ButtonBox["Formalizing Mathematics",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::FormalMathematics"],
 ". For instance, consider formula ",
 ButtonBox["(irreflexivity of <)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1308600401"],
 ": This formula states that ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["<", "D"], TraditionalForm]]],
 " is an irreflexive order relation whenever ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " is a totally-ordered domain, meaning that we never have ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["<", "D"], "x"}], TraditionalForm]]],
 ", which will be quite useful later on. However, later we will not need this \
formula directly in order to carry out an inference step, but simply make use \
of an",
 StyleBox[" inference rule",
  FontSlant->"Italic"],
 " (\[OpenCurlyDoubleQuote]Knowledge base contains contradicting formulae in \
theory of totally-ordered domains\[CloseCurlyDoubleQuote]): This rule \
basically does nothing else than proving ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " whenever ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["<", "D"], "x"}], TraditionalForm]]],
 " is among the formulas in the knowledge base, for then the proof is \
finished. We proceed like that for a simple reason:\nTypically, type \
information about some object ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " (e.g. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 ") is not among the list of assumptions when we want such a rule to be \
applied. So, the general rewriting mechanism (see section ",
 ButtonBox["Prover",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Prover"],
 ") does not work in this case, because rewriting (by elementary \
substitutions) only works if all the conditions of a rewrite rule follow \
immediately from the current knowledge base, which, in order words, means \
that they more or less have to be included in the knowledge base. The \
inference rule simply proves these conditions afterwards, so to say."
}], "Text",
 CellChangeTimes->{{3.605436167432696*^9, 3.605436238011138*^9}, {
  3.605436271719248*^9, 3.605436413505098*^9}, {3.605436450820495*^9, 
  3.605436871843695*^9}, {3.605436902275684*^9, 3.605437184101328*^9}, {
  3.605437242883256*^9, 3.605437312757911*^9}, {3.605437348601333*^9, 
  3.60543738126154*^9}, {3.60543815215649*^9, 3.605438161470833*^9}, {
  3.605438197979896*^9, 3.605438439538129*^9}, {3.605438476079082*^9, 
  3.605438598723995*^9}, {3.605438675379939*^9, 3.605438675403744*^9}, {
  3.605438903823556*^9, 3.605438924385984*^9}, {3.60552524571989*^9, 
  3.605525252974002*^9}, {3.611035986208798*^9, 3.611035989069389*^9}, {
  3.615025014359508*^9, 3.615025017616263*^9}, {3.615025124830101*^9, 
  3.615025200815633*^9}, {3.615025233825317*^9, 3.615025240421891*^9}, {
  3.615025277605984*^9, 3.615025319327569*^9}},
 CellID->1795547704],

Cell[TextData[{
 "So, one might now wonder why formula ",
 ButtonBox["(irreflexivity of <)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1308600401"],
 " is needed at all, and the answer is the following:\n",
 StyleBox["It is needed to establish the correctness of the corresponding \
inference rule.",
  FontSlant->"Italic"],
 "\nThe issue of verifying provers by other provers is of interest in its own \
right, and is addressed for instance in ",
 ButtonBox["[Giese2007]",
  BaseStyle->"Hyperlink",
  ButtonData->"Giese2007"],
 ". Only note that, clearly, an inference rule can only be verified if the \
corresponding formula is proved without the application of this inference \
rule, since otherwise the whole proof would be circular. This has been taken \
care of in all of the proofs in this section."
}], "Text",
 CellChangeTimes->{
  3.615025294564816*^9, {3.615025401645237*^9, 3.615025404999565*^9}, {
   3.61502544759612*^9, 3.615025461065408*^9}},
 CellID->541338696],

Cell["\<\
Please observe that the structure of the ComplexityProver, i.e. the \
arrangement of the (theory-specific) inference rules, follows exactly the \
section-structure below: Inference rules dealing with \
\[DoubleStruckCapitalZ], inference rules dealing with order relations, \
inference rules dealing with tuples, and inference rules dealing with min/max.\
\>", "Text",
 CellChangeTimes->{3.615025402977601*^9},
 CellID->765415560],

Cell[CellGroupData[{

Cell["General", "Subsection",
 CellChangeTimes->{{3.612862691365267*^9, 3.612862692524809*^9}},
 CellID->497241048],

Cell[TextData[{
 "The following lemma is of course entirely obvious (it basically states ",
 Cell[BoxData[
  FormBox[
   RowBox[{"A", "\[Or]", 
    RowBox[{"\[Not]", "A"}]}], TraditionalForm]]],
 "), but nevertheless it will be used a couple of times when proving the more \
advanced theorems in the subsequent sections."
}], "Text",
 CellChangeTimes->{{3.615025489737885*^9, 3.615025620821893*^9}},
 CellID->627429447],

Cell["", "OpenEnvironment",
 CellID->518018031],

Cell[CellGroupData[{

Cell["Lemma: Two Objects are either Equal or Distinct", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.612862707966013*^9, 3.612862722263294*^9}},
 CellID->1477491869],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"x", ",", "y"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[Equal]", "y"}], "\[Or]", 
    RowBox[{"x", "\[NotEqual]", "y"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1538785027", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600420867170865*^9, 3.600420870082918*^9}, {
  3.61284865559504*^9, 3.612848671637333*^9}},
 CellTags->{
  "ID:1538785027", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "3"},
 CellID->1538785027],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->238945331]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1212198655],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["x", "ExpressionVariable"], ",", 
            StyleBox["y", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[Equal]", 
             StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["x", "ExpressionVariable"], ",", 
            StyleBox["y", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[Equal]", 
             StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1538785027"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1538785027-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1538785027-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1538785027", "Proof|ID:1538785027-1"},
 CellID->1800356526],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1538785027-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1169871723]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\[DoubleStruckCapitalZ]", "Subsection",
 CellChangeTimes->{{3.60316885223127*^9, 3.603168853393327*^9}, {
   3.603176989722755*^9, 3.603176990908389*^9}, {3.610956208993248*^9, 
   3.610956221017746*^9}, 3.615021411398095*^9},
 CellID->777956154],

Cell[TextData[{
 "In the next lemma we summarize some simple properties of ",
 Cell[BoxData[
  FormBox["\[DoubleStruckCapitalZ]", TraditionalForm]]],
 " that follow immediately from its definition in section ",
 ButtonBox["Basic Notions / General Notions / \[DoubleStruckCapitalZ]",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::DefinitionZ"],
 ":"
}], "Text",
 CellChangeTimes->{{3.605435389719512*^9, 3.605435506169107*^9}, {
  3.614937483507525*^9, 3.614937484283202*^9}, {3.615025713337526*^9, 
  3.615025771601134*^9}, {3.615025806633771*^9, 3.615025806651202*^9}},
 CellID->1725363192],

Cell["", "OpenEnvironment",
 CellID->1655989806],

Cell[CellGroupData[{

Cell["Lemma: Simple Properties of \[DoubleStruckCapitalZ]", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.61206616870104*^9, 3.612066179792914*^9}, {
  3.61206621607232*^9, 3.612066219919531*^9}, {3.612086672385529*^9, 
  3.612086673492375*^9}},
 CellID->276004122],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"[", "a", "]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612066100512712*^9, 3.61206610806216*^9}},
 CellID->2116038141],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   SubscriptBox["\[DoubleStruckCapitalZ]", 
    RowBox[{"1", ",", "\[Ellipsis]", ",", "2"}]]], 
  RowBox[{"[", "1", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:968428080", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612086682778586*^9, 3.612086692570489*^9}, {
  3.612086728618803*^9, 3.612086731423012*^9}},
 CellTags->{
  "ID:968428080", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "4"},
 CellID->968428080],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   SubscriptBox["\[DoubleStruckCapitalZ]", 
    RowBox[{"1", ",", "\[Ellipsis]", ",", "2"}]]], 
  RowBox[{"[", "2", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:680012314", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.612086729909161*^9},
 CellTags->{
  "ID:680012314", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "5"},
 CellID->680012314],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], "a"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"1", ",", "\[Ellipsis]", ",", "a"}]]], 
   RowBox[{"[", "a", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:655751250", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611991792899589*^9, 3.611991852592421*^9}, {
  3.611992684383261*^9, 3.611992685470785*^9}, {3.611992748509068*^9, 
  3.611992762383859*^9}, {3.612066110273054*^9, 3.612066111379326*^9}, {
  3.61206665420918*^9, 3.612066665715434*^9}},
 CellTags->{
  "ID:655751250", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "6"},
 CellID->655751250],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"1", ",", "\[Ellipsis]", ",", "a"}]]], 
   RowBox[{"[", 
    RowBox[{"a", 
     UnderscriptBox["+", "\[DoubleStruckCapitalZ]"], "1"}], 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(7)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1963081417", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611989696390915*^9, 3.611989723974249*^9}, {
  3.611990974973115*^9, 3.61199098159357*^9}, {3.612066112781301*^9, 
  3.612066113979475*^9}},
 CellTags->{
  "ID:1963081417", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "7"},
 CellID->1963081417],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2087479351]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1363114882],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", 
          SubscriptBox["\[DoubleStruckCapitalN]", 
           RowBox[{"1", ",", "\[Ellipsis]", ",", "2"}]]], "[", "1", 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", 
          SubscriptBox["\[DoubleStruckCapitalN]", 
           RowBox[{"1", ",", "\[Ellipsis]", ",", "2"}]]], "[", "1", 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:968428080"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p968428080-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p968428080-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:968428080", "Proof|ID:968428080-1"},
 CellID->665537877],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p968428080-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1093697245]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", 
          SubscriptBox["\[DoubleStruckCapitalN]", 
           RowBox[{"1", ",", "\[Ellipsis]", ",", "2"}]]], "[", "2", 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", 
          SubscriptBox["\[DoubleStruckCapitalN]", 
           RowBox[{"1", ",", "\[Ellipsis]", ",", "2"}]]], "[", "2", 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:680012314"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p680012314-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p680012314-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:680012314", "Proof|ID:680012314-1"},
 CellID->134690164],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p680012314-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->633008601]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              SubscriptBox["\[DoubleStruckCapitalN]", 
               RowBox[{"1", ",", "\[Ellipsis]", ",", 
                 StyleBox["a", "ExpressionVariable"]}]]], "[", 
             StyleBox["a", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              SubscriptBox["\[DoubleStruckCapitalN]", 
               RowBox[{"1", ",", "\[Ellipsis]", ",", 
                 StyleBox["a", "ExpressionVariable"]}]]], "[", 
             StyleBox["a", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:655751250"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p655751250-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p655751250-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:655751250", "Proof|ID:655751250-1"},
 CellID->234671713],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p655751250-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->319014567]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(7)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(7)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               UnderscriptBox["\[Element]", 
                SubscriptBox["\[DoubleStruckCapitalN]", 
                 RowBox[{"1", ",", "\[Ellipsis]", ",", 
                   StyleBox["a", "ExpressionVariable"]}]]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[DoubleStruckCapitalZ]"], "1"}], "]"}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               UnderscriptBox["\[Element]", 
                SubscriptBox["\[DoubleStruckCapitalN]", 
                 RowBox[{"1", ",", "\[Ellipsis]", ",", 
                   StyleBox["a", "ExpressionVariable"]}]]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[DoubleStruckCapitalZ]"], "1"}], "]"}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1963081417"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1963081417-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1963081417-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1963081417", "Proof|ID:1963081417-1"},
 CellID->1541015864],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1963081417-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->845366744]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Other Coordinate", "Subsection",
 CellChangeTimes->{{3.61224042498025*^9, 3.612240426842207*^9}, {
  3.612240593779819*^9, 3.612240597553348*^9}},
 CellID->58313402],

Cell[TextData[{
 "Now we focus on the notion of \[OpenCurlyDoubleQuote]other coordinate\
\[CloseCurlyDoubleQuote], as introduced in section ",
 ButtonBox["Basic Notion / General Notions / Other Coordinate",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::DefinitionOC"],
 ". All three formulas are immediate consequences of the definition of ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[EmptySquare]", "-"], TraditionalForm]]],
 " and can be proved directly by distinguishing whether ",
 Cell[BoxData[
  FormBox["k", TraditionalForm]]],
 " is 1 or 2."
}], "Text",
 CellChangeTimes->{{3.615025838204724*^9, 3.615025951057755*^9}, {
  3.615026004273914*^9, 3.615026004297182*^9}, {3.615026092518317*^9, 
  3.615026093783914*^9}},
 CellID->732300639],

Cell["", "OpenEnvironment",
 CellID->885155644],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["k", "-"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.61224045354422*^9, 3.612240457989742*^9}},
 CellID->1272852245],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "2"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612240462969787*^9, 3.612240470345918*^9}},
 CellID->405924544],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   SubscriptBox["\[DoubleStruckCapitalZ]", 
    RowBox[{"1", ",", "\[Ellipsis]", ",", "2"}]]], 
  RowBox[{"[", 
   SuperscriptBox["k", "-"], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(8)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1457798778", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612240473660171*^9, 3.612240488374008*^9}},
 CellTags->{
  "ID:1457798778", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "8"},
 CellID->1457798778],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", "2"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"l", "\[NotEqual]", "k"}], 
    TagBox["\[DoubleLeftRightArrow]",
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{"l", "\[Equal]", 
     SuperscriptBox["k", "-"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(9)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1044840887", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612240503336774*^9, 3.612240510327392*^9}, {
   3.612241403562106*^9, 3.612241408126576*^9}, 3.61224143953748*^9, {
   3.612243808314797*^9, 3.612243810971586*^9}, 3.61311490980443*^9},
 CellTags->{
  "ID:1044840887", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "9"},
 CellID->1044840887],

Cell[BoxData[
 RowBox[{
  SuperscriptBox[
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     SuperscriptBox["k", "-"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], "-"], "\[Equal]", "k"}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(10)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2017090463", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612240519385915*^9, 3.612240536589382*^9}},
 CellTags->{
  "ID:2017090463", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "10"},
 CellID->2017090463],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1459353343]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->384181077],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(8)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(8)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", "2"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            SubscriptBox["\[DoubleStruckCapitalN]", 
             RowBox[{"1", ",", "\[Ellipsis]", ",", "2"}]]], "[", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             SuperscriptBox[
              StyleBox["k", "ExpressionVariable"], "-"], 
             TagBox[")", "AutoParentheses"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", "2"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            SubscriptBox["\[DoubleStruckCapitalN]", 
             RowBox[{"1", ",", "\[Ellipsis]", ",", "2"}]]], "[", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             SuperscriptBox[
              StyleBox["k", "ExpressionVariable"], "-"], 
             TagBox[")", "AutoParentheses"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1457798778"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1457798778-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1457798778-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1457798778", "Proof|ID:1457798778-1"},
 CellID->122079454],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1457798778-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->717730248]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(9)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(9)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", "2"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", "2"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["l", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["k", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["l", "ExpressionVariable"], "\[Equal]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox[
                StyleBox["k", "ExpressionVariable"], "-"], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", "2"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", "2"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["l", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["k", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["l", "ExpressionVariable"], "\[Equal]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox[
                StyleBox["k", "ExpressionVariable"], "-"], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1044840887"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1044840887-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1044840887-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1044840887", "Proof|ID:1044840887-1"},
 CellID->1585772568],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1044840887-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->240181621]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(10)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(10)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", "2"}]}}]], 
         RowBox[{
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             SuperscriptBox[
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                SuperscriptBox[
                 StyleBox["k", "ExpressionVariable"], "-"], 
                TagBox[")", "AutoParentheses"]}], "-"], 
             TagBox[")", "AutoParentheses"]}], "\[Equal]", 
           StyleBox[
           "k", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", "2"}]}}]], 
         RowBox[{
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             SuperscriptBox[
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                SuperscriptBox[
                 StyleBox["k", "ExpressionVariable"], "-"], 
                TagBox[")", "AutoParentheses"]}], "-"], 
             TagBox[")", "AutoParentheses"]}], "\[Equal]", 
           StyleBox[
           "k", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2017090463"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2017090463-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2017090463-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2017090463", "Proof|ID:2017090463-1"},
 CellID->1023157275],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2017090463-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1765168555]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Orderings", "Subsection",
 CellChangeTimes->{{3.600148066258234*^9, 3.60014807521599*^9}, {
   3.603014191505322*^9, 3.603014251499771*^9}, 3.60353077567831*^9, {
   3.610956204931847*^9, 3.610956219873381*^9}},
 CellTags->"sect::LemmataOrderings",
 CellID->1103427178],

Cell[TextData[{
 "The first lemma in this subsection summarizes some well-known properties of \
total order relations, and relations between reflexive and strict order \
relations. All the predicates occuring in the formulas are defined in section \
",
 ButtonBox["Basic Notions / Categories / Totally-Ordered Domain",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::DefinitionTotallyOrderedDomain"],
 ", meaning that absolutely no built-in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " knowledge is used."
}], "Text",
 CellChangeTimes->{{3.603532649045107*^9, 3.603532810732877*^9}, {
  3.603600216057317*^9, 3.603600234041742*^9}, {3.605526711092276*^9, 
  3.60552678424333*^9}, {3.610274133964856*^9, 3.61027414725188*^9}, {
  3.615026110908155*^9, 3.615026155377703*^9}, {3.615026225893879*^9, 
  3.615026225919891*^9}, {3.615026262542894*^9, 3.615026272309348*^9}},
 CellID->833726611],

Cell[TextData[{
 "Note that we prove the following two lemmata not only for our a.b.f. \
totally-ordered Abelian monoid ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ", but for ",
 StyleBox["all",
  FontSlant->"Italic"],
 " totally-ordered domains. This is because later, at some points, we need to \
know the formulas below also for the case where the domain is ",
 Cell[BoxData[
  FormBox["\[DoubleStruckCapitalZ]", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.605526791442166*^9, 3.605526910385688*^9}, {
  3.61027415787492*^9, 3.610274297355016*^9}, {3.615026287333762*^9, 
  3.615026318561669*^9}, {3.617952592620621*^9, 3.617952594335528*^9}, {
  3.617953180057379*^9, 3.617953181320832*^9}},
 CellID->591871091],

Cell["", "OpenEnvironment",
 CellID->1950877292],

Cell[CellGroupData[{

Cell["Lemma: Properties of totally-ordered domains", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.600059140830375*^9, 3.600059155920018*^9}, {
  3.600408888308157*^9, 3.600408893151963*^9}},
 CellID->1909682137],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isTotallyOrdered", "[", "dom", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.610273998476514*^9, 3.610274005426766*^9}, {
  3.610274318597181*^9, 3.610274327410667*^9}},
 CellID->1236064582],

Cell[BoxData[
 RowBox[{"isIrreflexive", "[", 
  UnderscriptBox["<", "dom"], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(irreflexivity of <)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1308600401", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600060444560332*^9, 3.600060453301771*^9}, {
   3.600408365726573*^9, 3.600408372535882*^9}, 3.600408793444093*^9, {
   3.610274046493358*^9, 3.610274046774463*^9}},
 CellTags->{
  "ID:1308600401", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "irreflexivity of <"},
 CellID->1308600401],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "dom"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y", ",", "z"}], "]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.600151733636729*^9, 3.600151739962263*^9}, {
  3.600152751294046*^9, 3.600152751479049*^9}, {3.610274049173316*^9, 
  3.610274049441543*^9}},
 CellID->1862701560],

Cell[BoxData[
 RowBox[{"x", 
  UnderscriptBox["\[LessEqual]", "dom"], "x"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reflexivity of \[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:914659821", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600059217899283*^9, 3.600059272284259*^9}, 
   3.60005934010537*^9, {3.600408352408255*^9, 3.600408363331265*^9}, 
   3.600408792530296*^9, {3.602934050141862*^9, 3.602934076388727*^9}, 
   3.602997133761764*^9, 3.610274053225811*^9},
 CellTags->{
  "ID:914659821", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of \[LessEqual]"},
 CellID->914659821],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["<", "dom"], "y"}], "\[Or]", 
    RowBox[{"x", "\[Equal]", "y"}]}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"x", 
   UnderscriptBox["\[LessEqual]", "dom"], "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(<\[Or]=)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:966362506", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600151597180127*^9, 3.600151662995488*^9}, {
  3.600151696642337*^9, 3.600151705919442*^9}, {3.600151743313932*^9, 
  3.600151744308773*^9}, {3.600152821204228*^9, 3.600152902955752*^9}, {
  3.600153177862607*^9, 3.600153201977138*^9}, {3.60040838083105*^9, 
  3.600408383932227*^9}, {3.600408802466825*^9, 3.600408804647085*^9}, {
  3.610274054519209*^9, 3.610274055521455*^9}},
 CellTags->{
  "ID:966362506", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<\[Or]="},
 CellID->966362506],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"x", 
     UnderscriptBox["\[LessEqual]", "dom"], "y"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLeftRightArrow]", 
  RowBox[{"y", 
   UnderscriptBox["<", "dom"], "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Not]\[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1288924674", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600151753904079*^9, 3.600151789263146*^9}, 
   3.600153260272639*^9, {3.600408393508057*^9, 3.600408396416486*^9}, {
   3.600408806524342*^9, 3.600408809404616*^9}, {3.60645946554104*^9, 
   3.606459474230033*^9}, {3.60973930210693*^9, 3.60973931116013*^9}, {
   3.610274057189733*^9, 3.610274058303236*^9}},
 CellTags->{
  "ID:1288924674", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[Not]\[LessEqual]"},
 CellID->1288924674],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"x", 
     UnderscriptBox["<", "dom"], "y"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLeftRightArrow]", 
  RowBox[{"y", 
   UnderscriptBox["\[LessEqual]", "dom"], "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Not]<)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1962569111", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600151811811287*^9, 3.600151826840223*^9}, 
   3.600153261525768*^9, {3.600408398900475*^9, 3.600408401925904*^9}, {
   3.600408810951854*^9, 3.600408812949665*^9}, {3.606459477381242*^9, 
   3.606459492825698*^9}, {3.609739314132977*^9, 3.609739318077506*^9}, {
   3.610274059861014*^9, 3.61027406245549*^9}},
 CellTags->{
  "ID:1962569111", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[Not]<"},
 CellID->1962569111],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["<", "dom"], "y"}], "\[And]", 
    RowBox[{"y", 
     UnderscriptBox["\[LessEqual]", "dom"], "z"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", 
   UnderscriptBox["<", "dom"], "z"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(transitivity <\[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:243913977", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600152759314976*^9, 3.600152779485783*^9}, 
   3.600153262485871*^9, {3.600408404505347*^9, 3.600408410369903*^9}, {
   3.60040881519463*^9, 3.600408819509857*^9}, {3.600408851713349*^9, 
   3.600408861573107*^9}, {3.610274063791624*^9, 3.610274067703878*^9}},
 CellTags->{
  "ID:243913977", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity <\[LessEqual]"},
 CellID->243913977],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["\[LessEqual]", "dom"], "y"}], "\[And]", 
    RowBox[{"y", 
     UnderscriptBox["<", "dom"], "z"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", 
   UnderscriptBox["<", "dom"], "z"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(transitivity \[LessEqual]<)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1623404218", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600152918959816*^9, 3.600152930028819*^9}, 
   3.600153263629775*^9, {3.600408412985893*^9, 3.600408418340538*^9}, {
   3.600408821241819*^9, 3.600408827771594*^9}, {3.600408864447555*^9, 
   3.600408867619521*^9}, {3.610274069113625*^9, 3.610274072046289*^9}},
 CellTags->{
  "ID:1623404218", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity \[LessEqual]<"},
 CellID->1623404218],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["<", "dom"], "y"}], "\[And]", 
    RowBox[{"y", 
     UnderscriptBox["<", "dom"], "z"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", 
   UnderscriptBox["<", "dom"], "z"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(transitivity <<)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1509384446", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600593978065586*^9, 3.600593980126014*^9}, {
  3.610274073344627*^9, 3.610274075459983*^9}},
 CellTags->{
  "ID:1509384446", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity <<"},
 CellID->1509384446],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[NotEqual]", "y"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["<", "dom"], "y"}], "\[Or]", 
    RowBox[{"y", 
     UnderscriptBox["<", "dom"], "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[NotEqual] in terms of <)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:972758509", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600421544454903*^9, 3.600421566675442*^9}, {
  3.6064594446844387`*^9, 3.60645945209377*^9}, {3.610274076684621*^9, 
  3.610274077687504*^9}},
 CellTags->{
  "ID:972758509", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[NotEqual] in terms of <"},
 CellID->972758509],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->931362008]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1684574786],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(irreflexivity of <)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(irreflexivity of <)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isIrreflexive", "[", 
           UnderscriptBox["<", 
            StyleBox["dom", "ExpressionVariable"]], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isIrreflexive", "[", 
           UnderscriptBox["<", 
            StyleBox["dom", "ExpressionVariable"]], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1308600401"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1308600401-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1308600401-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1308600401", "Proof|ID:1308600401-1"},
 CellID->21123431],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1308600401-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1409828609]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reflexivity of \[LessEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reflexivity of \[LessEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], 
           UnderscriptBox["\[LessEqual]", 
            StyleBox["dom", "ExpressionVariable"]], 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], 
           UnderscriptBox["\[LessEqual]", 
            StyleBox["dom", "ExpressionVariable"]], 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:914659821"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p914659821-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p914659821-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:914659821", "Proof|ID:914659821-1"},
 CellID->388106339],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p914659821-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->496671473]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(<\[Or]=)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(<\[Or]=)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["<", 
                  StyleBox["dom", "ExpressionVariable"]], 
                 StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["y", "ExpressionVariable"]}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[LessEqual]", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["<", 
                  StyleBox["dom", "ExpressionVariable"]], 
                 StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["y", "ExpressionVariable"]}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[LessEqual]", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:966362506"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p966362506-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p966362506-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:966362506", "Proof|ID:966362506-1"},
 CellID->831539329],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p966362506-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.61795983998648*^9, 3.617959842890036*^9}},
 CellID->440020905]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Not]\[LessEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Not]\[LessEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", 
                    StyleBox["dom", "ExpressionVariable"]], 
                   StyleBox["y", "ExpressionVariable"]}], 
                 TagBox[")", "AutoParentheses"]}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", 
                    StyleBox["dom", "ExpressionVariable"]], 
                   StyleBox["y", "ExpressionVariable"]}], 
                 TagBox[")", "AutoParentheses"]}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1288924674"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1288924674-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1288924674-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1288924674", "Proof|ID:1288924674-1"},
 CellID->1029351046],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1288924674-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1669987450]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Not]<)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Not]<)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], 
                   UnderscriptBox["<", 
                    StyleBox["dom", "ExpressionVariable"]], 
                   StyleBox["y", "ExpressionVariable"]}], 
                 TagBox[")", "AutoParentheses"]}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["\[LessEqual]", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], 
                   UnderscriptBox["<", 
                    StyleBox["dom", "ExpressionVariable"]], 
                   StyleBox["y", "ExpressionVariable"]}], 
                 TagBox[")", "AutoParentheses"]}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["\[LessEqual]", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1962569111"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1962569111-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1962569111-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1962569111", "Proof|ID:1962569111-1"},
 CellID->1777374244],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1962569111-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1678900439]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(transitivity <\[LessEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(transitivity <\[LessEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:243913977"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p243913977-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p243913977-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:243913977", "Proof|ID:243913977-1"},
 CellID->1987732976],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p243913977-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1471369452]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(transitivity \[LessEqual]<)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(transitivity \[LessEqual]<)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1623404218"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1623404218-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1623404218-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1623404218", "Proof|ID:1623404218-1"},
 CellID->1964215912],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1623404218-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1573508644]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(transitivity <<)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(transitivity <<)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1509384446"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1509384446-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1509384446-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1509384446", "Proof|ID:1509384446-1"},
 CellID->295324759],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1509384446-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->572410148]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[NotEqual] in terms of <)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[NotEqual] in terms of <)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["y", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["<", 
                  StyleBox["dom", "ExpressionVariable"]], 
                 StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], 
                 UnderscriptBox["<", 
                  StyleBox["dom", "ExpressionVariable"]], 
                 StyleBox["x", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTotallyOrdered", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["y", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["<", 
                  StyleBox["dom", "ExpressionVariable"]], 
                 StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], 
                 UnderscriptBox["<", 
                  StyleBox["dom", "ExpressionVariable"]], 
                 StyleBox["x", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:972758509"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p972758509-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p972758509-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:972758509", "Proof|ID:972758509-1"},
 CellID->1361125711],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p972758509-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->436371387]
}, Closed]],

Cell[TextData[{
 "Now that the first very basic properties of order relations have been \
established, it is time to turn to totally-ordered Abelian monoids. In \
particular, we want to prove further connections between the monoid operation \
",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["+", "dom"], TraditionalForm]]],
 " and the order relations ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "dom"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["<", "dom"], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.603600299033111*^9, 3.603600575835873*^9}, {
  3.604032463012604*^9, 3.604032467619874*^9}, {3.615026444679683*^9, 
  3.615026462792791*^9}, {3.617952576853622*^9, 3.617952578598996*^9}, {
  3.617953172263185*^9, 3.617953173540694*^9}},
 CellID->2080376934],

Cell["", "OpenEnvironment",
 CellID->652155032],

Cell[CellGroupData[{

Cell["Lemma: Properties of totally-ordered commutative monoids", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.600593215848107*^9, 3.600593234965672*^9}, {
  3.617952571960853*^9, 3.617952573903943*^9}},
 CellID->1674491660],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isTOM", "[", "dom", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.610274081606257*^9, 3.610274085367681*^9}},
 CellID->1774773558],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "dom"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y", ",", "u", ",", "v"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.600593258270857*^9, 3.600593269080491*^9}, 
   3.610274087555465*^9, {3.615611401834132*^9, 3.615611409623683*^9}},
 CellID->1799285719],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"u", 
    UnderscriptBox["+", "dom"], "x"}], 
   UnderscriptBox["\[LessEqual]", "dom"], 
   RowBox[{"u", 
    UnderscriptBox["+", "dom"], "y"}]}], "\[DoubleLeftRightArrow]", 
  RowBox[{"x", 
   UnderscriptBox["\[LessEqual]", "dom"], "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LessEqual]+)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:786186036", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.615611387269375*^9, 3.615611427756548*^9}, {
  3.615623483867785*^9, 3.615623485204834*^9}},
 CellTags->{
  "ID:786186036", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]+"},
 CellID->786186036],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"u", 
    UnderscriptBox["+", "dom"], "x"}], 
   UnderscriptBox["<", "dom"], 
   RowBox[{"u", 
    UnderscriptBox["+", "dom"], "y"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"x", 
   UnderscriptBox["<", "dom"], "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(<+)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1385901945", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6005933426447*^9, 3.600593380093335*^9}, {
  3.602999321965249*^9, 3.602999341687372*^9}, {3.610274088855683*^9, 
  3.610274091796108*^9}},
 CellTags->{
  "ID:1385901945", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<+"},
 CellID->1385901945],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["\[LessEqual]", "dom"], "y"}], "\[And]", 
    RowBox[{"u", 
     UnderscriptBox["\[LessEqual]", "dom"], "v"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"x", 
    UnderscriptBox["+", "dom"], "u"}], 
   UnderscriptBox["\[LessEqual]", "dom"], 
   RowBox[{"y", 
    UnderscriptBox["+", "dom"], "v"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LessEqual]\[LessEqual]+)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1581648423", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600593277775005*^9, 3.600593308418406*^9}, {
  3.61027409316867*^9, 3.61027409700149*^9}},
 CellTags->{
  "ID:1581648423", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]\[LessEqual]+"},
 CellID->1581648423],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["\[LessEqual]", "dom"], "y"}], "\[And]", 
    RowBox[{"u", 
     UnderscriptBox["<", "dom"], "v"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"x", 
    UnderscriptBox["+", "dom"], "u"}], 
   UnderscriptBox["<", "dom"], 
   RowBox[{"y", 
    UnderscriptBox["+", "dom"], "v"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LessEqual]<+)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1815472507", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600593866225422*^9, 3.600593882143146*^9}, {
  3.610274098521497*^9, 3.610274101976936*^9}},
 CellTags->{
  "ID:1815472507", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]<+"},
 CellID->1815472507],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["<", "dom"], "y"}], "\[And]", 
    RowBox[{"u", 
     UnderscriptBox["<", "dom"], "v"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"x", 
    UnderscriptBox["+", "dom"], "u"}], 
   UnderscriptBox["<", "dom"], 
   RowBox[{"y", 
    UnderscriptBox["+", "dom"], "v"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(<<+)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:443432455", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600593888023358*^9, 3.600593891927727*^9}, {
  3.610274103423263*^9, 3.610274107576602*^9}},
 CellTags->{
  "ID:443432455", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<<+"},
 CellID->443432455],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2069740564]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2028534090],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LessEqual]+)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LessEqual]+)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTOM", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox["\[LessEqual]", 
              StyleBox["dom", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[LessEqual]", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTOM", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox["\[LessEqual]", 
              StyleBox["dom", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[LessEqual]", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:786186036"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p786186036-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p786186036-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:786186036", "Proof|ID:786186036-1"},
 CellID->1923006109],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p786186036-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1549560626]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(<+)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(<+)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTOM", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTOM", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1385901945"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1385901945-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1385901945-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1385901945", "Proof|ID:1385901945-1"},
 CellID->263629365],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1385901945-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellID->701073402]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LessEqual]\[LessEqual]+)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LessEqual]\[LessEqual]+)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTOM", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["v", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["u", "ExpressionVariable"]}], 
             UnderscriptBox["\[LessEqual]", 
              StyleBox["dom", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox[
               "v", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTOM", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["v", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["u", "ExpressionVariable"]}], 
             UnderscriptBox["\[LessEqual]", 
              StyleBox["dom", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox[
               "v", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1581648423"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1581648423-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1581648423-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1581648423", "Proof|ID:1581648423-1"},
 CellID->422549508],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1581648423-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1666906397]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LessEqual]<+)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LessEqual]<+)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTOM", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["v", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["u", "ExpressionVariable"]}], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox[
               "v", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTOM", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["v", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["u", "ExpressionVariable"]}], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox[
               "v", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1815472507"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1815472507-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1815472507-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1815472507", "Proof|ID:1815472507-1"},
 CellID->1818000638],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1815472507-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellID->743904775]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(<<+)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(<<+)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTOM", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["v", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["u", "ExpressionVariable"]}], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox[
               "v", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTOM", "[", 
               StyleBox["dom", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["<", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["v", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox["u", "ExpressionVariable"]}], 
             UnderscriptBox["<", 
              StyleBox["dom", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["dom", "ExpressionVariable"]], 
               StyleBox[
               "v", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:443432455"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p443432455-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p443432455-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:443432455", "Proof|ID:443432455-1"},
 CellID->633496152],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p443432455-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1736685396]
}, Closed]],

Cell[TextData[{
 "It is important to note that for proving the latter three formulas one \
cannot use the inference rule \[OpenCurlyDoubleQuote]substitution based on \
quantified equalities and equivalences\[CloseCurlyDoubleQuote], because we \
make use of the commutativity law. Having the mentioned inference rule \
activated would cause an infinite loop of applications of exactly this rule, \
because every sum ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "+", "y"}], TraditionalForm]]],
 " could then be rewritten into ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", "+", "x"}], TraditionalForm]]],
 " and back."
}], "Text",
 CellChangeTimes->{{3.605435554698072*^9, 3.605435560231609*^9}, {
  3.605435602913091*^9, 3.60543571948091*^9}, {3.615026488975389*^9, 
  3.615026489718644*^9}},
 CellID->2067863327]
}, Closed]],

Cell[CellGroupData[{

Cell["Tuples", "Subsection",
 CellChangeTimes->{{3.600148092329953*^9, 3.600148093346615*^9}, {
  3.61095620182497*^9, 3.610956218729598*^9}},
 CellID->253205562],

Cell[TextData[{
 "Having some elementary properties and notions of tuples defined in section ",
 ButtonBox["Basic Notions / General Notions / Tuples",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::DefinitionTuples"],
 ", we can now proceed by proving simple properties of them. All of these \
properties are more or less trivial (meaning that they follow readily from \
the definitions), but later on, when really diving into the much harder \
proofs of the complexity analysis, it will be good to have them."
}], "Text",
 CellChangeTimes->{{3.603601914406233*^9, 3.603602079677798*^9}, {
  3.615026510049713*^9, 3.61502657465868*^9}},
 CellID->859597688],

Cell["", "OpenEnvironment",
 CellID->998146153],

Cell[CellGroupData[{

Cell["Lemma: Properties of tuples", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599283389008015*^9, 3.599283393510412*^9}, 
   3.603014574292994*^9},
 CellID->1763958272],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"isTuple", "[", "A", "]"}], ",", "x", ",", "y", ",", "u", ",", 
   "v"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.59928339993839*^9, 3.599283404973195*^9}, {
   3.60006569074781*^9, 3.600065691145244*^9}, {3.602918745500581*^9, 
   3.602918745793523*^9}, 3.602919233470488*^9, 3.610874625800742*^9, {
   3.612523205473127*^9, 3.612523206734194*^9}},
 CellID->224379101],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    "True"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1\[LessEqual]length)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1546357463", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599906541171012*^9, 3.59990656098238*^9}, {
  3.600062486260737*^9, 3.600062528327593*^9}, {3.6031776469111*^9, 
  3.6031776522722*^9}, {3.603603757032916*^9, 3.603603759150127*^9}},
 CellTags->{
  "ID:1546357463", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1\[LessEqual]length"},
 CellID->1546357463],

Cell[BoxData[
 RowBox[{
  RowBox[{"2", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  "\[Implies]", 
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(11)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:409966668", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600663751362094*^9, 3.600663755923966*^9}, {
  3.600670527770128*^9, 3.600670530544065*^9}, {3.612855237941746*^9, 
  3.612855246713245*^9}, {3.612860545259623*^9, 3.612860563829715*^9}},
 CellTags->{
  "ID:409966668", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "11"},
 CellID->409966668],

Cell[BoxData[
 RowBox[{
  RowBox[{"2", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  TagBox["\[Implies]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{"i", "\[NotEqual]", "j"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(2\[LessEqual]length)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1768549223", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612859820553953*^9, 3.612859829969424*^9}, 
   3.612859938054829*^9, {3.612859984737727*^9, 3.612859986597311*^9}, {
   3.612860124360581*^9, 3.612860125318914*^9}, {3.612862546795657*^9, 
   3.61286256251802*^9}, {3.61286294802849*^9, 3.612862948903371*^9}},
 CellTags->{
  "ID:1768549223", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "2\[LessEqual]length"},
 CellID->1768549223],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
  RowBox[{"[", 
   RowBox[{"\[LeftBracketingBar]", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "\[RightBracketingBar]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(length\:293a is integer)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1399858996", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611397835648872*^9, 3.611397847034197*^9}},
 CellTags->{
  "ID:1399858996", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a is integer"},
 CellID->1399858996],

Cell[BoxData[
 RowBox[{"1", 
  UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
  RowBox[{"\[LeftBracketingBar]", 
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], 
   "\[RightBracketingBar]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(length\:293a is positive)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1353010357", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.603605369684542*^9, 3.603605385393355*^9}},
 CellTags->{
  "ID:1353010357", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a is positive"},
 CellID->1353010357],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"1", ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]], 
    RowBox[{"[", "j", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"1", ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", 
        RowBox[{"A", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "x"}], "\[RightBracketingBar]"}]}]]], 
    RowBox[{"[", "j", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(12)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:763884008", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611989622426654*^9, 3.611989678276898*^9}, {
  3.611989778494546*^9, 3.611989785336346*^9}, {3.611990307058156*^9, 
  3.611990338636733*^9}, {3.611990960000793*^9, 3.611990997413845*^9}, {
  3.611991161894145*^9, 3.611991173128878*^9}, {3.611992682023747*^9, 
  3.611992682318707*^9}, {3.611995469966056*^9, 3.611995473472271*^9}, {
  3.612066000264276*^9, 3.612066003693505*^9}},
 CellTags->{
  "ID:763884008", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "12"},
 CellID->763884008],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"1", ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]], 
   RowBox[{"[", 
    RowBox[{"\[LeftBracketingBar]", 
     RowBox[{"A", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "x"}], "\[RightBracketingBar]"}], 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(13)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1442641059", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611992663944545*^9, 3.611992708269544*^9}, {
  3.612066028678908*^9, 3.612066030892237*^9}},
 CellTags->{
  "ID:1442641059", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "13"},
 CellID->1442641059],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "\[Equal]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"u", ",", "v"}], "\[RightAngleBracket]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[Equal]", "u"}], "\[And]", 
    RowBox[{"y", "\[Equal]", "v"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(14)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:215173194", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612523214202528*^9, 3.612523239895606*^9}},
 CellTags->{
  "ID:215173194", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "14"},
 CellID->215173194],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->274743996]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->183782654],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1\[LessEqual]length)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1\[LessEqual]length)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[DoubleLeftRightArrow]", 
           
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             "True"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[DoubleLeftRightArrow]", 
           
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             "True"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1546357463"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1546357463-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1546357463-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1546357463", "Proof|ID:1546357463-1"},
 CellID->1202974989],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1546357463-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1207870308]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(11)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(11)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:409966668"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p409966668-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p409966668-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:409966668", "Proof|ID:409966668-1"},
 CellID->577682993],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p409966668-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->924364061]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(2\[LessEqual]length)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(2\[LessEqual]length)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox[
                 "j", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox[
                 "j", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1768549223"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1768549223-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1768549223-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1768549223", "Proof|ID:1768549223-1"},
 CellID->1717553837],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1768549223-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->499654441]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(length\:293a is integer)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(length\:293a is integer)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], "[", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{"\[LeftBracketingBar]", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], 
               "\[RightBracketingBar]"}], 
             TagBox[")", "AutoParentheses"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], "[", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{"\[LeftBracketingBar]", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], 
               "\[RightBracketingBar]"}], 
             TagBox[")", "AutoParentheses"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1399858996"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1399858996-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1399858996-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1399858996", "Proof|ID:1399858996-1"},
 CellID->86355476],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1399858996-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1825116772]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(length\:293a is positive)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(length\:293a is positive)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{"1", 
           UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{"\[LeftBracketingBar]", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], 
               "\[RightBracketingBar]"}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{"1", 
           UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{"\[LeftBracketingBar]", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], 
               "\[RightBracketingBar]"}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1353010357"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1353010357-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1353010357-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1353010357", "Proof|ID:1353010357-1"},
 CellID->1804120756],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1353010357-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1951723367]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(12)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(12)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                SubscriptBox["\[DoubleStruckCapitalN]", 
                 RowBox[{"1", ",", "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]]], "[", 
               StyleBox["j", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            SubscriptBox["\[DoubleStruckCapitalN]", 
             RowBox[{"1", ",", "\[Ellipsis]", ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}]]], "[", 
           StyleBox["j", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                SubscriptBox["\[DoubleStruckCapitalN]", 
                 RowBox[{"1", ",", "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]]], "[", 
               StyleBox["j", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            SubscriptBox["\[DoubleStruckCapitalN]", 
             RowBox[{"1", ",", "\[Ellipsis]", ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}]]], "[", 
           StyleBox["j", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:763884008"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p763884008-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p763884008-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:763884008", "Proof|ID:763884008-1"},
 CellID->1513904826],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p763884008-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1881258437]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(13)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(13)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               UnderscriptBox["\[Element]", 
                SubscriptBox["\[DoubleStruckCapitalN]", 
                 RowBox[{"1", ",", "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]]], "[", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}], "]"}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               UnderscriptBox["\[Element]", 
                SubscriptBox["\[DoubleStruckCapitalN]", 
                 RowBox[{"1", ",", "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]]], "[", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}], "]"}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1442641059"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1442641059-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1442641059-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1442641059", "Proof|ID:1442641059-1"},
 CellID->1528692265],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1442641059-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1729488571]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(14)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(14)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["x", "ExpressionVariable"], ",", 
            StyleBox["y", "ExpressionVariable"], ",", 
            StyleBox["u", "ExpressionVariable"], ",", 
            StyleBox["v", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["u", "ExpressionVariable"], ",", 
                 StyleBox["v", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["u", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["v", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["x", "ExpressionVariable"], ",", 
            StyleBox["y", "ExpressionVariable"], ",", 
            StyleBox["u", "ExpressionVariable"], ",", 
            StyleBox["v", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["u", "ExpressionVariable"], ",", 
                 StyleBox["v", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["u", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["v", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:215173194"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p215173194-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p215173194-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:215173194", "Proof|ID:215173194-1"},
 CellID->1068613755],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p215173194-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 32.}}]], "ProofInfoBody",
 CellID->1165482879]
}, Closed]],

Cell[TextData[{
 "Finally, we state yet another useful lemma which relates the ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[SquareSubsetEqual]", "dom"], TraditionalForm]]],
 " predicate to ",
 Cell[BoxData[
  FormBox["\:293a", TraditionalForm]]],
 ". Apparently, formula ",
 ButtonBox["(\[SquareSubsetEqual]\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:759960416"],
 " is just the combination of formulas ",
 ButtonBox["(15)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:8303793"],
 " and ",
 ButtonBox["(16)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1339530720"],
 "."
}], "Text",
 CellChangeTimes->{{3.603602296693901*^9, 3.603602333242089*^9}, {
  3.603602444438198*^9, 3.603602577695429*^9}, {3.610274635789369*^9, 
  3.610274635789481*^9}, {3.615026605565886*^9, 3.615026696244036*^9}, {
  3.61578749253832*^9, 3.615787496160466*^9}},
 CellID->1820728087],

Cell["", "OpenEnvironment",
 CellID->131337638],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["\[SquareSubsetEqual]", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.601886407182152*^9, 3.601886423175953*^9}},
 CellID->210695348],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"dom", ",", "A", ",", "x"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.601886430958169*^9, 3.601886445468501*^9}, {
  3.61027461011321*^9, 3.610274610382131*^9}},
 CellID->924384146],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[SquareSubsetEqual]", "dom"], 
   RowBox[{"[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[SquareSubsetEqual]", "dom"], 
     RowBox[{"[", "A", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "dom"], 
     RowBox[{"[", "x", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(15)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:8303793", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.602921526477881*^9, 3.602921536175575*^9}, {
  3.610274613078844*^9, 3.610274623463276*^9}},
 CellTags->{
  "ID:8303793", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "15"},
 CellID->8303793],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[SquareSubsetEqual]", "dom"], 
     RowBox[{"[", "A", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "dom"], 
     RowBox[{"[", "x", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[SquareSubsetEqual]", "dom"], 
   RowBox[{"[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(16)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1339530720", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.602921544773176*^9, 3.602921550218191*^9}, {
  3.610274624983505*^9, 3.610274627131794*^9}},
 CellTags->{
  "ID:1339530720", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "16"},
 CellID->1339530720],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[SquareSubsetEqual]", "dom"], 
   RowBox[{"[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[SquareSubsetEqual]", "dom"], 
     RowBox[{"[", "A", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "dom"], 
     RowBox[{"[", "x", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[SquareSubsetEqual]\:293a)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:759960416", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601886450613625*^9, 3.601886491717598*^9}, {
  3.610274628318759*^9, 3.610274630361499*^9}},
 CellTags->{
  "ID:759960416", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[SquareSubsetEqual]\:293a"},
 CellID->759960416],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->730215235]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1781807047],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(15)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(15)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["dom", "ExpressionVariable"], ",", 
            StyleBox["A", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderscriptBox["\[SquareSubsetEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               TagBox[")", "AutoParentheses"]}], "[", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], "\:293a", 
               StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", 
                  StyleBox["dom", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["dom", "ExpressionVariable"], ",", 
            StyleBox["A", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderscriptBox["\[SquareSubsetEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               TagBox[")", "AutoParentheses"]}], "[", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], "\:293a", 
               StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", 
                  StyleBox["dom", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:8303793"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p8303793-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:8303793|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->222527391],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p8303793-display.m\"}]"]],
     None], StandardForm],
  ImageSizeCache->{298., {24., 31.}}]], "ProofInfoBody",
 CellID->1226520191]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(16)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(16)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["dom", "ExpressionVariable"], ",", 
            StyleBox["A", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", 
                  StyleBox["dom", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderscriptBox["\[SquareSubsetEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               TagBox[")", "AutoParentheses"]}], "[", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], "\:293a", 
               StyleBox["x", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["dom", "ExpressionVariable"], ",", 
            StyleBox["A", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", 
                  StyleBox["dom", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["dom", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderscriptBox["\[SquareSubsetEqual]", 
                StyleBox["dom", "ExpressionVariable"]], 
               TagBox[")", "AutoParentheses"]}], "[", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], "\:293a", 
               StyleBox["x", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1339530720"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1339530720-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1339530720|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->980530952],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1339530720-display.m\"}]"]], None], StandardForm],
  ImageSizeCache->{298., {24., 31.}}]], "ProofInfoBody",
 CellID->447677394]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[SquareSubsetEqual]\:293a)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[SquareSubsetEqual]\:293a)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["A", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["D", "ExpressionVariable"], "[", 
               "\[SquareSubsetEqual]", "]"}], "[", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], "\:293a", 
               StyleBox["x", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["D", "ExpressionVariable"], "[", 
                   "\[SquareSubsetEqual]", "]"}], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["D", "ExpressionVariable"], "[", "\[Element]", 
                   "]"}], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["A", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["D", "ExpressionVariable"], "[", 
               "\[SquareSubsetEqual]", "]"}], "[", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], "\:293a", 
               StyleBox["x", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["D", "ExpressionVariable"], "[", 
                   "\[SquareSubsetEqual]", "]"}], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["D", "ExpressionVariable"], "[", "\[Element]", 
                   "]"}], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:759960416"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p759960416-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:759960416|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1092143830],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p759960416-display.m\"}]"]], None], StandardForm],
  ImageSizeCache->{298., {24., 31.}}]], "ProofInfoBody",
 CellID->790789736]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["min, max", "Subsection",
 CellChangeTimes->{{3.600148080105234*^9, 3.600148081745968*^9}, {
   3.603602798845528*^9, 3.603602801984599*^9}, 3.604993279877072*^9, {
   3.610956189176521*^9, 3.610956217770388*^9}},
 CellTags->"sect::Lemmataminmax",
 CellID->473572412],

Cell[TextData[{
 "The last notions we have to deal with are ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["min", "D"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["max", "D"], TraditionalForm]]],
 ", as defined in section ",
 ButtonBox["Basic Notions / Categories / min-max Domain",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::DefinitionminmaxDomain"],
 ". We mainly focus on how ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["min", "D"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["max", "D"], TraditionalForm]]],
 " are related to ",
 Cell[BoxData[
  FormBox["\:293a", TraditionalForm]]],
 " (formulas ",
 ButtonBox["(min\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1460290836"],
 " and ",
 ButtonBox["(max\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1931124773"],
 ") and to the monoid operation ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["+", "D"], TraditionalForm]]],
 " (formulas ",
 ButtonBox["(min+)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:446229453"],
 " and ",
 ButtonBox["(max+)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:513276857"],
 ").\nIn the whole subsection let ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " a tuple of elements of the a.b.f. totally-ordered Abelian monoid ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.603602735074889*^9, 3.603602812990941*^9}, {
  3.603602933138134*^9, 3.603602958865078*^9}, {3.603602996737175*^9, 
  3.603603038514579*^9}, {3.603603130994815*^9, 3.603603338040845*^9}, {
  3.603603430387075*^9, 3.603603477450138*^9}, {3.603605079068479*^9, 
  3.60360517148865*^9}, {3.60499189704529*^9, 3.604991938620311*^9}, {
  3.604991988343157*^9, 3.604992027140191*^9}, {3.604993286111972*^9, 
  3.604993286628519*^9}, {3.60499332061915*^9, 3.604993372406368*^9}, {
  3.605587450938437*^9, 3.605587462277686*^9}, {3.605589115249729*^9, 
  3.605589118045876*^9}, {3.613965187241821*^9, 3.613965190394648*^9}, {
  3.614673581694599*^9, 3.614673623466027*^9}, {3.615026779836103*^9, 
  3.615026804369013*^9}, {3.6150272639469*^9, 3.6150273883205*^9}},
 CellID->1049858421],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[SquareSubsetEqual]", "D"], 
   RowBox[{"[", "A", "]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.605589096673661*^9, 3.605589104056554*^9}, {
  3.605589184206278*^9, 3.605589200472976*^9}},
 CellID->931803829],

Cell["", "OpenEnvironment",
 CellID->1523217451],

Cell[CellGroupData[{

Cell["Lemma: Defining Properties of min, max", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.61396511864146*^9, 3.613965129645724*^9}},
 CellID->1515714911],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.61396514844907*^9, 3.613965158034677*^9}},
 CellID->656269750],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   RowBox[{
    UnderscriptBox["min", "D"], "[", "A", "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   SubscriptBox["A", "i"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(defining property of min)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1448857518", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600584546336778*^9, 3.600584588978607*^9}, {
  3.604993026390648*^9, 3.60499302898441*^9}},
 CellTags->{
  "ID:1448857518", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of min"},
 CellID->1448857518],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   SubscriptBox["A", "i"], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{
    UnderscriptBox["max", "D"], "[", "A", "]"}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(defining property of max)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1578657786", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600584599158454*^9, 3.600584612360336*^9}, {
  3.604993032207085*^9, 3.604993034093701*^9}},
 CellTags->{
  "ID:1578657786", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of max"},
 CellID->1578657786],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["min", "D"], "[", "A", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of min)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1968885164", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601711833584371*^9, 3.601711848787337*^9}, {
  3.604993036383906*^9, 3.604993038363262*^9}},
 CellTags->{
  "ID:1968885164", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of min"},
 CellID->1968885164],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["max", "D"], "[", "A", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of max)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1151569277", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601711866490219*^9, 3.601711868419515*^9}, {
  3.604993041207148*^9, 3.604993043112611*^9}},
 CellTags->{
  "ID:1151569277", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of max"},
 CellID->1151569277],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1603322875]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1913696948],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(defining property of min)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(defining property of min)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1448857518"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1448857518-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1448857518-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1448857518", "Proof|ID:1448857518-1"},
 CellID->1149489714],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1448857518-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1171443524]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(defining property of max)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(defining property of max)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1578657786"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1578657786-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1578657786-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1578657786", "Proof|ID:1578657786-1"},
 CellID->1807688628],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1578657786-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1889510511]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of min)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of min)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1968885164"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1968885164-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1968885164-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1968885164", "Proof|ID:1968885164-1"},
 CellID->1971047889],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1968885164-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->773286761]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of max)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of max)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1151569277"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1151569277-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1151569277-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1151569277", "Proof|ID:1151569277-1"},
 CellID->2058731697],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1151569277-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->158595008]
}, Closed]],

Cell[TextData[{
 "In the rest of this subsection, let ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " always be two elements of ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.613965202387457*^9, 3.613965220368433*^9}},
 CellID->1149895741],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y"}], "]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.599194006684617*^9, 3.599194007529624*^9}, {
  3.599194081206328*^9, 3.59919409060399*^9}, {3.59919458809842*^9, 
  3.599194591536751*^9}, {3.600058827890865*^9, 3.600058870706449*^9}, {
  3.600058966065933*^9, 3.600058967657179*^9}, {3.600062819511432*^9, 
  3.600062826119807*^9}, {3.600064628457687*^9, 3.600064628862984*^9}, {
  3.600150967517769*^9, 3.600150967886192*^9}, {3.605589054800964*^9, 
  3.605589105752696*^9}},
 CellID->1335750375],

Cell[TextData[{
 "The next lemma deals exclusively with ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["min", "D"], TraditionalForm]]],
 "/",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["max", "D"], TraditionalForm]]],
 " of pairs:"
}], "Text",
 CellChangeTimes->{{3.615027442016286*^9, 3.615027463235165*^9}, {
  3.615027857687958*^9, 3.615027867771277*^9}},
 CellID->744956650],

Cell["", "OpenEnvironment",
 CellID->1583106624],

Cell[CellGroupData[{

Cell["Lemma: Properties of min, max of pairs", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.612693325088987*^9, 3.612693335506745*^9}},
 CellID->345801648],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["\[LessEqual]", "D"], "y"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["min", "D"], "[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
     "x"}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["max", "D"], "[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
     "y"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(17)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:232133128", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599194883433239*^9, 3.599194929686728*^9}, {
  3.599989342012971*^9, 3.599989348783473*^9}, {3.600064634752158*^9, 
  3.600064640984613*^9}, {3.600064693034253*^9, 3.600064710176384*^9}, {
  3.600064941530039*^9, 3.600064945374969*^9}, {3.600409851101407*^9, 
  3.600409857899563*^9}, {3.604993046483323*^9, 3.604993052842169*^9}},
 CellTags->{
  "ID:232133128", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "17"},
 CellID->232133128],

Cell[BoxData[
 RowBox[{
  RowBox[{"y", 
   UnderscriptBox["\[LessEqual]", "D"], "x"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["min", "D"], "[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
     "y"}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["max", "D"], "[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
     "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(18)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1533503972", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599194955167439*^9, 3.599194972079623*^9}, {
  3.599989365192291*^9, 3.599989376771379*^9}, {3.600064645768506*^9, 
  3.600064663000714*^9}, {3.600064723170302*^9, 3.600064727265368*^9}, {
  3.600409862319442*^9, 3.600409876917877*^9}, {3.60360341872803*^9, 
  3.603603419604774*^9}, {3.604993056018102*^9, 3.604993062160953*^9}},
 CellTags->{
  "ID:1533503972", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "18"},
 CellID->1533503972],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["<", "D"], "y"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["min", "D"], "[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
     "x"}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["max", "D"], "[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
     "y"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(17')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:881325315", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612692127115062*^9, 3.612692129509184*^9}},
 CellTags->{
  "ID:881325315", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "17'"},
 CellID->881325315],

Cell[BoxData[
 RowBox[{
  RowBox[{"y", 
   UnderscriptBox["<", "D"], "x"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["min", "D"], "[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
     "y"}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["max", "D"], "[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
     "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(18')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1667609383", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612692141471193*^9, 3.612692143608688*^9}},
 CellTags->{
  "ID:1667609383", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "18'"},
 CellID->1667609383],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["min", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["min", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"y", ",", "x"}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min is invariant under permutations)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1257756255", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613964912020554*^9, 3.613964930171185*^9}},
 CellTags->{
  "ID:1257756255", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min is invariant under permutations"},
 CellID->1257756255],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"y", ",", "x"}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max is invariant under permutations)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2071551441", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613964961841322*^9, 3.613964965836217*^9}},
 CellTags->{
  "ID:2071551441", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max is invariant under permutations"},
 CellID->2071551441],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1103232359]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->678471395],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(17)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(17)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[LessEqual]", "D"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox["x", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[LessEqual]", "D"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox["x", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:232133128"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p232133128-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p232133128-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:232133128", "Proof|ID:232133128-1"},
 CellID->557201343],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p232133128-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1057192124]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(18)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(18)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["\[LessEqual]", "D"], 
             StyleBox["x", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox[
               "x", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["\[LessEqual]", "D"], 
             StyleBox["x", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox[
               "x", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1533503972"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1533503972-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1533503972-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1533503972", "Proof|ID:1533503972-1"},
 CellID->1508553184],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1533503972-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->849429518]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(17')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(17')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["<", "D"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox["x", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["<", "D"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox["x", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:881325315"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p881325315-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p881325315-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:881325315", "Proof|ID:881325315-1"},
 CellID->803466674],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p881325315-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1693750335]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(18')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(18')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["<", "D"], 
             StyleBox["x", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox[
               "x", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["<", "D"], 
             StyleBox["x", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
               StyleBox[
               "x", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1667609383"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1667609383-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1667609383-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1667609383", "Proof|ID:1667609383-1"},
 CellID->1598215446],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1667609383-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->2061132157]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(min is invariant under permutations)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(min is invariant under permutations)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["min", "D"], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["min", "D"], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["min", "D"], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["min", "D"], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1257756255"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1257756255-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1257756255-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1257756255", "Proof|ID:1257756255-1"},
 CellID->725596050],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1257756255-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1826872359]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(max is invariant under permutations)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(max is invariant under permutations)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["max", "D"], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["max", "D"], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["max", "D"], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["max", "D"], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2071551441"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2071551441-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2071551441-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2071551441", "Proof|ID:2071551441-1"},
 CellID->1663975332],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2071551441-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1814758527]
}, Closed]],

Cell[TextData[{
 "The following lemma describes how ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["min", "D"], "[", "A", "]"}], 
    UnderscriptBox["\[LessEqual]", "D"], "x"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["\[LessEqual]", "D"], 
    RowBox[{
     UnderscriptBox["min", "D"], "[", "A", "]"}]}], TraditionalForm]]],
 " etc. can be expressed in terms of formulas where ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["min", "D"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["max", "D"], TraditionalForm]]],
 " do not appear any more. Please note that it justifies the four inferences \
rules \[OpenCurlyDoubleQuote]min[S]<x in the goal\[CloseCurlyDoubleQuote], \
\[OpenCurlyDoubleQuote]x<max[S] in the goal\[CloseCurlyDoubleQuote], \
\[OpenCurlyDoubleQuote]x<min[S] among the assumptions\[CloseCurlyDoubleQuote] \
and \[OpenCurlyDoubleQuote]max[S]<x among the assumptions\
\[CloseCurlyDoubleQuote] of the ComplexityProver."
}], "Text",
 CellChangeTimes->{
  3.60360368445187*^9, {3.603604194231604*^9, 3.603604291866544*^9}, {
   3.603604807024648*^9, 3.603604812076992*^9}, {3.604992223299622*^9, 
   3.604992255108216*^9}, 3.604993290094897*^9, {3.605436023298079*^9, 
   3.605436034837252*^9}, {3.615027473712696*^9, 3.615027630158996*^9}, {
   3.615027819537107*^9, 3.615027846098542*^9}},
 CellID->304451747],

Cell["", "OpenEnvironment",
 CellID->3199803],

Cell[CellGroupData[{

Cell["Lemma: General Properties of min, max", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599194265714582*^9, 3.599194271016751*^9}, {
   3.599195034518517*^9, 3.599195038322216*^9}, {3.600578822373432*^9, 
   3.600578823710445*^9}, {3.603604799048489*^9, 3.603604799493069*^9}, 
   3.615093169773813*^9},
 CellID->1111682865],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.600150948378549*^9, 3.600150954054202*^9}, {
  3.603177696425872*^9, 3.603177702224826*^9}, {3.603603735167146*^9, 
  3.603603737527564*^9}},
 CellID->1035719112],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["min", "D"], "[", "A", "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], "x"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    SubscriptBox["A", "i"], 
    UnderscriptBox["\[LessEqual]", "D"], "x"}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min\[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:48233088", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599194599440777*^9, 3.599194607712185*^9}, {
  3.599194642361054*^9, 3.599194669550129*^9}, {3.600061925463137*^9, 
  3.60006193461923*^9}, {3.600409791991877*^9, 3.6004098118111*^9}, {
  3.612693833247243*^9, 3.61269383612306*^9}},
 CellTags->{
  "ID:48233088", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\[LessEqual]"},
 CellID->48233088],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{
    UnderscriptBox["min", "D"], "[", "A", "]"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{"x", 
    UnderscriptBox["\[LessEqual]", "D"], 
    SubscriptBox["A", "i"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LessEqual]min)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1034128987", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599194599440777*^9, 3.599194607712185*^9}, {
  3.599194642361054*^9, 3.599194669550129*^9}, {3.599194703600382*^9, 
  3.599194722749772*^9}, {3.600063463239674*^9, 3.600063472594537*^9}, {
  3.60040981796556*^9, 3.600409828093958*^9}, {3.612693842751016*^9, 
  3.612693844964472*^9}},
 CellTags->{
  "ID:1034128987", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]min"},
 CellID->1034128987],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["max", "D"], "[", "A", "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], "x"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    SubscriptBox["A", "i"], 
    UnderscriptBox["\[LessEqual]", "D"], "x"}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max\[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1323776552", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599194732924373*^9, 3.59919474136413*^9}, {
  3.600063483121098*^9, 3.600063485926034*^9}, {3.600409830821453*^9, 
  3.600409834141379*^9}, {3.612693848022927*^9, 3.61269384997937*^9}},
 CellTags->{
  "ID:1323776552", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\[LessEqual]"},
 CellID->1323776552],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{
    UnderscriptBox["max", "D"], "[", "A", "]"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{"x", 
    UnderscriptBox["\[LessEqual]", "D"], 
    SubscriptBox["A", "i"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LessEqual]max)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1520079070", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599194746714421*^9, 3.599194763870446*^9}, {
  3.600063489314623*^9, 3.600063492450695*^9}, {3.600409835726986*^9, 
  3.600409840272565*^9}, {3.61269385171308*^9, 3.612693853587017*^9}},
 CellTags->{
  "ID:1520079070", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]max"},
 CellID->1520079070],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["min", "D"], "[", "A", "]"}], 
   UnderscriptBox["<", "D"], "x"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    SubscriptBox["A", "i"], 
    UnderscriptBox["<", "D"], "x"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min<)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1860476328", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.60057914529631*^9, 3.600579150782155*^9}, {
  3.612693855725451*^9, 3.612693857567084*^9}},
 CellTags->{
  "ID:1860476328", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min<"},
 CellID->1860476328],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["<", "D"], 
   RowBox[{
    UnderscriptBox["min", "D"], "[", "A", "]"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{"x", 
    UnderscriptBox["<", "D"], 
    SubscriptBox["A", "i"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(<min)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1907558186", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600579174274164*^9, 3.600579177111517*^9}, {
  3.612693859963209*^9, 3.612693861662419*^9}},
 CellTags->{
  "ID:1907558186", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<min"},
 CellID->1907558186],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["max", "D"], "[", "A", "]"}], 
   UnderscriptBox["<", "D"], "x"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    SubscriptBox["A", "i"], 
    UnderscriptBox["<", "D"], "x"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max<)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2008421481", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600579229087084*^9, 3.600579232329108*^9}, {
  3.612693863635026*^9, 3.612693865407859*^9}},
 CellTags->{
  "ID:2008421481", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max<"},
 CellID->2008421481],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["<", "D"], 
   RowBox[{
    UnderscriptBox["max", "D"], "[", "A", "]"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{"x", 
    UnderscriptBox["<", "D"], 
    SubscriptBox["A", "i"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(<max)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1141242040", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600579237519279*^9, 3.600579241350856*^9}, {
  3.612693868503028*^9, 3.612693870349116*^9}},
 CellTags->{
  "ID:1141242040", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<max"},
 CellID->1141242040],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->562397624]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->954132810],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(min\[LessEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(min\[LessEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox[
                 "x", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox[
                 "x", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:48233088"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p48233088-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p48233088-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:48233088", "Proof|ID:48233088-1"},
 CellID->2119117596],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p48233088-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1487833047]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LessEqual]min)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LessEqual]min)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1034128987"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1034128987-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1034128987-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1034128987", "Proof|ID:1034128987-1"},
 CellID->1605499322],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1034128987-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->980083213]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(max\[LessEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(max\[LessEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox[
                 "x", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox[
                 "x", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1323776552"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1323776552-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1323776552-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1323776552", "Proof|ID:1323776552-1"},
 CellID->87934740],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1323776552-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1396094928]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LessEqual]max)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LessEqual]max)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1520079070"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1520079070-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1520079070-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1520079070", "Proof|ID:1520079070-1"},
 CellID->650805455],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1520079070-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1652314721]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(min<)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(min<)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["<", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 StyleBox[
                 "x", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["<", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 StyleBox[
                 "x", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1860476328"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1860476328-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1860476328-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1860476328", "Proof|ID:1860476328-1"},
 CellID->489839113],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1860476328-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->58718705]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(<min)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(<min)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["<", "D"], 
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["<", "D"], 
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1907558186"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1907558186-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1907558186-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1907558186", "Proof|ID:1907558186-1"},
 CellID->846360522],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1907558186-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1661282637]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(max<)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(max<)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["<", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 StyleBox[
                 "x", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["<", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 StyleBox[
                 "x", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2008421481"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2008421481-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2008421481-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2008421481", "Proof|ID:2008421481-1"},
 CellID->1454120988],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2008421481-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->2055319065]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(<max)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(<max)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["<", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["<", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1141242040"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1141242040-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1141242040-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1141242040", "Proof|ID:1141242040-1"},
 CellID->155865952],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1141242040-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1159855003]
}, Closed]],

Cell[TextData[{
 "Lemma \[OpenCurlyDoubleQuote]min, max w.r.t. append Elements\
\[CloseCurlyDoubleQuote] below expresses terms of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["min", "D"], "[", 
    RowBox[{"A", " ", "\:293a", " ", "x"}], "]"}], TraditionalForm]]],
 " in terms of simpler expressions where the append-operator does not appear \
any more."
}], "Text",
 CellChangeTimes->{{3.615027677710081*^9, 3.615027783397811*^9}, {
  3.615027814451365*^9, 3.615027814847287*^9}},
 CellID->445509307],

Cell["", "OpenEnvironment",
 CellID->50622437],

Cell[CellGroupData[{

Cell["Lemma: min, max w.r.t. appending Elements", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.60360482282265*^9, 3.603604830745801*^9}, {
  3.614488159053882*^9, 3.614488180241833*^9}},
 CellID->409867065],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     SubscriptBox["A", "i"], 
     UnderscriptBox["\[LessEqual]", "D"], "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(19)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:572537117", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610875024919163*^9, 3.610875030593346*^9}},
 CellTags->{
  "ID:572537117", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "19"},
 CellID->572537117],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     SubscriptBox["A", "i"], 
     UnderscriptBox["\[LessEqual]", "D"], "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["min", "D"], "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], "\[Equal]", 
   RowBox[{
    UnderscriptBox["min", "D"], "[", "A", "]"}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min\:293a 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2073093776", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599194883433239*^9, 3.599194929686728*^9}, {
  3.599194984180437*^9, 3.599194989202079*^9}, {3.599989401109965*^9, 
  3.599989408595806*^9}, {3.600064741431346*^9, 3.600064778315018*^9}, {
  3.600409888556448*^9, 3.600409891646856*^9}, {3.60068847111703*^9, 
  3.60068849270616*^9}, {3.601785961470872*^9, 3.601786128471712*^9}, {
  3.603177855026215*^9, 3.603177860928117*^9}, {3.603603801195529*^9, 
  3.603603803653601*^9}, {3.603604889854322*^9, 3.603604921554757*^9}, {
  3.604993161022017*^9, 3.604993167066571*^9}, {3.610875032548386*^9, 
  3.61087503369069*^9}},
 CellTags->{
  "ID:2073093776", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a 1"},
 CellID->2073093776],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{"x", 
     UnderscriptBox["\[LessEqual]", "D"], 
     SubscriptBox["A", "i"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["min", "D"], "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], "\[Equal]", 
   "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min\:293a 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1632846240", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.603604891672144*^9, 3.603604900283053*^9}, {
  3.604993168912972*^9, 3.604993170688332*^9}},
 CellTags->{
  "ID:1632846240", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a 2"},
 CellID->1632846240],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{"x", 
     UnderscriptBox["\[LessEqual]", "D"], 
     SubscriptBox["A", "i"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(20)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1146565830", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610876042557911*^9, 3.610876047460598*^9}},
 CellTags->{
  "ID:1146565830", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "20"},
 CellID->1146565830],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{"x", 
     UnderscriptBox["\[LessEqual]", "D"], 
     SubscriptBox["A", "i"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["max", "D"], "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], "\[Equal]", 
   RowBox[{
    UnderscriptBox["max", "D"], "[", "A", "]"}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max\:293a 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1172583037", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599194955167439*^9, 3.599195002691232*^9}, {
   3.599989420067258*^9, 3.599989427012683*^9}, {3.600064796937985*^9, 
   3.600064804722176*^9}, {3.600409893342524*^9, 3.600409895484553*^9}, 
   3.600688498512469*^9, {3.601786180947954*^9, 3.601786200239542*^9}, {
   3.60317786545353*^9, 3.603177872443061*^9}, {3.603603806282537*^9, 
   3.603603808777815*^9}, 3.603604962189423*^9, {3.603604995770229*^9, 
   3.60360501123587*^9}, {3.604993173163022*^9, 3.604993179187822*^9}, {
   3.610876051220558*^9, 3.610876052252391*^9}},
 CellTags->{
  "ID:1172583037", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a 1"},
 CellID->1172583037],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     SubscriptBox["A", "i"], 
     UnderscriptBox["\[LessEqual]", "D"], "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["max", "D"], "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], "\[Equal]", 
   "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max\:293a 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2142810348", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.603604964081396*^9, 3.603604972395655*^9}, {
  3.604993181550184*^9, 3.604993183362601*^9}},
 CellTags->{
  "ID:2142810348", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a 2"},
 CellID->2142810348],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.613965897635568*^9, 3.613965907096552*^9}},
 CellID->419566097],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["min", "D"], "[", 
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["min", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["min", "D"], "[", "A", "]"}], ",", "x"}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min\:293a)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1460290836", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613965942424958*^9, 3.6139659697537*^9}},
 CellTags->{
  "ID:1460290836", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a"},
 CellID->1460290836],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["max", "D"], "[", "A", "]"}], ",", "x"}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max\:293a)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1931124773", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613971680754212*^9, 3.613971687707322*^9}},
 CellTags->{
  "ID:1931124773", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a"},
 CellID->1931124773],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["min", "D"], "[", 
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   UnderscriptBox["min", "D"], "[", "A", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min\:293a decreases)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:978507648", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.614488230318255*^9, 3.61448826292251*^9}},
 CellTags->{
  "ID:978507648", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a decreases"},
 CellID->978507648],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["max", "D"], "[", "A", "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max\:293a increases)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2097939500", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.614488281154134*^9, 3.614488290195345*^9}},
 CellTags->{
  "ID:2097939500", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a increases"},
 CellID->2097939500],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1315514221]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1827651235],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(19)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(19)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:572537117"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p572537117-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p572537117-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:572537117", "Proof|ID:572537117-1"},
 CellID->1768246037],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p572537117-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1139235207]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(min\:293a 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(min\:293a 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2073093776"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2073093776-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2073093776-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2073093776", "Proof|ID:2073093776-1"},
 CellID->470202616],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2073093776-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->986447845]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(min\:293a 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(min\:293a 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1632846240"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1632846240-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1632846240-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1632846240", "Proof|ID:1632846240-1"},
 CellID->163303679],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1632846240-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1508519340]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(20)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(20)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1146565830"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1146565830-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1146565830-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1146565830", "Proof|ID:1146565830-1"},
 CellID->1152786378],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1146565830-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->880629379]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(max\:293a 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(max\:293a 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1172583037"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1172583037-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1172583037-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1172583037", "Proof|ID:1172583037-1"},
 CellID->222384166],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1172583037-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->822954280]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(max\:293a 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(max\:293a 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2142810348"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2142810348-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2142810348-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2142810348", "Proof|ID:2142810348-1"},
 CellID->1174101328],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2142810348-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->653142793]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(min\:293a)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(min\:293a)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["min", "D"], "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["x", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["min", "D"], "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["x", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1460290836"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1460290836-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1460290836-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1460290836", "Proof|ID:1460290836-1"},
 CellID->1313465390],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1460290836-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.61467184671761*^9, 3.614671859808209*^9}},
 CellID->1745865902]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(max\:293a)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(max\:293a)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["x", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["x", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1931124773"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1931124773-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1931124773-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1931124773", "Proof|ID:1931124773-1"},
 CellID->70708585],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1931124773-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1543615849]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(min\:293a decreases)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(min\:293a decreases)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:978507648"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p978507648-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p978507648-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:978507648", "Proof|ID:978507648-1"},
 CellID->1954389032],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p978507648-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1594840844]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(max\:293a increases)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(max\:293a increases)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2097939500"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2097939500-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2097939500-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2097939500", "Proof|ID:2097939500-1"},
 CellID->910681388],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2097939500-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1555785782]
}, Closed]],

Cell[TextData[{
 "The only thing that still remains to be done is establishing the connection \
between ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["min", "D"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["max", "D"], TraditionalForm]]],
 " and the monoid operation ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["+", "D"], TraditionalForm]]],
 " in totally-ordered Abelian monoids. Please note that formulas ",
 ButtonBox["(min+)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:446229453"],
 " and ",
 ButtonBox["(max+)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:513276857"],
 " provide the theoretical justification of inference rule \
\[OpenCurlyDoubleQuote]Move addition into tuples\[CloseCurlyDoubleQuote] of \
the ComplexityProver."
}], "Text",
 CellChangeTimes->{{3.603605455841588*^9, 3.603605497108473*^9}, {
  3.603605532788448*^9, 3.603605548988338*^9}, {3.603605618629896*^9, 
  3.603605682299516*^9}, {3.603605717664525*^9, 3.603605717939148*^9}, {
  3.604993189766521*^9, 3.604993228334168*^9}, {3.615027804060902*^9, 
  3.615027810086763*^9}, {3.615027906681981*^9, 3.615027987819013*^9}},
 CellID->755189501],

Cell["", "OpenEnvironment",
 CellID->1830383482],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"min", "+", "x"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"max", "+", "x"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.600578834300765*^9, 3.600578853512872*^9}, {
  3.600688419359877*^9, 3.600688419993036*^9}, {3.613964777347385*^9, 
  3.613964786846366*^9}},
 CellID->524486684],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.603605978840075*^9, 3.603605984337759*^9}},
 CellID->1418774134],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[SquareSubsetEqual]", "D"], 
   RowBox[{"[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{
       SubscriptBox["A", "i"], 
       UnderscriptBox["+", "D"], "x"}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}], "]"}]}], "\[And]", 
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{
       SubscriptBox["A", "i"], 
       UnderscriptBox["+", "D"], "x"}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}], 
    "\[RightBracketingBar]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(21)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:900399941", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.615619934973438*^9, 3.615619976528263*^9}},
 CellTags->{
  "ID:900399941", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "21"},
 CellID->900399941],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{
     RowBox[{
      UnderscriptBox["min", "D"], "[", "A", "]"}], 
     UnderscriptBox["+", "D"], "x"}], "]"}]}], "\[And]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{
     UnderscriptBox["min", "D"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{
        SubscriptBox["A", "i"], 
        UnderscriptBox["+", "D"], "x"}], 
       UnderscriptBox["|", 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
      "\[RightAngleBracket]"}], "]"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min+ 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1833484473", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610946998135238*^9, 3.610947007894602*^9}, {
  3.610949591275794*^9, 3.610949607245697*^9}, {3.61094968727327*^9, 
  3.610949691776265*^9}},
 CellTags->{
  "ID:1833484473", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 1"},
 CellID->1833484473],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["min", "D"], "[", "A", "]"}], 
   UnderscriptBox["+", "D"], "x"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   UnderscriptBox["min", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      SubscriptBox["A", "i"], 
      UnderscriptBox["+", "D"], "x"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min+ 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1976058899", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61095441708428*^9, 3.610954421414012*^9}},
 CellTags->{
  "ID:1976058899", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 2"},
 CellID->1976058899],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["min", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      SubscriptBox["A", "i"], 
      UnderscriptBox["+", "D"], "x"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   RowBox[{
    UnderscriptBox["min", "D"], "[", "A", "]"}], 
   UnderscriptBox["+", "D"], "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min+ 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2067845318", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610954426758623*^9, 3.610954438824223*^9}},
 CellTags->{
  "ID:2067845318", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 3"},
 CellID->2067845318],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["min", "D"], "[", "A", "]"}], 
   UnderscriptBox["+", "D"], "x"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["min", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      SubscriptBox["A", "i"], 
      UnderscriptBox["+", "D"], "x"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min+)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:446229453", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600579715305568*^9, 3.600579731063783*^9}, {
  3.600579778852762*^9, 3.600579806311667*^9}, {3.600580041466303*^9, 
  3.600580079210564*^9}, {3.603605729417163*^9, 3.603605765950106*^9}, {
  3.604993199598869*^9, 3.604993205285691*^9}},
 CellTags->{
  "ID:446229453", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+"},
 CellID->446229453],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{
     RowBox[{
      UnderscriptBox["max", "D"], "[", "A", "]"}], 
     UnderscriptBox["+", "D"], "x"}], "]"}]}], "\[And]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{
     UnderscriptBox["max", "D"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{
        SubscriptBox["A", "i"], 
        UnderscriptBox["+", "D"], "x"}], 
       UnderscriptBox["|", 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
      "\[RightAngleBracket]"}], "]"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max+ 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:437038980", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610949181023388*^9, 3.610949183716164*^9}, {
  3.610955393105612*^9, 3.610955405396908*^9}},
 CellTags->{
  "ID:437038980", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 1"},
 CellID->437038980],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["max", "D"], "[", "A", "]"}], 
   UnderscriptBox["+", "D"], "x"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      SubscriptBox["A", "i"], 
      UnderscriptBox["+", "D"], "x"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max+ 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:367539508", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61095531370502*^9, 3.610955318164619*^9}},
 CellTags->{
  "ID:367539508", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 2"},
 CellID->367539508],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      SubscriptBox["A", "i"], 
      UnderscriptBox["+", "D"], "x"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   RowBox[{
    UnderscriptBox["max", "D"], "[", "A", "]"}], 
   UnderscriptBox["+", "D"], "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max+ 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1689218301", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610955325676592*^9, 3.610955329763939*^9}},
 CellTags->{
  "ID:1689218301", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 3"},
 CellID->1689218301],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["max", "D"], "[", "A", "]"}], 
   UnderscriptBox["+", "D"], "x"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      SubscriptBox["A", "i"], 
      UnderscriptBox["+", "D"], "x"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max+)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:513276857", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600579827108284*^9, 3.600579832792181*^9}, {
  3.600580058681718*^9, 3.600580080389373*^9}, {3.603605734420918*^9, 
  3.603605768178226*^9}, {3.604993207612353*^9, 3.60499321351154*^9}},
 CellTags->{
  "ID:513276857", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+"},
 CellID->513276857],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1359585033]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1284690125],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(21)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(21)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{"1", 
               UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{"1", 
               UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:900399941"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p900399941-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p900399941-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:900399941", "Proof|ID:900399941-1"},
 CellID->1837094901],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p900399941-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1605223748]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(min+ 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(min+ 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["min", "D"], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["+", "D"], 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                   "\[RightAngleBracket]"}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["min", "D"], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["+", "D"], 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                   "\[RightAngleBracket]"}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1833484473"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1833484473-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1833484473-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1833484473", "Proof|ID:1833484473-1"},
 CellID->209997353],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1833484473-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.615621805691964*^9, 3.615621832594963*^9}},
 CellID->1366892285]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(min+ 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(min+ 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1976058899"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1976058899-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1976058899-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1976058899", "Proof|ID:1976058899-1"},
 CellID->1890751220],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1976058899-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {50., 56.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.61562189232046*^9, 3.615621909862635*^9}},
 CellID->2116571744]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(min+ 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(min+ 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox[
               "x", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox[
               "x", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2067845318"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2067845318-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2067845318-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2067845318", "Proof|ID:2067845318-1"},
 CellID->1653738850],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2067845318-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {50., 56.}}]], "ProofInfoBody",
 CellID->2084769334]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(min+)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(min+)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["min", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["min", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:446229453"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p446229453-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p446229453-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:446229453", "Proof|ID:446229453-1"},
 CellID->357185825],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p446229453-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677195792352*^9},
 CellID->22513353]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(max+ 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(max+ 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["max", "D"], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["+", "D"], 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                   "\[RightAngleBracket]"}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["max", "D"], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["+", "D"], 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                   "\[RightAngleBracket]"}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:437038980"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p437038980-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p437038980-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:437038980", "Proof|ID:437038980-1"},
 CellID->132391776],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p437038980-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1014774542]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(max+ 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(max+ 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:367539508"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p367539508-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p367539508-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:367539508", "Proof|ID:367539508-1"},
 CellID->1111387014],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p367539508-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {50., 56.}}]], "ProofInfoBody",
 CellID->785534569]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(max+ 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(max+ 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox[
               "x", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox[
               "x", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1689218301"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1689218301-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1689218301-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1689218301", "Proof|ID:1689218301-1"},
 CellID->434528939],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1689218301-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {50., 56.}}]], "ProofInfoBody",
 CellID->984365216]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(max+)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(max+)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[SquareSubsetEqual]", "D"], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["x", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:513276857"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p513276857-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p513276857-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:513276857", "Proof|ID:513276857-1"},
 CellID->1335154081],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p513276857-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->15476564]
}, Closed]],

Cell[TextData[{
 "Now, all auxiliary lemmata on the basic notions introduced in section ",
 ButtonBox["Basic Notions",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::BasicNotions"],
 " are proved, and we proceed by introducing new notions directly related to \
the complexity analysis."
}], "Text",
 CellChangeTimes->{{3.603606078586127*^9, 3.603606168806884*^9}, {
  3.609832748995092*^9, 3.609832771481567*^9}, {3.615028081505333*^9, 
  3.6150281319033*^9}},
 CellID->2110478468]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Specific Notions for Complexity Analysis", "Section",
 CellChangeTimes->{{3.598252880227565*^9, 3.598252881565854*^9}, {
  3.59946082377149*^9, 3.599460824594164*^9}, {3.605522230710675*^9, 
  3.605522249398903*^9}},
 CellTags->"sect::SpecificNotions",
 CellID->1752495898],

Cell[TextData[{
 "In this section, more precisely in the definition below, we define all the \
notions that are needed for carrying out the complexity analysis. These \
notions include, for instance, ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["W", TraditionalForm]]],
 ", and they are defined in terms of the basic notions introduced in section ",
 ButtonBox["Basic Notions",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::BasicNotions"],
 "."
}], "Text",
 CellChangeTimes->{{3.605525550635837*^9, 3.605525612425576*^9}, {
  3.615087459480236*^9, 3.615087489286313*^9}},
 CellID->96197413],

Cell["", "OpenEnvironment",
 CellID->1718449027],

Cell[CellGroupData[{

Cell["Definition: Notions for Complexity Analysis", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.596364987131295*^9, 3.596364989493083*^9}, {
  3.598250861711047*^9, 3.598250865056572*^9}, {3.605522268058523*^9, 
  3.605522274848635*^9}, {3.615700925673659*^9, 3.615700925898709*^9}},
 CellID->1247544428],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"x", ",", "y", ",", "A", ",", "k", ",", "l"}]]], "GlobalDeclaration",\

 ShowCellTags->False,
 CellChangeTimes->{{3.596365014112432*^9, 3.596365022411622*^9}, {
   3.596365091802784*^9, 3.596365092207751*^9}, {3.596365182898036*^9, 
   3.596365186174687*^9}, 3.596365274446312*^9, {3.596365318084251*^9, 
   3.596365322115815*^9}, {3.596365539141289*^9, 3.596365539919199*^9}, {
   3.596365878006898*^9, 3.59636587924828*^9}, {3.596366308619101*^9, 
   3.596366310824057*^9}, {3.604834355963501*^9, 3.604834356325261*^9}},
 CellTags->{
  "ID:1076035631", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->1076035631],

Cell[BoxData[
 RowBox[{
  RowBox[{"isExponentVector", "[", "x", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[SquareSubsetEqual]", "D"], 
         RowBox[{"[", "x", "]"}]}]},
       {
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "x", "\[RightBracketingBar]"}], 
         "\[Equal]", "n"}]}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExponentVector)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:610987882", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596365028680867*^9, 3.596365080698778*^9}, {
   3.596365275847011*^9, 3.596365282222434*^9}, {3.598251485711934*^9, 
   3.59825151893873*^9}, 3.599282173940314*^9, {3.604834361967245*^9, 
   3.604834365214603*^9}, {3.605523197717904*^9, 3.605523198381931*^9}, {
   3.605523246716865*^9, 3.605523287566361*^9}, {3.605524429966975*^9, 
   3.605524443596571*^9}, {3.605589454681737*^9, 3.60558946676286*^9}, {
   3.605589716297259*^9, 3.605589717439931*^9}},
 CellTags->{
  "ID:610987882", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExponentVector"},
 CellID->610987882],

Cell[BoxData[
 RowBox[{
  RowBox[{"isExpVectorTuple", "[", "A", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"isTuple", "[", "A", "]"}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{"isExponentVector", "[", 
          SubscriptBox["A", "j"], "]"}]}]}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpVectorTuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:468249611", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.605589590712989*^9, 3.605589649131835*^9}, {
  3.611382039915812*^9, 3.611382041650769*^9}},
 CellTags->{
  "ID:468249611", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExpVectorTuple"},
 CellID->468249611],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[VerticalBar]", "y"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"d", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
   RowBox[{
    SubscriptBox["x", "d"], 
    UnderscriptBox["\[LessEqual]", "D"], 
    SubscriptBox["y", "d"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(divisibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:444732983", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596365146344831*^9, 3.596365241296704*^9}, {
   3.598251574314238*^9, 3.598251581313857*^9}, {3.59825163670231*^9, 
   3.598251687768917*^9}, {3.598252305088328*^9, 3.598252315870304*^9}, {
   3.598696507480502*^9, 3.598696518395009*^9}, 3.59928218240833*^9, {
   3.604834367162975*^9, 3.60483437015296*^9}, {3.605523322463986*^9, 
   3.6055233226309*^9}, {3.606791627877149*^9, 3.606791629291973*^9}, {
   3.613884278007499*^9, 3.613884280226245*^9}},
 CellTags->{
  "ID:444732983", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "divisibility"},
 CellID->444732983],

Cell[BoxData[
 RowBox[{
  RowBox[{"deg", "[", "x", "]"}], ":=", 
  RowBox[{
   UnderscriptBox[
    SubscriptBox["\[Sum]", "D"], 
    RowBox[{
     RowBox[{"d", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
   SubscriptBox["x", "d"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(degree)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1111389324", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.5963652593333*^9, 3.596365286679436*^9}, {
   3.598251808979941*^9, 3.598251830612073*^9}, 3.599282189696566*^9, {
   3.604834373685886*^9, 3.60483437541778*^9}, {3.605523330115784*^9, 
   3.605523330373743*^9}, {3.613884269250575*^9, 3.613884271240566*^9}},
 CellTags->{
  "ID:1111389324", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "degree"},
 CellID->1111389324],

Cell[BoxData[
 RowBox[{
  RowBox[{"maxdeg", "[", "A", "]"}], ":=", 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"deg", "[", 
      SubscriptBox["A", "i"], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(maxdeg)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1699354265", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599383459281503*^9, 3.599383495855803*^9}, {
  3.599383548944912*^9, 3.59938357971051*^9}, {3.599804839238223*^9, 
  3.59980484344901*^9}, {3.600407703471596*^9, 3.600407704367169*^9}, {
  3.60499309602203*^9, 3.604993098055469*^9}, {3.605523335037004*^9, 
  3.605523336772792*^9}},
 CellTags->{
  "ID:1699354265", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maxdeg"},
 CellID->1699354265],

Cell[BoxData[
 RowBox[{
  RowBox[{"lcm", "[", 
   RowBox[{"x", ",", "y"}], "]"}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["max", "D"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       SubscriptBox["x", "d"], ",", 
       SubscriptBox["y", "d"]}], "\[RightAngleBracket]"}], "]"}], 
    UnderscriptBox["|", 
     RowBox[{
      RowBox[{"d", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lcm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1614921088", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596365451316036*^9, 3.596365502229487*^9}, {
   3.598251857863278*^9, 3.598251892606659*^9}, {3.598696551944726*^9, 
   3.598696552995991*^9}, 3.599282196066469*^9, {3.60006088568891*^9, 
   3.600060888821185*^9}, {3.60483438540853*^9, 3.604834388543271*^9}, {
   3.604993100567157*^9, 3.604993102490536*^9}, {3.605523340385245*^9, 
   3.605523340791791*^9}, {3.613884291802206*^9, 3.613884294759443*^9}},
 CellTags->{
  "ID:1614921088", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "lcm"},
 CellID->1614921088],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["let", 
    RowBox[{"z", "=", 
     RowBox[{"lcm", "[", 
      RowBox[{"x", ",", "y"}], "]"}]}]], 
   RowBox[{"(", 
    RowBox[{"\[Not]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"\[And]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{
              SubscriptBox["A", "j"], "\[VerticalBar]", "z"}]},
            {
             RowBox[{
              RowBox[{"deg", "[", 
               RowBox[{"lcm", "[", 
                RowBox[{"x", ",", 
                 SubscriptBox["A", "j"]}], "]"}], "]"}], 
              UnderscriptBox["<", "D"], 
              RowBox[{"deg", "[", "z", "]"}]}]},
            {
             RowBox[{
              RowBox[{"deg", "[", 
               RowBox[{"lcm", "[", 
                RowBox[{
                 SubscriptBox["A", "j"], ",", "y"}], "]"}], "]"}], 
              UnderscriptBox["<", "D"], 
              RowBox[{"deg", "[", "z", "]"}]}]}
           }]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chain criterion)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1784038925", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599372674949866*^9, 3.599372709305928*^9}, {
  3.599372750361296*^9, 3.599372771709582*^9}, {3.599372939576456*^9, 
  3.599372947262178*^9}, {3.599388477318429*^9, 3.599388477530061*^9}, {
  3.599388602008721*^9, 3.599388607700214*^9}, {3.604825950258353*^9, 
  3.604825967421949*^9}, {3.604826464044654*^9, 3.60482647108681*^9}, {
  3.604834390460561*^9, 3.604834398127914*^9}, {3.604834606034249*^9, 
  3.604834628629839*^9}, {3.605523345365729*^9, 3.605523358434541*^9}},
 CellTags->{
  "ID:1784038925", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chain criterion"},
 CellID->1784038925],

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "[", 
   RowBox[{"A", ",", "k"}], "]"}], ":=", 
  RowBox[{
   UnderscriptBox["min", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["A", "i"], ")"}], "k"], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(width w.r.t. k-th coordinate)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:312341806", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596368569907185*^9, 3.596368586565551*^9}, {
   3.598251252045475*^9, 3.598251256611137*^9}, 3.598251626527007*^9, {
   3.598251971177821*^9, 3.598251979759195*^9}, {3.59869661862429*^9, 
   3.598696619735518*^9}, 3.599282298849689*^9, {3.599804018099186*^9, 
   3.599804030468335*^9}, {3.599804079223328*^9, 3.599804098334104*^9}, {
   3.599807079505942*^9, 3.599807103732306*^9}, {3.599807280462097*^9, 
   3.599807299365169*^9}, {3.600576879696671*^9, 3.600576906212763*^9}, {
   3.604826166768641*^9, 3.604826192914639*^9}, {3.604993106962017*^9, 
   3.604993108812067*^9}, {3.605523363686792*^9, 3.605523363872787*^9}, {
   3.611382182775124*^9, 3.611382189017481*^9}},
 CellTags->{
  "ID:312341806", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "width w.r.t. k-th coordinate"},
 CellID->312341806],

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "[", "A", "]"}], ":=", 
  RowBox[{
   UnderscriptBox[
    SubscriptBox["\[Sum]", "D"], 
    RowBox[{
     RowBox[{"d", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "d"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(width)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:282925385", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596366921058262*^9, 3.59636693191184*^9}, {
   3.596367063376717*^9, 3.596367080354524*^9}, {3.596367209655026*^9, 
   3.596367223089203*^9}, 3.596368200849081*^9, {3.596368585166435*^9, 
   3.596368621212362*^9}, {3.598252019264389*^9, 3.598252047304834*^9}, 
   3.599282302752813*^9, {3.604826309295523*^9, 3.604826311585924*^9}, {
   3.605523365162232*^9, 3.605523366621757*^9}, {3.612505323436774*^9, 
   3.612505326643771*^9}, {3.613884305488929*^9, 3.613884307220721*^9}},
 CellTags->{
  "ID:282925385", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "width"},
 CellID->282925385],

Cell[BoxData[
 RowBox[{
  RowBox[{"M", "[", 
   RowBox[{"x", ",", "A"}], "]"}], ":=", 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"deg", "[", 
      RowBox[{"lcm", "[", 
       RowBox[{"x", ",", 
        SubscriptBox["A", "i"]}], "]"}], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{"chainCrit", "[", 
      RowBox[{"x", ",", 
       SubscriptBox["A", "i"], ",", "A"}], "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(maximum degree of admissible pairs of one object)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1664335746", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599804184483785*^9, 3.59980427412232*^9}, {
  3.599804342275648*^9, 3.599804357594859*^9}, {3.599804388211136*^9, 
  3.599804395087615*^9}, {3.599804798992827*^9, 3.599804805154614*^9}, {
  3.60040773997708*^9, 3.60040774085841*^9}, {3.604826530981415*^9, 
  3.604826549704301*^9}, {3.604834428137694*^9, 3.604834432722756*^9}, {
  3.604834669543021*^9, 3.604834671131857*^9}, {3.604993114125255*^9, 
  3.60499311715455*^9}, {3.605523368023699*^9, 3.60552337431791*^9}, {
  3.611380441470282*^9, 3.611380628612227*^9}},
 CellTags->{
  "ID:1664335746", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maximum degree of admissible pairs of one object"},
 CellID->1664335746],

Cell[BoxData[
 RowBox[{
  RowBox[{"M", "[", "A", "]"}], ":=", 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"M", "[", 
      RowBox[{
       SubscriptBox["A", "i"], ",", "A"}], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(maximum degree of admissible pairs)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:640444143", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596367317863857*^9, 3.596367428870215*^9}, {
   3.596367669883805*^9, 3.596367691330864*^9}, {3.598251302623422*^9, 
   3.598251305773571*^9}, 3.598251621922846*^9, {3.598697018443303*^9, 
   3.598697019434379*^9}, 3.599282308791382*^9, {3.59937338658466*^9, 
   3.599373388463525*^9}, {3.599804430101492*^9, 3.599804469013744*^9}, {
   3.599804807347909*^9, 3.599804812106826*^9}, {3.600407744693688*^9, 
   3.60040774557884*^9}, {3.604993123691396*^9, 3.604993125467123*^9}, {
   3.605523378813808*^9, 3.605523380917964*^9}},
 CellTags->{
  "ID:640444143", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maximum degree of admissible pairs"},
 CellID->640444143],

Cell[BoxData[
 RowBox[{
  RowBox[{"L", "[", "A", "]"}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["max", "D"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       SubscriptBox[
        RowBox[{"(", 
         SubscriptBox["A", "j"], ")"}], "d"], 
       UnderscriptBox["|", 
        RowBox[{
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
      "\[RightAngleBracket]"}], "]"}], 
    UnderscriptBox["|", 
     RowBox[{
      RowBox[{"d", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(corner of rectangular region)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1923113076", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599896982323629*^9, 3.599897012462471*^9}, {
  3.599897057974966*^9, 3.599897080971896*^9}, {3.599897191366482*^9, 
  3.599897232478101*^9}, {3.604993127352049*^9, 3.604993129124376*^9}, {
  3.60552338283598*^9, 3.605523383058231*^9}, {3.611382166260786*^9, 
  3.611382173068829*^9}, {3.613884323296107*^9, 3.613884340404903*^9}},
 CellTags->{
  "ID:1923113076", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "corner of rectangular region"},
 CellID->1923113076],

Cell[TextData[{
 "The following predicates are the so-called ",
 StyleBox["cover predicates",
  FontSlant->"Italic"],
 ". The sets of points in the space of two-dimensional exponent vectors for \
which the predicates hold cover the whole space; For more details on this see \
formula ",
 ButtonBox["(cover)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:315092621"],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.603688378799354*^9, 3.603688486938563*^9}, {
  3.604037289484044*^9, 3.604037293295114*^9}, {3.604037520708736*^9, 
  3.604037533671932*^9}, {3.604037645329672*^9, 3.604037650024348*^9}, {
  3.604038926467767*^9, 3.604038927131242*^9}, {3.605523392302986*^9, 
  3.60552340846736*^9}, {3.615087801530394*^9, 3.615087838018748*^9}},
 CellID->2042263002],

Cell[BoxData[
 RowBox[{
  RowBox[{"isAbove", "[", 
   RowBox[{"x", ",", "A"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    SubscriptBox["A", "j"], "\[VerticalBar]", 
    "x"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(above)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1774661003", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596368405187606*^9, 3.596368463748702*^9}, {
  3.604834436741319*^9, 3.604834460553326*^9}, {3.604834744474092*^9, 
  3.604834745447281*^9}, {3.605523415841287*^9, 3.605523417816872*^9}, {
  3.606791639030389*^9, 3.606791640023573*^9}, {3.612252936224057*^9, 
  3.612252937771195*^9}},
 CellTags->{
  "ID:1774661003", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "above"},
 CellID->1774661003],

Cell[BoxData[
 RowBox[{
  RowBox[{"inFarExterior", "[", 
   RowBox[{"x", ",", "A", ",", "k"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["x", "k"], 
     UnderscriptBox["<", "D"], 
     RowBox[{"W", "[", 
      RowBox[{"A", ",", "k"}], "]"}]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
      RowBox[{
       SubscriptBox[
        RowBox[{"(", 
         SubscriptBox["A", "j"], ")"}], 
        SuperscriptBox["k", "-"]], 
       UnderscriptBox["<", "D"], 
       SubscriptBox["x", 
        SuperscriptBox["k", "-"]]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(far exterior w.r.t. k-th coordinate)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:285552196", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59981809034045*^9, 3.599818168490175*^9}, {
  3.599889827318324*^9, 3.599889827689205*^9}, {3.604834440430336*^9, 
  3.604834465644566*^9}, {3.604834749419303*^9, 3.604834763708775*^9}, {
  3.605523419183475*^9, 3.605523421619513*^9}, {3.611382202228463*^9, 
  3.611382211474788*^9}, {3.612239311229935*^9, 3.612239323023825*^9}, {
  3.612239736759293*^9, 3.612239739055941*^9}, {3.612239769274506*^9, 
  3.612239770456682*^9}, {3.612252940572783*^9, 3.612252941604598*^9}},
 CellTags->{
  "ID:285552196", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "far exterior w.r.t. k-th coordinate"},
 CellID->285552196],

Cell[BoxData[
 RowBox[{
  RowBox[{"inRectangle", "[", 
   RowBox[{"x", ",", "A"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"d", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     SubscriptBox["x", "d"], 
     UnderscriptBox["\[LessEqual]", "D"], 
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["A", "j"], ")"}], "d"]}]}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rectangular region)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:494773971", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599889760801803*^9, 3.599889776695528*^9}, {
  3.599889807065485*^9, 3.599889913536784*^9}, {3.599889973322619*^9, 
  3.599889974867637*^9}, {3.604834449502304*^9, 3.604834473837085*^9}, {
  3.604834757656903*^9, 3.604834766767886*^9}, {3.605523427613265*^9, 
  3.605523427909272*^9}, {3.606791642928588*^9, 3.606791644032154*^9}, {
  3.611382224640135*^9, 3.611382228187187*^9}, {3.612252950293386*^9, 
  3.612252957953208*^9}, {3.612253007143733*^9, 3.612253009975176*^9}},
 CellTags->{
  "ID:494773971", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "rectangular region"},
 CellID->494773971],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->544969917]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1228004538],

Cell[TextData[{
 "As can be easily observed, all the functions and predicates defined above \
depend on the totally-ordered Abelian monoid ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " and the natural number ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.60352136431935*^9, 3.603521408347714*^9}, {
  3.603521457470612*^9, 3.603521787013376*^9}, {3.604826443638505*^9, 
  3.604826448267863*^9}, {3.605523478510078*^9, 3.605523625414487*^9}, {
  3.605524406041989*^9, 3.605524416795674*^9}, {3.605524458462987*^9, 
  3.605524463186513*^9}, {3.605601262748522*^9, 3.605601282869522*^9}},
 CellID->1229204291],

Cell[TextData[{
 "All the functions and predicates mentioned in section \
\[OpenCurlyDoubleQuote]Problem Setting\[CloseCurlyDoubleQuote], in particular \
",
 Cell[BoxData[
  FormBox[
   SubscriptBox["M", "F"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["W", "F"], TraditionalForm]]],
 ", ",
 StyleBox["chainCrit",
  FontFamily->"Courier 10 Pitch"],
 ", etc., are defined in the definition above. It should be quite obvious \
that their formal ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " definitions agree with those in section \[OpenCurlyDoubleQuote]Problem \
Setting\[CloseCurlyDoubleQuote], up to a couple of slight modifications:"
}], "Text",
 CellChangeTimes->{{3.603521880934966*^9, 3.603521983252846*^9}, {
  3.603522018412145*^9, 3.603522090079833*^9}, {3.603526559651927*^9, 
  3.603526563938103*^9}, {3.605523634233438*^9, 3.605523642958437*^9}},
 CellID->955050321],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   SubscriptBox["M", "F"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["W", "F"], TraditionalForm]]],
 " are written as ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "F", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", "F", "]"}], TraditionalForm]]],
 ", respectively."
}], "Item",
 CellChangeTimes->{{3.603522097676148*^9, 3.603522156633486*^9}, {
  3.603522204457919*^9, 3.603522240100025*^9}, {3.60352227174345*^9, 
  3.603522393550974*^9}},
 CellID->1782480268],

Cell[TextData[{
 "Rather than considering ",
 StyleBox["sets",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 ", everything is based upon ",
 StyleBox["tuples",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ". This is because we have to take into account the multiplicity an element \
occurs in a collection with (\[Rule] Several polynomials might have same \
leading monomial), and this can faithfully only be accomplished with tuples. \
Also, the fact that every tuple is finite by definition does not cause any \
problems, since all the collections of monomials (resp. exponent vectors) \
that are dealt with in the complexity analysis are finite anyway."
}], "Item",
 CellChangeTimes->{{3.603522097676148*^9, 3.603522156633486*^9}, {
  3.603522204457919*^9, 3.603522240100025*^9}, {3.60352227174345*^9, 
  3.60352256931291*^9}, {3.603522615297966*^9, 3.60352268540662*^9}, {
  3.603522725685912*^9, 3.603522725702861*^9}, {3.603527253720015*^9, 
  3.603527262369459*^9}},
 CellID->1033322772]
}, Open  ]],

Cell[TextData[StyleBox["First Remark:",
 FontFamily->"Bitstream Charter",
 FontSize->14,
 FontWeight->"Bold"]], "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellChangeTimes->{{3.603522907354292*^9, 3.603522909426996*^9}, {
  3.603522955835897*^9, 3.603522963082172*^9}, {3.603793011527348*^9, 
  3.603793024231905*^9}, {3.605601401187481*^9, 3.60560140544953*^9}},
 CellID->121149513],

Cell[TextData[{
 "If one carefully investigates the definition of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " one realizes that also the cases ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "[", 
    RowBox[{
     SubscriptBox["A", "i"], ",", 
     SubscriptBox["A", "i"]}], "]"}], TraditionalForm]]],
 " are taken into account. This might appear strange at first sight, since in \
Buchberger\[CloseCurlyQuote]s algorithm ",
 ButtonBox["(Algorithm 1)",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 " only pairs of ",
 StyleBox["different",
  FontSlant->"Italic"],
 " polynomials are considered. There are two reasons for defining ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " in that way:\n(i) This is also how it is done in ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 ", and\n(ii) It makes things (especially notation) much easier, since it \
allows us to define the auxiliary binary ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 "-predicate in just the way we did, without having to care about indices.\n\
In any case, due to formula ",
 ButtonBox["(M=M*)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:872843574"],
 " it does not make a difference at all whether the two indices are required \
to be unequal or not: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " will always be the same, and this is the only thing that matters."
}], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellChangeTimes->{{3.603522955835983*^9, 3.603523016264877*^9}, {
   3.603523067993002*^9, 3.603523068774317*^9}, {3.603526511396674*^9, 
   3.603526537459005*^9}, {3.603526632531704*^9, 3.603526636839904*^9}, {
   3.603526720792316*^9, 3.60352672617742*^9}, {3.60352675821136*^9, 
   3.603526794949112*^9}, {3.603526838114751*^9, 3.60352684115434*^9}, {
   3.603526877383081*^9, 3.603527143643606*^9}, {3.603527173960989*^9, 
   3.603527187399389*^9}, {3.603793026985894*^9, 3.603793319090968*^9}, {
   3.603793361177027*^9, 3.60379351484412*^9}, 3.604035104537365*^9, {
   3.604035136662751*^9, 3.604035137669769*^9}, {3.604834858331772*^9, 
   3.604834958512149*^9}, {3.605601401187581*^9, 3.605601405449621*^9}, {
   3.615087678930005*^9, 3.615087689896127*^9}},
 CellID->787517132],

Cell[TextData[StyleBox["Second Remark:",
 FontFamily->"Bitstream Charter",
 FontSize->14,
 FontWeight->"Bold"]], "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15002.},
 CellChangeTimes->{{3.603522907354292*^9, 3.603522909426996*^9}, {
  3.603522955835897*^9, 3.603522963082172*^9}, {3.603793011527348*^9, 
  3.603793015707187*^9}, {3.605601401187649*^9, 3.605601413059204*^9}},
 CellID->2122574584],

Cell[TextData[{
 "All the quantified variables in definition \[OpenCurlyDoubleQuote]Notions \
of Complexity Analysis\[CloseCurlyDoubleQuote] are ",
 StyleBox["untyped",
  FontSlant->"Italic"],
 ": ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ", for instance, is nowhere required to be a tuple. This has something to do \
with the general principle of function definitions in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ", according to ",
 ButtonBox["[Buchberger2014]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger2014"],
 ": At the time of defining a function, as few constraints should be put on \
the variables as possible (ideally no constraints at all), even if the \
definition makes sense only if the variables involved have certain properties \
or types. Later, when computing/proving with the function, the user has to \
make sure that the variables are instantiated by suitable terms; However, \
even if this is not the case, due to the design of ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " there will hardly be any errors, since ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " simply leaves undefined expressions unevaluated (in most cases)."
}], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15002.},
 CellChangeTimes->{{3.603522955835983*^9, 3.603523016264877*^9}, {
  3.603523067993002*^9, 3.603523068774317*^9}, {3.603526511396674*^9, 
  3.603526537459005*^9}, {3.603526632531704*^9, 3.603526636839904*^9}, {
  3.603526720792316*^9, 3.60352672617742*^9}, {3.60352675821136*^9, 
  3.603526794949112*^9}, {3.603526838114751*^9, 3.60352684115434*^9}, {
  3.603526877383081*^9, 3.603527143643606*^9}, {3.603527173960989*^9, 
  3.603527187399389*^9}, {3.605523666671106*^9, 3.605523683579066*^9}, {
  3.605601401197036*^9, 3.605601413059299*^9}, {3.615087914458756*^9, 
  3.615087920195296*^9}},
 CellID->365888472],

Cell[TextData[{
 "There is one more definition we need in order to be able to carry out the \
complexity analysis: We have to introduce the auxiliary notion ",
 StyleBox["chainCritDiagonal",
  FontSlant->"Italic"],
 ", which covers one special case where the chain criterion for exponents ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " holds w.r.t. ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " in the bivariate case, i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 " (following the second remark, however, we do not impose any type/value \
restrictions in the definition). See also the first drawing in the next \
subsubsection."
}], "Text",
 CellChangeTimes->{{3.612521502991415*^9, 3.612521679551224*^9}, {
  3.612521714715012*^9, 3.612521730040865*^9}},
 CellID->408039679],

Cell["", "OpenEnvironment",
 CellID->1453057307],

Cell[CellGroupData[{

Cell["Definition: chainCritDiagonal", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.606468616245521*^9, 3.606468616416904*^9}, {
   3.60646881533475*^9, 3.606468840310203*^9}, {3.610189307833959*^9, 
   3.610189309809167*^9}, {3.612516679341083*^9, 3.612516683514747*^9}, 
   3.612517739769921*^9},
 CellID->2050713516],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"x", ",", "y", ",", "A", ",", "k"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612521468368179*^9, 3.612521472568313*^9}},
 CellID->852694876],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A", ",", "k"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         SubscriptBox["x", "k"], 
         UnderscriptBox["<", "D"], 
         SubscriptBox["y", "k"]}]},
       {
        RowBox[{
         SubscriptBox["y", 
          SuperscriptBox["k", "-"]], 
         UnderscriptBox["<", "D"], 
         SubscriptBox["x", 
          SuperscriptBox["k", "-"]]}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[Or]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                SubscriptBox["y", "k"], 
                UnderscriptBox["\[LessEqual]", "D"], 
                SubscriptBox[
                 RowBox[{"(", 
                  SubscriptBox["A", "j"], ")"}], "k"]}]},
              {
               RowBox[{
                SubscriptBox["x", 
                 SuperscriptBox["k", "-"]], 
                UnderscriptBox["\[LessEqual]", "D"], 
                SubscriptBox[
                 RowBox[{"(", 
                  SubscriptBox["A", "j"], ")"}], 
                 SuperscriptBox["k", "-"]]}]}
             }]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCritDiagonal w.r.t. k-th coordinate)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1278822860", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612517002900844*^9, 3.612517021691178*^9}, 
   3.612517064641383*^9, {3.612521476516618*^9, 3.612521477807876*^9}},
 CellTags->{
  "ID:1278822860", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal w.r.t. k-th coordinate"},
 CellID->1278822860],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", "2"}]], 
    RowBox[{"chainCritDiagonal", "[", 
     RowBox[{"x", ",", "y", ",", "A", ",", "k"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCritDiagonal)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1458824241", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612516483141697*^9, 3.612516578105695*^9}, {
  3.612517028143002*^9, 3.612517035007572*^9}},
 CellTags->{
  "ID:1458824241", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal"},
 CellID->1458824241],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1156889619]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->568715675],

Cell[CellGroupData[{

Cell["Two Drawings", "Subsection",
 CellChangeTimes->{{3.605601607701691*^9, 3.605601609397275*^9}, {
  3.612521694466364*^9, 3.612521698937286*^9}},
 CellID->1231732349],

Cell[TextData[{
 "In the first drawing, ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " lies strictly left and strictly above ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " and no element of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " lies in the region below the least common multiple of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ": This is exactly the case when ",
 Cell[BoxData[
  FormBox[
   RowBox[{"chainCritDiagonal", "[", 
    RowBox[{"x", ",", "y", ",", "A", ",", "1"}], "]"}], TraditionalForm]]],
 " holds. Please note that the relative positions of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " are important!\nThe drawing suggests that not only ",
 Cell[BoxData[
  FormBox["chainCritDiagonal", TraditionalForm]]],
 ", but also ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 " holds: This is indeed true in general and will be proved in section ",
 ButtonBox["Lemmata on Specific Notions / Special Case n=2 / chainCrit",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::chainCritSpecial"],
 ". If ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " were swapped, then ",
 Cell[BoxData[
  FormBox[
   RowBox[{"chainCritDiagonal", "[", 
    RowBox[{"x", ",", "y", ",", "A", ",", "2"}], "]"}], TraditionalForm]]],
 " would hold."
}], "Text",
 CellChangeTimes->{{3.604040897468361*^9, 3.604041043587647*^9}, {
   3.604990457470149*^9, 3.604990473835212*^9}, {3.605597947558214*^9, 
   3.605597961294178*^9}, {3.606469136478948*^9, 3.606469367906013*^9}, {
   3.606474037429156*^9, 3.606474038326828*^9}, {3.606474096858222*^9, 
   3.60647410497723*^9}, {3.606474470418671*^9, 3.606474470567337*^9}, {
   3.606474956845282*^9, 3.606474958676582*^9}, {3.612521091453391*^9, 
   3.61252109282526*^9}, 3.612521601426703*^9, {3.612521743561641*^9, 
   3.612521818987031*^9}, {3.612521850401604*^9, 3.612521959499496*^9}, {
   3.615087991873151*^9, 3.615088020292095*^9}, {3.615088050582038*^9, 
   3.615088116540501*^9}},
 CellID->1857632321],

Cell[BoxData[
 GraphicsBox[{
   {JoinForm[{"Miter", 2.0500000000000003`}], EdgeForm[{GrayLevel[0.], 
    Opacity[1.], AbsoluteThickness[1], 
    JoinForm[{"Miter", 2.0500000000000003`}]}], 
    ArrowBox[{{0.09722222222222224, 0.07499999999999996}, {
     0.09722222222222224, 0.9722222222222222}}]}, 
   {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
    ArrowBox[{{0.1, 0.07499999999999996}, {1.2972222222222225`, 
     0.07499999999999973}}]}, GraphicsGroupBox[{
     {EdgeForm[{RGBColor[0, 0.666667, 1], Opacity[1.], Thickness[
      0.003019951720402013]}], FaceForm[{RGBColor[0, 0.666667, 1], Opacity[
      0.1]}], DiskBox[{0.3388888888888891, 0.6138888888888888}, \
{0.015713484026367797`, 0.01571348402636774}]}, 
     StyleBox[InsetBox[
       StyleBox[
        TagBox["x",
         "RawBoxes"],
        TextAlignment->Left,
        Background->RGBColor[
         0.9570000762951095, 0.9600061036087587, 
          0.9700007629510948]], {0.28611111111111115`, 0.6472222222222223}, {
       Left, Baseline},
       Alignment->{Left, Top}],
      FontSize->24,
      FontColor->RGBColor[0, 0.666667, 1]]}], GeometricTransformationBox[{
     {EdgeForm[{RGBColor[0, 0.666667, 1], Opacity[1.], Thickness[
      0.003019951720402013]}], FaceForm[{RGBColor[0, 0.666667, 1], Opacity[
      0.1]}], DiskBox[{0.3388888888888891, 0.6138888888888888}, \
{0.015713484026367797`, 0.01571348402636774}]}, 
     StyleBox[InsetBox[
       StyleBox[
        TagBox["y",
         "RawBoxes"],
        TextAlignment->Left,
        Background->RGBColor[
         0.9570000762951095, 0.9600061036087587, 
          0.9700007629510948]], {0.36944444444444474`, 0.5749999999999998}, {
       Left, Baseline},
       Alignment->{Left, Top}],
      FontSize->24,
      FontColor->RGBColor[0, 0.666667, 1]]}, {{
    0.4560321468757051, -0.3532543690979274}}], 
   {Thickness[0.003019951720402013], Dashing[{Medium, Medium}], StrokeForm[
    Opacity[1.]], EdgeForm[Opacity[1.]], EdgeForm[None], 
    LineBox[{{0.338888888888889, 0.6138888888888889}, {0.7916666666666667, 
     0.6138888888888889}}]}, 
   {Thickness[0.003019951720402013], Dashing[{Medium, Medium}], StrokeForm[
    Opacity[1.]], EdgeForm[Opacity[1.]], EdgeForm[None], 
    LineBox[{{0.7944444444444445, 0.26111111111111107`}, {0.7944444444444445, 
     0.6111111111111109}}]}, 
   {EdgeForm[{Opacity[1.], Thickness[0.005370317963702522]}], FaceForm[None], 
    RectangleBox[{0.7750000000000001, 0.6361111111111112}, \
{0.8123195245626234, 0.5987915865484879}]}, 
   {EdgeForm[{Opacity[1.], Thickness[0.0004570881896148747]}], EdgeForm[None],
     FaceForm[Opacity[0.1]], 
    RectangleBox[{0.7694444444444446, 0.5916666666666669}, \
{0.09722222222222217, 0.07499999999999994}]}, InsetBox[
    StyleBox[
     TagBox[
      RowBox[{"lcm", "[", 
       RowBox[{"x", ",", "y"}], "]"}],
      "RawBoxes"],
     TextAlignment->Left,
     FontSize->24,
     FontSlant->"Italic",
     Background->RGBColor[
      0.9570000762951095, 0.9600061036087587, 
       0.9700007629510948]], {0.8243589743589745, 0.653846153846154}, {
    Left, Baseline},
    Alignment->{Left, Top}]},
  ContentSelectable->True,
  ImageMargins->{{3., 0.}, {0., 0.}},
  ImagePadding->{{0., 0.}, {0., 0.}},
  ImageSize->{348., 260.},
  PlotRange->{{0., 1.3333333333333335`}, {0., 1.}},
  PlotRangePadding->Automatic]], "Text",
 Editable->False,
 Deployed->True,
 CellChangeTimes->{{3.60560130456881*^9, 3.605601309150591*^9}, 
   3.60560148437655*^9, {3.605606346223402*^9, 3.605606348528124*^9}, {
   3.606474111921043*^9, 3.606474353264689*^9}, {3.606474516534484*^9, 
   3.606474533232435*^9}},
 TextAlignment->Center,
 CellID->1503678106],

Cell[TextData[{
 "The second drawing illustrates some of the notions defined above in the \
case of two dimensions (i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 "). For this, let ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " be a non-empty tuple of exponent vectors:"
}], "Text",
 CellChangeTimes->{{3.605601334558802*^9, 3.605601377348643*^9}, {
  3.605601818828955*^9, 3.605601836753585*^9}, {3.612521824115552*^9, 
  3.612521824867094*^9}},
 CellID->342302046],

Cell[BoxData[
 GraphicsBox[{
   {JoinForm[{"Miter", 2.0500000000000003`}], EdgeForm[{GrayLevel[0.], 
    Opacity[1.], AbsoluteThickness[1], 
    JoinForm[{"Miter", 2.0500000000000003`}]}], 
    ArrowBox[{{0.09722222222222224, 0.07499999999999996}, {
     0.09722222222222224, 0.9722222222222222}}]}, 
   {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
    ArrowBox[{{0.1, 0.07499999999999996}, {1.2972222222222225`, 
     0.07499999999999973}}]}, 
   {Thickness[0.003019951720402013], Dashing[{Medium, Medium}], StrokeForm[
    Opacity[1.]], EdgeForm[Opacity[1.]], EdgeForm[None], 
    LineBox[{{1.155758313057584, 0.08050036791758572}, {0.10000000000000021`, 
     0.902722590139808}}]}, GeometricTransformationBox[GraphicsGroupBox[{
      {EdgeForm[{RGBColor[0., 0.6666666666666666, 1.], Opacity[1.], 
       AbsoluteThickness[1]}], FaceForm[{RGBColor[0, 0.666667, 1], Opacity[
       0.1]}], PolygonBox[{{0.20277777777777783`, 0.625}, {
        0.20277777777777783`, 0.40555555555555556`}, {0.3333333333333334, 
        0.40555555555555556`}, {0.3333333333333334, 0.3472222222222222}, {
        0.3777777777777778, 0.3472222222222222}, {0.3777777777777778, 0.25}, {
        0.5111111111111113, 0.25}, {0.5111111111111113, 0.1694444444444444}, {
        0.9777777777777781, 0.1694444444444444}, {0.9777777777777781, 
        0.625}}]}, {
       {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
        FaceForm[RGBColor[0, 0.666667, 1]], 
        DiskBox[{0.20555555555555563`, 0.4055555555555552}, \
{0.00982092751648006, 0.009820927516480005}]}, 
       {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
        FaceForm[RGBColor[0, 0.666667, 1]], 
        DiskBox[{0.3333333333333334, 0.34722222222222193`}, \
{0.00982092751647995, 0.009820927516479894}]}, 
       {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
        FaceForm[RGBColor[0, 0.666667, 1]], 
        DiskBox[{0.38333333333333336`, 0.2499999999999997}, \
{0.00982092751647995, 0.009820927516479755}]}, 
       {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
        FaceForm[RGBColor[0, 0.666667, 1]], 
        DiskBox[{0.5111111111111113, 0.16944444444444412`}, \
{0.009820927516479894, 0.009820927516479977}]}, 
       {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
        FaceForm[RGBColor[0, 0.666667, 1]], 
        DiskBox[{0.9750000000000002, 0.19444444444444417`}, \
{0.009820927516479894, 0.009820927516479977}]}, 
       {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
        FaceForm[RGBColor[0, 0.666667, 1]], 
        DiskBox[{0.4027777777777779, 0.6222222222222218}, \
{0.00982092751647995, 0.009820927516480005}]}, 
       {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
        FaceForm[None], 
        RectangleBox[{0.31666666666666676`, 0.4166666666666664}, \
{0.3422010782095144, 0.3911322551238188}]}, 
       {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
        FaceForm[None], 
        RectangleBox[{0.363621683117465, 0.35998942799364564`}, \
{0.38915609466031276`, 0.3344550164507982}]}, 
       {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
        FaceForm[None], 
        RectangleBox[{0.4941772386730208, 0.25998942799364555`}, \
{0.5197116502158684, 0.23445501645079814`}]}}, {
       {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
        FaceForm[None], 
        RectangleBox[{0.38862168311746503`, 0.6349894279936457}, \
{0.4141560946603128, 0.6094550164507982}]}, 
       {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
        FaceForm[None], 
        RectangleBox[{0.9608439053396876, 0.2099894279936459}, \
{0.9863783168825351, 0.18445501645079831`}]}}}], {{-3.469446951953614*^-18, \
-0.0027777777777778546`}}], 
   StyleBox[InsetBox[
     TagBox[
      RowBox[{"M", "[", "A", "]"}],
      "RawBoxes"], {0.3305555555555555, 0.747222222222222}, {Left, Baseline},
     Alignment->{Left, Top}],
    FontSize->14,
    Background->RGBColor[1, 1, 1, 0.]], 
   {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
    None], GeometricTransformationBox[
     RectangleBox[{0.9636216831174651, 0.6349894279936459}, \
{0.9891560946603128, 0.6094550164507984}], {{{{
       0.699886702893344, -0.7142538786125588}, {0.7142538786125588, 
       0.699886702893344}}, {0.7374519242838966, -0.5106523882789543}}}]}, 
   {Thickness[0.003019951720402013], Dashing[{Medium, Medium}], StrokeForm[
    Opacity[1.]], EdgeForm[Opacity[1.]], EdgeForm[None], 
    LineBox[{{0.7472222222222225, 0.1694444444444444}, {0.7472222222222225, 
     0.07499999999999962}}]}, 
   {Thickness[0.003019951720402013], Dashing[{Medium, Medium}], StrokeForm[
    Opacity[1.]], EdgeForm[Opacity[1.]], EdgeForm[None], 
    LineBox[{{0.2, 0.5111111111111112}, {0.09722222222222227, 
     0.511111111111111}}]}, 
   StyleBox[InsetBox[
     StyleBox[
      TagBox[
       RowBox[{"W", "[", 
        RowBox[{"A", ",", "2"}], "]"}],
       "RawBoxes"],
      FontSize->14], {0.7555555555555556, 0.10833333333333309`}, {
     Left, Baseline},
     Alignment->{Left, Top}],
    Background->RGBColor[1, 1, 1, 0.]], 
   {EdgeForm[{RGBColor[0, 0.784314, 0], Opacity[1.], Thickness[
    0.0032734069487883794`]}], FaceForm[None], 
    RectangleBox[{0.9888888888888892, 0.6333333333333327}, \
{0.09722222222222231, 0.07499999999999962}]}, 
   StyleBox[InsetBox[
     StyleBox[
      TagBox[
       RowBox[{"L", "[", "A", "]"}],
       "RawBoxes"],
      FontSize->14], {0.9861111111111113, 0.6444444444444443}, {
     Left, Baseline},
     Alignment->{Left, Top}],
    Background->RGBColor[1, 1, 1, 0.]], 
   {EdgeForm[{RGBColor[1, 0, 0], Opacity[1.], Thickness[
    0.0032734069487883794`]}], FaceForm[None], 
    RectangleBox[{0.09722222222222227, 0.636111111111111}, \
{0.20277777777777783`, 0.9416666666666667}]}, 
   {EdgeForm[{RGBColor[1, 0, 0], Opacity[1.], Thickness[
    0.0032734069487883794`]}], FaceForm[None], 
    RectangleBox[{0.9916666666666669, 0.16944444444444423`}, 
     {1.2722222222222226`, 0.07499999999999979}]}, 
   StyleBox[InsetBox[
     TagBox["A",
      "RawBoxes"], {0.5555555555555558, 0.34722222222222204`}, {
     Left, Baseline},
     Alignment->{Left, Top}],
    FontSize->30,
    FontWeight->Bold,
    FontColor->RGBColor[0, 0.666667, 1],
    Background->RGBColor[1, 1, 1, 0.]], 
   StyleBox[InsetBox[
     StyleBox[
      TagBox[
       RowBox[{"W", "[", 
        RowBox[{"A", ",", "1"}], "]"}],
       "RawBoxes"],
      FontSize->14], {0.03611111111111098, 0.5402777777777776}, {
     Left, Baseline},
     Alignment->{Left, Top}],
    Background->RGBColor[1, 1, 1, 0]]},
  ContentSelectable->True,
  ImageMargins->{{3., 0.}, {0., 0.}},
  ImagePadding->{{0., 0.}, {0., 0.}},
  ImageSize->{480, 360},
  PlotRange->{{0., 1.3333333333333335`}, {0., 1.}},
  PlotRangePadding->Automatic]], "Text",
 Editable->False,
 Deployed->True,
 CellChangeTimes->{{3.60560130456881*^9, 3.605601309150591*^9}, 
   3.60560148437655*^9, {3.605606346223402*^9, 3.605606348528124*^9}},
 TextAlignment->Center,
 CellID->652519512],

Cell[CellGroupData[{

Cell[TextData[{
 "The light-blue region is the region all the points of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " belong to. This does not mean that every point of the region belongs to ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ", but only that there are no points of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " outside of it."
}], "Item",
 CellChangeTimes->{{3.605601622626967*^9, 3.605601691107909*^9}, {
  3.605602804895709*^9, 3.605602805108042*^9}},
 CellID->457187766],

Cell[TextData[{
 "The blue circles depict elements of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ". It is important to observe that for each edge of the light-blue region \
there is at least one point lying on it, meaning that the region is \
\[OpenCurlyDoubleQuote]smallest\[CloseCurlyDoubleQuote] in some sense."
}], "Item",
 CellChangeTimes->{{3.605601622626967*^9, 3.605601807106226*^9}, {
  3.605601848398149*^9, 3.605601848428904*^9}},
 CellID->574066102],

Cell[TextData[{
 "The empty squares depict LCMs of elements of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " which satisfy the chain criterion."
}], "Item",
 CellChangeTimes->{{3.605601622626967*^9, 3.605601807106226*^9}, {
  3.605601848398149*^9, 3.605601908321684*^9}},
 CellID->438806075],

Cell[TextData[{
 "The long diagonal dashed line illustrates ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 ", meaning that all points below have degree less than ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.605601622626967*^9, 3.605601807106226*^9}, {
  3.605601848398149*^9, 3.605601961649728*^9}},
 CellID->665266488],

Cell[TextData[{
 "This, by definition of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 ", means that all LCMs satisfying the chain criterion (empty squares) have \
to lie below the dashed line."
}], "Item",
 CellChangeTimes->{{3.605601622626967*^9, 3.605601807106226*^9}, {
  3.605601848398149*^9, 3.60560200755814*^9}},
 CellID->914798491],

Cell[TextData[{
 "However, formula ",
 ButtonBox["(maxdeg\[LessEqual]M)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:592779258"],
 " even ensures that all points of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " must lie below it, which means that the part of the light-blue region \
above ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " must in fact not contain any points of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.605601622626967*^9, 3.605601807106226*^9}, {
  3.605601848398149*^9, 3.60560201462731*^9}, {3.605602091088166*^9, 
  3.60560216675906*^9}, {3.615088226219383*^9, 3.615088233254544*^9}, {
  3.615088288725314*^9, 3.615088302772516*^9}},
 CellID->872492695],

Cell[TextData[{
 "The diamond in the upper-right corner of the light-blue region depicts ",
 Cell[BoxData[
  FormBox[
   RowBox[{"L", "[", "A", "]"}], TraditionalForm]]],
 ". Formula ",
 ButtonBox["(M\[LessEqual]L)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:341822661"],
 " guarantees that this point always lies above ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.605601622626967*^9, 3.605601807106226*^9}, {
  3.605601848398149*^9, 3.60560201462731*^9}, {3.605602091088166*^9, 
  3.605602124059186*^9}, {3.605602170009476*^9, 3.605602266090158*^9}, {
  3.615088327309345*^9, 3.615088349400572*^9}},
 CellID->41635902],

Cell[TextData[{
 "The short dashed lines depict ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "1"}], "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "2"}], "]"}], TraditionalForm]]],
 ", respectively."
}], "Item",
 CellChangeTimes->{{3.605601622626967*^9, 3.605601807106226*^9}, {
  3.605601848398149*^9, 3.60560201462731*^9}, {3.605602091088166*^9, 
  3.605602124059186*^9}, {3.605602170009476*^9, 3.605602316569396*^9}},
 CellID->484929528],

Cell[TextData[{
 "The big rectangle with green frame is the rectangular region of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ", i.e. the set of all points ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " for which ",
 Cell[BoxData[
  FormBox[
   RowBox[{"inRectangle", "[", 
    RowBox[{"x", ",", "A"}], "]"}], TraditionalForm]]],
 " holds. As can be seen, its upper right corner is just ",
 Cell[BoxData[
  FormBox[
   RowBox[{"L", "[", "A", "]"}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.605601622626967*^9, 3.605601807106226*^9}, {
  3.605601848398149*^9, 3.60560201462731*^9}, {3.605602091088166*^9, 
  3.605602124059186*^9}, {3.605602170009476*^9, 3.605602410530151*^9}},
 CellID->1357948457],

Cell[TextData[{
 "The two rectangles with red frame are the far exteriors of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " w.r.t. either coordinate (left one w.r.t. first coordinate, right one \
w.r.t. second coordinate). Note that they are in fact unbounded."
}], "Item",
 CellChangeTimes->{{3.605601622626967*^9, 3.605601807106226*^9}, {
  3.605601848398149*^9, 3.60560201462731*^9}, {3.605602091088166*^9, 
  3.605602124059186*^9}, {3.605602170009476*^9, 3.605602522982015*^9}},
 CellID->545799062],

Cell[TextData[{
 "The set of points lying above ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " is not depicted in the drawing. It is, however, simply the set of points \
lying, as the name suggests, \[OpenCurlyDoubleQuote]above\
\[CloseCurlyDoubleQuote] ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ": The points in the space except the ones lying in the far exterior or in \
the part of the rectangular region outside of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.605601622626967*^9, 3.605601807106226*^9}, {
  3.605601848398149*^9, 3.60560201462731*^9}, {3.605602091088166*^9, 
  3.605602124059186*^9}, {3.605602170009476*^9, 3.605602663413276*^9}},
 CellID->1522152467]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Computations", "Section",
 CellChangeTimes->{{3.599387990400902*^9, 3.599387994365326*^9}, {
   3.599388031433927*^9, 3.599388038137345*^9}, {3.603513767163762*^9, 
   3.603513771782143*^9}, 3.603529821324995*^9, {3.613896811844099*^9, 
   3.613896813864383*^9}},
 CellTags->"sect::Computations",
 CellID->1485320096],

Cell[TextData[{
 "This section is dedicated to some examples involving computations with the \
notions defined in the previous section. The underlying domain is always the ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " built-in domain ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 ", since this is also the domain of interest in the complexity analysis. \
However, in addition to the bivariate case we also consider the trivariate \
case here."
}], "Text",
 CellChangeTimes->{{3.603529710237187*^9, 3.603529843909071*^9}, {
  3.60352995288437*^9, 3.603529980426966*^9}, {3.605523765415267*^9, 
  3.605523791944326*^9}, {3.615089167308309*^9, 3.615089185904633*^9}},
 CellID->472378341],

Cell["", "OpenEnvironment",
 CellID->871319643],

Cell[CellGroupData[{

Cell[TextData[{
 "Definition: Exponent vectors over ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.60552429254527*^9, 3.605524308800393*^9}},
 CellID->1397499305],

Cell[BoxData[
 RowBox[{"D", ":=", 
  SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(D=N0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1559569657", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.605524316763954*^9, 3.605524325243367*^9}, 
   3.605524977078948*^9},
 CellTags->{
  "ID:1559569657", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D=N0"},
 CellID->1559569657],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->156157206]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1451763118],

Cell[CellGroupData[{

Cell["Auxiliary Definitions", "Subsection",
 CellChangeTimes->{{3.605432390654989*^9, 3.605432391548584*^9}, {
  3.605523986595478*^9, 3.605523995289431*^9}},
 CellID->589691952],

Cell[TextData[{
 "Since ",
 Cell[BoxData[
  FormBox["min", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["max", TraditionalForm]]],
 " are no built-in domain functions of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 ", we have to define them ourselves by referring to the standard ",
 Cell[BoxData[
  FormBox["min", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["max", TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.605432988093222*^9, 3.605433016447329*^9}, {
  3.605435197229728*^9, 3.605435239800556*^9}, {3.615700921500213*^9, 
  3.615700921649403*^9}},
 CellID->1351740566],

Cell["", "OpenEnvironment",
 CellID->256974579],

Cell[CellGroupData[{

Cell[TextData[{
 "Definition: min, max in ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.60543239872616*^9, 3.605432424365699*^9}},
 CellID->149993638],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "X"]], "GlobalDeclaration",
 ShowCellTags->False,
 CellChangeTimes->{{3.6054324152422*^9, 3.605432416262111*^9}},
 CellTags->{
  "ID:977095909", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->977095909],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["min", 
    SubscriptBox["\[DoubleStruckCapitalN]", "0"]], "[", "X", "]"}], ":=", 
  RowBox[{"min", "[", "X", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(min in N0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:615519889", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.605432432270682*^9, 3.605432445977601*^9}, {
  3.605432604111703*^9, 3.605432612922772*^9}, {3.605433029375688*^9, 
  3.605433031779452*^9}},
 CellTags->{
  "ID:615519889", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min in N0"},
 CellID->615519889],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["max", 
    SubscriptBox["\[DoubleStruckCapitalN]", "0"]], "[", "X", "]"}], ":=", 
  RowBox[{"max", "[", "X", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(max in N0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1245684717", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.605432454193218*^9, 3.605432458739274*^9}, {
  3.605432614686204*^9, 3.605432620916364*^9}, {3.605433034409856*^9, 
  3.60543303725545*^9}},
 CellTags->{
  "ID:1245684717", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max in N0"},
 CellID->1245684717],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1875602967]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1548667982]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{"n", "=", "2"}], TraditionalForm]]]], "Subsection",
 CellChangeTimes->{{3.600512764131949*^9, 3.600512770534628*^9}},
 CellID->586087969],

Cell[TextData[{
 "First we consider the bivariate case, i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.603530033673118*^9, 3.603530168595472*^9}, {
  3.603530272887217*^9, 3.603530295539124*^9}, {3.605526203590393*^9, 
  3.605526267740659*^9}, {3.605597563496636*^9, 3.605597588199734*^9}},
 CellID->1272854071],

Cell["", "OpenEnvironment",
 CellID->1322766950],

Cell[CellGroupData[{

Cell[TextData[{
 "Assumption: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.605597614847661*^9, 3.6055976199153*^9}},
 CellID->540238681],

Cell[BoxData[
 RowBox[{"n", ":=", "2"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(n=2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1526008058", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.605526057347753*^9, 3.605526058975346*^9}},
 CellTags->{
  "ID:1526008058", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n=2"},
 CellID->1526008058],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1365355815]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2067681119],

Cell["\<\
Now we are ready to perform computations. Make sure the necessary knowledge \
and built-ins are activated when carrying out the computations.\
\>", "Text",
 CellChangeTimes->{{3.603530178377585*^9, 3.60353021129481*^9}},
 CellID->2004358328],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"isExponentVector", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"0", ",", "0"}], "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.601115743228764*^9, 3.601115756583202*^9}, {
  3.605523926745312*^9, 3.605523930294938*^9}},
 CellID->363465800],

Cell[BoxData[
 FormBox["True", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.601115757342396*^9, 3.601115767738403*^9}, 
   3.601878031947992*^9, {3.604835119143655*^9, 3.604835133104729*^9}, {
   3.605432523554683*^9, 3.605432529928145*^9}, 3.605432891657551*^9, 
   3.605435101105245*^9, {3.605525075340141*^9, 3.605525082067889*^9}, 
   3.605526312431426*^9, {3.613897055352139*^9, 3.613897083039763*^9}, 
   3.615700963380702*^9, 3.615701000805568*^9, 3.617949236079895*^9},
 CellID->1355610580]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c363465800"]]]]}, {None, None}},
 CellChangeTimes->{3.617949236614443*^9},
 CellID->1483357400],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c363465800-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{{3.617949236746065*^9, 3.617949245691939*^9}},
 CellID->483700261]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"isExponentVector", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"0", ",", "0", ",", "0"}], "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.601115772522319*^9, 3.601115783048776*^9}, {
  3.605523936495394*^9, 3.605523939896007*^9}},
 CellID->301274866],

Cell[BoxData[
 FormBox["False", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.601115783528553*^9, 3.601878033994014*^9, 
  3.604835134949461*^9, 3.605432531764411*^9, 3.605432893657424*^9, 
  3.605435102852411*^9, 3.605525084299175*^9, 3.605526314295902*^9, 
  3.61389708508906*^9, 3.615701019596144*^9, 3.617949299971009*^9},
 CellID->43157872]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c301274866"]]]]}, {None, None}},
 CellChangeTimes->{3.617949300491625*^9},
 CellID->1539369667],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c301274866-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949300630791*^9},
 CellID->2109196125]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"isExponentVector", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"7", ",", 
    RowBox[{"-", "2"}]}], "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.601878038816563*^9, 3.601878041131601*^9}, {
  3.605523941862515*^9, 3.605523945135093*^9}},
 CellID->1843740659],

Cell[BoxData[
 FormBox["False", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.601878041847998*^9, 3.604835136941165*^9, 
  3.605432533611424*^9, 3.605432895722021*^9, 3.605435104476702*^9, 
  3.605525086457797*^9, 3.605526316167681*^9, 3.613897089445301*^9, 
  3.615701021849867*^9, 3.617949301956211*^9},
 CellID->494166568]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1843740659"]]]]}, {None, None}},
 CellChangeTimes->{3.617949302518542*^9},
 CellID->77268719],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1843740659-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949302654048*^9},
 CellID->699968735]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"lcm", "[", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"0", ",", "9"}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"7", ",", "8"}], "\[RightAngleBracket]"}]}], "]"}]], "Computation",\

 CellChangeTimes->{{3.601878051017001*^9, 3.601878079082249*^9}, {
  3.60552394801125*^9, 3.605523948196736*^9}},
 CellID->494647160],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"7", ",", "9"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.601878080057098*^9, 3.604835138934265*^9, 
  3.605432536069519*^9, 3.605432645705376*^9, 3.605432897748348*^9, 
  3.605435106649945*^9, 3.605525088744477*^9, 3.605526317840775*^9, 
  3.613897091658375*^9, 3.615701024066706*^9, 3.617949304021166*^9},
 CellID->1016883962]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c494647160"]]]]}, {None, None}},
 CellChangeTimes->{3.617949304550227*^9},
 CellID->1918883531],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c494647160-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949304684304*^9},
 CellID->1375661780]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"deg", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"3", ",", "6"}], "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.601878096195298*^9, 3.601878112344391*^9}, {
  3.605523949450544*^9, 3.605523949636455*^9}},
 CellID->1489184765],

Cell[BoxData[
 FormBox["9", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.601878113255244*^9, 3.604835141037605*^9, 
  3.60543293083956*^9, 3.605435110742955*^9, 3.605525090663598*^9, 
  3.613897095005581*^9, 3.615701027057651*^9, 3.617949306353029*^9},
 CellID->965057731]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1489184765"]]]]}, {None, None}},
 CellChangeTimes->{3.617949306866961*^9},
 CellID->1070988792],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1489184765-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949307005463*^9},
 CellID->1857348292]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"chainCrit", "[", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"11", ",", "11"}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"0", ",", "12"}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"10", ",", "0"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"11", ",", "11"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"0", ",", "12"}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.601878125235025*^9, 3.601878146827232*^9}, {
  3.604835143283946*^9, 3.604835144752724*^9}, {3.605523951039996*^9, 
  3.605523951225471*^9}},
 CellID->379223532],

Cell[BoxData[
 FormBox["False", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.601878147548241*^9, 3.60483514595773*^9, 
  3.605432933535255*^9, 3.605435113438399*^9, 3.605525092848789*^9, 
  3.613897272265045*^9, 3.615701029142676*^9, 3.617949308425264*^9},
 CellID->174076915]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c379223532"]]]]}, {None, None}},
 CellChangeTimes->{3.617949308935693*^9},
 CellID->612957880],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c379223532-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949309077387*^9},
 CellID->1453820005]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"chainCrit", "[", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"10", ",", "0"}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"0", ",", "12"}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"10", ",", "0"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"11", ",", "11"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"0", ",", "12"}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.601878164199744*^9, 3.601878181806055*^9}, {
   3.604835148074144*^9, 3.604835149426034*^9}, {3.605523952553766*^9, 
   3.605523952686832*^9}, 3.613897746820461*^9},
 CellID->1396655990],

Cell[BoxData[
 FormBox["True", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.601878183496613*^9, 3.604835149954922*^9, 3.605435116249574*^9, 
   3.605525095478031*^9, 3.613897274453038*^9, {3.613897748165644*^9, 
   3.613897758353124*^9}, 3.615701031571963*^9, 3.617949310887916*^9},
 CellID->1382976801]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1396655990"]]]]}, {None, None}},
 CellChangeTimes->{3.617949311425341*^9},
 CellID->488228445],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1396655990-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949311563928*^9},
 CellID->1091772466]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"M", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "7"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"5", ",", "6"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "1"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "6"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.601878284335909*^9, 3.601878292192804*^9}, {
  3.605523954527852*^9, 3.605523954676871*^9}},
 CellID->164423730],

Cell[BoxData[
 FormBox["11", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.601878294060684*^9, 3.604835153002011*^9, 
  3.605435118582031*^9, 3.605525097700857*^9, 3.613897276995798*^9, 
  3.615701034262824*^9, 3.617949313198843*^9},
 CellID->782644456]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c164423730"]]]]}, {None, None}},
 CellChangeTimes->{3.617949313700498*^9},
 CellID->252386576],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c164423730-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949313838144*^9},
 CellID->38275912]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"L", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "7"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"5", ",", "6"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "1"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "6"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{
  3.601878312257741*^9, {3.605523955783712*^9, 3.605523955933157*^9}},
 CellID->752433928],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"5", ",", "7"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.601878312639934*^9, 3.604835155551393*^9, 
  3.605435121025119*^9, 3.605525099638312*^9, 3.613897280426607*^9, 
  3.615701041458123*^9, 3.617949315904971*^9},
 CellID->1528747130]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c752433928"]]]]}, {None, None}},
 CellChangeTimes->{3.617949316402065*^9},
 CellID->417165553],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c752433928-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949316532954*^9},
 CellID->1567625520]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"W", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "7"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"5", ",", "6"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "1"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "6"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{
  3.601878316024851*^9, {3.605523957335238*^9, 3.605523957667729*^9}},
 CellID->20329479],

Cell[BoxData[
 FormBox["2", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.601878316789005*^9, 3.60483515757703*^9, 
  3.605435122937419*^9, 3.605525101403643*^9, 3.61389728719967*^9, 
  3.615701043493383*^9, 3.617949318260881*^9},
 CellID->1193416390]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c20329479"]]]]}, {None, None}},
 CellChangeTimes->{3.617949318742378*^9},
 CellID->837187957],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c20329479-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949318879964*^9},
 CellID->1896946096]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"maxdeg", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "7"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"5", ",", "6"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "1"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "6"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.60187832023622*^9, 3.601878321201934*^9}, {
  3.605523958811617*^9, 3.605523958997307*^9}},
 CellID->285678123],

Cell[BoxData[
 FormBox["11", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.601878321880057*^9, 3.60483515984087*^9, 
  3.60543512481568*^9, 3.605525103103094*^9, 3.613897289840957*^9, 
  3.615701045970061*^9, 3.617949320401263*^9},
 CellID->1225552093]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c285678123"]]]]}, {None, None}},
 CellChangeTimes->{3.617949320908198*^9},
 CellID->45314890],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c285678123-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949321043995*^9},
 CellID->1232819849]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{"n", "=", "3"}], TraditionalForm]]]], "Subsection",
 CellChangeTimes->{{3.599388002975222*^9, 3.599388008111171*^9}, {
  3.600512521474701*^9, 3.600512521474836*^9}},
 CellID->1098302827],

Cell[TextData[{
 "Now comes the trivariate case, i.e. again ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 " is the domain of exponents, but now ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "3"}], TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.603530033673118*^9, 3.603530168595472*^9}, {
  3.603530272887217*^9, 3.603530295539124*^9}, {3.605526203590393*^9, 
  3.605526267740659*^9}, {3.605597598521556*^9, 3.60559760017576*^9}},
 CellID->1778373308],

Cell["", "OpenEnvironment",
 CellID->2113767064],

Cell[CellGroupData[{

Cell[TextData[{
 "Assumption: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "3"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.60051253527466*^9, 3.600512538159241*^9}, {
  3.605525113113917*^9, 3.605525114901972*^9}, {3.605597605923809*^9, 
  3.605597607707896*^9}},
 CellID->1847813459],

Cell[BoxData[
 RowBox[{"n", ":=", "3"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(n=3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1544493310", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600512548947566*^9, 3.600512572334023*^9}, {
  3.605525119317718*^9, 3.605525121493079*^9}},
 CellTags->{
  "ID:1544493310", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n=3"},
 CellID->1544493310],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1619210422]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->766446153],

Cell["\<\
Now we are ready to perform computations. Make sure the necessary knowledge \
and built-ins are activated when carrying out the computations.\
\>", "Text",
 CellChangeTimes->{{3.603530178377585*^9, 3.60353021129481*^9}},
 CellID->1985946244],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"isExponentVector", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"0", ",", "0", ",", "0"}], "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.600512607877179*^9, 3.600512622328478*^9}, {
  3.605525146611391*^9, 3.605525149697491*^9}},
 CellID->189844734],

Cell[BoxData[
 FormBox["True", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.600512637647556*^9, 3.600512949588688*^9, 3.601025564994822*^9, {
   3.601115643825148*^9, 3.601115652274757*^9}, 3.604835174700382*^9, 
   3.605435152252834*^9, {3.605525142500856*^9, 3.605525150229751*^9}, 
   3.605526275570239*^9, 3.613897310102931*^9, 3.615701070186024*^9, 
   3.617949345880146*^9},
 CellID->1709252354]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c189844734"]]]]}, {None, None}},
 CellChangeTimes->{3.617949346327294*^9},
 CellID->88472762],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c189844734-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949346461581*^9},
 CellID->624705868]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"isExponentVector", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"0", ",", "0"}], "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.605525153498423*^9, 3.605525156778784*^9}},
 CellID->491362362],

Cell[BoxData[
 FormBox["False", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.600512646596345*^9, 3.60102556937385*^9, 
  3.601115657052293*^9, 3.60483517619999*^9, 3.605435154043991*^9, 
  3.605525157378192*^9, 3.613897312067526*^9, 3.615701072650364*^9, 
  3.617949347787594*^9},
 CellID->1670964831]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c491362362"]]]]}, {None, None}},
 CellChangeTimes->{3.617949348320895*^9},
 CellID->582639331],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c491362362-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949348459315*^9},
 CellID->701974587]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"isExponentVector", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"1", ",", "2", ",", 
    RowBox[{"-", "4"}]}], "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.605525160265766*^9, 3.605525164096499*^9}},
 CellID->1181188516],

Cell[BoxData[
 FormBox["False", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.600512657713841*^9, 3.601025571691861*^9, 
  3.601115659329306*^9, 3.604835178051014*^9, 3.605435155968534*^9, 
  3.605525164556492*^9, 3.613897314388962*^9, 3.615701074607543*^9, 
  3.617949350456913*^9},
 CellID->223449918]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1181188516"]]]]}, {None, None}},
 CellChangeTimes->{3.617949350968715*^9},
 CellID->1599523293],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1181188516-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949351105119*^9},
 CellID->473374470]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"lcm", "[", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"11", ",", "11", ",", "11"}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"0", ",", "11", ",", "0"}], "\[RightAngleBracket]"}]}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.605525167271285*^9, 3.605525167456579*^9}},
 CellID->307478637],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"11", ",", "11", ",", "11"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.60051266661517*^9, 3.601115663737609*^9, 
  3.604835179901068*^9, 3.605435158255158*^9, 3.605525167938258*^9, 
  3.613897316305152*^9, 3.615701077043992*^9, 3.617949352815575*^9},
 CellID->1599111698]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c307478637"]]]]}, {None, None}},
 CellChangeTimes->{3.61794935332693*^9},
 CellID->417097376],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c307478637-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949353466433*^9},
 CellID->1544363263]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"deg", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"11", ",", "11", ",", "11"}], "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.604835194884008*^9, 3.604835203592457*^9}, {
  3.60552516974588*^9, 3.605525169895309*^9}},
 CellID->1510116252],

Cell[BoxData[
 FormBox["33", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.600512674499244*^9, 3.601115667015734*^9, 
  3.60483520474956*^9, 3.605435159908387*^9, 3.605525170295448*^9, 
  3.613897318628084*^9, 3.615701079006834*^9, 3.617949355834442*^9},
 CellID->1308976343]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1510116252"]]]]}, {None, None}},
 CellChangeTimes->{3.617949356366496*^9},
 CellID->591174291],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1510116252-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949356503772*^9},
 CellID->9837495]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"chainCrit", "[", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"11", ",", "11", ",", "11"}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"0", ",", "12", ",", "0"}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"10", ",", "10", ",", "0"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"11", ",", "11", ",", "11"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"0", ",", "12", ",", "0"}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.604835207099503*^9, 3.604835208515657*^9}, {
  3.605525172370504*^9, 3.60552517255574*^9}},
 CellID->1391592445],

Cell[BoxData[
 FormBox["False", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.600512683654033*^9, 3.601115669646667*^9, 
  3.604835209097546*^9, 3.605435162440251*^9, 3.605525173040429*^9, 
  3.613897321098012*^9, 3.615701081233693*^9, 3.617949358679717*^9},
 CellID->615312700]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1391592445"]]]]}, {None, None}},
 CellChangeTimes->{3.617949359192309*^9},
 CellID->1974151093],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1391592445-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949359329646*^9},
 CellID->498065940]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"chainCrit", "[", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"10", ",", "10", ",", "0"}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"0", ",", "12", ",", "0"}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"10", ",", "10", ",", "0"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"11", ",", "10", ",", "10"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"0", ",", "12", ",", "0"}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.60483521132084*^9, 3.604835212594885*^9}, {
  3.605525175269191*^9, 3.6055251754546*^9}},
 CellID->1172710914],

Cell[BoxData[
 FormBox["True", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.600512692852726*^9, 3.601115672120932*^9, 
  3.604835213042053*^9, 3.605435164544155*^9, 3.605525175926458*^9, 
  3.61389732318408*^9, 3.615701083721064*^9, 3.617949361490819*^9},
 CellID->1707085166]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1172710914"]]]]}, {None, None}},
 CellChangeTimes->{3.617949361991752*^9},
 CellID->609311901],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1172710914-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949362125847*^9},
 CellID->750452517]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"deg", "[", 
  RowBox[{"lcm", "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"10", ",", "10", ",", "0"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"0", ",", "12", ",", "0"}], "\[RightAngleBracket]"}]}], "]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.604835218384527*^9, 3.604835225983872*^9}, {
  3.605525178112261*^9, 3.605525180547489*^9}},
 CellID->823265960],

Cell[BoxData[
 FormBox["22", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.600512698986936*^9, 3.60111567441766*^9, 
  3.604835226444816*^9, 3.605435166535926*^9, 3.605525181836971*^9, 
  3.613897325366897*^9, 3.615701086104881*^9, 3.617949364037864*^9},
 CellID->1745361079]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c823265960"]]]]}, {None, None}},
 CellChangeTimes->{3.61794936453266*^9},
 CellID->2138059031],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c823265960-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949364675832*^9},
 CellID->1167618070]
}, Closed]],

Cell[TextData[{
 "The following two examples provide a counterexample for formula ",
 ButtonBox["(88)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:477242065"],
 " if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", ">", "2"}], TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.603530449639257*^9, 3.603530537012363*^9}, {
  3.615089371827481*^9, 3.615089382016161*^9}, {3.615787614211652*^9, 
  3.615787615047699*^9}},
 CellID->1360124193],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"M", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "7", ",", "0"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"5", ",", "3", ",", "6"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "1", ",", "1"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.605525184550674*^9, 3.605525184809483*^9}},
 CellID->1205577273],

Cell[BoxData[
 FormBox["14", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.600512713980499*^9, 3.601115676658191*^9, 
  3.604835230567797*^9, 3.605435168840617*^9, 3.605525185954836*^9, 
  3.613897332122318*^9, 3.615701088825176*^9, 3.617949367006567*^9},
 CellID->1165502951]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1205577273"]]]]}, {None, None}},
 CellChangeTimes->{3.617949367481005*^9},
 CellID->1183991773],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1205577273-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949367624825*^9},
 CellID->2004806027]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"M", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "7", ",", "0"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"5", ",", "3", ",", "6"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "1", ",", "1"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "3", ",", "6"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.605525187865652*^9, 3.605525188052024*^9}},
 CellID->1005372790],

Cell[BoxData[
 FormBox["16", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.600512725356736*^9, 3.600512954500685*^9, 
  3.601115679562591*^9, 3.604835233773611*^9, 3.605435171103611*^9, 
  3.60552518899306*^9, 3.613897382356773*^9, 3.615701092057393*^9, 
  3.617949370390164*^9},
 CellID->914965399]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1005372790"]]]]}, {None, None}},
 CellChangeTimes->{3.617949370906698*^9},
 CellID->262060218],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1005372790-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949371046961*^9},
 CellID->2050120917]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"W", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "7", ",", "0"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"5", ",", "3", ",", "6"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "1", ",", "1"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "3", ",", "6"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.605525190851335*^9, 3.605525191036464*^9}},
 CellID->317313208],

Cell[BoxData[
 FormBox["2", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.600512733235714*^9, 3.601115681890609*^9, 
  3.604835235772602*^9, 3.605435173133614*^9, 3.605525191507057*^9, 
  3.61389739989741*^9, 3.615701094322555*^9, 3.61794937267128*^9},
 CellID->301600]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c317313208"]]]]}, {None, None}},
 CellChangeTimes->{3.617949373172506*^9},
 CellID->1971312572],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c317313208-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949373315352*^9},
 CellID->2138593680]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"L", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "7", ",", "0"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"5", ",", "3", ",", "6"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "1", ",", "1"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "3", ",", "6"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.605525194026539*^9, 3.60552519421142*^9}},
 CellID->176496894],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"5", ",", "7", ",", "6"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.600512741017115*^9, 3.601115684949202*^9, 
  3.604835237949885*^9, 3.60543517508717*^9, 3.605525194653648*^9, 
  3.613897402287431*^9, 3.615701096521772*^9, 3.617949374990123*^9},
 CellID->625007468]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c176496894"]]]]}, {None, None}},
 CellChangeTimes->{3.617949375481509*^9},
 CellID->1494793755],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c176496894-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949375626513*^9},
 CellID->573749363]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"maxdeg", "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "7", ",", "0"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"5", ",", "3", ",", "6"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "1", ",", "1"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "3", ",", "6"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.605525197551146*^9, 3.605525197773375*^9}},
 CellID->1743340434],

Cell[BoxData[
 FormBox["14", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.600512748573631*^9, 3.60111568726345*^9, 
  3.604835240016594*^9, 3.60543517700334*^9, 3.605525198270354*^9, 
  3.613897404739142*^9, 3.615701099031617*^9, 3.617949377126554*^9},
 CellID->1326982056]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1743340434"]]]]}, {None, None}},
 CellChangeTimes->{3.61794937765477*^9},
 CellID->1718243673],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1743340434-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {58., 65.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.617949377798136*^9},
 CellID->25789008]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmata on Specific Notions", "Section",
 CellChangeTimes->{{3.596372033631333*^9, 3.596372037489846*^9}, 
   3.605435273552693*^9, 3.612766296808673*^9, 3.613885820089048*^9, {
   3.613971548063376*^9, 3.613971551645987*^9}},
 CellTags->"sect::LemmataSpecific",
 CellID->1812043342],

Cell[TextData[{
 "This section is about proving auxiliary lemmata on the various functions \
and predicates introduced in section ",
 ButtonBox["Specific Notions for Complexity Analysis",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::SpecificNotions"],
 ". In particular, the most important functions/predicates that are dealt \
with are ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["W", TraditionalForm]]],
 " and the cover predicates ",
 Cell[BoxData[
  FormBox["isAbove", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["inFarExterior", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["inRectangle", TraditionalForm]]],
 ".\nNote that, although the main complexity analysis is only carried out for \
the bivariate case (i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 "), some of the lemmata can be proved in the general setting as well."
}], "Text",
 CellChangeTimes->{{3.603688134889966*^9, 3.603688165302273*^9}, {
  3.603688210456125*^9, 3.603688336934789*^9}, {3.603690323881535*^9, 
  3.603690350718791*^9}, {3.603690383151159*^9, 3.603690395117629*^9}, {
  3.603690446347239*^9, 3.60369052282772*^9}, {3.604835359924966*^9, 
  3.604835407751827*^9}, {3.605437417276657*^9, 3.605437422509679*^9}, {
  3.611568589885774*^9, 3.611568606019848*^9}, {3.612089743860581*^9, 
  3.612089745429292*^9}, {3.615089420091734*^9, 3.615089432373523*^9}},
 CellID->207929894],

Cell[TextData[{
 "Also note that unlike the lemmata above (lemmata on notions), the lemmata \
in this section will always be used ",
 StyleBox["directly",
  FontSlant->"Italic"],
 " later on. This means that there are no special inference rules in the \
ComplexityProver that somehow \[OpenCurlyDoubleQuote]mimic\
\[CloseCurlyDoubleQuote] their contents. The reason is that the lemmata in \
section ",
 ButtonBox["Lemmata on Basic Notions",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataBasic"],
 " are very general, at a very low level: They hold ",
 StyleBox["for all",
  FontSlant->"Italic"],
 " tuples, ",
 StyleBox["for all",
  FontSlant->"Italic"],
 " order relations (with certain properties), etc. The lemmata in this \
section, however, are far more specific: They involve notions like ",
 Cell[BoxData[
  FormBox["deg", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["lcm", TraditionalForm]]],
 ", etc. which can easily be dealt with by simple rewriting later on, using \
the formulas below."
}], "Text",
 CellChangeTimes->{{3.605437426411547*^9, 3.605437499473424*^9}, {
  3.605437555378677*^9, 3.605437557652117*^9}, {3.605437592433831*^9, 
  3.605437722928037*^9}, {3.605437755185778*^9, 3.605437764855305*^9}, {
  3.60543783529942*^9, 3.605437982167955*^9}, {3.60543803302663*^9, 
  3.605438046217642*^9}, {3.605438077167332*^9, 3.605438118554006*^9}, {
  3.612497276243194*^9, 3.61249728251045*^9}, {3.615089450149391*^9, 
  3.615089525286666*^9}},
 CellID->58811915],

Cell[CellGroupData[{

Cell["General Case", "Subsection",
 CellChangeTimes->{{3.598270390679023*^9, 3.598270393136361*^9}, {
   3.612082339273864*^9, 3.612082339933762*^9}, 3.612766293128424*^9},
 CellTags->"sect::LemmataSpecificGeneral",
 CellID->1771021382],

Cell[TextData[{
 "All the lemmata in this subsection hold for all dimensions (i.e. number of \
indeterminates) ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ", and they are also stated and proved in this general setting (apart from \
some explicitly mentioned exceptions).\n\nTherefore, let in the sequel always \
",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " be an ",
 StyleBox["arbitrary",
  FontSlant->"Italic"],
 " natural number (not necessarily 2) and let ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " be a tuple of exponent vectors:"
}], "Text",
 CellChangeTimes->{{3.603692485394801*^9, 3.603692542506256*^9}, {
  3.603693027140089*^9, 3.6036931311555*^9}, {3.603694302657077*^9, 
  3.603694318941622*^9}, {3.603694948414311*^9, 3.603694949724773*^9}, {
  3.605588627985058*^9, 3.605588650324534*^9}, {3.605589746488864*^9, 
  3.605589777396371*^9}, {3.611035997179003*^9, 3.611035999721217*^9}, {
  3.611568452090481*^9, 3.611568452307457*^9}},
 CellID->506293920],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExpVectorTuple", "[", "A", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.598270600205748*^9, 3.598270615950374*^9}, 
   3.598329751179035*^9, {3.60558976899632*^9, 3.605589772787629*^9}},
 CellID->1507641767],

Cell[CellGroupData[{

Cell["isExponentVector, isExpVectorTuple", "Subsubsection",
 CellChangeTimes->{{3.611397122371382*^9, 3.611397125459381*^9}, {
   3.611397319730462*^9, 3.611397321852264*^9}, {3.611986885852834*^9, 
   3.611986886369014*^9}, 3.612086918107184*^9, {3.613963913119446*^9, 
   3.613963917715923*^9}},
 CellID->2076981791],

Cell[TextData[{
 "We start by proving some very simple and very general properties of \
exponent vectors and tuples thereof. Formula ",
 ButtonBox["(isExpVectorTuple\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:313023917"],
 " is probably the most important result in this subsection."
}], "Text",
 CellChangeTimes->{{3.615089595413315*^9, 3.615089624677504*^9}, {
  3.615089674039861*^9, 3.615089718367127*^9}},
 CellID->1783959097],

Cell["", "OpenEnvironment",
 CellID->2089232889],

Cell[CellGroupData[{

Cell["Lemma: Useful type information about Exponent Vectors", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.613107642672458*^9, 3.613107645119662*^9}, {
  3.613107977396778*^9, 3.61310799422081*^9}, {3.613108113413301*^9, 
  3.613108114591258*^9}},
 CellID->1181039367],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"isExponentVector", "[", "u", "]"}]], 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     SubscriptBox["u", "k"], "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(22)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:18996081", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612686415244599*^9, 3.612686430972688*^9}, {
   3.613107806111335*^9, 3.613107809894197*^9}, 3.613963896639178*^9},
 CellTags->{
  "ID:18996081", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "22"},
 CellID->18996081],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->489695439]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->496071433],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(22)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(22)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", "n"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "D"], "[", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             SubscriptBox[
              StyleBox["u", "ExpressionVariable"], 
              StyleBox["k", "ExpressionVariable"]], 
             TagBox[")", "AutoParentheses"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", "n"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "D"], "[", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             SubscriptBox[
              StyleBox["u", "ExpressionVariable"], 
              StyleBox["k", "ExpressionVariable"]], 
             TagBox[")", "AutoParentheses"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:18996081"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p18996081-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p18996081-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:18996081", "Proof|ID:18996081-1"},
 CellID->1450282884],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p18996081-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->517597611]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->844735177],

Cell[CellGroupData[{

Cell["\<\
Lemma: Useful type information about Tuples of Exponent Vectors\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.613963924083634*^9, 3.613963928677909*^9}},
 CellID->1405903066],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.611397133048281*^9, 3.611397139772798*^9}},
 CellID->1825557675],

Cell[BoxData[
 RowBox[{"isExpVectorTuple", "[", 
  RowBox[{"A", 
   TagBox["\:293a",
    Identity,
    SyntaxForm->"a*b"], "x"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpVectorTuple\:293a)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:313023917", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611397142693071*^9, 3.611397151665424*^9}},
 CellTags->{
  "ID:313023917", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExpVectorTuple\:293a"},
 CellID->313023917],

Cell[BoxData[
 RowBox[{"isTuple", "[", "A", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(23)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1697603634", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612080636687067*^9, 3.612080639954524*^9}},
 CellTags->{
  "ID:1697603634", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "23"},
 CellID->1697603634],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->173042086]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1570375643],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpVectorTuple\:293a)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpVectorTuple\:293a)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isExpVectorTuple", "[", 
           RowBox[{
             StyleBox["A", "ExpressionVariable"], "\:293a", 
             StyleBox["x", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isExpVectorTuple", "[", 
           RowBox[{
             StyleBox["A", "ExpressionVariable"], "\:293a", 
             StyleBox["x", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:313023917"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p313023917-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:313023917|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->360389676],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p313023917-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{298., {24., 31.}}]], "ProofInfoBody",
 CellID->511935660]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(23)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(23)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTuple", "[", 
           StyleBox["A", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTuple", "[", 
           StyleBox["A", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1697603634"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1697603634-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1697603634-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1697603634", "Proof|ID:1697603634-1"},
 CellID->1339747217],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1697603634-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 32.}}]], "ProofInfoBody",
 CellID->918500356]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Divisibility, lcm", "Subsubsection",
 CellChangeTimes->{{3.603697387554676*^9, 3.603697391538114*^9}, {
   3.603794799628197*^9, 3.603794800587016*^9}, {3.610962056411503*^9, 
   3.610962056941672*^9}, 3.612086919434023*^9},
 CellID->513543808],

Cell[TextData[{
 "Now, we concentrate on divisibility and ",
 Cell[BoxData[
  FormBox["lcm", TraditionalForm]]],
 ". In particular, we first show the very elementary property of divisibility \
being a partial order relation. For this we cannot use the definition of \
partial orderings in section \[OpenCurlyDoubleQuote]Definitions\
\[CloseCurlyDoubleQuote], but really have to state and prove the three \
properties individually. The reason for this lies in the fact that we defined \
being a partial ordering for relations of ",
 StyleBox["domains",
  FontSlant->"Italic"],
 ", but divisibility is not a relation of a particular domain!"
}], "Text",
 CellChangeTimes->{{3.603695133411118*^9, 3.603695148403477*^9}, {
  3.60369518038386*^9, 3.603695245105789*^9}, {3.60369583948984*^9, 
  3.603695840768871*^9}, {3.60369742806676*^9, 3.603697443773691*^9}, {
  3.603698574534117*^9, 3.603698603834412*^9}, {3.605590297343503*^9, 
  3.605590397292874*^9}, {3.615089898132538*^9, 3.615089898468882*^9}, {
  3.615089947101388*^9, 3.615089947920825*^9}},
 CellID->1070024737],

Cell["", "OpenEnvironment",
 CellID->91834526],

Cell[CellGroupData[{

Cell["\<\
Lemma: \[OpenCurlyDoubleQuote]Divisibility is partial order relation\
\[CloseCurlyDoubleQuote]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.603697526276883*^9, 3.603697555599883*^9}, {
  3.603697622636361*^9, 3.603697622805005*^9}},
 CellID->1104374192],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", 
   RowBox[{"x", ",", "y", ",", "z"}], "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.605590417790837*^9, 3.605590426210914*^9}},
 CellID->1959503054],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[VerticalBar]", "y"}], "\[And]", 
    RowBox[{"y", "\[VerticalBar]", "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", "\[Equal]", "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(antisymmetry of divisibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1313148627", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.603697562999945*^9, 3.60369760346191*^9}, {
  3.605589787972725*^9, 3.605589788157185*^9}, {3.60559044039032*^9, 
  3.605590466353946*^9}},
 CellTags->{
  "ID:1313148627", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "antisymmetry of divisibility"},
 CellID->1313148627],

Cell[BoxData[
 RowBox[{"x", "\[VerticalBar]", "x"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reflexivity of divisibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:785542046", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.605590611796865*^9, 3.605590615665492*^9}},
 CellTags->{
  "ID:785542046", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of divisibility"},
 CellID->785542046],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[VerticalBar]", "y"}], "\[And]", 
    RowBox[{"y", "\[VerticalBar]", "z"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", "\[VerticalBar]", "z"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(transitivity of divisibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1055765423", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.605590634637005*^9, 3.605590645293616*^9}},
 CellTags->{
  "ID:1055765423", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity of divisibility"},
 CellID->1055765423],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1453338991]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2087869455],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(antisymmetry of divisibility)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(antisymmetry of divisibility)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1313148627"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1313148627-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1313148627-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1313148627", "Proof|ID:1313148627-1"},
 CellID->513842368],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1313148627-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->847402406]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reflexivity of divisibility)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reflexivity of divisibility)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:785542046"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p785542046-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p785542046-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:785542046", "Proof|ID:785542046-1"},
 CellID->530754490],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p785542046-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->133907625]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(transitivity of divisibility)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(transitivity of divisibility)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1055765423"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1055765423-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1055765423-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1055765423", "Proof|ID:1055765423-1"},
 CellID->985028352],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1055765423-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->794383855]
}, Closed]],

Cell[TextData[{
 "Now we move on to the elementary properties of the ",
 Cell[BoxData[
  FormBox["lcm", TraditionalForm]]],
 " function, where for instance we show that in the way it was defined in \
section ",
 ButtonBox["Specific Notions for Complexity Analysis",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::SpecificNotions"],
 " it really possesses all the necessary properties of a least common \
multiple (formulas ",
 ButtonBox["(defining property of lcm 1)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1065951991"],
 ", ",
 ButtonBox["(defining property of lcm 2)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1071215803"],
 " and ",
 ButtonBox["(defining property of lcm 3)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1028467786"],
 "). Also, we show that ",
 Cell[BoxData[
  FormBox["lcm", TraditionalForm]]],
 " is both commutative and idempotent."
}], "Text",
 CellChangeTimes->{{3.615089960560773*^9, 3.615089964901522*^9}, {
  3.615090026373386*^9, 3.615090154541116*^9}, {3.615090192695064*^9, 
  3.615090200142339*^9}, {3.615090586325248*^9, 3.615090643429489*^9}, {
  3.61509071986217*^9, 3.615090739983838*^9}},
 CellID->326509238],

Cell["", "OpenEnvironment",
 CellID->1361910085],

Cell[CellGroupData[{

Cell["Lemma: Properties of lcm", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599385254886341*^9, 3.599385265479904*^9}, {
  3.600670458242626*^9, 3.600670462716899*^9}},
 CellID->113793713],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", 
   RowBox[{"x", ",", "y"}], "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.600591837947372*^9, 3.600591845662776*^9}, {
  3.60558980097286*^9, 3.605589804850805*^9}},
 CellID->2077457411],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"l", "=", 
   RowBox[{"lcm", "[", 
    RowBox[{"x", ",", "y"}], "]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.600591850114923*^9, 3.600591858644207*^9}, {
  3.60558980959669*^9, 3.605589809781825*^9}},
 CellID->1354533749],

Cell[BoxData[
 RowBox[{"isExponentVector", "[", "l", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of lcm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1918380148", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601024795747565*^9, 3.601024801468948*^9}, {
   3.601025311754523*^9, 3.601025315465251*^9}, 3.60369350510125*^9, {
   3.605589813391252*^9, 3.605589816956359*^9}},
 CellTags->{
  "ID:1918380148", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of lcm"},
 CellID->1918380148],

Cell[BoxData[
 RowBox[{"x", "\[VerticalBar]", "l"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(defining property of lcm 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1065951991", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600591869787397*^9, 3.600591879169281*^9}, {
  3.605589819488541*^9, 3.605589819674039*^9}},
 CellTags->{
  "ID:1065951991", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 1"},
 CellID->1065951991],

Cell[BoxData[
 RowBox[{"y", "\[VerticalBar]", "l"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(defining property of lcm 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1071215803", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600591888488988*^9, 3.600591890440147*^9}, {
  3.605589820694284*^9, 3.605589820873533*^9}},
 CellTags->{
  "ID:1071215803", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 2"},
 CellID->1071215803],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"isExponentVector", "[", "z", "]"}]], 
   RowBox[{
    RowBox[{"x", "\[VerticalBar]", "z"}], "\[And]", 
    RowBox[{"y", "\[VerticalBar]", "z"}]}]], 
  RowBox[{"l", "\[VerticalBar]", "z"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(defining property of lcm 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1028467786", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599385274313068*^9, 3.59938540485549*^9}, {
  3.600591896362789*^9, 3.6005918980189*^9}, {3.605589823870581*^9, 
  3.605589831433238*^9}},
 CellTags->{
  "ID:1028467786", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 3"},
 CellID->1028467786],

Cell[BoxData[
 RowBox[{"l", "\[Equal]", 
  RowBox[{"lcm", "[", 
   RowBox[{"y", ",", "x"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(commutativity of lcm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:98270815", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600670472278188*^9, 3.600670479700486*^9}, {
   3.605589833091376*^9, 3.60558983327699*^9}, 3.614938363879716*^9},
 CellTags->{
  "ID:98270815", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "commutativity of lcm"},
 CellID->98270815],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[VerticalBar]", "y"}], "\[Implies]", 
  RowBox[{"l", "\[Equal]", "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(24)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:622866956", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.603694808624698*^9, 3.603694819240047*^9}, {
  3.604135567083501*^9, 3.604135573192501*^9}, {3.605589834714409*^9, 
  3.605589834918952*^9}},
 CellTags->{
  "ID:622866956", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "24"},
 CellID->622866956],

Cell[BoxData[
 RowBox[{
  RowBox[{"lcm", "[", 
   RowBox[{"x", ",", "x"}], "]"}], "\[Equal]", 
  "x"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(idempotency of lcm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:810936059", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.604033473348478*^9, 3.604033491164522*^9}, {
  3.604135562798208*^9, 3.604135569569877*^9}, {3.605589836005819*^9, 
  3.605589836191254*^9}},
 CellTags->{
  "ID:810936059", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "idempotency of lcm"},
 CellID->810936059],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1802578373]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2058203328],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of lcm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of lcm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isExponentVector", "[", 
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isExponentVector", "[", 
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1918380148"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1918380148-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1918380148-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1918380148", "Proof|ID:1918380148-1"},
 CellID->1760699056],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1918380148-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->524153589]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(defining property of lcm 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(defining property of lcm 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1065951991"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1065951991-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1065951991-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1065951991", "Proof|ID:1065951991-1"},
 CellID->1695070992],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1065951991-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellID->542892896]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(defining property of lcm 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(defining property of lcm 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1071215803"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1071215803-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1071215803-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1071215803", "Proof|ID:1071215803-1"},
 CellID->1919630296],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1071215803-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellID->59184554]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(defining property of lcm 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(defining property of lcm 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["z", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"lcm", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "\[VerticalBar]", 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["z", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"lcm", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "\[VerticalBar]", 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1028467786"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1028467786-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1028467786-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1028467786", "Proof|ID:1028467786-1"},
 CellID->488707796],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1028467786-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1260793235]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(commutativity of lcm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(commutativity of lcm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:98270815"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p98270815-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p98270815-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:98270815", "Proof|ID:98270815-1"},
 CellID->615308222],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p98270815-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->65135107]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(24)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(24)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"lcm", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"lcm", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:622866956"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p622866956-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p622866956-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:622866956", "Proof|ID:622866956-1"},
 CellID->947602302],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p622866956-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1486588817]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(idempotency of lcm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(idempotency of lcm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"lcm", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:810936059"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p810936059-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p810936059-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:810936059", "Proof|ID:810936059-1"},
 CellID->898800181],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p810936059-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1345421969]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["deg, maxdeg", "Subsubsection",
 CellChangeTimes->{{3.599385138401371*^9, 3.599385138808661*^9}, {
   3.59938524762949*^9, 3.599385248816717*^9}, {3.59938542311225*^9, 
   3.599385426622009*^9}, {3.600594133912526*^9, 3.60059413441043*^9}, {
   3.600663712880736*^9, 3.600663713585332*^9}, {3.603695261414675*^9, 
   3.603695262731344*^9}, {3.603697411045795*^9, 3.603697411233159*^9}, {
   3.610962832890458*^9, 3.610962833554021*^9}, 3.612086920576863*^9},
 CellID->222564893],

Cell[TextData[{
 "Now we turn to ",
 Cell[BoxData[
  FormBox["deg", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["maxdeg", TraditionalForm]]],
 ". Note already here that some statements cannot be proved with the \
ComplexityProver in the general setting of arbitrary ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ", although they certainly hold."
}], "Text",
 CellChangeTimes->{{3.60369745050124*^9, 3.603697468625272*^9}, {
  3.615090755241047*^9, 3.61509075676264*^9}},
 CellID->1443198008],

Cell["", "OpenEnvironment",
 CellID->1609431449],

Cell[CellGroupData[{

Cell["Lemma: Properties of deg", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.600663721780016*^9, 3.600663731643733*^9}, {
  3.600670506943062*^9, 3.600670509708354*^9}},
 CellID->483808494],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", 
   RowBox[{"x", ",", "y"}], "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.600663738612628*^9, 3.60066374503771*^9}, {
  3.600670513433602*^9, 3.600670513589461*^9}, {3.605591898840703*^9, 
  3.60559190207814*^9}},
 CellID->1370144613],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{"deg", "[", "x", "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of deg)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1822207066", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601024677294421*^9, 3.601024684268393*^9}, 
   3.601024767970024*^9, {3.601025328818547*^9, 3.601025329521119*^9}, {
   3.605591906161902*^9, 3.605591906420405*^9}},
 CellTags->{
  "ID:1822207066", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of deg"},
 CellID->1822207066],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[VerticalBar]", "y"}], "\[Implies]", 
  RowBox[{
   RowBox[{"deg", "[", "x", "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"deg", "[", "y", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(25)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:455032720", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.600592754338011*^9, {3.601026778529277*^9, 3.601026779450594*^9}, {
   3.605591907748075*^9, 3.605591911622167*^9}},
 CellTags->{
  "ID:455032720", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "25"},
 CellID->455032720],

Cell[BoxData[
 RowBox[{
  RowBox[{"deg", "[", "x", "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{"deg", "[", 
   RowBox[{"lcm", "[", 
    RowBox[{"x", ",", "y"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(26)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:839534060", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601026804297156*^9, 3.601026819704311*^9}, {
  3.604835472801213*^9, 3.604835478508821*^9}, {3.605591913667157*^9, 
  3.605591918166856*^9}},
 CellTags->{
  "ID:839534060", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "26"},
 CellID->839534060],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->994235688]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1479111596],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of deg)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of deg)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "D"], "[", 
           RowBox[{"deg", "[", 
             StyleBox["x", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "D"], "[", 
           RowBox[{"deg", "[", 
             StyleBox["x", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1822207066"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1822207066-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1822207066-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1822207066", "Proof|ID:1822207066-1"},
 CellID->694957922],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1822207066-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->2069252785]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(25)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(25)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"deg", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"deg", "[", 
               StyleBox["y", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"deg", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"deg", "[", 
               StyleBox["y", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:455032720"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p455032720-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:455032720|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1987093019],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p455032720-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{371., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.60559770122359*^9, 3.614934125022828*^9},
 CellID->841372092]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(26)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(26)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"deg", "[", 
             StyleBox["x", "ExpressionVariable"], "]"}], 
           UnderscriptBox["\[LessEqual]", "D"], 
           RowBox[{"deg", "[", 
             RowBox[{"lcm", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"deg", "[", 
             StyleBox["x", "ExpressionVariable"], "]"}], 
           UnderscriptBox["\[LessEqual]", "D"], 
           RowBox[{"deg", "[", 
             RowBox[{"lcm", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:839534060"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p839534060-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:839534060|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->849905264],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p839534060-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{298., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677424085212*^9, 3.614934129301827*^9},
 CellID->135010444]
}, Closed]],

Cell[TextData[{
 "Unfortunately, formulas ",
 ButtonBox["(type of deg)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1822207066"],
 " and ",
 ButtonBox["(25)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:455032720"],
 " cannot be proved with the ComplexityProver, since induction and an \
inductive definition of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Sum]", "D"], TraditionalForm]]],
 " would be needed (at the meta-level, i.e. in the prover). Still, we can \
prove them for the special case of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 ", i.e. we add formula ",
 ButtonBox["(n=2)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1526008058"],
 " to our assumptions. Formula ",
 ButtonBox["(26)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:839534060"],
 " is an immediate consequence of ",
 ButtonBox["(25)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:455032720"],
 " and knowledge about the least common multiple."
}], "Text",
 CellChangeTimes->{{3.603693889460199*^9, 3.60369398306365*^9}, {
  3.603694237482784*^9, 3.603694273702211*^9}, {3.603694997535412*^9, 
  3.603695015563224*^9}, {3.60369509129421*^9, 3.603695091704905*^9}, {
  3.603711345306229*^9, 3.603711346796203*^9}, {3.605330379715028*^9, 
  3.605330520057546*^9}, {3.605592781713303*^9, 3.605592782008118*^9}, {
  3.605597767251704*^9, 3.605597787520424*^9}, {3.615090922269304*^9, 
  3.61509105913941*^9}, {3.61578765518744*^9, 3.615787672689558*^9}},
 CellID->1710837182],

Cell[TextData[{
 "The following lemma is completely obvious, but it is quite useful for \
proving formula ",
 ButtonBox["(100)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:724101256"],
 ":"
}], "Text",
 CellChangeTimes->{{3.601111745825865*^9, 3.601111771506509*^9}, {
   3.603693244367541*^9, 3.60369324437682*^9}, 3.60369590760846*^9, {
   3.615091112888886*^9, 3.615091121902472*^9}, {3.61578768653654*^9, 
   3.615787686980058*^9}},
 CellID->210483232],

Cell["", "OpenEnvironment",
 CellID->613968158],

Cell[CellGroupData[{

Cell["Lemma: Properties of maxdeg", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.601111670667011*^9, 3.601111677620015*^9}, {
  3.603693570786659*^9, 3.603693572739841*^9}},
 CellID->151723375],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.603693592515487*^9, 3.603693601609433*^9}, {
  3.605592979788017*^9, 3.605592981049454*^9}},
 CellID->717930824],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{"maxdeg", "[", "A", "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of maxdeg)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2123737964", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601025373736176*^9, 3.601025410339598*^9}, {
  3.605592976988099*^9, 3.605592977209148*^9}},
 CellTags->{
  "ID:2123737964", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of maxdeg"},
 CellID->2123737964],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   RowBox[{"deg", "[", 
    SubscriptBox["A", "i"], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"maxdeg", "[", "A", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(27)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:647995479", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601111690861085*^9, 3.601111736186579*^9}, {
  3.605592984897065*^9, 3.605592986470962*^9}},
 CellTags->{
  "ID:647995479", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "27"},
 CellID->647995479],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1382258393]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1369097639],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of maxdeg)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of maxdeg)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{"maxdeg", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{"maxdeg", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2123737964"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2123737964-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:2123737964|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1501203891],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2123737964-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{456., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.61493413569407*^9},
 CellID->1309062189]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(27)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(27)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{"deg", "[", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{"maxdeg", "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{"deg", "[", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{"maxdeg", "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:647995479"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p647995479-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:647995479|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->878198538],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p647995479-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{683., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677463139236*^9, 3.614934138973728*^9},
 CellID->1033778194]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["chainCrit", "Subsubsection",
 CellChangeTimes->{{3.598590308706875*^9, 3.598590309052363*^9}, {
   3.60483553002857*^9, 3.60483553140692*^9}, {3.610969712068877*^9, 
   3.610969712621627*^9}, 3.612086921609421*^9},
 CellTags->"sect::LemmataSpecificGeneralM",
 CellID->940386836],

Cell[TextData[{
 "Next we turn to predicate ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 ", one of the essential predicates in the complexity analysis. As can be \
seen, even for this predicate some general (i.e. for arbitrary ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ") properties can be proved (semi-) automatically with the ComplexityProver, \
where some of them are maybe not entirely obvious."
}], "Text",
 CellChangeTimes->{{3.603695939046775*^9, 3.603695968444433*^9}, {
  3.603696032324396*^9, 3.603696100885395*^9}, {3.603696162018489*^9, 
  3.603696166710084*^9}, {3.60403240607523*^9, 3.60403240644466*^9}, {
  3.604033539499506*^9, 3.60403354496173*^9}, {3.604835535622192*^9, 
  3.604835540908836*^9}},
 CellID->78870630],

Cell["", "OpenEnvironment",
 CellID->446526914],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Simple Properties of ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.598270427429897*^9, 3.598270443285998*^9}, {
  3.59869752717395*^9, 3.598697541847145*^9}, {3.600064493679118*^9, 
  3.600064495564889*^9}, {3.604835914664884*^9, 3.604835928385334*^9}, {
  3.614591033564793*^9, 3.614591039655972*^9}},
 CellID->915159816],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", 
   RowBox[{"x", ",", "y", ",", "z"}], "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.600594299491009*^9, 3.600594306885377*^9}, {
  3.605592997971095*^9, 3.605593001446338*^9}},
 CellID->2015253534],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[VerticalBar]", "y"}], "\[Implies]", 
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(28)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1452963633", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610964231848454*^9, 3.610964235390336*^9}},
 CellTags->{
  "ID:1452963633", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "28"},
 CellID->1452963633],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}], "\[Implies]", 
  RowBox[{"chainCrit", "[", 
   RowBox[{"y", ",", "x", ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(29)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1500097429", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.604988066065134*^9, 3.604988074700359*^9}, {
  3.605593009701047*^9, 3.605593011086523*^9}},
 CellTags->{
  "ID:1500097429", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "29"},
 CellID->1500097429],

Cell[TextData[{
 ButtonBox["(symmetry of chainCrit)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:279915377"],
 " is of course equivalent to ",
 ButtonBox["(29)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1500097429"],
 ", but for automated proving it is more convenient."
}], "EnvironmentText",
 CellChangeTimes->{{3.60498823029802*^9, 3.604988270451617*^9}, {
  3.615092066974498*^9, 3.615092097937807*^9}, {3.615787704136662*^9, 
  3.615787705058462*^9}},
 CellID->1103072691],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"chainCrit", "[", 
   RowBox[{"y", ",", "x", ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symmetry of chainCrit)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:279915377", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600670600949087*^9, 3.600670616566415*^9}, {
  3.604835559652795*^9, 3.604835561241135*^9}, {3.605593013739291*^9, 
  3.605593015142968*^9}},
 CellTags->{
  "ID:279915377", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCrit"},
 CellID->279915377],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[VerticalBar]", "y"}], "\[Or]", 
    RowBox[{"y", "\[VerticalBar]", "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(divisibility\[DoubleRightArrow]chainCrit)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1391995733", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.598270535060209*^9, {3.598270663721381*^9, 3.59827078110672*^9}, 
   3.599373645759145*^9, {3.600594290768886*^9, 3.600594339317884*^9}, {
   3.604835552786752*^9, 3.604835554099426*^9}, {3.605593005642005*^9, 
   3.605593008484257*^9}},
 CellTags->{
  "ID:1391995733", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "divisibility\[DoubleRightArrow]chainCrit"},
 CellID->1391995733],

Cell[BoxData[
 RowBox[{"chainCrit", "[", 
  RowBox[{"x", ",", "x", ",", "A"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reflexivity of chainCrit)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1271079949", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61398682409283*^9, 3.613986828513092*^9}},
 CellTags->{
  "ID:1271079949", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of chainCrit"},
 CellID->1271079949],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", 
    RowBox[{"\[LeftAngleBracket]", "z", "\[RightAngleBracket]"}]}], "]"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["let", 
     RowBox[{"l", "=", 
      RowBox[{"lcm", "[", 
       RowBox[{"x", ",", "y"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Or]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"\[Not]", 
           RowBox[{"z", "\[VerticalBar]", "l"}]}]},
         {
          RowBox[{
           RowBox[{"deg", "[", "l", "]"}], 
           UnderscriptBox["\[LessEqual]", "D"], 
           RowBox[{"deg", "[", 
            RowBox[{"lcm", "[", 
             RowBox[{"x", ",", "z"}], "]"}], "]"}]}]},
         {
          RowBox[{
           RowBox[{"deg", "[", "l", "]"}], 
           UnderscriptBox["\[LessEqual]", "D"], 
           RowBox[{"deg", "[", 
            RowBox[{"lcm", "[", 
             RowBox[{"z", ",", "y"}], "]"}], "]"}]}]}
        }]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit w.r.t. singleton)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:488309194", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600594464539363*^9, 3.600594474475368*^9}, {
  3.600594528086685*^9, 3.600594648138259*^9}, {3.603696880528362*^9, 
  3.603696913435873*^9}, {3.604835564748418*^9, 3.604835585411392*^9}, {
  3.605593017210365*^9, 3.605593027438672*^9}},
 CellTags->{
  "ID:488309194", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit w.r.t. singleton"},
 CellID->488309194],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "z"}]}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", "y", ",", "A"}], "]"}], "\[And]", 
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", "y", ",", 
      RowBox[{"\[LeftAngleBracket]", "z", "\[RightAngleBracket]"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit\:293a 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1322351588", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61096757834707*^9, 3.610967588178448*^9}},
 CellTags->{
  "ID:1322351588", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a 1"},
 CellID->1322351588],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", "y", ",", "A"}], "]"}], "\[And]", 
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", "y", ",", 
      RowBox[{"\[LeftAngleBracket]", "z", "\[RightAngleBracket]"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "z"}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit\:293a 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1269486125", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610967597791647*^9, 3.610967603640737*^9}},
 CellTags->{
  "ID:1269486125", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a 2"},
 CellID->1269486125],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "z"}]}], "]"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", "y", ",", "A"}], "]"}], "\[And]", 
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", "y", ",", 
      RowBox[{"\[LeftAngleBracket]", "z", "\[RightAngleBracket]"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit\:293a)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:941129646", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600594384102059*^9, 3.600594418141308*^9}, {
  3.600663332230509*^9, 3.600663340855289*^9}, {3.604835592315911*^9, 
  3.604835595790569*^9}, {3.605593029462042*^9, 3.605593032599564*^9}},
 CellTags->{
  "ID:941129646", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a"},
 CellID->941129646],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}]}], "]"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit\:293a I)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1657568255", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600680130889949*^9, 3.600680156069857*^9}, {
  3.604835597340518*^9, 3.604835598671049*^9}, {3.605593033926529*^9, 
  3.605593035918118*^9}},
 CellTags->{
  "ID:1657568255", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a I"},
 CellID->1657568255],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->41255160]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1530445502],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(28)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(28)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1452963633"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1452963633-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1452963633|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1783184226],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1452963633-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{557., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677486386739*^9, 3.614934154188781*^9},
 CellID->1043672466]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(29)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(29)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1500097429"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1500097429-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1500097429|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1859160805],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1500097429-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{323., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677491785848*^9, 3.614934159565059*^9},
 CellID->1975569240]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symmetry of chainCrit)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symmetry of chainCrit)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:279915377"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p279915377-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:279915377|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1592316742],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p279915377-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{298., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.61467749597781*^9, 3.614934163452816*^9},
 CellID->1362235388]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(divisibility\[DoubleRightArrow]chainCrit)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(divisibility\[DoubleRightArrow]chainCrit)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1391995733"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1391995733-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1391995733|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->218343730],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1391995733-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{298., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677508897563*^9, 3.614934168500771*^9},
 CellID->84993298]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reflexivity of chainCrit)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reflexivity of chainCrit)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"chainCrit", "[", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], ",", 
             StyleBox["x", "ExpressionVariable"], ",", 
             StyleBox["A", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"chainCrit", "[", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], ",", 
             StyleBox["x", "ExpressionVariable"], ",", 
             StyleBox["A", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1271079949"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1271079949-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1271079949-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1271079949", "Proof|ID:1271079949-1"},
 CellID->1431739973],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1271079949-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934172252087*^9},
 CellID->1498238151]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit w.r.t. singleton)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit w.r.t. singleton)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox["z", "ExpressionVariable"], 
                 "\[RightAngleBracket]"}]}], "]"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["let", 
              GridBox[{{
                 RowBox[{
                   StyleBox["l", "ExpressionVariable"], "=", 
                   RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}]}}]], 
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["l", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"deg", "[", 
                   StyleBox["l", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"deg", "[", 
                   RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["z", "ExpressionVariable"]}], "]"}], "]"}]}], 
               "\[Or]", 
               RowBox[{
                 RowBox[{"deg", "[", 
                   StyleBox["l", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"deg", "[", 
                   RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox["z", "ExpressionVariable"], 
                 "\[RightAngleBracket]"}]}], "]"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["let", 
              GridBox[{{
                 RowBox[{
                   StyleBox["l", "ExpressionVariable"], "=", 
                   RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}]}}]], 
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["l", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"deg", "[", 
                   StyleBox["l", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"deg", "[", 
                   RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["z", "ExpressionVariable"]}], "]"}], "]"}]}], 
               "\[Or]", 
               RowBox[{
                 RowBox[{"deg", "[", 
                   StyleBox["l", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"deg", "[", 
                   RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:488309194"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p488309194-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:488309194|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1697069805],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p488309194-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{298., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677515977001*^9, 3.614934180044251*^9},
 CellID->1563635678]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit\:293a 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit\:293a 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["z", "ExpressionVariable"]}]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"chainCrit", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{"chainCrit", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox["z", "ExpressionVariable"], 
                   "\[RightAngleBracket]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["z", "ExpressionVariable"]}]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"chainCrit", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{"chainCrit", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox["z", "ExpressionVariable"], 
                   "\[RightAngleBracket]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1322351588"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1322351588-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1322351588|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1376405718],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1322351588-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{473., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677519840895*^9, 3.614934184299744*^9},
 CellID->1107600169]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit\:293a 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit\:293a 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"chainCrit", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{"chainCrit", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox["z", "ExpressionVariable"], 
                   "\[RightAngleBracket]"}]}], "]"}]}], "\[Implies]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["z", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"chainCrit", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{"chainCrit", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox["z", "ExpressionVariable"], 
                   "\[RightAngleBracket]"}]}], "]"}]}], "\[Implies]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["z", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1269486125"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1269486125-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1269486125|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->779731268],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1269486125-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{473., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677525360569*^9, 3.614934188507079*^9},
 CellID->1608520609]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit\:293a)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit\:293a)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["z", "ExpressionVariable"]}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox["z", "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["z", "ExpressionVariable"]}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox["z", "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:941129646"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p941129646-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:941129646|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->581146426],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p941129646-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{298., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677530151874*^9, 3.614934192618885*^9},
 CellID->471861581]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit\:293a I)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit\:293a I)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isExponentVector", "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:293a", 
                 StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"chainCrit", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1657568255"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1657568255-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1657568255|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1717614685],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1657568255-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{575., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677538092409*^9, 3.614934196435303*^9},
 CellID->485455762]
}, Closed]],

Cell[TextData[{
 "The importance of formula ",
 ButtonBox["(divisibility\[DoubleRightArrow]chainCrit)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1391995733"],
 " lies in the fact that it guarantees that for many pairs of points the \
chain criterion holds for which relatively little is known (apart from one \
dividing the other), a fact which will be exploited later in section ",
 ButtonBox["Theorems",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Theorems"],
 ".\nThe most important formula, however, is probably ",
 ButtonBox["(chainCrit\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:941129646"],
 ", which allows to deduce that the chain criterion holds for a pair of \
points w.r.t. a tuple where a new element is appended. This will also be very \
useful in section ",
 ButtonBox["Theorems",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Theorems"],
 ", since almost the entire complexity analysis does nothing else than \
expressing the quantities ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["W", TraditionalForm]]],
 " where some new element is appended to the original tuple (see sections \
below)."
}], "Text",
 CellChangeTimes->{{3.604033631538728*^9, 3.604034082791525*^9}, {
  3.604034129362665*^9, 3.604034129510831*^9}, {3.604835619077284*^9, 
  3.604835653759527*^9}, {3.613898330372367*^9, 3.613898354734772*^9}, {
  3.615092364710769*^9, 3.615092458206911*^9}},
 CellID->633826383],

Cell[TextData[{
 "The following theorem (in particular formula ",
 ButtonBox["(chainCrit holds for some element)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1053941118"],
 ") is nowhere stated explicitly in the literature, despite its \
significance:"
}], "Text",
 CellChangeTimes->{{3.596372357203384*^9, 3.596372395076858*^9}, {
  3.600755220595843*^9, 3.600755221452762*^9}, {3.600758016583369*^9, 
  3.600758205672081*^9}, {3.6007583454538*^9, 3.600758370906001*^9}, {
  3.601021921649681*^9, 3.601021922240854*^9}, {3.604034767639808*^9, 
  3.604034782747032*^9}, {3.61509269992161*^9, 3.615092720606138*^9}, {
  3.61509275712434*^9, 3.615092757144065*^9}},
 CellID->897047203],

Cell["", "OpenEnvironment",
 CellChangeTimes->{3.600758233370479*^9},
 CellID->1158939693],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[OpenCurlyDoubleQuote]Every point is part of pair for which Chain \
Criterion holds\[CloseCurlyDoubleQuote]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.600755207345921*^9, 3.600755224346956*^9}, {
   3.600755261200723*^9, 3.60075528445457*^9}, 3.600758233371297*^9, {
   3.604034740503336*^9, 3.60403474138623*^9}, {3.613963569956995*^9, 
   3.613963576333874*^9}, {3.615626802750759*^9, 3.615626819188397*^9}},
 CellID->1661037347],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.611986865501834*^9, 3.611986870199394*^9}},
 CellID->1227578886],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.601021988869747*^9, 3.601021998525553*^9}, {
  3.605593674192302*^9, 3.605593677801931*^9}},
 CellID->1399902871],

Cell[BoxData[
 RowBox[{"chainCrit", "[", 
  RowBox[{"x", ",", 
   SubscriptBox["A", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      RowBox[{
       UnderscriptBox[
        SubscriptBox["argmin", "D"], 
        RowBox[{
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
       RowBox[{"deg", "[", 
        RowBox[{"lcm", "[", 
         RowBox[{"x", ",", 
          SubscriptBox["A", "j"]}], "]"}], "]"}]}],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}]], ",", "A"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(30)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:469812703", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601022014095622*^9, 3.601022081749156*^9}, {
  3.604835670496218*^9, 3.604835684155265*^9}, {3.604993501662679*^9, 
  3.604993503824559*^9}, {3.605593682735767*^9, 3.605593685615502*^9}, {
  3.611036077839767*^9, 3.611036117293099*^9}, {3.611045374633202*^9, 
  3.611045379539194*^9}},
 CellTags->{
  "ID:469812703", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "30"},
 CellID->469812703],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Exists]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", 
    SubscriptBox["A", "i"], ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit holds for some element)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1053941118", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600755096499711*^9, 3.600755136820753*^9}, 
   3.600755365432022*^9, 3.600758233372402*^9, {3.600758398427107*^9, 
   3.600758404103785*^9}, 3.601022001143143*^9, 3.604835674109994*^9, {
   3.605593687346806*^9, 3.605593687532257*^9}},
 CellTags->{
  "ID:1053941118", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit holds for some element"},
 CellID->1053941118],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellChangeTimes->{3.600758233372592*^9},
 CellID->1542713411]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellChangeTimes->{3.600758233372653*^9},
 CellID->1684866112],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(30)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(30)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{"chainCrit", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["argmin", "D"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}]}]], 
                   TagBox[")", "AutoParentheses"]}], ",", 
                 StyleBox["A", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{"chainCrit", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["argmin", "D"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}]}]], 
                   TagBox[")", "AutoParentheses"]}], ",", 
                 StyleBox["A", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:469812703"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p469812703-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:469812703|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1066023196],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p469812703-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{402., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{
  3.612862069332567*^9, 3.614677552999552*^9, {3.6149342024988*^9, 
   3.614934208890158*^9}},
 CellID->370971114]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit holds for some element)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit holds for some element)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1053941118"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1053941118-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1053941118|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->875669745],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1053941118-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{347., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614677558151284*^9, 3.61493421208177*^9},
 CellID->419994512]
}, Closed]],

Cell[TextData[{
 "The significance of formula ",
 ButtonBox["(chainCrit holds for some element)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1053941118"],
 " lies in the fact that provides the basis for proving formula ",
 ButtonBox["(M=M*)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:872843574"],
 ", which in turn is of utmost importance for establishing the link between \
the definition of ",
 StyleBox["chainCrit",
  FontFamily->"Courier 10 Pitch"],
 " as it is done in formula ",
 ButtonBox["(chain criterion)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1784038925"],
 ", where nothing is known about the objects constituing the critical pair, \
and the usual definition of ",
 StyleBox["chainCrit",
  FontFamily->"Courier 10 Pitch"],
 " in Buchberger\[CloseCurlyQuote]s algorithm, where the two objects have to \
appear at distinct positions in the tuple. See also the first remark in \
section ",
 ButtonBox["Specific Notions for Complexity Analysis",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::SpecificNotions"],
 "."
}], "Text",
 CellChangeTimes->{{3.604034806927134*^9, 3.604035071894406*^9}, {
   3.609039868818578*^9, 3.60903987221449*^9}, {3.609039916737065*^9, 
   3.609039929186407*^9}, {3.611044813576878*^9, 3.611044846497457*^9}, {
   3.611044887205434*^9, 3.611044964939574*^9}, 3.613898376261543*^9, {
   3.615092540434605*^9, 3.6150925555846*^9}, {3.615092763486395*^9, 
   3.615092837707637*^9}, {3.615092891670516*^9, 3.61509292642324*^9}},
 CellID->499366557]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox["M", TraditionalForm]]]], "Subsubsection",
 CellChangeTimes->{
  3.598590330647606*^9, {3.59859303601266*^9, 3.598593036751502*^9}, {
   3.599803628216441*^9, 3.599803662444739*^9}, {3.600663389387906*^9, 
   3.60066339752832*^9}, {3.600669606867778*^9, 3.600669608663923*^9}, {
   3.603693802301153*^9, 3.603693802674752*^9}, {3.611396517719187*^9, 
   3.611396518560385*^9}, 3.61198689599401*^9, {3.611986944188105*^9, 
   3.611986944594561*^9}, 3.612086922862954*^9},
 CellID->1911387542],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.600669704002672*^9, 3.600669710800482*^9}, {
  3.603178007957734*^9, 3.603178010202393*^9}, {3.603693164679785*^9, 
  3.603693166840385*^9}},
 CellID->1584445324],

Cell[TextData[{
 "The following lemma proves the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", 
    RowBox[{"x", ",", "A"}], "]"}], TraditionalForm]]],
 " as defined in ",
 ButtonBox["Specific Notions for Complexity Analysis",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::SpecificNotions"],
 ", as well as their \[OpenCurlyDoubleQuote]types\[CloseCurlyDoubleQuote]. \
Note that formula ",
 ButtonBox["(31)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1635347335"],
 " follows immediately from formula ",
 ButtonBox["(chainCrit holds for some element)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1053941118"],
 " in the previous subsection."
}], "Text",
 CellChangeTimes->{{3.600671049597399*^9, 3.600671130116441*^9}, {
  3.604034558349981*^9, 3.604034722540442*^9}, {3.604035187620704*^9, 
  3.604035195323783*^9}, {3.609039946721228*^9, 3.609039959605225*^9}, {
  3.6150929502196*^9, 3.615092976044572*^9}, {3.615093006738627*^9, 
  3.615093053519283*^9}, {3.615787725517148*^9, 3.615787726850404*^9}},
 CellID->855235021],

Cell["", "OpenEnvironment",
 CellID->615718314],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.600670117424204*^9, 3.600670128481216*^9}, {
  3.600681751017215*^9, 3.600681757509243*^9}, {3.603693724983408*^9, 
  3.603693725702934*^9}},
 CellID->529250115],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.600669621415873*^9, 3.600669627220381*^9}, {
   3.600681719738835*^9, 3.600681719863338*^9}, 3.600687192294685*^9, {
   3.605593729875551*^9, 3.605593733418346*^9}},
 CellID->1727832338],

Cell[BoxData[
 RowBox[{"1", 
  UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
  RowBox[{"\[LeftBracketingBar]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"deg", "[", 
      RowBox[{"lcm", "[", 
       RowBox[{"x", ",", 
        SubscriptBox["A", "i"]}], "]"}], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{"chainCrit", "[", 
      RowBox[{"x", ",", 
       SubscriptBox["A", "i"], ",", "A"}], "]"}]}], "\[RightAngleBracket]"}], 
   "\[RightBracketingBar]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(31)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1635347335", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600669654055028*^9, 3.600669677763621*^9}, {
  3.603178026460968*^9, 3.603178028773384*^9}, {3.603693178382441*^9, 
  3.603693180544087*^9}, {3.604835700756337*^9, 3.6048357084895*^9}, {
  3.605593737361095*^9, 3.605593740327825*^9}},
 CellTags->{
  "ID:1635347335", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "31"},
 CellID->1635347335],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[SquareSubsetEqual]", "D"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"deg", "[", 
      RowBox[{"lcm", "[", 
       RowBox[{"x", ",", 
        SubscriptBox["A", "i"]}], "]"}], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{"chainCrit", "[", 
      RowBox[{"x", ",", 
       SubscriptBox["A", "i"], ",", "A"}], "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(32)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1901113035", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611391180604876*^9, 3.61139120529767*^9}},
 CellTags->{
  "ID:1901113035", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "32"},
 CellID->1901113035],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{"M", "[", 
    RowBox[{"x", ",", "A"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of M (binary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1760547878", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601025428318825*^9, 3.601025449781367*^9}, {
  3.605593742574004*^9, 3.60559374301684*^9}},
 CellTags->{
  "ID:1760547878", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M (binary)"},
 CellID->1760547878],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{"M", "[", "A", "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of M (unary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1896797254", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601025455970993*^9, 3.601025457957251*^9}, {
  3.605593744970331*^9, 3.60559374521053*^9}},
 CellTags->{
  "ID:1896797254", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M (unary)"},
 CellID->1896797254],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   RowBox[{"M", "[", 
    RowBox[{
     SubscriptBox["A", "i"], ",", "A"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"M", "[", "A", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(33)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1611647919", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.60068172735016*^9, 3.600681742301058*^9}, 
   3.60068717188043*^9, {3.605593747386035*^9, 3.605593748973873*^9}},
 CellTags->{
  "ID:1611647919", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "33"},
 CellID->1611647919],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->481543849]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1953187198],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(31)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(31)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{"1", 
               UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{"1", 
               UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1635347335"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1635347335-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1635347335|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->786500781],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1635347335-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{422., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934228809499*^9},
 CellID->937118081]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(32)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(32)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1901113035"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1901113035-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1901113035|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->573625489],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1901113035-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{527., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934233697447*^9},
 CellID->1065390053]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of M (binary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of M (binary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{"M", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{"M", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1760547878"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1760547878-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1760547878|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->177141490],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1760547878-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{633., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.61493423753748*^9},
 CellID->1878691620]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of M (unary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of M (unary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1896797254"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1896797254-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 GeneratedCell->False,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.611046968621162*^9, 3.611046971643871*^9}},
 CellTags->
  "Proof|ID:1896797254|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->367799976],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1896797254-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{663., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{
  3.614934241928323*^9, {3.615007330614058*^9, 3.61500733576884*^9}},
 CellID->2136307825]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(33)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(33)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{"M", "[", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{"M", "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{"M", "[", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{"M", "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1611647919"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1611647919-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1611647919|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->2107085874],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1611647919-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{827., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934246352989*^9},
 CellID->1300790402]
}, Closed]],

Cell[TextData[{
 "The next lemma describes how ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], 
    UnderscriptBox["\[LessEqual]", "D"], "u"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", 
    UnderscriptBox["\[LessEqual]", "D"], 
    RowBox[{"M", "[", "A", "]"}]}], TraditionalForm]]],
 " etc. can be expressed in terms of formulas where ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 " does not appear any more, in close similarity to lemma \
\[OpenCurlyDoubleQuote]General Properties of min, max\[CloseCurlyDoubleQuote] \
in section ",
 ButtonBox["Lemmata on Basic Notions / min, max",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Lemmataminmax"],
 "."
}], "Text",
 CellChangeTimes->{{3.615093116234991*^9, 3.615093150158695*^9}, {
  3.615093202534156*^9, 3.615093235088885*^9}},
 CellID->1704242160],

Cell["", "OpenEnvironment",
 CellID->1170519931],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "\[LessEqual]", "u"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", "\[LessEqual]", "M"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.612767145769546*^9, 3.61276716372312*^9}, {
  3.612767737721349*^9, 3.612767738923732*^9}},
 CellID->1280894414],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", "u", "]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612767172779694*^9, 3.6127671866784*^9}, 
   3.612767501521475*^9},
 CellID->664584010],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612767486028775*^9, 3.612767498432565*^9}},
 CellID->1887894187],

Cell[BoxData[
 RowBox[{
  RowBox[{"u", 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"M", "[", "A", "]"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"r", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{"u", 
    UnderscriptBox["\[LessEqual]", "D"], 
    RowBox[{"M", "[", 
     RowBox[{
      SubscriptBox["A", "r"], ",", "A"}], 
     "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LessEqual]M (unary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1545946275", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600148644275676*^9, 3.600148717385941*^9}, {
   3.60014875851393*^9, 3.600148849997594*^9}, {3.600409639616649*^9, 
   3.600409663874679*^9}, 3.600688366509062*^9, {3.612767361373108*^9, 
   3.612767397598277*^9}, {3.61276743295501*^9, 3.612767453144141*^9}, {
   3.612767502993384*^9, 3.612767504023866*^9}, {3.612767537557185*^9, 
   3.612767537890943*^9}},
 CellTags->{
  "ID:1545946275", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M (unary)"},
 CellID->1545946275],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"M", "[", "A", "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"r", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    RowBox[{"M", "[", 
     RowBox[{
      SubscriptBox["A", "r"], ",", "A"}], "]"}], 
    UnderscriptBox["\[LessEqual]", "D"], "u"}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\[LessEqual] (unary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1018591902", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.60014820740983*^9, 3.600148241519262*^9}, {
   3.600148274081915*^9, 3.600148389956722*^9}, {3.600148820171261*^9, 
   3.600148827272071*^9}, {3.600148938757218*^9, 3.600148941969407*^9}, {
   3.60014906006257*^9, 3.600149061755045*^9}, {3.600405746880696*^9, 
   3.600405753217857*^9}, {3.60040967359848*^9, 3.600409688162742*^9}, 
   3.600688352024196*^9, 3.61276736262328*^9, {3.612767513907566*^9, 
   3.61276754927719*^9}},
 CellTags->{
  "ID:1018591902", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual] (unary)"},
 CellID->1018591902],

Cell[BoxData[
 RowBox[{
  RowBox[{"u", 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"M", "[", 
    RowBox[{"x", ",", "A"}], "]"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"s", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", 
      SubscriptBox["A", "s"], ",", "A"}], "]"}]], 
   RowBox[{"u", 
    UnderscriptBox["\[LessEqual]", "D"], 
    RowBox[{"deg", "[", 
     RowBox[{"lcm", "[", 
      RowBox[{"x", ",", 
       SubscriptBox["A", "s"]}], "]"}], "]"}]}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LessEqual]M (binary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:382103250", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600148644275676*^9, 3.600148717385941*^9}, {
   3.60014875851393*^9, 3.600148849997594*^9}, {3.600149073969659*^9, 
   3.600149085676266*^9}, {3.600409695019163*^9, 3.600409712869742*^9}, 
   3.600688369415438*^9, 3.612767364076105*^9, {3.612767554093445*^9, 
   3.612767639781588*^9}, {3.61276791056231*^9, 3.612767913393056*^9}},
 CellTags->{
  "ID:382103250", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M (binary)"},
 CellID->382103250],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"M", "[", 
    RowBox[{"x", ",", "A"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"s", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", 
      SubscriptBox["A", "s"], ",", "A"}], "]"}]], 
   RowBox[{
    RowBox[{"deg", "[", 
     RowBox[{"lcm", "[", 
      RowBox[{"x", ",", 
       SubscriptBox["A", "s"]}], "]"}], "]"}], 
    UnderscriptBox["\[LessEqual]", "D"], "u"}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\[LessEqual] (binary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:450457679", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612767643468274*^9, 3.612767682842289*^9}, {
  3.61276791504931*^9, 3.612767918615729*^9}},
 CellTags->{
  "ID:450457679", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual] (binary)"},
 CellID->450457679],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "v", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"s", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
        RowBox[{"chainCrit", "[", 
         RowBox[{"x", ",", 
          SubscriptBox["A", "s"], ",", "A"}], "]"}]], 
       RowBox[{
        RowBox[{
         RowBox[{"deg", "[", 
          RowBox[{"lcm", "[", 
           RowBox[{"x", ",", 
            SubscriptBox["A", "s"]}], "]"}], "]"}], 
         UnderscriptBox["+", "D"], "v"}], 
        UnderscriptBox["\[LessEqual]", "D"], "u"}]}],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     RowBox[{
      RowBox[{"M", "[", 
       RowBox[{"x", ",", "A"}], "]"}], 
      UnderscriptBox["+", "D"], "v"}], 
     UnderscriptBox["\[LessEqual]", "D"], "u"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M+\[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:643714041", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613806103643453*^9, 3.613806141164186*^9}, {
  3.613806903153958*^9, 3.613806942408813*^9}},
 CellTags->{
  "ID:643714041", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M+\[LessEqual]"},
 CellID->643714041],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1721951835]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->671143984],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LessEqual]M (unary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LessEqual]M (unary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["u", "ExpressionVariable"], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["r", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["u", "ExpressionVariable"], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["r", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1545946275"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1545946275-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1545946275-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1545946275", "Proof|ID:1545946275-1"},
 CellID->1890851245],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1545946275-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.612850913276157*^9, 3.614934254048005*^9},
 CellID->735815697]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\[LessEqual] (unary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\[LessEqual] (unary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox["u", "ExpressionVariable"]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["r", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox["u", "ExpressionVariable"]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["r", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1018591902"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1018591902-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1018591902-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1018591902", "Proof|ID:1018591902-1"},
 CellID->1462209435],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1018591902-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934257960189*^9},
 CellID->209766788]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LessEqual]M (binary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LessEqual]M (binary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    StyleBox["u", "ExpressionVariable"], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    StyleBox["u", "ExpressionVariable"], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:382103250"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p382103250-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p382103250-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:382103250", "Proof|ID:382103250-1"},
 CellID->280851943],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p382103250-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.612856101751161*^9, 3.614934262167618*^9},
 CellID->1182486729]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\[LessEqual] (binary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\[LessEqual] (binary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox[
                    "u", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox[
                    "u", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:450457679"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p450457679-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p450457679-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:450457679", "Proof|ID:450457679-1"},
 CellID->478833348],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p450457679-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.613806464432735*^9, 3.614934266623499*^9},
 CellID->967557673]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M+\[LessEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M+\[LessEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:643714041"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p643714041-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p643714041-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:643714041", "Proof|ID:643714041-1"},
 CellID->105955449],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p643714041-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {50., 56.}}]], "ProofInfoBody",
 CellID->754929079]
}, Closed]],

Cell[TextData[{
 "The the following lemma please note that formula ",
 ButtonBox["(38)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1342720412"],
 " cannot be proved formally with ",
 StyleBox["Theorema ",
  FontSlant->"Italic"],
 "at the present stage, since no inference rule for proving equality of two \
abstraction tuples has been implemented (but could be added in the future). \
However, it is apparent that it follows immediately from formula ",
 ButtonBox["(chainCrit\:293a I)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1657568255"],
 " which asserts that the two conditions are in fact equivalent (and the \
remaining parts of the tuples are identical anyway)."
}], "Text",
 CellChangeTimes->{{3.614511883603783*^9, 3.614511915377674*^9}, {
  3.614512081945541*^9, 3.614512254310266*^9}, {3.61459079855561*^9, 
  3.614590811026589*^9}, {3.61509402832415*^9, 3.615094045424702*^9}, {
  3.615094081205803*^9, 3.615094120011187*^9}, {3.61509455721967*^9, 
  3.615094557699413*^9}, {3.615705830130551*^9, 3.615705831391366*^9}},
 CellID->512688545],

Cell["", "OpenEnvironment",
 CellID->791144920],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Preparations for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.611399150152572*^9, 3.611399160357111*^9}},
 CellID->1432366078],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.611834200189312*^9, 3.611834221459058*^9}, {
   3.611834341351356*^9, 3.611834352609097*^9}, {3.6144001654331*^9, 
   3.614400165654842*^9}, 3.614400263506983*^9},
 CellID->140598250],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[SquareSubsetEqual]", "D"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"M", "[", 
      RowBox[{"x", ",", "A"}], "]"}], ",", 
     RowBox[{"deg", "[", "x", "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(34)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1934562955", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611399969906967*^9, 3.611399985991983*^9}, {
  3.611834356414742*^9, 3.611834357431609*^9}, {3.614400267705203*^9, 
  3.614400268699629*^9}},
 CellTags->{
  "ID:1934562955", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "34"},
 CellID->1934562955],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["max", "D"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"M", "[", 
       RowBox[{"x", ",", "A"}], "]"}], ",", 
      RowBox[{"deg", "[", "x", "]"}]}], "\[RightAngleBracket]"}], "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(35)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1993966556", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611396487871175*^9, 3.611396496574395*^9}, {
  3.611400787757913*^9, 3.611400794883039*^9}, {3.61183435897819*^9, 
  3.611834359999977*^9}, {3.614400270137465*^9, 3.614400271353214*^9}},
 CellTags->{
  "ID:1993966556", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "35"},
 CellID->1993966556],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[SquareSubsetEqual]", "D"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"M", "[", "A", "]"}], ",", 
     RowBox[{"M", "[", 
      RowBox[{"x", ",", "A"}], "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(36)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1265284403", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611809548334079*^9, 3.611809583306326*^9}, {
   3.611834361565907*^9, 3.611834363717009*^9}, {3.614400155273377*^9, 
   3.614400162009705*^9}, 3.61440027286517*^9},
 CellTags->{
  "ID:1265284403", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "36"},
 CellID->1265284403],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[SquareSubsetEqual]", "D"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"M", "[", "A", "]"}], ",", 
     RowBox[{"M", "[", 
      RowBox[{"x", ",", "A"}], "]"}], ",", 
     RowBox[{"deg", "[", "x", "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(37)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:934408883", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611898839221769*^9, 3.611898855985822*^9}, {
  3.614400274264511*^9, 3.614400275295961*^9}},
 CellTags->{
  "ID:934408883", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "37"},
 CellID->934408883],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"deg", "[", 
     RowBox[{"lcm", "[", 
      RowBox[{"x", ",", 
       SubscriptBox["A", "i"]}], "]"}], "]"}], 
    UnderscriptBox["|", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", 
      SubscriptBox["A", "i"], ",", 
      RowBox[{"A", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "x"}]}], "]"}]}], "\[RightAngleBracket]"}], 
  "\[Equal]", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"deg", "[", 
     RowBox[{"lcm", "[", 
      RowBox[{"x", ",", 
       SubscriptBox["A", "i"]}], "]"}], "]"}], 
    UnderscriptBox["|", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", 
      SubscriptBox["A", "i"], ",", "A"}], "]"}]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(38)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1342720412", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.614491753106091*^9, 3.614491811074791*^9}, {
  3.614511846632074*^9, 3.614511850862019*^9}},
 CellTags->{
  "ID:1342720412", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "38"},
 CellID->1342720412],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExpVectorTuple", "[", "B", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.611399163315347*^9, 3.611399173364835*^9}, {
  3.6118341692390127`*^9, 3.61183417204741*^9}, {3.614400203464762*^9, 
  3.614400205563581*^9}, {3.614402673702878*^9, 3.614402683766404*^9}},
 CellID->432451001],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[SquareSubsetEqual]", "D"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"M", "[", 
      RowBox[{
       SubscriptBox["B", "i"], ",", "A"}], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(39)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:192359742", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611985154675231*^9, 3.611985180212731*^9}, {
   3.613982724793444*^9, 3.613982778000987*^9}, 3.613983604819195*^9},
 CellTags->{
  "ID:192359742", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "39"},
 CellID->192359742],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[SquareSubsetEqual]", "D"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["max", "D"], "[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"M", "[", 
         RowBox[{
          SubscriptBox["A", "j"], ",", 
          RowBox[{"A", 
           TagBox["\:293a",
            Identity,
            SyntaxForm->"a*b"], "x"}]}], "]"}], 
        UnderscriptBox["|", 
         RowBox[{
          RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}],
        "\[RightAngleBracket]"}], "]"}], ",", 
     RowBox[{"M", "[", 
      RowBox[{"x", ",", 
       RowBox[{"A", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "x"}]}], "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(40)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2005322529", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613983559736269*^9, 3.613983588322741*^9}, {
  3.614587984472508*^9, 3.61458803586976*^9}, {3.614588176270243*^9, 
  3.614588191092522*^9}},
 CellTags->{
  "ID:2005322529", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "40"},
 CellID->2005322529],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->469218948]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1792755754],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(34)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(34)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                   RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                   RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1934562955"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1934562955-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1934562955|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->721371033],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1934562955-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{298., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.611982366811307*^9, 3.614934283110276*^9},
 CellID->1356713102]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(35)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(35)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1993966556"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1993966556-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1993966556|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->899634188],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1993966556-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{298., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934317533258*^9},
 CellID->599689410]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(36)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(36)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1265284403"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1265284403-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1265284403-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1265284403", "Proof|ID:1265284403-1"},
 CellID->612613216],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1265284403-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1433050870]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(37)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(37)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["u", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                   RowBox[{"deg", "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["u", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                   RowBox[{"deg", "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:934408883"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p934408883-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p934408883-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:934408883", "Proof|ID:934408883-1"},
 CellID->716604140],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p934408883-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->2015567188]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(39)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(39)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExpVectorTuple", "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExpVectorTuple", "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:192359742"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p192359742-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p192359742-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:192359742", "Proof|ID:192359742-1"},
 CellID->89073571],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p192359742-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1022528965]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(40)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(40)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "]"}], ",", 
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "]"}], ",", 
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2005322529"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2005322529-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2005322529-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2005322529", "Proof|ID:2005322529-1"},
 CellID->592718415],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2005322529-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1252787252]
}, Closed]],

Cell[TextData[{
 "Before we can turn to the most import result of this subsection, we first \
have to prove that the degree of the least common multiple of no elements of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " for which the chain criterion holds is greater than ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 ". This is stated in formula ",
 ButtonBox["(42)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1616352426"],
 "."
}], "Text",
 CellChangeTimes->{{3.615094191316877*^9, 3.615094341469816*^9}, {
  3.615705923753176*^9, 3.61570592500617*^9}},
 CellID->1468024158],

Cell["", "OpenEnvironment",
 CellID->744151466],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "[", 
     RowBox[{"lcm", "[", "]"}], "]"}], "\[LessEqual]", "M"}], 
   TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.614397771063195*^9, 3.61439779583394*^9}},
 CellID->1514731226],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.614397819917766*^9, 3.6143978260923*^9}},
 CellID->1447892216],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.614397830085128*^9, 3.614397837156578*^9}},
 CellID->1051935310],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCrit", "[", 
   RowBox[{
    SubscriptBox["A", "i"], ",", "x", ",", "A"}], "]"}], "\[Implies]", 
  RowBox[{
   RowBox[{"deg", "[", 
    RowBox[{"lcm", "[", 
     RowBox[{
      SubscriptBox["A", "i"], ",", "x"}], "]"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"M", "[", 
    RowBox[{"x", ",", "A"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(41)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:984666222", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611896555189641*^9, 3.611896567068462*^9}, {
  3.611896939532438*^9, 3.611896956591397*^9}, {3.614397855975424*^9, 
  3.614397858260236*^9}},
 CellTags->{
  "ID:984666222", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "41"},
 CellID->984666222],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{"chainCrit", "[", 
    RowBox[{
     SubscriptBox["A", "i"], ",", 
     SubscriptBox["A", "j"], ",", "A"}], "]"}]], 
  RowBox[{
   RowBox[{"deg", "[", 
    RowBox[{"lcm", "[", 
     RowBox[{
      SubscriptBox["A", "i"], ",", 
      SubscriptBox["A", "j"]}], "]"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"M", "[", "A", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(42)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1616352426", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611896325188142*^9, 3.611896340737321*^9}, 
   3.614397853290083*^9, 3.614398689868743*^9},
 CellTags->{
  "ID:1616352426", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "42"},
 CellID->1616352426],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->850223971]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->272111712],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(41)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(41)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:984666222"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p984666222-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p984666222-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:984666222", "Proof|ID:984666222-1"},
 CellID->740590150],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p984666222-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->238534644]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(42)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(42)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                RowBox[{"chainCrit", "[", 
                  RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 RowBox[{"deg", "[", 
                   RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                RowBox[{"chainCrit", "[", 
                  RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 RowBox[{"deg", "[", 
                   RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1616352426"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1616352426-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1616352426-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1616352426", "Proof|ID:1616352426-1"},
 CellID->1148420320],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1616352426-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1897583466]
}, Closed]],

Cell[TextData[{
 "The next theorem (in particular formula ",
 ButtonBox["(M\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1628665526"],
 ") is one of the most important results in this whole section, because it \
gives an upper bound of the quantity ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], TraditionalForm]]],
 " in terms of a simpler expression without the append-operator. The other \
formulas in the lemma are basically only auxiliary results needed for proving \
formula ",
 ButtonBox["(M\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1628665526"],
 "."
}], "Text",
 CellChangeTimes->{{3.604035238984633*^9, 3.604035418585084*^9}, {
  3.614399252452192*^9, 3.61439925361651*^9}, {3.615094359818354*^9, 
  3.615094392107335*^9}, {3.615095153216816*^9, 3.615095208270611*^9}},
 CellID->624927140],

Cell["", "OpenEnvironment",
 CellID->667675291],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.598590345427241*^9, 3.598590357376431*^9}, {
  3.600663410057905*^9, 3.60066341154616*^9}, {3.614397643716404*^9, 
  3.614397644802114*^9}},
 CellID->1051433415],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.600665442733193*^9, 3.600665449633493*^9}, {
  3.605593754281345*^9, 3.605593757918864*^9}},
 CellID->962205404],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"M", "[", 
     RowBox[{
      SubscriptBox["A", "j"], ",", 
      RowBox[{"A", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "x"}]}], "]"}], 
    UnderscriptBox["\[LessEqual]", "D"], 
    RowBox[{
     UnderscriptBox["max", "D"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{"M", "[", "A", "]"}], ",", 
       RowBox[{"M", "[", 
        RowBox[{"x", ",", "A"}], "]"}]}], "\[RightAngleBracket]"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\:293a 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1399613937", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613985270915189*^9, 3.613985326940014*^9}, {
  3.613985607751859*^9, 3.613985610662832*^9}, {3.613985763106558*^9, 
  3.613985777163497*^9}, {3.613985899826938*^9, 3.61398592070923*^9}, {
  3.613986130101821*^9, 3.613986149674512*^9}, {3.613986194283795*^9, 
  3.613986195720667*^9}, {3.613986330271553*^9, 3.613986333043106*^9}},
 CellTags->{
  "ID:1399613937", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 1"},
 CellID->1399613937],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"M", "[", 
      RowBox[{
       SubscriptBox["A", "j"], ",", 
       RowBox[{"A", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "x"}]}], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"M", "[", "A", "]"}], ",", 
     RowBox[{"M", "[", 
      RowBox[{"x", ",", "A"}], "]"}], ",", 
     RowBox[{"deg", "[", "x", "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\:293a 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1210171526", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.614402306006074*^9, 3.614402361490506*^9}},
 CellTags->{
  "ID:1210171526", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 2"},
 CellID->1210171526],

Cell[BoxData[
 RowBox[{
  RowBox[{"M", "[", 
   RowBox[{"x", ",", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}]}], "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"M", "[", 
      RowBox[{"x", ",", "A"}], "]"}], ",", 
     RowBox[{"deg", "[", "x", "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\:293a 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1251355025", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.614490933834486*^9},
 CellTags->{
  "ID:1251355025", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 3"},
 CellID->1251355025],

Cell[BoxData[
 RowBox[{
  RowBox[{"M", "[", 
   RowBox[{"x", ",", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}]}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"M", "[", "A", "]"}], ",", 
     RowBox[{"M", "[", 
      RowBox[{"x", ",", "A"}], "]"}], ",", 
     RowBox[{"deg", "[", "x", "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\:293a 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1093640190", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611898432978738*^9, 3.611898448418264*^9}, 
   3.611898575175533*^9},
 CellTags->{
  "ID:1093640190", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 4"},
 CellID->1093640190],

Cell[BoxData[
 RowBox[{
  RowBox[{"M", "[", 
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["max", "D"], "[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"M", "[", 
         RowBox[{
          SubscriptBox["A", "j"], ",", 
          RowBox[{"A", 
           TagBox["\:293a",
            Identity,
            SyntaxForm->"a*b"], "x"}]}], "]"}], 
        UnderscriptBox["|", 
         RowBox[{
          RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}],
        "\[RightAngleBracket]"}], "]"}], ",", 
     RowBox[{"M", "[", 
      RowBox[{"x", ",", 
       RowBox[{"A", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "x"}]}], "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\:293a 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:606212147", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61458731189311*^9, 3.614587350659333*^9}},
 CellTags->{
  "ID:606212147", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 5"},
 CellID->606212147],

Cell[BoxData[
 RowBox[{
  RowBox[{"M", "[", 
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"M", "[", "A", "]"}], ",", 
     RowBox[{"M", "[", 
      RowBox[{"x", ",", "A"}], "]"}], ",", 
     RowBox[{"deg", "[", "x", "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\:293a)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1628665526", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmViYGAQBWIQXRp1xuTI+VeOPSWvrEC06KlH9iDadYKfM4heyxLq
DaIzvMrA9Dc3qVAQHVRsHQOio3PnJ4Hozao6uSC6VulHEYhe9Y6lBGye/AeG
Y0B670dDJhD9/7Zt6A8QrZ8SDqKzP9Q9Fr34ynGDW8xTEP3gu/06JyBtZfBw
PYhe5Jy21x1Ipx9eAqZ3KMyVsL/0yvFEfLkUiL6jECgHoo9HzATTPSdlVEF0
x6mZYDpv+l7mKUD60hYeFhA9/fi0/B2XXzkmxnEVgGjeLleenUC68Y8ML4h+
5bgxF0Q/7rwApsWcVWeAaFOXbjC96WbG/l+3Xzl+vLYITPP2zbkqf/eVo4TB
lmsgGgD3fbzj
  "],
 CellTags->{
  "ID:1628665526", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a"},
 CellID->1628665526],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1798898406]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->39175054],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\:293a 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\:293a 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{"M", "[", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   UnderscriptBox["max", "D"], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                    "\[RightAngleBracket]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{"M", "[", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   UnderscriptBox["max", "D"], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                    "\[RightAngleBracket]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1399613937"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1399613937-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1399613937-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1399613937", "Proof|ID:1399613937-1"},
 CellID->1391233604],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1399613937-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellID->713233342]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\:293a 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\:293a 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                   "\[RightAngleBracket]"}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                   "\[RightAngleBracket]"}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1210171526"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1210171526-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1210171526-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1210171526", "Proof|ID:1210171526-1"},
 CellID->2096401104],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1210171526-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellID->848867484]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\:293a 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\:293a 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"M", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
               "\[Equal]", 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"M", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
               "\[Equal]", 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1251355025"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1251355025-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1251355025-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1251355025", "Proof|ID:1251355025-1"},
 CellID->185930261],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1251355025-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {50., 56.}}]], "ProofInfoBody",
 CellID->2087850823]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\:293a 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\:293a 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"M", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"M", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1093640190"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1093640190-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1093640190-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1093640190", "Proof|ID:1093640190-1"},
 CellID->808217224],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1093640190-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->348117997]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\:293a 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\:293a 5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"M", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"M", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}]}], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:606212147"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p606212147-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p606212147-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:606212147", "Proof|ID:606212147-1"},
 CellID->1011788619],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p606212147-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->288491612]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\:293a)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\:293a)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"M", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"M", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1628665526"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1628665526-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1628665526-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1628665526", "Proof|ID:1628665526-1"},
 CellID->975115047],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1628665526-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1079430099]
}, Closed]],

Cell[TextData[{
 "In fact, in the rest of this report and in particular for proving the main \
theorems, rather than ",
 ButtonBox["(M\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1628665526"],
 " we will need some immediate consequences of it, namely ",
 ButtonBox["(M\:293a\[LessEqual])",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2057269766"],
 " and ",
 ButtonBox["(M\:293a\[LessEqual])",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1065272829"],
 "."
}], "Text",
 CellChangeTimes->{{3.615095219561342*^9, 3.615095336491208*^9}},
 CellID->327532369],

Cell["", "OpenEnvironment",
 CellID->1255006385],

Cell[CellGroupData[{

Cell[TextData[{
 "Corollary: Consequences of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.613808929455596*^9, 3.61380894344347*^9}},
 CellID->542844259],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.613808983047625*^9, 3.613808988622415*^9}},
 CellID->1379329847],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"u", ",", "v"}], "]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.613808993510719*^9, 3.613809020200205*^9}},
 CellID->1704783354],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["max", "D"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"M", "[", "A", "]"}], ",", 
      RowBox[{"M", "[", 
       RowBox[{"x", ",", "A"}], "]"}], ",", 
      RowBox[{"deg", "[", "x", "]"}]}], "\[RightAngleBracket]"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}], "\[Implies]", 
  RowBox[{
   RowBox[{"M", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\:293a I)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1351181837", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613193458349521*^9, 3.613193523784312*^9}, 
   3.613193846451017*^9, 3.61319389064761*^9, {3.613193936969411*^9, 
   3.613193947903361*^9}, {3.613809004987857*^9, 3.61380900605605*^9}},
 CellTags->{
  "ID:1351181837", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a I"},
 CellID->1351181837],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"M", "[", "A", "]"}], 
     UnderscriptBox["\[LessEqual]", "D"], "u"}], "\[And]", 
    RowBox[{
     RowBox[{"M", "[", 
      RowBox[{"x", ",", "A"}], "]"}], 
     UnderscriptBox["\[LessEqual]", "D"], "u"}], "\[And]", 
    RowBox[{
     RowBox[{"deg", "[", "x", "]"}], 
     UnderscriptBox["\[LessEqual]", "D"], "u"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"M", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\:293a\[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2057269766", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613194062496612*^9, 3.613194092037152*^9}, {
  3.61380900872732*^9, 3.613809009907186*^9}},
 CellTags->{
  "ID:2057269766", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a\[LessEqual]"},
 CellID->2057269766],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     UnderscriptBox["max", "D"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{"M", "[", "A", "]"}], ",", 
       RowBox[{"M", "[", 
        RowBox[{"x", ",", "A"}], "]"}], ",", 
       RowBox[{"deg", "[", "x", "]"}]}], "\[RightAngleBracket]"}], "]"}], 
    UnderscriptBox["+", "D"], "v"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}], "\[Implies]", 
  RowBox[{
   RowBox[{
    RowBox[{"M", "[", 
     RowBox[{"A", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "x"}], "]"}], 
    UnderscriptBox["+", "D"], "v"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\:293a II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:94917498", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613808648360169*^9, 3.613808673081212*^9}, {
  3.61380901377617*^9, 3.613809072533361*^9}},
 CellTags->{
  "ID:94917498", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a II"},
 CellID->94917498],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"M", "[", "A", "]"}], 
      UnderscriptBox["+", "D"], "v"}], 
     UnderscriptBox["\[LessEqual]", "D"], "u"}], "\[And]", 
    RowBox[{
     RowBox[{
      RowBox[{"M", "[", 
       RowBox[{"x", ",", "A"}], "]"}], 
      UnderscriptBox["+", "D"], "v"}], 
     UnderscriptBox["\[LessEqual]", "D"], "u"}], "\[And]", 
    RowBox[{
     RowBox[{
      RowBox[{"deg", "[", "x", "]"}], 
      UnderscriptBox["+", "D"], "v"}], 
     UnderscriptBox["\[LessEqual]", "D"], "u"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    RowBox[{"M", "[", 
     RowBox[{"A", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "x"}], "]"}], 
    UnderscriptBox["+", "D"], "v"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\:293a+\[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1065272829", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.61380905290215*^9},
 CellTags->{
  "ID:1065272829", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a+\[LessEqual]"},
 CellID->1065272829],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->659421005]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2130518472],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\:293a I)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\:293a I)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                    "\[RightAngleBracket]"}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                    "\[RightAngleBracket]"}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1351181837"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1351181837-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1351181837-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1351181837", "Proof|ID:1351181837-1"},
 CellID->1829706754],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1351181837-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.613809144156783*^9},
 CellID->468214080]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\:293a\[LessEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\:293a\[LessEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox["u", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox["u", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox["u", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox["u", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2057269766"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2057269766-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2057269766-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2057269766", "Proof|ID:2057269766-1"},
 CellID->706843964],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2057269766-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.613808778567021*^9, 3.613809726136895*^9},
 CellID->951036626]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\:293a II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\:293a II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                    "\[RightAngleBracket]"}], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                    "\[RightAngleBracket]"}], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:94917498"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p94917498-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p94917498-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:94917498", "Proof|ID:94917498-1"},
 CellID->1107419424],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p94917498-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->196846389]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\:293a+\[LessEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\:293a+\[LessEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "D"], 
                   StyleBox["v", "ExpressionVariable"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox[
                 "u", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "D"], 
                   StyleBox["v", "ExpressionVariable"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox[
                 "u", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1065272829"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1065272829-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1065272829-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1065272829", "Proof|ID:1065272829-1"},
 CellID->1792160777],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1065272829-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->994864864]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox["W", TraditionalForm]]]], "Subsubsection",
 CellChangeTimes->{{3.611986907818323*^9, 3.611986909562036*^9}, {
   3.612082283586991*^9, 3.612082284104392*^9}, 3.612086924300553*^9},
 CellID->362221155],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.600669704002672*^9, 3.600669710800482*^9}, {
  3.603178007957734*^9, 3.603178010202393*^9}, {3.603693164679785*^9, 
  3.603693166840385*^9}},
 CellID->2013493621],

Cell[TextData[{
 "We are now done with ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 " and turn to ",
 Cell[BoxData[
  FormBox["W", TraditionalForm]]],
 " by first stating a lemma which is completely obvious, but at the same time \
quite useful for proving formula ",
 ButtonBox["(100)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:724101256"],
 ":"
}], "Text",
 CellChangeTimes->{{3.60111140079956*^9, 3.601111434565702*^9}, {
  3.604035702573645*^9, 3.60403578107437*^9}, {3.615095527416707*^9, 
  3.615095538443984*^9}, {3.615787771259991*^9, 3.615787771518902*^9}},
 CellID->231154263],

Cell["", "OpenEnvironment",
 CellID->593345801],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["W", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.601111319195435*^9, 3.601111325187767*^9}, {
  3.603693630909316*^9, 3.60369364223314*^9}},
 CellID->1877952819],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "n"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.603693764106042*^9, 3.603693770050994*^9}},
 CellID->1922840150],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "k"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of W (binary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1824645721", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601025468989107*^9, 3.601025486025434*^9}, 
   3.603693773297586*^9, {3.605593815823108*^9, 3.605593816115767*^9}},
 CellTags->{
  "ID:1824645721", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of W (binary)"},
 CellID->1824645721],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{"W", "[", "A", "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of W (unary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1345307293", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601025495704209*^9, 3.601025517761702*^9}, {
  3.605593817792844*^9, 3.605593817978217*^9}},
 CellTags->{
  "ID:1345307293", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of W (unary)"},
 CellID->1345307293],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "k"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   SubscriptBox[
    RowBox[{"(", 
     SubscriptBox["A", "i"], ")"}], "k"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(43)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1185705395", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601111337152073*^9, 3.60111138906261*^9}, 
   3.603693775376122*^9, {3.605593819675214*^9, 3.605593820007525*^9}, {
   3.611382017887186*^9, 3.611382024032292*^9}},
 CellTags->{
  "ID:1185705395", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "43"},
 CellID->1185705395],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1776977096]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2030887648],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of W (binary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of W (binary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{"W", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{"W", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1824645721"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1824645721-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1824645721-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1824645721", "Proof|ID:1824645721-1"},
 CellID->286621138],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1824645721-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1193162250]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of W (unary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of W (unary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{"W", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{"W", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1345307293"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1345307293-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1345307293-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1345307293", "Proof|ID:1345307293-1"},
 CellID->28034114],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1345307293-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1750002018]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(43)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(43)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1185705395"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1185705395-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1185705395-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1185705395", "Proof|ID:1185705395-1"},
 CellID->2067392509],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1185705395-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellID->508522961]
}, Closed]],

Cell[TextData[{
 "Unfortunately, formula ",
 ButtonBox["(type of W (unary))",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1345307293"],
 " cannot be proved with the \[OpenCurlyDoubleQuote]ComplexityProver\
\[CloseCurlyDoubleQuote] for arbitrary ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ", for the same reasons why formula ",
 ButtonBox["(type of deg)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1822207066"],
 " cannot be proved. Therefore, we only prove it if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 ", i.e. with formula ",
 ButtonBox["(n=2)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1526008058"],
 " among our assumptions."
}], "Text",
 CellChangeTimes->{{3.603694351747491*^9, 3.603694365499796*^9}, {
   3.603694400238164*^9, 3.603694463880298*^9}, {3.603792071750582*^9, 
   3.603792072084231*^9}, {3.604035458212397*^9, 3.604035460635172*^9}, {
   3.604035797728349*^9, 3.604035811126106*^9}, 3.605593849388855*^9, {
   3.605597814649339*^9, 3.605597854094041*^9}, {3.612082779628796*^9, 
   3.612082885985329*^9}, {3.615095596138771*^9, 3.615095619939392*^9}, {
   3.615095674456064*^9, 3.615095742210941*^9}},
 CellID->2017434267],

Cell[TextData[{
 "The next lemma is analogous to lemma \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "\[LessEqual]", "u"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", "\[LessEqual]", "M"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] in the previous subsection:"
}], "Text",
 CellChangeTimes->{{3.615095788027053*^9, 3.615095835675157*^9}, 
   3.615706665045584*^9},
 CellID->1656499058],

Cell["", "OpenEnvironment",
 CellID->1706918605],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "\[LessEqual]", "u"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", "\[LessEqual]", "W"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.612842619796348*^9, 3.612842630238352*^9}},
 CellID->938515413],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "n"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.61284265784251*^9, 3.612842662287327*^9}},
 CellID->1091363795],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", "u", "]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612842665302489*^9, 3.612842673977023*^9}},
 CellID->110581283],

Cell[BoxData[
 RowBox[{
  RowBox[{"u", 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "k"}], "]"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"r", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{"u", 
      UnderscriptBox["\[LessEqual]", "D"], 
      SubscriptBox[
       RowBox[{"(", 
        SubscriptBox["A", "r"], ")"}], "k"]}]}],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LessEqual]W)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:354042430", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600589526269121*^9, 3.600589585141293*^9}, 
   3.600688553231018*^9, {3.60403736225809*^9, 3.604037362779233*^9}, 
   3.604037422933242*^9, {3.612842723536942*^9, 3.612842772239634*^9}},
 CellTags->{
  "ID:354042430", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]W"},
 CellID->354042430],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "k"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"r", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["A", "r"], ")"}], "k"], 
     UnderscriptBox["\[LessEqual]", "D"], "u"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(W\[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:551974343", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600589602848213*^9, 3.600589608006454*^9}, 
   3.60068855477745*^9, {3.6040373652995*^9, 3.604037365821325*^9}, {
   3.604037435283146*^9, 3.604037440643015*^9}, {3.612842777070752*^9, 
   3.61284281198694*^9}},
 CellTags->{
  "ID:551974343", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\[LessEqual]"},
 CellID->551974343],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->135917884]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1118462943],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LessEqual]W)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LessEqual]W)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    StyleBox["u", "ExpressionVariable"], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["r", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    StyleBox["u", "ExpressionVariable"], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["r", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:354042430"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p354042430-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p354042430-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:354042430", "Proof|ID:354042430-1"},
 CellID->1517092650],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p354042430-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934352194862*^9},
 CellID->624328162]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(W\[LessEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(W\[LessEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["r", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox[
                    "u", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "D"], "[", 
                    StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["r", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox[
                    "u", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:551974343"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p551974343-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p551974343-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:551974343", "Proof|ID:551974343-1"},
 CellID->766317937],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p551974343-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934356899044*^9},
 CellID->604160603]
}, Closed]],

Cell[TextData[{
 "As with ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 ", the main thing we are interested in is how ",
 Cell[BoxData[
  FormBox["W", TraditionalForm]]],
 " changes if some new element is appended to the original tuple. \
Unfortunately, in general (i.e. for arbitrary ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ") we only know very little about this, namely how ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", 
    RowBox[{
     RowBox[{"A", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "x"}], ",", "k"}], "]"}], TraditionalForm]]],
 " can be expressed in terms of a minimum. This is for sure completely \
obvious, but nonetheless quite useful for proving lemma ",
 ButtonBox["W[A\:293ax]",
  BaseStyle->"Hyperlink",
  ButtonData->"Lem::W[A<-x]"],
 " in section ",
 ButtonBox["Lemmata on Specific Notions / Special Case n=2",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataSpecificSpecial"],
 " below."
}], "Text",
 CellChangeTimes->{{3.604035836172386*^9, 3.604035995550676*^9}, {
  3.604036038970223*^9, 3.604036167537348*^9}, {3.609040226499744*^9, 
  3.609040228450418*^9}, {3.615096039275944*^9, 3.615096084447372*^9}},
 CellID->1123187194],

Cell["", "OpenEnvironment",
 CellID->622337541],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Preparations for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", 
    RowBox[{
     RowBox[{"A", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "x"}], ",", "k"}], "]"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.611989101509112*^9, 3.611989115194239*^9}},
 CellID->416872943],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "n"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.611989191633725*^9, 3.611989195669033*^9}},
 CellID->1770721675],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "y", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.611994328880067*^9, 3.611994367237766*^9}},
 CellID->887909859],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[SquareSubsetEqual]", "D"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"W", "[", 
      RowBox[{"A", ",", "k"}], "]"}], ",", 
     SubscriptBox["y", "k"]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(44)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:894623443", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611994352168352*^9, 3.611994369820624*^9}},
 CellTags->{
  "ID:894623443", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "44"},
 CellID->894623443],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[SquareSubsetEqual]", "D"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["A", "i"], ")"}], "k"], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(45)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1995654247", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61198914302419*^9, 3.611989210792842*^9}},
 CellTags->{
  "ID:1995654247", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "45"},
 CellID->1995654247],

Cell[BoxData[
 RowBox[{"1", 
  UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
  RowBox[{"\[LeftBracketingBar]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["A", "i"], ")"}], "k"], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], 
   "\[RightBracketingBar]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(46)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1916087968", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611989430598586*^9, 3.611989446744533*^9}, {
  3.611995655987835*^9, 3.611995657175353*^9}},
 CellTags->{
  "ID:1916087968", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "46"},
 CellID->1916087968],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1446445473]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->523519669],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(44)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(44)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                   TagBox[")", "AutoParentheses"]}], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                   TagBox[")", "AutoParentheses"]}], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:894623443"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p894623443-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p894623443-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:894623443", "Proof|ID:894623443-1"},
 CellID->456641578],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p894623443-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934362971043*^9},
 CellID->314453048]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(45)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(45)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1995654247"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1995654247-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1995654247-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1995654247", "Proof|ID:1995654247-1"},
 CellID->1030159721],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1995654247-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934367626059*^9},
 CellID->645849534]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(46)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(46)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{"1", 
               UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "n"}]}}]], 
             RowBox[{"1", 
               UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1916087968"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1916087968-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1916087968-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1916087968", "Proof|ID:1916087968-1"},
 CellID->1865547860],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1916087968-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934372050758*^9},
 CellID->1576636564]
}, Closed]],

Cell[TextData[{
 "In the following lemma, the most interesting formulae are ",
 ButtonBox["(W\:293a)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2064198302"],
 " and ",
 ButtonBox["(W\:293a decreases)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:144808276"],
 ". Note that, for the same reasons as for ",
 ButtonBox["(type of deg)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1822207066"],
 " and ",
 ButtonBox["(type of W (unary))",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1345307293"],
 ", also ",
 ButtonBox["(W\:293a decreases)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:144808276"],
 " cannot be proved for arbitrary ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " but only for fixed ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " (in this case ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 "), although it certainly holds for all natural numbers ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.61208269903479*^9, 3.612082758903938*^9}, {
   3.61208289638675*^9, 3.612083029977774*^9}, {3.613963729926941*^9, 
   3.613963734198717*^9}, 3.613980863641772*^9, {3.615096408365362*^9, 
   3.615096559128243*^9}},
 CellID->1668358382],

Cell["", "OpenEnvironment",
 CellID->1003278763],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", 
    RowBox[{
     RowBox[{"A", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "x"}], ",", "k"}], "]"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.598593044080209*^9, 3.59859305753018*^9}, {
  3.598593301179483*^9, 3.598593301390831*^9}, {3.599384013845329*^9, 
  3.599384013845421*^9}, {3.600663439415896*^9, 3.600663441407262*^9}, {
  3.600687418400543*^9, 3.600687418572485*^9}},
 CellID->966932569],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.598593093620895*^9, 3.598593100252844*^9}, {
  3.605593856485326*^9, 3.605593860102635*^9}},
 CellID->930876469],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "n"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.611386534038763*^9, 3.611386538157381*^9}},
 CellID->1247889975],

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "[", 
   RowBox[{
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], ",", "k"}], "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["min", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"W", "[", 
      RowBox[{"A", ",", "k"}], "]"}], ",", 
     SubscriptBox["x", "k"]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(W\:293a)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2064198302", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.598593117086516*^9, 3.598593216032782*^9}, {
   3.598594047092754*^9, 3.59859406754681*^9}, {3.598697140374051*^9, 
   3.598697141319739*^9}, 3.599282572014103*^9, 3.599384021945888*^9, {
   3.599890658851067*^9, 3.599890663166188*^9}, {3.600064552046279*^9, 
   3.60006455459549*^9}, {3.60041318901093*^9, 3.6004131907754803`*^9}, 
   3.600663430865221*^9, {3.604993544152984*^9, 3.604993545965013*^9}, {
   3.605593863338554*^9, 3.605593865147467*^9}, {3.611386540707731*^9, 
   3.611386542813251*^9}},
 CellTags->{
  "ID:2064198302", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a"},
 CellID->2064198302],

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "[", 
   RowBox[{
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], ",", "k"}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{"W", "[", 
   RowBox[{"A", ",", "k"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(W\:293a I)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2144277606", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.611986721159836*^9},
 CellTags->{
  "ID:2144277606", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a I"},
 CellID->2144277606],

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "[", 
   RowBox[{
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], ",", "k"}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  SubscriptBox["x", "k"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(W\:293a II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:388422476", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.611387356358292*^9, 3.611387376763782*^9}, {
  3.61138763763378*^9, 3.611387638374151*^9}, {3.611986719720509*^9, 
  3.611986722416396*^9}},
 CellTags->{
  "ID:388422476", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a II"},
 CellID->388422476],

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "[", 
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{"W", "[", "A", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(W\:293a decreases)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:144808276", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599383928275332*^9, 3.599383962781804*^9}, 
   3.599384007620543*^9, 3.60066343234006*^9, {3.605597906114887*^9, 
   3.605597907445204*^9}},
 CellTags->{
  "ID:144808276", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a decreases"},
 CellID->144808276],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->572158620]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->944992805],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(W\:293a)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(W\:293a)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "n"}]}}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   UnderscriptBox["min", "D"], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "n"}]}}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   UnderscriptBox["min", "D"], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2064198302"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2064198302-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2064198302-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2064198302", "Proof|ID:2064198302-1"},
 CellID->595904185],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2064198302-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->472964561]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(W\:293a I)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(W\:293a I)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "n"}]}}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "n"}]}}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2144277606"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2144277606-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2144277606-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2144277606", "Proof|ID:2144277606-1"},
 CellID->391021459],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2144277606-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1450565738]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(W\:293a II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(W\:293a II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "n"}]}}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "n"}]}}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:388422476"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p388422476-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p388422476-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:388422476", "Proof|ID:388422476-1"},
 CellID->743656091],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p388422476-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1071526442]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(W\:293a decreases)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(W\:293a decreases)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"W", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{"W", "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"W", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{"W", "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:144808276"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p144808276-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p144808276-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:144808276", "Proof|ID:144808276-1"},
 CellID->1758271591],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p144808276-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->672863167]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox["L", TraditionalForm]]]], "Subsubsection",
 CellChangeTimes->{{3.611989005569711*^9, 3.611989006325164*^9}, {
   3.61199466269424*^9, 3.611994663018992*^9}, 3.612086925664602*^9},
 CellID->138333285],

Cell[TextData[{
 "We finally conclude this subsection by stating what the \
\[OpenCurlyDoubleQuote]type\[CloseCurlyDoubleQuote] of ",
 Cell[BoxData[
  FormBox["L", TraditionalForm]]],
 " is:"
}], "Text",
 CellChangeTimes->{{3.604036184486121*^9, 3.604036246399575*^9}, {
  3.615096663994958*^9, 3.615096670935243*^9}},
 CellID->1953748933],

Cell["", "OpenEnvironment",
 CellID->474703228],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Type of ",
 Cell[BoxData[
  FormBox["L", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.60369381633023*^9, 3.603693823207377*^9}},
 CellID->925860718],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  "\[Implies]", 
  RowBox[{"isExponentVector", "[", 
   RowBox[{"L", "[", "A", "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of L)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1149216477", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601025524160443*^9, 3.601025528293469*^9}, {
  3.605593870385807*^9, 3.605593875916284*^9}, {3.611989010813419*^9, 
  3.611989023727773*^9}},
 CellTags->{
  "ID:1149216477", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of L"},
 CellID->1149216477],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->990555961]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->880453912],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of L)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of L)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{"isExponentVector", "[", 
             RowBox[{"L", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{"isExponentVector", "[", 
             RowBox[{"L", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1149216477"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 ":   ",
 Cell[BoxData[
  ButtonBox["\<\"Show proof\"\>",
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1149216477-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[]]]]]}, {None, None}},
 CellTags->
  "Proof|ID:1149216477|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->187961443],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    Get[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1149216477-display.m\"}]"]], None], 
   StandardForm],
  ImageSizeCache->{674., {24., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6146775725503*^9},
 CellID->1995973371]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Special Case ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]]
}], "Subsection",
 CellChangeTimes->{{3.59825635832593*^9, 3.598256374241795*^9}},
 CellTags->"sect::LemmataSpecificSpecial",
 CellID->1466386606],

Cell[TextData[{
 "Some properties of the functions and predicates introduced in definition ",
 ButtonBox["Specific Notions for Complexity Analysis",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::SpecificNotions"],
 " only hold for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 " (i.e. in the bivariate case). This of course does not come as a surprise, \
for otherwise we would not have to restrict ourselves to the bivariate case \
in the complexity analysis.\n\nLet in this subsection ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " be an arbitrary tuple of two-dimensional exponent vectors, and always \
assume in this subsection that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.604036347860602*^9, 3.604036474466333*^9}, {
  3.604036509148521*^9, 3.604036535638315*^9}, {3.604036571120963*^9, 
  3.604036576562706*^9}, {3.604036636453291*^9, 3.604036752713609*^9}, {
  3.605593918835398*^9, 3.605593981262867*^9}, {3.605594043309577*^9, 
  3.605594064202632*^9}, {3.605597397080306*^9, 3.605597397265716*^9}, {
  3.615096688874368*^9, 3.615096732463669*^9}},
 CellID->765254624],

Cell[BoxData[
 RowBox[{
  RowBox[{"n", "\[Equal]", "2"}], "\[Implies]"}]], "GlobalDeclaration",
 ShowCellTags->False,
 CellChangeTimes->{{3.605597420531152*^9, 3.60559742757369*^9}},
 CellTags->{
  "ID:504163160", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->504163160],

Cell["", "Text",
 CellID->853094610],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isExpVectorTuple", "[", "A", "]"}]], 
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 ShowCellTags->False,
 CellChangeTimes->{{3.598270600205748*^9, 3.598270615950374*^9}, 
   3.598329737810229*^9, {3.600063928375089*^9, 3.600063950485659*^9}, {
   3.605593989963293*^9, 3.605593997379387*^9}, {3.612499970523137*^9, 
   3.612499981746875*^9}},
 CellTags->{
  "ID:649536122", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->649536122],

Cell[CellGroupData[{

Cell["Cover of Space of Exponent Vectors", "Subsubsection",
 CellChangeTimes->{{3.599890065113158*^9, 3.599890073669934*^9}, 
   3.604038906576945*^9, {3.604038940081265*^9, 3.604038940907026*^9}, {
   3.605594028217259*^9, 3.605594033249987*^9}, {3.61250476147172*^9, 
   3.61250476224741*^9}, 3.612766278488132*^9},
 CellTags->"sect::Cover",
 CellID->676471010],

Cell[TextData[{
 "The following lemma states that the four ",
 StyleBox["cover predicates",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox["isAbove", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["inRectangle", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["inFarExterior", TraditionalForm]]],
 " cover the whole space of two-dimensional exponent vectors, i.e. for every \
two-dimensional exponent vector at least one such predicate holds (if ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " is not empty).\nNote that the four sets really only cover the whole space, \
but in general they are ",
 StyleBox["not",
  FontSlant->"Italic"],
 " disjoint (in particular ",
 Cell[BoxData[
  FormBox["isAbove", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["inRectangle", TraditionalForm]]],
 ")!"
}], "Text",
 CellChangeTimes->{{3.604037556133724*^9, 3.604037627837926*^9}, {
  3.604037662831336*^9, 3.604037750732977*^9}, {3.604037800004974*^9, 
  3.604037811306887*^9}, {3.604038789368493*^9, 3.604038888389874*^9}, {
  3.604038943454295*^9, 3.604038973711825*^9}, {3.604039157632706*^9, 
  3.604039161900186*^9}, {3.60483575613036*^9, 3.604835790520428*^9}, {
  3.605594010415785*^9, 3.605594023649514*^9}, {3.605594082482572*^9, 
  3.605594088590249*^9}, {3.612497489013008*^9, 3.61249749269533*^9}},
 CellID->466566289],

Cell["", "OpenEnvironment",
 CellID->813984505],

Cell[CellGroupData[{

Cell["Lemma: Preparations for Cover", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.612086051165697*^9, 3.612086056200504*^9}},
 CellID->294236068],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612086067192006*^9, 3.612086074107478*^9}},
 CellID->432121352],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "2"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612244373547252*^9, 3.612244378219429*^9}},
 CellID->2038539178],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    SubscriptBox["x", "k"], "]"}]}], "\[And]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    SubscriptBox["x", 
     SuperscriptBox["k", "-"]], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(47)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:465693853", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61208607676345*^9, 3.612086082821167*^9}, {
  3.612244380982805*^9, 3.612244414733122*^9}},
 CellTags->{
  "ID:465693853", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "47"},
 CellID->465693853],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[Not]", 
     RowBox[{
      SubscriptBox["x", "k"], 
      UnderscriptBox["<", "D"], 
      RowBox[{"W", "[", 
       RowBox[{"A", ",", "k"}], "]"}]}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Not]", 
      RowBox[{
       SubscriptBox["x", 
        SuperscriptBox["k", "-"]], 
       UnderscriptBox["\[LessEqual]", "D"], 
       SubscriptBox[
        RowBox[{"(", 
         SubscriptBox["A", "j"], ")"}], 
        SuperscriptBox["k", "-"]]}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       SubscriptBox[
        RowBox[{"(", 
         SubscriptBox["A", "j"], ")"}], "k"], 
       UnderscriptBox["\[LessEqual]", "D"], 
       SubscriptBox["x", "k"]}], "\[And]", 
      RowBox[{
       SubscriptBox[
        RowBox[{"(", 
         SubscriptBox["A", "j"], ")"}], 
        SuperscriptBox["k", "-"]], 
       UnderscriptBox["\[LessEqual]", "D"], 
       SubscriptBox["x", 
        SuperscriptBox["k", "-"]]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(48)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:777205621", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612085854183242*^9, 3.612086022621791*^9}, {
  3.612090030525613*^9, 3.612090033421242*^9}, {3.612244429401541*^9, 
  3.612244457232687*^9}},
 CellTags->{
  "ID:777205621", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "48"},
 CellID->777205621],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[Not]", 
     RowBox[{
      SubscriptBox["x", "k"], 
      UnderscriptBox["<", "D"], 
      RowBox[{"W", "[", 
       RowBox[{"A", ",", "k"}], "]"}]}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Not]", 
      RowBox[{
       SubscriptBox["x", 
        SuperscriptBox["k", "-"]], 
       UnderscriptBox["\[LessEqual]", "D"], 
       SubscriptBox[
        RowBox[{"(", 
         SubscriptBox["A", "j"], ")"}], 
        SuperscriptBox["k", "-"]]}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"isAbove", "[", 
   RowBox[{"x", ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(49)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:42909295", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612086131517634*^9, 3.612086138484388*^9}, {
  3.612090035974819*^9, 3.612090039019638*^9}, {3.612246280683835*^9, 
  3.612246288902186*^9}, {3.612252799582649*^9, 3.612252807513144*^9}, {
  3.612252859062991*^9, 3.612252863657443*^9}},
 CellTags->{
  "ID:42909295", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "49"},
 CellID->42909295],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
      RowBox[{"\[Not]", 
       RowBox[{
        SubscriptBox[
         RowBox[{"(", 
          SubscriptBox["A", "j"], ")"}], "k"], 
        UnderscriptBox["<", "D"], 
        SubscriptBox["x", "k"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
      RowBox[{"\[Not]", 
       RowBox[{
        SubscriptBox["x", "k"], 
        UnderscriptBox["\[LessEqual]", "D"], 
        SubscriptBox[
         RowBox[{"(", 
          SubscriptBox["A", "j"], ")"}], "k"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"isAbove", "[", 
   RowBox[{"x", ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(50)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:112873953", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61225369060599*^9, 3.612253796031857*^9}},
 CellTags->{
  "ID:112873953", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "50"},
 CellID->112873953],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->505349263]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1249260285],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(47)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(47)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "2"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "D"], "[", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", "D"], "[", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                   TagBox[")", "AutoParentheses"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "2"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "D"], "[", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", "D"], "[", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                   TagBox[")", "AutoParentheses"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:465693853"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p465693853-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p465693853-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:465693853", "Proof|ID:465693853-1"},
 CellID->2114883905],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p465693853-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934395161031*^9},
 CellID->1227222626]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(48)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(48)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"1", 
                   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
                   
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    TagBox[")", "AutoParentheses"]}]}]}]}], "\[Implies]", 
                   RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"1", 
                   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
                   
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    TagBox[")", "AutoParentheses"]}]}]}]}], "\[Implies]", 
                   RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:777205621"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p777205621-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p777205621-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:777205621", "Proof|ID:777205621-1"},
 CellID->118523572],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p777205621-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->1502988398]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(49)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(49)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"1", 
                   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
                   
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    TagBox[")", "AutoParentheses"]}]}]}]}], "\[Implies]", 
                   RowBox[{"isAbove", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"1", 
                   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
                   
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    TagBox[")", "AutoParentheses"]}]}]}]}], "\[Implies]", 
                   RowBox[{"isAbove", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:42909295"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p42909295-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p42909295-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:42909295", "Proof|ID:42909295-1"},
 CellID->234727221],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p42909295-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934406704375*^9, 3.614935628084181*^9},
 CellID->849552024]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(50)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(50)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"1", 
                   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
                   
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    TagBox[")", "AutoParentheses"]}]}]}]}], "\[Implies]", 
                   RowBox[{"isAbove", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"1", 
                   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
                   
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    TagBox[")", "AutoParentheses"]}]}]}]}], "\[Implies]", 
                   RowBox[{"isAbove", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:112873953"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p112873953-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p112873953-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:112873953", "Proof|ID:112873953-1"},
 CellID->334296225],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p112873953-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934414512407*^9, 3.6149356370836*^9},
 CellID->1258513368]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1565002200],

Cell[CellGroupData[{

Cell["\<\
Lemma: isAbove, inRectangle and inFarExterior cover Space of Exponent Vectors\
\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599890111919245*^9, 3.599890141096925*^9}, {
  3.599890212328756*^9, 3.599890215700862*^9}, {3.603688580639262*^9, 
  3.603688585443306*^9}, {3.612255858892362*^9, 3.612255885182489*^9}},
 CellID->1334741891],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isExponentVector", "[", "x", "]"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"isAbove", "[", 
         RowBox[{"x", ",", "A"}], "]"}]},
       {
        RowBox[{"inRectangle", "[", 
         RowBox[{"x", ",", "A"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", "2"}]], 
         RowBox[{"inFarExterior", "[", 
          RowBox[{"x", ",", "A", ",", "k"}], "]"}]}]}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cover)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:315092621", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59989014765398*^9, 3.599890227830912*^9}, {
  3.604835796809836*^9, 3.604835822394441*^9}, {3.605594097880554*^9, 
  3.60559410669906*^9}, {3.612254384694562*^9, 3.61225439066407*^9}, {
  3.612254434680953*^9, 3.612254462320536*^9}},
 CellTags->{
  "ID:315092621", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "cover"},
 CellID->315092621],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1801799012]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->768394677],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cover)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cover)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{"1", 
               UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{"isAbove", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[Or]", 
                 RowBox[{"inRectangle", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[Or]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
                   RowBox[{"inFarExterior", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{"1", 
               UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{"isAbove", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[Or]", 
                 RowBox[{"inRectangle", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[Or]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
                   RowBox[{"inFarExterior", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:315092621"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p315092621-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p315092621-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:315092621", "Proof|ID:315092621-1"},
 CellID->875989985],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p315092621-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614934420296407*^9, 3.61493564192335*^9},
 CellID->1116871946]
}, Closed]],

Cell[TextData[{
 "If one compares the cover in this report to the partition in ",
 ButtonBox["[Buchberger1979]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979"],
 " and ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 ", one realizes that there a different cover/partition was chosen: There, \
the space of two-component exponents is divided w.r.t. ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " into"
}], "Text",
 CellChangeTimes->{{3.604037903675905*^9, 3.604038021988471*^9}, {
  3.60403839160167*^9, 3.604038435540428*^9}, {3.604038989030782*^9, 
  3.604039014679988*^9}, {3.605594119853312*^9, 3.605594151285301*^9}, {
  3.605597480929869*^9, 3.605597481417363*^9}, {3.615096874055312*^9, 
  3.615096875357876*^9}},
 CellID->1219107783],

Cell[CellGroupData[{

Cell[TextData[{
 "The \[OpenCurlyDoubleQuote]interior\[CloseCurlyDoubleQuote] of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ": Contains all the points that are multiples of the point ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"W", "[", 
      RowBox[{"A", ",", "1"}], "]"}], ",", 
     RowBox[{"W", "[", 
      RowBox[{"A", ",", "2"}], "]"}]}], "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 ". In particular, all the points in ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " lie in the interior of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.604038420721773*^9, 3.604038453812675*^9}, {
  3.604038564412401*^9, 3.604038631936407*^9}, {3.604039046112522*^9, 
  3.604039054905654*^9}},
 CellID->1150682902],

Cell[TextData[{
 "The space \[OpenCurlyDoubleQuote]below\[CloseCurlyDoubleQuote] ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ": Contains all the points that properly divide the point ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"W", "[", 
      RowBox[{"A", ",", "1"}], "]"}], ",", 
     RowBox[{"W", "[", 
      RowBox[{"A", ",", "2"}], "]"}]}], "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.604038420721773*^9, 3.604038453812675*^9}, {
  3.604038564412401*^9, 3.604038633485767*^9}, {3.604038683791597*^9, 
  3.604038750404537*^9}, {3.604039058818762*^9, 3.604039062252803*^9}},
 CellID->1824552351],

Cell[TextData[{
 "The \[OpenCurlyDoubleQuote]exterior\[CloseCurlyDoubleQuote] of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ": Contains all remaining points."
}], "Item",
 CellChangeTimes->{{3.604038420721773*^9, 3.604038453812675*^9}, {
  3.604038564412401*^9, 3.604038633485767*^9}, {3.604038683791597*^9, 
  3.604038770850693*^9}, {3.604039050143122*^9, 3.604039067462063*^9}},
 CellID->1894858166]
}, Open  ]],

Cell[TextData[{
 "One can easily see that \[OpenCurlyDoubleQuote]interior\
\[CloseCurlyDoubleQuote], \
\[OpenCurlyDoubleQuote]below\[CloseCurlyDoubleQuote] and \
\[OpenCurlyDoubleQuote]exterior\[CloseCurlyDoubleQuote] really form a \
partition of the two-dimensional ",
 StyleBox["ExponentSpace",
  FontFamily->"Courier 10 Pitch"],
 " (i.e. they are a disjoint cover).\nThe reason why this report deviates \
from ",
 ButtonBox["[Buchberger1979]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979"],
 " and ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 " is quite simple: The proof of \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " does not increase\[CloseCurlyDoubleQuote] proceeds differently here, and \
the cover ",
 Cell[BoxData[
  FormBox["isAbove", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["inRectangle", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["inFarExterior", TraditionalForm]]],
 " is better suited for the needs of the new proof than the partition in the \
referenced literature. For more details on the new proof see section ",
 ButtonBox["M[A]+W[A] Does Not Increase",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Invariant"],
 "."
}], "Text",
 CellChangeTimes->{{3.604039023572123*^9, 3.604039039652761*^9}, {
  3.604039072854227*^9, 3.60403912021155*^9}, {3.604039213569566*^9, 
  3.604039498415615*^9}, {3.604835835936537*^9, 3.604835854244079*^9}, {
  3.605597501225891*^9, 3.605597504912108*^9}, {3.612498781585596*^9, 
  3.612498785694463*^9}, {3.615096908632053*^9, 3.615096929122603*^9}},
 CellID->857096204]
}, Closed]],

Cell[CellGroupData[{

Cell["Divisibility, lcm", "Subsubsection",
 CellChangeTimes->{{3.613107577000648*^9, 3.613107577267338*^9}, {
  3.613107947897656*^9, 3.613107952497507*^9}},
 CellID->526429181],

Cell[TextData[{
 "This subsection summarizes some basic results in connection with ",
 Cell[BoxData[
  FormBox["lcm", TraditionalForm]]],
 " in the two-dimensional case."
}], "Text",
 CellChangeTimes->{{3.615097091641074*^9, 3.615097123845159*^9}},
 CellID->656737055],

Cell["", "OpenEnvironment",
 CellID->1128068551],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Some Properties of lcm for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.613108010992308*^9, 3.613108040008119*^9}, {
  3.613108421017795*^9, 3.613108421315032*^9}, {3.613108536880964*^9, 
  3.613108539965416*^9}, {3.613190939737612*^9, 3.613190950657227*^9}},
 CellID->624744263],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", 
   RowBox[{"x", ",", "y"}], "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.613107596607635*^9, 3.613107602520929*^9}},
 CellID->413827810],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["x", "1"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "1"]}], "\[And]", 
    RowBox[{
     SubscriptBox["y", "2"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["x", "2"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"lcm", "[", 
    RowBox[{"x", ",", "y"}], "]"}], "\[Equal]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     SubscriptBox["y", "1"], ",", 
     SubscriptBox["x", "2"]}], 
    "\[RightAngleBracket]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(51)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:744007071", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612596403557722*^9, 3.612596436460739*^9}, {
  3.61259678989797*^9, 3.612596790963714*^9}, {3.613108061593152*^9, 
  3.61310806642583*^9}, {3.61310843377879*^9, 3.613108447933819*^9}},
 CellTags->{
  "ID:744007071", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "51"},
 CellID->744007071],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["y", "1"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["x", "1"]}], "\[And]", 
    RowBox[{
     SubscriptBox["x", "2"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "2"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"lcm", "[", 
    RowBox[{"x", ",", "y"}], "]"}], "\[Equal]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     SubscriptBox["x", "1"], ",", 
     SubscriptBox["y", "2"]}], 
    "\[RightAngleBracket]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(52)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1720661645", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612691679751678*^9, 3.612691690107485*^9}, 
   3.613108104476849*^9, {3.613108455888164*^9, 3.613108467677952*^9}},
 CellTags->{
  "ID:1720661645", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "52"},
 CellID->1720661645],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isExponentVector", "[", "z", "]"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"z", "\[VerticalBar]", 
     RowBox[{"lcm", "[", 
      RowBox[{"x", ",", "y"}], "]"}]}], 
    TagBox["\[DoubleLeftRightArrow]",
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", "2"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[Or]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             SubscriptBox["z", "k"], 
             UnderscriptBox["\[LessEqual]", "D"], 
             SubscriptBox["x", "k"]}]},
           {
            RowBox[{
             SubscriptBox["z", "k"], 
             UnderscriptBox["\[LessEqual]", "D"], 
             SubscriptBox["y", "k"]}]}
          }]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[VerticalBar]lcm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1022330126", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613190788017524*^9, 3.613190856822939*^9}, {
  3.613191420585861*^9, 3.613191442756489*^9}},
 CellTags->{
  "ID:1022330126", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[VerticalBar]lcm"},
 CellID->1022330126],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->73373163]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->37865400],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(51)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(51)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
             RowBox[{
               RowBox[{"lcm", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
             RowBox[{
               RowBox[{"lcm", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:744007071"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p744007071-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p744007071-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:744007071", "Proof|ID:744007071-1"},
 CellID->1991259271],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p744007071-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614935649322968*^9},
 CellID->1421118479]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(52)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(52)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
             RowBox[{
               RowBox[{"lcm", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
             RowBox[{
               RowBox[{"lcm", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1720661645"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1720661645-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1720661645-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1720661645", "Proof|ID:1720661645-1"},
 CellID->491886609],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1720661645-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614935653242906*^9},
 CellID->814496018]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[VerticalBar]lcm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[VerticalBar]lcm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], "\[VerticalBar]", 
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], "\[VerticalBar]", 
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1022330126"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1022330126-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1022330126-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1022330126", "Proof|ID:1022330126-1"},
 CellID->448824604],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1022330126-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614935657123327*^9},
 CellID->1669274429]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["deg", "Subsubsection",
 CellChangeTimes->{{3.612691345659289*^9, 3.612691354501964*^9}, 
   3.612766261655032*^9},
 CellID->497056006],

Cell["\<\
This subsection is dedicated to alternative descriptions of the degree \
function, as well as to relations between the degrees of least common \
multiples under various conditions.\
\>", "Text",
 CellChangeTimes->{{3.615103697450025*^9, 3.615103791850552*^9}},
 CellID->2048305540],

Cell["", "OpenEnvironment",
 CellID->537258638],

Cell[CellGroupData[{

Cell["Lemma: Alternative Description of deg", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.612587960326744*^9, 3.612587974250631*^9}, {
  3.612588028307439*^9, 3.612588034484156*^9}},
 CellID->575367155],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", "2"}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"isExponentVector", "[", "x", "]"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"deg", "[", "x", "]"}], "\[Equal]", 
     RowBox[{
      SubscriptBox["x", "k"], 
      UnderscriptBox["+", "D"], 
      SubscriptBox["x", 
       SuperscriptBox["k", "-"]]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(53)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:528222346", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613811231916492*^9, 3.613811263208528*^9}},
 CellTags->{
  "ID:528222346", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "53"},
 CellID->528222346],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"u", ",", "v"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"deg", "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"u", ",", "v"}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
    RowBox[{"u", 
     UnderscriptBox["+", "D"], "v"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(54)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1598627200", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612587990952733*^9, 3.612588045566493*^9}},
 CellTags->{
  "ID:1598627200", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "54"},
 CellID->1598627200],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->728081085]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->949543245],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(53)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(53)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                 "\[Ellipsis]", ",", "2"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"deg", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["+", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                 "\[Ellipsis]", ",", "2"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"deg", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["+", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:528222346"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p528222346-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p528222346-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:528222346", "Proof|ID:528222346-1"},
 CellID->1476471173],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p528222346-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{
  3.614935663794234*^9, {3.615103810103247*^9, 3.61510381378585*^9}, 
   3.615103850672856*^9},
 CellID->876298637]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(54)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(54)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
              StyleBox["u", "ExpressionVariable"], ",", 
              StyleBox["v", "ExpressionVariable"]}]], 
           RowBox[{
             RowBox[{"deg", "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], ",", 
                   StyleBox["v", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["+", "D"], 
               StyleBox[
               "v", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
              StyleBox["u", "ExpressionVariable"], ",", 
              StyleBox["v", "ExpressionVariable"]}]], 
           RowBox[{
             RowBox[{"deg", "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], ",", 
                   StyleBox["v", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["+", "D"], 
               StyleBox[
               "v", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1598627200"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1598627200-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1598627200-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1598627200", "Proof|ID:1598627200-1"},
 CellID->605461210],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1598627200-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 32.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614935667905987*^9},
 CellID->162691003]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1347133480],

Cell[CellGroupData[{

Cell["Lemma: deg in some cases", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.613108531334453*^9, 3.613108544157678*^9}},
 CellID->761168772],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", 
   RowBox[{"x", ",", "y", ",", "z"}], "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.61310855284029*^9, 3.613108558866702*^9}, {
  3.61310869836834*^9, 3.613108699733393*^9}},
 CellID->2063990149],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["x", "1"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "1"]}], "\[And]", 
    RowBox[{
     SubscriptBox["y", "2"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["x", "2"]}], "\[And]", 
    RowBox[{
     SubscriptBox["y", "1"], 
     UnderscriptBox["\[LessEqual]", "D"], 
     SubscriptBox["z", "1"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"deg", "[", 
    RowBox[{"lcm", "[", 
     RowBox[{"x", ",", "y"}], "]"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"deg", "[", 
    RowBox[{"lcm", "[", 
     RowBox[{"x", ",", "z"}], "]"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(55)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1537560967", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613108570762324*^9, 3.613108620282472*^9}},
 CellTags->{
  "ID:1537560967", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "55"},
 CellID->1537560967],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["y", "1"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["x", "1"]}], "\[And]", 
    RowBox[{
     SubscriptBox["x", "2"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "2"]}], "\[And]", 
    RowBox[{
     SubscriptBox["y", "2"], 
     UnderscriptBox["\[LessEqual]", "D"], 
     SubscriptBox["z", "2"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"deg", "[", 
    RowBox[{"lcm", "[", 
     RowBox[{"x", ",", "y"}], "]"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"deg", "[", 
    RowBox[{"lcm", "[", 
     RowBox[{"x", ",", "z"}], "]"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(56)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:507460317", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613108639336791*^9, 3.613108654680444*^9}},
 CellTags->{
  "ID:507460317", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "56"},
 CellID->507460317],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->565711075]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1713796769],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(55)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(55)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
             RowBox[{
               RowBox[{"deg", "[", 
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{"deg", "[", 
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["z", "ExpressionVariable"]}], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
             RowBox[{
               RowBox[{"deg", "[", 
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{"deg", "[", 
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["z", "ExpressionVariable"]}], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1537560967"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1537560967-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1537560967-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1537560967", "Proof|ID:1537560967-1"},
 CellID->1213329230],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1537560967-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614935672770359*^9},
 CellID->1707761366]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(56)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(56)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
             RowBox[{
               RowBox[{"deg", "[", 
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{"deg", "[", 
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["z", "ExpressionVariable"]}], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExponentVector", "[", 
                 StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
             RowBox[{
               RowBox[{"deg", "[", 
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{"deg", "[", 
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["z", "ExpressionVariable"]}], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:507460317"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p507460317-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p507460317-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:507460317", "Proof|ID:507460317-1"},
 CellID->688346270],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p507460317-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614935676785557*^9},
 CellID->1314245273]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox["W", TraditionalForm]]]], "Subsubsection",
 CellChangeTimes->{{3.599890082065975*^9, 3.599890095287219*^9}, {
   3.604835862678826*^9, 3.604835864241992*^9}, 3.606468725898529*^9, {
   3.612511412574478*^9, 3.612511421503115*^9}, 3.612766243444576*^9},
 CellTags->"sect::LemmataSpecificSpecialW",
 CellID->1209432111],

Cell[TextData[{
 "The following lemma states the properties of ",
 Cell[BoxData[
  FormBox["W", TraditionalForm]]],
 " that were postponed in subsection ",
 ButtonBox["General Case",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataSpecificGeneral"],
 ". As can be seen, expressing ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], TraditionalForm]]],
 " in terms of simpler expressions free of the append-operator requires \
knowledge about ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " lying in the far exterior (w.r.t. any coordinate) w.r.t. ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ".\nOf course, one could also try to find an exact correspondence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", 
    RowBox[{"A", " ", "\:293a", " ", "x"}], "]"}], TraditionalForm]]],
 " to a simpler expression if ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " does not lie in the far exterior, but knowing formula ",
 ButtonBox["(W\:293a decreases)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:144808276"],
 " is absolutely sufficient for our purpose in such case."
}], "Text",
 CellChangeTimes->{{3.604040355342039*^9, 3.604040556196657*^9}, {
  3.604040616003144*^9, 3.604040804863452*^9}, {3.604040884587621*^9, 
  3.604040884998184*^9}, {3.615109505313048*^9, 3.615109601378091*^9}},
 CellID->434235030],

Cell["", "OpenEnvironment",
 CellID->551873],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", 
    RowBox[{
     RowBox[{"A", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "x"}], ",", "k"}], "]"}], TraditionalForm]]],
 " if ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " is in Far Exterior"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.598593887608899*^9, 3.598593896585044*^9}, {
   3.600688563257754*^9, 3.600688563257865*^9}, {3.612083824880569*^9, 
   3.612083860597567*^9}, 3.612499247307758*^9, {3.612504105440568*^9, 
   3.612504105662772*^9}},
 CellTags->"Lem::W[A<-x]",
 CellID->1283608134],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "2"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612498822545963*^9, 3.612498827232749*^9}},
 CellID->1093627130],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isExponentVector", "[", "x", "]"}]], 
  RowBox[{"inFarExterior", "[", 
   RowBox[{"x", ",", "A", ",", "k"}], "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.59859391525424*^9, 3.598593921638134*^9}, {
  3.605597883321216*^9, 3.605597887195657*^9}, {3.612083873957923*^9, 
  3.612083881936188*^9}, {3.612498832020167*^9, 3.61249883613436*^9}},
 CellID->1464211085],

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "[", 
   RowBox[{"A", ",", 
    SuperscriptBox["k", "-"]}], "]"}], 
  UnderscriptBox["<", "D"], 
  SubscriptBox["x", 
   SuperscriptBox["k", "-"]]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(57)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1442383873", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.612504213850235*^9},
 CellTags->{
  "ID:1442383873", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "57"},
 CellID->1442383873],

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "[", 
   RowBox[{
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], ",", "k"}], "]"}], "\[Equal]", 
  SubscriptBox["x", "k"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(W\:293a (binary) FarExterior 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:523824248", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.598593946676362*^9, 3.598593987487744*^9}, {
  3.598594029597392*^9, 3.598594031117108*^9}, {3.598594189073072*^9, 
  3.598594194564183*^9}, {3.600688543261228*^9, 3.600688546502059*^9}, {
  3.604037419751375*^9, 3.604037428458472*^9}, {3.60483587752374*^9, 
  3.604835892446223*^9}, {3.605597889878956*^9, 3.605597892518302*^9}, {
  3.606465617504588*^9, 3.606465641466087*^9}, {3.612083918476685*^9, 
  3.612083934007108*^9}, {3.612498865469325*^9, 3.612498866761954*^9}},
 CellTags->{
  "ID:523824248", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (binary) FarExterior 1"},
 CellID->523824248],

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "[", 
   RowBox[{
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], ",", 
    SuperscriptBox["k", "-"]}], "]"}], "\[Equal]", 
  RowBox[{"W", "[", 
   RowBox[{"A", ",", 
    SuperscriptBox["k", "-"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(W\:293a (binary) FarExterior 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1663623856", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.598594211416615*^9, 3.598594279583917*^9}, 
   3.60068855076712*^9, {3.604037439235703*^9, 3.604037444116539*^9}, {
   3.604835881290353*^9, 3.604835896240283*^9}, {3.605597893882441*^9, 
   3.605597896320043*^9}, {3.60646567442324*^9, 3.606465709811249*^9}, 
   3.612083932713483*^9, {3.612498869964683*^9, 3.612498876487371*^9}},
 CellTags->{
  "ID:1663623856", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (binary) FarExterior 2"},
 CellID->1663623856],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->249399937]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->532566392],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(57)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(57)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "2"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{"inFarExterior", "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "2"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{"inFarExterior", "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1442383873"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1442383873-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1442383873-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1442383873", "Proof|ID:1442383873-1"},
 CellID->1192136911],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1442383873-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614935684329338*^9, 3.614935798611512*^9},
 CellID->709403126]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(W\:293a (binary) FarExterior 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(W\:293a (binary) FarExterior 1)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "2"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{"inFarExterior", "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "2"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{"inFarExterior", "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:523824248"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p523824248-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p523824248-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:523824248", "Proof|ID:523824248-1"},
 CellID->842417444],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p523824248-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->960603845]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(W\:293a (binary) FarExterior 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(W\:293a (binary) FarExterior 2)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "2"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{"inFarExterior", "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "\[Equal]", 
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "2"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{"inFarExterior", "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "\[Equal]", 
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1663623856"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1663623856-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1663623856-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1663623856", "Proof|ID:1663623856-1"},
 CellID->658477440],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1663623856-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->430830718]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->693346110],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], TraditionalForm]]],
 " if ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " is in Far Exterior"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.612504114451341*^9, {3.6125041488253*^9, 3.612504148825388*^9}, {
   3.615109956632921*^9, 3.615109958123635*^9}, {3.615110201747692*^9, 
   3.615110202568332*^9}},
 CellID->1916344980],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.61250416126573*^9, 3.612504168251356*^9}},
 CellID->888020555],

Cell[BoxData[
 RowBox[{
  RowBox[{"inFarExterior", "[", 
   RowBox[{"x", ",", "A", ",", "1"}], "]"}], "\[Implies]", 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], "\[Equal]", 
   RowBox[{
    SubscriptBox["x", "1"], 
    UnderscriptBox["+", "D"], 
    RowBox[{"W", "[", 
     RowBox[{"A", ",", "2"}], "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(W\:293a (unary) FarExterior 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1829996465", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.598596357652608*^9, 3.59859643995701*^9}, {
   3.598596493005891*^9, 3.598596529215342*^9}, {3.598596634649827*^9, 
   3.598596650781329*^9}, 3.599281885157292*^9, 3.600688556360987*^9, {
   3.604037368296731*^9, 3.604037368781679*^9}, {3.604835883582531*^9, 
   3.604835897312566*^9}, {3.605597897573712*^9, 3.605597900692427*^9}, {
   3.612083938618029*^9, 3.61208393976263*^9}, {3.612498929248834*^9, 
   3.612498932049232*^9}, {3.612504172599393*^9, 3.612504192816515*^9}},
 CellTags->{
  "ID:1829996465", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (unary) FarExterior 1"},
 CellID->1829996465],

Cell[BoxData[
 RowBox[{
  RowBox[{"inFarExterior", "[", 
   RowBox[{"x", ",", "A", ",", "2"}], "]"}], "\[Implies]", 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{"W", "[", 
     RowBox[{"A", ",", "1"}], "]"}], 
    UnderscriptBox["+", "D"], 
    SubscriptBox["x", "2"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(W\:293a (unary) FarExterior 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:842592091", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.598606017390757*^9, 3.598606067009717*^9}, {
   3.598694830959859*^9, 3.598694831255875*^9}, 3.599281888840397*^9, 
   3.600688558678363*^9, {3.604037401805295*^9, 3.604037402261161*^9}, {
   3.604835885872616*^9, 3.604835898566688*^9}, {3.605597902240226*^9, 
   3.605597904862335*^9}, 3.606466190795*^9, {3.612083948992944*^9, 
   3.612083950173821*^9}, {3.61249932742829*^9, 3.612499347677366*^9}, 
   3.612500025849412*^9, 3.612504211747241*^9, {3.612504292203072*^9, 
   3.612504307973578*^9}},
 CellTags->{
  "ID:842592091", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (unary) FarExterior 2"},
 CellID->842592091],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", "2"}]], 
   RowBox[{"inFarExterior", "[", 
    RowBox[{"x", ",", "A", ",", "k"}], "]"}]], 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], "\[Equal]", 
   RowBox[{
    SubscriptBox["x", "k"], 
    UnderscriptBox["+", "D"], 
    RowBox[{"W", "[", 
     RowBox[{"A", ",", 
      SuperscriptBox["k", "-"]}], "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(W\:293a FarExterior)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1051080023", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612504347934946*^9, 3.612504377879122*^9}},
 CellTags->{
  "ID:1051080023", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a FarExterior"},
 CellID->1051080023],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->965026936]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->686705729],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(W\:293a (unary) FarExterior 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(W\:293a (unary) FarExterior 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"inFarExterior", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", "2"}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"inFarExterior", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", "2"}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1829996465"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1829996465-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1829996465-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1829996465", "Proof|ID:1829996465-1"},
 CellID->590566551],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1829996465-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614936924772471*^9},
 CellID->1758652648]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(W\:293a (unary) FarExterior 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(W\:293a (unary) FarExterior 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"inFarExterior", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "2"}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"inFarExterior", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "2"}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:842592091"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p842592091-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p842592091-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:842592091", "Proof|ID:842592091-1"},
 CellID->917529032],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p842592091-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614936932627936*^9},
 CellID->864478433]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(W\:293a FarExterior)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(W\:293a FarExterior)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
                RowBox[{"inFarExterior", "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
                RowBox[{"inFarExterior", "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1051080023"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1051080023-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1051080023-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1051080023", "Proof|ID:1051080023-1"},
 CellID->148396000],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1051080023-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614936936747369*^9},
 CellID->1486101954]
}, Closed]],

Cell[TextData[{
 "Formula ",
 ButtonBox["(59)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1988188117"],
 ", though completely obvious (at least when knowing formula ",
 ButtonBox["(58)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:631716597"],
 "), is convenient to have when proving ",
 ButtonBox["(invariant)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2102726322"],
 "."
}], "Text",
 CellChangeTimes->{{3.613816901023554*^9, 3.613816981506516*^9}, {
  3.615110437856232*^9, 3.615110530968196*^9}, {3.615787860952046*^9, 
  3.615787872911283*^9}},
 CellID->1033758105],

Cell["", "OpenEnvironment",
 CellID->1236240386],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Alternative Description of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", "A", "]"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.61381588597408*^9, 3.613815898391836*^9}},
 CellID->1017694636],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "2"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.613816856508299*^9, 3.613816860878884*^9}},
 CellID->1667409472],

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "[", "A", "]"}], "\[Equal]", 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "k"}], "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"W", "[", 
    RowBox[{"A", ",", 
     SuperscriptBox["k", "-"]}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(58)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:631716597", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613815906491857*^9, 3.61381593192524*^9}, {
  3.613816863275997*^9, 3.613816864492171*^9}},
 CellTags->{
  "ID:631716597", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "58"},
 CellID->631716597],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "k"}], "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"W", "[", 
    RowBox[{"A", ",", 
     SuperscriptBox["k", "-"]}], "]"}]}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{"W", "[", "A", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(59)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1988188117", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613816885254933*^9, 3.613816892658059*^9}},
 CellTags->{
  "ID:1988188117", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "59"},
 CellID->1988188117],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1087276993]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->42773662],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(58)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(58)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "2"}]}}]], 
             RowBox[{
               RowBox[{"W", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "D"], 
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", "2"}]}}]], 
             RowBox[{
               RowBox[{"W", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "D"], 
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:631716597"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p631716597-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p631716597-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:631716597", "Proof|ID:631716597-1"},
 CellID->1960584677],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p631716597-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614936941747842*^9},
 CellID->1246911685]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(59)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(59)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                RowBox[{
                  RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                  "\[Ellipsis]", ",", "2"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{
               RowBox[{"W", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{"W", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}]}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"W", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                RowBox[{
                  RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                  "\[Ellipsis]", ",", "2"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{
               RowBox[{"W", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{"W", "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}]}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"W", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1988188117"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1988188117-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1988188117-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1988188117", "Proof|ID:1988188117-1"},
 CellID->1726608150],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1988188117-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->737511693]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox["L", TraditionalForm]]]], "Subsubsection",
 CellChangeTimes->{{3.613817898546031*^9, 3.613817902762465*^9}},
 CellID->414836815],

Cell[TextData[{
 "Formulas ",
 ButtonBox["(61)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:914554779"],
 " and ",
 ButtonBox["(62)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1336064727"],
 " in the following lemma are only needed to rename some bound variables. The \
essential result is formula ",
 ButtonBox["(L+\[LessEqual])",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1651077201"],
 ", which will be needed for proving ",
 ButtonBox["(bound)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1990359651"],
 "."
}], "Text",
 CellChangeTimes->{{3.613819083064443*^9, 3.61381922815396*^9}, {
  3.615110713094331*^9, 3.615110798821338*^9}, {3.615711970557035*^9, 
  3.615711975046902*^9}},
 CellID->648452100],

Cell["", "OpenEnvironment",
 CellID->641967087],

Cell[CellGroupData[{

Cell["Lemma", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellID->1971506263],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"u", ",", "v"}], "]"}]}]]], "GlobalDeclaration",
 ShowCellTags->False,
 CellChangeTimes->{{3.613817959390174*^9, 3.613817966547287*^9}},
 CellTags->{
  "ID:590870546", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->590870546],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       SubscriptBox[
        RowBox[{"(", 
         SubscriptBox["A", "i"], ")"}], "1"], 
       UnderscriptBox["+", "D"], 
       SubscriptBox[
        RowBox[{"(", 
         SubscriptBox["A", "j"], ")"}], "2"]}], ")"}], 
     UnderscriptBox["+", "D"], "v"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(60)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:501569980", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613879642949423*^9, 3.613879671076161*^9}},
 CellTags->{
  "ID:501569980", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "60"},
 CellID->501569980],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["A", "j"], ")"}], "1"], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["A", "i"], ")"}], "1"], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(61)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:914554779", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613819062468697*^9, 3.613819066637232*^9}, {
   3.613819232609869*^9, 3.613819234563484*^9}, 3.613819661906123*^9, 
   3.613819922418474*^9},
 CellTags->{
  "ID:914554779", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "61"},
 CellID->914554779],

Cell[BoxData[
 RowBox[{
  RowBox[{"deg", "[", 
   RowBox[{"L", "[", "A", "]"}], "]"}], "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["max", "D"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      SubscriptBox[
       RowBox[{"(", 
        SubscriptBox["A", "i"], ")"}], "1"], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}], "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{
    UnderscriptBox["max", "D"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      SubscriptBox[
       RowBox[{"(", 
        SubscriptBox["A", "j"], ")"}], "2"], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(62)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1336064727", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613818814868281*^9, 3.613818902629689*^9}, {
   3.613819242780876*^9, 3.613819252336206*^9}, 3.613819669087589*^9, 
   3.613819917583534*^9},
 CellTags->{
  "ID:1336064727", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "62"},
 CellID->1336064727],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", ",", 
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        SubscriptBox[
         RowBox[{"(", 
          SubscriptBox["A", "i"], ")"}], "1"], 
        UnderscriptBox["+", "D"], 
        SubscriptBox[
         RowBox[{"(", 
          SubscriptBox["A", "j"], ")"}], "2"]}], ")"}], 
      UnderscriptBox["+", "D"], "v"}], 
     UnderscriptBox["\[LessEqual]", "D"], "u"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    RowBox[{"deg", "[", 
     RowBox[{"L", "[", "A", "]"}], "]"}], 
    UnderscriptBox["+", "D"], "v"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(L+\[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1651077201", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613817981614898*^9, 3.613818011048619*^9}, {
  3.613818296151806*^9, 3.613818340846981*^9}, {3.613821876703694*^9, 
  3.613821881127746*^9}},
 CellTags->{
  "ID:1651077201", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "L+\[LessEqual]"},
 CellID->1651077201],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1659661723]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->455739212],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(60)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(60)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                RowBox[{
                  UnderscriptBox["\[Element]", "D"], "[", 
                  StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                 UnderscriptBox["+", "D"], 
                 StyleBox["v", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                RowBox[{
                  UnderscriptBox["\[Element]", "D"], "[", 
                  StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                 UnderscriptBox["+", "D"], 
                 StyleBox["v", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:501569980"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p501569980-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p501569980-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:501569980", "Proof|ID:501569980-1"},
 CellID->1185891342],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p501569980-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1281666503]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(61)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(61)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:914554779"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p914554779-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p914554779-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:914554779", "Proof|ID:914554779-1"},
 CellID->5884268],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p914554779-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.61493695749908*^9},
 CellID->830333471]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(62)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(62)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{"deg", "[", 
               RowBox[{"L", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                   "\[RightAngleBracket]"}], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{"deg", "[", 
               RowBox[{"L", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                   "\[RightAngleBracket]"}], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{
                 UnderscriptBox["max", "D"], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1336064727"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1336064727-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1336064727-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1336064727", "Proof|ID:1336064727-1"},
 CellID->970257786],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1336064727-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614936961498324*^9},
 CellID->1589057661]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(L+\[LessEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(L+\[LessEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"deg", "[", 
                    RowBox[{"L", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], "]"}], 
                   UnderscriptBox["+", "D"], 
                   StyleBox["v", "ExpressionVariable"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox[
                 "u", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["+", "D"], 
                    StyleBox["v", "ExpressionVariable"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"deg", "[", 
                    RowBox[{"L", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], "]"}], 
                   UnderscriptBox["+", "D"], 
                   StyleBox["v", "ExpressionVariable"]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox[
                 "u", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1651077201"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1651077201-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1651077201-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1651077201", "Proof|ID:1651077201-1"},
 CellID->451306320],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1651077201-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614936975113832*^9},
 CellID->965210338]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["chainCrit", "Subsubsection",
 CellChangeTimes->{{3.599890082065975*^9, 3.599890095287219*^9}, {
   3.604835862678826*^9, 3.604835864241992*^9}, 3.606468728940797*^9},
 CellTags->"sect::chainCritSpecial",
 CellID->226754043],

Cell[TextData[{
 "We conclude the whole section about auxiliary lemmata with an alternative \
description of the ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 "-predicate in the special case ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 ". This description is better suited for the proofs of the main theorems, \
where we heavily rely on covering the space of two-component exponent vectors \
by ",
 Cell[BoxData[
  FormBox["isAbove", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["inRectangle", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["inFarExterior", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.604040897468361*^9, 3.604041043587647*^9}, {
  3.604990457470149*^9, 3.604990473835212*^9}, {3.605597947558214*^9, 
  3.605597961294178*^9}, {3.612516432161397*^9, 3.612516433582868*^9}, {
  3.612521106235057*^9, 3.612521111453918*^9}},
 CellID->436231197],

Cell[TextData[{
 "Let in this subsection ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " always be two exponent vectors:"
}], "Text",
 CellChangeTimes->{{3.615181079812698*^9, 3.615181101620145*^9}},
 CellID->826985685],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", 
   RowBox[{"x", ",", "y"}], "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.606468856363545*^9, 3.606468864087621*^9}},
 CellID->905478503],

Cell["\<\
We start by proving some auxiliary results about types of tuples:\
\>", "Text",
 CellChangeTimes->{{3.615181043918216*^9, 3.61518106478312*^9}},
 CellID->1009550453],

Cell["", "OpenEnvironment",
 CellID->28139740],

Cell[CellGroupData[{

Cell["\<\
Lemma: Useful Type Information for proving the Alternative Description of \
chainCrit\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.612582998549204*^9, 3.612583012209914*^9}, {
   3.612587720570058*^9, 3.612587747000272*^9}, 3.612595414110258*^9},
 CellID->755849276],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExpVectorTuple", "[", "B", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612583692415651*^9, 3.612583698959736*^9}},
 CellID->2145956178],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "2"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612583087013432*^9, 3.612583093382982*^9}},
 CellID->1284341245],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "B", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612583060070735*^9, 3.612583067929712*^9}, 
   3.612583703155001*^9},
 CellID->325823042],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[SquareSubsetEqual]", "D"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["B", "j"], ")"}], "k"], ",", 
     SubscriptBox["x", "k"]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(63)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:53741691", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612583026101914*^9, 3.612583036577552*^9}, {
   3.61258307119908*^9, 3.612583106136735*^9}, 3.612583705143969*^9},
 CellTags->{
  "ID:53741691", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "63"},
 CellID->53741691],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[SquareSubsetEqual]", "D"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     SubscriptBox["x", "k"], ",", 
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["B", "j"], ")"}], "k"]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(64)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1045358942", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61258311561964*^9, 3.612583122730213*^9}, 
   3.612583707499003*^9},
 CellTags->{
  "ID:1045358942", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "64"},
 CellID->1045358942],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["max", "D"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      SubscriptBox[
       RowBox[{"(", 
        SubscriptBox["B", "j"], ")"}], "k"], ",", 
      SubscriptBox["x", "k"]}], "\[RightAngleBracket]"}], "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(65)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1132937314", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612583138786656*^9, 3.612583157365686*^9}, 
   3.612583709597523*^9},
 CellTags->{
  "ID:1132937314", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "65"},
 CellID->1132937314],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["max", "D"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      SubscriptBox["x", "k"], ",", 
      SubscriptBox[
       RowBox[{"(", 
        SubscriptBox["B", "j"], ")"}], "k"]}], "\[RightAngleBracket]"}], 
    "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(66)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:152254973", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612583166888525*^9, 3.612583196416608*^9}, 
   3.612583712283871*^9},
 CellTags->{
  "ID:152254973", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "66"},
 CellID->152254973],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->204886373]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1003232302],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(63)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(63)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExpVectorTuple", "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExpVectorTuple", "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:53741691"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p53741691-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p53741691-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:53741691", "Proof|ID:53741691-1"},
 CellID->1005730830],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p53741691-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614936993377059*^9},
 CellID->1697566003]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(64)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(64)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExpVectorTuple", "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExpVectorTuple", "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1045358942"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1045358942-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1045358942-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1045358942", "Proof|ID:1045358942-1"},
 CellID->1783727900],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1045358942-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{
  3.614936997512561*^9, {3.617966458511518*^9, 3.617966459830585*^9}},
 CellID->74922882]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(65)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(65)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExpVectorTuple", "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], "]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExpVectorTuple", "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], "]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1132937314"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1132937314-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1132937314-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1132937314", "Proof|ID:1132937314-1"},
 CellID->279556406],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1132937314-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937001736975*^9},
 CellID->982255660]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(66)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(66)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExpVectorTuple", "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], "]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExponentVector", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExpVectorTuple", "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], "]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:152254973"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p152254973-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p152254973-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:152254973", "Proof|ID:152254973-1"},
 CellID->1134289223],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p152254973-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937005736169*^9},
 CellID->39010603]
}, Closed]],

Cell[TextData[{
 "For the first time, we now consider the ",
 Cell[BoxData[
  FormBox["chainCritDiagonal", TraditionalForm]]],
 "-predicate that was introduced in section ",
 ButtonBox["Specific Notions for Complexity Analysis",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::SpecificNotions"],
 ". The reason why we did not consider it in the general case is that it only \
makes sense if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 ".\nJust as the ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 "-predicate, also the ",
 Cell[BoxData[
  FormBox["chainCritDiagonal", TraditionalForm]]],
 "-predicate is symmetric (formula ",
 ButtonBox["(symmetry of chainCritDiagonal)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1490326930"],
 "), and if it holds for ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " w.r.t. the first coordinate, we may draw some conclusions about some least \
common multiples:"
}], "Text",
 CellChangeTimes->{{3.615181117890623*^9, 3.615181335219767*^9}},
 CellID->891029648],

Cell["", "OpenEnvironment",
 CellID->1820021401],

Cell[CellGroupData[{

Cell["Lemma: Properties of chainCritDiagonal", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.612517329723412*^9, 3.612517335038017*^9}, {
  3.612517625392013*^9, 3.612517625651682*^9}, {3.612758141092217*^9, 
  3.612758141277645*^9}},
 CellID->739473411],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "2"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612517341421053*^9, 3.612517344976177*^9}},
 CellID->804277790],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A", ",", "k"}], "]"}], "\[Implies]", 
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"y", ",", "x", ",", "A", ",", 
    SuperscriptBox["k", "-"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symmetry of chainCritDiagonal 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:836877533", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612517353996914*^9, 3.612517380069741*^9}},
 CellTags->{
  "ID:836877533", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 1"},
 CellID->836877533],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A", ",", "k"}], "]"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"y", ",", "x", ",", "A", ",", 
    SuperscriptBox["k", "-"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symmetry of chainCritDiagonal 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1226632396", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612517543139844*^9, 3.612517549580035*^9}},
 CellTags->{
  "ID:1226632396", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 2"},
 CellID->1226632396],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A", ",", 
    SuperscriptBox["k", "-"]}], "]"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"y", ",", "x", ",", "A", ",", "k"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symmetry of chainCritDiagonal 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:457131364", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612519267263262*^9, 3.612519302843754*^9}, {
  3.612519338420067*^9, 3.61251934222553*^9}, {3.612520969654069*^9, 
  3.612520970578807*^9}, {3.612686268681049*^9, 3.612686273784097*^9}, {
  3.612758192838723*^9, 3.612758193977484*^9}},
 CellTags->{
  "ID:457131364", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 3"},
 CellID->457131364],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"y", ",", "x", ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symmetry of chainCritDiagonal)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1490326930", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61251739258087*^9, 3.612517409152721*^9}},
 CellTags->{
  "ID:1490326930", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal"},
 CellID->1490326930],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A", ",", "1"}], "]"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[And]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{
        SubscriptBox["x", "1"], 
        UnderscriptBox["<", "D"], 
        SubscriptBox["y", "1"]}]},
      {
       RowBox[{
        SubscriptBox["y", "2"], 
        UnderscriptBox["<", "D"], 
        SubscriptBox["x", "2"]}]},
      {
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[Or]", 
          RowBox[{"\[Piecewise]", GridBox[{
             {
              RowBox[{
               SubscriptBox["y", "1"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               SubscriptBox[
                RowBox[{"(", 
                 SubscriptBox["A", "j"], ")"}], "1"]}]},
             {
              RowBox[{
               SubscriptBox["x", "2"], 
               UnderscriptBox["\[LessEqual]", "D"], 
               SubscriptBox[
                RowBox[{"(", 
                 SubscriptBox["A", "j"], ")"}], "2"]}]}
            }]}]}], 
         TagBox[")",
          "AutoParentheses"]}]}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCritDiagonal w.r.t. first coordinate)", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:749802963", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.612524776000106*^9, {3.612590347962481*^9, 3.612590360597978*^9}},
 CellTags->{
  "ID:749802963", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal w.r.t. first coordinate"},
 CellID->749802963],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A", ",", "1"}], "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.612517660964982*^9, 3.612517665049375*^9}},
 CellID->2008961285],

Cell[BoxData[
 RowBox[{
  RowBox[{"lcm", "[", 
   RowBox[{"x", ",", "y"}], "]"}], "\[Equal]", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    SubscriptBox["y", "1"], ",", 
    SubscriptBox["x", "2"]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(67)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:505706706", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612517152142541*^9, 3.612517176831862*^9}, {
  3.612517249879813*^9, 3.612517281443537*^9}, {3.612517668152471*^9, 
  3.612517669519227*^9}},
 CellTags->{
  "ID:505706706", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "67"},
 CellID->505706706],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["A", "j"], "\[VerticalBar]", 
     RowBox[{"lcm", "[", 
      RowBox[{"x", ",", "y"}], "]"}]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            RowBox[{"lcm", "[", 
             RowBox[{"x", ",", 
              SubscriptBox["A", "j"]}], "]"}], "\[Equal]", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
              RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{
                 SubscriptBox["x", "1"], ",", 
                 SubscriptBox[
                  RowBox[{"(", 
                   SubscriptBox["A", "j"], ")"}], "1"]}], 
                "\[RightAngleBracket]"}], "]"}], ",", 
              SubscriptBox["x", "2"]}], "\[RightAngleBracket]"}]}]},
          {
           RowBox[{
            RowBox[{"lcm", "[", 
             RowBox[{
              SubscriptBox["A", "j"], ",", "y"}], "]"}], "\[Equal]", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
              SubscriptBox["y", "1"], ",", 
              RowBox[{
               UnderscriptBox["max", "D"], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{
                 SubscriptBox[
                  RowBox[{"(", 
                   SubscriptBox["A", "j"], ")"}], "2"], ",", 
                 SubscriptBox["y", "2"]}], "\[RightAngleBracket]"}], "]"}]}], 
             "\[RightAngleBracket]"}]}]}
         }]}]}],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(68)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1262287499", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612583734669499*^9, 3.612583741199088*^9}, {
   3.612587231271161*^9, 3.612587340606954*^9}, {3.612587391757716*^9, 
   3.612587418879793*^9}, {3.612587470082844*^9, 3.612587562335671*^9}, 
   3.612587682645532*^9, {3.612587826159916*^9, 3.612587830795328*^9}, {
   3.612588799218152*^9, 3.612588801841494*^9}, {3.612588856873998*^9, 
   3.612588872487032*^9}},
 CellTags->{
  "ID:1262287499", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "68"},
 CellID->1262287499],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->263104304]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->350642411],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symmetry of chainCritDiagonal 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symmetry of chainCritDiagonal 1)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:836877533"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p836877533-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p836877533-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:836877533", "Proof|ID:836877533-1"},
 CellID->313941599],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p836877533-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937010760441*^9},
 CellID->1689572409]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symmetry of chainCritDiagonal 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symmetry of chainCritDiagonal 2)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1226632396"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1226632396-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1226632396-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1226632396", "Proof|ID:1226632396-1"},
 CellID->346366551],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1226632396-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937014576304*^9},
 CellID->13987266]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symmetry of chainCritDiagonal 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symmetry of chainCritDiagonal 3)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{"chainCritDiagonal", "[", 
                   RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:457131364"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p457131364-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p457131364-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:457131364", "Proof|ID:457131364-1"},
 CellID->1718476676],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p457131364-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 32.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937022159988*^9},
 CellID->89734148]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symmetry of chainCritDiagonal)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symmetry of chainCritDiagonal)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1490326930"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1490326930-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1490326930-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1490326930", "Proof|ID:1490326930-1"},
 CellID->758177161],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1490326930-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937018431651*^9},
 CellID->1459213616]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCritDiagonal w.r.t. first coordinate)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCritDiagonal w.r.t. first coordinate)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"(", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}]}]}]}]}], 
                 ")"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"(", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}]}]}]}]}], 
                 ")"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:749802963"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p749802963-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p749802963-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:749802963", "Proof|ID:749802963-1"},
 CellID->1283930748],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p749802963-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937025767775*^9},
 CellID->443166725]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(67)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(67)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "\[RightAngleBracket]"}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"lcm", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   "\[RightAngleBracket]"}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:505706706"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p505706706-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p505706706-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:505706706", "Proof|ID:505706706-1"},
 CellID->292146520],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p505706706-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937029703782*^9},
 CellID->389998305]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(68)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(68)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "\[VerticalBar]", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "\[Equal]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], "]"}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}]}], "\[And]", 
                    RowBox[{
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], "]"}]}], 
                    "\[RightAngleBracket]"}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "\[VerticalBar]", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "\[Equal]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], "]"}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}]}], "\[And]", 
                    RowBox[{
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    UnderscriptBox["max", "D"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "\[RightAngleBracket]"}], "]"}]}], 
                    "\[RightAngleBracket]"}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1262287499"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1262287499-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1262287499-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1262287499", "Proof|ID:1262287499-1"},
 CellID->1935805160],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1262287499-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.61493703329524*^9},
 CellID->1719653774]
}, Closed]],

Cell["\<\
Before we can eventually prove the main result of this subsection, we first \
have to prove some more auxiliary formulas that are of no particular interest \
by themselves:\
\>", "Text",
 CellChangeTimes->{{3.615181429657214*^9, 3.61518148239603*^9}},
 CellID->6932400],

Cell["", "OpenEnvironment",
 CellID->1921149736],

Cell[CellGroupData[{

Cell["\<\
Lemma: Preparations for the Alternative Description of chaincrit\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.612517633924766*^9, 3.612517634899988*^9}, {
  3.612600457791613*^9, 3.612600458968208*^9}, {3.612689621825973*^9, 
  3.612689633815606*^9}},
 CellID->436470413],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612758208342206*^9, 3.612758215018646*^9}},
 CellID->1928470162],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["y", "1"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["x", "1"]}], "\[And]", 
    RowBox[{
     SubscriptBox["x", "2"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "2"]}], "\[And]", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["A", "j"], ")"}], "1"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["x", "1"]}], "\[And]", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["A", "j"], ")"}], "2"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "2"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"\[Not]", 
   RowBox[{"chainCrit", "[", 
    RowBox[{"x", ",", "y", ",", "A"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(69)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1909233179", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612689641397396*^9, 3.6126897146331*^9}, 
   3.612689899299775*^9, {3.612758219962083*^9, 3.612758223379278*^9}},
 CellTags->{
  "ID:1909233179", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "69"},
 CellID->1909233179],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["x", "1"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "1"]}], "\[And]", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["A", "j"], ")"}], "1"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "1"]}], "\[And]", 
    RowBox[{
     RowBox[{
      SubscriptBox["y", "1"], 
      UnderscriptBox["+", "D"], 
      SubscriptBox["x", "2"]}], 
     UnderscriptBox["\[LessEqual]", "D"], 
     RowBox[{"deg", "[", 
      RowBox[{"lcm", "[", 
       RowBox[{"x", ",", 
        SubscriptBox["A", "j"]}], "]"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   SubscriptBox["x", "2"], 
   UnderscriptBox["\[LessEqual]", "D"], 
   SubscriptBox[
    RowBox[{"(", 
     SubscriptBox["A", "j"], ")"}], "2"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(70)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2125747942", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.598594992744306*^9, 3.598595043890403*^9}, {
  3.612756669745426*^9, 3.612756830983343*^9}, {3.612758386255733*^9, 
  3.61275839945802*^9}},
 CellTags->{
  "ID:2125747942", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "70"},
 CellID->2125747942],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["y", "2"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["x", "2"]}], "\[And]", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["A", "j"], ")"}], "1"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "1"]}], "\[And]", 
    RowBox[{
     RowBox[{
      SubscriptBox["y", "1"], 
      UnderscriptBox["+", "D"], 
      SubscriptBox["x", "2"]}], 
     UnderscriptBox["\[LessEqual]", "D"], 
     RowBox[{"deg", "[", 
      RowBox[{"lcm", "[", 
       RowBox[{
        SubscriptBox["A", "j"], ",", "y"}], "]"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   SubscriptBox["x", "2"], 
   UnderscriptBox["\[LessEqual]", "D"], 
   SubscriptBox[
    RowBox[{"(", 
     SubscriptBox["A", "j"], ")"}], "2"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(71)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1211430036", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.598595060646004*^9, 3.59859507785788*^9}, {
  3.612758083673637*^9, 3.612758108638558*^9}, {3.612758226169994*^9, 
  3.612758340257703*^9}, {3.612758404311161*^9, 3.612758407145511*^9}, {
  3.612761089742928*^9, 3.61276109314906*^9}},
 CellTags->{
  "ID:1211430036", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "71"},
 CellID->1211430036],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[Not]", 
    RowBox[{"chainCritDiagonal", "[", 
     RowBox[{"x", ",", "y", ",", "A", ",", "2"}], "]"}]}], "\[And]", 
   RowBox[{"chainCrit", "[", 
    RowBox[{"x", ",", "y", ",", "A"}], "]"}], "\[And]", 
   RowBox[{"\[Not]", 
    RowBox[{"x", "\[VerticalBar]", "y"}]}], "\[And]", 
   RowBox[{"\[Not]", 
    RowBox[{"y", "\[VerticalBar]", "x"}]}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.612687545291148*^9, 3.612687560438697*^9}, {
  3.612688239453942*^9, 3.612688265734522*^9}, {3.612688814886776*^9, 
  3.612688868842518*^9}},
 CellID->643000006],

Cell[BoxData[
 RowBox[{
  SubscriptBox["x", "1"], 
  UnderscriptBox["<", "D"], 
  SubscriptBox["y", "1"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(72)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2061027934", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612600468726154*^9, 3.612600517834534*^9}, {
  3.612600775865662*^9, 3.612600782603854*^9}, {3.612687552036024*^9, 
  3.612687570936288*^9}},
 CellTags->{
  "ID:2061027934", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "72"},
 CellID->2061027934],

Cell[BoxData[
 RowBox[{
  SubscriptBox["y", "2"], 
  UnderscriptBox["<", "D"], 
  SubscriptBox["x", "2"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(73)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1653917255", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.612687569571729*^9},
 CellTags->{
  "ID:1653917255", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "73"},
 CellID->1653917255],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->638144522]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1008755936],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(69)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(69)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1909233179"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1909233179-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1909233179-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1909233179", "Proof|ID:1909233179-1"},
 CellID->1389502986],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1909233179-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937038574528*^9},
 CellID->972044349]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(70)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(70)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}]}]}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}]}]}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2125747942"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2125747942-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2125747942-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2125747942", "Proof|ID:2125747942-1"},
 CellID->673958282],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2125747942-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937042151169*^9},
 CellID->71081292]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(71)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(71)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]}]}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]}]}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1211430036"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1211430036-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1211430036-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1211430036", "Proof|ID:1211430036-1"},
 CellID->1679514424],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1211430036-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.61493704583089*^9},
 CellID->321998331]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(72)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(72)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", "2"}], "]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", "2"}], "]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2061027934"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2061027934-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2061027934-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2061027934", "Proof|ID:2061027934-1"},
 CellID->381202233],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2061027934-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937049374195*^9},
 CellID->1850172725]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(73)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(73)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", "2"}], "]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", "2"}], "]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                   TagBox[")", "AutoParentheses"]}], 
                 UnderscriptBox["<", "D"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1653917255"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1653917255-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1653917255-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1653917255", "Proof|ID:1653917255-1"},
 CellID->929231870],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1653917255-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937053085645*^9},
 CellID->2027019073]
}, Closed]],

Cell[TextData[{
 "Finally, we can state and prove the main result of this subsection, namely ",
 ButtonBox["(chainCrit bivariate)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:810037327"],
 ": This formula tells us how ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 " can be expressed in terms of ",
 Cell[BoxData[
  FormBox["chainCritDiagonal", TraditionalForm]]],
 " and divisibility relations. The five formulas before are needed for \
showing the two directions \[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\
\[CloseCurlyDoubleQuote] and \[OpenCurlyDoubleQuote]\[DoubleLongRightArrow]\
\[CloseCurlyDoubleQuote], and formula ",
 ButtonBox["(chainCrit bivariate I)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2111681775"],
 " is an immediate consequence of ",
 ButtonBox["(chainCrit bivariate)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:810037327"],
 " and the definition of the ternary ",
 Cell[BoxData[
  FormBox["chainCritDiagonal", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.615181495537924*^9, 3.615181595289997*^9}, {
  3.61518164177523*^9, 3.615181715944172*^9}, {3.61518179362147*^9, 
  3.615181815054329*^9}, {3.615181910970602*^9, 3.615181911008917*^9}},
 CellID->1006068998],

Cell["", "OpenEnvironment",
 CellID->695996923],

Cell[CellGroupData[{

Cell["Lemma: Alternative description of chainCrit", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599280526873647*^9, 3.599280541933178*^9}, {
  3.599280810884097*^9, 3.599280817283692*^9}, {3.599890020612568*^9, 
  3.599890024701671*^9}, {3.604835904811762*^9, 3.604835906146348*^9}, {
  3.612690044117126*^9, 3.612690049564246*^9}},
 CellID->750017357],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A", ",", "1"}], "]"}], "\[Implies]", 
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit bivariate 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1690170957", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.606469083016791*^9, 3.606469102770933*^9}, {
  3.606474967602486*^9, 3.606474969425221*^9}, {3.612517794739827*^9, 
  3.612517799142436*^9}, {3.612524660070936*^9, 3.612524676612719*^9}, {
  3.612524773974228*^9, 3.612524794760395*^9}},
 CellTags->{
  "ID:1690170957", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 1"},
 CellID->1690170957],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}], "\[Implies]", 
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit bivariate 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:271680825", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612520480690622*^9, 3.612520482687706*^9}},
 CellTags->{
  "ID:271680825", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 2"},
 CellID->271680825],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["x", "1"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "1"]}], "\[And]", 
    RowBox[{
     SubscriptBox["y", "2"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["x", "2"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", "y", ",", "A"}], "]"}], 
    TagBox["\[DoubleLeftRightArrow]",
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[Or]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            SubscriptBox["y", "1"], 
            UnderscriptBox["<", "D"], 
            SubscriptBox[
             RowBox[{"(", 
              SubscriptBox["A", "j"], ")"}], "1"]}]},
          {
           RowBox[{
            SubscriptBox["x", "2"], 
            UnderscriptBox["<", "D"], 
            SubscriptBox[
             RowBox[{"(", 
              SubscriptBox["A", "j"], ")"}], "2"]}]},
          {
           RowBox[{
            RowBox[{
             SubscriptBox["y", "1"], 
             UnderscriptBox["+", "D"], 
             SubscriptBox["x", "2"]}], 
            UnderscriptBox["\[LessEqual]", "D"], 
            RowBox[{"deg", "[", 
             RowBox[{"lcm", "[", 
              RowBox[{"x", ",", 
               SubscriptBox["A", "j"]}], "]"}], "]"}]}]},
          {
           RowBox[{
            RowBox[{
             SubscriptBox["y", "1"], 
             UnderscriptBox["+", "D"], 
             SubscriptBox["x", "2"]}], 
            UnderscriptBox["\[LessEqual]", "D"], 
            RowBox[{"deg", "[", 
             RowBox[{"lcm", "[", 
              RowBox[{
               SubscriptBox["A", "j"], ",", "y"}], "]"}], "]"}]}]}
         }]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit bivariate 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1572461042", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.60014820740983*^9, 3.600148241519262*^9}, {
   3.600148274081915*^9, 3.600148389956722*^9}, {3.600148820171261*^9, 
   3.600148827272071*^9}, {3.600151469727669*^9, 3.600151474744439*^9}, {
   3.600405417339228*^9, 3.600405428109575*^9}, {3.600405625346005*^9, 
   3.600405628204237*^9}, {3.600405714399906*^9, 3.600405724686401*^9}, {
   3.600409619652453*^9, 3.600409635556766*^9}, 3.600688347465829*^9, 
   3.612758389382903*^9},
 CellTags->{
  "ID:1572461042", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 3"},
 CellID->1572461042],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", "y", ",", "A"}], "]"}], "\[And]", 
    RowBox[{
     SubscriptBox["x", "1"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "1"]}], "\[And]", 
    RowBox[{
     SubscriptBox["y", "2"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["x", "2"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A", ",", "1"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit bivariate 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2046865956", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59826256218055*^9, 3.598262650100715*^9}, 
   3.598270628887717*^9, 3.600067214992343*^9, {3.612697057654262*^9, 
   3.612697105681789*^9}},
 CellTags->{
  "ID:2046865956", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 4"},
 CellID->2046865956],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"chainCrit", "[", 
     RowBox[{"x", ",", "y", ",", "A"}], "]"}], "\[And]", 
    RowBox[{"\[Not]", 
     RowBox[{"x", "\[VerticalBar]", "y"}]}], "\[And]", 
    RowBox[{"\[Not]", 
     RowBox[{"y", "\[VerticalBar]", "x"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit bivariate 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:917844397", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612520501585424*^9, 3.612520508767699*^9}, {
  3.612524385951644*^9, 3.612524413166591*^9}},
 CellTags->{
  "ID:917844397", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 5"},
 CellID->917844397],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", "y", ",", "A"}], "]"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[Or]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{"x", "\[VerticalBar]", "y"}]},
      {
       RowBox[{"y", "\[VerticalBar]", "x"}]},
      {
       RowBox[{"chainCritDiagonal", "[", 
        RowBox[{"x", ",", "y", ",", "A"}], "]"}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit bivariate)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:810037327", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599280601061648*^9, 3.599280670109322*^9}, {
   3.599280766298709*^9, 3.599280793348982*^9}, {3.599280827143442*^9, 
   3.59928083213403*^9}, {3.599280922938172*^9, 3.599281005278566*^9}, {
   3.599281146734771*^9, 3.599281154163102*^9}, {3.599281674478076*^9, 
   3.599281800777444*^9}, {3.599281892826056*^9, 3.599281900927132*^9}, {
   3.599373420654823*^9, 3.599373468337973*^9}, {3.599890391005658*^9, 
   3.599890408458549*^9}, {3.600063994308919*^9, 3.600064003153391*^9}, {
   3.601029114075683*^9, 3.601029172974543*^9}, {3.601029247632683*^9, 
   3.601029341755698*^9}, {3.601029411458361*^9, 3.601029413444835*^9}, {
   3.601029457088123*^9, 3.601029459966897*^9}, {3.601029517123304*^9, 
   3.601029518284217*^9}, {3.601029582708264*^9, 3.601029701408114*^9}, {
   3.60483594624284*^9, 3.604835947582091*^9}, {3.605597966356155*^9, 
   3.605597969272574*^9}, 3.606468932161004*^9, {3.60646901228062*^9, 
   3.606469044769605*^9}, {3.606474972486705*^9, 3.606474978785237*^9}, {
   3.612517815570393*^9, 3.612517820274622*^9}},
 CellTags->{
  "ID:810037327", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate"},
 CellID->810037327],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", "2"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    RowBox[{
     RowBox[{"chainCrit", "[", 
      RowBox[{"x", ",", "y", ",", "A"}], "]"}], 
     TagBox["\[DoubleLeftRightArrow]",
      Identity,
      SyntaxForm->"a\[Implies]b"], 
     RowBox[{"\[Or]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"x", "\[VerticalBar]", "y"}]},
         {
          RowBox[{"y", "\[VerticalBar]", "x"}]},
         {
          RowBox[{"chainCritDiagonal", "[", 
           RowBox[{"x", ",", "y", ",", "A", ",", "k"}], "]"}]},
         {
          RowBox[{"chainCritDiagonal", "[", 
           RowBox[{"x", ",", "y", ",", "A", ",", 
            SuperscriptBox["k", "-"]}], "]"}]}
        }]}]}]}],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit bivariate I)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2111681775", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613799588515583*^9, 3.613799620567419*^9}},
 CellTags->{
  "ID:2111681775", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate I"},
 CellID->2111681775],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->311506677]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1706489092],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit bivariate 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit bivariate 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
               "\[Implies]", 
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], "]"}], 
               "\[Implies]", 
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1690170957"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1690170957-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1690170957-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1690170957", "Proof|ID:1690170957-1"},
 CellID->2085296512],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1690170957-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937057806163*^9},
 CellID->346853541]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit bivariate 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit bivariate 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:271680825"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p271680825-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p271680825-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:271680825", "Proof|ID:271680825-1"},
 CellID->1844198183],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p271680825-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937061973369*^9},
 CellID->1602314431]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit bivariate 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit bivariate 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["<", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["<", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}]}], 
                    "\[Or]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["<", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["<", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}]}], 
                    "\[Or]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1572461042"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1572461042-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1572461042-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1572461042", "Proof|ID:1572461042-1"},
 CellID->1050587621],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1572461042-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->966276044]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit bivariate 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit bivariate 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["<", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["<", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["<", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "1"], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}], 
                   UnderscriptBox["<", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", "1"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2046865956"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2046865956-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2046865956-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2046865956", "Proof|ID:2046865956-1"},
 CellID->1535230662],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2046865956-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937069077264*^9},
 CellID->155732320]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit bivariate 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit bivariate 5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
               RowBox[{"chainCritDiagonal", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:917844397"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p917844397-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p917844397-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:917844397", "Proof|ID:917844397-1"},
 CellID->665088749],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p917844397-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937073068794*^9},
 CellID->537828057]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit bivariate)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit bivariate)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"(", 
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
                   RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Or]", 
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                 ")"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"(", 
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
                   RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Or]", 
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                 ")"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:810037327"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p810037327-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p810037327-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:810037327", "Proof|ID:810037327-1"},
 CellID->1607548979],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p810037327-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937076637316*^9},
 CellID->1719284110]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit bivariate I)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit bivariate I)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Or]", 
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}]}], 
                   ")"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"chainCrit", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Or]", 
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}]}], 
                   ")"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2111681775"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2111681775-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2111681775-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2111681775", "Proof|ID:2111681775-1"},
 CellID->215040822],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2111681775-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937080348314*^9},
 CellID->1969341009]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Theorems", "Section",
 CellChangeTimes->{{3.596372061906974*^9, 3.596372065397917*^9}, {
   3.599812565923628*^9, 3.599812567005927*^9}, 3.605435275026304*^9, {
   3.6138129088728228`*^9, 3.613812909218138*^9}, 3.613885819000445*^9},
 CellTags->"sect::Theorems",
 CellID->1928916681],

Cell[TextData[{
 "This section contains some very important theorms needed for proving the \
three main theorms of section ",
 ButtonBox["Main Results",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::MainResults"],
 ". Since they are of significance also in their own right, they are not \
included in the section about the lemmata on specific notions, but put \
separately.\nHowever, again we distinguish between theorms that hold in the \
general case of arbitrary ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " and those that only hold if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 ", but in addition to that we also dedicate a separate subsection to some \
new function ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.604041153981664*^9, 3.604041394349798*^9}, {
  3.604045766834079*^9, 3.604045785621736*^9}, {3.604045821020906*^9, 
  3.604045866329819*^9}, {3.604046516513538*^9, 3.604046580723011*^9}, {
  3.604046641762563*^9, 3.604046642558337*^9}, {3.605439615726791*^9, 
  3.605439618498282*^9}, {3.615181939566492*^9, 3.615181952772684*^9}, {
  3.615182025770862*^9, 3.615182051452964*^9}},
 CellID->1182788500],

Cell[TextData[{
 "Apart from subsection ",
 ButtonBox["M*[A]",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::M*"],
 " below, we are now done with all the basic lemmata that are mostly \
concerned with the relations between \[OpenCurlyDoubleQuote]new\
\[CloseCurlyDoubleQuote] notions (e.g. ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 ") and \[OpenCurlyDoubleQuote]old\[CloseCurlyDoubleQuote] notions (e.g. ",
 Cell[BoxData[
  FormBox[
   TagBox["\:293a",
    Identity,
    SyntaxForm->"a*b"], TraditionalForm]]],
 "), which are crucial in the subsequent proofs of the more advanced \
theorems: Having those relations at hand (by means of formulas) allows us to \
do most parts of these proofs by straightforward rewriting techniques that do \
not require a lot of user interaction."
}], "Text",
 CellChangeTimes->{{3.605439027112185*^9, 3.605439040151144*^9}, {
  3.605439076069877*^9, 3.605439204849295*^9}, {3.605439392407854*^9, 
  3.6054395794341*^9}, {3.61518206563526*^9, 3.615182117337081*^9}},
 CellID->976970424],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   SuperscriptBox["M", "*"], "[", "A", "]"}], 
  TraditionalForm]]]], "Subsection",
 CellChangeTimes->{{3.604045700695384*^9, 3.604045706216661*^9}, {
   3.604046504418922*^9, 3.604046507435342*^9}, {3.613194731080368*^9, 
   3.613194731449261*^9}, 3.613812915153608*^9},
 CellTags->"sect::M*",
 CellID->391063855],

Cell[TextData[{
 "First of all we want to establish the link between the definition of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " as it is done in formula ",
 ButtonBox["(maximum degree of admissible pairs)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:640444143"],
 ", i.e. where the two indices might be equal, and the other (more usual) \
definition of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " (or, equivalently, pairs for which the chain criterion holds) in \
Buchberger\[CloseCurlyQuote]s algorithm ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ", where the two indices must never be equal.\n",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " is the function defined using ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 " without the constraint on the indices being unequal, and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}], TraditionalForm]]],
 " is its analogue where the indices have to be unequal, as defined below.\n\n\
Therefore, let ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " be a tuple of exponent vectors:"
}], "Text",
 CellChangeTimes->{{3.604044643312318*^9, 3.604044988466062*^9}, {
  3.604046596434764*^9, 3.604046615838505*^9}, {3.604047558132737*^9, 
  3.604047754799265*^9}, {3.604047813918387*^9, 3.604047826012426*^9}, {
  3.604052501921089*^9, 3.60405253646247*^9}, {3.604052587276354*^9, 
  3.604052632822396*^9}, {3.604835973443503*^9, 3.604836026897605*^9}, {
  3.605598007241839*^9, 3.605598016732382*^9}, {3.605598058997836*^9, 
  3.605598113051154*^9}, {3.615182217452797*^9, 3.615182259769317*^9}},
 CellID->457008407],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExpVectorTuple", "[", "A", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.604046742535876*^9, 3.60404675456358*^9}, {
  3.604053128454892*^9, 3.604053128603556*^9}, {3.605598121656811*^9, 
  3.605598126140853*^9}},
 CellID->1760611200],

Cell[TextData[{
 "Below comes the definition of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}], TraditionalForm]]],
 ". As for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " there are two versions, one binary and one unary, and in fact the only \
difference between ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " is that the pairs of elements of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " that are considered in ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}], TraditionalForm]]],
 " must have unequal indices. Note, however, that the elements themselves \
might still be equal!"
}], "Text",
 CellChangeTimes->{{3.604052675729518*^9, 3.604052794706757*^9}, {
  3.604053026254339*^9, 3.604053026729693*^9}, {3.60559813158609*^9, 
  3.60559813629521*^9}},
 CellID->1013604180],

Cell["", "OpenEnvironment",
 CellID->1145076435],

Cell[CellGroupData[{

Cell[TextData[{
 "Definition: ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "*"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.604046679843281*^9, 3.604046685633793*^9}, {
  3.604052934322038*^9, 3.604052934322137*^9}},
 CellID->800620613],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SuperscriptBox["M", "*"], "[", 
     RowBox[{"i", ",", "A"}], "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     UnderscriptBox["max", "D"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{"deg", "[", 
        RowBox[{"lcm", "[", 
         RowBox[{
          SubscriptBox["A", "i"], ",", 
          SubscriptBox["A", "j"]}], "]"}], "]"}], 
       UnderscriptBox["|", 
        RowBox[{
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
       RowBox[{
        RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
        RowBox[{"chainCrit", "[", 
         RowBox[{
          SubscriptBox["A", "i"], ",", 
          SubscriptBox["A", "j"], ",", "A"}], "]"}]}]}], 
      "\[RightAngleBracket]"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M* binary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2973275", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.604046785549076*^9, 3.604046839582149*^9}, {
   3.604053200869017*^9, 3.604053204358087*^9}, 3.604053261691595*^9, {
   3.604053315056565*^9, 3.604053319851742*^9}, {3.604836040967275*^9, 
   3.604836066028746*^9}, {3.604993579961453*^9, 3.604993581875498*^9}, {
   3.605598145468502*^9, 3.605598149656389*^9}, {3.605598190692953*^9, 
   3.605598192909882*^9}, {3.605598311279195*^9, 3.605598358498991*^9}, {
   3.612847339366419*^9, 3.612847339769441*^9}},
 CellTags->{
  "ID:2973275", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M* binary"},
 CellID->2973275],

Cell[BoxData[
 RowBox[{
  RowBox[{
   SuperscriptBox["M", "*"], "[", "A", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["max", "D"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      SuperscriptBox["M", "*"], "[", 
      RowBox[{"i", ",", "A"}], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M* unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1258784171", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.604046857939352*^9, 3.60404689107252*^9}, {
  3.604053208449621*^9, 3.604053214916298*^9}, {3.604053323356438*^9, 
  3.604053332250682*^9}, {3.604993585077984*^9, 3.604993586890466*^9}, {
  3.605598151239332*^9, 3.605598152606369*^9}},
 CellTags->{
  "ID:1258784171", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M* unary"},
 CellID->1258784171],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->83476417]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2104356565],

Cell[TextData[{
 "The following lemma is in the same spirit as lemma \
\[OpenCurlyDoubleQuote]Properties of ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] in section ",
 ButtonBox["Lemmata on Specific Notions / General Case",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataSpecificGeneral"],
 "; However, unlike there, the following lemma only holds if ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " contains at least two elements:"
}], "Text",
 CellChangeTimes->{{3.604052875479042*^9, 3.604052958935782*^9}, {
  3.612848269892556*^9, 3.612848289405156*^9}, {3.615182312353739*^9, 
  3.615182376348169*^9}, {3.615182939342313*^9, 3.615182963903327*^9}},
 CellID->1285990268],

Cell[BoxData[
 RowBox[{
  RowBox[{"2", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.612848294763469*^9, 3.612848301226974*^9}},
 CellID->927538807],

Cell["", "OpenEnvironment",
 CellID->1868882626],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "*"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.604047112824473*^9, 3.604047122024182*^9}, {
  3.604052927194387*^9, 3.604052927926519*^9}},
 CellID->2108232500],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.604047338923003*^9, 3.604047348816677*^9}},
 CellID->1941261269],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{"i", "\[NotEqual]", "l"}]], 
  RowBox[{"chainCrit", "[", 
   RowBox[{
    SubscriptBox["A", "i"], ",", 
    SubscriptBox["A", "l"], ",", "A"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit holds for distinct indices)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1549582561", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.604047300501734*^9, 3.604047363402215*^9}, {
  3.604047398203909*^9, 3.604047429192339*^9}, {3.604836088773266*^9, 
  3.604836096884074*^9}, {3.605598422493442*^9, 3.605598436598476*^9}, {
  3.612847341866354*^9, 3.612847342233838*^9}, {3.612864046952406*^9, 
  3.612864049784126*^9}},
 CellTags->{
  "ID:1549582561", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit holds for distinct indices"},
 CellID->1549582561],

Cell[BoxData[
 RowBox[{"1", 
  UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
  RowBox[{"\[LeftBracketingBar]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"deg", "[", 
      RowBox[{"lcm", "[", 
       RowBox[{
        SubscriptBox["A", "i"], ",", 
        SubscriptBox["A", "j"]}], "]"}], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{
      RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
      RowBox[{"chainCrit", "[", 
       RowBox[{
        SubscriptBox["A", "i"], ",", 
        SubscriptBox["A", "j"], ",", "A"}], "]"}]}]}], 
    "\[RightAngleBracket]"}], 
   "\[RightBracketingBar]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(74)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1966722449", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612856923032386*^9, 3.612856950632509*^9}},
 CellTags->{
  "ID:1966722449", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "74"},
 CellID->1966722449],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[SquareSubsetEqual]", "D"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"deg", "[", 
      RowBox[{"lcm", "[", 
       RowBox[{
        SubscriptBox["A", "i"], ",", 
        SubscriptBox["A", "j"]}], "]"}], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{
      RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
      RowBox[{"chainCrit", "[", 
       RowBox[{
        SubscriptBox["A", "i"], ",", 
        SubscriptBox["A", "j"], ",", "A"}], "]"}]}]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(75)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:164254989", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612856966160465*^9, 3.612856978241077*^9}},
 CellTags->{
  "ID:164254989", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "75"},
 CellID->164254989],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{
    SuperscriptBox["M", "*"], "[", 
    RowBox[{"i", ",", "A"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of M* (binary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:290364666", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.604047189592442*^9, 3.604047215344882*^9}, 
   3.604047353945165*^9, {3.604053378214055*^9, 3.604053378733757*^9}, {
   3.605598440464774*^9, 3.605598440650546*^9}},
 CellTags->{
  "ID:290364666", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M* (binary)"},
 CellID->290364666],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", 
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(type of M* (unary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1063863080", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.604047250231873*^9, 3.604047252853883*^9}, {
  3.604053381319026*^9, 3.604053381797543*^9}, {3.605598442051859*^9, 
  3.605598442309759*^9}},
 CellTags->{
  "ID:1063863080", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M* (unary)"},
 CellID->1063863080],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1097467670]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1183566659],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit holds for distinct indices)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit holds for distinct indices)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                  StyleBox["l", "ExpressionVariable"]}]], 
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["l", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                  StyleBox["l", "ExpressionVariable"]}]], 
               RowBox[{"chainCrit", "[", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["l", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1549582561"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1549582561-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1549582561-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1549582561", "Proof|ID:1549582561-1"},
 CellID->1962722847],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1549582561-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937100555303*^9},
 CellID->2031482966]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(74)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(74)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{"1", 
               UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["j", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{"1", 
               UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["j", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1966722449"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1966722449-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1966722449-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1966722449", "Proof|ID:1966722449-1"},
 CellID->938794098],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1966722449-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937108123202*^9},
 CellID->49039736]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(75)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(75)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["j", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 UnderscriptBox["\[SquareSubsetEqual]", "D"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["j", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}]}], 
                 "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:164254989"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p164254989-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p164254989-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:164254989", "Proof|ID:164254989-1"},
 CellID->895811688],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p164254989-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937112035136*^9},
 CellID->774539108]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of M* (binary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of M* (binary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SuperscriptBox["M", "*"], 
                   TagBox[")", "AutoParentheses"]}], "[", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "D"], "[", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SuperscriptBox["M", "*"], 
                   TagBox[")", "AutoParentheses"]}], "[", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:290364666"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p290364666-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p290364666-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:290364666", "Proof|ID:290364666-1"},
 CellID->1493869216],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p290364666-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937115722979*^9},
 CellID->1990230967]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(type of M* (unary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(type of M* (unary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SuperscriptBox["M", "*"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "D"], "[", 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SuperscriptBox["M", "*"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1063863080"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1063863080-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1063863080-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1063863080", "Proof|ID:1063863080-1"},
 CellID->391529240],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1063863080-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937119434819*^9},
 CellID->690333880]
}, Closed]],

Cell[TextData[{
 "The reason why we have to state and prove formula ",
 ButtonBox["(chainCrit holds for distinct indices)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1549582561"],
 " is the following: Although it looks very similar to formula ",
 ButtonBox["(chainCrit holds for some element)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1053941118"],
 " and can also be proved in exactly the same fashion, they are different. \
Formula ",
 ButtonBox["(chainCrit holds for some element)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1053941118"],
 ", on the one hand, holds if an arbitrary object ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " is taken as the first argument of ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 ", whereas formula ",
 ButtonBox["(chainCrit holds for distinct indices)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1549582561"],
 " requires an element of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " to be the first argument of ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 ", so to say. Thus, ",
 ButtonBox["(chainCrit holds for some element)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1053941118"],
 " is more general than ",
 ButtonBox["(chainCrit holds for distinct indices)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1549582561"],
 " in this respect.\nOn the other hand, however, ",
 ButtonBox["(chainCrit holds for distinct indices)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1549582561"],
 " can be regarded to be more general than ",
 ButtonBox["(chainCrit holds for some element)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1053941118"],
 " because in ",
 ButtonBox["(chainCrit holds for some element)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1053941118"],
 " the object ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " typically is no member of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " (although it might be), and hence is not included in the last argument of ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 ". In ",
 ButtonBox["(chainCrit holds for distinct indices)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1549582561"],
 ", both arguments of ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 " are in there.\nOf course, with some effort and additional knowledge about \
permutations of tuples etc. one could in principle prove",
 " ",
 ButtonBox["(chainCrit holds for distinct indices)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1549582561"],
 " ",
 "directly from ",
 ButtonBox["(chainCrit holds for some element).",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1053941118"]
}], "Text",
 CellChangeTimes->{{3.604055320860594*^9, 3.604055483609233*^9}, {
  3.604055528663722*^9, 3.604055893432886*^9}, {3.604227715365214*^9, 
  3.604227716090736*^9}, {3.605598486797372*^9, 3.605598558522558*^9}, {
  3.615182579976371*^9, 3.615182616334096*^9}, {3.615182710152936*^9, 
  3.615182710181573*^9}, {3.615182754992964*^9, 3.615182851470761*^9}},
 CellID->1583197414],

Cell[TextData[{
 "The next lemma is in exactly the same spirit as lemma \
\[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "\[LessEqual]", "u"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", "\[LessEqual]", "M"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] in section ",
 ButtonBox["Lemmata on Specific Notions / General Case",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataSpecificGeneral"],
 ":"
}], "Text",
 CellChangeTimes->{{3.615182892673886*^9, 3.615182924104667*^9}, {
  3.615182971661105*^9, 3.615182973874603*^9}},
 CellID->809498243],

Cell["", "OpenEnvironment",
 CellID->378652997],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "*"], "\[LessEqual]", "u"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", "\[LessEqual]", 
    SuperscriptBox["M", "*"]}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.612847849383792*^9, 3.612847864355224*^9}},
 CellID->1036416147],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", "u", "]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612847891302412*^9, 3.612847898042949*^9}},
 CellID->255426543],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612767486028775*^9, 3.612767498432565*^9}, {
  3.612847944637742*^9, 3.61284796899503*^9}},
 CellID->1342888001],

Cell[BoxData[
 RowBox[{
  RowBox[{"u", 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"r", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{"u", 
    UnderscriptBox["\[LessEqual]", "D"], 
    RowBox[{
     SuperscriptBox["M", "*"], "[", 
     RowBox[{"r", ",", "A"}], "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LessEqual]M* (unary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1680577066", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600148644275676*^9, 3.600148717385941*^9}, {
   3.60014875851393*^9, 3.600148849997594*^9}, {3.600409639616649*^9, 
   3.600409663874679*^9}, 3.600688366509062*^9, {3.612767361373108*^9, 
   3.612767397598277*^9}, {3.61276743295501*^9, 3.612767453144141*^9}, {
   3.612767502993384*^9, 3.612767504023866*^9}, {3.612767537557185*^9, 
   3.612767537890943*^9}, {3.612847953812527*^9, 3.612847962234629*^9}},
 CellTags->{
  "ID:1680577066", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M* (unary)"},
 CellID->1680577066],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"r", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    RowBox[{
     SuperscriptBox["M", "*"], "[", 
     RowBox[{"r", ",", "A"}], "]"}], 
    UnderscriptBox["\[LessEqual]", "D"], "u"}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M*\[LessEqual] (unary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1866021817", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.60014820740983*^9, 3.600148241519262*^9}, {
   3.600148274081915*^9, 3.600148389956722*^9}, {3.600148820171261*^9, 
   3.600148827272071*^9}, {3.600148938757218*^9, 3.600148941969407*^9}, {
   3.60014906006257*^9, 3.600149061755045*^9}, {3.600405746880696*^9, 
   3.600405753217857*^9}, {3.60040967359848*^9, 3.600409688162742*^9}, 
   3.600688352024196*^9, 3.61276736262328*^9, {3.612767513907566*^9, 
   3.61276754927719*^9}, {3.612847974728469*^9, 3.612847982979097*^9}},
 CellTags->{
  "ID:1866021817", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M*\[LessEqual] (unary)"},
 CellID->1866021817],

Cell[BoxData[
 RowBox[{
  RowBox[{"u", 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{
    SuperscriptBox["M", "*"], "[", 
    RowBox[{"i", ",", "A"}], "]"}]}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"s", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{"i", "\[NotEqual]", "s"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"chainCrit", "[", 
      RowBox[{
       SubscriptBox["A", "i"], ",", 
       SubscriptBox["A", "s"], ",", "A"}], "]"}], "\[And]", 
     RowBox[{"u", 
      UnderscriptBox["\[LessEqual]", "D"], 
      RowBox[{"deg", "[", 
       RowBox[{"lcm", "[", 
        RowBox[{
         SubscriptBox["A", "i"], ",", 
         SubscriptBox["A", "s"]}], "]"}], "]"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LessEqual]M* (binary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1272742036", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600148644275676*^9, 3.600148717385941*^9}, {
   3.60014875851393*^9, 3.600148849997594*^9}, {3.600149073969659*^9, 
   3.600149085676266*^9}, {3.600409695019163*^9, 3.600409712869742*^9}, 
   3.600688369415438*^9, 3.612767364076105*^9, {3.612767554093445*^9, 
   3.612767639781588*^9}, {3.61276791056231*^9, 3.612767913393056*^9}, {
   3.612847988676881*^9, 3.612848002040864*^9}, {3.612848126733596*^9, 
   3.61284816166502*^9}},
 CellTags->{
  "ID:1272742036", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M* (binary)"},
 CellID->1272742036],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    SuperscriptBox["M", "*"], "[", 
    RowBox[{"i", ",", "A"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], "u"}], 
  TagBox["\[DoubleLeftRightArrow]",
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"s", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
    RowBox[{"i", "\[NotEqual]", "s"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"chainCrit", "[", 
      RowBox[{
       SubscriptBox["A", "i"], ",", 
       SubscriptBox["A", "s"], ",", "A"}], "]"}], "\[Implies]", 
     RowBox[{
      RowBox[{"deg", "[", 
       RowBox[{"lcm", "[", 
        RowBox[{
         SubscriptBox["A", "i"], ",", 
         SubscriptBox["A", "s"]}], "]"}], "]"}], 
      UnderscriptBox["\[LessEqual]", "D"], "u"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M*\[LessEqual] (binary))", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:462890005", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612767643468274*^9, 3.612767682842289*^9}, {
  3.61276791504931*^9, 3.612767918615729*^9}, {3.612848005544075*^9, 
  3.612848015117038*^9}, {3.612848175560492*^9, 3.61284818659775*^9}, {
  3.61285039183959*^9, 3.612850401726694*^9}},
 CellTags->{
  "ID:462890005", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M*\[LessEqual] (binary)"},
 CellID->462890005],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->982813600]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->934500559],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LessEqual]M* (unary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LessEqual]M* (unary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["u", "ExpressionVariable"], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["u", "ExpressionVariable"], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1680577066"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1680577066-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1680577066-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1680577066", "Proof|ID:1680577066-1"},
 CellID->631558205],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1680577066-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937124842518*^9},
 CellID->503475645]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M*\[LessEqual] (unary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M*\[LessEqual] (unary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox["u", "ExpressionVariable"]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 StyleBox["u", "ExpressionVariable"]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox[
                   "u", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1866021817"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1866021817-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1866021817-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1866021817", "Proof|ID:1866021817-1"},
 CellID->860652560],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1866021817-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937129538181*^9},
 CellID->917867914]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LessEqual]M* (binary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LessEqual]M* (binary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    StyleBox["u", "ExpressionVariable"], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    StyleBox["u", "ExpressionVariable"], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1272742036"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1272742036-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1272742036-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1272742036", "Proof|ID:1272742036-1"},
 CellID->135701756],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1272742036-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937133705709*^9},
 CellID->1766556048]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M*\[LessEqual] (binary))\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M*\[LessEqual] (binary))\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox[
                    "u", "ExpressionVariable"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox["M", "*"], 
                    TagBox[")", "AutoParentheses"]}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   StyleBox["u", "ExpressionVariable"]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    StyleBox[
                    "u", "ExpressionVariable"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:462890005"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p462890005-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p462890005-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:462890005", "Proof|ID:462890005-1"},
 CellID->829755125],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p462890005-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.61493713703366*^9},
 CellID->421237545]
}, Closed]],

Cell[TextData[{
 "The following theorem (in particular formula ",
 ButtonBox["(M=M*)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:872843574"],
 ") plays the central role in this subsection. As can be seen from it, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}], TraditionalForm]]],
 " always have the same value, meaning that all the statements of section ",
 ButtonBox["Main Results",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::MainResults"],
 " would also be true if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " was replaced by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}], TraditionalForm]]],
 ", given that ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " contains at least two elements."
}], "Text",
 CellChangeTimes->{{3.604052542700734*^9, 3.604052574649884*^9}, 
   3.612847841044133*^9, {3.612848686332077*^9, 3.612848745929811*^9}, {
   3.612862662034638*^9, 3.612862662256517*^9}, {3.615183348846089*^9, 
   3.615183424031091*^9}, {3.615183591461499*^9, 3.615183606106134*^9}},
 CellID->1211263146],

Cell["", "OpenEnvironment",
 CellChangeTimes->{3.600758233382262*^9},
 CellID->1043926589],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " equals ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}], TraditionalForm]], "Input"],
 "\[CloseCurlyDoubleQuote]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.59637207087823*^9, 3.596372088473265*^9}, {
   3.596372278374362*^9, 3.596372280551172*^9}, {3.598694956355545*^9, 
   3.598694959118789*^9}, {3.599885609224245*^9, 3.599885614314564*^9}, {
   3.600755227695421*^9, 3.600755229380012*^9}, 3.600758233382553*^9, {
   3.604034532226787*^9, 3.604034535114983*^9}, {3.604047884161863*^9, 
   3.604047919377103*^9}},
 CellID->1274870949],

Cell[BoxData[
 RowBox[{
  RowBox[{"M", "[", "A", "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   SuperscriptBox["M", "*"], "[", "A", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\[LessEqual]M*)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:929561290", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612846515680337*^9, 3.612846524274778*^9}},
 CellTags->{
  "ID:929561290", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual]M*"},
 CellID->929561290],

Cell[BoxData[
 RowBox[{
  RowBox[{
   SuperscriptBox["M", "*"], "[", "A", "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{"M", "[", "A", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M*\[LessEqual]M)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:795795393", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.612846528871564*^9, 3.61284653531279*^9}},
 CellTags->{
  "ID:795795393", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M*\[LessEqual]M"},
 CellID->795795393],

Cell[BoxData[
 RowBox[{
  RowBox[{"M", "[", "A", "]"}], "\[Equal]", 
  RowBox[{
   SuperscriptBox["M", "*"], "[", "A", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M=M*)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:872843574", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596372187671416*^9, 3.596372253551077*^9}, {
   3.598252912092934*^9, 3.598252914753788*^9}, {3.599373730492249*^9, 
   3.599373751141782*^9}, {3.599386067761938*^9, 3.599386085260058*^9}, 
   3.600758233383315*^9, {3.603181323303326*^9, 3.603181330760711*^9}, 
   3.604034504671064*^9, {3.604047028592247*^9, 3.604047033245926*^9}, {
   3.604053384639008*^9, 3.60405338504832*^9}, {3.605598568452229*^9, 
   3.605598569634252*^9}, 3.612846434904177*^9},
 CellTags->{
  "ID:872843574", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M=M*"},
 CellID->872843574],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellChangeTimes->{3.600758233383519*^9},
 CellID->325406659]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellChangeTimes->{3.600758233383582*^9},
 CellID->692941230],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\[LessEqual]M*)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\[LessEqual]M*)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SuperscriptBox["M", "*"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SuperscriptBox["M", "*"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:929561290"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p929561290-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p929561290-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:929561290", "Proof|ID:929561290-1"},
 CellID->286309708],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p929561290-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {50., 56.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937141537473*^9},
 CellID->2072324147]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M*\[LessEqual]M)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M*\[LessEqual]M)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SuperscriptBox["M", "*"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SuperscriptBox["M", "*"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:795795393"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p795795393-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p795795393-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:795795393", "Proof|ID:795795393-1"},
 CellID->1883422493],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p795795393-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937146897467*^9},
 CellID->1637224044]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M=M*)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M=M*)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SuperscriptBox["M", "*"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"2", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SuperscriptBox["M", "*"], 
                 TagBox[")", "AutoParentheses"]}], "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:872843574"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p872843574-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p872843574-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:872843574", "Proof|ID:872843574-1"},
 CellID->1077588120],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p872843574-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937150649165*^9},
 CellID->1552594980]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["General Case", "Subsection",
 CellChangeTimes->{{3.599385866169655*^9, 3.599385868591121*^9}, {
   3.612846329600006*^9, 3.612846330298451*^9}, 3.613812913605022*^9},
 CellTags->"sect::TheoremsGeneral",
 CellID->1847669776],

Cell[TextData[{
 "We now continue with some other important theorems in the general case, \
i.e. ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " is not restricted to 2.\nLet ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " as usual be a ",
 StyleBox["non-empty",
  FontSlant->"Italic"],
 " tuple of exponent vectors:"
}], "Text",
 CellChangeTimes->{{3.604041661557845*^9, 3.604041749548723*^9}, {
  3.604045724610373*^9, 3.604045727409038*^9}, {3.604055933204168*^9, 
  3.604055963933535*^9}, {3.605598592609788*^9, 3.605598643804934*^9}, {
  3.615183627355078*^9, 3.615183628620897*^9}},
 CellID->2105593863],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isExpVectorTuple", "[", "A", "]"}]], 
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.596372114271475*^9, 3.59637214156681*^9}, {
   3.598252779840109*^9, 3.598252779995368*^9}, 3.59832976766527*^9, {
   3.599386063658651*^9, 3.599386063990499*^9}, {3.599388842688274*^9, 
   3.599388847151938*^9}, {3.600065312590323*^9, 3.600065316728138*^9}, {
   3.603181285915295*^9, 3.603181288273784*^9}, {3.603693428975579*^9, 
   3.603693431285988*^9}, {3.605598651105867*^9, 3.605598654694686*^9}},
 CellTags->"???",
 CellID->560938714],

Cell[TextData[{
 "The following corollary is completely trivial if in the definition of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " one also considers the degrees of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "[", 
    RowBox[{
     SubscriptBox["A", "i"], ",", 
     SubscriptBox["A", "i"]}], "]"}], TraditionalForm]]],
 ", i.e. elements at the same position in ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " (as it is the case in this report). However, theorem \
\[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " equals ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "*"], "[", "A", "]"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] guarantees validity even if additionally the \
requirement for the two indices being distinct is imposed."
}], "Text",
 CellChangeTimes->{{3.599384880208054*^9, 3.599384952397325*^9}, {
   3.599385038903413*^9, 3.599385050251889*^9}, {3.59938612078368*^9, 
   3.599386136062413*^9}, {3.599885638552988*^9, 3.599885697931137*^9}, {
   3.599887264498942*^9, 3.599887287936167*^9}, {3.600757981916449*^9, 
   3.600757982916512*^9}, 3.60075823338511*^9, {3.604041501395907*^9, 
   3.604041540096427*^9}, {3.604055977963639*^9, 3.604056013547891*^9}, {
   3.604836157280508*^9, 3.60483620594706*^9}, {3.604836239969153*^9, 
   3.604836273953945*^9}, {3.615183644721765*^9, 3.615183665090118*^9}},
 CellID->72362671],

Cell["", "OpenEnvironment",
 CellChangeTimes->{3.600758233385224*^9},
 CellID->1437260198],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"maxdeg", "[", "A", "]"}], "\[LessEqual]", 
    RowBox[{"M", "[", "A", "]"}]}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599384233414378*^9, 3.59938425045447*^9}, {
   3.599384993288948*^9, 3.59938499481993*^9}, {3.599385062866613*^9, 
   3.59938507600866*^9}, 3.600758233385938*^9, {3.60405618464114*^9, 
   3.604056185709853*^9}, {3.613189642738586*^9, 3.613189643772126*^9}},
 CellID->852970621],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   RowBox[{"deg", "[", 
    SubscriptBox["A", "l"], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"M", "[", "A", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(76)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:642765392", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61318969489363*^9, 3.613189722201644*^9}},
 CellTags->{
  "ID:642765392", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "76"},
 CellID->642765392],

Cell[BoxData[
 RowBox[{
  RowBox[{"maxdeg", "[", "A", "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{"M", "[", "A", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(maxdeg\[LessEqual]M)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:592779258", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599384254328952*^9, 3.599384272085472*^9}, 
   3.60075823338634*^9, {3.605598662575745*^9, 3.605598663972577*^9}},
 CellTags->{
  "ID:592779258", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maxdeg\[LessEqual]M"},
 CellID->592779258],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellChangeTimes->{3.600758233386499*^9},
 CellID->459797040]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellChangeTimes->{3.600758233386559*^9},
 CellID->52345895],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(76)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(76)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"deg", "[", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["l", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"deg", "[", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 SubscriptBox[
                  StyleBox["A", "ExpressionVariable"], 
                  StyleBox["l", "ExpressionVariable"]], 
                 TagBox[")", "AutoParentheses"]}], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:642765392"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p642765392-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p642765392-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:642765392", "Proof|ID:642765392-1"},
 CellID->1075234758],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p642765392-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937162640277*^9},
 CellID->1622776262]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(maxdeg\[LessEqual]M)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(maxdeg\[LessEqual]M)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"maxdeg", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"maxdeg", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:592779258"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p592779258-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p592779258-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:592779258", "Proof|ID:592779258-1"},
 CellID->582763895],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p592779258-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.613189831044322*^9, 3.614937166744296*^9},
 CellID->1924730576]
}, Closed]],

Cell[TextData[{
 "The following theorem and corollary are needed for obtaining the bound ",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    RowBox[{"maxdeg", "[", "A", "]"}]}], TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.599904958319131*^9, 3.599904988101663*^9}, {
  3.60405632634405*^9, 3.604056334329266*^9}},
 CellID->703221276],

Cell["", "OpenEnvironment",
 CellID->527748213],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"L", "[", "A", "]"}], TraditionalForm]]],
 " is a multiple of all LCMs"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599900107837213*^9, 3.599900119963766*^9}, {
  3.599900183298295*^9, 3.599900193082077*^9}, {3.601025861188753*^9, 
  3.601025863815385*^9}},
 CellID->1751511295],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.612845274975768*^9, 3.612845281970974*^9}},
 CellID->1051492688],

Cell[BoxData[
 RowBox[{
  SubscriptBox["A", "i"], "\[VerticalBar]", 
  RowBox[{"L", "[", "A", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(77)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:691015215", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.600067228189778*^9, 3.600067298463982*^9}, {
  3.60042085759157*^9, 3.600420860734803*^9}, {3.61284528854674*^9, 
  3.612845299334045*^9}},
 CellTags->{
  "ID:691015215", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "77"},
 CellID->691015215],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   RowBox[{"lcm", "[", 
    RowBox[{
     SubscriptBox["A", "i"], ",", 
     SubscriptBox["A", "j"]}], "]"}], "\[VerticalBar]", 
   RowBox[{"L", "[", "A", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(78)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:402511110", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599900135555024*^9, 3.59990015539571*^9}, {
  3.599900200544446*^9, 3.599900242199953*^9}, {3.604836280368443*^9, 
  3.604836285690142*^9}, {3.605598668466977*^9, 3.605598672413557*^9}, {
  3.612845284763647*^9, 3.612845284984356*^9}},
 CellTags->{
  "ID:402511110", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "78"},
 CellID->402511110],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2092750413]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->938769538],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(77)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(77)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                StyleBox["A", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], 
               TagBox[")", "AutoParentheses"]}], "\[VerticalBar]", 
             RowBox[{"L", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                StyleBox["A", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], 
               TagBox[")", "AutoParentheses"]}], "\[VerticalBar]", 
             RowBox[{"L", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:691015215"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p691015215-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p691015215-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:691015215", "Proof|ID:691015215-1"},
 CellID->31238831],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p691015215-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937171208405*^9},
 CellID->2057268967]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(78)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(78)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"lcm", "[", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], ",", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}]}], "]"}], 
             "\[VerticalBar]", 
             RowBox[{"L", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], 
                 TagBox[")", "AutoParentheses"]}]}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"lcm", "[", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}], ",", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                   TagBox[")", "AutoParentheses"]}]}], "]"}], 
             "\[VerticalBar]", 
             RowBox[{"L", "[", 
               StyleBox["A", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:402511110"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p402511110-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p402511110-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:402511110", "Proof|ID:402511110-1"},
 CellID->1203109870],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p402511110-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937175096076*^9},
 CellID->1020066581]
}, Closed]],

Cell[TextData[{
 "Finally, this subsection is concluded with a simple corollary of formula ",
 ButtonBox["(78)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:402511110"],
 ":"
}], "Text",
 CellChangeTimes->{{3.604056527563752*^9, 3.604056559262777*^9}, {
  3.612845325859452*^9, 3.612845338271509*^9}, {3.612846314239598*^9, 
  3.612846314249751*^9}, {3.61319516566137*^9, 3.613195168684978*^9}, {
  3.613195506090375*^9, 3.613195514502379*^9}, {3.615785711308331*^9, 
  3.615785725555487*^9}},
 CellID->1406772823],

Cell["", "OpenEnvironment",
 CellID->1681684092],

Cell[CellGroupData[{

Cell["Corollary", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellID->1926603590],

Cell[BoxData[
 RowBox[{
  RowBox[{"M", "[", "A", "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{"deg", "[", 
   RowBox[{"L", "[", "A", "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(M\[LessEqual]L)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:341822661", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599900601604664*^9, 3.599900618490361*^9}, {
  3.599901264670568*^9, 3.599901264907847*^9}, {3.60559867981776*^9, 
  3.60559868223619*^9}},
 CellTags->{
  "ID:341822661", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual]L"},
 CellID->341822661],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1589678083]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->71356467],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(M\[LessEqual]L)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(M\[LessEqual]L)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"deg", "[", 
               RowBox[{"L", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpVectorTuple", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"1", 
             UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"M", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{"deg", "[", 
               RowBox[{"L", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:341822661"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p341822661-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p341822661-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:341822661", "Proof|ID:341822661-1"},
 CellID->1958581994],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p341822661-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937179511652*^9},
 CellID->1851915245]
}, Closed]],

Cell[TextData[{
 "Apparently, there are not many theorems we can prove for arbitrary ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ". This, however, is not surprising: Most of the important theorems just do \
not hold in the general case."
}], "Text",
 CellChangeTimes->{{3.615183803842208*^9, 3.615183870721519*^9}},
 CellID->2026828730]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Special Case ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]]
}], "Subsection",
 CellChangeTimes->{{3.599385847316572*^9, 3.599385853822335*^9}},
 CellTags->"sect::TheoremsSpecial",
 CellID->240164494],

Cell[TextData[{
 "Finally we turn to other important theorems that only hold in the case ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "2"}], TraditionalForm]]],
 ".\nLet ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " as usual be a ",
 StyleBox["non-empty",
  FontSlant->"Italic"],
 " tuple of exponent vectors, and let ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " be an exponent vector:"
}], "Text",
 CellChangeTimes->{{3.604041661557845*^9, 3.604041749548723*^9}, {
  3.604045724610373*^9, 3.604045727409038*^9}, {3.604055933204168*^9, 
  3.604055963933535*^9}, {3.60405663771598*^9, 3.604056700667701*^9}, {
  3.604057356528236*^9, 3.604057384595083*^9}, {3.605598702714526*^9, 
  3.605598724217063*^9}, {3.615188005184238*^9, 3.615188007127323*^9}},
 CellID->1113483777],

Cell[BoxData[
 RowBox[{
  RowBox[{"n", "\[Equal]", "2"}], "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.605598754300117*^9, 3.605598757344803*^9}},
 CellID->1550838377],

Cell["", "Text",
 CellID->1114476607],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isExpVectorTuple", "[", "A", "]"}]], 
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.596372114271475*^9, 3.59637214156681*^9}, {
   3.598252779840109*^9, 3.598252779995368*^9}, 3.59832976766527*^9, {
   3.599386063658651*^9, 3.599386063990499*^9}, {3.599388830749669*^9, 
   3.599388835943664*^9}, {3.600065302859017*^9, 3.600065307169814*^9}, {
   3.603181377359428*^9, 3.603181380630552*^9}, {3.604056674066639*^9, 
   3.604056676589355*^9}, {3.605598730377319*^9, 3.605598735958171*^9}},
 CellTags->"???",
 CellID->299478056],

Cell["", "Text",
 CellID->2116435651],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "x", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.604057345708286*^9, 3.604057352588092*^9}, {
  3.605598740108604*^9, 3.605598743458281*^9}},
 CellID->1547442960],

Cell[CellGroupData[{

Cell["Rectangular Region", "Subsubsection",
 CellChangeTimes->{{3.599385847316572*^9, 3.599385853822335*^9}, {
  3.606466922424492*^9, 3.6064669277125*^9}, {3.613196148457555*^9, 
  3.613196149081402*^9}, {3.613812904632773*^9, 3.613812904929047*^9}},
 CellID->679066213],

Cell[TextData[{
 "The first theorem in this subsection (in particular formula ",
 ButtonBox["(86)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:789654176"],
 ") already states one elementary result needed for proving that the sum ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " does not increase in Buchberger\[CloseCurlyQuote]s algorithm, if ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " is considered the tuple consisting of the exponent vectors of the leading \
monomials in the current basis.\nThe formulas in the lemma below might at \
first glance appear a bit complicated, but they are anyway only auxiliary \
results needed for proving formula ",
 ButtonBox["(86)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:789654176"],
 "."
}], "Text",
 CellChangeTimes->{{3.604056755933717*^9, 3.604056811056344*^9}, {
  3.604056900208633*^9, 3.60405694446583*^9}, {3.604057245781061*^9, 
  3.604057267459331*^9}, {3.604057758164433*^9, 3.604057778952591*^9}, {
  3.604057815216394*^9, 3.604057951350331*^9}, {3.606552593091217*^9, 
  3.606552610009646*^9}, {3.615188305614306*^9, 3.615188357637552*^9}, {
  3.615786316590817*^9, 3.615786321947124*^9}},
 CellID->1198933695],

Cell["", "OpenEnvironment",
 CellID->140872528],

Cell[CellGroupData[{

Cell["Theorem: Preparations for Rectangular Region", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.613103488633493*^9, 3.613103503761249*^9}},
 CellID->260198431],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "2"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.613104047519357*^9, 3.613104055255814*^9}},
 CellID->323697053],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExponentVector", "[", "y", "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.606548245573721*^9, 3.606548261445572*^9}, {
  3.60654830709384*^9, 3.606548313687865*^9}, {3.606552544032277*^9, 
  3.606552549475805*^9}, {3.606552661639689*^9, 3.606552666902264*^9}},
 CellID->1753587106],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", "y", ",", "A", ",", "k"}], "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.606548407943701*^9, 3.606548417432847*^9}, {
   3.613103597860123*^9, 3.613103603088411*^9}, 3.613104058308526*^9},
 CellID->1153504918],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"isExponentVector", "[", "z", "]"}]], 
   RowBox[{
    SubscriptBox["x", 
     SuperscriptBox["k", "-"]], 
    UnderscriptBox["\[LessEqual]", "D"], 
    SubscriptBox["z", 
     SuperscriptBox["k", "-"]]}]], 
  RowBox[{
   RowBox[{"deg", "[", 
    RowBox[{"lcm", "[", 
     RowBox[{"x", ",", "y"}], "]"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"deg", "[", 
    RowBox[{"lcm", "[", 
     RowBox[{"y", ",", "z"}], "]"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(79)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:245124153", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.606549074894179*^9, 3.606549134140615*^9}, {
  3.613104270597351*^9, 3.613104271922169*^9}, {3.613129143226326*^9, 
  3.613129154590997*^9}, {3.613130111787077*^9, 3.613130112266746*^9}, {
  3.613130229971759*^9, 3.613130230377745*^9}},
 CellTags->{
  "ID:245124153", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "79"},
 CellID->245124153],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"isExponentVector", "[", "z", "]"}]], 
   RowBox[{
    SubscriptBox["x", 
     SuperscriptBox["k", "-"]], 
    UnderscriptBox["\[LessEqual]", "D"], 
    SubscriptBox["z", 
     SuperscriptBox["k", "-"]]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{"y", "\[VerticalBar]", "z"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     SubscriptBox["z", "k"], 
     UnderscriptBox["<", "D"], 
     SubscriptBox["y", "k"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(80)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:952627996", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613105944755883*^9, 3.613105974282663*^9}, {
  3.613129161319429*^9, 3.613129197938506*^9}},
 CellTags->{
  "ID:952627996", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "80"},
 CellID->952627996],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    SubscriptBox["x", 
     SuperscriptBox["k", "-"]], 
    UnderscriptBox["\[LessEqual]", "D"], 
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["A", "j"], ")"}], 
     SuperscriptBox["k", "-"]]}]}], "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.613120755461824*^9, 3.613120763123582*^9}, {
  3.613129009894638*^9, 3.613129016507677*^9}},
 CellID->301509033],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"l", "=", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     RowBox[{
      UnderscriptBox[
       UnderscriptBox[
        SubscriptBox["argmin", "D"], 
        RowBox[{
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
       RowBox[{
        SubscriptBox["x", 
         SuperscriptBox["k", "-"]], 
        UnderscriptBox["\[LessEqual]", "D"], 
        SubscriptBox[
         RowBox[{"(", 
          SubscriptBox["A", "j"], ")"}], 
         SuperscriptBox["k", "-"]]}]], 
      SubscriptBox[
       RowBox[{"(", 
        SubscriptBox["A", "j"], ")"}], 
       SuperscriptBox["k", "-"]]}],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.613120774100889*^9, 3.613120781593787*^9}, {
  3.613129028244778*^9, 3.613129037559601*^9}},
 CellID->1728650251],

Cell[BoxData[
 RowBox[{"chainCrit", "[", 
  RowBox[{"y", ",", 
   SubscriptBox["A", "l"], ",", "A"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(81)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2013760956", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601182777313121*^9, 3.601182984183058*^9}, {
   3.601183658973657*^9, 3.601183713120301*^9}, {3.60118559939252*^9, 
   3.601185650377952*^9}, {3.601185744505231*^9, 3.601185747778637*^9}, {
   3.604836309120777*^9, 3.604836345795948*^9}, {3.605598814573376*^9, 
   3.6055988454947*^9}, 3.606552254701735*^9, 3.606554092737983*^9, {
   3.610957160542973*^9, 3.610957208854896*^9}, 3.611035531218395*^9, {
   3.611035638782283*^9, 3.61103566193718*^9}, {3.613103823014441*^9, 
   3.613103826655224*^9}, {3.613104208821321*^9, 3.613104216879475*^9}, {
   3.613116284632279*^9, 3.613116305315059*^9}, {3.613120759191659*^9, 
   3.613120816559275*^9}, {3.613129042816546*^9, 3.613129050406961*^9}},
 CellTags->{
  "ID:2013760956", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "81"},
 CellID->2013760956],

Cell[BoxData[
 RowBox[{
  RowBox[{"deg", "[", 
   RowBox[{"lcm", "[", 
    RowBox[{"x", ",", "y"}], "]"}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{"deg", "[", 
   RowBox[{"lcm", "[", 
    RowBox[{"y", ",", 
     SubscriptBox["A", "l"]}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(82)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1011330545", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.613120815013231*^9, 3.613129054190832*^9},
 CellTags->{
  "ID:1011330545", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "82"},
 CellID->1011330545],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"m", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
   RowBox[{"chainCrit", "[", 
    RowBox[{"y", ",", 
     SubscriptBox["A", "m"], ",", "A"}], "]"}]], 
  RowBox[{
   RowBox[{"deg", "[", 
    RowBox[{"lcm", "[", 
     RowBox[{"x", ",", "y"}], "]"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"deg", "[", 
    RowBox[{"lcm", "[", 
     RowBox[{"y", ",", 
      SubscriptBox["A", "m"]}], "]"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(83)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:718586393", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.61313036330654*^9, 3.613130429662397*^9}},
 CellTags->{
  "ID:718586393", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "83"},
 CellID->718586393],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1994638750]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->159419655],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(79)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(79)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["z", "ExpressionVariable"]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["z", "ExpressionVariable"]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:245124153"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p245124153-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p245124153-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:245124153", "Proof|ID:245124153-1"},
 CellID->1061920785],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p245124153-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937188479855*^9},
 CellID->1202825855]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(80)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(80)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    RowBox[{
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["z", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    RowBox[{
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["z", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["z", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["<", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["y", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:952627996"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p952627996-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p952627996-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:952627996", "Proof|ID:952627996-1"},
 CellID->921465591],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p952627996-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937191687111*^9},
 CellID->189186220]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(81)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(81)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox[
                    SubscriptBox["argmin", "D"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox[
                    SubscriptBox["argmin", "D"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2013760956"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2013760956-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2013760956-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2013760956", "Proof|ID:2013760956-1"},
 CellID->605609267],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2013760956-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937195550695*^9},
 CellID->47464659]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(82)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(82)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox[
                    SubscriptBox["argmin", "D"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox[
                    SubscriptBox["argmin", "D"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1011330545"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1011330545-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1011330545-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1011330545", "Proof|ID:1011330545-1"},
 CellID->1676671824],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1011330545-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937199807358*^9},
 CellID->1372119806]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(83)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(83)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["m", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]], 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["m", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{"isExponentVector", "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["m", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]], 
                    RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["m", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                    "]"}]}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:718586393"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p718586393-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p718586393-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:718586393", "Proof|ID:718586393-1"},
 CellID->1047095018],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p718586393-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.61493720362155*^9},
 CellID->866926986]
}, Closed]],

Cell[TextData[{
 "The diagram below visualizes formulas ",
 ButtonBox["(81)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2013760956"],
 " and ",
 ButtonBox["(82)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1011330545"],
 ":\nW.l.o.g. assume ",
 Cell[BoxData[
  FormBox["chainCritDiagonal", TraditionalForm]]],
 " holds for ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " w.r.t. the second coordinate, i.e. ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " lies right-below of ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " and the red-shaded region does not contain any elements of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " (because the chain criterion holds for ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 "). By definition of ",
 Cell[BoxData[
  FormBox["l", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "l"], TraditionalForm]]],
 " is one of the elements of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " lying not on the left of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " ",
 StyleBox["with minimal first coordinate",
  FontSlant->"Italic"],
 ", i.e. the blue-shaded region (which is in fact unbounded towards the top) \
does not contain any elements of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " either. This, however, in turn implies that the chain criterion also holds \
for ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "l"], TraditionalForm]]],
 ": Either ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "l"], TraditionalForm]]],
 " lies below ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", then ",
 Cell[BoxData[
  FormBox["chainCritDiagonal", TraditionalForm]]],
 " holds w.r.t. the second coordinate because the region \
\[OpenCurlyDoubleQuote]under\[CloseCurlyDoubleQuote] their least common \
multiple does not conatain any element of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " (\[Rule] subset of union of red- and blue-shaded regions), or it lies \
above ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", in which case ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " divides ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "l"], TraditionalForm]]],
 " and therefore the chain criterion holds due to formula ",
 ButtonBox["(divisibility\[Implies]chainCrit)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1391995733"],
 ". This finishes the informal \[OpenCurlyDoubleQuote]proof\
\[CloseCurlyDoubleQuote] of formula ",
 ButtonBox["(81)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2013760956"],
 ".\n",
 ButtonBox["(82)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1011330545"],
 " is even more obvious: Clearly, the lcm of ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "l"], TraditionalForm]]],
 " (blue square) is not located ",
 StyleBox["left",
  FontSlant->"Italic"],
 " of the lcm of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " (red square), because ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "l"], TraditionalForm]]],
 " is not located left of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ". Also, by definition it cannot lie below ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", which immediately implies that it does not lie ",
 StyleBox["below",
  FontSlant->"Italic"],
 " the lcm of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " either. Hence, its degree is at least as high as the degree of the lcm of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.615521744364631*^9, 3.615521765414933*^9}, {
  3.615521796168173*^9, 3.615522494359512*^9}, {3.615522542983201*^9, 
  3.615522743135044*^9}, {3.615522787855705*^9, 3.615522849718984*^9}, {
  3.615522912168401*^9, 3.615522918086908*^9}, {3.615786247223982*^9, 
  3.615786251507527*^9}, {3.615786293590515*^9, 3.615786301975064*^9}},
 CellID->1277801707],

Cell[BoxData[
 GraphicsBox[{
   {JoinForm[{"Miter", 2.0500000000000003`}], EdgeForm[{GrayLevel[0.], 
    Opacity[1.], AbsoluteThickness[1], 
    JoinForm[{"Miter", 2.0500000000000003`}]}], 
    ArrowBox[{{0.06815245478036216, 0.24878009425575315`}, {
     0.06815245478036216, 0.9514378574517769}}]}, 
   {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
    ArrowBox[{{0.0709302325581399, 0.24878009425575315`}, {
     1.2681524547803622`, 0.24878009425575293`}}]}, 
   GeometricTransformationBox[{
     {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
      GrayLevel[1]], 
      DiskBox[{0.23333333333333342`, 0.42222222222222194`}, \
{0.009820927516479921, 0.009820927516479727}]}, 
     {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
      RGBColor[0, 0.666667, 1]], 
      DiskBox[{0.8223514211886307, 0.2749999999999996}, \
{0.009820927516480116, 0.00982092751648006}]}, 
     {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
      RGBColor[1, 0, 0]], 
      DiskBox[{0.5583333333333336, 0.16388888888888858`}, \
{0.009820927516479672, 0.009820927516479977}]}, 
     {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
      RGBColor[0, 0.666667, 1]], 
      RectangleBox[{0.8052883497841318, 0.43498942799364565`}, \
{0.8308227613269795, 0.40945501645079824`}]}, 
     {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
      RGBColor[1, 0, 0]], 
      RectangleBox[{0.5441772386730208, 0.4349894279936456}, \
{0.5697116502158684, 0.4094550164507982}]}}, {{{{1., 0.}, {0., 
      1.1192778528786218`}}, {-0.029069767441860104`, 
      0.16172515014297145`}}}], 
   StyleBox[InsetBox[
     RowBox[{"y", 
      StyleBox[Cell["",
        CellBaseline->Baseline,
        TextAlignment->Left],
       TextAlignment->Left,
       TextJustification->0,
       Background->RGBColor[
        0.9570000762951095, 0.9600061036087587, 
         0.9700007629510948]]}], {0.14470284237726166`, 0.7403889645409887}, {
     Left, Top}, {0.054780361757106016`, 0.0901640492596668}, {{1., 0.}, {0., 
     1.}},
     Alignment->{Left, Top}],
    FontSize->24], 
   StyleBox[InsetBox[
     RowBox[{"x", 
      StyleBox[Cell["",
        CellBaseline->Baseline,
        TextAlignment->Left],
       TextAlignment->Left,
       TextJustification->0,
       Background->RGBColor[
        0.9570000762951095, 0.9600061036087587, 
         0.9700007629510948]]}], {0.5417635658914735, 0.362262432117058}, {
     Left, Top}, {0.07222222222222233, 0.09016404925966678}, {{1., 0.}, {0., 
     1.}},
     Alignment->{Left, Top}],
    FontSize->24], 
   StyleBox[InsetBox[
     RowBox[{
      SubscriptBox[
       StyleBox["A",
        TextAlignment->Left,
        TextJustification->0,
        Background->RGBColor[
         0.9570000762951095, 0.9600061036087587, 0.9700007629510948]], "l"], 
      StyleBox[Cell["",
        CellBaseline->Baseline,
        TextAlignment->Left],
       TextAlignment->Left,
       TextJustification->0,
       Background->RGBColor[
        0.9570000762951095, 0.9600061036087587, 
         0.9700007629510948]]}], {0.8223191214470289, 0.47108111225803473`}, {
     Left, Top}, {0.07222222222222224, 0.09016404925966678}, {{1., 0.}, {0., 
     1.}},
     Alignment->{Left, Top}],
    FontSize->24], 
   {AbsoluteThickness[1], StrokeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], 
    EdgeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{0.5292635658914734, 0.6343091324695007}, {0.5292635658914734, 
     0.3451623538091902}}]}, 
   {AbsoluteThickness[1], StrokeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], 
    EdgeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{0.20426356589147326`, 0.6343091324695007}, {0.7903746770025845, 
     0.6343091324695007}}]}, 
   {AbsoluteThickness[1], StrokeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], 
    EdgeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{0.7931524547803622, 0.6343091324695007}, {0.7931524547803622, 
     0.47263566483147773`}}]}, 
   {EdgeForm[GrayLevel[0]], FaceForm[{RGBColor[1, 0, 0], Opacity[0.1]}], 
    RectangleBox[{0.06928294573643422, 0.6169590643274863}, \
{0.5082364341085275, 0.248538011695907}]}, 
   {EdgeForm[GrayLevel[0]], FaceForm[{RGBColor[0, 0.666667, 1], Opacity[
    0.1]}], RectangleBox[{0.5082364341085275, 0.24853801169590697`}, \
{0.7698643410852718, 0.9093567251461991}]}},
  AspectRatio->0.6008676789587852,
  ContentSelectable->True,
  ImageMargins->{{3., 0.}, {0., 0.}},
  ImagePadding->{{0., 0.}, {0., 0.}},
  ImageSize->{461., 277.},
  PlotRange->{{0., 1.3333333333333335`}, {0.19005847953216373`, 1.}},
  PlotRangePadding->None]], "Text",
 Editable->False,
 Deployed->True,
 CellChangeTimes->{{3.60560130456881*^9, 3.605601309150591*^9}, 
   3.60560148437655*^9, {3.605606346223402*^9, 3.605606348528124*^9}, {
   3.61552110910792*^9, 3.615521323217483*^9}, {3.615521356258374*^9, 
   3.615521389589831*^9}, {3.615521564300829*^9, 3.615521568828444*^9}, {
   3.615521622761972*^9, 3.615521714600076*^9}},
 TextAlignment->Center,
 CellID->696386456],

Cell[TextData[{
 "The next theorem states the most important result of this subsection, \
namely that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", 
    RowBox[{"A", " ", "\:293a", " ", "x"}], "]"}], TraditionalForm]]],
 " does not increase under certain circumstances (formula ",
 ButtonBox["(86)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:789654176"],
 ")."
}], "Text",
 CellChangeTimes->{{3.615188481824006*^9, 3.615188550958078*^9}, {
  3.615786028925536*^9, 3.615786029313824*^9}},
 CellID->981796601],

Cell["", "OpenEnvironment",
 CellID->435957245],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: \[OpenCurlyDoubleQuote]Points which divide LCMs of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " do not increase ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 "\[CloseCurlyDoubleQuote]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599813863230678*^9, 3.599813932646038*^9}, {
  3.599815769924101*^9, 3.599815783930462*^9}, {3.599887515498872*^9, 
  3.599887535072322*^9}, {3.604057498189491*^9, 3.60405750010472*^9}},
 CellID->2089484077],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"deg", "[", "x", "]"}], 
     UnderscriptBox["\[LessEqual]", "D"], 
     RowBox[{"M", "[", "A", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{"x", "\[VerticalBar]", 
      RowBox[{"lcm", "[", 
       RowBox[{
        SubscriptBox["A", "i"], ",", 
        SubscriptBox["A", "j"]}], "]"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.613126240702134*^9, 3.613126249692261*^9}, {
  3.613126595672313*^9, 3.613126603109546*^9}},
 CellID->2065858242],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"s", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"chainCritDiagonal", "[", 
     RowBox[{"x", ",", 
      SubscriptBox["A", "s"], ",", "A"}], "]"}], "\[Implies]", 
    RowBox[{
     RowBox[{"deg", "[", 
      RowBox[{"lcm", "[", 
       RowBox[{"x", ",", 
        SubscriptBox["A", "s"]}], "]"}], "]"}], 
     UnderscriptBox["\[LessEqual]", "D"], 
     RowBox[{"M", "[", "A", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(84)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:696507084", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613190054477989*^9, 3.613190107076762*^9}},
 CellTags->{
  "ID:696507084", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "84"},
 CellID->696507084],

Cell[BoxData[
 RowBox[{
  RowBox[{"M", "[", 
   RowBox[{"x", ",", "A"}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{"M", "[", "A", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(85)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:626970451", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613126266046262*^9, 3.613126273991831*^9}, {
  3.613126590484572*^9, 3.613126590812798*^9}},
 CellTags->{
  "ID:626970451", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "85"},
 CellID->626970451],

Cell[BoxData[
 RowBox[{
  RowBox[{"M", "[", 
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{"M", "[", "A", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(86)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:789654176", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599386178864095*^9, 3.599386293139529*^9}, {
  3.599386345120476*^9, 3.599386348105769*^9}, {3.599386426462421*^9, 
  3.599386452462884*^9}, {3.599386505511611*^9, 3.59938653642992*^9}, {
  3.599813250551335*^9, 3.599813250773232*^9}, {3.599815751346653*^9, 
  3.599815764498039*^9}, {3.599815965418804*^9, 3.59981598759373*^9}, {
  3.600688796251583*^9, 3.600688799534651*^9}, {3.601183090490615*^9, 
  3.601183190968485*^9}, {3.601183231618369*^9, 3.601183238386104*^9}, {
  3.604836394575993*^9, 3.604836402198217*^9}, {3.60559886732349*^9, 
  3.605598873359937*^9}, {3.613126247929947*^9, 3.613126256369088*^9}, {
  3.613126601741142*^9, 3.613126604600106*^9}},
 CellTags->{
  "ID:789654176", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "86"},
 CellID->789654176],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->672392106]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->649471340],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(84)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(84)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["s", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:696507084"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p696507084-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p696507084-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:696507084", "Proof|ID:696507084-1"},
 CellID->785497368],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p696507084-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937219894242*^9},
 CellID->1321524629]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(85)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(85)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"M", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"M", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:626970451"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p626970451-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p626970451-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:626970451", "Proof|ID:626970451-1"},
 CellID->1008962236],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p626970451-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937223870064*^9},
 CellID->254095126]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(86)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(86)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"M", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"M", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:789654176"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p789654176-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p789654176-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:789654176", "Proof|ID:789654176-1"},
 CellID->543282043],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p789654176-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937227277881*^9},
 CellID->1999610841]
}, Closed]],

Cell[TextData[StyleBox["Remark:",
 FontFamily->"Bitstream Charter",
 FontSize->14,
 FontWeight->"Bold"]], "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.603522907354292*^9, 3.603522909426996*^9}, {
   3.603522955835897*^9, 3.603522963082172*^9}, {3.603793011527348*^9, 
   3.603793024231905*^9}, 3.604052995320351*^9, 3.604056263647482*^9, {
   3.604056976471727*^9, 3.604056993288131*^9}, 3.604057052315235*^9},
 CellID->995033752],

Cell["In all formulas of the above theorem:", "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.604056279050277*^9, 3.604056285058309*^9}, {
  3.604056999138058*^9, 3.6040570017543*^9}, {3.604057278847682*^9, 
  3.604057278854649*^9}, {3.613194675170232*^9, 3.613194684937353*^9}},
 CellID->8175233],

Cell[TextData[{
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 " may be equal, in which case ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " divides some element of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 "."
}], "Item",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.604056254509397*^9, 3.604056263647619*^9}, {
  3.604057007430305*^9, 3.604057031825153*^9}},
 CellID->1251264984],

Cell[TextData[{
 "Nothing is said about whether the chain criterion holds for ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "i"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "j"], TraditionalForm]]],
 "."
}], "Item",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.604056254509397*^9, 3.604056263647793*^9}, {
  3.604057037864601*^9, 3.604057037869041*^9}, {3.60483641635254*^9, 
  3.604836428997536*^9}},
 CellID->1241360375],

Cell[TextData[{
 "Also, formula ",
 ButtonBox["(86)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:789654176"],
 " does not hold in the general case of arbitrary ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " because of the following counterexample for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "3"}], TraditionalForm]]],
 ", as can be easily verified by computation: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"A", "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", ",", "7", ",", "0"}], ")"}], ",", 
      RowBox[{"(", 
       RowBox[{"5", ",", "3", ",", "6"}], ")"}], ",", 
      RowBox[{"(", 
       RowBox[{"4", ",", "1", ",", "1"}], ")"}]}], "\[RightAngleBracket]"}]}],
    TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", 
    RowBox[{"(", 
     RowBox[{"3", ",", "3", ",", "6"}], ")"}]}], TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.599387497964957*^9, 3.599387547390328*^9}, {
  3.599388854616552*^9, 3.59938895559994*^9}, {3.599544418766914*^9, 
  3.599544440714715*^9}, {3.599802104620125*^9, 3.599802176696523*^9}, {
  3.59980339419387*^9, 3.599803396761692*^9}, {3.601183270233227*^9, 
  3.601183278377561*^9}, {3.604057105561947*^9, 3.604057175976999*^9}, {
  3.604057286578313*^9, 3.604057286590137*^9}, {3.604058031598409*^9, 
  3.604058033812078*^9}, {3.615188567224545*^9, 3.61518856862013*^9}, {
  3.615786035142434*^9, 3.615786035328131*^9}},
 CellID->1860606636],

Cell[TextData[{
 "The following corollary (in particular formula ",
 ButtonBox["(88)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:477242065"],
 ") is in fact equivalent to formula ",
 ButtonBox["(86)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:789654176"],
 " above, only that the premise of the implication is formulated slightly \
differently. It is one of the two main results for proving that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " does not increase in the course of Buchberger\[CloseCurlyQuote]s \
algorithm, because it states that any point lying in the rectangular region \
of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " and \[OpenCurlyDoubleQuote]below\[CloseCurlyDoubleQuote] ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " does not increase ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 " when added to ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.599887620133906*^9, 3.599887659459411*^9}, {
  3.604057199522655*^9, 3.604057218970398*^9}, {3.604057294991378*^9, 
  3.604057301879796*^9}, {3.60405764858964*^9, 3.604057715719259*^9}, {
  3.604058097191682*^9, 3.604058238601855*^9}, {3.61518858314721*^9, 
  3.615188599411484*^9}, {3.61578604146944*^9, 3.615786041655508*^9}, {
  3.615786419213074*^9, 3.615786420156703*^9}},
 CellID->707373101],

Cell["", "OpenEnvironment",
 CellID->454483030],

Cell[CellGroupData[{

Cell[TextData[{
 "Corollary: \[OpenCurlyDoubleQuote]Points in the rectangular region of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " do not increase ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 "\[CloseCurlyDoubleQuote]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599816705851175*^9, 3.59981671376198*^9}, {
  3.604057976778585*^9, 3.604057985909831*^9}},
 CellID->334142738],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isExponentVector", "[", "x", "]"}]], 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"inRectangle", "[", 
     RowBox[{"x", ",", "A"}], "]"}], 
    TagBox["\[Implies]",
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
      RowBox[{"x", "\[VerticalBar]", 
       RowBox[{"lcm", "[", 
        RowBox[{
         SubscriptBox["A", "i"], ",", 
         SubscriptBox["A", "j"]}], "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ")"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(87)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1437590439", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.601102353419661*^9, 3.601102418868956*^9}, {
  3.604057401994073*^9, 3.604057404423133*^9}, {3.604836438075568*^9, 
  3.604836450571683*^9}, {3.60559899481642*^9, 3.605598999867949*^9}, {
  3.613194910205427*^9, 3.613194911609383*^9}, {3.613195087101627*^9, 
  3.613195093140737*^9}},
 CellTags->{
  "ID:1437590439", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "87"},
 CellID->1437590439],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"deg", "[", "x", "]"}], 
     UnderscriptBox["\[LessEqual]", "D"], 
     RowBox[{"M", "[", "A", "]"}]}], "\[And]", 
    RowBox[{"inRectangle", "[", 
     RowBox[{"x", ",", "A"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"M", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"M", "[", "A", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(88)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:477242065", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599816720624339*^9, 3.59981673870893*^9}, {
   3.599816834567939*^9, 3.599816907974399*^9}, {3.59988756996878*^9, 
   3.599887589119038*^9}, {3.599891031289947*^9, 3.599891039454976*^9}, 
   3.600752394919691*^9, {3.604057410970432*^9, 3.60405744461789*^9}, {
   3.604836454161382*^9, 3.604836458333802*^9}, {3.605599001121929*^9, 
   3.605599009478842*^9}},
 CellTags->{
  "ID:477242065", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "88"},
 CellID->477242065],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->776728042]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1701216104],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(87)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(87)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"inRectangle", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                   RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"inRectangle", "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                   RowBox[{"lcm", "[", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1437590439"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1437590439-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1437590439-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1437590439", "Proof|ID:1437590439-1"},
 CellID->773586293],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1437590439-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.61493723182107*^9},
 CellID->140107252]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(88)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(88)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{"inRectangle", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"M", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{"inRectangle", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{"M", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:477242065"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p477242065-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p477242065-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:477242065", "Proof|ID:477242065-1"},
 CellID->700215890],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p477242065-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937235060931*^9},
 CellID->1979693258]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Far Exterior", "Subsubsection",
 CellChangeTimes->{{3.599385847316572*^9, 3.599385853822335*^9}, {
  3.606466922424492*^9, 3.606466958101395*^9}},
 CellTags->"sect::TheoremsFarExterior",
 CellID->1174010248],

Cell[TextData[{
 "Since the rectangular region of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " has been dealt with in the theorem/corollary above, the only part of the \
cover of the two-dimensional space of exponents that still remains is ",
 Cell[BoxData[
  FormBox["inFarExterior", TraditionalForm]]],
 " (and also ",
 Cell[BoxData[
  FormBox["isAbove", TraditionalForm]]],
 ", but this can be handled without any auxiliary results). Therefore, the \
next two theorems are all about points lying in the far exterior."
}], "Text",
 CellChangeTimes->{{3.604058283935259*^9, 3.604058555117129*^9}, {
   3.604836468089176*^9, 3.604836496737076*^9}, {3.605599018599544*^9, 
   3.605599025812846*^9}, {3.613196175924531*^9, 3.613196186595511*^9}, 
   3.615188673089529*^9},
 CellID->1038672145],

Cell[TextData[{
 "Please note that although ",
 Cell[BoxData[
  FormBox["inFarExterior", TraditionalForm]]],
 " consists of two disjoint subsets (far exterior w.r.t. first coordinate, \
far exterior w.r.t. second coordinate) which, in principle, would have to be \
dealt with separately, due to the use of function ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[EmptySquare]", "-"], TraditionalForm]]],
 " (\[OpenCurlyDoubleQuote]other coordinate\[CloseCurlyDoubleQuote]) they can \
be considered at once, without any additional effort."
}], "Text",
 CellChangeTimes->{{3.615188678582117*^9, 3.615188809377078*^9}, {
  3.615188961297454*^9, 3.615188962762936*^9}},
 CellID->167638010],

Cell[TextData[{
 "The first theorem characterizes elements ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "i"], TraditionalForm]]],
 " of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " such that the chain criterion holds for ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "i"], TraditionalForm]]],
 ", if ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " lies in the far exterior (formula ",
 ButtonBox["(94)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1370308823"],
 ")."
}], "Text",
 CellChangeTimes->{{3.604058629568643*^9, 3.604058703609933*^9}, {
   3.604836524485254*^9, 3.604836544074096*^9}, {3.615188840267976*^9, 
   3.615188908083812*^9}, {3.615188953490907*^9, 3.615188953512703*^9}, 
   3.615523198400226*^9, {3.615787129824091*^9, 3.615787130856051*^9}},
 CellID->42774035],

Cell["", "OpenEnvironment",
 CellID->370524994],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: Characterization of admissible pairs if ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " lies in the far exterior"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.5998177464167*^9, 3.599817767618709*^9}, {
  3.59988754103089*^9, 3.599887547135335*^9}, {3.604057482503943*^9, 
  3.604057484455871*^9}, {3.613205044470315*^9, 3.613205045466067*^9}},
 CellID->1808105650],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "2"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.613196278519109*^9, 3.613196282500853*^9}},
 CellID->616105124],

Cell[BoxData[
 RowBox[{
  RowBox[{"inFarExterior", "[", 
   RowBox[{"x", ",", "A", ",", "k"}], "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.61320437504479*^9, 3.613204382960508*^9}},
 CellID->1179235710],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.613204402860491*^9, 3.613204409945779*^9}},
 CellID->1322949089],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{"x", "\[VerticalBar]", 
   SubscriptBox["A", "i"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(89)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1041359623", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613204414418051*^9, 3.613204421295282*^9}},
 CellTags->{
  "ID:1041359623", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "89"},
 CellID->1041359623],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{
   SubscriptBox["A", "i"], "\[VerticalBar]", 
   "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(90)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2130272336", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613204427159571*^9, 3.613204437590165*^9}},
 CellTags->{
  "ID:2130272336", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "90"},
 CellID->2130272336],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", 
    SubscriptBox["A", "i"], ",", "A", ",", 
    SuperscriptBox["k", "-"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(91)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1584641667", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613204442032434*^9, 3.613204454918414*^9}},
 CellTags->{
  "ID:1584641667", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "91"},
 CellID->1584641667],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCritDiagonal", "[", 
   RowBox[{"x", ",", 
    SubscriptBox["A", "i"], ",", "A", ",", "k"}], "]"}], "\[Implies]", 
  RowBox[{
   SubscriptBox[
    RowBox[{"(", 
     SubscriptBox["A", "i"], ")"}], "k"], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "k"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(92)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1175076282", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613204482979075*^9, 3.613204490036357*^9}, {
  3.613735290316331*^9, 3.613735294074118*^9}, {3.613798245165952*^9, 
  3.613798252417094*^9}},
 CellTags->{
  "ID:1175076282", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "92"},
 CellID->1175076282],

Cell[BoxData[
 RowBox[{
  RowBox[{"chainCrit", "[", 
   RowBox[{"x", ",", 
    SubscriptBox["A", "i"], ",", "A"}], "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.613799427468845*^9, 3.613799435826997*^9}},
 CellID->1303947124],

Cell[BoxData[
 RowBox[{
  SubscriptBox[
   RowBox[{"(", 
    SubscriptBox["A", "i"], ")"}], "k"], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{"W", "[", 
   RowBox[{"A", ",", "k"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(93)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:333980150", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmViYGAQBmIQfXqvyZlpF185usUbXAXRrH5JL0H093LubyB61fJd
v0B01sa9f0G0VPR03ulA+uStClEQXbznrxGIbktiNgbR2Z8FmEQuvXJkur6S
GURfvC7ZAaLLCuaD6SUrm2OlgfQmxt9xIFreqqEBRGfZrAPThbaJU0D0vH93
poHovD//DoJoR43awyD6QbLNesNbrxxniRSC6c1HSiXW3X7l2HuwUApEB/FH
ZCjdfeXY73YUTM/7efr7/NevHH+ZnPsJok88DP6yGEgX72oA05dyq/+C6Fb1
hWD6pI5KwLE3rxyNA0+CaV6p540g+sn8ma0gGgDpw66L
  "],
 CellTags->{
  "ID:333980150", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "93"},
 CellID->333980150],

Cell[BoxData[
 RowBox[{
  SubscriptBox[
   RowBox[{"(", 
    SubscriptBox["A", "i"], ")"}], "k"], "\[Equal]", 
  RowBox[{"W", "[", 
   RowBox[{"A", ",", "k"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(94)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1370308823", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613799473924809*^9, 3.613799476365003*^9}},
 CellTags->{
  "ID:1370308823", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "94"},
 CellID->1370308823],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1131751420]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1349563567],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(89)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(89)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[VerticalBar]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1041359623"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1041359623-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1041359623-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1041359623", "Proof|ID:1041359623-1"},
 CellID->650148495],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1041359623-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937243820222*^9},
 CellID->114116961]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(90)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(90)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "\[VerticalBar]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2130272336"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2130272336-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2130272336-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2130272336", "Proof|ID:2130272336-1"},
 CellID->874306688],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2130272336-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937268763511*^9},
 CellID->1181186212]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(91)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(91)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1584641667"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1584641667-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1584641667-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1584641667", "Proof|ID:1584641667-1"},
 CellID->813414581],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1584641667-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937272162874*^9},
 CellID->1535779146]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(92)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(92)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{"chainCritDiagonal", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1175076282"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1175076282-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1175076282-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1175076282", "Proof|ID:1175076282-1"},
 CellID->54321453],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1175076282-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937275667116*^9},
 CellID->1473601407]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(93)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(93)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:333980150"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p333980150-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p333980150-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:333980150", "Proof|ID:333980150-1"},
 CellID->749915393],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p333980150-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937279378953*^9},
 CellID->1548804265]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(94)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(94)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "\[Equal]", 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{"chainCrit", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], "\[Equal]", 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1370308823"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1370308823-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1370308823-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1370308823", "Proof|ID:1370308823-1"},
 CellID->1775388908],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1370308823-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937283322412*^9},
 CellID->1029312850]
}, Closed]],

Cell[TextData[{
 "The diagram below illustrates formulas ",
 ButtonBox["(93)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:333980150"],
 " and ",
 ButtonBox["(94)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1370308823"],
 ": Assume ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " lies in the far exterior w.r.t. the second coordinate, i.e. strictly ",
 StyleBox["below",
  FontSlant->"Italic"],
 " the horizontal thick line and strictly ",
 StyleBox["right",
  FontSlant->"Italic"],
 " of the vertical thick line. Assume further that the blue-shaded rectangle \
is the ",
 StyleBox["smallest",
  FontSlant->"Italic"],
 " rectangle containing all elements of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ". This means that for each of the four sides of the rectangle there is at \
least one element of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " lying exactly on it; This also holds in particular for the bottom side, \
where ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "j"], TraditionalForm]]],
 " lies (possibly amongst others). Now assume the contrary of formula ",
 ButtonBox["(93)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:333980150"],
 ", i.e. ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "i"], TraditionalForm]]],
 " is located strictly above the horizontal thick line (whose distance to the \
",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 "-axis is exactly the width of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " w.r.t. the second coordinate). Clearly, if the chain criterion holds for \
",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "i"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["chainCritDiagonal", TraditionalForm]]],
 " must hold w.r.t. the second coordinate. Hence, in such case the red-shaded \
rectangle \[OpenCurlyDoubleQuote]under\[CloseCurlyDoubleQuote] the lcm of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "i"], TraditionalForm]]],
 " should not contain any elements of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 "; However, ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "j"], TraditionalForm]]],
 " apparently lies in it - contradiction. Therefore, ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "i"], TraditionalForm]]],
 " must have smallest second coordinate among all elements of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ", which is precisely the width of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " w.r.t. the second coordinate."
}], "Text",
 CellChangeTimes->{{3.615523674012916*^9, 3.61552415155907*^9}, {
  3.615524202494799*^9, 3.615524287373781*^9}, {3.61552437383494*^9, 
  3.615524407228254*^9}, {3.615785562294069*^9, 3.615785578045749*^9}},
 CellID->1096841450],

Cell[BoxData[
 GraphicsBox[{
   StyleBox[InsetBox[
     RowBox[{
      SubscriptBox[
       StyleBox["A",
        TextAlignment->Left,
        TextJustification->0,
        Background->RGBColor[
         0.9570000762951095, 0.9600061036087587, 0.9700007629510948]], "j"], 
      StyleBox[Cell["",
        CellBaseline->Baseline,
        TextAlignment->Left],
       TextAlignment->Left,
       TextJustification->0,
       Background->RGBColor[
        0.9570000762951095, 0.9600061036087587, 
         0.9700007629510948]]}], {0.252543271844715, 0.3804378374042352}, {
     Left, Top}, {0.07222222222222224, 0.09016404925966678}, {{1., 0.}, {0., 
     1.}},
     Alignment->{Left, Top}],
    FontSize->24], 
   {EdgeForm[GrayLevel[0]], FaceForm[{RGBColor[1, 0, 0], Opacity[0.1]}], 
    RectangleBox[{0.8503253796095445, 0.46783625730994205`}, \
{0.0694143167028198, 0.2485380116959069}]}, 
   {EdgeForm[GrayLevel[0]], FaceForm[{RGBColor[0, 0.666667, 1], Opacity[
    0.1]}], RectangleBox[{0.14091900822277165`, 0.39181286549707905`}, \
{0.6975577611865009, 0.7836257309941536}]}, 
   {JoinForm[{"Miter", 2.0500000000000003`}], EdgeForm[{GrayLevel[0.], 
    Opacity[1.], AbsoluteThickness[1], 
    JoinForm[{"Miter", 2.0500000000000003`}]}], 
    ArrowBox[{{0.06815245478036216, 0.24878009425575315`}, {
     0.06815245478036216, 0.9514378574517769}}]}, 
   {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
    ArrowBox[{{0.0709302325581399, 0.24878009425575315`}, {
     1.2681524547803622`, 0.24878009425575293`}}]}, 
   GeometricTransformationBox[{
     {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
      RGBColor[0, 0.666667, 1]], 
      DiskBox[{0.4694953112826292, 0.2932866445580688}, \
{0.009820927516480116, 0.00982092751647995}]}, 
     {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
      RGBColor[1, 0, 0]], 
      DiskBox[{0.9025126536514826, 0.1560517555068591}, \
{0.00982092751647956, 0.009820927516479755}]}, 
     {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
      RGBColor[1, 0, 0]], 
      RectangleBox[{0.8891639824667059, 0.3095952938811732}, \
{0.9146983940095537, 0.2840608823383256}]}}, {{{{1., 0.}, {0., 
      1.1192778528786218`}}, {-0.029069767441860104`, 
      0.16172515014297145`}}}], 
   StyleBox[InsetBox[
     RowBox[{
      SubscriptBox["A", "i"], 
      StyleBox[Cell["",
        CellBaseline->Baseline,
        TextAlignment->Left],
       TextAlignment->Left,
       TextJustification->0,
       Background->RGBColor[
        0.9570000762951095, 0.9600061036087587, 
         0.9700007629510948]]}], {0.3905452140330823, 0.6029620639561941}, {
     Left, Top}, {0.09527204650041765, 0.0901640492596668}, {{1., 0.}, {0., 
     1.}},
     Alignment->{Left, Top}],
    FontSize->24], 
   StyleBox[InsetBox[
     RowBox[{"x", 
      StyleBox[Cell["",
        CellBaseline->Baseline,
        TextAlignment->Left],
       TextAlignment->Left,
       TextJustification->0,
       Background->RGBColor[
        0.9570000762951095, 0.9600061036087587, 
         0.9700007629510948]]}], {0.8772660966217699, 0.3564144789006841}, {
     Left, Top}, {0.07222222222222233, 0.09016404925966678}, {{1., 0.}, {0., 
     1.}},
     Alignment->{Left, Top}],
    FontSize->24], 
   {AbsoluteThickness[1], StrokeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], 
    EdgeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{0.4396240057845263, 0.4912280701754391}, {0.8734634851771511, 
     0.49122807017543957`}}]}, 
   {AbsoluteThickness[1], StrokeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], 
    EdgeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{0.8734634851771512, 0.3391812865497077}, {0.8734634851771512, 
     0.48830409356725224`}}]}, 
   {Thickness[0.004265795188015926], StrokeForm[{RGBColor[0., 0., 0.], 
    Opacity[1.]}], EdgeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[
    None], LineBox[{{0.07230657989877072, 0.39181286549707656`}, {
     1.2581344902386118`, 0.391812865497077}}]}, 
   {Thickness[0.004415704473533121], StrokeForm[{RGBColor[0., 0., 0.], 
    Opacity[1.]}], EdgeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[
    None], LineBox[{{0.6999276934201013, 0.2543859649122808}, {
     0.6999276934201013, 0.9473684210526323}}]}, GeometricTransformationBox[
    {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
     GrayLevel[1]], 
     DiskBox[{0.3316702819956617, 0.2053948653194046}, \
{0.00982092751647995, 0.009820927516479894}]}, {{{1., 0.}, {0., 
     1.1192778528786218`}}, {-0.029069767441860104`, 0.16172515014297145`}}]},
  
  AspectRatio->0.6008676789587852,
  ContentSelectable->True,
  ImageMargins->{{3., 0.}, {0., 0.}},
  ImagePadding->{{0., 0.}, {0., 0.}},
  ImageSize->{461., 277.},
  PlotRange->{{0., 1.3333333333333335`}, {0.19005847953216373`, 1.}},
  PlotRangePadding->Automatic]], "Text",
 Editable->False,
 Deployed->True,
 CellChangeTimes->{{3.60560130456881*^9, 3.605601309150591*^9}, 
   3.60560148437655*^9, {3.605606346223402*^9, 3.605606348528124*^9}, {
   3.61552110910792*^9, 3.615521323217483*^9}, {3.615521356258374*^9, 
   3.615521389589831*^9}, {3.615521564300829*^9, 3.615521568828444*^9}, {
   3.615521622761972*^9, 3.615521714600076*^9}, {3.615523333465935*^9, 
   3.615523556838936*^9}, {3.615523605577046*^9, 3.615523629330637*^9}, {
   3.615524790106071*^9, 3.615524806545388*^9}},
 TextAlignment->Center,
 CellID->100422464],

Cell[TextData[{
 "Eventually we are able to state and prove the second of the two main \
results based upon which theorem \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " does not increase\[CloseCurlyDoubleQuote] can be proved. As already \
indicated above, now the case where ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " lies in the far exterior is considered. Unfortunately, we cannot derive a \
result similar to formula ",
 ButtonBox["(88)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:477242065"],
 " (where ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " lies in the rectangular region) that ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 " does not increase: ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 " ",
 StyleBox["does",
  FontSlant->"Italic"],
 " increase, but luckily only up to a certain small amount that still allows \
us to conclude that the sum ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " does ",
 StyleBox["not",
  FontSlant->"Italic"],
 " increase:"
}], "Text",
 CellChangeTimes->{{3.604058916507*^9, 3.604059195263811*^9}, {
  3.604059237835544*^9, 3.604059238726389*^9}, {3.604059642379295*^9, 
  3.604059642527374*^9}, {3.615188994813476*^9, 3.615188996784771*^9}, {
  3.615189039755407*^9, 3.615189039778875*^9}, {3.615786684965713*^9, 
  3.615786685816557*^9}},
 CellID->443594684],

Cell["", "OpenEnvironment",
 CellID->552391168],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: Extent to which ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 " increases if ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " lies in the far exterior"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599888433165719*^9, 3.599888481678889*^9}, {
  3.604057487180299*^9, 3.604057489353393*^9}, {3.613205047014627*^9, 
  3.613205047826574*^9}},
 CellID->1097894802],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   "2"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.613204652841628*^9, 3.613204656548159*^9}},
 CellID->1460528271],

Cell[BoxData[
 RowBox[{
  RowBox[{"inFarExterior", "[", 
   RowBox[{"x", ",", "A", ",", "k"}], "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.604057569031152*^9, 3.604057573629476*^9}, {
  3.605599056420323*^9, 3.605599058127599*^9}, {3.613204660560921*^9, 
  3.613204667077748*^9}, {3.613801038243596*^9, 3.613801041194964*^9}},
 CellID->58178322],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   RowBox[{"deg", "[", 
    RowBox[{"lcm", "[", 
     RowBox[{"x", ",", 
      SubscriptBox["A", "i"]}], "]"}], "]"}], "\[Equal]", 
   RowBox[{
    SubscriptBox["x", 
     SuperscriptBox["k", "-"]], 
    UnderscriptBox["+", "D"], 
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["A", "i"], ")"}], "k"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(95)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1434774204", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613205826620106*^9, 3.613205881973755*^9}},
 CellTags->{
  "ID:1434774204", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "95"},
 CellID->1434774204],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"M", "[", "A", "]"}], 
   UnderscriptBox["+", "D"], 
   SubscriptBox["x", "k"]}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   RowBox[{"M", "[", "A", "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "k"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(96)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1867780171", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.613810162879502*^9},
 CellTags->{
  "ID:1867780171", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "96"},
 CellID->1867780171],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"deg", "[", "x", "]"}], 
   UnderscriptBox["\[LessEqual]", "D"], 
   RowBox[{"M", "[", "A", "]"}]}], "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.613801033085107*^9, 3.613801039830106*^9}},
 CellID->329161725],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"deg", "[", "x", "]"}], 
   UnderscriptBox["+", "D"], 
   SubscriptBox["x", "k"]}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   RowBox[{"M", "[", "A", "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "k"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(97)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:786584693", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613810092302236*^9, 3.613810094699308*^9}, {
  3.613810161550508*^9, 3.613810169643946*^9}},
 CellTags->{
  "ID:786584693", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "97"},
 CellID->786584693],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"M", "[", 
    RowBox[{"x", ",", "A"}], "]"}], 
   UnderscriptBox["+", "D"], 
   SubscriptBox["x", "k"]}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   RowBox[{"M", "[", "A", "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "k"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(98)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:183897269", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613810106452905*^9, 3.613810108188724*^9}},
 CellTags->{
  "ID:183897269", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "98"},
 CellID->183897269],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"M", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], 
   UnderscriptBox["+", "D"], 
   SubscriptBox["x", "k"]}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   RowBox[{"M", "[", "A", "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"W", "[", 
    RowBox[{"A", ",", "k"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(99)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1246401530", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599888340023496*^9, 3.599888425884846*^9}, {
   3.599888608502744*^9, 3.599888668521211*^9}, 3.600752425401211*^9, {
   3.604057517036633*^9, 3.604057534598771*^9}, {3.604057572370733*^9, 
   3.604057576136678*^9}, {3.604836580940151*^9, 3.604836588825862*^9}, {
   3.605599063800728*^9, 3.605599068339616*^9}, {3.613204686615053*^9, 
   3.613204692474531*^9}},
 CellTags->{
  "ID:1246401530", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "99"},
 CellID->1246401530],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->146116537]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->838044464],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(95)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(95)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    "\[Equal]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}}]], 
                   RowBox[{
                    RowBox[{"deg", "[", 
                    RowBox[{"lcm", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], "]"}], "]"}], 
                    "\[Equal]", 
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SuperscriptBox[
                    StyleBox["k", "ExpressionVariable"], "-"], 
                    TagBox[")", "AutoParentheses"]}]], 
                    TagBox[")", "AutoParentheses"]}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1434774204"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1434774204-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1434774204-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1434774204", "Proof|ID:1434774204-1"},
 CellID->4587845],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1434774204-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.61493728862633*^9},
 CellID->704261409]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(96)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(96)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                   UnderscriptBox["\[LessEqual]", "D"], 
                   RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1867780171"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1867780171-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1867780171-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1867780171", "Proof|ID:1867780171-1"},
 CellID->975302909],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1867780171-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937292626551*^9},
 CellID->925555814]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(97)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(97)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isExponentVector", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "2"}]}}]], 
               RowBox[{
                 RowBox[{"inFarExterior", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], 
                   "\[Implies]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                    UnderscriptBox["\[LessEqual]", "D"], 
                    RowBox[{
                    RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "D"], 
                    RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:786584693"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p786584693-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p786584693-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:786584693", "Proof|ID:786584693-1"},
 CellID->1192023621],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p786584693-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937296090362*^9},
 CellID->1973863468]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(98)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(98)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                RowBox[{"isExponentVector", "[", 
                  StyleBox["x", "ExpressionVariable"], "]"}]}, {
                RowBox[{
                  RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                  "\[Ellipsis]", ",", "2"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{"inFarExterior", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 RowBox[{"deg", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                RowBox[{"isExponentVector", "[", 
                  StyleBox["x", "ExpressionVariable"], "]"}]}, {
                RowBox[{
                  RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                  "\[Ellipsis]", ",", "2"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{"inFarExterior", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 RowBox[{"deg", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:183897269"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p183897269-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p183897269-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:183897269", "Proof|ID:183897269-1"},
 CellID->1230628713],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p183897269-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->90775516]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(99)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(99)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                RowBox[{"isExponentVector", "[", 
                  StyleBox["x", "ExpressionVariable"], "]"}]}, {
                RowBox[{
                  RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                  "\[Ellipsis]", ",", "2"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{"inFarExterior", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 RowBox[{"deg", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                RowBox[{"isExponentVector", "[", 
                  StyleBox["x", "ExpressionVariable"], "]"}]}, {
                RowBox[{
                  RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                  "\[Ellipsis]", ",", "2"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{"inFarExterior", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 RowBox[{"deg", "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"M", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    TagBox[")", "AutoParentheses"]}]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1246401530"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1246401530-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1246401530-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1246401530", "Proof|ID:1246401530-1"},
 CellID->879169295],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1246401530-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellID->637478936]
}, Closed]],

Cell[TextData[{
 "The diagram below illustrates formula ",
 ButtonBox["(99)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1246401530"],
 ": Assume w.l.o.g. that ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " lies in the far exterior w.r.t. the second coordinate and, in addition, \
strictly below the black diagonal line representing ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 ". From formula ",
 ButtonBox["(94)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1370308823"],
 " we know that the lcm of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 StyleBox["any",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "i"], TraditionalForm]]],
 ", such that the chain criterion holds for ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "i"], TraditionalForm]]],
 ", needs to lie exactly on the horizontal black line representing the width \
of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " w.r.t. the second coordinate (red square). Obviously, such an lcm might \
well be located above ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 ", meaning that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", 
    RowBox[{"A", " ", "\:293a", " ", "x"}], "]"}], TraditionalForm]]],
 " increases by an amount corresponding to the length of the short arrow \
pointing upwards, but at the same time the width of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " w.r.t. the second coordinate also needs to decrease by an amount \
corresponding to the length of the short arrow pointing downwards."
}], "Text",
 CellChangeTimes->{{3.615525398197647*^9, 3.6155256028156*^9}, {
  3.615525633393255*^9, 3.6155257948757*^9}, {3.615786653756124*^9, 
  3.615786668722226*^9}},
 CellID->1441855581],

Cell[BoxData[
 GraphicsBox[{
   {EdgeForm[GrayLevel[0]], FaceForm[{RGBColor[0, 0.666667, 1], Opacity[
    0.1]}], RectangleBox[{0.14959579781062404`, 0.46198830409357067`}, \
{0.7062345507743534, 0.8538011695906449}]}, 
   {JoinForm[{"Miter", 2.0500000000000003`}], EdgeForm[{GrayLevel[0.], 
    Opacity[1.], AbsoluteThickness[1], 
    JoinForm[{"Miter", 2.0500000000000003`}]}], 
    ArrowBox[{{0.06813821909895013, 0.24879114366719876`}, {
     0.06813821909895013, 0.9835204230256968}}]}, 
   {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], 
    ArrowBox[{{0.0709302325581399, 0.24878009425575315`}, {
     1.2681524547803622`, 0.24878009425575293`}}]}, GeometricTransformationBox[
    {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
     RGBColor[1, 0, 0]], 
     DiskBox[{0.9025126536514826, 0.1560517555068591}, \
{0.00982092751647956, 0.009820927516479755}]}, {{{{1., 0.}, {0., 
      1.1192778528786218`}}, {-0.029069767441860104`, 
      0.16172515014297145`}}}], 
   StyleBox[InsetBox[
     RowBox[{
      SubscriptBox["A", "i"], 
      StyleBox[Cell["",
        CellBaseline->Baseline,
        TextAlignment->Left],
       TextAlignment->Left,
       TextJustification->0,
       Background->RGBColor[
        0.9570000762951095, 0.9600061036087587, 
         0.9700007629510948]]}], \
{0.35294579248572144`, 0.46261118676321233`}, {
     Left, Top}, {0.09527204650041765, 0.0901640492596668}, {{1., 0.}, {0., 
     1.}},
     Alignment->{Left, Top}],
    FontSize->24], 
   StyleBox[InsetBox[
     RowBox[{"x", 
      StyleBox[Cell["",
        CellBaseline->Baseline,
        TextAlignment->Left],
       TextAlignment->Left,
       TextJustification->0,
       Background->RGBColor[
        0.9570000762951095, 0.9600061036087587, 
         0.9700007629510948]]}], {0.8772660966217699, 0.3564144789006841}, {
     Left, Top}, {0.07222222222222233, 0.09016404925966678}, {{1., 0.}, {0., 
     1.}},
     Alignment->{Left, Top}],
    FontSize->24], GeometricTransformationBox[
    {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
     RGBColor[1, 0, 0]], 
     RectangleBox[{0.8891639824667059, 0.22077444888483844`}, \
{0.9146983940095537, 0.1952400373419908}]}, {{{{1., 0.}, {0., 
      1.1192778528786218`}}, {-0.029069767441860152`, 0.2289766121312757}}}], 
   GeometricTransformationBox[
    {EdgeForm[{GrayLevel[0.], Opacity[1.], AbsoluteThickness[1]}], FaceForm[
     RGBColor[0, 0.666667, 1]], 
     DiskBox[{0.4579262584988259, 0.2018534217677242}, \
{0.00982092751648006, 0.009820927516480005}]}, {{{{1., 0.}, {0., 
      1.1192778528786218`}}, {-0.029069767441860152`, 
      0.23190058873946284`}}}], 
   {AbsoluteThickness[1], StrokeForm[{GrayLevel[0.], Opacity[1.]}], EdgeForm[{
    GrayLevel[0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{1.0730296456977586`, 0.25146198830409394`}, {
     0.06941431670281979, 0.9649122807017549}}]}, 
   {AbsoluteThickness[1], StrokeForm[{RGBColor[1, 0, 0], Opacity[1.]}], 
    EdgeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{1.1684743311641361`, 0.2514619883040945}, {0.1648590021691971, 
     0.9649122807017556}}]}, 
   {AbsoluteThickness[1], StrokeForm[{GrayLevel[0.], Opacity[1.]}], EdgeForm[{
    GrayLevel[0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{0.8734634851771512, 0.3391812865497077}, {0.8734634851771512, 
     0.4590643274853805}}]}, 
   {Thickness[0.002], StrokeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], 
    EdgeForm[{RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{0.06941431670281982, 0.4619883040935677}, {1.255242227042661, 
     0.46198830409356795`}}]}, 
   {Thickness[0.002], StrokeForm[{RGBColor[1, 0, 0], Opacity[1.]}], EdgeForm[{
    RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[None], 
    LineBox[{{0.0737527114967459, 0.3362573099415215}, {1.2595806218365873`, 
     0.3362573099415217}}]}, 
   {Thickness[0.002], Arrowheads[{{0.020044720273651614`, 1, {
        GraphicsBox[{
          EdgeForm[None], 
          Dashing[{}], 
          LineBox[{{-1, 0.5}, {0, 0}, {-1, -0.5}}]}], 0}}}], StrokeForm[{
    RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[{RGBColor[0., 0., 0.], 
    Opacity[1.], Thickness[0.002]}], FaceForm[{RGBColor[0, 0.666667, 1], 
    Opacity[0.1]}], 
    ArrowBox[{{0.7780187997107739, 0.4649122807017553}, {0.7780187997107739, 
     0.3362573099415217}}]}, 
   {Thickness[0.002], Arrowheads[{{0.020044720273651614`, 1, {
        GraphicsBox[{
          EdgeForm[None], 
          Dashing[{}], 
          LineBox[{{-1, 0.5}, {0, 0}, {-1, -0.5}}]}], 0}}}], StrokeForm[{
    RGBColor[0., 0., 0.], Opacity[1.]}], EdgeForm[{RGBColor[0., 0., 0.], 
    Opacity[1.], Thickness[0.002]}], FaceForm[{RGBColor[0, 0.666667, 1], 
    Opacity[0.1]}], 
    ArrowBox[{{0.9515545914678236, 0.339181286549708}, {0.9515545914678236, 
     0.40643274853801226`}}]}},
  AspectRatio->0.6008676789587852,
  ContentSelectable->True,
  ImageMargins->{{3., 0.}, {0., 0.}},
  ImagePadding->{{0., 0.}, {0., 0.}},
  ImageSize->{461., 277.},
  PlotRange->{{0., 1.3333333333333335`}, {0.19005847953216373`, 1.}},
  PlotRangePadding->Automatic]], "Text",
 CellChangeTimes->{{3.60560130456881*^9, 3.605601309150591*^9}, 
   3.60560148437655*^9, {3.605606346223402*^9, 3.605606348528124*^9}, {
   3.61552110910792*^9, 3.615521323217483*^9}, {3.615521356258374*^9, 
   3.615521389589831*^9}, {3.615521564300829*^9, 3.615521568828444*^9}, {
   3.615521622761972*^9, 3.615521714600076*^9}, {3.615523333465935*^9, 
   3.615523556838936*^9}, {3.615523605577046*^9, 3.615523629330637*^9}, {
   3.615524714194783*^9, 3.61552476129146*^9}, {3.615524829873021*^9, 
   3.615524972631296*^9}, {3.615525005287313*^9, 3.615525035992225*^9}, {
   3.615525067996641*^9, 3.615525074560066*^9}, {3.615525127570002*^9, 
   3.615525129616955*^9}, {3.61552517954059*^9, 3.615525193997916*^9}, {
   3.615525230076402*^9, 3.615525275236545*^9}},
 TextAlignment->Center,
 CellID->482923784]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Results", "Section",
 CellChangeTimes->{{3.615181959474289*^9, 3.615181962041616*^9}},
 CellTags->"sect::MainResults",
 CellID->356961332],

Cell[TextData[{
 "In this section we state and prove the main theorems of the complexity \
analysis of Buchberger\[CloseCurlyQuote]s algorithm in the bivariate case. As \
can be seen, they are divided into the three subsections ",
 ButtonBox["M[A]+W[A] Does Not Increase",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Invariant"],
 ", ",
 ButtonBox["M[A]+W[A]\[LessEqual]2maxdeg[A]",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Bound"],
 " and ",
 ButtonBox["maxdeg[G]\[LessEqual]2maxdeg[F]",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::MainResult"],
 ". This reflects they way they were originally presented in papers and \
technical reports by Buchberger."
}], "Text",
 CellChangeTimes->{{3.615189062316002*^9, 3.615189221952533*^9}, {
  3.615189296257134*^9, 3.615189312581844*^9}, {3.615189342809443*^9, 
  3.615189365110321*^9}, {3.615189450949014*^9, 3.615189475402708*^9}},
 CellID->52869115],

Cell["\<\
Also note that most of the theorems presented in this section are in fact \
rather simple corollaries of the theorems proved in the previous section.\
\>", "Text",
 CellChangeTimes->{{3.615189388948973*^9, 3.61518943512756*^9}},
 CellID->1355349683],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " Does Not Increase"
}], "Subsection",
 CellChangeTimes->{{3.596370465703707*^9, 3.596370487571546*^9}, {
   3.599892156578454*^9, 3.599892163220907*^9}, {3.59989219354854*^9, 
   3.599892202345566*^9}, {3.599892667564456*^9, 3.599892672123718*^9}, {
   3.613817522492873*^9, 3.613817523119793*^9}, 3.613885817524959*^9},
 CellTags->"sect::Invariant",
 CellID->1255092584],

Cell[TextData[{
 "This is one of the main theorems on the complexity of Buchberger\
\[CloseCurlyQuote]s algorithm in the bivariate case, as stated in ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 ": The quantity ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " does not increase in the course of Buchberger\[CloseCurlyQuote]s algorithm \
(with chain criterion), if ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " is considered the tuple of exponent vectors of the leading monomials in \
the current basis in ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ". As already indicated in section ",
 ButtonBox["Basic Notions / Categories",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::DefinitionCategories"],
 ", this result only provides an actual bound of the number of elementary \
operations in ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 " if the totally-ordered Abelian monoid ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " is instantiated by some discrete domain with least element, like ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.596370656108682*^9, 3.596370729871311*^9}, {
  3.59637079700174*^9, 3.596370797070632*^9}, {3.596370982148037*^9, 
  3.59637098218082*^9}, {3.596371028121195*^9, 3.596371035272396*^9}, {
  3.596371191523393*^9, 3.596371191529109*^9}, {3.596371870909039*^9, 
  3.596371926500803*^9}, {3.598253797897262*^9, 3.598253804887453*^9}, {
  3.604059663358121*^9, 3.604059844211648*^9}, {3.604059874410319*^9, 
  3.604059985654835*^9}, {3.604060034136529*^9, 3.60406003615014*^9}, {
  3.604060221670283*^9, 3.604060239657068*^9}, {3.604118344469948*^9, 
  3.604118347399361*^9}, {3.604118514959529*^9, 3.604118529767606*^9}, {
  3.615189543473211*^9, 3.615189568548824*^9}, {3.61518966200418*^9, 
  3.615189681050814*^9}},
 CellID->1014400362],

Cell["", "OpenEnvironment",
 CellID->307314990],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " does not increase\[CloseCurlyDoubleQuote]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.596368172349451*^9, 3.596368208361205*^9}, {
  3.596370493084639*^9, 3.596370493343258*^9}, {3.599892653739602*^9, 
  3.599892658990007*^9}, {3.604058960451599*^9, 3.604058962365221*^9}},
 CellID->1542489228],

Cell[BoxData[
 RowBox[{
  RowBox[{"n", "\[Equal]", "2"}], "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.60559915772879*^9, 3.605599160552943*^9}},
 CellID->1656946773],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isExpVectorTuple", "[", "A", "]"}]], 
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.59637042436002*^9, 3.596370438371085*^9}, {
   3.5963721477457*^9, 3.596372155296349*^9}, {3.598252963903893*^9, 
   3.598252964089517*^9}, 3.598329777102635*^9, {3.599891158899264*^9, 
   3.59989116274029*^9}, {3.600065328430302*^9, 3.600065332666945*^9}, {
   3.603181426342165*^9, 3.603181429143798*^9}, {3.60404776798574*^9, 
   3.604047770359271*^9}, {3.605599133537191*^9, 3.605599143263116*^9}},
 CellTags->"???",
 CellID->1004739293],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isExponentVector", "[", "x", "]"}]], 
  RowBox[{
   RowBox[{
    RowBox[{"deg", "[", "x", "]"}], 
    UnderscriptBox["\[LessEqual]", "D"], 
    RowBox[{"M", "[", "A", "]"}]}], "\[And]", 
   RowBox[{"\[Not]", 
    RowBox[{"isAbove", "[", 
     RowBox[{"x", ",", "A"}], "]"}]}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.596370520757964*^9, 3.596370530125892*^9}, {
   3.596370562222863*^9, 3.596370578550646*^9}, {3.598252968363254*^9, 
   3.598252973461396*^9}, {3.598593405378374*^9, 3.598593405712128*^9}, 
   3.599282606025327*^9, {3.59989082024125*^9, 3.599890830845575*^9}, {
   3.60111542350063*^9, 3.601115424547888*^9}, {3.604061331116803*^9, 
   3.604061336251944*^9}, {3.604836618473571*^9, 3.604836622404072*^9}, {
   3.605599145980008*^9, 3.605599171354808*^9}},
 CellTags->"???",
 CellID->1633843438],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"M", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"W", "[", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "x"}], "]"}]}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   RowBox[{"M", "[", "A", "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"W", "[", "A", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(invariant)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2102726322", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596368998570278*^9, 3.596369018302079*^9}, {
  3.596369079919278*^9, 3.596369111192611*^9}, {3.598253024946394*^9, 
  3.598253032545874*^9}, {3.598253090913661*^9, 3.598253128390848*^9}, {
  3.59869716485476*^9, 3.598697177466296*^9}, {3.599282611147875*^9, 
  3.599282616083893*^9}, {3.600752444637037*^9, 3.600752449792815*^9}, {
  3.604061337393735*^9, 3.604061338645591*^9}, {3.605599172719271*^9, 
  3.60559917631963*^9}},
 CellTags->{
  "ID:2102726322", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "invariant"},
 CellID->2102726322],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->643723384]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1917787441],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(invariant)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(invariant)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                  UnderscriptBox["\[LessEqual]", "D"], 
                  RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], "\[And]", 
                RowBox[{"\[Not]", 
                  RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"isAbove", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"M", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "D"], 
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}]}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["+", "D"], 
                 RowBox[{"W", "[", 
                   StyleBox["A", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{"isExponentVector", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  RowBox[{"deg", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                  UnderscriptBox["\[LessEqual]", "D"], 
                  RowBox[{"M", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], "\[And]", 
                RowBox[{"\[Not]", 
                  RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"isAbove", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    TagBox[")", "AutoParentheses"]}]}]}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"M", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "D"], 
                 RowBox[{"W", "[", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}]}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 RowBox[{"M", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["+", "D"], 
                 RowBox[{"W", "[", 
                   StyleBox["A", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2102726322"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2102726322-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2102726322-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2102726322", "Proof|ID:2102726322-1"},
 CellID->987040057],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p2102726322-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6149373171609*^9},
 CellID->204517399]
}, Closed]],

Cell[TextData[{
 "Knowing the results from section \[OpenCurlyDoubleQuote]Theorems\
\[CloseCurlyDoubleQuote] the proof of ",
 ButtonBox["(invariant)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2102726322"],
 " is by no means difficult. The overall strategy is the same as in ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 ", although some modifications and simplifications were applied:"
}], "Text",
 CellChangeTimes->{{3.604060084412119*^9, 3.60406012978147*^9}, {
   3.604060347659611*^9, 3.604060420585194*^9}, {3.604060716936477*^9, 
   3.604060718022776*^9}, {3.604060801596133*^9, 3.604060801611026*^9}, {
   3.604061674051232*^9, 3.604061674606942*^9}, 3.605599318072529*^9, {
   3.615189712339735*^9, 3.615189730058745*^9}},
 CellID->1493025671],

Cell[CellGroupData[{

Cell[TextData[{
 "In ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 " a strong emphasis is on the notion of ",
 StyleBox["contour",
  FontSlant->"Italic"],
 ", a notion which turned out to be completely irrelevant for the statements \
and proofs of both ",
 ButtonBox["(invariant)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2102726322"],
 " and ",
 ButtonBox["(bound)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1990359651"],
 "."
}], "Item",
 CellChangeTimes->{{3.604060427908289*^9, 3.604060521833078*^9}, {
  3.604060814761954*^9, 3.604060814778733*^9}, {3.604061654732355*^9, 
  3.604061662809774*^9}, {3.615189796019922*^9, 3.615189829300707*^9}},
 CellID->714527305],

Cell[TextData[{
 "As mentioned already in section ",
 ButtonBox["Lemmata on Specific Notions / Special Case n=2 / Cover of Space \
of Exponent Vector",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Cover"],
 ", a different cover/partition was chosen in this report. This has also \
something to do with the emphasis on contours in ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 " and in fact reduces the number of cases that have to be considered from \
three to two (essentially, i.e. by exploiting symmetry)."
}], "Item",
 CellChangeTimes->{{3.604060427908289*^9, 3.604060692802177*^9}, {
  3.604060739446286*^9, 3.604060756527116*^9}, {3.604061593234331*^9, 
  3.60406160205443*^9}, {3.615189900054249*^9, 3.615189933125608*^9}},
 CellID->1061151323]
}, Open  ]],

Cell[TextData[{
 "There are some further remarks on formula ",
 ButtonBox["(invariant)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2102726322"],
 ", always assuming that ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " is the tuple of exponent vectors of leading monomials in the current basis \
",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " in ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ":"
}], "Text",
 CellChangeTimes->{{3.604060848484888*^9, 3.604060859558236*^9}, {
  3.604061199813502*^9, 3.604061225583248*^9}, {3.604061421034936*^9, 
  3.604061431549881*^9}, {3.604061626141128*^9, 3.604061628337635*^9}, {
  3.615189802475899*^9, 3.615189803227017*^9}},
 CellID->1082726349],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " plays the role of the exponent vector of the leading monomial of the new \
polynomial ",
 Cell[BoxData[
  FormBox["h", TraditionalForm]]],
 " that is added to ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " in line 11 of ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 "."
}], "Item",
 CellChangeTimes->{{3.604060867284621*^9, 3.604060931827997*^9}, {
   3.604060987426975*^9, 3.604060988747237*^9}, {3.604061176811459*^9, 
   3.604061193633928*^9}, {3.604061230845614*^9, 3.604061231141072*^9}, {
   3.60406134244968*^9, 3.604061344866696*^9}, {3.604061451754997*^9, 
   3.604061451764041*^9}, 3.613817512730195*^9},
 CellID->44989876],

Cell[TextData[{
 "Hence, we know that its degree must be ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LessEqual]", 
    RowBox[{"M", "[", "A", "]"}]}], TraditionalForm]]],
 " (in fact it must even be strictly less), because\[LineSeparator](i) ",
 Cell[BoxData[
  FormBox["h", TraditionalForm]]],
 " originated from the reduction of the S-polynomial of two polynomials ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 " in lines 7 and 8,\[LineSeparator](ii) we use a ",
 StyleBox["graded",
  FontSlant->"Italic"],
 " order of monomials, and\[LineSeparator](iii) the degree of the least \
common multiple of ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 " certainly is ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LessEqual]", 
    RowBox[{"M", "[", "A", "]"}]}], TraditionalForm]]],
 ", since the chain criterion holds for ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.604060867284621*^9, 3.604061167069658*^9}, 
   3.604061239751203*^9, {3.604061349470649*^9, 3.604061373402476*^9}, {
   3.604061706019273*^9, 3.604061756876307*^9}, {3.60483666211748*^9, 
   3.604836670165178*^9}, {3.615189968132219*^9, 3.615189968354516*^9}},
 CellID->496865089],

Cell[TextData[{
 "We also know that ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " cannot lie above ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ", for otherwise there would be an element of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " dividing it, which in other words means that ",
 Cell[BoxData[
  FormBox["h", TraditionalForm]]],
 " is not irreducible."
}], "Item",
 CellChangeTimes->{{3.604060867284621*^9, 3.604061069313167*^9}, {
  3.604061254651788*^9, 3.604061307318953*^9}, {3.604061384500349*^9, 
  3.604061389785069*^9}},
 CellID->1437698621]
}, Open  ]],

Cell[TextData[{
 "If in some iteration of the main loop of ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 " we do not arrive at lines 10 and 11, i.e. the basis ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " is not completed by a new element, then the invariant holds vacuously ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", "A"}], TraditionalForm]]],
 " remains unchanged)."
}], "Text",
 CellChangeTimes->{{3.604061478784103*^9, 3.604061561716188*^9}, {
  3.615189995627651*^9, 3.615189998383379*^9}},
 CellID->1200874253]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], "\[LessEqual]", 
   RowBox[{"2", " ", 
    RowBox[{"maxdeg", "[", "A", "]"}]}]}], TraditionalForm]]]], "Subsection",
 CellChangeTimes->{{3.599891943917156*^9, 3.599891945546906*^9}, {
  3.599892000966938*^9, 3.59989200844095*^9}, {3.599892169810305*^9, 
  3.599892210851052*^9}, {3.599899990123857*^9, 3.599899991805805*^9}, {
  3.599904725399062*^9, 3.599904725399195*^9}},
 CellTags->"sect::Bound",
 CellID->485945541],

Cell[TextData[{
 "This is now the second main theorem on complexity, as stated in ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 ": The quantity ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 ", which was shown not to increase in the section above, can be bounded by \
",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    RowBox[{"maxdeg", "[", "A", "]"}]}], TraditionalForm]]],
 ". Again, ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " has to be considered the tuple of exponent vectors of the leading \
monomials in the current basis ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " in ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ".\nAlthough ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " would in principle also serve as a good bound itself (both ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", "A", "]"}], TraditionalForm]]],
 " can be computed rather easily knowing ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 "), the initial goal of the complexity analysis was to give a bound in terms \
of the degrees of the input polynomials of ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ".\nAlso note that the bound ",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    RowBox[{"maxdeg", "[", "A", "]"}]}], TraditionalForm]]],
 " is tight, meaning that there are (infinitely) many examples of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " where one has equality of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    RowBox[{"maxdeg", "[", "A", "]"}]}], TraditionalForm]]],
 "; Take, for instance, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"A", "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"2", ",", "5"}], ")"}], ",", 
      RowBox[{"(", 
       RowBox[{"6", ",", "1"}], ")"}]}], "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 " over the ground domain ",
 Cell[BoxData[
  FormBox[
   RowBox[{"D", "=", 
    SubscriptBox["\[DoubleStruckCapitalN]", "0"]}], TraditionalForm]]],
 ".\n\nAs usual, let ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " be a ",
 StyleBox["non-empty",
  FontSlant->"Italic"],
 " tuple of exponent vectors in two dimensions:"
}], "Text",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmViYGAQB2IQHVzFHNNx+pXjiYfWqSDakW9BLoie57QITO/Yc6QF
RLufPAqmrzxo7gLRFdszukH04v0f5oHosncQ+qbY3c8gegrbhV8gOn7bnjvy
5185sqTNuQuiZwY3n7K/BKTDr1wF0dueOM1yBtJngkPAtMRZhjUgms+2dS2I
FthX/glEi++5DqYzd0u7uwBpXyFNDxB9Rv36b1cgPSNL6i+IDr1/Odf/1itH
e8GZtSA6bsKLRhB9/iuEnlvyqh1Ez1G3WQKinXkPbQbR7Ju27gTRCsdi9oPo
ZyW9R0G03qTiYxFAupw/4gyIfjBnX53S3VeO7/fcaATRy7hbMy+8eeX4df3S
MhBt9kTiwQcgHSzy/BWI3tZ3sD73/SvHF7e520E0ADQ61dA=
  "],
 CellID->1437968482],

Cell[BoxData[
 RowBox[{
  RowBox[{"n", "\[Equal]", "2"}], "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.598252730147874*^9, 3.598252744133661*^9}, {
   3.598252775962493*^9, 3.598252776074589*^9}, 3.598253014232589*^9, 
   3.599282599620074*^9, 3.599900060435769*^9, 3.599904749174281*^9, {
   3.605599253218846*^9, 3.605599258254866*^9}},
 CellID->746038455],

Cell["", "Text",
 CellID->1576662597],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isExpVectorTuple", "[", "A", "]"}]], 
  RowBox[{"1", 
   UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
   RowBox[{
   "\[LeftBracketingBar]", "A", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.59637042436002*^9, 3.596370438371085*^9}, {
   3.5963721477457*^9, 3.596372155296349*^9}, {3.598252963903893*^9, 
   3.598252964089517*^9}, 3.598329777102635*^9, {3.599891158899264*^9, 
   3.59989116274029*^9}, {3.600065348312459*^9, 3.600065352356407*^9}, {
   3.603181436467237*^9, 3.603181440960043*^9}, {3.604047776372204*^9, 
   3.604047778594484*^9}, {3.605599261017014*^9, 3.605599264747295*^9}},
 CellTags->"???",
 CellID->954349703],

Cell[TextData[{
 "Since multiplication is not defined in ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ", the bound is not formulated as ",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    RowBox[{"maxdeg", "[", "A", "]"}]}], TraditionalForm]]],
 " but rather as ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"maxdeg", "[", "A", "]"}], "+", 
    RowBox[{"maxdeg", "[", "A", "]"}]}], TraditionalForm]]],
 ".\nFormula ",
 ButtonBox["(bound)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1990359651"],
 " follows readily from the theorem below, together with formula ",
 ButtonBox["(M\[LessEqual]L)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:341822661"],
 " in section ",
 ButtonBox["Theorems / General Case",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::TheoremsGeneral"],
 ":"
}], "Text",
 CellChangeTimes->{{3.60412027329952*^9, 3.604120317139632*^9}, {
   3.604120354352381*^9, 3.604120386398115*^9}, 3.605599298702115*^9, {
   3.615190094884689*^9, 3.615190110347474*^9}, {3.615190144644058*^9, 
   3.615190230190575*^9}},
 CellID->2127870379],

Cell["", "OpenEnvironment",
 CellID->247339406],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "[", 
     RowBox[{"L", "[", "A", "]"}], "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " can be bounded by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    RowBox[{"maxdeg", "[", "A", "]"}]}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599900580033329*^9, 3.599900583508669*^9}, {
  3.599900625911455*^9, 3.599900692640759*^9}, {3.599905093927323*^9, 
  3.599905095712719*^9}},
 CellID->1208461823],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"deg", "[", 
    RowBox[{"L", "[", "A", "]"}], "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"W", "[", "A", "]"}]}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   RowBox[{"maxdeg", "[", "A", "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"maxdeg", "[", "A", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(100)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:724101256", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.599900700683196*^9, 3.599900732831045*^9}, {
  3.599904876792496*^9, 3.599904882616439*^9}, {3.605599277955549*^9, 
  3.605599283160148*^9}},
 CellTags->{
  "ID:724101256", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "100"},
 CellID->724101256],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->733284948]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->498946658],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(100)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(100)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{
               RowBox[{"deg", "[", 
                 RowBox[{"L", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{"W", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               RowBox[{"maxdeg", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{"maxdeg", "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{
               RowBox[{"deg", "[", 
                 RowBox[{"L", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{"W", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               RowBox[{"maxdeg", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{"maxdeg", "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:724101256"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p724101256-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p724101256-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:724101256", "Proof|ID:724101256-1"},
 CellID->205519987],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p724101256-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellID->835900380]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->957582836],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "A", "]"}], "+", 
    RowBox[{"W", "[", "A", "]"}]}], TraditionalForm]]],
 " can be bounded by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    RowBox[{"maxdeg", "[", "A", "]"}]}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.599892681582202*^9, 3.599892696161255*^9}, {
  3.599900646542795*^9, 3.599900647140486*^9}, {3.599905098431268*^9, 
  3.599905124223662*^9}},
 CellID->1193318642],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"M", "[", "A", "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"W", "[", "A", "]"}]}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   RowBox[{"maxdeg", "[", "A", "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"maxdeg", "[", "A", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(bound)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1990359651", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59989276059418*^9, 3.599892805087654*^9}, {
  3.599900065874013*^9, 3.599900067857441*^9}, {3.599904756747322*^9, 
  3.599904765026015*^9}, {3.605599272365912*^9, 3.605599276223223*^9}},
 CellTags->{
  "ID:1990359651", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "bound"},
 CellID->1990359651],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2055477639]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->292236524],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(bound)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(bound)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{
               RowBox[{"M", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{"W", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               RowBox[{"maxdeg", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{"maxdeg", "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isExpVectorTuple", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"1", 
              UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["A", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}], 
                TagBox[")", "AutoParentheses"]}]}]], 
           RowBox[{
             RowBox[{
               RowBox[{"M", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{"W", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], 
             UnderscriptBox["\[LessEqual]", "D"], 
             RowBox[{
               RowBox[{"maxdeg", "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "D"], 
               RowBox[{"maxdeg", "[", 
                 StyleBox["A", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1990359651"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1990359651-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1990359651-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1990359651", "Proof|ID:1990359651-1"},
 CellID->2132280415],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1990359651-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {33., 39.}}]], "ProofInfoBody",
 CellChangeTimes->{3.614937329112316*^9},
 CellID->1143064906]
}, Closed]],

Cell[TextData[StyleBox["Remark:",
 FontFamily->"Bitstream Charter",
 FontSize->14,
 FontWeight->"Bold"]], "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.603522907354292*^9, 3.603522909426996*^9}, {
  3.603522955835897*^9, 3.603522963082172*^9}, {3.603793011527348*^9, 
  3.603793024231905*^9}, {3.604121445370575*^9, 3.604121448015606*^9}},
 CellID->618421757],

Cell[TextData[{
 "For arbitrary ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " it is very easy to see that one always gets at least ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"M", "[", "A", "]"}], "+", 
     RowBox[{"W", "[", "A", "]"}]}], "\[LessEqual]", 
    RowBox[{"3", 
     RowBox[{"maxdeg", "[", "A", "]"}]}]}], TraditionalForm]]],
 ". This is simply because ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " can always be bounded by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    RowBox[{"maxdeg", "[", "A", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"W", "[", "A", "]"}], TraditionalForm]]],
 " can always be bounded by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"maxdeg", "[", "A", "]"}], TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.599908318315582*^9, 3.599908420153387*^9}, {
   3.599908452471986*^9, 3.599908456153075*^9}, 3.604120629238266*^9, {
   3.604121445370662*^9, 3.604121465885604*^9}},
 CellID->1618954571],

Cell[TextData[{
 "The overall strategy for proving ",
 ButtonBox["(bound)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1990359651"],
 " is the same as in the proof of (5) in ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 ": There, basically the same as formula ",
 ButtonBox["(bound)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1990359651"],
 " is stated for ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " being a contour. The crucial observation that allows to drastically \
shorten the proof of the main theorem in ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 " is the following: ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " does ",
 StyleBox["not",
  FontSlant->"Italic"],
 " have to be a contour in order for (5) to hold, but may be completely \
arbitrary. The biggest and most difficult part in the proof of the main \
theorem in ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 " is to express ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 " in terms of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["M", "C"], "[", "A", "]"}], TraditionalForm]]],
 " (subproof (2)), where ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["M", "C"], "[", "A", "]"}], TraditionalForm]]],
 " is a function similar to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", "A", "]"}], TraditionalForm]]],
 ", but only takes into account neighbouring points on the contour of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 "; This difficult elaboration can hence be skipped entirely, because there \
is absolutely no need to consider contours at all."
}], "Text",
 CellChangeTimes->{{3.604060084412119*^9, 3.60406012978147*^9}, {
  3.604060347659611*^9, 3.604060420585194*^9}, {3.604060716936477*^9, 
  3.604060718022776*^9}, {3.604060801596133*^9, 3.604060801611026*^9}, {
  3.604061674051232*^9, 3.604061674606942*^9}, {3.604120660877115*^9, 
  3.604120857661005*^9}, {3.604120916583312*^9, 3.604121114208817*^9}, {
  3.604121154181511*^9, 3.604121390895914*^9}, {3.604121485510354*^9, 
  3.604121556037774*^9}, {3.615190281580752*^9, 3.615190309034282*^9}},
 CellID->1876059057]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"maxdeg", "[", "G", "]"}], "\[LessEqual]", 
   RowBox[{"2", 
    RowBox[{"maxdeg", "[", "F", "]"}]}]}], TraditionalForm]]]], "Subsection",
 CellChangeTimes->{{3.613886039549366*^9, 3.613886059204792*^9}},
 CellTags->"sect::MainResult",
 CellID->1121720170],

Cell[TextData[{
 "Let ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " be a finite set of polynomials input to ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ", and let ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " be the resulting Gr\[ODoubleDot]bner basis. Combining the results from the \
previous sections, we are finally able to state and prove the following main \
theorem:"
}], "Text",
 CellChangeTimes->{{3.613886094773713*^9, 3.613886172977064*^9}, {
  3.61388630675481*^9, 3.613886313620106*^9}, {3.613888460082856*^9, 
  3.613888496303825*^9}},
 CellID->297111079],

Cell["", "OpenEnvironment",
 CellID->1239024718],

Cell[CellGroupData[{

Cell["Theorem: Main Theorem of Complexity Analysis", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.613886764834894*^9, 3.613886777773231*^9}, {
  3.613887350164594*^9, 3.613887350423541*^9}},
 CellID->1918271715],

Cell[BoxData[
 RowBox[{
  RowBox[{"n", "\[Equal]", "2"}], "\[Implies]"}]], "GlobalDeclaration",
 CellChangeTimes->{{3.613887416155671*^9, 3.613887419229606*^9}},
 CellID->1879526953],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isExpVectorTuple", "[", 
   RowBox[{"F", ",", "G"}], "]"}]]], "GlobalDeclaration",
 CellChangeTimes->{{3.613886786874897*^9, 3.613886805864675*^9}, {
  3.61388746324239*^9, 3.613887467492855*^9}},
 CellID->592262441],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"1", 
    UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
    RowBox[{"\[LeftBracketingBar]", "F", "\[RightBracketingBar]"}]}], 
   "\[And]", 
   RowBox[{"1", 
    UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
    RowBox[{"\[LeftBracketingBar]", "G", "\[RightBracketingBar]"}]}], 
   "\[And]", 
   RowBox[{
    UnderscriptBox["0", "D"], 
    UnderscriptBox["\[LessEqual]", "D"], 
    RowBox[{"W", "[", "G", "]"}]}], "\[And]", 
   RowBox[{
    RowBox[{
     RowBox[{"M", "[", "G", "]"}], 
     UnderscriptBox["+", "D"], 
     RowBox[{"W", "[", "G", "]"}]}], 
    UnderscriptBox["\[LessEqual]", "D"], 
    RowBox[{
     RowBox[{"M", "[", "F", "]"}], 
     UnderscriptBox["+", "D"], 
     RowBox[{"W", "[", "F", "]"}]}]}]}], "\[Implies]"}]], "GlobalDeclaration",\

 CellChangeTimes->{{3.613886810609276*^9, 3.613886871183879*^9}, {
  3.613886966220091*^9, 3.613886970624194*^9}, {3.613887138352167*^9, 
  3.613887150562748*^9}, {3.61388739865829*^9, 3.613887413417514*^9}},
 CellID->1452306355],

Cell[BoxData[
 RowBox[{
  RowBox[{"maxdeg", "[", "G", "]"}], 
  UnderscriptBox["\[LessEqual]", "D"], 
  RowBox[{
   RowBox[{"maxdeg", "[", "F", "]"}], 
   UnderscriptBox["+", "D"], 
   RowBox[{"maxdeg", "[", "F", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(main theorem)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1891911403", 
            "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity \
(no functor).nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.613886873230992*^9, 3.613886898220677*^9}, {
  3.613886969146696*^9, 3.613886972506525*^9}},
 CellTags->{
  "ID:1891911403", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "main theorem"},
 CellID->1891911403],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2069358189]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->116419994],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(main theorem)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(main theorem)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["F", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["G", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{"1", 
                 UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["F", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{"1", 
                 UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["G", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["0", "D"], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"W", "[", 
                   StyleBox["G", "ExpressionVariable"], "]"}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"M", "[", 
                    StyleBox["G", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    StyleBox["G", "ExpressionVariable"], "]"}]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   RowBox[{"M", "[", 
                    StyleBox["F", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    StyleBox["F", "ExpressionVariable"], "]"}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{"maxdeg", "[", 
                 StyleBox["G", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 RowBox[{"maxdeg", "[", 
                   StyleBox["F", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["+", "D"], 
                 RowBox[{"maxdeg", "[", 
                   StyleBox["F", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"n", "\[Equal]", "2"}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["F", "ExpressionVariable"], "]"}]}, {
               RowBox[{"isExpVectorTuple", "[", 
                 StyleBox["G", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{"1", 
                 UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["F", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{"1", 
                 UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["G", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["0", "D"], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{"W", "[", 
                   StyleBox["G", "ExpressionVariable"], "]"}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"M", "[", 
                    StyleBox["G", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    StyleBox["G", "ExpressionVariable"], "]"}]}], 
                 UnderscriptBox["\[LessEqual]", "D"], 
                 RowBox[{
                   RowBox[{"M", "[", 
                    StyleBox["F", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["+", "D"], 
                   RowBox[{"W", "[", 
                    StyleBox["F", "ExpressionVariable"], "]"}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{"maxdeg", "[", 
                 StyleBox["G", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[LessEqual]", "D"], 
               RowBox[{
                 RowBox[{"maxdeg", "[", 
                   StyleBox["F", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["+", "D"], 
                 RowBox[{"maxdeg", "[", 
                   StyleBox["F", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1891911403"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1891911403-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1891911403-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1891911403", "Proof|ID:1891911403-1"},
 CellID->233183322],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"],FileBaseName[\
CurrentValue[\"NotebookFileName\"]],\"p1891911403-1-display.m\"}]"]], 
     Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {41., 48.}}]], "ProofInfoBody",
 CellID->311475437]
}, Closed]],

Cell["\<\
Some remarks on the third and the fourth assumption in the above theorem:\
\>", "Text",
 CellChangeTimes->{{3.613887842180135*^9, 3.613887862379607*^9}},
 CellID->1678387744],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "\[LessEqual]", 
    RowBox[{"W", "[", "G", "]"}]}], TraditionalForm]]],
 " does not hold in general in totally-ordered Abelian monoids. However, the \
interesting case anyway is when the domain ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " is taken as ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 ", in which case ",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "\[LessEqual]", 
    RowBox[{"W", "[", "G", "]"}]}], TraditionalForm]]],
 " vacuously holds. Still, it has to be pointed out that there are also other \
totally-ordered Abelian monoids where ",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "\[LessEqual]", 
    RowBox[{"W", "[", "G", "]"}]}], TraditionalForm]]],
 " always is fulfilled (like, for instance, ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["\[DoubleStruckCapitalQ]", "0", "+"], 
   TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["\[DoubleStruckCapitalR]", "0", "+"], 
   TraditionalForm]]],
 ")."
}], "Item",
 CellChangeTimes->{{3.613887868845257*^9, 3.613888104793449*^9}},
 CellID->149821961],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"M", "[", "G", "]"}], "+", 
     RowBox[{"W", "[", "G", "]"}]}], "\[LessEqual]", 
    RowBox[{
     RowBox[{"M", "[", "F", "]"}], "+", 
     RowBox[{"W", "[", "F", "]"}]}]}], TraditionalForm]]],
 " follows from ",
 ButtonBox["(invariant)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2102726322"],
 " using an inductive argument: ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 " computes ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " by successively appending elements ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " satisfying the required conditions in ",
 ButtonBox["(invariant)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2102726322"],
 " (i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"deg", "[", "x", "]"}], "\[LessEqual]", 
    RowBox[{"M", "[", "G", "]"}]}], TraditionalForm]]],
 " etc.). Hence, ",
 ButtonBox["(invariant)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:2102726322"],
 " gurantess that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", "G", "]"}], "+", 
    RowBox[{"W", "[", "G", "]"}]}], TraditionalForm]]],
 " does not increase in ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ", and since ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " is initialized by ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " in the very beginning, we can conclude ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"M", "[", "G", "]"}], "+", 
     RowBox[{"W", "[", "G", "]"}]}], "\[LessEqual]", 
    RowBox[{
     RowBox[{"M", "[", "F", "]"}], "+", 
     RowBox[{"W", "[", "F", "]"}]}]}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.613887868845257*^9, 3.613888047755346*^9}, {
  3.613888141496578*^9, 3.613888359451737*^9}, {3.613888441562855*^9, 
  3.61388844158532*^9}, {3.614496996753434*^9, 3.614496997910257*^9}, {
  3.615190394834694*^9, 3.61519043032014*^9}},
 CellID->938510761]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Prover", "Section",
 CellChangeTimes->{{3.598330476073441*^9, 3.598330479540067*^9}, {
   3.60353121631317*^9, 3.603531219310287*^9}, 3.604121789389984*^9},
 CellTags->"sect::Prover",
 CellID->1903011022],

Cell[TextData[{
 "This section aims at presenting the details of the prover used for formally \
verifying the theory developed in this report. The prover will, as before, be \
referred to as \
\[OpenCurlyDoubleQuote]ComplexityProver\[CloseCurlyDoubleQuote].\nIn ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ", a prover consists of two parts:"
}], "Text",
 CellChangeTimes->{{3.604121677139974*^9, 3.604121742825842*^9}, {
  3.604121807284087*^9, 3.604121854050562*^9}, {3.61519051629214*^9, 
  3.615190521848093*^9}},
 CellID->1608846916],

Cell[CellGroupData[{

Cell["Collection of inference rules", "Item",
 CellChangeTimes->{{3.604121866165128*^9, 3.604121873718603*^9}, 
   3.604122027029031*^9, {3.615190498525467*^9, 3.615190500158017*^9}},
 CellID->2142089810],

Cell["Strategy", "Item",
 CellChangeTimes->{{3.604121866165128*^9, 3.60412187687138*^9}, 
   3.604122029754306*^9},
 CellID->565617232]
}, Open  ]],

Cell["\<\
Both parts are more or less independent of each other, and a new prover might \
in fact only provide several new inference rules but use an existing \
strategy, or vice versa. However, the ComplexityProver both provides a new \
strategy and new inference rules; All of them will be discussed in the \
remaing part of this section.
In total there are 57 inference rules divided into 10 categories. However, \
only 29 of these rules are rules especially tailored for the needs of the \
present theory exploration (Integers, Tuples, Addition in TOMs, Order \
Relation in TOMs, Minimum and Maximum, some of Logical Quantifiers). The \
remaining 28 rules are general-purpose rules of propositional and quantified \
reasoning.\
\>", "Text",
 CellChangeTimes->{{3.604121882401362*^9, 3.604122015643768*^9}, {
  3.615190510083899*^9, 3.615190524654536*^9}, {3.615273863721316*^9, 
  3.615273874756041*^9}, {3.615274682800276*^9, 3.615274880050536*^9}},
 CellID->116836426],

Cell[TextData[StyleBox["Remarks regarding notation:",
 FontFamily->"Bitstream Charter",
 FontSize->14,
 FontWeight->"Bold"]], "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603522907354292*^9, 3.603522909426996*^9}, {
   3.603522955835897*^9, 3.603522963082172*^9}, {3.603793011527348*^9, 
   3.603793024231905*^9}, {3.604122440773405*^9, 3.604122447629318*^9}, {
   3.61519262242572*^9, 3.615192644618701*^9}, {3.615194241215107*^9, 
   3.615194245548585*^9}, {3.615266413022323*^9, 3.615266418643915*^9}, {
   3.615266950452785*^9, 3.615266950712943*^9}, 3.615274197536723*^9, {
   3.615274419530756*^9, 3.615274420194926*^9}, {3.615274518405911*^9, 
   3.615274518591731*^9}, {3.615274645763265*^9, 3.615274669080743*^9}},
 CellID->1090859055],

Cell[TextData[{
 "In the informal descriptions of some of the inference rules, ranges of \
bound variables are given as something like ",
 Cell[BoxData[
  FormBox[
   RowBox[{"R", "[", "i", "]"}], TraditionalForm]]],
 " (where ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 " is the bound variable). There, the ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 " stands for an arbitary set/step/predicate/simple range. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"RC", "[", "i", "]"}], TraditionalForm]]],
 " then refers to the corresponding range condition, e.g. in case of a step \
range ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "a"}], ",", "\[Ellipsis]", ",", "b"}], 
   TraditionalForm]]],
 " the corresponding range condition is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]]], 
    RowBox[{"[", "i", "]"}]}], TraditionalForm]]],
 "."
}], "Item",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.615274206778811*^9, 3.615274422044468*^9}, 
   3.615274649272402*^9},
 CellID->550862969],

Cell["\<\
The term \[OpenCurlyDoubleQuote]occurrence\[CloseCurlyDoubleQuote] always \
refers to the occurrence of a particular expression somewhere in the current \
proof situation, be it as a whole formula or as part of a formula. \
\[OpenCurlyDoubleQuote]Formula\[CloseCurlyDoubleQuote], on the other hand, \
always refers to whole formulas appearing in the proof situation.\
\>", "Item",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.615274206778811*^9, 3.615274520695787*^9}, 
   3.61527464927248*^9},
 CellID->1857320962],

Cell[TextData[{
 "Tuples where all elements are explicitly enumerated, as for instance in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 ", are called ",
 StyleBox["enumeration",
  FontSlant->"Italic"],
 " tuples. Tuples that are given in the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"t", 
     UnderscriptBox["|", 
      RowBox[{"R", "[", "i", "]"}]], "\[Phi]"}], "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 ", i.e. tuples that are constructed by a quantifier, are called ",
 StyleBox["abstraction",
  FontSlant->"Italic"],
 " tuples (following the notion of abstraction terms from set theory)."
}], "Item",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.615274206778811*^9, 3.615274649272547*^9}},
 CellID->824107932],

Cell[CellGroupData[{

Cell["Strategy", "Subsection",
 CellChangeTimes->{{3.604121762633619*^9, 3.604121763949824*^9}},
 CellID->78883476],

Cell["\<\
The strategy used to guide the proof search can be described as follows:\
\>", "Text",
 CellChangeTimes->{{3.615269693518047*^9, 3.615269719535872*^9}, {
  3.615273800860315*^9, 3.615273808558532*^9}},
 CellID->2054152417],

Cell[CellGroupData[{

Cell[TextData[{
 "First, try all non-interactive inference rules with priority ",
 Cell[BoxData[
  FormBox[
   RowBox[{"<", "5"}], TraditionalForm]]],
 ". If several rules are applicable, only apply the ",
 StyleBox["first",
  FontSlant->"Italic"],
 " one (i.e. the one with lowest priority, and in case there are several with \
lowest priority, the one that comes first in the list of rules). Hence, no \
alternatives are created."
}], "Item",
 CellChangeTimes->{{3.615269725320968*^9, 3.61526974306926*^9}, {
  3.615269781701885*^9, 3.615269883239532*^9}},
 CellID->1458172368],

Cell[TextData[{
 "If none of the non-interactive rules with priority ",
 Cell[BoxData[
  FormBox[
   RowBox[{"<", "5"}], TraditionalForm]]],
 " is applicable, try all non-interactive rules with priority ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[GreaterEqual]", "5"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LessEqual]", "90"}], TraditionalForm]]],
 ". In case that several rules are applicable, apply all of them and create \
alternative branches."
}], "Item",
 CellChangeTimes->{{3.615269725320968*^9, 3.61526974306926*^9}, {
  3.615269781701885*^9, 3.615269961987569*^9}},
 CellID->2016699663],

Cell[TextData[{
 "If still no rule could be applied until now, try all the remaining \
non-interactive rules with priority ",
 Cell[BoxData[
  FormBox[
   RowBox[{">", "90"}], TraditionalForm]]],
 ". Again, apply all applicable ones and create alternative branches."
}], "Item",
 CellChangeTimes->{{3.615269725320968*^9, 3.61526974306926*^9}, {
  3.615269781701885*^9, 3.615270047018383*^9}},
 CellID->1058238047],

Cell["\<\
Finally, if no non-interactive rule could be applied, try the interactive \
ones. If several rules are applicable, the user may choose the one that \
should actually be applied.\
\>", "Item",
 CellChangeTimes->{{3.615269725320968*^9, 3.61526974306926*^9}, {
  3.615269781701885*^9, 3.615270058318071*^9}, {3.615273758828892*^9, 
  3.615273794401494*^9}, {3.615273825425659*^9, 3.615273826898902*^9}},
 CellID->260116386]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Propositional Logic (13)", "Subsection",
 CellChangeTimes->{{3.615190807769428*^9, 3.615190833948463*^9}, 
   3.615190931739522*^9},
 CellID->538287563],

Cell[CellGroupData[{

Cell["\<\
\[OpenCurlyDoubleQuote]goalInKB\[CloseCurlyDoubleQuote]: If the current goal \
appears in the current list of assumptions, the proof is finished.\
\>", "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268158859027*^9}},
 CellID->608926717],

Cell["\<\
\[OpenCurlyDoubleQuote]falseInKB\[CloseCurlyDoubleQuote]: If the current \
knowledge base conatins the truth constant False, the proof is finished.\
\>", "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268223953125*^9}},
 CellID->917489538],

Cell["\<\
\[OpenCurlyDoubleQuote]trueGoal\[CloseCurlyDoubleQuote]: If the current proof \
goal is the truth constant True, the proof is finished.\
\>", "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268242009092*^9}},
 CellID->527072755],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]contradictionKB\[CloseCurlyDoubleQuote]: If the \
current knowledge base is evidently contradictory, i.e. contains both ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", "\[Phi]"}], TraditionalForm]]],
 " for some formula ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 ", then the proof is finished."
}], "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268283520005*^9}},
 CellID->86175489],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]notGoal\[CloseCurlyDoubleQuote]: If the current goal \
is a negated formula ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", "\[Phi]"}], TraditionalForm]]],
 ", then assume ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " and prove False."
}], "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268331207144*^9}},
 CellID->1110939375],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]andGoal\[CloseCurlyDoubleQuote]: If the current goal \
is a conjunction ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Phi]", "1"], "\[And]", 
    SubscriptBox["\[Phi]", "2"], "\[And]", "\[Ellipsis]", "\[And]", 
    SubscriptBox["\[Phi]", "n"]}], TraditionalForm]]],
 ", then prove each individual ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "i"], TraditionalForm]]],
 " in a subproof, where the ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 "-th subproof (showing ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "i"], TraditionalForm]]],
 ") already has all of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "1"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "2"], TraditionalForm]]],
 ", ..., ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", 
    RowBox[{"i", "-", "1"}]], TraditionalForm]]],
 " in the list of assumptions."
}], "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268447407612*^9}},
 CellID->333958307],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]andKB\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Phi]", "1"], "\[And]", 
    SubscriptBox["\[Phi]", "2"], "\[And]", "\[Ellipsis]", "\[And]", 
    SubscriptBox["\[Phi]", "n"]}], TraditionalForm]]],
 " is known, then split the conjunction into the individual conjuncts and \
continue with all of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "1"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "2"], TraditionalForm]]],
 ", ..., ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "n"], TraditionalForm]]],
 " in the list of assumptions."
}], "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268518020502*^9}},
 CellID->225969300],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]orGoal\[CloseCurlyDoubleQuote]: If the current goal \
is a disjunction ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Phi]", "1"], "\[Or]", 
    SubscriptBox["\[Phi]", "2"], "\[Or]", "\[Ellipsis]", "\[Or]", 
    SubscriptBox["\[Phi]", "n"]}], TraditionalForm]]],
 ", then assume all of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    SubscriptBox["\[Phi]", "1"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    SubscriptBox["\[Phi]", "2"]}], TraditionalForm]]],
 ", ..., ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    SubscriptBox["\[Phi]", 
     RowBox[{"n", "-", "1"}]]}], TraditionalForm]]],
 " and prove ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "n"], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268599018849*^9}},
 CellID->5933658],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]orGoalInteractive\[CloseCurlyDoubleQuote]: Similar \
as \[OpenCurlyDoubleQuote]orGoal\[CloseCurlyDoubleQuote], but offer the user \
the opportunity to interactively choose the disjunct ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "i"], TraditionalForm]]],
 " to prove."
}], "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268643765103*^9}},
 CellID->832000632],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]orKB\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Phi]", "1"], "\[Or]", 
    SubscriptBox["\[Phi]", "2"], "\[Or]", "\[Ellipsis]", "\[Or]", 
    SubscriptBox["\[Phi]", "n"]}], TraditionalForm]]],
 " is known, distinguish ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " cases (i.e. make ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " subproofs), where in the ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 "-th subproof formula ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "i"], TraditionalForm]]],
 " is known insetad of the whole disjunction."
}], "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268594212729*^9}, {
  3.615268664464856*^9, 3.615268736494768*^9}},
 CellID->413197361],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]implGoalDirect\[CloseCurlyDoubleQuote]: If the \
current goal is of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "\[Implies]", "\[Psi]"}], TraditionalForm]]],
 ", add ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " to the knowledge base and continue proving ",
 Cell[BoxData[
  FormBox["\[Psi]", TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268594212729*^9}, {
  3.615268664464856*^9, 3.615268781818237*^9}},
 CellID->1765219953],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]equivGoal\[CloseCurlyDoubleQuote]: If the current \
goal is of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "\[DoubleLeftRightArrow]", "\[Psi]"}], TraditionalForm]]],
 ", show both of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "\[Implies]", "\[Psi]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Psi]", "\[Implies]", "\[Phi]"}], TraditionalForm]]],
 " in two subproofs."
}], "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268594212729*^9}, {
  3.615268664464856*^9, 3.615268830400582*^9}},
 CellID->1448323563],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]equivKB\[CloseCurlyDoubleQuote]: Replaces \
equivalences ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "\[DoubleLeftRightArrow]", "\[Psi]"}], TraditionalForm]]],
 " in the knowledge base by the two formulas ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "\[Implies]", "\[Psi]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Psi]", "\[Implies]", "\[Phi]"}], TraditionalForm]]],
 ". This rule has to be used only very rarely due to the powerful rewriting \
capabilities of the prover that are able to deal with equivalences in a very \
efficient way. It basically only has to be used if, for some reason, the \
\[OpenCurlyDoubleQuote]elementarySubstitution\[CloseCurlyDoubleQuote]-rule is \
deactivated."
}], "Item",
 CellChangeTimes->{{3.615268133571582*^9, 3.615268594212729*^9}, {
  3.615268664464856*^9, 3.615268989894664*^9}, {3.615269028366948*^9, 
  3.615269032407586*^9}},
 CellID->1522961883]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Miscellaneous (3)", "Subsection",
 CellChangeTimes->{{3.60014846233428*^9, 3.600148465600321*^9}, {
   3.603175300207177*^9, 3.603175300764675*^9}, 3.614314640034643*^9, 
   3.61431477287776*^9, 3.61519077309146*^9},
 CellID->1059323634],

Cell[CellGroupData[{

Cell["\<\
\[OpenCurlyDoubleQuote]expandNegationGoal\[CloseCurlyDoubleQuote]: Expand \
negations in the current goal, i.e. transform negated disjunctions into \
conjunctions, negated existential quantifiers into universal ones, etc.\
\>", "Item",
 CellChangeTimes->{{3.600148475532892*^9, 3.600148580893873*^9}, {
   3.60058575459615*^9, 3.600585792658378*^9}, {3.60059576857011*^9, 
   3.600595768718341*^9}, {3.601120300352522*^9, 3.601120301580128*^9}, {
   3.601120744031234*^9, 3.60112075711463*^9}, {3.601195575880556*^9, 
   3.601195575889746*^9}, 3.60172586824445*^9, {3.601871401957421*^9, 
   3.601871411255443*^9}, {3.603175297498109*^9, 3.603175298402373*^9}, {
   3.603601153038665*^9, 3.603601159955712*^9}, 3.603688014020759*^9, {
   3.604045337403222*^9, 3.604045340186588*^9}, {3.606017297664663*^9, 
   3.606017298724373*^9}, {3.606037034692231*^9, 3.606037035707945*^9}, {
   3.615190645126857*^9, 3.615190656898645*^9}, {3.615190691102536*^9, 
   3.615190712614108*^9}},
 CellID->1967607739],

Cell["\<\
\[OpenCurlyDoubleQuote]expandNegationKB\[CloseCurlyDoubleQuote]: Expand \
negations in the current knowledge base, in analogy to \
\[OpenCurlyDoubleQuote]expandNegationGoal\[CloseCurlyDoubleQuote].\
\>", "Item",
 CellChangeTimes->{{3.600148475532892*^9, 3.600148580893873*^9}, {
   3.60058575459615*^9, 3.600585792658378*^9}, {3.60059576857011*^9, 
   3.600595768718341*^9}, {3.601120300352522*^9, 3.601120301580128*^9}, {
   3.601120744031234*^9, 3.60112075711463*^9}, {3.601195575880556*^9, 
   3.601195575889746*^9}, 3.60172586824445*^9, {3.601871401957421*^9, 
   3.601871411255443*^9}, {3.603175297498109*^9, 3.603175298402373*^9}, {
   3.603601153038665*^9, 3.603601159955712*^9}, 3.603688014020759*^9, {
   3.604045337403222*^9, 3.604045340186588*^9}, {3.606017297664663*^9, 
   3.606017298724373*^9}, {3.606037034692231*^9, 3.606037035707945*^9}, {
   3.615190645126857*^9, 3.615190656898645*^9}, {3.615190691102536*^9, 
   3.61519074668846*^9}, {3.615191504529729*^9, 3.615191507036328*^9}},
 CellID->1509344252],

Cell[TextData[{
 StyleBox["\[OpenCurlyDoubleQuote]exchangeGoalKBInteractive\
\[CloseCurlyDoubleQuote]: Offer the user the opportunity to interactively \
exchange the current goal and an assumption: If the goal ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " is exchanged with assumption ",
 Cell[BoxData[
  FormBox["\[Psi]", TraditionalForm]]],
 ", then the new goal becomes ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", "\[Psi]"}], TraditionalForm]]],
 " and formula ",
 Cell[BoxData[
  FormBox["\[Psi]", TraditionalForm]]],
 " in the knowledge base is replaced by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", "\[Phi]"}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.598595360836577*^9, 3.598595584656642*^9}, {
  3.598595880178237*^9, 3.598595894749849*^9}, {3.598596276104618*^9, 
  3.598596276498315*^9}, {3.599812883090181*^9, 3.599812892118937*^9}, {
  3.599976707489124*^9, 3.599976707499688*^9}, {3.601184274564189*^9, 
  3.601184286699556*^9}, {3.601874435803879*^9, 3.601874464381835*^9}, {
  3.602391711843116*^9, 3.60239211728956*^9}, {3.602392273229614*^9, 
  3.602392331155631*^9}, {3.602585696175345*^9, 3.602585724789441*^9}, {
  3.615190752703684*^9, 3.615190752934894*^9}, {3.615191367796549*^9, 
  3.615191493547308*^9}, {3.615195874461551*^9, 3.61519587528134*^9}},
 CellID->624399819]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Integers (2)", "Subsection",
 CellChangeTimes->{{3.603175138512284*^9, 3.603175139672855*^9}, {
   3.60317525025398*^9, 3.603175253187499*^9}, 3.609823423505788*^9, 
   3.614314641320505*^9, {3.615190847294507*^9, 3.615190849612111*^9}, {
   3.615190934246938*^9, 3.615190935243719*^9}},
 CellID->1361386280],

Cell[CellGroupData[{

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]integerIntervalKB\[CloseCurlyDoubleQuote]:",
 StyleBox[" If ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]]], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" is known, for some ground terms ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 StyleBox[":\[LineSeparator]If both ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["a", TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["b", TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" are concrete integers, replace ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]]], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" by ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and add ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"x", "\[Equal]", "a"}], "\[Or]", 
      RowBox[{"x", "\[Equal]", 
       RowBox[{"a", "+", "1"}]}], "\[Or]"}], "..."}], "\[Or]", 
    RowBox[{"x", "\[Equal]", "b"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" to the knowledge base.",
  FontColor->GrayLevel[0]],
 "\[LineSeparator]Otherwise,",
 StyleBox[" if ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", 
    RowBox[{"-", "\[Infinity]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "\[NotEqual]", 
    RowBox[{"+", "\[Infinity]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[", remove ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]]], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and add ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[", ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   StyleBox[
    RowBox[{"a", 
     UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], "x"}],
    FontWeight->"Plain"], TraditionalForm]], "Input",
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   StyleBox[
    RowBox[{"x", 
     UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], "b"}],
    FontWeight->"Plain"], TraditionalForm]], "Input",
  FontColor->GrayLevel[0]],
 StyleBox[" to the knowledge base instead.",
  FontColor->GrayLevel[0]],
 StyleBox["\[LineSeparator]",
  FontColor->RGBColor[1, 0, 0]],
 StyleBox["The reason why we cannot let computation do the job is because \
computation would translate ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 " into ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isInteger", "[", "x", "]"}], TraditionalForm]]],
 "; This we do not want!"
}], "Item",
 CellChangeTimes->CompressedData["
1:eJwlz1sog2EAxvGvkUaOWyRqWdTcyKEmuRivYSE5rcasyTkXiBgRNudDJBcU
cyinpCRCalwScuGYNLTWUh9SoxDF+7wunn43z81fWlKbWy7gOE5CBxu86n4t
5zwxT5o996gr4ryusQue1Pr590ONQlQ5QV36CWBaVUM9E5c8qRrdYN7/cnvQ
a3V+H3ZH/KRaqNNvoVnQKWzWwKPyf9uTewzQV/DBnLU4WuDma6AJ+kk4lds1
/Rnf0+BiZ4vc84YnWeO6OJiwcmrvoJq0z8wg2bHVRH1SnjHrM/U2GLv+/gLl
tkFBJ0yucofxInUYbDLkRcGp0ooiOBPezWwc0bt2UcXDUm+o0m4R+LDTq4Qe
BTEZ0FF2WwhD1Ibszzue9IUl5sBqFzcNdFXL86Fxeybpi3rn72TutmrXgu9p
R9vCLlw/KNyHPiOHJ3CuOiT9wc6TkgJ9LixekuXDN8UVM8WwHBjt4IkwTRYE
v3VGCXyyDTCnIpsVNa88eVSKU+AfHlAGgg==
  "],
 CellID->521411084],

Cell[TextData[{
 StyleBox["\[OpenCurlyDoubleQuote]integerIntervalGoal\[CloseCurlyDoubleQuote]:\
 If ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]]], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" has to be shown (for some ground terms ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 StyleBox["with ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", 
    RowBox[{"-", "\[Infinity]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "\[NotEqual]", 
    RowBox[{"+", "\[Infinity]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox["): Show ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
     RowBox[{"[", "x", "]"}]}], "\[And]", 
    RowBox[{"a", 
     UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], "x"}], 
    "\[And]", 
    RowBox[{"x", 
     UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], "b"}]}], 
   TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" instead.\[LineSeparator]In this case, there is no need to \
distinguish between concrete integers and other terms.",
  FontColor->GrayLevel[0]]
}], "Item",
 CellChangeTimes->CompressedData["
1:eJwlz1sow2EABfC/kRAyci8sCkmhJimXDxlJw1ZjRGHkAbmNCJt7RPJAYTy4
NJFEbrV5kAjtwSVJLrns5Y9WKJTiO5+H0+/l1OmIiqtzVAKO4/xpYL1Tza/h
lCeT45OORuqCm6xz5Iwn1UL3PqiIdy0fo87/eDCvJQPdY+c8qRheZd7+ckbo
tDizA7vCf1INVN17oBS+2TUp4KHq37aUbjV0EXwypw3mZrhm8dJCoR8nsb2g
Pc1HOpzraBY7XvJEOloQAxMWTI/tVK3yhekTfHStpT4nnzBrMwvvYfTKxysU
3/cLOmBKhT2MdZUHwUa1LAJOlJQVwamQLmbDUKFNJ9VtUOQMJcp1Au82e5Kh
Q15UBjSXXuXDALk66+uGJ71Bidmw0tpWAW3k4lyo2ZhK+qbeuL8xt1uUy763
9Efr7DZc2c/fgVtHeyZ4bKU6hfFxXlewrl2hvHvkyYxQz9QvWWY3nuiO7lgP
Q5usvSPNPDnQmJiyZU9/mBUjYobteqRWWXhin/SQBv8AM7MO9w==
  "],
 CellID->132142128]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Tuples (7)", "Subsection",
 CellChangeTimes->{{3.598330486778203*^9, 3.598330487826708*^9}, {
   3.601872567977281*^9, 3.601872574050446*^9}, 3.601872605287873*^9, {
   3.601875629344592*^9, 3.601875630909377*^9}, 3.601875722613394*^9, 
   3.610853731108092*^9, 3.614314668846904*^9, {3.615190856835415*^9, 
   3.615190861519133*^9}, 3.615190937824673*^9},
 CellTags->"sect::ProverTuples",
 CellID->1679395489],

Cell[CellGroupData[{

Cell["Subscripts (2)", "Subsubsection",
 CellChangeTimes->{{3.601881593312263*^9, 3.601881595732731*^9}, {
   3.609572505133442*^9, 3.609572505759392*^9}, 3.614314646901135*^9},
 CellID->417228170],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]subscriptTupleOf\[CloseCurlyDoubleQuote]: Let ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " denote the ground abstraction tuple ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"t", "[", "i", "]"}], 
     UnderscriptBox["|", 
      RowBox[{"R", "[", "i", "]"}]], 
     RowBox[{"\[Phi]", "[", "i", "]"}]}], "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 " and let ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["j", "0"], TraditionalForm]]],
 " be an a.b.f. constant. If ",
 Cell[BoxData[
  FormBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
     RowBox[{"[", 
      SubscriptBox["j", "0"], "]"}]}],
    FontColor->GrayLevel[0]], TraditionalForm]],
  FontColor->RGBColor[1, 0, 0]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"1", 
    UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
    SubscriptBox["j", "0"]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["j", "0"], 
    UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], "a"}], 
   TraditionalForm]]],
 " are known (for some term ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 ") ",
 StyleBox["and apart from that ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   SubscriptBox["j", "0"], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" only appears as subscript of ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["T", TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" in the entire proof situation (and it really appears as such), \
then first show ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isTuple", "[", "T", "]"}], "\[And]", 
    RowBox[{"a", "\[Equal]", 
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]}], 
   TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" without ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
    RowBox[{"[", 
     SubscriptBox["j", "0"], "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"1", 
    UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
    SubscriptBox["j", "0"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["j", "0"], 
    UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], "a"}], 
   TraditionalForm]]],
 StyleBox[" among the assumptions, and then continue with the main branch, \
where some new a.b.f. constant ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   SubscriptBox["i", "0"], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" is introduced, all of ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
    RowBox[{"[", 
     SubscriptBox["j", "0"], "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"1", 
    UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
    SubscriptBox["j", "0"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["j", "0"], 
    UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], "a"}], 
   TraditionalForm]]],
 StyleBox[" got removed, ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"RC", "[", 
    SubscriptBox["i", "0"], "]"}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "[", 
    SubscriptBox["i", "0"], "]"}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" are in the knowledge base instead, and every occurrence of ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   SubscriptBox["T", 
    SubscriptBox["j", "0"]], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" got replaced by ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "[", 
    SubscriptBox["i", "0"], "]"}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[".",
  FontColor->GrayLevel[0]]
}], "Item",
 CellChangeTimes->{{3.601881750836237*^9, 3.601881940946668*^9}, {
   3.601885379938052*^9, 3.601885392872744*^9}, {3.601885458301719*^9, 
   3.601885469994751*^9}, {3.603176200314907*^9, 3.603176242533435*^9}, {
   3.603790048589066*^9, 3.603790154860987*^9}, {3.603790188958662*^9, 
   3.603790189520324*^9}, 3.603790221189398*^9, {3.60379041929382*^9, 
   3.603790426039811*^9}, {3.603790762719559*^9, 3.6037907761177*^9}, {
   3.603790948213904*^9, 3.603790948639166*^9}, {3.60379098684483*^9, 
   3.603791056613961*^9}, {3.603791099040071*^9, 3.60379110084946*^9}, {
   3.60543167365855*^9, 3.605431808433391*^9}, 3.605944771580036*^9, {
   3.606108422440668*^9, 3.606108513529595*^9}, {3.606108544024*^9, 
   3.606108549065972*^9}, {3.606108594209542*^9, 3.606108606504943*^9}, {
   3.606108652171467*^9, 3.606108684044693*^9}, {3.606109423458546*^9, 
   3.606109429160481*^9}, {3.606110241466822*^9, 3.606110250772665*^9}, {
   3.606111329209966*^9, 3.606111329954908*^9}, {3.606117729413963*^9, 
   3.606117762665714*^9}, {3.606117794835476*^9, 3.606117879710505*^9}, {
   3.606118088779096*^9, 3.606118089821166*^9}, {3.609131794329653*^9, 
   3.609131934102621*^9}, 3.609572500908309*^9, {3.615191645525678*^9, 
   3.61519164725307*^9}, {3.615191696473743*^9, 3.615191751750707*^9}},
 CellID->1268545251],

Cell[TextData[{
 "In the above rule it is crucial that nothing is known about ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["j", "0"], TraditionalForm]]],
 " apart from being an integer in some interval! In particular, this is the \
reason why we cannot \[OpenCurlyDoubleQuote]merge\[CloseCurlyDoubleQuote] the \
rule above and the one below. Note that the rule above can be considered the \
\[OpenCurlyDoubleQuote]dual\[CloseCurlyDoubleQuote] to rules \
\[OpenCurlyDoubleQuote]existentialGoalTupleOf\[CloseCurlyDoubleQuote] and \
\[OpenCurlyDoubleQuote]universalKBTupleOf\[CloseCurlyDoubleQuote]."
}], "Text",
 CellChangeTimes->{{3.606118002074067*^9, 3.606118074211876*^9}, {
  3.6064580333996058`*^9, 3.606458104785471*^9}, {3.609131956435614*^9, 
  3.609131960572621*^9}, {3.614509440500405*^9, 3.614509443811528*^9}},
 CellID->219033739],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]elementTupleOf\[CloseCurlyDoubleQuote]: Let ",
 Cell[BoxData[
  FormBox["S", TraditionalForm]]],
 " denote the ground abstraction tuple ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"t", "[", "i", "]"}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "u"}]], "True"}], 
     "\[RightAngleBracket]"}], " "}], TraditionalForm]]],
 " and let ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 " be a ground term (not necessarily an a.b.f. constant).\[LineSeparator]If \
somewhere in the proof situation ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "j"], TraditionalForm]]],
 " appears, first show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isTuple", "[", "S", "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", 
      SubscriptBox["\[DoubleStruckCapitalZ]", 
       RowBox[{"1", ",", "\[Ellipsis]", ",", "u"}]]], 
     RowBox[{"[", "j", "]"}]}]}], TraditionalForm]]],
 ", and then continue with the main branch where every occurrence of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "j"], TraditionalForm]]],
 " got replaced by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "[", "j", "]"}], TraditionalForm]]],
 StyleBox[" and both ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "S", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"1", ",", "\[Ellipsis]", ",", "u"}]]], 
    RowBox[{"[", "j", "]"}]}], TraditionalForm]]],
 " are among the assumptions."
}], "Item",
 CellChangeTimes->{{3.600580441270837*^9, 3.600580526367975*^9}, {
   3.600580560411833*^9, 3.600580643332298*^9}, {3.600582900863939*^9, 
   3.600582911057041*^9}, {3.600582973733313*^9, 3.600582998740082*^9}, {
   3.600753700333034*^9, 3.600753707315737*^9}, {3.601121017354355*^9, 
   3.601121020289213*^9}, {3.60112107139583*^9, 3.601121132722918*^9}, {
   3.601196208543596*^9, 3.601196230134478*^9}, 3.601725922548328*^9, {
   3.601725959572663*^9, 3.601725963505483*^9}, {3.601790615298827*^9, 
   3.60179070706807*^9}, {3.601878864951594*^9, 3.601878881872009*^9}, {
   3.602390780807338*^9, 3.602390979317202*^9}, {3.602392680064516*^9, 
   3.602392749785584*^9}, {3.602394195737057*^9, 3.602394258184583*^9}, {
   3.602394317201169*^9, 3.602394317385884*^9}, {3.603176356000268*^9, 
   3.603176396498599*^9}, {3.6061101035162*^9, 3.606110133622632*^9}, {
   3.606110190032601*^9, 3.606110232584209*^9}, {3.606110275952682*^9, 
   3.60611029984363*^9}, {3.606110345128669*^9, 3.606110432866633*^9}, {
   3.606110476808692*^9, 3.60611056756991*^9}, {3.606110680390012*^9, 
   3.606110681431857*^9}, {3.60611132621512*^9, 3.606111366564305*^9}, {
   3.606111489623015*^9, 3.606111629556641*^9}, {3.606116225909365*^9, 
   3.606116244396612*^9}, 3.609572503956924*^9, {3.615191774761361*^9, 
   3.615191823207888*^9}, 3.615191854538124*^9, 3.615192450057722*^9},
 CellID->1468431481],

Cell[TextData[StyleBox["Remark:",
 FontFamily->"Bitstream Charter",
 FontSize->14,
 FontWeight->"Bold"]], "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603522907354292*^9, 3.603522909426996*^9}, {
   3.603522955835897*^9, 3.603522963082172*^9}, {3.603793011527348*^9, 
   3.603793024231905*^9}, {3.604122440773405*^9, 3.604122447629318*^9}, 
   3.615191908916092*^9},
 CellID->1935127905],

Cell[TextData[{
 "Rules \[OpenCurlyDoubleQuote]subscriptTupleOf\[CloseCurlyDoubleQuote] and \
\[OpenCurlyDoubleQuote]elementTupleOf\[CloseCurlyDoubleQuote] are examples of \
inference rules that may only be applied if some additional condition is \
fulfilled (in this case: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "T", "]"}], TraditionalForm]]],
 " resp. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "S", "]"}], TraditionalForm]]],
 ", among others). The overall strategy to tackle such rules in the \
ComplexityProver is to create a new branch in the proof tree, ",
 StyleBox["first",
  FontSlant->"Italic"],
 " show the conditions in a subproof, and ",
 StyleBox["then",
  FontSlant->"Italic"],
 " continue with the \[OpenCurlyDoubleQuote]main branch\
\[CloseCurlyDoubleQuote] where the inference rule has actually been applied.\n\
Still, there is one more detail concerning the implementation of such rules: \
If, for instance, in \[OpenCurlyDoubleQuote]subscriptTupleOf\
\[CloseCurlyDoubleQuote] the term ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["T", 
    SubscriptBox["j", "0"]], TraditionalForm]]],
 " occurs in the knowledge base, then it still does so in the subproof where \
the condition of the inference rule is checked. Hence, the inference rule \
could be applied ",
 StyleBox["again",
  FontSlant->"Italic"],
 " on exactly the same term in exactly the same way, yielding another \
subproof, and so on, finally resulting in an infinite loop. Hence, it is of \
utmost importance to make sure that the rule is not applied on the same term \
in condition-checking subproofs; It may, however, still be applied on other \
terms."
}], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.601877703631141*^9, 3.601877715932562*^9}, {
  3.601888275960589*^9, 3.601888309333694*^9}, {3.605944166565875*^9, 
  3.605944172833125*^9}, {3.615191912308668*^9, 3.615192126228526*^9}, {
  3.615192156231898*^9, 3.615192437485356*^9}},
 CellID->1355311681]
}, Closed]],

Cell[CellGroupData[{

Cell["Abstraction Tuples (3)", "Subsubsection",
 CellChangeTimes->{{3.600580430160945*^9, 3.600580433027842*^9}, {
   3.600582894331026*^9, 3.60058289780495*^9}, 3.610779071677217*^9, 
   3.614314650938627*^9},
 CellID->1023287520],

Cell[CellGroupData[{

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]lengthTupleOf\[CloseCurlyDoubleQuote]: Let ",
 Cell[BoxData[
  FormBox["S", TraditionalForm]]],
 " denote the ground abstraction tuple ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"t", "[", "i", "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "u"}]], "True"}], 
    "\[RightAngleBracket]"}], TraditionalForm]]],
 ".\[LineSeparator]If somehwere in the proof situation ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}], 
   TraditionalForm]]],
 " appears, first show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "S", "]"}], TraditionalForm]]],
 " and then continue with the main branch where",
 StyleBox[" every occurrence of ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}], 
   TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" got replaced by ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["u", TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "S", "]"}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" is in the knowledge base.",
  FontColor->GrayLevel[0]]
}], "Item",
 CellChangeTimes->{{3.600580441270837*^9, 3.600580526367975*^9}, {
   3.600580560411833*^9, 3.600580643332298*^9}, {3.600582900863939*^9, 
   3.600582900877019*^9}, {3.600582954677519*^9, 3.600582961396924*^9}, {
   3.600669939876018*^9, 3.600669939887615*^9}, {3.600753681116602*^9, 
   3.600753688535045*^9}, {3.601121006708324*^9, 3.601121009518744*^9}, {
   3.601725904589971*^9, 3.601725948634447*^9}, {3.601790407456865*^9, 
   3.601790457440473*^9}, {3.601790601974711*^9, 3.601790604948616*^9}, {
   3.601878705657401*^9, 3.601878721345344*^9}, {3.601878853240664*^9, 
   3.60187888423975*^9}, {3.601879318844998*^9, 3.601879328819725*^9}, {
   3.602392581280226*^9, 3.602392664202556*^9}, {3.602394279193469*^9, 
   3.60239431082836*^9}, {3.606110610546812*^9, 3.60611064989092*^9}, {
   3.606110705247775*^9, 3.606110706363266*^9}, {3.606111168251109*^9, 
   3.606111183387753*^9}, {3.60611133294217*^9, 3.606111333761005*^9}, {
   3.606111381640804*^9, 3.606111462885424*^9}, {3.60611164446376*^9, 
   3.606111671878468*^9}, 3.609572515715035*^9, {3.615192465283612*^9, 
   3.615192488701584*^9}},
 CellID->567474730],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]nonemptyTupleOf\[CloseCurlyDoubleQuote]: Let ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " denote the abstraction tuple ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"t", "[", "i", "]"}], 
     UnderscriptBox["|", 
      RowBox[{"R", "[", "i", "]"}]], 
     RowBox[{"\[Phi]", "[", "i", "]"}]}], "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 ". If",
 StyleBox[" ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"1", 
    UnderscriptBox["\[LessEqual]", "\[DoubleStruckCapitalZ]"], 
    RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 " has to be shown, show instead ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isTuple", "[", "T", "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{"R", "[", "i", "]"}]], 
     RowBox[{"\[Phi]", "[", "i", "]"}]}]}], TraditionalForm]]],
 ".\[LineSeparator]This rule is a combination of rule \
\[OpenCurlyDoubleQuote]existsTupleOfGoal\[CloseCurlyDoubleQuote] and formula ",
 ButtonBox["(1\[LessEqual]length).",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1546357463"]
}], "Item",
 CellChangeTimes->{{3.600580441270837*^9, 3.600580526367975*^9}, {
   3.600580560411833*^9, 3.600580643332298*^9}, {3.600582900863939*^9, 
   3.600582900877019*^9}, {3.600582954677519*^9, 3.600582961396924*^9}, {
   3.600669939876018*^9, 3.600670015639011*^9}, {3.601121140669054*^9, 
   3.601121155864541*^9}, {3.601121229952318*^9, 3.601121229959841*^9}, {
   3.601194831388522*^9, 3.601194831396609*^9}, {3.601725969656473*^9, 
   3.601725974672322*^9}, {3.601790589147228*^9, 3.601790596588914*^9}, {
   3.601790723795847*^9, 3.6017907288209*^9}, {3.601875141624138*^9, 
   3.601875183102734*^9}, {3.603176427343548*^9, 3.603176435336974*^9}, {
   3.60369334923838*^9, 3.603693351335943*^9}, {3.606110962497376*^9, 
   3.606110972532651*^9}, {3.606111031359927*^9, 3.606111065957898*^9}, {
   3.60611113310671*^9, 3.60611115390122*^9}, {3.606111335715463*^9, 
   3.606111336467607*^9}, 3.610779076028482*^9, {3.615192492311817*^9, 
   3.615192494008318*^9}, {3.615265618975472*^9, 3.615265694910359*^9}},
 CellID->2055491952],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]isTupleTupleOf\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"t", "[", "i", "]"}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "a"}], ",", "\[Ellipsis]", ",", "b"}]], 
      RowBox[{"\[Phi]", "[", "i", "]"}]}], "\[RightAngleBracket]"}], "]"}], 
   TraditionalForm]]],
 " has to be shown, show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
     RowBox[{"[", "a", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
     RowBox[{"[", "b", "]"}]}]}], TraditionalForm]]],
 " instead.\[LineSeparator]This rule is certainly ",
 StyleBox["correct",
  FontSlant->"Italic"],
 " but not ",
 StyleBox["complete",
  FontSlant->"Italic"],
 ": For instance ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " could also be \[Infinity] if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "[", "i", "]"}], TraditionalForm]]],
 " fails for large enough ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 ". Also, ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and/or ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " could be real numbers. However, in our setting ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " will always be integers."
}], "Item",
 CellChangeTimes->{{3.600580441270837*^9, 3.600580526367975*^9}, {
   3.600580560411833*^9, 3.600580643332298*^9}, {3.600582900863939*^9, 
   3.600582900877019*^9}, {3.600582954677519*^9, 3.600582961396924*^9}, {
   3.600669939876018*^9, 3.600670015639011*^9}, {3.601121140669054*^9, 
   3.601121155864541*^9}, {3.601121229952318*^9, 3.601121229959841*^9}, {
   3.601194831388522*^9, 3.60119486333804*^9}, {3.601194904477428*^9, 
   3.601194940822009*^9}, {3.601195135628024*^9, 3.601195261122734*^9}, {
   3.601195309486309*^9, 3.601195377316245*^9}, {3.601195478290921*^9, 
   3.601195487858064*^9}, {3.601196360281655*^9, 3.601196361823235*^9}, {
   3.601725989370004*^9, 3.60172599475552*^9}, {3.601871640427273*^9, 
   3.601871641095283*^9}, {3.601871746186025*^9, 3.601871770930354*^9}, {
   3.602391012047455*^9, 3.602391012232325*^9}, {3.603176445169021*^9, 
   3.603176461407112*^9}, {3.606110977580585*^9, 3.606110998522778*^9}, 
   3.610779078646118*^9, {3.615192499796426*^9, 3.615192592571846*^9}},
 CellID->1853217663]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["\:293a (2)", "Subsubsection",
 CellChangeTimes->{{3.598331094541346*^9, 3.598331096501145*^9}, {
  3.601122115216563*^9, 3.601122120660624*^9}, {3.609572531476811*^9, 
  3.609572532003016*^9}, {3.613966975585701*^9, 3.61396697580834*^9}, {
  3.614314655632965*^9, 3.614314665530623*^9}},
 CellID->542161598],

Cell[CellGroupData[{

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]appendElement\[CloseCurlyDoubleQuote]: Let ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " be an a.b.f. constant or an abstraction tuple (in either case without free \
variables), and let ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " be an arbitrary ground object.",
 StyleBox[" If somewhere in the proof situation the term ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   SubscriptBox[
    RowBox[{"(", 
     RowBox[{"A", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "x"}], ")"}], "t"], TraditionalForm]]],
 StyleBox[" appears, where ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " is some some ground term, then ",
 StyleBox["make three branches, where in the first branch ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isTuple", "[", "A", "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", 
      SubscriptBox["\[DoubleStruckCapitalZ]", 
       RowBox[{"1", ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", 
         RowBox[{"A", 
          TagBox["\:293a",
           Identity,
           SyntaxForm->"a*b"], "x"}], "\[RightBracketingBar]"}]}]]], 
     RowBox[{"[", "t", "]"}]}]}], TraditionalForm]]],
 " is checked, in the second branch the original goal is shown with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "A", "]"}], TraditionalForm]]],
 " and",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"1", ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 " in the knowledge base and every occurrence of",
 StyleBox[" ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   SubscriptBox[
    RowBox[{"(", 
     RowBox[{"A", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "x"}], ")"}], "t"], TraditionalForm]]],
 " replaced by ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["A", "t"], TraditionalForm]]],
 ", and in the third branch again the original goal is shown with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "A", "]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[DoubleStruckCapitalZ]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Equal]", 
    RowBox[{
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}], 
     UnderscriptBox["+", "\[DoubleStruckCapitalZ]"], "1"}]}], 
   TraditionalForm]]],
 " in the knowledge base and every occurrence of",
 StyleBox[" ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   SubscriptBox[
    RowBox[{"(", 
     RowBox[{"A", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "x"}], ")"}], "t"], TraditionalForm]]],
 " replaced by ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ".\[LineSeparator]This inference rule is derived from and justified by \
formula ",
 ButtonBox["(element\:293a).",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:585712945"]
}], "Item",
 CellChangeTimes->{{3.601122128894205*^9, 3.60112220452954*^9}, {
   3.601122414002012*^9, 3.601122462141045*^9}, {3.601196407986636*^9, 
   3.601196422846312*^9}, {3.601726345437134*^9, 3.601726358968725*^9}, {
   3.601726467461077*^9, 3.601726477073788*^9}, {3.601789050002401*^9, 
   3.60178920104329*^9}, {3.60178930714549*^9, 3.60178946054288*^9}, {
   3.601790827274899*^9, 3.601790837824048*^9}, {3.601790916690058*^9, 
   3.601791166907928*^9}, {3.601871500540934*^9, 3.601871500583633*^9}, {
   3.601886762980718*^9, 3.601886787108744*^9}, {3.601886992423673*^9, 
   3.601886998188101*^9}, {3.601887389707345*^9, 3.601887398088709*^9}, {
   3.602392952640022*^9, 3.602392961065172*^9}, {3.602393391794092*^9, 
   3.602393485324304*^9}, {3.603176579507345*^9, 3.603176598903847*^9}, {
   3.604925098837256*^9, 3.604925101405356*^9}, {3.605945143221865*^9, 
   3.605945158196799*^9}, {3.605945363540733*^9, 3.605945365603698*^9}, {
   3.606111829531788*^9, 3.606111869745199*^9}, {3.606119018162998*^9, 
   3.606119328828779*^9}, {3.606119421567849*^9, 3.60611942916483*^9}, {
   3.606458396133337*^9, 3.606458396366856*^9}, {3.606557350608612*^9, 
   3.606557372427245*^9}, {3.609572536058998*^9, 3.609572542429635*^9}, {
   3.613966373279728*^9, 3.613966373301181*^9}, {3.613966671571418*^9, 
   3.613966671833861*^9}, {3.613966973156179*^9, 3.61396697433398*^9}, 
   3.615192603789713*^9, {3.615192662722921*^9, 3.615192717624879*^9}, {
   3.615265719439065*^9, 3.615265736797232*^9}, {3.615265778928988*^9, 
   3.61526579498551*^9}},
 CellID->1872632954],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]appendTupleOf\[CloseCurlyDoubleQuote]: If somewhere \
in the proof situation the ground abstraction tuple ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"t", "[", 
      SubscriptBox[
       RowBox[{"(", 
        RowBox[{"A", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "x"}], ")"}], "i"], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", 
        RowBox[{"(", 
         RowBox[{"A", 
          TagBox["\:293a",
           Identity,
           SyntaxForm->"a*b"], "x"}], ")"}], "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Phi]", "[", 
      SubscriptBox[
       RowBox[{"(", 
        RowBox[{"A", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "x"}], ")"}], "i"], "]"}]}], 
    "\[RightAngleBracket]"}], TraditionalForm]]],
 " appears, where ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 " ranges over the whole tuple (with ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " appended) and only appears as subscript of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"A", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "x"}], TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " and \[Phi], then first check ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "A", "]"}], TraditionalForm]]],
 " and then make a case distinction, where in the first branch every \
occurrence of the abstraction tuple got replaced by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"t", "[", 
       SubscriptBox["A", "i"], "]"}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
      RowBox[{"\[Phi]", "[", 
       SubscriptBox["A", "i"], "]"}]}], "\[RightAngleBracket]"}], "\:293a", 
    " ", 
    RowBox[{"t", "[", "x", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "[", "x", "]"}], TraditionalForm]]],
 " is in the knowledge base, and in the second one every occurrence got \
replaced by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"t", "[", 
      SubscriptBox["A", "i"], "]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Phi]", "[", 
      SubscriptBox["A", "i"], "]"}]}], "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"\[Phi]", "[", "x", "]"}]}], TraditionalForm]]],
 " is in the knowledge base."
}], "Item",
 CellChangeTimes->CompressedData["
1:eJwdz2kow3EAxvFhvNFQJkfu2itXSnlD+cmdayKF0ByNFxjejHJPojFTSCmt
RMwtytEQS46UW2OKlfzcZ7xw/J7/i6dPPe++XpLSlAJzHo/nzgZT+12Lew8p
+X1RVcBquxUllDWud8Fm9d/4EvNHaJiEvB8/me6Ikp0/YzkUrZbUQXmWph6O
2ofZGpkaiZc3NAzlB8ConM5w+L0tfYB92QuP8EPY/QV1ggvHS2ZaUqfY6pgS
jzE1p7UkZ0PAPKrK24SL7YIreM/Xc6o1c/Y2zNnjNSG0rRY+8U8oMTl4P8Og
om+RJXMmdyQQ+oaUqRpPKRHfn6vh7lZG2pmBkrEvLafC+aZw+4ISv3i9FE4k
yBpgjXiGU8qf2nI1UqI9nd+DMcGRCZ7MLIVLJdz3ma6FcVGDdTA+oslCz4xd
buNckR+kfjJ/R5XpMPnarGfOxDp0u5wtb8N8pwdKum+TneH7dKIbHGj1F8Ge
DW0o9z+SJHhjGsyEdy6vnHFy1X7JEyUd0Zbn8B+U8xId
  "],
 CellID->914117023]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Addition in TOMs (2)", "Subsection",
 CellChangeTimes->{{3.605943954791955*^9, 3.605943956423859*^9}, {
   3.605944230513113*^9, 3.605944231546885*^9}, 3.605944482643708*^9, {
   3.606108008475721*^9, 3.606108008772338*^9}, 3.614314680001017*^9, {
   3.615190866274284*^9, 3.615190872955358*^9}, 3.615190939521522*^9},
 CellTags->"sect::ProverAddition",
 CellID->285731009],

Cell[TextData[{
 "Assume in this subsection that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTOM", "[", "D", "]"}], TraditionalForm]]],
 " is known."
}], "Text",
 CellChangeTimes->{{3.600403980755826*^9, 3.600404060905324*^9}, {
   3.600404326703542*^9, 3.600404338289435*^9}, {3.600404424635476*^9, 
   3.60040443258387*^9}, {3.600591020642235*^9, 3.60059103180283*^9}, {
   3.600591100316223*^9, 3.600591108718622*^9}, {3.60075310004409*^9, 
   3.600753109141402*^9}, {3.601120370512073*^9, 3.601120370697124*^9}, {
   3.601188169664918*^9, 3.601188183768708*^9}, {3.605341646737563*^9, 
   3.605341654158266*^9}, {3.605341771100681*^9, 3.605341781758591*^9}, {
   3.605943975743863*^9, 3.605943979487895*^9}, 3.605944517714909*^9, {
   3.615192819087663*^9, 3.615192850875904*^9}},
 CellID->2061042824],

Cell[CellGroupData[{

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]cancellationGoal\[CloseCurlyDoubleQuote]: If the \
current goal is of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         SubscriptBox["x", "1"], 
         UnderscriptBox["+", "D"], 
         SubscriptBox["x", "2"], 
         UnderscriptBox["+", "D"]}], "..."}], 
       UnderscriptBox["+", "D"], 
       SubscriptBox["x", "n"]}], "\[Tilde]", 
      RowBox[{
       SubscriptBox["y", "1"], 
       UnderscriptBox["+", "D"], 
       SubscriptBox["y", "2"], 
       UnderscriptBox["+", "D"]}]}], "..."}], 
    UnderscriptBox["+", "D"], 
    SubscriptBox["y", "m"]}], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox["\[Tilde]", TraditionalForm]]],
 " is one of ",
 Cell[BoxData[
  FormBox["\[Equal]", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "D"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["<", "D"], TraditionalForm]]],
 ", then remove all arguments ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["x", "i"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["y", "j"], TraditionalForm]]],
 " which appear on both sides, but check all of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     SubscriptBox["x", "i"], "]"}]}], TraditionalForm]]],
 " (for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}], 
   TraditionalForm]]],
 ") and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     SubscriptBox["y", "j"], "]"}]}], TraditionalForm]]],
 " (for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", "m"}], 
   TraditionalForm]]],
 ") before (in a subproof).\[LineSeparator]Of course, ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["+", "D"], TraditionalForm]]],
 " should be considered in \[OpenCurlyDoubleQuote]flattened\
\[CloseCurlyDoubleQuote] form in this case, but should be kept as a binary \
operation afterwards."
}], "Item",
 CellChangeTimes->{{3.601182411873045*^9, 3.601182545914791*^9}, {
   3.601182581622921*^9, 3.601182595337548*^9}, {3.601706871551639*^9, 
   3.601706871848742*^9}, {3.601706991517001*^9, 3.601707048115601*^9}, {
   3.601711398877698*^9, 3.601711441767816*^9}, {3.601711548361139*^9, 
   3.601711563168953*^9}, {3.601722998555549*^9, 3.601722999036331*^9}, {
   3.601791759216896*^9, 3.601791760170382*^9}, {3.601791846001401*^9, 
   3.601791846706061*^9}, {3.601872119418201*^9, 3.601872125027553*^9}, {
   3.601878604052081*^9, 3.601878651250653*^9}, {3.601878685637463*^9, 
   3.601878685862165*^9}, {3.602391376104166*^9, 3.602391377322075*^9}, {
   3.605414268423476*^9, 3.605414269052563*^9}, 3.605414300074664*^9, {
   3.60542884851008*^9, 3.605428849026328*^9}, {3.605847021758101*^9, 
   3.605847022386169*^9}, {3.605847079971088*^9, 3.605847093857389*^9}, {
   3.605847129689712*^9, 3.605847140276743*^9}, {3.60585607647526*^9, 
   3.605856099271857*^9}, {3.605861830691729*^9, 3.605861878195673*^9}, 
   3.606108004897205*^9, {3.615192855082964*^9, 3.615192948550631*^9}, {
   3.615192990470506*^9, 3.615192991030657*^9}},
 CellID->600703031],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]cancellationKB\[CloseCurlyDoubleQuote]: If a formula \
among the assumptions is of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         SubscriptBox["x", "1"], 
         UnderscriptBox["+", "D"], 
         SubscriptBox["x", "2"], 
         UnderscriptBox["+", "D"]}], "..."}], 
       UnderscriptBox["+", "D"], 
       SubscriptBox["x", "n"]}], "\[Tilde]", 
      RowBox[{
       SubscriptBox["y", "1"], 
       UnderscriptBox["+", "D"], 
       SubscriptBox["y", "2"], 
       UnderscriptBox["+", "D"]}]}], "..."}], 
    UnderscriptBox["+", "D"], 
    SubscriptBox["y", "m"]}], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox["\[Tilde]", TraditionalForm]]],
 " is one of ",
 Cell[BoxData[
  FormBox["\[Equal]", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "D"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["<", "D"], TraditionalForm]]],
 ", then proceed similarly as in \[OpenCurlyDoubleQuote]cancellationGoal\
\[CloseCurlyDoubleQuote]."
}], "Item",
 CellChangeTimes->{{3.601182411873045*^9, 3.601182545914791*^9}, {
   3.601182581622921*^9, 3.601182595337548*^9}, {3.601706871551639*^9, 
   3.601706871848742*^9}, {3.601706991517001*^9, 3.601707048115601*^9}, {
   3.601711398877698*^9, 3.601711441767816*^9}, {3.601711548361139*^9, 
   3.601711563168953*^9}, {3.601722998555549*^9, 3.601722999036331*^9}, {
   3.601791759216896*^9, 3.601791760170382*^9}, {3.601791846001401*^9, 
   3.601791846706061*^9}, {3.601872119418201*^9, 3.601872125027553*^9}, {
   3.601878604052081*^9, 3.601878651250653*^9}, {3.601878685637463*^9, 
   3.601878685862165*^9}, {3.602391376104166*^9, 3.602391377322075*^9}, {
   3.605414268423476*^9, 3.605414269052563*^9}, 3.605414300074664*^9, {
   3.60542884851008*^9, 3.605428849026328*^9}, {3.605847021758101*^9, 
   3.605847022386169*^9}, {3.605847079971088*^9, 3.605847093857389*^9}, {
   3.605847129689712*^9, 3.605847140276743*^9}, {3.60585607647526*^9, 
   3.605856099271857*^9}, {3.605861830691729*^9, 3.605861878195673*^9}, {
   3.605861949386553*^9, 3.605861963088341*^9}, 3.606108006668779*^9, {
   3.615192896859924*^9, 3.615192899017621*^9}, {3.615192958486667*^9, 
   3.615193023680778*^9}},
 CellID->1067746268]
}, Open  ]],

Cell[TextData[{
 "The two formulas in this subsection are derived from and justified by ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " being a totally-ordered Abelian monoid (and in particular having the \
cancellation property for all of ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Equal]", "D"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "D"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["<", "D"], TraditionalForm]]],
 ")."
}], "Text",
 CellChangeTimes->{{3.615195481340058*^9, 3.615195564672107*^9}},
 CellID->1062121198]
}, Closed]],

Cell[CellGroupData[{

Cell["Order Relation in TOMs (7)", "Subsection",
 CellChangeTimes->{{3.600403971947538*^9, 3.600403973659915*^9}, {
   3.601872581327787*^9, 3.601872606688297*^9}, {3.601875632843384*^9, 
   3.601875634328862*^9}, 3.60187571852201*^9, 3.610872355209138*^9, 
   3.614314688689277*^9, {3.615190889169784*^9, 3.615190904599385*^9}, 
   3.615190941992267*^9, 3.615191060834458*^9},
 CellTags->"sect::ProverOrdering",
 CellID->915360441],

Cell[TextData[{
 "Assume that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTOM", "[", "D", "]"}], TraditionalForm]]],
 " is known, and let ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["z", TraditionalForm]]],
 " be ground terms."
}], "Text",
 CellChangeTimes->{{3.600403980755826*^9, 3.600404060905324*^9}, {
  3.600404326703542*^9, 3.600404338289435*^9}, {3.600404424635476*^9, 
  3.60040443258387*^9}, {3.600591020642235*^9, 3.60059103180283*^9}, {
  3.600591100316223*^9, 3.600591108718622*^9}, {3.60075310004409*^9, 
  3.600753109141402*^9}, {3.601120370512073*^9, 3.601120370697124*^9}, {
  3.601188169664918*^9, 3.601188183768708*^9}, {3.605341646737563*^9, 
  3.605341654158266*^9}, {3.605341771100681*^9, 3.605341781758591*^9}, {
  3.615193039828716*^9, 3.615193047723334*^9}, {3.615193346043711*^9, 
  3.615193358204846*^9}, {3.615193422234392*^9, 3.615193432231201*^9}},
 CellID->1330669066],

Cell[CellGroupData[{

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]contradictionOrdering\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["<", "D"], "x"}], TraditionalForm]]],
 " is known, try to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 ", because we would have a contradiction in the knowledge base then.\
\[LineSeparator]This inference rule is derived from and justified by formula ",
 ButtonBox["(irreflexivity of <)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1308600401"],
 "."
}], "Item",
 CellChangeTimes->{{3.601032750923368*^9, 3.601032797498435*^9}, {
  3.601707298140935*^9, 3.601707300804262*^9}, {3.601791234386223*^9, 
  3.601791264220375*^9}, {3.605341659887742*^9, 3.605341712794471*^9}, {
  3.605413674978262*^9, 3.605413683962958*^9}, {3.615193063030247*^9, 
  3.615193092524228*^9}, {3.615193361413111*^9, 3.615193362485049*^9}, {
  3.615194832914441*^9, 3.615194890392152*^9}},
 CellID->2002455362],

Cell[TextData[{
 StyleBox["\[OpenCurlyDoubleQuote]notLessEqual\[CloseCurlyDoubleQuote]: If ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"(", 
     RowBox[{"x", 
      UnderscriptBox["\[LessEqual]", "D"], "y"}], ")"}]}], 
   TraditionalForm]]],
 " appears somewhere in the proof situation: First check ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", "x", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", "y", "]"}]}]}], TraditionalForm]]],
 " without ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"(", 
     RowBox[{"x", 
      UnderscriptBox["\[LessEqual]", "D"], "y"}], ")"}]}], 
   TraditionalForm]]],
 " in the proof situation and then continue with the main branch, where",
 StyleBox[" every occurrence of ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"(", 
     RowBox[{"x", 
      UnderscriptBox["\[LessEqual]", "D"], "y"}], ")"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" got replaced by ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"y", 
    UnderscriptBox["<", "D"], "x"}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 " and both ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "y", "]"}]}], TraditionalForm]]],
 " are in the knowledge base. Analoguous for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"(", 
     RowBox[{"x", 
      UnderscriptBox["<", "D"], "y"}], ")"}]}], TraditionalForm]]],
 ".\[LineSeparator]This inference rule is derived from and justified by \
formulas ",
 ButtonBox["(\[Not]\[LessEqual])",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1288924674"],
 " and ",
 ButtonBox["(\[Not]<)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1962569111"],
 "."
}], "Item",
 CellChangeTimes->{{3.600404064292387*^9, 3.600404133244512*^9}, {
   3.600404229356722*^9, 3.600404252609709*^9}, {3.600406896040203*^9, 
   3.600406906619077*^9}, {3.600591126916682*^9, 3.60059113301456*^9}, {
   3.601707329928132*^9, 3.601707333829496*^9}, {3.601707890816441*^9, 
   3.601707895541413*^9}, 3.602330697621379*^9, {3.602330738140521*^9, 
   3.602330780067987*^9}, {3.602330998507676*^9, 3.602330999730026*^9}, {
   3.605341370717936*^9, 3.6053413718633*^9}, 3.605342179563943*^9, {
   3.606454988798038*^9, 3.60645507236316*^9}, {3.60645532481022*^9, 
   3.606455462089822*^9}, {3.6064555328376827`*^9, 3.606455561680675*^9}, {
   3.6064555921800823`*^9, 3.606455593146291*^9}, 3.61085620420214*^9, {
   3.61519309425847*^9, 3.615193319967909*^9}, 3.615193371627729*^9, {
   3.615193937931285*^9, 3.615193984398123*^9}, {3.615194900485313*^9, 
   3.615194969035036*^9}},
 CellID->1349325019],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]reflexiveClosure\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["<", "D"], "y"}], TraditionalForm]]],
 " is known but ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["\[LessEqual]", "D"], "y"}], TraditionalForm]]],
 " is not, first check ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", "x", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", "y", "]"}]}]}], TraditionalForm]]],
 " without ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["<", "D"], "y"}], TraditionalForm]]],
 " in the knowledge base, and then continue with the main branch, where all \
of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "y", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["\[LessEqual]", "D"], "y"}], TraditionalForm]]],
 " are added to the asumptions.\[LineSeparator]This inference rule is derived \
from and justified by formula ",
 ButtonBox["(<\[Or]=)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:966362506"],
 "."
}], "Item",
 CellChangeTimes->{{3.600404064292387*^9, 3.600404133244512*^9}, {
   3.600404229356722*^9, 3.600404252606188*^9}, {3.600404291584313*^9, 
   3.600404306851026*^9}, {3.600591135635194*^9, 3.600591142266255*^9}, {
   3.600591415024049*^9, 3.600591530937902*^9}, {3.600591573848447*^9, 
   3.600591573857703*^9}, {3.600591656283947*^9, 3.600591656288824*^9}, {
   3.600591722920783*^9, 3.600591724995784*^9}, {3.601120322696248*^9, 
   3.601120323546765*^9}, {3.601120637969905*^9, 3.601120639147845*^9}, {
   3.601120768713363*^9, 3.601120769760843*^9}, {3.60172656007751*^9, 
   3.601726570260891*^9}, {3.606455917365295*^9, 3.606456038799696*^9}, {
   3.606456120265275*^9, 3.606456202929986*^9}, {3.606456299333848*^9, 
   3.6064563839112263`*^9}, {3.606456454244034*^9, 3.606456491013652*^9}, {
   3.60645669097643*^9, 3.606456717325057*^9}, {3.6064567676041613`*^9, 
   3.606456778883088*^9}, {3.606625213318291*^9, 3.606625215638453*^9}, 
   3.610856206842857*^9, {3.615193332330954*^9, 3.615193343665095*^9}, {
   3.61519337737866*^9, 3.615193402201606*^9}, {3.615195000499296*^9, 
   3.615195035213977*^9}},
 CellID->234874177],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]transitivityKB\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["\[LessEqual]", "D"], "y"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", 
    UnderscriptBox["\[LessEqual]", "D"], "z"}], TraditionalForm]]],
 " are known, ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " is syntactically different from ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " is syntactically different from ",
 Cell[BoxData[
  FormBox["z", TraditionalForm]]],
 ",",
 StyleBox[" first check ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "y", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[", and then continue with the main branch with ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["\[LessEqual]", "D"], "z"}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" (or, if ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["z", TraditionalForm]]],
 " are identical, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Equal]", "y"}], TraditionalForm]]],
 ") ",
 StyleBox["added to the knowledge base. Similar for all other combinations of \
",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   UnderscriptBox["<", "D"], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "D"], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[".\[LineSeparator]However, do not apply this rule if the new \
formula is already known!\[LineSeparator]",
  FontColor->GrayLevel[0]],
 "The reason why this rule cannot simply be dealt with by rewriting, but \
really has to be implemented as an inference rule, is because first of all ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["<", "D"], "z"}], TraditionalForm]]],
 " can be inferred if ",
 StyleBox["some",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " satisfies the conditions, and also we do not require type information \
about any of the occuring objects."
}], "Item",
 CellChangeTimes->{{3.600404064292387*^9, 3.600404133244512*^9}, {
   3.600404229356722*^9, 3.600404252606188*^9}, {3.600404291584313*^9, 
   3.600404315492769*^9}, {3.600404352087138*^9, 3.600404401191081*^9}, {
   3.600591171100865*^9, 3.600591179331023*^9}, {3.601120682174582*^9, 
   3.601120682953483*^9}, {3.601120775614251*^9, 3.601120776612519*^9}, {
   3.601726575748499*^9, 3.601726591626074*^9}, 3.601726676312003*^9, {
   3.605341812460337*^9, 3.605341906250266*^9}, {3.605342149080443*^9, 
   3.605342172044697*^9}, {3.605412931714202*^9, 3.605413092249145*^9}, {
   3.605413222185462*^9, 3.605413224083292*^9}, {3.605414416406889*^9, 
   3.605414443037129*^9}, {3.605424762525804*^9, 3.605424848067059*^9}, {
   3.605428795779314*^9, 3.605428836098505*^9}, {3.606456293112068*^9, 
   3.60645631148739*^9}, {3.610188209339673*^9, 3.610188232522932*^9}, {
   3.610188283686701*^9, 3.610188284549318*^9}, {3.615193406482133*^9, 
   3.615193409026381*^9}, {3.615193453256687*^9, 3.615193453842141*^9}, {
   3.615193487125506*^9, 3.615193506468586*^9}, {3.615195058041781*^9, 
   3.615195127776154*^9}, {3.615195161159616*^9, 3.615195166771424*^9}},
 CellID->1971664220],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]transitivityGoal\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["\[LessEqual]", "D"], "y"}], TraditionalForm]]],
 " (",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " syntactically different from ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ") has to be shown and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["\[LessEqual]", "D"], "z"}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["<", "D"], "z"}], TraditionalForm]]],
 " is known (",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " syntactically different from ",
 Cell[BoxData[
  FormBox["z", TraditionalForm]]],
 "), try to prove",
 StyleBox[" ",
  FontColor->RGBColor[1, 0, 0]],
 Cell[BoxData[
  FormBox[
   StyleBox[
    RowBox[{
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", "x", "]"}]}], "\[And]", 
     RowBox[{"z", 
      UnderscriptBox["\[LessEqual]", "D"], "y"}]}],
    FontColor->GrayLevel[0]], TraditionalForm]],
  FontColor->RGBColor[1, 0, 0]],
 " instead.\[LineSeparator]Similar for all other combinations of ",
 Cell[BoxData[
  FormBox["\[LessEqual]", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["<", TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.600752840603072*^9, 3.600752905968175*^9}, {
   3.601120792421282*^9, 3.601120818860075*^9}, {3.601707542983797*^9, 
   3.601707543369402*^9}, {3.601707586793931*^9, 3.60170758701601*^9}, {
   3.601791289412301*^9, 3.601791302853948*^9}, {3.605342096052163*^9, 
   3.605342103105723*^9}, {3.605413947640805*^9, 3.605413972320764*^9}, 
   3.605414051655663*^9, {3.605427250389996*^9, 3.605427268854468*^9}, {
   3.606108149309412*^9, 3.606108164385231*^9}, {3.610872361768136*^9, 
   3.610872420629075*^9}, {3.615193522139155*^9, 3.615193523947047*^9}},
 CellID->1238552456]
}, Open  ]],

Cell[TextData[{
 "The last two inference rules above are both derived from and justified by \
",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "D"], TraditionalForm]]],
 " being a total order relation (and in particular transitive), as well as \
formulas ",
 ButtonBox["(transitivity \[LessEqual]<)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1623404218"],
 ", ",
 ButtonBox["(transitivity <\[LessEqual])",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:243913977"],
 " and ",
 ButtonBox["(transitivity <<)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:1509384446"],
 "."
}], "Text",
 CellChangeTimes->{{3.615195138617751*^9, 3.615195271292103*^9}, {
  3.615195460500959*^9, 3.615195460723046*^9}},
 CellID->547361213],

Cell[CellGroupData[{

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]transitivityGoalAddition\[CloseCurlyDoubleQuote]: If \
",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["\[LessEqual]", "D"], "a"}], TraditionalForm]]],
 " is known (for some ground term ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 ") and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["+", "D"], "y"}], 
    UnderscriptBox["\[LessEqual]", "D"], "z"}], TraditionalForm]]],
 " has to be shown, try to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["+", "D"], "y"}], 
    UnderscriptBox["\[LessEqual]", "D"], "z"}], TraditionalForm]]],
 " instead. Same for all other combinations of ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["+", "D"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "D"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["<", "D"], TraditionalForm]]],
 ".\[LineSeparator]Also, if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    UnderscriptBox["\[LessEqual]", "D"], "x"}], TraditionalForm]]],
 " is known and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"z", 
    UnderscriptBox["\[LessEqual]", "D"], 
    RowBox[{"x", 
     UnderscriptBox["+", "D"], "y"}]}], TraditionalForm]]],
 " has to be shown, try to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"z", 
    UnderscriptBox["\[LessEqual]", "D"], 
    RowBox[{"a", 
     UnderscriptBox["+", "D"], "y"}]}], TraditionalForm]]],
 " instead. Again, same for all other combinations of ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["+", "D"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "D"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["<", "D"], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.601107990994296*^9, 3.601108005225212*^9}, {
   3.601108322958489*^9, 3.601108399297316*^9}, {3.601110134900465*^9, 
   3.601110142141281*^9}, {3.601120958251007*^9, 3.601120961041627*^9}, {
   3.601706697592047*^9, 3.601706717433937*^9}, {3.60170675108374*^9, 
   3.601706776307349*^9}, 3.601706859123699*^9, {3.601707059302217*^9, 
   3.601707060440521*^9}, {3.601711466012571*^9, 3.601711566866832*^9}, 
   3.601711671662247*^9, {3.601722995569925*^9, 3.601722996050357*^9}, {
   3.601791762418184*^9, 3.601791763413546*^9}, {3.601791844030278*^9, 
   3.601791844876304*^9}, {3.60187219822986*^9, 3.601872238706825*^9}, {
   3.605428851344699*^9, 3.605428851828445*^9}, {3.605606137652218*^9, 
   3.605606167186234*^9}, {3.605944290661063*^9, 3.605944335282814*^9}, {
   3.615193555988692*^9, 3.615193583184845*^9}, {3.615193690085728*^9, 
   3.61519369176654*^9}, {3.615193728891647*^9, 3.615193850598498*^9}},
 CellID->567605550],

Cell["\<\
\[OpenCurlyDoubleQuote]transitivityGoalAdditionLeft\[CloseCurlyDoubleQuote]: \
Similar to \[OpenCurlyDoubleQuote]transitivityGoalAddition\
\[CloseCurlyDoubleQuote], but perform replacements only on the left-hand-side \
of the current proof goal.\
\>", "Item",
 CellChangeTimes->{{3.601107990994296*^9, 3.601108005225212*^9}, {
   3.601108322958489*^9, 3.601108399297316*^9}, {3.601110134900465*^9, 
   3.601110142141281*^9}, {3.601120958251007*^9, 3.601120961041627*^9}, {
   3.601706697592047*^9, 3.601706717433937*^9}, {3.60170675108374*^9, 
   3.601706776307349*^9}, 3.601706859123699*^9, {3.601707059302217*^9, 
   3.601707060440521*^9}, {3.601711466012571*^9, 3.601711566866832*^9}, 
   3.601711671662247*^9, {3.601722995569925*^9, 3.601722996050357*^9}, {
   3.601791762418184*^9, 3.601791763413546*^9}, {3.601791844030278*^9, 
   3.601791844876304*^9}, {3.60187219822986*^9, 3.601872238706825*^9}, {
   3.605428851344699*^9, 3.605428851828445*^9}, {3.605606137652218*^9, 
   3.605606167186234*^9}, {3.605944290661063*^9, 3.605944335282814*^9}, {
   3.615191070124301*^9, 3.615191070758148*^9}, {3.615193586764778*^9, 
   3.615193588498395*^9}, {3.61519371772096*^9, 3.615193718016563*^9}, {
   3.615193873308051*^9, 3.615193912866752*^9}},
 CellID->906766808]
}, Open  ]],

Cell[TextData[{
 "The last two inference rules above are both derived from and justified by \
lemma \[OpenCurlyDoubleQuote]Properties of Totally-Ordered Abelian Monoids\
\[CloseCurlyDoubleQuote] in section ",
 ButtonBox["Lemmata on Basic Notions / Orderings",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::LemmataOrderings"],
 "."
}], "Text",
 CellChangeTimes->{{3.615195344172114*^9, 3.61519544380524*^9}},
 CellID->1487453809],

Cell[TextData[StyleBox["Remark:",
 FontFamily->"Bitstream Charter",
 FontSize->14,
 FontWeight->"Bold"]], "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellChangeTimes->{{3.603522907354292*^9, 3.603522909426996*^9}, {
   3.603522955835897*^9, 3.603522963082172*^9}, {3.603793011527348*^9, 
   3.603793024231905*^9}, {3.605601401187481*^9, 3.60560140544953*^9}, 
   3.615269090046624*^9},
 CellID->907544784],

Cell[TextData[{
 "Most of the inference rules in this subsection are almost identical to the \
formulas they are derived from, meaning that the formulas could, in \
principle, be used directly instead (by rewriting). The reason why we decided \
to add the rules nevertheless is the following: If we want to replace, for \
instance, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"(", 
     RowBox[{"x", 
      UnderscriptBox["\[LessEqual]", "D"], "y"}], ")"}]}], 
   TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", "<", "x"}], TraditionalForm]]],
 ", then formula (\[Not]\[LessEqual]) can only be used for rewriting if both ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "y", "]"}]}], TraditionalForm]]],
 " follow ",
 StyleBox["immediately",
  FontSlant->"Italic"],
 " from the current knowledge (see subsection on rewriting). This, however, \
is not the case most of the times: ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " are arbitrary terms which, of course, ",
 StyleBox["are",
  FontSlant->"Italic"],
 " elements of the domain ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " (otherwise the expression ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["\[LessEqual]", "D"], "y"}], TraditionalForm]]],
 " would not be defined), but this fact is not ",
 StyleBox["known",
  FontSlant->"Italic"],
 " in the proof yet. Hence, rewriting would leave the expression unchanged.\n\
Therefore, if we really want to carry out the replacement ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     RowBox[{"(", 
      RowBox[{"x", 
       UnderscriptBox["\[LessEqual]", "D"], "y"}], ")"}]}], "\[Rule]", 
    RowBox[{"y", 
     UnderscriptBox["<", "D"], "x"}]}], TraditionalForm]]],
 " as soon as ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"(", 
     RowBox[{"x", 
      UnderscriptBox["\[LessEqual]", "D"], "y"}], ")"}]}], 
   TraditionalForm]]],
 " appears in the proof situation, then we have to employ a different \
startegy, by means of an inference rule. The rule, rather than ",
 StyleBox["requiring",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " to be domain elements of ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ", simply ",
 StyleBox["proves",
  FontSlant->"Italic"],
 " this in a subproof."
}], "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellChangeTimes->{{3.603522955835983*^9, 3.603523016264877*^9}, {
   3.603523067993002*^9, 3.603523068774317*^9}, {3.603526511396674*^9, 
   3.603526537459005*^9}, {3.603526632531704*^9, 3.603526636839904*^9}, {
   3.603526720792316*^9, 3.60352672617742*^9}, {3.60352675821136*^9, 
   3.603526794949112*^9}, {3.603526838114751*^9, 3.60352684115434*^9}, {
   3.603526877383081*^9, 3.603527143643606*^9}, {3.603527173960989*^9, 
   3.603527187399389*^9}, {3.603793026985894*^9, 3.603793319090968*^9}, {
   3.603793361177027*^9, 3.60379351484412*^9}, 3.604035104537365*^9, {
   3.604035136662751*^9, 3.604035137669769*^9}, {3.604834858331772*^9, 
   3.604834958512149*^9}, {3.605601401187581*^9, 3.605601405449621*^9}, {
   3.615087678930005*^9, 3.615087689896127*^9}, {3.61526909277344*^9, 
   3.615269410744973*^9}, {3.61526944185774*^9, 3.615269459587545*^9}, {
   3.615269506653889*^9, 3.615269638933427*^9}},
 CellID->47814498]
}, Closed]],

Cell[CellGroupData[{

Cell["Minimum and Maximum (7)", "Subsection",
 CellChangeTimes->{{3.598590931675801*^9, 3.59859093339705*^9}, {
   3.603176738898614*^9, 3.603176741162154*^9}, {3.606038092419596*^9, 
   3.606038109017907*^9}, 3.606038383690134*^9, 3.608961808821316*^9, 
   3.614314746074089*^9, {3.615190959591014*^9, 3.615190966337248*^9}, 
   3.61519109880579*^9},
 CellTags->"sect::ProverMinMax",
 CellID->200612129],

Cell[TextData[{
 "Let ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " be a domain for which ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTOM", "[", "D", "]"}], TraditionalForm]]],
 " is known."
}], "Text",
 CellChangeTimes->{{3.601718382487292*^9, 3.60171840182645*^9}, {
  3.601885887118976*^9, 3.60188590863976*^9}, {3.604992043718783*^9, 
  3.604992047865769*^9}, {3.605342599144289*^9, 3.605342618131107*^9}, {
  3.606556326208996*^9, 3.606556329260534*^9}, {3.608972401683148*^9, 
  3.60897240525843*^9}, {3.615194691560793*^9, 3.615194697503868*^9}},
 CellID->949427665],

Cell[CellGroupData[{

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]ArgMin\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox[
      SubscriptBox["argmin", "D"], 
      RowBox[{"R", "[", "i", "]"}]], 
     RowBox[{"\[Phi]", "[", "i", "]"}]], 
    RowBox[{"t", "[", "i", "]"}]}], TraditionalForm]]],
 " appears somewhere in the current proof situation, where ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " is an arbitrary term: First check ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{"R", "[", "i", "]"}]], 
    RowBox[{"\[Phi]", "[", "i", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"R", "[", "i", "]"}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Phi]", "[", "i", "]"}], "\[Implies]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", 
        RowBox[{"t", "[", "i", "]"}], "]"}]}]}], ")"}]}], TraditionalForm]]],
 " in subproofs, and then continue with the main branch, where every \
occurrence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox[
      SubscriptBox["argmin", "D"], 
      RowBox[{"R", "[", "i", "]"}]], 
     RowBox[{"\[Phi]", "[", "i", "]"}]], 
    RowBox[{"t", "[", "i", "]"}]}], TraditionalForm]]],
 " got replaced by some new a.b.f. constant ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["i", "0"], TraditionalForm]]],
 " and additional knowledge ",
 Cell[BoxData[
  FormBox[
   RowBox[{"RC", "[", 
    SubscriptBox["i", "0"], "]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "[", 
    SubscriptBox["i", "0"], "]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"R", "[", "i", "]"}]], 
     RowBox[{"\[Phi]", "[", "i", "]"}]], 
    RowBox[{
     RowBox[{"t", "[", 
      SubscriptBox["i", "0"], "]"}], "\[LessEqual]", 
     RowBox[{"t", "[", "i", "]"}]}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"R", "[", "i", "]"}]], 
     RowBox[{"\[Phi]", "[", "i", "]"}]}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", 
      RowBox[{"t", "[", "i", "]"}], "]"}]}]}], TraditionalForm]]],
 " is in the list of assumptions (we do not need ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{"R", "[", "i", "]"}]], 
    RowBox[{"\[Phi]", "[", "i", "]"}]}], TraditionalForm]]],
 ", as this is covered by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"RC", "[", 
    SubscriptBox["i", "0"], "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "[", 
    SubscriptBox["i", "0"], "]"}], TraditionalForm]]],
 " anyway)."
}], "Item",
 CellChangeTimes->{{3.611035024774206*^9, 3.611035332732543*^9}, {
  3.611035399478875*^9, 3.611035434017837*^9}, {3.614334812864578*^9, 
  3.614334828839136*^9}, {3.615194006827499*^9, 3.615194126710483*^9}},
 CellID->1140449286],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]ArgMax\[CloseCurlyDoubleQuote]: Analoguous to \
\[OpenCurlyDoubleQuote]ArgMin\[CloseCurlyDoubleQuote] for appearances of the ",
 Cell[BoxData[
  FormBox["argmax", TraditionalForm]]],
 "-quantifier."
}], "Item",
 CellChangeTimes->{{3.611035024774206*^9, 3.611035332732543*^9}, {
   3.611035399478875*^9, 3.611035434017837*^9}, {3.614334812864578*^9, 
   3.614334828839136*^9}, {3.615194006827499*^9, 3.615194167231734*^9}, 
   3.615194203640003*^9},
 CellID->1761540416]
}, Open  ]],

Cell[TextData[{
 "The two inference rules above can be regarded \
\[OpenCurlyDoubleQuote]abbreviation\[CloseCurlyDoubleQuote] rules: Rather \
than always writing ",
 Cell[BoxData[
  FormBox["argmin", TraditionalForm]]],
 ", we replace it by some new constant ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["i", "0"], TraditionalForm]]],
 " for which we collect all knowledge available."
}], "Text",
 CellChangeTimes->{{3.606013999381632*^9, 3.606014087139855*^9}, {
  3.606458554834075*^9, 3.606458720831595*^9}, {3.6064588295372677`*^9, 
  3.606458829926656*^9}, {3.611035368517553*^9, 3.61103538311999*^9}, {
  3.6149375473474*^9, 3.614937551492182*^9}, {3.615191125729112*^9, 
  3.615191140580913*^9}, {3.615194173834164*^9, 3.615194185498004*^9}},
 CellID->10740590],

Cell[TextData[{
 StyleBox["\[OpenCurlyDoubleQuote]tupleAdditionGoal\[CloseCurlyDoubleQuote]: \
Let ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " be either an abstraction- or an enumeration tuple. ",
 StyleBox["Replace all occurrences of ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    UnderscriptBox["+", "D"], 
    RowBox[{
     UnderscriptBox["min", "D"], "[", "T", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" in the proof situation by ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["min", "D"], "[", "S", "]"}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 ", where ",
 Cell[BoxData[
  FormBox["S", TraditionalForm]]],
 " is a new tuple constructed from ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " by moving the \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "+"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] \
\[OpenCurlyDoubleQuote]into\[CloseCurlyDoubleQuote] the tuple. ",
 StyleBox["Analoguous for ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   SubscriptBox["max", "D"], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[".\[LineSeparator]This inference rule is derived from and justified \
by formulas ",
  FontColor->GrayLevel[0]],
 ButtonBox["(min+)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:446229453"],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 ButtonBox["(max+)",
  BaseStyle->"Hyperlink",
  ButtonData->"ID:513276857"],
 StyleBox[".\[LineSeparator]Note that if ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["T", TraditionalForm]],
  FormatType->"TraditionalForm"],
 " is an abstraction tuple and ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]],
  FormatType->"TraditionalForm"],
 " contains free variables, these variables must not get bound when moving ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]],
  FormatType->"TraditionalForm"],
 " into ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]],
  FormatType->"TraditionalForm"],
 ". This might happen in situations like ",
 Cell[BoxData[
  FormBox[
   RowBox[{"max", "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{
       RowBox[{"max", "[", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"s", "[", "i", "]"}], 
          UnderscriptBox["|", 
           RowBox[{"Q", "[", "i", "]"}]], 
          RowBox[{"\[Psi]", "[", "i", "]"}]}], "\[RightAngleBracket]"}], 
        "]"}], "+", 
       RowBox[{"t", "[", "i", "]"}]}], 
      UnderscriptBox["|", 
       RowBox[{"R", "[", "i", "]"}]], 
      RowBox[{"\[Phi]", "[", "i", "]"}]}], "\[RightAngleBracket]"}], "]"}], 
   TraditionalForm]],
  FormatType->"TraditionalForm"],
 "."
}], "Item",
 CellChangeTimes->{{3.598590951235552*^9, 3.598590997687429*^9}, {
   3.598591055395675*^9, 3.598591129885771*^9}, {3.598591177405831*^9, 
   3.598591249384736*^9}, {3.598591354801624*^9, 3.598591358481397*^9}, {
   3.598591407062554*^9, 3.598591464492893*^9}, {3.598591522216549*^9, 
   3.598591525871287*^9}, {3.598591809356681*^9, 3.598591862695048*^9}, {
   3.598591895192172*^9, 3.598591908028887*^9}, {3.598591979682789*^9, 
   3.598591990550081*^9}, {3.59859459695961*^9, 3.59859459696829*^9}, {
   3.599383291492991*^9, 3.599383291510159*^9}, {3.599805304577141*^9, 
   3.599805307382005*^9}, {3.600577054790803*^9, 3.600577140423743*^9}, {
   3.60057720621877*^9, 3.60057720952608*^9}, {3.600577250569121*^9, 
   3.600577262273926*^9}, {3.600577349214705*^9, 3.600577361097341*^9}, {
   3.601101905584679*^9, 3.60110199473978*^9}, {3.601108967980312*^9, 
   3.601109040932447*^9}, {3.601123459688401*^9, 3.601123459941486*^9}, {
   3.601197432969106*^9, 3.601197444697175*^9}, {3.601722674968618*^9, 
   3.60172270721309*^9}, {3.601872535782513*^9, 3.601872552397443*^9}, {
   3.601874759622533*^9, 3.601874769769402*^9}, {3.60188127124024*^9, 
   3.60188127190753*^9}, {3.601881337827681*^9, 3.601881338453982*^9}, {
   3.603176736351541*^9, 3.603176736982929*^9}, {3.604924613662981*^9, 
   3.604924623402014*^9}, {3.60498698441336*^9, 3.604987022939194*^9}, {
   3.60498706158977*^9, 3.604987112925446*^9}, {3.605855656951869*^9, 
   3.605855712013719*^9}, {3.605855874217542*^9, 3.605855896331088*^9}, {
   3.605861459558701*^9, 3.605861460154695*^9}, {3.605861495745945*^9, 
   3.605861632409263*^9}, {3.605861666789877*^9, 3.605861666900934*^9}, {
   3.606556775528048*^9, 3.606556810160855*^9}, {3.606556847798179*^9, 
   3.606556848166942*^9}, {3.608961827329256*^9, 3.608961864204154*^9}, 
   3.608961912234512*^9, {3.615194198016609*^9, 3.615194266090461*^9}, {
   3.615194306936194*^9, 3.615194438072764*^9}, {3.615194470794778*^9, 
   3.615194513630517*^9}, {3.621056011464853*^9, 3.621056138920353*^9}},
 CellID->150731591],

Cell[TextData[{
 "Let in the following ",
 Cell[BoxData[
  FormBox["S", TraditionalForm]]],
 " be an arbitrary term (not necessarily an abstraction/enumeration tuple), \
and let ",
 Cell[BoxData[
  FormBox["\[Precedes]", TraditionalForm]]],
 " be one of ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[LessEqual]", "D"], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["<", "D"], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.606555789060725*^9, 3.606555824151847*^9}, {
  3.606556860359372*^9, 3.606556863373284*^9}, {3.615194525393943*^9, 
  3.615194548137482*^9}, {3.615194677442376*^9, 3.61519469459037*^9}},
 CellID->293040326],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["\[OpenCurlyDoubleQuote]minGoalLeft\[CloseCurlyDoubleQuote]: If ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["min", "D"], "[", "S", "]"}], "\[Precedes]", "x"}], 
   TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" has to be shown, show ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[SquareSubsetEqual]", "D"], 
     RowBox[{"[", "S", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
     RowBox[{
      SubscriptBox["S", "i"], "\[Precedes]", "x"}]}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" instead.",
  FontColor->GrayLevel[0]]
}], "Item",
 CellChangeTimes->{{3.598590951235552*^9, 3.598590997687429*^9}, {
   3.598591055395675*^9, 3.598591129885771*^9}, {3.598591177405831*^9, 
   3.598591249384736*^9}, {3.598591354801624*^9, 3.598591358481397*^9}, {
   3.598591407062554*^9, 3.598591464492893*^9}, {3.598591522216549*^9, 
   3.598591525871287*^9}, {3.598591809356681*^9, 3.598591862695048*^9}, {
   3.598591895192172*^9, 3.598591908028887*^9}, {3.598591979682789*^9, 
   3.598591990550081*^9}, {3.59859459695961*^9, 3.59859459696829*^9}, {
   3.599383291492991*^9, 3.599383291510159*^9}, {3.599805304577141*^9, 
   3.599805307382005*^9}, {3.600577054790803*^9, 3.600577054803883*^9}, {
   3.600577474945249*^9, 3.60057748555418*^9}, {3.600577943355371*^9, 
   3.6005779461934*^9}, {3.600680449795398*^9, 3.600680641803613*^9}, {
   3.600680682416*^9, 3.600680713333048*^9}, {3.601122800822648*^9, 
   3.601122820488424*^9}, {3.601123183543547*^9, 3.60112318383904*^9}, {
   3.601188519289023*^9, 3.601188542679154*^9}, {3.60119465065298*^9, 
   3.601194676107137*^9}, {3.601194707692878*^9, 3.601194749029313*^9}, {
   3.601197273369169*^9, 3.601197274520638*^9}, {3.604923897090558*^9, 
   3.604923899225614*^9}, {3.604924004840649*^9, 3.604924085888489*^9}, {
   3.604924123088354*^9, 3.604924193700508*^9}, {3.606107660173972*^9, 
   3.606107668557477*^9}, {3.606555910176355*^9, 3.606555936326524*^9}, 
   3.608961896872073*^9, 3.608961950421352*^9, {3.615194560267882*^9, 
   3.615194563181316*^9}},
 CellID->3623347],

Cell[TextData[{
 StyleBox["\[OpenCurlyDoubleQuote]maxGoalRight\[CloseCurlyDoubleQuote]: If ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Precedes]", 
    RowBox[{
     UnderscriptBox["max", "D"], "[", "S", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" has to be shown, show ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[SquareSubsetEqual]", "D"], 
     RowBox[{"[", "S", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
     RowBox[{"x", "\[Precedes]", 
      SubscriptBox["S", "i"]}]}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" instead.",
  FontColor->GrayLevel[0]]
}], "Item",
 CellChangeTimes->{{3.598590951235552*^9, 3.598590997687429*^9}, {
   3.598591055395675*^9, 3.598591129885771*^9}, {3.598591177405831*^9, 
   3.598591249384736*^9}, {3.598591354801624*^9, 3.598591358481397*^9}, {
   3.598591407062554*^9, 3.598591464492893*^9}, {3.598591522216549*^9, 
   3.598591525871287*^9}, {3.598591809356681*^9, 3.598591862695048*^9}, {
   3.598591895192172*^9, 3.598591908028887*^9}, {3.598591979682789*^9, 
   3.598591990550081*^9}, {3.59859459695961*^9, 3.59859459696829*^9}, {
   3.599383291492991*^9, 3.599383291510159*^9}, {3.599805304577141*^9, 
   3.599805307382005*^9}, {3.600577054790803*^9, 3.600577054803883*^9}, {
   3.600577474945249*^9, 3.60057748555418*^9}, {3.600577943355371*^9, 
   3.6005779461934*^9}, {3.600680449795398*^9, 3.600680641803613*^9}, {
   3.600680682416*^9, 3.600680713333048*^9}, {3.601122800822648*^9, 
   3.601122820488424*^9}, {3.601123183543547*^9, 3.60112318383904*^9}, {
   3.601188519289023*^9, 3.601188542679154*^9}, {3.60119465065298*^9, 
   3.601194676107137*^9}, {3.601194707692878*^9, 3.601194749029313*^9}, {
   3.601197273369169*^9, 3.601197274520638*^9}, {3.604923897090558*^9, 
   3.604923899225614*^9}, {3.604924004840649*^9, 3.604924113182796*^9}, {
   3.604924200269719*^9, 3.604924211179594*^9}, {3.606107674044416*^9, 
   3.606107686136467*^9}, {3.606555974562094*^9, 3.606555996567017*^9}, 
   3.608961897910951*^9, 3.608961951792013*^9, {3.615194567606253*^9, 
   3.61519456956135*^9}},
 CellID->1634809851],

Cell[TextData[{
 StyleBox["\[OpenCurlyDoubleQuote]minKBRight\[CloseCurlyDoubleQuote]: If ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Precedes]", 
    RowBox[{
     SubscriptBox["min", "D"], "[", "S", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" is known, first check ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[SquareSubsetEqual]", "D"], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" without ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Precedes]", 
    RowBox[{
     SubscriptBox["min", "D"], "[", "S", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" in the knowledge base, and then continue with the main branch, \
where ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Precedes]", 
    RowBox[{
     SubscriptBox["min", "D"], "[", "S", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" got replaced by ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{"x", "\[Precedes]", 
     SubscriptBox["S", "i"]}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[SquareSubsetEqual]", "D"], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" is added to the list of assumptions.",
  FontColor->GrayLevel[0]]
}], "Item",
 CellChangeTimes->{{3.598590951235552*^9, 3.598590997687429*^9}, {
   3.598591055395675*^9, 3.598591129885771*^9}, {3.598591177405831*^9, 
   3.598591249384736*^9}, {3.598591354801624*^9, 3.598591358481397*^9}, {
   3.598591407062554*^9, 3.598591464492893*^9}, {3.598591522216549*^9, 
   3.598591525871287*^9}, {3.598591809356681*^9, 3.598591862695048*^9}, {
   3.598591895192172*^9, 3.598591908028887*^9}, {3.598591979682789*^9, 
   3.598591990550081*^9}, {3.59859459695961*^9, 3.59859459696829*^9}, {
   3.599383291492991*^9, 3.599383291510159*^9}, {3.599805304577141*^9, 
   3.599805307382005*^9}, {3.600577054790803*^9, 3.600577054803883*^9}, {
   3.600577474945249*^9, 3.60057748555418*^9}, {3.600577943355371*^9, 
   3.6005779461934*^9}, {3.600680449795398*^9, 3.600680641803613*^9}, {
   3.600680682416*^9, 3.600680713333048*^9}, {3.601122800822648*^9, 
   3.601122820488424*^9}, {3.601123183543547*^9, 3.60112318383904*^9}, {
   3.601188519289023*^9, 3.601188542679154*^9}, {3.60119465065298*^9, 
   3.601194676107137*^9}, {3.601194707692878*^9, 3.601194749029313*^9}, {
   3.601197273369169*^9, 3.601197274520638*^9}, {3.604923897090558*^9, 
   3.604923899225614*^9}, {3.604924004840649*^9, 3.604924085888489*^9}, {
   3.604924123088354*^9, 3.604924193700508*^9}, {3.604924293182153*^9, 
   3.604924329975882*^9}, {3.604924366714484*^9, 3.60492436904155*^9}, {
   3.604924456979537*^9, 3.604924476050724*^9}, {3.604924855731753*^9, 
   3.604924918631447*^9}, {3.605430975119181*^9, 3.60543104843514*^9}, {
   3.606106782125582*^9, 3.606106789142739*^9}, {3.606107623741606*^9, 
   3.606107632111128*^9}, {3.606556003401909*^9, 3.606556039423817*^9}, {
   3.606556102062807*^9, 3.60655611397418*^9}, {3.606556205178735*^9, 
   3.606556233385698*^9}, {3.606625372922651*^9, 3.606625411684756*^9}, {
   3.608961899432459*^9, 3.608961900666267*^9}, 3.608961952805673*^9, {
   3.615194572623006*^9, 3.615194597869759*^9}},
 CellID->785212987],

Cell[TextData[{
 StyleBox["\[OpenCurlyDoubleQuote]maxKBLeft\[CloseCurlyDoubleQuote]: If ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["max", "D"], "[", "S", "]"}], "\[Precedes]", "x"}], 
   TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" is known, first show ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[SquareSubsetEqual]", "D"], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" without ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["max", "D"], "[", "S", "]"}], "\[Precedes]", "x"}], 
   TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" in the knowledge base, and then continue with the main branch, \
where ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["max", "D"], "[", "S", "]"}], "\[Precedes]", "x"}], 
   TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" got replaced by ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     SubscriptBox["S", "i"], "\[Precedes]", "x"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[SquareSubsetEqual]", "D"], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" is added to the list of assumptions.",
  FontColor->GrayLevel[0]]
}], "Item",
 CellChangeTimes->{{3.598590951235552*^9, 3.598590997687429*^9}, {
   3.598591055395675*^9, 3.598591129885771*^9}, {3.598591177405831*^9, 
   3.598591249384736*^9}, {3.598591354801624*^9, 3.598591358481397*^9}, {
   3.598591407062554*^9, 3.598591464492893*^9}, {3.598591522216549*^9, 
   3.598591525871287*^9}, {3.598591809356681*^9, 3.598591862695048*^9}, {
   3.598591895192172*^9, 3.598591908028887*^9}, {3.598591979682789*^9, 
   3.598591990550081*^9}, {3.59859459695961*^9, 3.59859459696829*^9}, {
   3.599383291492991*^9, 3.599383291510159*^9}, {3.599805304577141*^9, 
   3.599805307382005*^9}, {3.600577054790803*^9, 3.600577054803883*^9}, {
   3.600577474945249*^9, 3.60057748555418*^9}, {3.600577943355371*^9, 
   3.6005779461934*^9}, {3.600680449795398*^9, 3.600680641803613*^9}, {
   3.600680682416*^9, 3.600680713333048*^9}, {3.601122800822648*^9, 
   3.601122820488424*^9}, {3.601123183543547*^9, 3.60112318383904*^9}, {
   3.601188519289023*^9, 3.601188542679154*^9}, {3.60119465065298*^9, 
   3.601194676107137*^9}, {3.601194707692878*^9, 3.601194749029313*^9}, {
   3.601197273369169*^9, 3.601197274520638*^9}, {3.604923897090558*^9, 
   3.604923899225614*^9}, {3.604924004840649*^9, 3.604924085888489*^9}, {
   3.604924123088354*^9, 3.604924193700508*^9}, {3.604924293182153*^9, 
   3.604924329975882*^9}, {3.604924366714484*^9, 3.60492436904155*^9}, {
   3.604924456979537*^9, 3.604924520835388*^9}, {3.604924929222957*^9, 
   3.604924933773508*^9}, {3.605431094926849*^9, 3.605431137208239*^9}, {
   3.606106744093363*^9, 3.606106748036133*^9}, {3.606106793307857*^9, 
   3.606106800232665*^9}, {3.606107638219164*^9, 3.606107646731812*^9}, {
   3.606556117134263*^9, 3.606556151003877*^9}, {3.606556208260194*^9, 
   3.606556236795792*^9}, {3.606625424607087*^9, 3.60662542494048*^9}, {
   3.608961918835192*^9, 3.608961919172217*^9}, 3.608961953957549*^9, {
   3.61519460037957*^9, 3.615194629807612*^9}},
 CellID->919021924]
}, Open  ]],

Cell[TextData[{
 "The four rules above are derived from and justified by lemma \
\[OpenCurlyDoubleQuote]General Properties of min, max\[CloseCurlyDoubleQuote] \
in section ",
 ButtonBox["Lemmata on Basic Notions / min, max",
  BaseStyle->"Hyperlink",
  ButtonData->"sect::Lemmataminmax"],
 "."
}], "Text",
 CellChangeTimes->{
  3.604924036442597*^9, {3.604924236588355*^9, 3.604924271846669*^9}, {
   3.606038396667116*^9, 3.606038439207433*^9}, {3.606556378567653*^9, 
   3.606556384374937*^9}, {3.61493756270872*^9, 3.61493756657727*^9}, {
   3.615194641452619*^9, 3.615194665092651*^9}, {3.615194705779324*^9, 
   3.61519477844072*^9}, 3.615194815542494*^9},
 CellID->1322323722]
}, Closed]],

Cell[CellGroupData[{

Cell["Logical Quantifiers (9)", "Subsection",
 CellChangeTimes->{{3.598592356946731*^9, 3.598592371704864*^9}, {
   3.601875708584831*^9, 3.601875713206451*^9}, {3.606118807906453*^9, 
   3.606118809430715*^9}, 3.610856145401972*^9, 3.614314716212415*^9, {
   3.615190923701344*^9, 3.615190946876231*^9}, 3.615191201197637*^9},
 CellTags->"sect::ProverQuantifiers",
 CellID->141176790],

Cell[CellGroupData[{

Cell["General (5) ", "Subsubsection",
 CellChangeTimes->{{3.600584152503577*^9, 3.600584153552682*^9}, {
   3.603176819329482*^9, 3.603176821838684*^9}, 3.604987432608838*^9, {
   3.605943807229989*^9, 3.6059438102837*^9}, 3.615195689617871*^9},
 CellID->592950640],

Cell[CellGroupData[{

Cell["\<\
\[OpenCurlyDoubleQuote]forallGoal\[CloseCurlyDoubleQuote]: Prove universally \
quantified goals by introducing a.b.f. constants. This, of course, is a \
general-purpose rule for predicate logic.\
\>", "Item",
 CellChangeTimes->{{3.615195732725015*^9, 3.615195736385069*^9}, {
  3.615195926863486*^9, 3.615195980049874*^9}},
 CellID->1506480535],

Cell["\<\
\[OpenCurlyDoubleQuote]existsKB\[CloseCurlyDoubleQuote]: Handle existentially \
quantified assumptions by introducing a.b.f. constants. Again, this is a \
general-purpose rule for predicate logic.\
\>", "Item",
 CellChangeTimes->{{3.615195732725015*^9, 3.615195740707096*^9}, {
  3.615195986726607*^9, 3.615196027673129*^9}},
 CellID->1743222265],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]existsGoal\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{"R", "[", "x", "]"}]], 
    RowBox[{"\[Psi]", "[", "x", "]"}]}], TraditionalForm]]],
 " has to be shown, try to find some formula ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Psi]", "[", "a", "]"}], TraditionalForm]]],
 " in the knowledge base (by matching) and try to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"RC", "[", "a", "]"}], TraditionalForm]]],
 " instead. If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Psi]", "[", "x", "]"}], TraditionalForm]]],
 " is a conjunction, try all conjuncts individually."
}], "Item",
 CellChangeTimes->{{3.600584165256539*^9, 3.600584251504224*^9}, {
   3.600584289342545*^9, 3.60058441062025*^9}, 3.601123883988031*^9, {
   3.601705511284627*^9, 3.601705627076142*^9}, {3.602585676211578*^9, 
   3.602585676509736*^9}, {3.604987429595718*^9, 3.604987557362627*^9}, {
   3.604987591318352*^9, 3.604987610150983*^9}, {3.605429216023087*^9, 
   3.605429249479229*^9}, 3.605943801293285*^9, {3.605943834268779*^9, 
   3.605943834984318*^9}, {3.615195752384282*^9, 3.615195765879388*^9}},
 CellID->1446434590],

Cell["\<\
\[OpenCurlyDoubleQuote]forallKBInteractive\[CloseCurlyDoubleQuote]: Offer the \
user the opportunity to interactively instantiate a universally quantified \
formula in the knowledge base by an arbitrary term.\
\>", "Item",
 CellChangeTimes->{{3.600584165256539*^9, 3.600584251504224*^9}, {
   3.600584289342545*^9, 3.60058441062025*^9}, 3.601123883988031*^9, {
   3.601705511284627*^9, 3.601705562292672*^9}, {3.602585673722177*^9, 
   3.602585674222251*^9}, {3.615195791351919*^9, 3.615195833485679*^9}, {
   3.61519588103339*^9, 3.615195881706364*^9}},
 CellID->1820681756],

Cell["\<\
\[OpenCurlyDoubleQuote]existsGoalInteractive\[CloseCurlyDoubleQuote]: Offer \
the user the opportunity to interactively instantiate an existentially \
quantified goal by an arbitrary term.\
\>", "Item",
 CellChangeTimes->{{3.600584165256539*^9, 3.600584251504224*^9}, {
   3.600584289342545*^9, 3.60058441062025*^9}, 3.601123883988031*^9, {
   3.601705511284627*^9, 3.601705627076142*^9}, {3.602585676211578*^9, 
   3.602585676509736*^9}, {3.604987429595718*^9, 3.604987429599878*^9}, {
   3.615195840936488*^9, 3.615195922508849*^9}},
 CellID->860887451]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Tuples (4)", "Subsubsection",
 CellChangeTimes->{{3.599998832832859*^9, 3.599998833844636*^9}, {
   3.610856146841501*^9, 3.610856147247268*^9}, 3.614314714260134*^9},
 CellID->942392630],

Cell[TextData[{
 "Let in the sequel always ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " denote the abstraction tuple ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"t", "[", "i", "]"}], 
     UnderscriptBox["|", 
      StyleBox[
       RowBox[{"R", "[", "i", "]"}],
       FontColor->GrayLevel[0]]], 
     RowBox[{"\[Phi]", "[", "i", "]"}]}], "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.599997793493615*^9, 3.599997806728233*^9}, {
  3.601123892465786*^9, 3.601123908416566*^9}, {3.601723127578074*^9, 
  3.601723131050642*^9}, {3.601723256596641*^9, 3.601723276568659*^9}, {
  3.60610703569077*^9, 3.606107037248928*^9}},
 CellID->2122823176],

Cell[CellGroupData[{

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]existsTupleOfGoal\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[Exists]", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Psi]", "[", 
      SubscriptBox["T", "j"], "]"}]], 
    RowBox[{"\[Chi]", "[", 
     SubscriptBox["T", "j"], "]"}]}], TraditionalForm]]],
 " has to be shown and the free occurrences of ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["\[Psi]", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["\[Chi]", TraditionalForm]]],
 " are only as subscript of ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 ", then show",
 StyleBox[" ",
  FontColor->RGBColor[1, 0, 0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    StyleBox[
     RowBox[{"isTuple", "[", "T", "]"}],
     FontColor->GrayLevel[0]], 
    StyleBox["\[And]",
     FontColor->GrayLevel[0]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{"R", "[", "i", "]"}]], 
      RowBox[{"\[Phi]", "[", "i", "]"}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"\[Psi]", "[", 
        RowBox[{"t", "[", "i", "]"}], "]"}], "\[And]", 
       RowBox[{"\[Chi]", "[", 
        RowBox[{"t", "[", "i", "]"}], "]"}]}], ")"}]}]}], TraditionalForm]]],
 " instead.\[LineSeparator]Please note that the quantified variable ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 " refers to the variable ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 " in the definition of ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " and ",
 StyleBox["not",
  FontSlant->"Italic"],
 " to the variable ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 " bounded by the existential quantifier! Therefore, it is of utmost \
importance that the only free occurrences of ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["\[Psi]", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["\[Chi]", TraditionalForm]]],
 " are as the subscript of ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " and that it ranges over the whole tuple.\[LineSeparator]The meaning of \
this rule is the following: If we have to show that ",
 StyleBox["some",
  FontSlant->"Italic"],
 " element of tuple ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " satisfies a condition ",
 Cell[BoxData[
  FormBox["\[Chi]", TraditionalForm]]],
 ", then, even without knowing the element of ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " at some concrete position ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 ", we can instead show that some ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "[", "i", "]"}], TraditionalForm]]],
 " satisfies ",
 Cell[BoxData[
  FormBox["\[Chi]", TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.598592394152113*^9, 3.598592461424908*^9}, {
   3.598592592738388*^9, 3.598592671098834*^9}, {3.598592758889508*^9, 
   3.598592785700652*^9}, {3.599285879148644*^9, 3.599285992871696*^9}, {
   3.599286361305899*^9, 3.599286364946977*^9}, {3.599286420143661*^9, 
   3.599286432904006*^9}, {3.599384757028697*^9, 3.599384757039084*^9}, 
   3.599996556599184*^9, {3.599997455840449*^9, 3.599997455846392*^9}, {
   3.599998637148942*^9, 3.599998687912178*^9}, {3.601123924161216*^9, 
   3.601123926963412*^9}, {3.601123976807669*^9, 3.601123978457824*^9}, {
   3.601124033434925*^9, 3.601124035945053*^9}, {3.601723166242113*^9, 
   3.601723168841783*^9}, {3.601791989074866*^9, 3.601791994829715*^9}, {
   3.601887547452641*^9, 3.601887585781316*^9}, {3.602393681233813*^9, 
   3.602393681567418*^9}, {3.606107040913272*^9, 3.606107051051655*^9}, {
   3.606107284937045*^9, 3.606107297503148*^9}, {3.60912641596815*^9, 
   3.609126416668632*^9}, {3.60912646614911*^9, 3.609126489491675*^9}, {
   3.609128809573387*^9, 3.609128818988264*^9}, {3.609130639517646*^9, 
   3.609130645776288*^9}, 3.609130714907609*^9, {3.615196037244059*^9, 
   3.615196142098297*^9}, {3.615196423644739*^9, 3.615196434621442*^9}, {
   3.615197283596573*^9, 3.615197285278178*^9}, {3.615265914532441*^9, 
   3.615266037853689*^9}},
 CellID->195555088],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]forallTupleOfKB\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Psi]", "[", 
      SubscriptBox["T", "j"], "]"}]], 
    RowBox[{"\[Chi]", "[", 
     SubscriptBox["T", "j"], "]"}]}], TraditionalForm]]],
 " is known",
 " and the free occurrences of ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["\[Psi]", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["\[Chi]", TraditionalForm]]],
 " are only as subscript of ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 ", first check ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "T", "]"}], TraditionalForm]]],
 " without ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Psi]", "[", 
      SubscriptBox["T", "j"], "]"}]], 
    RowBox[{"\[Chi]", "[", 
     SubscriptBox["T", "j"], "]"}]}], TraditionalForm]]],
 " in the knowledge base and then continue with the main branch where ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Psi]", "[", 
      SubscriptBox["T", "j"], "]"}]], 
    RowBox[{"\[Chi]", "[", 
     SubscriptBox["T", "j"], "]"}]}], TraditionalForm]]],
 " got replaced by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"R", "[", "i", "]"}]], 
      RowBox[{"\[Phi]", "[", "i", "]"}]], 
     RowBox[{"\[Psi]", "[", 
      RowBox[{"t", "[", "i", "]"}], "]"}]}], "\[Implies]", 
    RowBox[{"\[Chi]", "[", 
     RowBox[{"t", "[", "i", "]"}], "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "T", "]"}], TraditionalForm]]],
 " is added to the list of assumptions.\[LineSeparator]The same comments \
apply as for \[OpenCurlyDoubleQuote]existsTupleOfGoal\[CloseCurlyDoubleQuote]."
}], "Item",
 CellChangeTimes->{{3.598592394152113*^9, 3.598592461424908*^9}, {
   3.598592592738388*^9, 3.598592671098834*^9}, {3.598592758889508*^9, 
   3.598592785700652*^9}, {3.599285879148644*^9, 3.599285992871696*^9}, {
   3.599286361305899*^9, 3.599286364946977*^9}, {3.599286420143661*^9, 
   3.599286432904006*^9}, {3.599384757028697*^9, 3.599384757039084*^9}, 
   3.599996556599184*^9, {3.599997455840449*^9, 3.599997602982351*^9}, {
   3.599998711682987*^9, 3.59999872997333*^9}, {3.599998899016842*^9, 
   3.599998903294621*^9}, {3.601124038711374*^9, 3.601124078157575*^9}, {
   3.601124110442112*^9, 3.601124111076097*^9}, {3.601723177982893*^9, 
   3.601723180347282*^9}, {3.601791997116298*^9, 3.601792001330695*^9}, {
   3.606107101060448*^9, 3.606107251075941*^9}, {3.606107317426577*^9, 
   3.606107323237191*^9}, {3.606107489367409*^9, 3.606107573963766*^9}, {
   3.609126476582974*^9, 3.609126499647416*^9}, {3.609130685758996*^9, 
   3.609130696660461*^9}, 3.610872457033045*^9, {3.615196150132445*^9, 
   3.615196212666953*^9}, {3.615196441757883*^9, 3.615196453728816*^9}},
 CellID->1141312465]
}, Open  ]],

Cell[TextData[{
 "Let ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " be an a.b.f. constant or an abstraction tuple, and let ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " be an arbitrary term."
}], "Text",
 CellChangeTimes->{{3.600594923473379*^9, 3.600595020460938*^9}, {
  3.60170515893966*^9, 3.601705161615566*^9}, {3.601723411127004*^9, 
  3.601723442952857*^9}, {3.601872731097259*^9, 3.601872742895632*^9}, {
  3.602394075816891*^9, 3.602394097297812*^9}, {3.61519622255783*^9, 
  3.615196222931189*^9}},
 CellID->1396162367],

Cell[CellGroupData[{

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]forallKBAppend\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", 
        RowBox[{"A", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "x"}], "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Psi]", "[", 
      SubscriptBox[
       RowBox[{"(", 
        RowBox[{"A", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "x"}], ")"}], "i"], "]"}]], 
    RowBox[{"\[Chi]", "[", 
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{"A", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "x"}], ")"}], "i"], "]"}]}], TraditionalForm]]],
 " is known and the free occurrences of ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["\[Psi]", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["\[Chi]", TraditionalForm]]],
 " are only as subscript of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"A", " ", "\:293a", " ", "x"}], TraditionalForm]]],
 ", ",
 StyleBox["first show ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "A", "]"}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" without ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", 
        RowBox[{"A", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "x"}], "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Psi]", "[", 
      SubscriptBox[
       RowBox[{"(", 
        RowBox[{"A", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "x"}], ")"}], "i"], "]"}]], 
    RowBox[{"\[Chi]", "[", 
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{"A", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "x"}], ")"}], "i"], "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" in the knowledge base, and then continue with the main branch \
where ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", 
        RowBox[{"A", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "x"}], "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Psi]", "[", 
      SubscriptBox[
       RowBox[{"(", 
        RowBox[{"A", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "x"}], ")"}], "i"], "]"}]], 
    RowBox[{"\[Chi]", "[", 
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{"A", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "x"}], ")"}], "i"], "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" got replaced by ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
       RowBox[{"\[Psi]", "[", 
        SubscriptBox["A", "i"], "]"}]], 
      RowBox[{"\[Chi]", "[", 
       SubscriptBox["A", "i"], "]"}]}], ")"}], "\[And]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Psi]", "[", "x", "]"}], "\[Implies]", 
      RowBox[{"\[Chi]", "[", "x", "]"}]}], ")"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "A", "]"}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" is added to the list of assumptions.\[LineSeparator]The meaning \
of this rule is the following: If we know that ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox["\[Chi]", TraditionalForm]]],
 " holds for ",
 StyleBox["all",
  FontSlant->"Italic"],
 " elements of the tuple ",
 Cell[BoxData[
  FormBox[
   RowBox[{"A", " ", "\:293a", " ", "x"}], TraditionalForm]]],
 ", then we know that it holds in particular for all elements of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 ", as well as for ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.600594999163565*^9, 3.600595112130239*^9}, {
   3.600595213532617*^9, 3.600595450283225*^9}, {3.600595502189967*^9, 
   3.600595533073041*^9}, {3.60059559574161*^9, 3.600595615945768*^9}, {
   3.600595660052221*^9, 3.600595694779342*^9}, {3.60067250595661*^9, 
   3.600672580558069*^9}, {3.600684689256722*^9, 3.600684725759828*^9}, {
   3.601121765268354*^9, 3.601121798674764*^9}, {3.601197821139896*^9, 
   3.601197856105034*^9}, {3.601198449301652*^9, 3.601198450496138*^9}, {
   3.601787840415256*^9, 3.601787867781433*^9}, {3.601791900477154*^9, 
   3.601791903865362*^9}, {3.601872781637668*^9, 3.601872842457034*^9}, {
   3.601873511742547*^9, 3.601873589743851*^9}, {3.601873689553971*^9, 
   3.601873738326754*^9}, {3.601873806499796*^9, 3.601873809349706*^9}, {
   3.601875658520848*^9, 3.601875659152608*^9}, {3.601888190193463*^9, 
   3.601888190642067*^9}, {3.603176910926911*^9, 3.603176912950066*^9}, {
   3.605429410453501*^9, 3.605429517455282*^9}, 3.609126558718467*^9, {
   3.610853806411248*^9, 3.610853835083355*^9}, {3.610856150348152*^9, 
   3.610856178957407*^9}, {3.61519623484276*^9, 3.615196312380381*^9}, 
   3.615196418224599*^9, {3.61526605156931*^9, 3.615266123442472*^9}},
 CellID->628443461],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]existsGoalAppend\[CloseCurlyDoubleQuote]: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[Exists]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", 
        RowBox[{"A", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "x"}], "\[RightBracketingBar]"}]}]], 
     RowBox[{"\[Psi]", "[", 
      SubscriptBox[
       RowBox[{"(", 
        RowBox[{"A", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "x"}], ")"}], "i"], "]"}]], 
    RowBox[{"\[Chi]", "[", 
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{"A", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "x"}], ")"}], "i"], "]"}]}], TraditionalForm]]],
 " has to be shown and the free occurrences of ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["\[Psi]", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["\[Chi]", TraditionalForm]]],
 " are only as subscript of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"A", " ", "\:293a", " ", "x"}], TraditionalForm]]],
 ", show instead ",
 Cell[BoxData[
  FormBox[
   StyleBox[
    RowBox[{
     RowBox[{"isTuple", "[", "A", "]"}], "\[And]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         UnderscriptBox[
          UnderscriptBox["\[Exists]", 
           RowBox[{
            RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
            RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
          RowBox[{"\[Psi]", "[", 
           SubscriptBox["A", "i"], "]"}]], 
         RowBox[{"\[Chi]", "[", 
          SubscriptBox["A", "i"], "]"}]}], ")"}], "\[Or]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Psi]", "[", "x", "]"}], "\[And]", 
         RowBox[{"\[Chi]", "[", "x", "]"}]}], ")"}]}], ")"}]}],
    FontWeight->"Plain"], TraditionalForm]], "Input"],
 ".\[LineSeparator]",
 StyleBox["The meaning of this rule is similar to that of \
\[OpenCurlyDoubleQuote]forallKBAppend\[CloseCurlyDoubleQuote].",
  FontColor->GrayLevel[0]]
}], "Item",
 CellChangeTimes->{{3.600594999163565*^9, 3.600595112130239*^9}, {
   3.600595213532617*^9, 3.600595450283225*^9}, {3.600595502189967*^9, 
   3.600595533073041*^9}, {3.60059559574161*^9, 3.600595615945768*^9}, {
   3.600595660052221*^9, 3.600595694779342*^9}, {3.60067250595661*^9, 
   3.600672580558069*^9}, {3.600684689256722*^9, 3.600684725759828*^9}, {
   3.601121765268354*^9, 3.601121798674764*^9}, {3.601197821139896*^9, 
   3.601197856105034*^9}, {3.601198449301652*^9, 3.601198450496138*^9}, {
   3.601787840415256*^9, 3.601787867781433*^9}, {3.601791900477154*^9, 
   3.601791903865362*^9}, {3.601872781637668*^9, 3.601872842457034*^9}, {
   3.601873511742547*^9, 3.601873589743851*^9}, {3.601873689553971*^9, 
   3.601873738323456*^9}, {3.601873898224398*^9, 3.601874057818667*^9}, {
   3.601875662359577*^9, 3.60187566302833*^9}, {3.606107349590801*^9, 
   3.606107371221008*^9}, {3.610853788484851*^9, 3.610853802739024*^9}, 
   3.610856152924866*^9, {3.615196316181214*^9, 3.615196400159559*^9}, {
   3.615266136558107*^9, 3.615266175689809*^9}},
 CellID->1730837153]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Implications (2)", "Subsection",
 CellChangeTimes->{{3.598595353146819*^9, 3.59859535546102*^9}, {
   3.598595909951857*^9, 3.598595912131435*^9}, {3.603175279603071*^9, 
   3.603175280159508*^9}, 3.615191211107828*^9},
 CellID->721887106],

Cell[CellGroupData[{

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]modusPonensTollens\[CloseCurlyDoubleQuote]: If an \
implication in the knowledge base has the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{
        SubscriptBox["\[Phi]", "1"], "\[And]", 
        SubscriptBox["\[Phi]", "2"], "\[And]"}], "..."}], "\[And]", 
      SubscriptBox["\[Phi]", "n"]}], ")"}], "\[DoubleLongRightArrow]", 
    "\[Psi]"}], TraditionalForm]]],
 ", then remove all the ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "i"], TraditionalForm]]],
 " that ",
 StyleBox["follow immediately",
  FontSlant->"Italic"],
 " from the other assumptions. If no ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Phi]", "i"], TraditionalForm]]],
 " remains on the left-hand-side, add ",
 Cell[BoxData[
  FormBox["\[Psi]", TraditionalForm]]],
 " to the list of assumptions. Also, if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", "\[Psi]"}], TraditionalForm]]],
 " follows immediately from the other assumptions, add ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     SubscriptBox["\[Phi]", "1"]}], "\[Or]", 
    RowBox[{"\[Not]", 
     SubscriptBox["\[Phi]", "2"]}], "\[Or]", "\[Ellipsis]", "\[Or]", 
    RowBox[{"\[Not]", 
     SubscriptBox["\[Phi]", "n"]}]}], TraditionalForm]]],
 " to the list of assumptions."
}], "Item",
 CellChangeTimes->{{3.6000649903621*^9, 3.600065067971493*^9}, {
   3.601184235822677*^9, 3.601184246548843*^9}, {3.60118430268836*^9, 
   3.601184398770976*^9}, {3.602331090539752*^9, 3.602331091840338*^9}, {
   3.602392471757397*^9, 3.602392481017806*^9}, 3.602413664108626*^9, {
   3.61526620052062*^9, 3.615266332146642*^9}, {3.6152663779909*^9, 
   3.615266405393267*^9}},
 CellID->1647031003],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]selectImplicationInteractive\[CloseCurlyDoubleQuote]:\
 Offer the user the opportunity to interactively select an implication ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "\[Implies]", "\[Psi]"}], TraditionalForm]]],
 " in the knowledge base. Then, first check ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " in a subproof, and then continue showing the original goal, where the \
implication ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "\[Implies]", "\[Psi]"}], TraditionalForm]]],
 " got removed from the knowledge base and both of ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["\[Psi]", TraditionalForm]]],
 " are added instead."
}], "Item",
 CellChangeTimes->{{3.598595360836577*^9, 3.598595584656642*^9}, {
   3.598595880178237*^9, 3.598595894749849*^9}, {3.598596276104618*^9, 
   3.598596276498315*^9}, {3.599812883090181*^9, 3.599812892118937*^9}, {
   3.599976707489124*^9, 3.599976707499688*^9}, {3.601184274564189*^9, 
   3.601184286699556*^9}, {3.601874435803879*^9, 3.601874464381835*^9}, {
   3.602391711843116*^9, 3.60239211728956*^9}, {3.602392273229614*^9, 
   3.602392280717751*^9}, {3.602392356613139*^9, 3.602392390273979*^9}, {
   3.602392431134975*^9, 3.602392448676128*^9}, 3.60258564629137*^9, {
   3.615266433416904*^9, 3.615266576835161*^9}},
 CellID->1971611069]
}, Open  ]],

Cell["\<\
For a more detailed definition of \[OpenCurlyDoubleQuote]follows immediately\
\[CloseCurlyDoubleQuote] please see the next subsection on rewriting.\
\>", "Text",
 CellChangeTimes->{{3.615266909958769*^9, 3.615266940248469*^9}},
 CellID->799082915]
}, Closed]],

Cell[CellGroupData[{

Cell["Rewriting (5)", "Subsection",
 CellChangeTimes->{{3.600153335224692*^9, 3.600153337114085*^9}, {
   3.600404031087199*^9, 3.600404031989547*^9}, {3.601706303708319*^9, 
   3.601706304202345*^9}, {3.60317527042872*^9, 3.603175273340782*^9}, 
   3.615191222679845*^9},
 CellTags->"sect::ProverRewriting",
 CellID->1387546506],

Cell[CellGroupData[{

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]generateRewriteRules\[CloseCurlyDoubleQuote]: \
Generate rewrite rules from formulas in the knowledge. This in particular \
happens for (conditional) equalities, (conditional) equivalences, and \
implications, but some rewrite rules (like ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "\[Rule]", "True"}], TraditionalForm]]],
 ") are in fact generated for ",
 StyleBox["all",
  FontSlant->"Italic"],
 " formulas ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 ". Note that equalities and equivalences only give rise to rewrite rules \
from left to right, e.g. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", "b"}], TraditionalForm]]],
 " gives rise to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Rule]", "b"}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.615266628387319*^9, 3.615266853811293*^9}},
 CellID->681646954],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]goalRewriting\[CloseCurlyDoubleQuote]: Implications ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "\[Implies]", "\[Psi]"}], TraditionalForm]]],
 " (maybe quantified) yield so-called ",
 StyleBox["backward",
  FontSlant->"Italic"],
 " rewrite rules ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Psi]", "\[Rule]", "\[Phi]"}], TraditionalForm]]],
 ". Apply such backward rules on the current goal, if possible, i.e. try to \
show ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " rather than ",
 Cell[BoxData[
  FormBox["\[Psi]", TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.600153343404018*^9, 3.60015362613783*^9}, {
   3.600153696975778*^9, 3.600153701697688*^9}, 3.601033181094123*^9, {
   3.60103338362047*^9, 3.601033411116065*^9}, {3.601033515791998*^9, 
   3.601033625941186*^9}, 3.601033675025119*^9, {3.601033706609492*^9, 
   3.601033709805635*^9}, {3.601110001515985*^9, 3.601110001533205*^9}, {
   3.601874247476016*^9, 3.601874300879415*^9}, 3.602331128323561*^9, {
   3.615266871981376*^9, 3.615266880302475*^9}, {3.615266989415106*^9, 
   3.615267080188779*^9}},
 CellID->1996689018],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]elementarySubstitution\[CloseCurlyDoubleQuote]: \
Apply rewrite rules originating from ",
 StyleBox["quantified",
  FontSlant->"Italic"],
 " (and possibly conditional) equalities and equivalences to formulas/terms \
in the whole proof situation."
}], "Item",
 CellChangeTimes->{{3.600153343404018*^9, 3.60015362613783*^9}, {
   3.600153696975778*^9, 3.600153701697688*^9}, 3.601033181094123*^9, {
   3.60103338362047*^9, 3.601033411116065*^9}, {3.601033515791998*^9, 
   3.601033625941186*^9}, 3.601033675025119*^9, {3.601033706609492*^9, 
   3.601033709805635*^9}, {3.601110001515985*^9, 3.601110001533205*^9}, {
   3.601874247476016*^9, 3.601874300879415*^9}, 3.602331128323561*^9, {
   3.615266871981376*^9, 3.615266880302475*^9}, {3.615266989415106*^9, 
   3.615267159729576*^9}, {3.61526900230855*^9, 3.615269011445504*^9}},
 CellID->563991546],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]groundSubstituation\[CloseCurlyDoubleQuote]: Apply \
rewrite rules originating from (possibly conditional, ",
 StyleBox["but not quantified",
  FontSlant->"Italic"],
 ") equalities and equivalences to formulas/terms in the whole proof \
situation."
}], "Item",
 CellChangeTimes->{{3.600153343404018*^9, 3.60015362613783*^9}, {
   3.600153696975778*^9, 3.600153701697688*^9}, 3.601033181094123*^9, {
   3.60103338362047*^9, 3.601033411116065*^9}, {3.601033515791998*^9, 
   3.601033625941186*^9}, 3.601033675025119*^9, {3.601033706609492*^9, 
   3.601033709805635*^9}, {3.601110001515985*^9, 3.601110001533205*^9}, {
   3.601874247476016*^9, 3.601874300879415*^9}, 3.602331128323561*^9, {
   3.615266871981376*^9, 3.615266880302475*^9}, {3.615266989415106*^9, 
   3.615267212072227*^9}},
 CellID->1703267404],

Cell["\<\
\[OpenCurlyDoubleQuote]expandDef\[CloseCurlyDoubleQuote]: Apply rewrite rules \
originating from definitions (:=, :\[DoubleLeftRightArrow]) to formulas/terms \
in the whole proof situation.\
\>", "Item",
 CellChangeTimes->{{3.600153343404018*^9, 3.60015362613783*^9}, {
   3.600153696975778*^9, 3.600153701697688*^9}, 3.601033181094123*^9, {
   3.60103338362047*^9, 3.601033411116065*^9}, {3.601033515791998*^9, 
   3.601033625941186*^9}, 3.601033675025119*^9, {3.601033706609492*^9, 
   3.601033709805635*^9}, {3.601110001515985*^9, 3.601110001533205*^9}, {
   3.601874247476016*^9, 3.601874300879415*^9}, 3.602331128323561*^9, {
   3.615266871981376*^9, 3.615266880302475*^9}, {3.615266989415106*^9, 
   3.615267261108139*^9}},
 CellID->389717320]
}, Open  ]],

Cell[CellGroupData[{

Cell["Control", "Subsubsection",
 CellChangeTimes->{{3.615267509872139*^9, 3.615267510998789*^9}},
 CellID->1394287770],

Cell["The control of the rewrite system is as follows:", "Text",
 CellChangeTimes->{{3.615267298798421*^9, 3.615267308303354*^9}},
 CellID->1647651918],

Cell[TextData[{
 "Bigger parts of expressions are tried to be rewritten before smaller ones, \
meaning that a breadth-first search of rewritable positions is employed. \
Rules are tried in an order that reflects the order they have been generated \
in; Rules that have been generated later will be tried earlier.\nAlso, \
usually only the ",
 StyleBox["first",
  FontSlant->"Italic"],
 " applicable rule is applied to rewrite an expression, meaning that ",
 StyleBox["one",
  FontSlant->"Italic"],
 " old formula results in exactly ",
 StyleBox["one",
  FontSlant->"Italic"],
 " new formula and ",
 StyleBox["one",
  FontSlant->"Italic"],
 " proof situation, after rewriting, is replaced by ",
 StyleBox["one",
  FontSlant->"Italic"],
 " new proof situation. An exception of this are backward rules for \
simplifying the goal: There, really all possibilities are tried, resulting in \
possibly several alternative branches in the proof."
}], "Text",
 CellChangeTimes->{{3.615267316115904*^9, 3.615267327066819*^9}, {
  3.615267369466706*^9, 3.615267465982354*^9}, {3.615267745180208*^9, 
  3.615267900027703*^9}},
 CellID->1285921573],

Cell[TextData[{
 "A conditional rules is applied if and only if all of its conditions follow ",
 StyleBox["immediately",
  FontSlant->"Italic"],
 " from the current knowledge base. This, in more detail, means that backward \
rules are used for exhaustively transforming each individual condition in all \
possible ways, until either the truth constant True is reached (success) or \
the depth of the search exceeds a certain limit (failure). Please note that \
as soon as some formula ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " is reached which appears in the knowledge base, then the backward rule ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Phi]", "\[Rule]", "True"}], TraditionalForm]]],
 " generated from ",
 Cell[BoxData[
  FormBox["\[Phi]", TraditionalForm]]],
 " immediately yields success."
}], "Text",
 CellChangeTimes->{{3.615267519324456*^9, 3.615267633744194*^9}, {
  3.615267673364715*^9, 3.615267734353558*^9}, {3.615267917558795*^9, 
  3.615268002534794*^9}, {3.615268058072778*^9, 3.615268060809721*^9}},
 CellID->705235995]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Conclusion", "Section",
 CellChangeTimes->{{3.603530574120141*^9, 3.603530576199951*^9}},
 CellTags->"sect::Conclusion",
 CellID->1304797350],

Cell[TextData[{
 "In this report, we formalized the entire complexity analysis of Buchberger\
\[CloseCurlyQuote]s algorithm with chain criterion in the bivariate case in \
",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ", following the papers ",
 ButtonBox["[Buchberger1979]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979"],
 ", ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 " and ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 ".\nFor this we proceeded by first introducing the basic notions, such as \
the basic categories and facts about tuples, and then proving many auxiliary \
lemmata about these notions. Later, we went on introducing the notions \
directly related to the complexity analysis (e.g. ",
 Cell[BoxData[
  FormBox["chainCrit", TraditionalForm]]],
 ") and again proved some elementary lemmata about them (following the \
concept of the Theory Exploration Cycle). Finally, we were able to formulate \
and prove the three main theorems in the complexity analysis, as stated and \
proved in ",
 ButtonBox["[Buchberger1979a]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979a"],
 " and ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 ". However, the strategy pursued here slightly deviates from the one in the \
referenced literature; This is not because otherwise the theorems could not \
be proved, but simply because the new strategy allows to shorten and simplify \
the original proofs."
}], "Text",
 CellChangeTimes->{{3.604122553242186*^9, 3.604123081864733*^9}, {
  3.604123120857133*^9, 3.604123224061311*^9}, {3.604123490738593*^9, 
  3.604123601650569*^9}, {3.604124179827939*^9, 3.604124356416344*^9}, {
  3.604124405092871*^9, 3.604124514426458*^9}, {3.604124565335992*^9, 
  3.604124565560527*^9}, {3.604125967259874*^9, 3.604125968670049*^9}, {
  3.615274944360173*^9, 3.61527514762327*^9}},
 CellID->162614402],

Cell[TextData[{
 "One of the most powerful aspects of ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " is that once one has introduced some basic domains, one can, in one \
coherent software-framework, prove properties ",
 StyleBox["of",
  FontSlant->"Italic"],
 " them and also compute ",
 StyleBox["with",
  FontSlant->"Italic"],
 " them. Hence, we also performed some sample computations, just to have an \
indication that our notions really behave as they should. But even more is \
possible in Theorema: Not only can one reason ",
 StyleBox["about",
  FontSlant->"Italic"],
 " new notions, but even ",
 StyleBox["with",
  FontSlant->"Italic"],
 " new notions by means of a special prover designed for natural and \
efficient proving in a certain theory. In our case, we implemented the \
ComplexityProver capable of handling basic notions such as total order \
relation, tuples, and minimum and maximum."
}], "Text",
 CellChangeTimes->{
  3.615275106331673*^9, {3.61527515934448*^9, 3.615275226611365*^9}, {
   3.615275275990509*^9, 3.615275284924655*^9}, {3.615275320457048*^9, 
   3.615275515770415*^9}},
 CellID->106646726],

Cell[TextData[{
 "For the readers not familiar with the referenced literature it should \
perhaps be pointed out that the bound ",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    RowBox[{"maxdeg", "[", "F", "]"}]}], TraditionalForm]]],
 " for the degrees of the polynomials in the final Groebner basis can easily \
be converted into a bound on the number of steps that have to be carried out \
in ",
 ButtonBox["Algorithm 1",
  BaseStyle->"Hyperlink",
  ButtonData->"Alg::BB"],
 ":\nIf ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is the input set of the algorithm and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "F"], TraditionalForm]]],
 " is the maximum degree of the leading monomials of ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " (i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "F"], "=", 
    RowBox[{"maxdeg", "[", "A", "]"}]}], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 " is as usual the tuple of exponent vectors of leading monomials of ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 "), then the number of steps needed to compute the Groebner basis ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 " of ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " can roughly be bounded by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    FractionBox["3", "2"], 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "F", "\[RightBracketingBar]"}], "+", 
       RowBox[{"2", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           SubscriptBox["m", "F"], "+", "2"}], ")"}], "2"]}]}], ")"}], "4"]}],
    TraditionalForm]]],
 ".\nFor more details on this we refer to ",
 ButtonBox["[Buchberger1979]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1979"],
 " and ",
 ButtonBox["[Buchberger1983b]",
  BaseStyle->"Hyperlink",
  ButtonData->"Buchberger1983b"],
 ", but in any case note that all this is only true if a ",
 StyleBox["graded",
  FontSlant->"Italic"],
 " ordering term ordering is used."
}], "Text",
 CellChangeTimes->{{3.604126082342734*^9, 3.604126183274321*^9}, {
  3.604126491803887*^9, 3.604126648691576*^9}, {3.604126713360804*^9, 
  3.604126962110786*^9}, {3.604127088761217*^9, 3.604127088761337*^9}, {
  3.604127702737751*^9, 3.604127725469127*^9}, {3.60412775890933*^9, 
  3.604127760268932*^9}, {3.615275424939883*^9, 3.615275425049654*^9}, {
  3.615275528711246*^9, 3.615275597710889*^9}},
 CellID->215946755],

Cell[CellGroupData[{

Cell["Future Work", "Subsection",
 CellChangeTimes->{{3.604123667594846*^9, 3.604123672824708*^9}},
 CellID->807851642],

Cell["\<\
The work described in this report could be extended in various ways, \
including the following:\
\>", "Text",
 CellChangeTimes->{{3.604124586190402*^9, 3.604124616071094*^9}, {
  3.604125050189567*^9, 3.604125054251112*^9}},
 CellID->1748677645],

Cell[CellGroupData[{

Cell[TextData[{
 "Formalize the complexity analysis of the ",
 StyleBox["trivariate",
  FontSlant->"Italic"],
 " Buchberger\[CloseCurlyQuote]s algorithm, following ",
 ButtonBox["[Winkler1984]",
  BaseStyle->"Hyperlink",
  ButtonData->"Winkler1984"],
 ". The strategy pursued there, however, is quite different from the one \
pursued in this report."
}], "Item",
 CellChangeTimes->{{3.60412509633989*^9, 3.604125139785982*^9}, {
   3.604125178974143*^9, 3.604125305795701*^9}, {3.604125348573316*^9, 
   3.604125455841755*^9}, {3.604125974899225*^9, 3.604125976238177*^9}, 
   3.615275625663406*^9},
 CellID->1950615972],

Cell[TextData[{
 "Try to find even more elegant/shorter proofs, making use of the (semi-) \
automated proving capabilities of ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ". Maybe it is even possible to find a proof for the trivariate case in the \
same spirit and using the same approach as in the bivariate case."
}], "Item",
 CellChangeTimes->{{3.60412509633989*^9, 3.604125139785982*^9}, {
  3.604125178974143*^9, 3.604125305795701*^9}, {3.604125348573316*^9, 
  3.604125502907304*^9}, {3.604125834659222*^9, 3.604125847395353*^9}, {
  3.604125881497363*^9, 3.604125927258571*^9}},
 CellID->532188132],

Cell[TextData[{
 "Formalize the entire theory of Groebner bases in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ". This is certainly a very ambitious undertaking, simply because of the \
immense size of the theory. The author\[CloseCurlyQuote]s PhD project (name: \
\[OpenCurlyDoubleQuote]Formal Theory and Algorithmics of Noncommutative \
Groebner Bases\[CloseCurlyDoubleQuote]) aims at exactly this."
}], "Item",
 CellChangeTimes->{{3.60412509633989*^9, 3.604125139785982*^9}, {
  3.604125178974143*^9, 3.604125319096228*^9}, {3.604125581089095*^9, 
  3.604125676019727*^9}, {3.604125708913293*^9, 3.604125787517249*^9}},
 CellID->83015438]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Acknowledgements", "Subsection",
 CellChangeTimes->{{3.604124083829086*^9, 3.604124087539053*^9}},
 CellID->1975960852],

Cell[TextData[{
 "The author gratefully acknowledges the many valuable discussions about all \
aspects of the formalization of mathematics and related topics with Bruno \
Buchberger, and about ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " with ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 "\[CloseCurlyQuote]s chief developer Wolfgang Windsteiger. Also, the support \
in implementing the prover by Ren\[EAcute] de Alvarado y Petterson has to be \
mentioned."
}], "Text",
 CellChangeTimes->{{3.604124100541449*^9, 3.604124103971066*^9}, 
   3.604124153277186*^9, {3.604133656921827*^9, 3.604133689868887*^9}, {
   3.604133745792001*^9, 3.604133752129759*^9}, {3.604133782451439*^9, 
   3.604133814555758*^9}, {3.604133870519476*^9, 3.604133871333841*^9}, {
   3.604291875897877*^9, 3.604291883852471*^9}, 3.621057329834812*^9},
 CellID->1518974484],

Cell["\<\
The research was funded by the Austrian Science Fund (FWF): grant no. \
W1214-N15, project DK1\
\>", "Text",
 CellChangeTimes->{{3.604124114033624*^9, 3.604124121919355*^9}, {
  3.621061297277305*^9, 3.621061298439921*^9}, {3.621075492030087*^9, 
  3.621075494633593*^9}},
 CellID->1954625252]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["References", "Section",
 CellChangeTimes->{{3.596370825209113*^9, 3.596370827027887*^9}},
 CellTags->"sect::References",
 CellID->664836793],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Buchberger1965]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.60345611059905*^9, 3.603456115593603*^9}, 
   3.603456320278039*^9, 3.619842098906154*^9},
 CellTags->"Buchberger1965",
 CellID->1095181227],

Cell[TextData[{
 "B. Buchberger. ",
 StyleBox["Ein Algorithmus zum Auffinden der Basiselemente des \
Restklassenringes nach einem nulldimensionalen Polynomideal (An Algorithm for \
Finding the Basis Elements in the Residue Class Ring Modulo a Zero \
Dimensional Polynomial Ideal)",
  FontSlant->"Italic"],
 ". PhD thesis, Mathematical Institute, University of Innsbruck, Austria, \
1965. English translation in J. of Symbolic Computation, Special Issue on \
Logic, Mathematics, and Computer Science: Interactions. Vol. 41, Number 3-4, \
Pages 475-511, 2006."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.603456120493424*^9, 3.603456320278138*^9}, 
   3.619842098906268*^9},
 CellID->279883113]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Buchberger1979]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.596371597446167*^9, 3.596371611486461*^9}, 
   3.596371824429163*^9, {3.604061831286485*^9, 3.604061831471585*^9}, 
   3.619842103036083*^9},
 CellTags->"Buchberger1979",
 CellID->1664980398],

Cell[TextData[{
 "B. Buchberger.",
 StyleBox[" A Criterion for Detecting Unnecessary Reductions in the \
Construction of Gr\[ODoubleDot]bner Bases",
  FontSlant->"Italic"],
 ". In E. W. Ng, editor, ",
 StyleBox["Proceedings of the EUROSAM \[OpenCurlyQuote]79",
  FontSlant->"Italic"],
 ", Marseille, June 26-28, 1979. Volume 72 of ",
 StyleBox["Lecture Notes in Computer Science",
  FontSlant->"Italic"],
 ", pages 3-21. Springer Berlin Heidelberg New York, 1979."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.59637161899924*^9, 3.596371785373132*^9}, 
   3.596371824429256*^9, {3.603456433535892*^9, 3.603456503206051*^9}, 
   3.603456725545584*^9, 3.61984210303619*^9},
 CellID->472274594]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Buchberger1979a]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10002.},
 CellChangeTimes->{{3.596371044766668*^9, 3.596371054877347*^9}, 
   3.596371088061005*^9, {3.596371507908625*^9, 3.596371531085864*^9}, {
   3.604061833274983*^9, 3.604061833497113*^9}, 3.619842116000024*^9},
 CellTags->"Buchberger1979a",
 CellID->1739310831],

Cell[TextData[{
 "B. Buchberger and F. Winkler.",
 StyleBox[" ",
  FontSlant->"Italic"],
 "Miscellaneous Results on the Construction of Gr\[ODoubleDot]bner Bases for \
Polynomial Ideals. Technical Report 137, Insitute of Mathematics, University \
of Linz, 1979."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10002.},
 CellChangeTimes->{{3.596371088061115*^9, 3.596371142293485*^9}, {
   3.596371210313045*^9, 3.596371234752764*^9}, {3.596371327728629*^9, 
   3.596371351497704*^9}, {3.596371471616689*^9, 3.596371471984742*^9}, {
   3.603456535202491*^9, 3.60345653648724*^9}, 3.603456724660614*^9, 
   3.61984211600016*^9},
 CellID->695030642]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Buchberger1983a]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10003.},
 CellChangeTimes->{{3.603456578015535*^9, 3.603456581924968*^9}, 
   3.603456761996477*^9, 3.603513934874871*^9, 3.619842118817697*^9},
 CellTags->"Buchberger1983a",
 CellID->1873989285],

Cell[TextData[{
 "B. Buchberger. A Note on the Complexity of Constructing \
Gr\[ODoubleDot]bner-Bases. In J. A. van Hulzen, editor, ",
 StyleBox["Computer Algebra (Proceedings of EUROCAL \[OpenCurlyQuote]83",
  FontSlant->"Italic"],
 ", London, March 28-30, 1983. Volume 162 of ",
 StyleBox["Lecture Notes in Computer Science",
  FontSlant->"Italic"],
 ", pages 137-145. Springer Berlin Heidelberg New York Tokyo, 1983."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10003.},
 CellChangeTimes->{{3.603456587353917*^9, 3.603456723444704*^9}, 
   3.603456761996576*^9, 3.619842118817834*^9},
 CellID->1667055178]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Buchberger1983b]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10004.},
 CellChangeTimes->{{3.596371264749744*^9, 3.596371274424445*^9}, {
   3.596371537417806*^9, 3.596371549776122*^9}, {3.604061835631491*^9, 
   3.60406183581701*^9}, 3.619842121640693*^9},
 CellTags->"Buchberger1983b",
 CellID->212964405],

Cell["\<\
B. Buchberger. Miscellaneous Results on Gr\[ODoubleDot]bner Bases for \
Polynomial Ideals II. Technical Report, Department of Computer and \
Information Science, University of Delaware, 1983.\
\>", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10004.},
 CellChangeTimes->{{3.596371264749825*^9, 3.596371303461798*^9}, {
   3.596371376137117*^9, 3.596371397623846*^9}, 3.603456792429498*^9, 
   3.619842121640839*^9},
 CellID->1076453871]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Buchberger1998]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10005.},
 CellChangeTimes->{{3.603456804621774*^9, 3.603456808369957*^9}, 
   3.6034568719703*^9, 3.619842125368368*^9},
 CellTags->"Buchberger1998",
 CellID->748631075],

Cell[TextData[{
 "B. Buchberger. ",
 StyleBox["Introduction to Gr\[ODoubleDot]bner Bases",
  FontSlant->"Italic"],
 ". London Mathematical Society Lecture Notes Series 251. Cambridge \
University Press, April 1998."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10005.},
 CellChangeTimes->{{3.603456812938151*^9, 3.603456871970427*^9}, 
   3.619842125368503*^9},
 CellID->998593826]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Buchberger2003]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10006.},
 CellChangeTimes->{{3.603456946023747*^9, 3.603456949934347*^9}, 
   3.603457026451921*^9, 3.619842127494289*^9},
 CellTags->"Buchberger2003",
 CellID->1719245896],

Cell["\<\
B. Buchberger. Groebner Rings in Theorema: A Case Study in Functors and \
Categories. Technical Report 203-49, Johannes Kepler University Linz, \
Spezialforschungsbereich F013, November 2003.\
\>", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10006.},
 CellChangeTimes->{{3.603456957741322*^9, 3.60345702645201*^9}, 
   3.619842127494428*^9},
 CellID->2121810688]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Buchberger2006]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10007.},
 CellChangeTimes->{{3.603457059166129*^9, 3.603457067313015*^9}, 
   3.603457150023366*^9, 3.619842130778006*^9},
 CellTags->"Buchberger2006",
 CellID->1062875853],

Cell[TextData[{
 "B. Buchberger et al. Theorema: Towards Computer-Aided Mathematical Theory \
Exploration. ",
 StyleBox["Journal of Applied Logic,",
  FontSlant->"Italic"],
 " 4(4):470-504, 2006."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10007.},
 CellChangeTimes->{{3.603457072395706*^9, 3.603457117160242*^9}, 
   3.603457150023503*^9, {3.604126028793165*^9, 3.604126030598848*^9}, 
   3.61984213077815*^9},
 CellID->1317313158]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Buchberger2014]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10008.},
 CellChangeTimes->{{3.603523085186647*^9, 3.603523126894945*^9}, 
   3.619842134530241*^9},
 CellTags->"Buchberger2014",
 CellID->473088647],

Cell["\<\
B. Buchberger. Personal communication. Research Institute for Symbolic \
Computation, Johannes Kepler University Linz, 2014.\
\>", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10008.},
 CellChangeTimes->{{3.6035231021142*^9, 3.603523115799922*^9}, {
   3.603523163914543*^9, 3.603523201193668*^9}, {3.60352649470859*^9, 
   3.603526495298742*^9}, 3.61984213453035*^9},
 CellID->1592513850]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Giese2007]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10009.},
 CellChangeTimes->{{3.605438633609752*^9, 3.605438643494227*^9}, {
   3.605438738226962*^9, 3.605438738898523*^9}, 3.61984213688646*^9},
 CellTags->"Giese2007",
 CellID->723325741],

Cell["\<\
M. Giese and B. Buchberger. Towards Practical Reflection for Formal \
Mathematics. Technical Report 07-05, Research Institute for Symbolic \
Computation (RISC), Johannes Kepler University Linz, 2007.\
\>", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10009.},
 CellChangeTimes->{
  3.60543864349432*^9, {3.605438684934698*^9, 3.60543869292397*^9}, {
   3.605438740923809*^9, 3.605438850127956*^9}, 3.619842136886604*^9},
 CellID->899421835]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Mayr-Meyer1983]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10011.},
 CellChangeTimes->{{3.619842153313454*^9, 3.619842167818294*^9}, 
   3.619842317180252*^9},
 CellTags->"Mayr-Meyer1982",
 CellID->1957558217],

Cell[TextData[{
 "E. W. Mayr and A. R. Meyer. The complexity of the word problems for \
commutative semigroups and polynomial ideals. ",
 StyleBox["Advances in Mathematics",
  FontSlant->"Italic"],
 ", 46(3):305-329, December 1982."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10011.},
 CellChangeTimes->{{3.619842193090257*^9, 3.619842317180379*^9}},
 CellID->1492494109]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Theorema]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10010.},
 CellChangeTimes->{{3.604131960925735*^9, 3.604131963175378*^9}, 
   3.604132061221108*^9, 3.619842139276478*^9},
 CellTags->"Theorema",
 CellID->367095808],

Cell[TextData[{
 "The ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " System. ",
 ButtonBox["http://www.risc.jku.at/research/theorema/description/",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["http://www.risc.jku.at/research/theorema/description/"], None},
  ButtonNote->"http://www.risc.jku.at/research/theorema/description/"]
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10010.},
 CellChangeTimes->{{3.604131969301822*^9, 3.60413198100055*^9}, {
   3.604132032352632*^9, 3.604132061221199*^9}, {3.604132093683083*^9, 
   3.604132093689797*^9}, 3.619842139276586*^9, {3.621223994694447*^9, 
   3.621224004006975*^9}},
 CellID->1720721919]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Windsteiger1999]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10011.},
 CellChangeTimes->{{3.603457181033961*^9, 3.603457186403518*^9}, 
   3.603513673926154*^9, 3.619842141615631*^9},
 CellTags->"Windsteiger1999",
 CellID->470851853],

Cell[TextData[{
 "W. Windsteiger. Building Up Hierarchical Mathematical Domains Using \
Functors in THEOREMA. In A. Armando and T. Jebelean, editors, ",
 StyleBox["Electronic Notes in Theoretical Computer Science",
  FontSlant->"Italic"],
 ", volume 23 of ",
 StyleBox["ENTCS",
  FontSlant->"Italic"],
 ", pages 401-419. Elsevier, 1999."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10011.},
 CellChangeTimes->{{3.603457198368583*^9, 3.603457230666555*^9}, {
   3.603513596488019*^9, 3.603513673926308*^9}, {3.604061798399996*^9, 
   3.604061798621634*^9}, 3.619842141615737*^9},
 CellID->1151878676]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["[Winkler1984]",
 FontWeight->"Bold"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10012.},
 CellChangeTimes->{{3.604124696501531*^9, 3.604124704962548*^9}, 
   3.604124864840077*^9, 3.619842143838105*^9},
 CellTags->"Winkler1984",
 CellID->805682494],

Cell[TextData[{
 "F. Winkler. On the Complexity of the Gr\[ODoubleDot]bner-Bases Algorithm \
over K[x,y,z]. In J. Fitch, editor,",
 StyleBox[" International Symposium on Symbolic and Algebraic Computation \
(EUROSAM \[CloseCurlyQuote]84)",
  FontSlant->"Italic"],
 ", Cambridge, England, June 1984. Springer-Verlag, pages 184-194, 1984."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10012.},
 CellChangeTimes->{{3.604124709527592*^9, 3.604124766748305*^9}, {
   3.604124810791434*^9, 3.604124864840193*^9}, 3.619842143838273*^9},
 CellID->696375225]
}, Open  ]]
}, Closed]]
}, Open  ]]
},
WindowSize->{1061, 969},
WindowMargins->{{112, Automatic}, {Automatic, 0}},
PrivateNotebookOptions->{"NoteStyle"->"NumberedNote"},
ShowSelection->True,
CounterAssignments->{{"TheoremaFormulaCounter" -> 311}},
Magnification->1.,
FrontEndVersion->"9.0 for Linux x86 (32-bit) (February 7, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Theorema"}, "TheoremaCore.nb"]]], 
   
   Cell["Modify the definitions below to change the default appearance of all \
cells in a given style. \nMake modifications to any definition using commands \
in the Format menu.", "Text"], 
   Cell[
    CellGroupData[{
      Cell["Notebook Options Settings", "Section"], 
      Cell[
      "The options defined for the style below will be used at the Notebook \
level.", "Text"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Notebook"], InputAliases -> {"and" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[And]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "or" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", "\[Or]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "not" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Not]", "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "impl" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Implies]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[DoubleLongLeftRightArrow]", 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Equal]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox[
                  
                  RowBox[{
                   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                  "a\[DoubleLongLeftRightArrow]b"], "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", ":=", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "andn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[And]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "orn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Or]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "equivn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[DoubleLeftRightArrow]", "\[Piecewise]", 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "cdist" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Piecewise]", 
                 
                 GridBox[{{
                   "\[SelectionPlaceholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}, {
                   "\[Placeholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "far" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exr" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "farc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exrc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "let" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"\[Placeholder]", "=", "\[Placeholder]"}]], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "app" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293a", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "prep" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293b", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "join" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22c8", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "elemT" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22ff", Identity, SyntaxForm -> "a+b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "subop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "supop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "subsupop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubsuperscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "oop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "uoop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderoverscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "osop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
                "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "op" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               TagBox["\[SelectionPlaceholder]", Identity], 
               TagBox[")", "AutoParentheses"]}], "[]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e114", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e115", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "()" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:fd3e", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:fd3f", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e8", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e9", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e117", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e118", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "[[]]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e6", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e7", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(())" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2e28", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2e29", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<<>>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27ea", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27eb", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{{}}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2983", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2984", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(" -> 
            TagBox["(", "AutoParentheses"], ")" -> 
            TagBox[")", "AutoParentheses"]}, Background -> 
          RGBColor[0.957, 0.96, 0.97]], 
         Cell[
          StyleData["Notebook", "Printout"], Background -> None]}, Open]]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell["Style Environment Names", "Section"], 
      Cell[
       StyleData[All, "Working"], DockedCells -> {
         Cell[
         "Theorema 2.0", "SBO", FontTracking -> 8, 
          CellMargins -> {{0, 0}, {0, 0}}, 
          CellFrameMargins -> {{0, 50}, {0, 0}}, FontSize -> 28, FontSlant -> 
          "Plain", FontColor -> RGBColor[0.29, 0.46, 0.6], TextAlignment -> 
          Right, Background -> RGBColor[0.36, 0.54, 0.69]]}], 
      Cell[
       StyleData[
       All, "Layout", StyleDefinitions -> StyleData[All, "Working"]], 
       MenuSortingValue -> 1200]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Title and Section Cells", "Section"], 
      Cell[
       CellGroupData[{
         Cell["Title", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, 27}, {10, 30}}, 
             TextAlignment -> Center, MenuSortingValue -> 1100, 
             MenuCommandKey -> "1", FontFamily -> "Bitstream Charter", 
             FontSize -> 36, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Title", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{6, 10}, {5, 30}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, 27}, {20, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1150, FontFamily -> 
             "Bitstream Charter", FontSize -> 24, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{6, 10}, {5, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], CellMargins -> {{27, 27}, {8, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], CellMargins -> {{27, 27}, {18, 20}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Sectioning", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], ShowGroupOpener -> True, MenuSortingValue -> 
             1250, MenuCommandKey -> "4", FontFamily -> "Bitstream Charter", 
             FontSize -> 20, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Section", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{2, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {8, 12}}, 
             MenuSortingValue -> 1300, MenuCommandKey -> "5", FontFamily -> 
             "Bitstream Charter", FontSize -> 14, FontWeight -> "Bold", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Subsection", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{9, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {2, 10}}, 
             MenuSortingValue -> 1350, MenuCommandKey -> "6", FontFamily -> 
             "Bitstream Charter", FontSize -> 12, FontWeight -> "Bold", 
             FontSlant -> "Italic", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsection", "Printout"], ShowGroupOpener -> False,
              CellMargins -> {{35, 0}, {7, 14}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsubsection"], CellDingbat -> None, 
             ShowGroupOpener -> True, 
             CellMargins -> {{27, Inherited}, {Inherited, Inherited}}, 
             FontFamily -> "Bitstream Charter", FontSize -> 11, FontWeight -> 
             "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsubsection", "Printout"], 
             CellMargins -> {{50, 0}, {7, 14}}]}, Closed]]}, Open]]}, Open]], 
   
   Cell[
    CellGroupData[{
      Cell["Styles for Publication\[Hyphen]ready Documents", "Section"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Author"], CellMargins -> {{27, 30}, {20, 5}}, 
          MenuSortingValue -> 1900, FontFamily -> "Helvetica", FontSize -> 14,
           FontSlant -> "Italic"], 
         Cell[
          StyleData["Author", "Printout"], CellMargins -> {{6, 10}, {0, 10}}, 
          FontSize -> 10]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Abstract"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
          CellMargins -> {{27, 140}, {20, 20}}, LineSpacing -> {0.9, 0}, 
          MenuSortingValue -> 2050, FontFamily -> "Times", FontSize -> 12], 
         Cell[
          StyleData["Abstract", "Printout"], Hyphenation -> True, 
          LineSpacing -> {1, 2}, FontSize -> 10]}, Closed]]}, Closed]], 
   Cell[
    CellGroupData[{
      Cell["Text", "Section"], 
      Cell[
       StyleData["Text"], CellMargins -> {{27, 27}, {Inherited, Inherited}}, 
       TextJustification -> 1., ParagraphIndent -> 15, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["EnvironmentText"], ShowCellBracket -> Automatic, 
       CellMargins -> {{82, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextJustification -> 1., ParagraphSpacing -> {0.3, 0.}, 
       MenuSortingValue -> 1480, MenuCommandKey -> "T", FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Item"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "Item", 
          CounterAssignments -> {{"Subitem", 0}, {"Subsubitem", 0}}, 
          MenuSortingValue -> 1600, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Item", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemParagraph"], CellMargins -> {{81, 10}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          DefaultNewCellStyle -> "Item", MenuSortingValue -> 1600, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemParagraph", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "Subitem", 
          CounterAssignments -> {{"Subsubitem", 0}}, MenuSortingValue -> 1610,
           FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subitem", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemParagraph"], CellMargins -> {{105, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subitem", 
          MenuSortingValue -> 1610, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemParagraph", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subsubitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1620, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subsubitem", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemParagraph"], 
          CellMargins -> {{129, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1625, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemParagraph", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "ItemNumbered", 
          CounterAssignments -> {{"SubitemNumbered", 0}, {
            "SubsubitemNumbered", 0}}, MenuSortingValue -> 1630, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemNumbered", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "SubitemNumbered", 
          CounterAssignments -> {{"SubsubitemNumbered", 0}}, MenuSortingValue -> 
          1640, FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemNumbered", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], ".", 
              CounterBox["SubsubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> 
          "SubitemNumbered", CounterIncrements -> "SubsubitemNumbered", 
          MenuSortingValue -> 1645, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemNumbered", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Theorema", "Section"], 
      Cell[
       StyleData["GlobalDeclaration"], CellFrame -> {{0, 2}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {2, 2}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       InitializationCell -> True, CellFrameMargins -> 10, CellFrameColor -> 
       RGBColor[0.73, 0.65, 0.61], CellFrameLabelMargins -> 33, 
       ShowAutoStyles -> False, MenuSortingValue -> 1441, MenuCommandKey -> 
       "G", FontSize -> 14], 
      Cell[
       StyleData["OpenEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {-4, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       StyleData["EnvironmentHeader"], 
       CellFrame -> {{False, False}, {False, False}}, 
       CellMargins -> {{76, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 6, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       ShowAutoStyles -> False, MenuSortingValue -> 1450, FontSize -> 14, 
       FontWeight -> "Bold", FontVariations -> {"CapsType" -> "SmallCaps"}, 
       Background -> RGBColor[0.89, 0.8, 0.69]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FormalTextInputFormula"], ShowCellBracket -> Automatic, 
          CellMargins -> {{96, 10}, {2, 2}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
          InitializationCell -> True, CellFrameLabelMargins -> 53, 
          ShowAutoStyles -> False, LimitsPositioningTokens -> {}, 
          MenuSortingValue -> 1470, MenuCommandKey -> "F", FontWeight -> 
          "Bold", FontColor -> RGBColor[0.25, 0.19, 0.2]], 
         Cell[
          StyleData["FormalTextInputFormula", "Layout"], Editable -> False], 
         Cell[
          StyleData["FormalTextInputFormula", "Presentation"], Editable -> 
          False]}, Open]], 
      Cell[
       StyleData["EndEnvironmentMarker"], 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextAlignment -> Right, MenuSortingValue -> 1480], 
      Cell[
       StyleData["CloseEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {False, True}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {Inherited, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Computation"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 7}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          MenuSortingValue -> 1510, MenuCommandKey -> "C", FontWeight -> 
          "Bold", FontColor -> RGBColor[0.23, 0.25, 0.34], Background -> 
          RGBColor[0.88, 0.88, 0.92]], 
         Cell[
          StyleData["Computation", "Layout"], Editable -> False], 
         Cell[
          StyleData["Computation", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       StyleData["ComputationResult"], CellFrame -> {{1.5, 0}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
       CellFrameMargins -> {{8, 8}, {4, 3}}, CellFrameColor -> 
       RGBColor[0.486, 0.33, 0.32], StyleMenuListing -> None, FontColor -> 
       RGBColor[0.23, 0.25, 0.34], Background -> RGBColor[0.88, 0.88, 0.92]], 
      
      Cell[
       StyleData["ComputationInfo"], ShowCellBracket -> False, 
       CellMargins -> {{120, 10}, {10, 2}}, StyleMenuListing -> None, 
       FontFamily -> $CellContext`Arial], 
      Cell[
       StyleData["ComputationInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{144, 10}, {10, 2}}, StyleMenuListing -> None], 
      Cell[
       StyleData["ProofInfo"], ShowCellBracket -> False, 
       CellMargins -> {{76, 10}, {6, 6}}, CellFrameLabelMargins -> 15, 
       StyleMenuListing -> None, FontFamily -> "Arial"], 
      Cell[
       StyleData["ProofInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{100, 10}, {0, 0}}, StyleMenuListing -> None], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FrameLabel"], LanguageCategory -> None, StyleMenuListing -> 
          None, FontFamily -> "Times New Roman", FontSize -> 10, FontWeight -> 
          "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
          FontVariations -> {
           "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
            "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
          RGBColor[0.55, 0.44, 0.42], Background -> None], 
         Cell[
          StyleData["FrameLabel", "Printout"]]}, Open]], 
      Cell[
       StyleData["OpenArchive"], CellFrame -> {{False, False}, {False, True}},
        ShowCellBracket -> Automatic, CellMargins -> {{70, 70}, {0, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 20}, CellFrameMargins -> 3, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], CellFrameLabelMargins -> 1, 
       DefaultFormatType -> DefaultInputFormatType, ShowAutoStyles -> False, 
       MenuSortingValue -> 1210, FontWeight -> Bold, Background -> 
       RGBColor[0.73, 0.74, 0.84]], 
      Cell[
       StyleData["ArchiveInfo"], ShowCellBracket -> Automatic, 
       CellMargins -> {{70, 70}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 5, CellFrameLabelMargins -> 3, ShowCellTags -> 
       False, ShowAutoStyles -> False, MenuSortingValue -> 1220, FontWeight -> 
       "Bold", Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["CloseArchive"], 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> 
       Automatic, CellMargins -> {{0, 0}, {5, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 25}, CellFrameMargins -> 2, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], TextAlignment -> Center, 
       MenuSortingValue -> 1230, FontFamily -> "Arial Black", FontSize -> 24, 
       FontWeight -> "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       Background -> RGBColor[0.73, 0.65, 0.61]], 
      Cell[
       StyleData["IncludeArchive"], CellFrame -> {{4, 0}, {False, False}}, 
       ShowCellBracket -> Automatic, 
       CellMargins -> {{81, 100}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameColor -> RGBColor[0.55, 0.44, 0.42], 
       CellFrameLabels -> {{"\[RightSkeleton]", None}, {None, None}}, 
       CellFrameLabelMargins -> 25, MenuSortingValue -> 1445], 
      Cell[
       StyleData["ExpressionVariable"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionABF"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionMeta"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["AutoParentheses"], SpanMaxSize -> 3, FontOpacity -> 
          0.15], 
         Cell[
          StyleData["AutoParentheses", "Layout"], FontOpacity -> 0], 
         Cell[
          StyleData["AutoParentheses", "SlideShow"], FontOpacity -> 0]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Emphasized"], MenuCommandKey -> "E", FontSlant -> 
          "Italic"], 
         Cell[
          StyleData["Emphasized", "SlideShow"], FontSlant -> "Plain", 
          FontColor -> RGBColor[0.36, 0.54, 0.69]]}, Open]], 
      Cell[
       StyleData["GenericButton"], FontFamily -> "Bitstream Charter", 
       FontWeight -> "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
       FontVariations -> {
        "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
         "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
       RGBColor[0.23, 0.25, 0.34]], 
      Cell[
       StyleData["TheoremaForm"], "TwoByteSyntaxCharacterAutoReplacement" -> 
       True, LineSpacing -> {1.25, 0, 2}, SingleLetterItalics -> True, 
       StyleMenuListing -> None, FontFamily -> "Courier"], 
      Cell[
       StyleData["ComponentActive"], FontFamily -> "Times New Roman"], 
      Cell[
       StyleData["ComponentInactive"], FontFamily -> "Times New Roman", 
       FontOpacity -> 0.4], 
      Cell[
       StyleData["ComponentEmpty"], FontFamily -> "Times New Roman", 
       FontVariations -> {"StrikeThrough" -> True}], 
      Cell[
       StyleData["TabView"], FontFamily -> "Times New Roman", FontSize -> 10], 
      Cell[
       StyleData["TabViewLabel"], FontFamily -> "Times New Roman", FontSize -> 
       10, FontWeight -> "Bold"], 
      Cell[
       StyleData["FormReference", StyleDefinitions -> StyleData["Text"]], 
       FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> "Plain",
        FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["FormReferenceHover", StyleDefinitions -> StyleData["Text"]],
        FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> 
       "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> True}, 
       FontColor -> RGBColor[0.404, 0.43, 0.545]], 
      Cell[
       StyleData["DisplayFormula"], CellMargins -> {{36, 6}, {7, 7}}, 
       ConversionRules :> {
        "HTML" -> {{"", ""}, {"<p class=\"Equation\">", "</p>"}}}, 
       CellHorizontalScrolling -> True, DefaultFormatType -> TraditionalForm, 
       InputAutoReplacements -> {}, TextAlignment -> Center, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       ButtonBoxOptions -> {Appearance -> {Automatic, None}}]}, Open]]}, 
  Visible -> False, FrontEndVersion -> 
  "9.0 for Linux x86 (32-bit) (February 7, 2013)", StyleDefinitions -> 
  "StylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "sect::FormalMathematics"->{
  Cell[29794, 775, 263, 5, 43, "Section",
   CellTags->"sect::FormalMathematics",
   CellID->1882786356]},
 "sect::Conditional"->{
  Cell[52337, 1276, 175, 3, 38, "Subsection",
   CellTags->"sect::Conditional",
   CellID->1621317523]},
 "sect::ProblemSetting"->{
  Cell[66413, 1642, 155, 3, 43, "Section",
   CellTags->"sect::ProblemSetting",
   CellID->264353757]},
 "Alg::BB"->{
  Cell[71242, 1782, 473, 13, 41, "Text",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->"Alg::BB",
   CellID->1191384728]},
 "Eqn::ChainCrit"->{
  Cell[81296, 2113, 3499, 96, 70, "Text",
   CellTags->"Eqn::ChainCrit",
   CellID->1279890854]},
 "Eqn::M"->{
  Cell[94957, 2567, 658, 20, 27, "Text",
   CellTags->"Eqn::M",
   CellID->862257598]},
 "Eqn::W"->{
  Cell[95618, 2589, 671, 21, 28, "Text",
   CellTags->"Eqn::W",
   CellID->2012096862]},
 "sect::BasicNotions"->{
  Cell[99119, 2709, 200, 4, 43, "Section",
   CellTags->"sect::BasicNotions",
   CellID->2131707921]},
 "sect::DefinitionTuples"->{
  Cell[100369, 2750, 179, 4, 28, "Subsubsection",
   CellTags->"sect::DefinitionTuples",
   CellID->25850532]},
 "ID:1969097229"->{
  Cell[103698, 2839, 1563, 37, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1969097229", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length is natural"},
   CellID->1969097229]},
 "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"->{
  Cell[103698, 2839, 1563, 37, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1969097229", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length is natural"},
   CellID->1969097229],
  Cell[105264, 2878, 2068, 66, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1508742051", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "tuple equality"},
   CellID->1508742051],
  Cell[107335, 2946, 1125, 31, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1264372773", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isTuple\:293a"},
   CellID->1264372773],
  Cell[108463, 2979, 1139, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2086441677", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a"},
   CellID->2086441677],
  Cell[109605, 3012, 1431, 43, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:585712945", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "element\:293a 1"},
   CellID->585712945],
  Cell[111039, 3057, 1123, 36, 24, "FormalTextInputFormula",
   CellTags->{
    "ID:1075236060", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "element\:293a 2"},
   CellID->1075236060],
  Cell[115173, 3180, 446, 11, 32, "GlobalDeclaration",
   CellTags->{
    "ID:1405024037", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->1405024037],
  Cell[115622, 3193, 1703, 48, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:379346204", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "domain tuple"},
   CellID->379346204],
  Cell[118776, 3290, 791, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1473378407", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "other coordinate 1"},
   CellID->1473378407],
  Cell[119570, 3313, 788, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:406630295", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "other coordinate 2"},
   CellID->406630295],
  Cell[124938, 3463, 950, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1091916109", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Z is TOM"},
   CellID->1091916109],
  Cell[125891, 3489, 841, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1845867532", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Z is totally ordered"},
   CellID->1845867532],
  Cell[126735, 3513, 1282, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:492790093", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "some integers"},
   CellID->492790093],
  Cell[128020, 3547, 844, 23, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1907318616", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1\[LessEqual]2"},
   CellID->1907318616],
  Cell[129366, 3584, 1402, 41, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1708561859", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1"},
   CellID->1708561859],
  Cell[130771, 3627, 1152, 32, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:49894867", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "2"},
   CellID->49894867],
  Cell[135727, 3769, 1253, 39, 45, "FormalTextInputFormula",
   CellTags->{
    "ID:349901806", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "closure"},
   CellID->349901806],
  Cell[136983, 3810, 1145, 33, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1476372605", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "0 in domain"},
   CellID->1476372605],
  Cell[138131, 3845, 1343, 42, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1359514637", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "0 is neutral"},
   CellID->1359514637],
  Cell[139477, 3889, 1589, 50, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1215514258", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "associativity"},
   CellID->1215514258],
  Cell[141069, 3941, 1403, 44, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1531751179", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "commutativity"},
   CellID->1531751179],
  Cell[142475, 3987, 1532, 48, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:929558765", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "cancellativity"},
   CellID->929558765],
  Cell[144974, 4068, 2225, 54, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:1003575752", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Abelian monoid"},
   CellID->1003575752],
  Cell[151151, 4235, 3526, 109, 145, "FormalTextInputFormula",
   CellTags->{
    "ID:1939013113", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "total ordering"},
   CellID->1939013113],
  Cell[154680, 4346, 1624, 51, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1585090210", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "irreflexive relation"},
   CellID->1585090210],
  Cell[156307, 4399, 2287, 74, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1535164863", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "strict version of ordering"},
   CellID->1535164863],
  Cell[160050, 4522, 1695, 53, 72, "FormalTextInputFormula",
   CellTags->{
    "ID:795999034", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "totally-ordered domain"},
   CellID->795999034],
  Cell[164686, 4665, 1780, 44, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:616702295", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min-max domain"},
   CellID->616702295],
  Cell[166469, 4711, 1888, 54, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:1204372660", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min domain"},
   CellID->1204372660],
  Cell[168360, 4767, 1940, 55, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:1875462394", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max domain"},
   CellID->1875462394],
  Cell[176329, 4994, 3839, 73, 73, "FormalTextInputFormula",
   CellTags->{
    "ID:743220188", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "totally-ordered Abelian monoid"},
   CellID->743220188],
  Cell[180171, 5069, 1577, 48, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1329821571", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "monotonicity"},
   CellID->1329821571],
  Cell[224169, 6514, 755, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2137457113", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is TOM"},
   CellID->2137457113],
  Cell[224927, 6536, 787, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1975341048", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is Abelian monoid"},
   CellID->1975341048],
  Cell[225717, 6558, 787, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:275749992", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is totally ordered"},
   CellID->275749992],
  Cell[226507, 6580, 785, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1904592343", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is min-max domain"},
   CellID->1904592343],
  Cell[228999, 6660, 919, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:589296925", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n is integer"},
   CellID->589296925],
  Cell[229921, 6686, 810, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:54512703", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n is positive"},
   CellID->54512703],
  Cell[238465, 6906, 1019, 31, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1538785027", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "3"},
   CellID->1538785027],
  Cell[244464, 7095, 919, 25, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:968428080", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "4"},
   CellID->968428080],
  Cell[245386, 7122, 846, 24, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:680012314", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "5"},
   CellID->680012314],
  Cell[246235, 7148, 1182, 31, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:655751250", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "6"},
   CellID->655751250],
  Cell[247420, 7181, 1076, 30, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1963081417", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "7"},
   CellID->1963081417],
  Cell[262614, 7647, 898, 25, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1457798778", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "8"},
   CellID->1457798778],
  Cell[263515, 7674, 1264, 37, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1044840887", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "9"},
   CellID->1044840887],
  Cell[264782, 7713, 920, 30, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2017090463", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "10"},
   CellID->2017090463],
  Cell[279429, 8158, 931, 23, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:1308600401", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "irreflexivity of <"},
   CellID->1308600401],
  Cell[280734, 8194, 1009, 24, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:914659821", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of \[LessEqual]"},
   CellID->914659821],
  Cell[281746, 8220, 1422, 39, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:966362506", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<\[Or]="},
   CellID->966362506],
  Cell[283171, 8261, 1285, 34, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1288924674", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[Not]\[LessEqual]"},
   CellID->1288924674],
  Cell[284459, 8297, 1265, 34, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1962569111", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[Not]<"},
   CellID->1962569111],
  Cell[285727, 8333, 1296, 35, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:243913977", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity <\[LessEqual]"},
   CellID->243913977],
  Cell[287026, 8370, 1300, 35, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1623404218", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity \[LessEqual]<"},
   CellID->1623404218],
  Cell[288329, 8407, 1098, 33, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1509384446", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity <<"},
   CellID->1509384446],
  Cell[289430, 8442, 1215, 36, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:972758509", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[NotEqual] in terms of <"},
   CellID->972758509],
  Cell[333513, 9691, 1050, 29, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:786186036", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]+"},
   CellID->786186036],
  Cell[334566, 9722, 1111, 33, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1385901945", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<+"},
   CellID->1385901945],
  Cell[335680, 9757, 1253, 37, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1581648423", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]\[LessEqual]+"},
   CellID->1581648423],
  Cell[336936, 9796, 1211, 37, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1815472507", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]<+"},
   CellID->1815472507],
  Cell[338150, 9835, 1175, 37, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:443432455", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<<+"},
   CellID->443432455],
  Cell[371442, 10744, 1421, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1546357463", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1\[LessEqual]length"},
   CellID->1546357463],
  Cell[372866, 10785, 1184, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:409966668", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "11"},
   CellID->409966668],
  Cell[374053, 10818, 1710, 45, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1768549223", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "2\[LessEqual]length"},
   CellID->1768549223],
  Cell[375766, 10865, 1001, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1399858996", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a is integer"},
   CellID->1399858996],
  Cell[376770, 10895, 982, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1353010357", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a is positive"},
   CellID->1353010357],
  Cell[377755, 10924, 1769, 47, 56, "FormalTextInputFormula",
   CellTags->{
    "ID:763884008", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "12"},
   CellID->763884008],
  Cell[379527, 10973, 1175, 33, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:1442641059", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "13"},
   CellID->1442641059],
  Cell[380705, 11008, 1294, 41, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:215173194", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "14"},
   CellID->215173194],
  Cell[419400, 12112, 1256, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:8303793", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "15"},
   CellID->8303793],
  Cell[420659, 12154, 1265, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1339530720", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "16"},
   CellID->1339530720],
  Cell[421927, 12196, 1378, 43, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:759960416", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[SquareSubsetEqual]\:293a"},
   CellID->759960416],
  Cell[439932, 12724, 1125, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1448857518", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of min"},
   CellID->1448857518],
  Cell[441060, 12756, 1128, 31, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1578657786", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of max"},
   CellID->1578657786],
  Cell[442191, 12789, 913, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1968885164", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of min"},
   CellID->1968885164],
  Cell[443107, 12817, 913, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1151569277", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of max"},
   CellID->1151569277],
  Cell[464763, 13435, 1591, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:232133128", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "17"},
   CellID->232133128],
  Cell[466357, 13481, 1593, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1533503972", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "18"},
   CellID->1533503972],
  Cell[467953, 13527, 1294, 40, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:881325315", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "17'"},
   CellID->881325315],
  Cell[469250, 13569, 1297, 40, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1667609383", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "18'"},
   CellID->1667609383],
  Cell[470550, 13611, 1093, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1257756255", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min is invariant under permutations"},
   CellID->1257756255],
  Cell[471646, 13642, 1093, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2071551441", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max is invariant under permutations"},
   CellID->2071551441],
  Cell[503790, 14546, 1410, 39, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:48233088", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\[LessEqual]"},
   CellID->48233088],
  Cell[505203, 14587, 1466, 39, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1034128987", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]min"},
   CellID->1034128987],
  Cell[506672, 14628, 1369, 38, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1323776552", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\[LessEqual]"},
   CellID->1323776552],
  Cell[508044, 14668, 1368, 37, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1520079070", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]max"},
   CellID->1520079070],
  Cell[509415, 14707, 1229, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1860476328", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min<"},
   CellID->1860476328],
  Cell[510647, 14745, 1230, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1907558186", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<min"},
   CellID->1907558186],
  Cell[511880, 14783, 1230, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:2008421481", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max<"},
   CellID->2008421481],
  Cell[513113, 14821, 1230, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1141242040", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<max"},
   CellID->1141242040],
  Cell[566074, 16250, 1266, 38, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:572537117", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "19"},
   CellID->572537117],
  Cell[567343, 16290, 1892, 51, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2073093776", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a 1"},
   CellID->2073093776],
  Cell[569238, 16343, 1359, 42, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1632846240", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a 2"},
   CellID->1632846240],
  Cell[570600, 16387, 1269, 38, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1146565830", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "20"},
   CellID->1146565830],
  Cell[571872, 16427, 1849, 50, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1172583037", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a 1"},
   CellID->1172583037],
  Cell[573724, 16479, 1359, 42, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2142810348", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a 2"},
   CellID->2142810348],
  Cell[575387, 16532, 1096, 33, 36, "FormalTextInputFormula",
   CellTags->{
    "ID:1460290836", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a"},
   CellID->1460290836],
  Cell[576486, 16567, 1098, 33, 36, "FormalTextInputFormula",
   CellTags->{
    "ID:1931124773", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a"},
   CellID->1931124773],
  Cell[577587, 16602, 990, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:978507648", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a decreases"},
   CellID->978507648],
  Cell[578580, 16633, 994, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2097939500", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a increases"},
   CellID->2097939500],
  Cell[634688, 18183, 1661, 47, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:900399941", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "21"},
   CellID->900399941],
  Cell[636352, 18232, 1523, 44, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1833484473", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 1"},
   CellID->1833484473],
  Cell[637878, 18278, 1271, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1976058899", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 2"},
   CellID->1976058899],
  Cell[639152, 18317, 1272, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:2067845318", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 3"},
   CellID->2067845318],
  Cell[640427, 18356, 1430, 39, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:446229453", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+"},
   CellID->446229453],
  Cell[641860, 18397, 1472, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:437038980", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 1"},
   CellID->437038980],
  Cell[643335, 18442, 1268, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:367539508", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 2"},
   CellID->367539508],
  Cell[644606, 18481, 1272, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1689218301", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 3"},
   CellID->1689218301],
  Cell[645881, 18520, 1380, 38, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:513276857", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+"},
   CellID->513276857],
  Cell[712804, 20275, 718, 15, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1076035631", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->1076035631],
  Cell[713525, 20292, 1777, 49, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:610987882", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExponentVector"},
   CellID->610987882],
  Cell[715305, 20343, 1542, 47, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:468249611", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExpVectorTuple"},
   CellID->468249611],
  Cell[716850, 20392, 1581, 41, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:444732983", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "divisibility"},
   CellID->444732983],
  Cell[718434, 20435, 1154, 30, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1111389324", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "degree"},
   CellID->1111389324],
  Cell[719591, 20467, 1384, 36, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1699354265", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maxdeg"},
   CellID->1699354265],
  Cell[720978, 20505, 1538, 39, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:1614921088", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "lcm"},
   CellID->1614921088],
  Cell[722519, 20546, 2791, 78, 77, "FormalTextInputFormula",
   CellTags->{
    "ID:1784038925", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chain criterion"},
   CellID->1784038925],
  Cell[725313, 20626, 1856, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:312341806", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "width w.r.t. k-th coordinate"},
   CellID->312341806],
  Cell[727172, 20671, 1398, 34, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:282925385", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "width"},
   CellID->282925385],
  Cell[728573, 20707, 1913, 47, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1664335746", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maximum degree of admissible pairs of one object"},
   CellID->1664335746],
  Cell[730489, 20756, 1712, 41, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:640444143", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maximum degree of admissible pairs"},
   CellID->640444143],
  Cell[732204, 20799, 1697, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1923113076", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "corner of rectangular region"},
   CellID->1923113076],
  Cell[734674, 20863, 1447, 40, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1774661003", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "above"},
   CellID->1774661003],
  Cell[736124, 20905, 2246, 63, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:285552196", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "far exterior w.r.t. k-th coordinate"},
   CellID->285552196],
  Cell[738373, 20970, 1894, 50, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:494773971", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "rectangular region"},
   CellID->494773971],
  Cell[750499, 21295, 2667, 81, 112, "FormalTextInputFormula",
   CellTags->{
    "ID:1278822860", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal w.r.t. k-th coordinate"},
   CellID->1278822860],
  Cell[753169, 21378, 1356, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1458824241", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal"},
   CellID->1458824241],
  Cell[776188, 21987, 809, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1559569657", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D=N0"},
   CellID->1559569657],
  Cell[778561, 22070, 296, 8, 30, "GlobalDeclaration",
   CellTags->{
    "ID:977095909", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->977095909],
  Cell[778860, 22080, 975, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:615519889", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min in N0"},
   CellID->615519889],
  Cell[779838, 22108, 977, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1245684717", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max in N0"},
   CellID->1245684717],
  Cell[782089, 22185, 737, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1526008058", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n=2"},
   CellID->1526008058],
  Cell[806526, 22925, 786, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1544493310", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n=3"},
   CellID->1544493310],
  Cell[839862, 23909, 1096, 31, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:18996081", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "22"},
   CellID->18996081],
  Cell[845263, 24078, 868, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:313023917", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExpVectorTuple\:293a"},
   CellID->313023917],
  Cell[846134, 24104, 745, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1697603634", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "23"},
   CellID->1697603634],
  Cell[854797, 24365, 1119, 31, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1313148627", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "antisymmetry of divisibility"},
   CellID->1313148627],
  Cell[855919, 24398, 794, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:785542046", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of divisibility"},
   CellID->785542046],
  Cell[856716, 24420, 1029, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1055765423", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity of divisibility"},
   CellID->1055765423],
  Cell[870776, 24838, 893, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1918380148", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of lcm"},
   CellID->1918380148],
  Cell[871672, 24862, 844, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1065951991", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 1"},
   CellID->1065951991],
  Cell[872519, 24885, 844, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1071215803", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 2"},
   CellID->1071215803],
  Cell[873366, 24908, 1115, 29, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:1028467786", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 3"},
   CellID->1028467786],
  Cell[874484, 24939, 896, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:98270815", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "commutativity of lcm"},
   CellID->98270815],
  Cell[875383, 24964, 904, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:622866956", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "24"},
   CellID->622866956],
  Cell[876290, 24990, 925, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:810936059", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "idempotency of lcm"},
   CellID->810936059],
  Cell[903499, 25794, 956, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1822207066", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of deg"},
   CellID->1822207066],
  Cell[904458, 25821, 972, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:455032720", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "25"},
   CellID->455032720],
  Cell[905433, 25850, 980, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:839534060", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "26"},
   CellID->839534060],
  Cell[919004, 26260, 892, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2123737964", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of maxdeg"},
   CellID->2123737964],
  Cell[919899, 26286, 1088, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:647995479", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "27"},
   CellID->647995479],
  Cell[932058, 26632, 849, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1452963633", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "28"},
   CellID->1452963633],
  Cell[932910, 26657, 932, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1500097429", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "29"},
   CellID->1500097429],
  Cell[934329, 26699, 1084, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:279915377", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCrit"},
   CellID->279915377],
  Cell[935416, 26730, 1291, 35, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1391995733", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "divisibility\[DoubleRightArrow]chainCrit"},
   CellID->1391995733],
  Cell[936710, 26767, 823, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1271079949", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of chainCrit"},
   CellID->1271079949],
  Cell[937536, 26790, 2164, 64, 72, "FormalTextInputFormula",
   CellTags->{
    "ID:488309194", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit w.r.t. singleton"},
   CellID->488309194],
  Cell[939703, 26856, 1261, 37, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1322351588", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a 1"},
   CellID->1322351588],
  Cell[940967, 26895, 1262, 37, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1269486125", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a 2"},
   CellID->1269486125],
  Cell[942232, 26934, 1467, 42, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:941129646", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a"},
   CellID->941129646],
  Cell[943702, 26978, 1164, 33, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1657568255", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a I"},
   CellID->1657568255],
  Cell[993412, 28333, 1538, 43, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:469812703", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "30"},
   CellID->469812703],
  Cell[994953, 28378, 1240, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1053941118", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit holds for some element"},
   CellID->1053941118],
  Cell[1013396, 28870, 1529, 39, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1635347335", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "31"},
   CellID->1635347335],
  Cell[1014928, 28911, 1279, 36, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1901113035", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "32"},
   CellID->1901113035],
  Cell[1016210, 28949, 919, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1760547878", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M (binary)"},
   CellID->1760547878],
  Cell[1017132, 28976, 892, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1896797254", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M (unary)"},
   CellID->1896797254],
  Cell[1018027, 29002, 1131, 31, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1611647919", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "33"},
   CellID->1611647919],
  Cell[1049345, 29860, 1589, 42, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1545946275", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M (unary)"},
   CellID->1545946275],
  Cell[1050937, 29904, 1654, 43, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1018591902", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual] (unary)"},
   CellID->1018591902],
  Cell[1052594, 29949, 1747, 47, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:382103250", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M (binary)"},
   CellID->382103250],
  Cell[1054344, 29998, 1507, 44, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:450457679", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual] (binary)"},
   CellID->450457679],
  Cell[1055854, 30044, 1966, 61, 57, "FormalTextInputFormula",
   CellTags->{
    "ID:643714041", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M+\[LessEqual]"},
   CellID->643714041],
  Cell[1100282, 31190, 1074, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1934562955", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "34"},
   CellID->1934562955],
  Cell[1101359, 31222, 1171, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1993966556", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "35"},
   CellID->1993966556],
  Cell[1102533, 31256, 1095, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1265284403", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "36"},
   CellID->1265284403],
  Cell[1103631, 31288, 1063, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:934408883", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "37"},
   CellID->934408883],
  Cell[1104697, 31320, 1826, 53, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:1342720412", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "38"},
   CellID->1342720412],
  Cell[1106873, 31383, 1204, 33, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:192359742", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "39"},
   CellID->192359742],
  Cell[1108080, 31418, 1677, 49, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:2005322529", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "40"},
   CellID->2005322529],
  Cell[1145936, 32447, 1160, 33, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:984666222", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "41"},
   CellID->984666222],
  Cell[1147099, 32482, 1321, 38, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:1616352426", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "42"},
   CellID->1616352426],
  Cell[1166853, 33005, 1768, 49, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1399613937", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 1"},
   CellID->1399613937],
  Cell[1168624, 33056, 1552, 46, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1210171526", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 2"},
   CellID->1210171526],
  Cell[1170179, 33104, 1104, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1251355025", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 3"},
   CellID->1251355025],
  Cell[1171286, 33140, 1223, 37, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1093640190", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 4"},
   CellID->1093640190],
  Cell[1172512, 33179, 1700, 52, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:606212147", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 5"},
   CellID->606212147],
  Cell[1174215, 33233, 1521, 43, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1628665526", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a"},
   CellID->1628665526],
  Cell[1216960, 34365, 1412, 39, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1351181837", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a I"},
   CellID->1351181837],
  Cell[1218375, 34406, 1446, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:2057269766", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a\[LessEqual]"},
   CellID->2057269766],
  Cell[1219824, 34452, 1427, 42, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:94917498", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a II"},
   CellID->94917498],
  Cell[1221254, 34496, 1596, 51, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1065272829", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a+\[LessEqual]"},
   CellID->1065272829],
  Cell[1254008, 35360, 943, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1824645721", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of W (binary)"},
   CellID->1824645721],
  Cell[1254954, 35387, 893, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1345307293", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of W (unary)"},
   CellID->1345307293],
  Cell[1255850, 35413, 1201, 33, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1185705395", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "43"},
   CellID->1185705395],
  Cell[1274630, 35950, 1549, 46, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:354042430", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]W"},
   CellID->354042430],
  Cell[1276182, 35998, 1538, 45, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:551974343", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\[LessEqual]"},
   CellID->551974343],
  Cell[1295731, 36531, 965, 28, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:894623443", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "44"},
   CellID->894623443],
  Cell[1296699, 36561, 1128, 32, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1995654247", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "45"},
   CellID->1995654247],
  Cell[1297830, 36595, 1240, 34, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1916087968", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "46"},
   CellID->1916087968],
  Cell[1321436, 37240, 1577, 41, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2064198302", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a"},
   CellID->2064198302],
  Cell[1323016, 37283, 948, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2144277606", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a I"},
   CellID->2144277606],
  Cell[1323967, 37314, 1038, 30, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:388422476", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a II"},
   CellID->388422476],
  Cell[1325008, 37346, 1030, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:144808276", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a decreases"},
   CellID->144808276],
  Cell[1349301, 38020, 1074, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1149216477", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of L"},
   CellID->1149216477],
  Cell[1355455, 38202, 322, 9, 21, "GlobalDeclaration",
   CellTags->{
    "ID:504163160", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->504163160],
  Cell[1355819, 38216, 691, 18, 57, "GlobalDeclaration",
   CellTags->{
    "ID:649536122", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->649536122],
  Cell[1359150, 38314, 1011, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:465693853", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "47"},
   CellID->465693853],
  Cell[1360164, 38346, 2424, 78, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:777205621", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "48"},
   CellID->777205621],
  Cell[1362591, 38426, 1737, 51, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:42909295", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "49"},
   CellID->42909295],
  Cell[1364331, 38479, 1942, 62, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:112873953", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "50"},
   CellID->112873953],
  Cell[1410709, 39640, 1579, 46, 66, "FormalTextInputFormula",
   CellTags->{
    "ID:315092621", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "cover"},
   CellID->315092621],
  Cell[1424417, 40031, 1434, 44, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:744007071", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "51"},
   CellID->744007071],
  Cell[1425854, 40077, 1368, 43, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1720661645", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "52"},
   CellID->1720661645],
  Cell[1427225, 40122, 1938, 62, 55, "FormalTextInputFormula",
   CellTags->{
    "ID:1022330126", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[VerticalBar]lcm"},
   CellID->1022330126],
  Cell[1451346, 40772, 1221, 38, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:528222346", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "53"},
   CellID->528222346],
  Cell[1452570, 40812, 1083, 33, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1598627200", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "54"},
   CellID->1598627200],
  Cell[1463450, 41134, 1434, 46, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1537560967", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "55"},
   CellID->1537560967],
  Cell[1464887, 41182, 1431, 46, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:507460317", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "56"},
   CellID->507460317],
  Cell[1485178, 41731, 859, 26, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1442383873", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "57"},
   CellID->1442383873],
  Cell[1486040, 41759, 1384, 33, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:523824248", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (binary) FarExterior 1"},
   CellID->523824248],
  Cell[1487427, 41794, 1326, 34, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1663623856", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (binary) FarExterior 2"},
   CellID->1663623856],
  Cell[1510194, 42406, 1579, 39, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1829996465", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (unary) FarExterior 1"},
   CellID->1829996465],
  Cell[1511776, 42447, 1598, 40, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:842592091", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (unary) FarExterior 2"},
   CellID->842592091],
  Cell[1513377, 42489, 1269, 38, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1051080023", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a FarExterior"},
   CellID->1051080023],
  Cell[1535244, 43091, 997, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:631716597", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "58"},
   CellID->631716597],
  Cell[1536244, 43122, 980, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1988188117", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "59"},
   CellID->1988188117],
  Cell[1549691, 43508, 399, 12, 41, "GlobalDeclaration",
   CellTags->{
    "ID:590870546", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->590870546],
  Cell[1550093, 43522, 1281, 38, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:501569980", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "60"},
   CellID->501569980],
  Cell[1551377, 43562, 1614, 46, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:914554779", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "61"},
   CellID->914554779],
  Cell[1552994, 43610, 1748, 50, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1336064727", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "62"},
   CellID->1336064727],
  Cell[1554745, 43662, 1710, 51, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1651077201", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "L+\[LessEqual]"},
   CellID->1651077201],
  Cell[1595418, 44724, 1054, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:53741691", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "63"},
   CellID->53741691],
  Cell[1596475, 44756, 1014, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1045358942", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "64"},
   CellID->1045358942],
  Cell[1597492, 44788, 1068, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1132937314", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "65"},
   CellID->1132937314],
  Cell[1598563, 44822, 1066, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:152254973", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "66"},
   CellID->152254973],
  Cell[1630384, 45669, 1000, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:836877533", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 1"},
   CellID->836877533],
  Cell[1631387, 45696, 1071, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1226632396", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 2"},
   CellID->1226632396],
  Cell[1632461, 45726, 1264, 32, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:457131364", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 3"},
   CellID->457131364],
  Cell[1633728, 45760, 1020, 27, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1490326930", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal"},
   CellID->1490326930],
  Cell[1634751, 45789, 2256, 68, 112, "FormalTextInputFormula",
   CellTags->{
    "ID:749802963", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal w.r.t. first coordinate"},
   CellID->749802963],
  Cell[1637256, 45867, 1026, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:505706706", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "67"},
   CellID->505706706],
  Cell[1638285, 45898, 2989, 83, 71, "FormalTextInputFormula",
   CellTags->{
    "ID:1262287499", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "68"},
   CellID->1262287499],
  Cell[1700637, 47472, 1576, 50, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1909233179", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "69"},
   CellID->1909233179],
  Cell[1702216, 47524, 1682, 54, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:2125747942", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "70"},
   CellID->2125747942],
  Cell[1703901, 47580, 1776, 55, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1211430036", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "71"},
   CellID->1211430036],
  Cell[1706305, 47655, 900, 25, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:2061027934", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "72"},
   CellID->2061027934],
  Cell[1707208, 47682, 778, 23, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1653917255", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "73"},
   CellID->1653917255],
  Cell[1762262, 49041, 1132, 27, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1690170957", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 1"},
   CellID->1690170957],
  Cell[1763397, 49070, 926, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:271680825", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 2"},
   CellID->271680825],
  Cell[1764326, 49096, 3340, 100, 111, "FormalTextInputFormula",
   CellTags->{
    "ID:1572461042", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 3"},
   CellID->1572461042],
  Cell[1767669, 49198, 1385, 41, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:2046865956", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 4"},
   CellID->2046865956],
  Cell[1769057, 49241, 1244, 35, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:917844397", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 5"},
   CellID->917844397],
  Cell[1770304, 49278, 2181, 49, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:810037327", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate"},
   CellID->810037327],
  Cell[1772488, 49329, 1642, 50, 71, "FormalTextInputFormula",
   CellTags->{
    "ID:2111681775", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate I"},
   CellID->2111681775],
  Cell[1836634, 50957, 2387, 65, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:2973275", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar] binary"},
   CellID->2973275],
  Cell[1839024, 51024, 1541, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1258784171", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar] unary"},
   CellID->1258784171],
  Cell[1842616, 51136, 1393, 35, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1549582561", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit holds for distinct indices"},
   CellID->1549582561],
  Cell[1844012, 51173, 1470, 41, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1966722449", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "74"},
   CellID->1966722449],
  Cell[1845485, 51216, 1405, 40, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:164254989", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "75"},
   CellID->164254989],
  Cell[1846893, 51258, 1018, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:290364666", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M\[RawStar] (binary)"},
   CellID->290364666],
  Cell[1847914, 51287, 974, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1063863080", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M\[RawStar] (unary)"},
   CellID->1063863080],
  Cell[1890170, 52400, 1658, 42, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1680577066", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M\[RawStar] (unary)"},
   CellID->1680577066],
  Cell[1891831, 52444, 1729, 44, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1866021817", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual] (unary)"},
   CellID->1866021817],
  Cell[1893563, 52490, 2109, 58, 54, "FormalTextInputFormula",
   CellTags->{
    "ID:1272742036", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M\[RawStar] (binary)"},
   CellID->1272742036],
  Cell[1895675, 52550, 1913, 55, 54, "FormalTextInputFormula",
   CellTags->{
    "ID:462890005", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual] (binary)"},
   CellID->462890005],
  Cell[1936278, 53591, 873, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:929561290", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual]M\[RawStar]"},
   CellID->929561290],
  Cell[1937154, 53617, 872, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:795795393", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual]M"},
   CellID->795795393],
  Cell[1938029, 53643, 1234, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:872843574", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M=M\[RawStar]"},
   CellID->872843574],
  Cell[1956073, 54152, 1033, 29, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:642765392", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "76"},
   CellID->642765392],
  Cell[1957109, 54183, 932, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:592779258", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maxdeg\[LessEqual]M"},
   CellID->592779258],
  Cell[1968850, 54524, 890, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:691015215", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "77"},
   CellID->691015215],
  Cell[1969743, 54550, 1254, 33, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:402511110", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "78"},
   CellID->402511110],
  Cell[1983724, 54935, 973, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:341822661", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual]L"},
   CellID->341822661],
  Cell[1994463, 55247, 1421, 40, 56, "FormalTextInputFormula",
   CellTags->{
    "ID:245124153", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "79"},
   CellID->245124153],
  Cell[1995887, 55289, 1437, 47, 63, "FormalTextInputFormula",
   CellTags->{
    "ID:952627996", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "80"},
   CellID->952627996],
  Cell[1998901, 55390, 1453, 31, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2013760956", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "81"},
   CellID->2013760956],
  Cell[2000357, 55423, 964, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1011330545", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "82"},
   CellID->1011330545],
  Cell[2001324, 55453, 1278, 37, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:718586393", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "83"},
   CellID->718586393],
  Cell[2079705, 57400, 1342, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:696507084", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "84"},
   CellID->696507084],
  Cell[2081050, 57442, 895, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:626970451", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "85"},
   CellID->626970451],
  Cell[2081948, 57469, 1525, 36, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:789654176", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "86"},
   CellID->789654176],
  Cell[2119489, 58422, 1706, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1437590439", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "87"},
   CellID->1437590439],
  Cell[2121198, 58472, 1572, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:477242065", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "88"},
   CellID->477242065],
  Cell[2140991, 59012, 793, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1041359623", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "89"},
   CellID->1041359623],
  Cell[2141787, 59036, 797, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2130272336", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "90"},
   CellID->2130272336],
  Cell[2142587, 59061, 872, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1584641667", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "91"},
   CellID->1584641667],
  Cell[2143462, 59087, 1139, 32, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1175076282", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "92"},
   CellID->1175076282],
  Cell[2144856, 59130, 1215, 33, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:333980150", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "93"},
   CellID->333980150],
  Cell[2146074, 59165, 863, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1370308823", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "94"},
   CellID->1370308823],
  Cell[2208836, 60784, 1216, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1434774204", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "95"},
   CellID->1434774204],
  Cell[2210055, 60822, 976, 30, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1867780171", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "96"},
   CellID->1867780171],
  Cell[2211307, 60863, 1048, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:786584693", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "97"},
   CellID->786584693],
  Cell[2212358, 60896, 1022, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:183897269", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "98"},
   CellID->183897269],
  Cell[2213383, 60929, 1403, 39, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1246401530", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "99"},
   CellID->1246401530],
  Cell[2268552, 62338, 1549, 43, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2102726322", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "invariant"},
   CellID->2102726322],
  Cell[2290623, 62988, 1120, 32, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:724101256", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "100"},
   CellID->724101256],
  Cell[2297499, 63192, 1140, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1990359651", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "bound"},
   CellID->1990359651],
  Cell[2310387, 63573, 987, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1891911403", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "main theorem"},
   CellID->1891911403]},
 "length is natural"->{
  Cell[103698, 2839, 1563, 37, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1969097229", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length is natural"},
   CellID->1969097229]},
 "ID:1508742051"->{
  Cell[105264, 2878, 2068, 66, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1508742051", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "tuple equality"},
   CellID->1508742051]},
 "tuple equality"->{
  Cell[105264, 2878, 2068, 66, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1508742051", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "tuple equality"},
   CellID->1508742051]},
 "ID:1264372773"->{
  Cell[107335, 2946, 1125, 31, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1264372773", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isTuple\:293a"},
   CellID->1264372773]},
 "isTuple\:293a"->{
  Cell[107335, 2946, 1125, 31, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1264372773", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isTuple\:293a"},
   CellID->1264372773]},
 "ID:2086441677"->{
  Cell[108463, 2979, 1139, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2086441677", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a"},
   CellID->2086441677]},
 "length\:293a"->{
  Cell[108463, 2979, 1139, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2086441677", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a"},
   CellID->2086441677]},
 "ID:585712945"->{
  Cell[109605, 3012, 1431, 43, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:585712945", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "element\:293a 1"},
   CellID->585712945]},
 "element\:293a 1"->{
  Cell[109605, 3012, 1431, 43, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:585712945", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "element\:293a 1"},
   CellID->585712945]},
 "ID:1075236060"->{
  Cell[111039, 3057, 1123, 36, 24, "FormalTextInputFormula",
   CellTags->{
    "ID:1075236060", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "element\:293a 2"},
   CellID->1075236060]},
 "element\:293a 2"->{
  Cell[111039, 3057, 1123, 36, 24, "FormalTextInputFormula",
   CellTags->{
    "ID:1075236060", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "element\:293a 2"},
   CellID->1075236060]},
 "ID:1405024037"->{
  Cell[115173, 3180, 446, 11, 32, "GlobalDeclaration",
   CellTags->{
    "ID:1405024037", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->1405024037]},
 "ID:379346204"->{
  Cell[115622, 3193, 1703, 48, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:379346204", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "domain tuple"},
   CellID->379346204]},
 "domain tuple"->{
  Cell[115622, 3193, 1703, 48, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:379346204", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "domain tuple"},
   CellID->379346204]},
 "sect::DefinitionOC"->{
  Cell[117494, 3253, 160, 3, 22, "Subsubsection",
   CellTags->"sect::DefinitionOC",
   CellID->157546407]},
 "ID:1473378407"->{
  Cell[118776, 3290, 791, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1473378407", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "other coordinate 1"},
   CellID->1473378407]},
 "other coordinate 1"->{
  Cell[118776, 3290, 791, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1473378407", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "other coordinate 1"},
   CellID->1473378407]},
 "ID:406630295"->{
  Cell[119570, 3313, 788, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:406630295", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "other coordinate 2"},
   CellID->406630295]},
 "other coordinate 2"->{
  Cell[119570, 3313, 788, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:406630295", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "other coordinate 2"},
   CellID->406630295]},
 "sect::DefinitionZ"->{
  Cell[121454, 3367, 167, 3, 22, "Subsubsection",
   CellTags->"sect::DefinitionZ",
   CellID->1036621313]},
 "ID:1091916109"->{
  Cell[124938, 3463, 950, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1091916109", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Z is TOM"},
   CellID->1091916109]},
 "Z is TOM"->{
  Cell[124938, 3463, 950, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1091916109", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Z is TOM"},
   CellID->1091916109]},
 "ID:1845867532"->{
  Cell[125891, 3489, 841, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1845867532", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Z is totally ordered"},
   CellID->1845867532]},
 "Z is totally ordered"->{
  Cell[125891, 3489, 841, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1845867532", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Z is totally ordered"},
   CellID->1845867532]},
 "ID:492790093"->{
  Cell[126735, 3513, 1282, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:492790093", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "some integers"},
   CellID->492790093]},
 "some integers"->{
  Cell[126735, 3513, 1282, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:492790093", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "some integers"},
   CellID->492790093]},
 "ID:1907318616"->{
  Cell[128020, 3547, 844, 23, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1907318616", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1\[LessEqual]2"},
   CellID->1907318616]},
 "1\[LessEqual]2"->{
  Cell[128020, 3547, 844, 23, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1907318616", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1\[LessEqual]2"},
   CellID->1907318616]},
 "ID:1708561859"->{
  Cell[129366, 3584, 1402, 41, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1708561859", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1"},
   CellID->1708561859]},
 "1"->{
  Cell[129366, 3584, 1402, 41, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1708561859", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1"},
   CellID->1708561859]},
 "ID:49894867"->{
  Cell[130771, 3627, 1152, 32, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:49894867", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "2"},
   CellID->49894867]},
 "2"->{
  Cell[130771, 3627, 1152, 32, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:49894867", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "2"},
   CellID->49894867]},
 "sect::DefinitionCategories"->{
  Cell[132187, 3674, 159, 3, 30, "Subsection",
   CellTags->"sect::DefinitionCategories",
   CellID->1469832235]},
 "sect::DefinitionAbelianMonoid"->{
  Cell[134051, 3723, 266, 5, 28, "Subsubsection",
   CellTags->"sect::DefinitionAbelianMonoid",
   CellID->1828699765]},
 "ID:349901806"->{
  Cell[135727, 3769, 1253, 39, 45, "FormalTextInputFormula",
   CellTags->{
    "ID:349901806", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "closure"},
   CellID->349901806]},
 "closure"->{
  Cell[135727, 3769, 1253, 39, 45, "FormalTextInputFormula",
   CellTags->{
    "ID:349901806", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "closure"},
   CellID->349901806]},
 "ID:1476372605"->{
  Cell[136983, 3810, 1145, 33, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1476372605", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "0 in domain"},
   CellID->1476372605]},
 "0 in domain"->{
  Cell[136983, 3810, 1145, 33, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1476372605", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "0 in domain"},
   CellID->1476372605]},
 "ID:1359514637"->{
  Cell[138131, 3845, 1343, 42, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1359514637", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "0 is neutral"},
   CellID->1359514637]},
 "0 is neutral"->{
  Cell[138131, 3845, 1343, 42, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1359514637", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "0 is neutral"},
   CellID->1359514637]},
 "ID:1215514258"->{
  Cell[139477, 3889, 1589, 50, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1215514258", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "associativity"},
   CellID->1215514258]},
 "associativity"->{
  Cell[139477, 3889, 1589, 50, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1215514258", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "associativity"},
   CellID->1215514258]},
 "ID:1531751179"->{
  Cell[141069, 3941, 1403, 44, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1531751179", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "commutativity"},
   CellID->1531751179]},
 "commutativity"->{
  Cell[141069, 3941, 1403, 44, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1531751179", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "commutativity"},
   CellID->1531751179]},
 "ID:929558765"->{
  Cell[142475, 3987, 1532, 48, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:929558765", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "cancellativity"},
   CellID->929558765]},
 "cancellativity"->{
  Cell[142475, 3987, 1532, 48, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:929558765", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "cancellativity"},
   CellID->929558765]},
 "ID:1003575752"->{
  Cell[144974, 4068, 2225, 54, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:1003575752", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Abelian monoid"},
   CellID->1003575752]},
 "Abelian monoid"->{
  Cell[144974, 4068, 2225, 54, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:1003575752", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Abelian monoid"},
   CellID->1003575752]},
 "sect::DefinitionTotallyOrderedDomain"->{
  Cell[149256, 4173, 184, 3, 22, "Subsubsection",
   CellTags->"sect::DefinitionTotallyOrderedDomain",
   CellID->898592553]},
 "ID:1939013113"->{
  Cell[151151, 4235, 3526, 109, 145, "FormalTextInputFormula",
   CellTags->{
    "ID:1939013113", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "total ordering"},
   CellID->1939013113]},
 "total ordering"->{
  Cell[151151, 4235, 3526, 109, 145, "FormalTextInputFormula",
   CellTags->{
    "ID:1939013113", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "total ordering"},
   CellID->1939013113]},
 "ID:1585090210"->{
  Cell[154680, 4346, 1624, 51, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1585090210", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "irreflexive relation"},
   CellID->1585090210]},
 "irreflexive relation"->{
  Cell[154680, 4346, 1624, 51, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1585090210", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "irreflexive relation"},
   CellID->1585090210]},
 "ID:1535164863"->{
  Cell[156307, 4399, 2287, 74, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1535164863", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "strict version of ordering"},
   CellID->1535164863]},
 "strict version of ordering"->{
  Cell[156307, 4399, 2287, 74, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1535164863", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "strict version of ordering"},
   CellID->1535164863]},
 "ID:795999034"->{
  Cell[160050, 4522, 1695, 53, 72, "FormalTextInputFormula",
   CellTags->{
    "ID:795999034", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "totally-ordered domain"},
   CellID->795999034]},
 "totally-ordered domain"->{
  Cell[160050, 4522, 1695, 53, 72, "FormalTextInputFormula",
   CellTags->{
    "ID:795999034", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "totally-ordered domain"},
   CellID->795999034]},
 "sect::DefinitionminmaxDomain"->{
  Cell[161914, 4587, 169, 3, 22, "Subsubsection",
   CellTags->"sect::DefinitionminmaxDomain",
   CellID->2071321112]},
 "ID:616702295"->{
  Cell[164686, 4665, 1780, 44, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:616702295", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min-max domain"},
   CellID->616702295]},
 "min-max domain"->{
  Cell[164686, 4665, 1780, 44, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:616702295", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min-max domain"},
   CellID->616702295]},
 "ID:1204372660"->{
  Cell[166469, 4711, 1888, 54, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:1204372660", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min domain"},
   CellID->1204372660]},
 "min domain"->{
  Cell[166469, 4711, 1888, 54, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:1204372660", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min domain"},
   CellID->1204372660]},
 "ID:1875462394"->{
  Cell[168360, 4767, 1940, 55, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:1875462394", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max domain"},
   CellID->1875462394]},
 "max domain"->{
  Cell[168360, 4767, 1940, 55, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:1875462394", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max domain"},
   CellID->1875462394]},
 "sect::DefinitionTOM"->{
  Cell[173683, 4919, 224, 4, 22, "Subsubsection",
   CellTags->"sect::DefinitionTOM",
   CellID->991805917]},
 "ID:743220188"->{
  Cell[176329, 4994, 3839, 73, 73, "FormalTextInputFormula",
   CellTags->{
    "ID:743220188", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "totally-ordered Abelian monoid"},
   CellID->743220188]},
 "totally-ordered Abelian monoid"->{
  Cell[176329, 4994, 3839, 73, 73, "FormalTextInputFormula",
   CellTags->{
    "ID:743220188", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "totally-ordered Abelian monoid"},
   CellID->743220188]},
 "ID:1329821571"->{
  Cell[180171, 5069, 1577, 48, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1329821571", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "monotonicity"},
   CellID->1329821571]},
 "monotonicity"->{
  Cell[180171, 5069, 1577, 48, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1329821571", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "monotonicity"},
   CellID->1329821571]},
 "ID:2137457113"->{
  Cell[224169, 6514, 755, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2137457113", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is TOM"},
   CellID->2137457113]},
 "D is TOM"->{
  Cell[224169, 6514, 755, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2137457113", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is TOM"},
   CellID->2137457113]},
 "ID:1975341048"->{
  Cell[224927, 6536, 787, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1975341048", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is Abelian monoid"},
   CellID->1975341048]},
 "D is Abelian monoid"->{
  Cell[224927, 6536, 787, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1975341048", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is Abelian monoid"},
   CellID->1975341048]},
 "ID:275749992"->{
  Cell[225717, 6558, 787, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:275749992", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is totally ordered"},
   CellID->275749992]},
 "D is totally ordered"->{
  Cell[225717, 6558, 787, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:275749992", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is totally ordered"},
   CellID->275749992]},
 "ID:1904592343"->{
  Cell[226507, 6580, 785, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1904592343", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is min-max domain"},
   CellID->1904592343]},
 "D is min-max domain"->{
  Cell[226507, 6580, 785, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1904592343", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is min-max domain"},
   CellID->1904592343]},
 "ID:589296925"->{
  Cell[228999, 6660, 919, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:589296925", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n is integer"},
   CellID->589296925]},
 "n is integer"->{
  Cell[228999, 6660, 919, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:589296925", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n is integer"},
   CellID->589296925]},
 "ID:54512703"->{
  Cell[229921, 6686, 810, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:54512703", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n is positive"},
   CellID->54512703]},
 "n is positive"->{
  Cell[229921, 6686, 810, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:54512703", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n is positive"},
   CellID->54512703]},
 "sect::LemmataBasic"->{
  Cell[230912, 6721, 305, 5, 43, "Section",
   CellTags->"sect::LemmataBasic",
   CellID->1171017268]},
 "ID:1538785027"->{
  Cell[238465, 6906, 1019, 31, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1538785027", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "3"},
   CellID->1538785027]},
 "3"->{
  Cell[238465, 6906, 1019, 31, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1538785027", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "3"},
   CellID->1538785027]},
 "Proof|ID:1538785027"->{
  Cell[239641, 6948, 2604, 79, 29, "ProofInfo",
   CellTags->{"Proof|ID:1538785027", "Proof|ID:1538785027-1"},
   CellID->1800356526]},
 "Proof|ID:1538785027-1"->{
  Cell[239641, 6948, 2604, 79, 29, "ProofInfo",
   CellTags->{"Proof|ID:1538785027", "Proof|ID:1538785027-1"},
   CellID->1800356526]},
 "ID:968428080"->{
  Cell[244464, 7095, 919, 25, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:968428080", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "4"},
   CellID->968428080]},
 "4"->{
  Cell[244464, 7095, 919, 25, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:968428080", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "4"},
   CellID->968428080]},
 "ID:680012314"->{
  Cell[245386, 7122, 846, 24, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:680012314", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "5"},
   CellID->680012314]},
 "5"->{
  Cell[245386, 7122, 846, 24, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:680012314", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "5"},
   CellID->680012314]},
 "ID:655751250"->{
  Cell[246235, 7148, 1182, 31, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:655751250", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "6"},
   CellID->655751250]},
 "6"->{
  Cell[246235, 7148, 1182, 31, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:655751250", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "6"},
   CellID->655751250]},
 "ID:1963081417"->{
  Cell[247420, 7181, 1076, 30, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1963081417", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "7"},
   CellID->1963081417]},
 "7"->{
  Cell[247420, 7181, 1076, 30, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1963081417", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "7"},
   CellID->1963081417]},
 "Proof|ID:968428080"->{
  Cell[248654, 7222, 1963, 60, 29, "ProofInfo",
   CellTags->{"Proof|ID:968428080", "Proof|ID:968428080-1"},
   CellID->665537877]},
 "Proof|ID:968428080-1"->{
  Cell[248654, 7222, 1963, 60, 29, "ProofInfo",
   CellTags->{"Proof|ID:968428080", "Proof|ID:968428080-1"},
   CellID->665537877]},
 "Proof|ID:680012314"->{
  Cell[251138, 7301, 1963, 60, 27, "ProofInfo",
   CellTags->{"Proof|ID:680012314", "Proof|ID:680012314-1"},
   CellID->134690164]},
 "Proof|ID:680012314-1"->{
  Cell[251138, 7301, 1963, 60, 27, "ProofInfo",
   CellTags->{"Proof|ID:680012314", "Proof|ID:680012314-1"},
   CellID->134690164]},
 "Proof|ID:655751250"->{
  Cell[253621, 7380, 3047, 84, 27, "ProofInfo",
   CellTags->{"Proof|ID:655751250", "Proof|ID:655751250-1"},
   CellID->234671713]},
 "Proof|ID:655751250-1"->{
  Cell[253621, 7380, 3047, 84, 27, "ProofInfo",
   CellTags->{"Proof|ID:655751250", "Proof|ID:655751250-1"},
   CellID->234671713]},
 "Proof|ID:1963081417"->{
  Cell[257188, 7483, 3196, 90, 27, "ProofInfo",
   CellTags->{"Proof|ID:1963081417", "Proof|ID:1963081417-1"},
   CellID->1541015864]},
 "Proof|ID:1963081417-1"->{
  Cell[257188, 7483, 3196, 90, 27, "ProofInfo",
   CellTags->{"Proof|ID:1963081417", "Proof|ID:1963081417-1"},
   CellID->1541015864]},
 "ID:1457798778"->{
  Cell[262614, 7647, 898, 25, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1457798778", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "8"},
   CellID->1457798778]},
 "8"->{
  Cell[262614, 7647, 898, 25, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1457798778", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "8"},
   CellID->1457798778]},
 "ID:1044840887"->{
  Cell[263515, 7674, 1264, 37, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1044840887", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "9"},
   CellID->1044840887]},
 "9"->{
  Cell[263515, 7674, 1264, 37, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1044840887", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "9"},
   CellID->1044840887]},
 "ID:2017090463"->{
  Cell[264782, 7713, 920, 30, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2017090463", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "10"},
   CellID->2017090463]},
 "10"->{
  Cell[264782, 7713, 920, 30, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2017090463", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "10"},
   CellID->2017090463]},
 "Proof|ID:1457798778"->{
  Cell[265859, 7754, 2868, 85, 29, "ProofInfo",
   CellTags->{"Proof|ID:1457798778", "Proof|ID:1457798778-1"},
   CellID->122079454]},
 "Proof|ID:1457798778-1"->{
  Cell[265859, 7754, 2868, 85, 29, "ProofInfo",
   CellTags->{"Proof|ID:1457798778", "Proof|ID:1457798778-1"},
   CellID->122079454]},
 "Proof|ID:1044840887"->{
  Cell[269248, 7858, 3408, 99, 27, "ProofInfo",
   CellTags->{"Proof|ID:1044840887", "Proof|ID:1044840887-1"},
   CellID->1585772568]},
 "Proof|ID:1044840887-1"->{
  Cell[269248, 7858, 3408, 99, 27, "ProofInfo",
   CellTags->{"Proof|ID:1044840887", "Proof|ID:1044840887-1"},
   CellID->1585772568]},
 "Proof|ID:2017090463"->{
  Cell[273177, 7976, 3004, 91, 27, "ProofInfo",
   CellTags->{"Proof|ID:2017090463", "Proof|ID:2017090463-1"},
   CellID->1023157275]},
 "Proof|ID:2017090463-1"->{
  Cell[273177, 7976, 3004, 91, 27, "ProofInfo",
   CellTags->{"Proof|ID:2017090463", "Proof|ID:2017090463-1"},
   CellID->1023157275]},
 "sect::LemmataOrderings"->{
  Cell[276715, 8087, 275, 5, 30, "Subsection",
   CellTags->"sect::LemmataOrderings",
   CellID->1103427178]},
 "ID:1308600401"->{
  Cell[279429, 8158, 931, 23, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:1308600401", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "irreflexivity of <"},
   CellID->1308600401]},
 "irreflexivity of <"->{
  Cell[279429, 8158, 931, 23, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:1308600401", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "irreflexivity of <"},
   CellID->1308600401]},
 "ID:914659821"->{
  Cell[280734, 8194, 1009, 24, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:914659821", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of \[LessEqual]"},
   CellID->914659821]},
 "reflexivity of \[LessEqual]"->{
  Cell[280734, 8194, 1009, 24, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:914659821", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of \[LessEqual]"},
   CellID->914659821]},
 "ID:966362506"->{
  Cell[281746, 8220, 1422, 39, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:966362506", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<\[Or]="},
   CellID->966362506]},
 "<\[Or]="->{
  Cell[281746, 8220, 1422, 39, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:966362506", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<\[Or]="},
   CellID->966362506]},
 "ID:1288924674"->{
  Cell[283171, 8261, 1285, 34, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1288924674", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[Not]\[LessEqual]"},
   CellID->1288924674]},
 "\[Not]\[LessEqual]"->{
  Cell[283171, 8261, 1285, 34, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1288924674", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[Not]\[LessEqual]"},
   CellID->1288924674]},
 "ID:1962569111"->{
  Cell[284459, 8297, 1265, 34, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1962569111", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[Not]<"},
   CellID->1962569111]},
 "\[Not]<"->{
  Cell[284459, 8297, 1265, 34, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1962569111", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[Not]<"},
   CellID->1962569111]},
 "ID:243913977"->{
  Cell[285727, 8333, 1296, 35, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:243913977", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity <\[LessEqual]"},
   CellID->243913977]},
 "transitivity <\[LessEqual]"->{
  Cell[285727, 8333, 1296, 35, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:243913977", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity <\[LessEqual]"},
   CellID->243913977]},
 "ID:1623404218"->{
  Cell[287026, 8370, 1300, 35, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1623404218", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity \[LessEqual]<"},
   CellID->1623404218]},
 "transitivity \[LessEqual]<"->{
  Cell[287026, 8370, 1300, 35, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1623404218", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity \[LessEqual]<"},
   CellID->1623404218]},
 "ID:1509384446"->{
  Cell[288329, 8407, 1098, 33, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1509384446", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity <<"},
   CellID->1509384446]},
 "transitivity <<"->{
  Cell[288329, 8407, 1098, 33, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1509384446", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity <<"},
   CellID->1509384446]},
 "ID:972758509"->{
  Cell[289430, 8442, 1215, 36, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:972758509", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[NotEqual] in terms of <"},
   CellID->972758509]},
 "\[NotEqual] in terms of <"->{
  Cell[289430, 8442, 1215, 36, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:972758509", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[NotEqual] in terms of <"},
   CellID->972758509]},
 "Proof|ID:1308600401"->{
  Cell[290802, 8489, 2289, 69, 29, "ProofInfo",
   CellTags->{"Proof|ID:1308600401", "Proof|ID:1308600401-1"},
   CellID->21123431]},
 "Proof|ID:1308600401-1"->{
  Cell[290802, 8489, 2289, 69, 29, "ProofInfo",
   CellTags->{"Proof|ID:1308600401", "Proof|ID:1308600401-1"},
   CellID->21123431]},
 "Proof|ID:914659821"->{
  Cell[293613, 8577, 2858, 82, 27, "ProofInfo",
   CellTags->{"Proof|ID:914659821", "Proof|ID:914659821-1"},
   CellID->388106339]},
 "Proof|ID:914659821-1"->{
  Cell[293613, 8577, 2858, 82, 27, "ProofInfo",
   CellTags->{"Proof|ID:914659821", "Proof|ID:914659821-1"},
   CellID->388106339]},
 "Proof|ID:966362506"->{
  Cell[296991, 8678, 4234, 114, 27, "ProofInfo",
   CellTags->{"Proof|ID:966362506", "Proof|ID:966362506-1"},
   CellID->831539329]},
 "Proof|ID:966362506-1"->{
  Cell[296991, 8678, 4234, 114, 27, "ProofInfo",
   CellTags->{"Proof|ID:966362506", "Proof|ID:966362506-1"},
   CellID->831539329]},
 "Proof|ID:1288924674"->{
  Cell[301810, 8812, 4244, 115, 27, "ProofInfo",
   CellTags->{"Proof|ID:1288924674", "Proof|ID:1288924674-1"},
   CellID->1029351046]},
 "Proof|ID:1288924674-1"->{
  Cell[301810, 8812, 4244, 115, 27, "ProofInfo",
   CellTags->{"Proof|ID:1288924674", "Proof|ID:1288924674-1"},
   CellID->1029351046]},
 "Proof|ID:1962569111"->{
  Cell[306576, 8946, 4222, 115, 27, "ProofInfo",
   CellTags->{"Proof|ID:1962569111", "Proof|ID:1962569111-1"},
   CellID->1777374244]},
 "Proof|ID:1962569111-1"->{
  Cell[306576, 8946, 4222, 115, 27, "ProofInfo",
   CellTags->{"Proof|ID:1962569111", "Proof|ID:1962569111-1"},
   CellID->1777374244]},
 "Proof|ID:243913977"->{
  Cell[311320, 9080, 4669, 122, 27, "ProofInfo",
   CellTags->{"Proof|ID:243913977", "Proof|ID:243913977-1"},
   CellID->1987732976]},
 "Proof|ID:243913977-1"->{
  Cell[311320, 9080, 4669, 122, 27, "ProofInfo",
   CellTags->{"Proof|ID:243913977", "Proof|ID:243913977-1"},
   CellID->1987732976]},
 "Proof|ID:1623404218"->{
  Cell[316510, 9221, 4678, 123, 27, "ProofInfo",
   CellTags->{"Proof|ID:1623404218", "Proof|ID:1623404218-1"},
   CellID->1964215912]},
 "Proof|ID:1623404218-1"->{
  Cell[316510, 9221, 4678, 123, 27, "ProofInfo",
   CellTags->{"Proof|ID:1623404218", "Proof|ID:1623404218-1"},
   CellID->1964215912]},
 "Proof|ID:1509384446"->{
  Cell[321710, 9363, 4633, 123, 27, "ProofInfo",
   CellTags->{"Proof|ID:1509384446", "Proof|ID:1509384446-1"},
   CellID->295324759]},
 "Proof|ID:1509384446-1"->{
  Cell[321710, 9363, 4633, 123, 27, "ProofInfo",
   CellTags->{"Proof|ID:1509384446", "Proof|ID:1509384446-1"},
   CellID->295324759]},
 "Proof|ID:972758509"->{
  Cell[326864, 9505, 4234, 110, 27, "ProofInfo",
   CellTags->{"Proof|ID:972758509", "Proof|ID:972758509-1"},
   CellID->1361125711]},
 "Proof|ID:972758509-1"->{
  Cell[326864, 9505, 4234, 110, 27, "ProofInfo",
   CellTags->{"Proof|ID:972758509", "Proof|ID:972758509-1"},
   CellID->1361125711]},
 "ID:786186036"->{
  Cell[333513, 9691, 1050, 29, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:786186036", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]+"},
   CellID->786186036]},
 "\[LessEqual]+"->{
  Cell[333513, 9691, 1050, 29, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:786186036", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]+"},
   CellID->786186036]},
 "ID:1385901945"->{
  Cell[334566, 9722, 1111, 33, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1385901945", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<+"},
   CellID->1385901945]},
 "<+"->{
  Cell[334566, 9722, 1111, 33, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1385901945", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<+"},
   CellID->1385901945]},
 "ID:1581648423"->{
  Cell[335680, 9757, 1253, 37, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1581648423", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]\[LessEqual]+"},
   CellID->1581648423]},
 "\[LessEqual]\[LessEqual]+"->{
  Cell[335680, 9757, 1253, 37, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1581648423", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]\[LessEqual]+"},
   CellID->1581648423]},
 "ID:1815472507"->{
  Cell[336936, 9796, 1211, 37, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1815472507", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]<+"},
   CellID->1815472507]},
 "\[LessEqual]<+"->{
  Cell[336936, 9796, 1211, 37, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1815472507", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]<+"},
   CellID->1815472507]},
 "ID:443432455"->{
  Cell[338150, 9835, 1175, 37, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:443432455", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<<+"},
   CellID->443432455]},
 "<<+"->{
  Cell[338150, 9835, 1175, 37, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:443432455", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<<+"},
   CellID->443432455]},
 "Proof|ID:786186036"->{
  Cell[339483, 9883, 4851, 128, 29, "ProofInfo",
   CellTags->{"Proof|ID:786186036", "Proof|ID:786186036-1"},
   CellID->1923006109]},
 "Proof|ID:786186036-1"->{
  Cell[339483, 9883, 4851, 128, 29, "ProofInfo",
   CellTags->{"Proof|ID:786186036", "Proof|ID:786186036-1"},
   CellID->1923006109]},
 "Proof|ID:1385901945"->{
  Cell[344855, 10030, 4793, 129, 27, "ProofInfo",
   CellTags->{"Proof|ID:1385901945", "Proof|ID:1385901945-1"},
   CellID->263629365]},
 "Proof|ID:1385901945-1"->{
  Cell[344855, 10030, 4793, 129, 27, "ProofInfo",
   CellTags->{"Proof|ID:1385901945", "Proof|ID:1385901945-1"},
   CellID->263629365]},
 "Proof|ID:1581648423"->{
  Cell[350169, 10178, 5767, 147, 27, "ProofInfo",
   CellTags->{"Proof|ID:1581648423", "Proof|ID:1581648423-1"},
   CellID->422549508]},
 "Proof|ID:1581648423-1"->{
  Cell[350169, 10178, 5767, 147, 27, "ProofInfo",
   CellTags->{"Proof|ID:1581648423", "Proof|ID:1581648423-1"},
   CellID->422549508]},
 "Proof|ID:1815472507"->{
  Cell[356458, 10344, 5702, 147, 27, "ProofInfo",
   CellTags->{"Proof|ID:1815472507", "Proof|ID:1815472507-1"},
   CellID->1818000638]},
 "Proof|ID:1815472507-1"->{
  Cell[356458, 10344, 5702, 147, 27, "ProofInfo",
   CellTags->{"Proof|ID:1815472507", "Proof|ID:1815472507-1"},
   CellID->1818000638]},
 "Proof|ID:443432455"->{
  Cell[362681, 10510, 5648, 146, 27, "ProofInfo",
   CellTags->{"Proof|ID:443432455", "Proof|ID:443432455-1"},
   CellID->633496152]},
 "Proof|ID:443432455-1"->{
  Cell[362681, 10510, 5648, 146, 27, "ProofInfo",
   CellTags->{"Proof|ID:443432455", "Proof|ID:443432455-1"},
   CellID->633496152]},
 "ID:1546357463"->{
  Cell[371442, 10744, 1421, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1546357463", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1\[LessEqual]length"},
   CellID->1546357463]},
 "1\[LessEqual]length"->{
  Cell[371442, 10744, 1421, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1546357463", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1\[LessEqual]length"},
   CellID->1546357463]},
 "ID:409966668"->{
  Cell[372866, 10785, 1184, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:409966668", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "11"},
   CellID->409966668]},
 "11"->{
  Cell[372866, 10785, 1184, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:409966668", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "11"},
   CellID->409966668]},
 "ID:1768549223"->{
  Cell[374053, 10818, 1710, 45, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1768549223", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "2\[LessEqual]length"},
   CellID->1768549223]},
 "2\[LessEqual]length"->{
  Cell[374053, 10818, 1710, 45, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1768549223", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "2\[LessEqual]length"},
   CellID->1768549223]},
 "ID:1399858996"->{
  Cell[375766, 10865, 1001, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1399858996", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a is integer"},
   CellID->1399858996]},
 "length\:293a is integer"->{
  Cell[375766, 10865, 1001, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1399858996", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a is integer"},
   CellID->1399858996]},
 "ID:1353010357"->{
  Cell[376770, 10895, 982, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1353010357", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a is positive"},
   CellID->1353010357]},
 "length\:293a is positive"->{
  Cell[376770, 10895, 982, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1353010357", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a is positive"},
   CellID->1353010357]},
 "ID:763884008"->{
  Cell[377755, 10924, 1769, 47, 56, "FormalTextInputFormula",
   CellTags->{
    "ID:763884008", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "12"},
   CellID->763884008]},
 "12"->{
  Cell[377755, 10924, 1769, 47, 56, "FormalTextInputFormula",
   CellTags->{
    "ID:763884008", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "12"},
   CellID->763884008]},
 "ID:1442641059"->{
  Cell[379527, 10973, 1175, 33, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:1442641059", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "13"},
   CellID->1442641059]},
 "13"->{
  Cell[379527, 10973, 1175, 33, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:1442641059", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "13"},
   CellID->1442641059]},
 "ID:215173194"->{
  Cell[380705, 11008, 1294, 41, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:215173194", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "14"},
   CellID->215173194]},
 "14"->{
  Cell[380705, 11008, 1294, 41, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:215173194", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "14"},
   CellID->215173194]},
 "Proof|ID:1546357463"->{
  Cell[382155, 11060, 3999, 109, 29, "ProofInfo",
   CellTags->{"Proof|ID:1546357463", "Proof|ID:1546357463-1"},
   CellID->1202974989]},
 "Proof|ID:1546357463-1"->{
  Cell[382155, 11060, 3999, 109, 29, "ProofInfo",
   CellTags->{"Proof|ID:1546357463", "Proof|ID:1546357463-1"},
   CellID->1202974989]},
 "Proof|ID:409966668"->{
  Cell[386676, 11188, 3542, 98, 27, "ProofInfo",
   CellTags->{"Proof|ID:409966668", "Proof|ID:409966668-1"},
   CellID->577682993]},
 "Proof|ID:409966668-1"->{
  Cell[386676, 11188, 3542, 98, 27, "ProofInfo",
   CellTags->{"Proof|ID:409966668", "Proof|ID:409966668-1"},
   CellID->577682993]},
 "Proof|ID:1768549223"->{
  Cell[390738, 11305, 5387, 141, 27, "ProofInfo",
   CellTags->{"Proof|ID:1768549223", "Proof|ID:1768549223-1"},
   CellID->1717553837]},
 "Proof|ID:1768549223-1"->{
  Cell[390738, 11305, 5387, 141, 27, "ProofInfo",
   CellTags->{"Proof|ID:1768549223", "Proof|ID:1768549223-1"},
   CellID->1717553837]},
 "Proof|ID:1399858996"->{
  Cell[396646, 11465, 3015, 85, 27, "ProofInfo",
   CellTags->{"Proof|ID:1399858996", "Proof|ID:1399858996-1"},
   CellID->86355476]},
 "Proof|ID:1399858996-1"->{
  Cell[396646, 11465, 3015, 85, 27, "ProofInfo",
   CellTags->{"Proof|ID:1399858996", "Proof|ID:1399858996-1"},
   CellID->86355476]},
 "Proof|ID:1353010357"->{
  Cell[400183, 11569, 3016, 86, 27, "ProofInfo",
   CellTags->{"Proof|ID:1353010357", "Proof|ID:1353010357-1"},
   CellID->1804120756]},
 "Proof|ID:1353010357-1"->{
  Cell[400183, 11569, 3016, 86, 27, "ProofInfo",
   CellTags->{"Proof|ID:1353010357", "Proof|ID:1353010357-1"},
   CellID->1804120756]},
 "Proof|ID:763884008"->{
  Cell[403721, 11674, 4378, 112, 27, "ProofInfo",
   CellTags->{"Proof|ID:763884008", "Proof|ID:763884008-1"},
   CellID->1513904826]},
 "Proof|ID:763884008-1"->{
  Cell[403721, 11674, 4378, 112, 27, "ProofInfo",
   CellTags->{"Proof|ID:763884008", "Proof|ID:763884008-1"},
   CellID->1513904826]},
 "Proof|ID:1442641059"->{
  Cell[408620, 11805, 4110, 110, 27, "ProofInfo",
   CellTags->{"Proof|ID:1442641059", "Proof|ID:1442641059-1"},
   CellID->1528692265]},
 "Proof|ID:1442641059-1"->{
  Cell[408620, 11805, 4110, 110, 27, "ProofInfo",
   CellTags->{"Proof|ID:1442641059", "Proof|ID:1442641059-1"},
   CellID->1528692265]},
 "Proof|ID:215173194"->{
  Cell[413252, 11934, 3994, 106, 27, "ProofInfo",
   CellTags->{"Proof|ID:215173194", "Proof|ID:215173194-1"},
   CellID->1068613755]},
 "Proof|ID:215173194-1"->{
  Cell[413252, 11934, 3994, 106, 27, "ProofInfo",
   CellTags->{"Proof|ID:215173194", "Proof|ID:215173194-1"},
   CellID->1068613755]},
 "ID:8303793"->{
  Cell[419400, 12112, 1256, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:8303793", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "15"},
   CellID->8303793]},
 "15"->{
  Cell[419400, 12112, 1256, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:8303793", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "15"},
   CellID->8303793]},
 "ID:1339530720"->{
  Cell[420659, 12154, 1265, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1339530720", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "16"},
   CellID->1339530720]},
 "16"->{
  Cell[420659, 12154, 1265, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1339530720", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "16"},
   CellID->1339530720]},
 "ID:759960416"->{
  Cell[421927, 12196, 1378, 43, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:759960416", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[SquareSubsetEqual]\:293a"},
   CellID->759960416]},
 "\[SquareSubsetEqual]\:293a"->{
  Cell[421927, 12196, 1378, 43, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:759960416", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[SquareSubsetEqual]\:293a"},
   CellID->759960416]},
 "Proof|ID:8303793|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[423462, 12250, 4020, 107, 29, "ProofInfo",
   CellTags->
    "Proof|ID:8303793|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->222527391]},
 "Proof|ID:1339530720|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[427839, 12373, 4025, 107, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1339530720|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->980530952]},
 "Proof|ID:759960416|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[432220, 12496, 3839, 105, 27, "ProofInfo",
   CellTags->
    "Proof|ID:759960416|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->1092143830]},
 "sect::Lemmataminmax"->{
  Cell[436426, 12618, 272, 5, 30, "Subsection",
   CellTags->"sect::Lemmataminmax",
   CellID->473572412]},
 "ID:1448857518"->{
  Cell[439932, 12724, 1125, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1448857518", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of min"},
   CellID->1448857518]},
 "defining property of min"->{
  Cell[439932, 12724, 1125, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1448857518", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of min"},
   CellID->1448857518]},
 "ID:1578657786"->{
  Cell[441060, 12756, 1128, 31, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1578657786", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of max"},
   CellID->1578657786]},
 "defining property of max"->{
  Cell[441060, 12756, 1128, 31, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1578657786", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of max"},
   CellID->1578657786]},
 "ID:1968885164"->{
  Cell[442191, 12789, 913, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1968885164", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of min"},
   CellID->1968885164]},
 "type of min"->{
  Cell[442191, 12789, 913, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1968885164", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of min"},
   CellID->1968885164]},
 "ID:1151569277"->{
  Cell[443107, 12817, 913, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1151569277", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of max"},
   CellID->1151569277]},
 "type of max"->{
  Cell[443107, 12817, 913, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1151569277", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of max"},
   CellID->1151569277]},
 "Proof|ID:1448857518"->{
  Cell[444178, 12854, 5017, 132, 29, "ProofInfo",
   CellTags->{"Proof|ID:1448857518", "Proof|ID:1448857518-1"},
   CellID->1149489714]},
 "Proof|ID:1448857518-1"->{
  Cell[444178, 12854, 5017, 132, 29, "ProofInfo",
   CellTags->{"Proof|ID:1448857518", "Proof|ID:1448857518-1"},
   CellID->1149489714]},
 "Proof|ID:1578657786"->{
  Cell[449719, 13005, 5013, 130, 27, "ProofInfo",
   CellTags->{"Proof|ID:1578657786", "Proof|ID:1578657786-1"},
   CellID->1807688628]},
 "Proof|ID:1578657786-1"->{
  Cell[449719, 13005, 5013, 130, 27, "ProofInfo",
   CellTags->{"Proof|ID:1578657786", "Proof|ID:1578657786-1"},
   CellID->1807688628]},
 "Proof|ID:1968885164"->{
  Cell[455256, 13154, 3317, 92, 27, "ProofInfo",
   CellTags->{"Proof|ID:1968885164", "Proof|ID:1968885164-1"},
   CellID->1971047889]},
 "Proof|ID:1968885164-1"->{
  Cell[455256, 13154, 3317, 92, 27, "ProofInfo",
   CellTags->{"Proof|ID:1968885164", "Proof|ID:1968885164-1"},
   CellID->1971047889]},
 "Proof|ID:1151569277"->{
  Cell[459096, 13265, 3317, 92, 27, "ProofInfo",
   CellTags->{"Proof|ID:1151569277", "Proof|ID:1151569277-1"},
   CellID->2058731697]},
 "Proof|ID:1151569277-1"->{
  Cell[459096, 13265, 3317, 92, 27, "ProofInfo",
   CellTags->{"Proof|ID:1151569277", "Proof|ID:1151569277-1"},
   CellID->2058731697]},
 "ID:232133128"->{
  Cell[464763, 13435, 1591, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:232133128", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "17"},
   CellID->232133128]},
 "17"->{
  Cell[464763, 13435, 1591, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:232133128", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "17"},
   CellID->232133128]},
 "ID:1533503972"->{
  Cell[466357, 13481, 1593, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1533503972", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "18"},
   CellID->1533503972]},
 "18"->{
  Cell[466357, 13481, 1593, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1533503972", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "18"},
   CellID->1533503972]},
 "ID:881325315"->{
  Cell[467953, 13527, 1294, 40, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:881325315", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "17'"},
   CellID->881325315]},
 "17'"->{
  Cell[467953, 13527, 1294, 40, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:881325315", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "17'"},
   CellID->881325315]},
 "ID:1667609383"->{
  Cell[469250, 13569, 1297, 40, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1667609383", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "18'"},
   CellID->1667609383]},
 "18'"->{
  Cell[469250, 13569, 1297, 40, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1667609383", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "18'"},
   CellID->1667609383]},
 "ID:1257756255"->{
  Cell[470550, 13611, 1093, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1257756255", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min is invariant under permutations"},
   CellID->1257756255]},
 "min is invariant under permutations"->{
  Cell[470550, 13611, 1093, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1257756255", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min is invariant under permutations"},
   CellID->1257756255]},
 "ID:2071551441"->{
  Cell[471646, 13642, 1093, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2071551441", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max is invariant under permutations"},
   CellID->2071551441]},
 "max is invariant under permutations"->{
  Cell[471646, 13642, 1093, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2071551441", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max is invariant under permutations"},
   CellID->2071551441]},
 "Proof|ID:232133128"->{
  Cell[472896, 13682, 4488, 120, 29, "ProofInfo",
   CellTags->{"Proof|ID:232133128", "Proof|ID:232133128-1"},
   CellID->557201343]},
 "Proof|ID:232133128-1"->{
  Cell[472896, 13682, 4488, 120, 29, "ProofInfo",
   CellTags->{"Proof|ID:232133128", "Proof|ID:232133128-1"},
   CellID->557201343]},
 "Proof|ID:1533503972"->{
  Cell[477907, 13821, 4494, 120, 27, "ProofInfo",
   CellTags->{"Proof|ID:1533503972", "Proof|ID:1533503972-1"},
   CellID->1508553184]},
 "Proof|ID:1533503972-1"->{
  Cell[477907, 13821, 4494, 120, 27, "ProofInfo",
   CellTags->{"Proof|ID:1533503972", "Proof|ID:1533503972-1"},
   CellID->1508553184]},
 "Proof|ID:881325315"->{
  Cell[482924, 13960, 4468, 120, 27, "ProofInfo",
   CellTags->{"Proof|ID:881325315", "Proof|ID:881325315-1"},
   CellID->803466674]},
 "Proof|ID:881325315-1"->{
  Cell[482924, 13960, 4468, 120, 27, "ProofInfo",
   CellTags->{"Proof|ID:881325315", "Proof|ID:881325315-1"},
   CellID->803466674]},
 "Proof|ID:1667609383"->{
  Cell[487915, 14099, 4474, 120, 27, "ProofInfo",
   CellTags->{"Proof|ID:1667609383", "Proof|ID:1667609383-1"},
   CellID->1598215446]},
 "Proof|ID:1667609383-1"->{
  Cell[487915, 14099, 4474, 120, 27, "ProofInfo",
   CellTags->{"Proof|ID:1667609383", "Proof|ID:1667609383-1"},
   CellID->1598215446]},
 "Proof|ID:1257756255"->{
  Cell[492913, 14238, 3684, 102, 27, "ProofInfo",
   CellTags->{"Proof|ID:1257756255", "Proof|ID:1257756255-1"},
   CellID->725596050]},
 "Proof|ID:1257756255-1"->{
  Cell[492913, 14238, 3684, 102, 27, "ProofInfo",
   CellTags->{"Proof|ID:1257756255", "Proof|ID:1257756255-1"},
   CellID->725596050]},
 "Proof|ID:2071551441"->{
  Cell[497121, 14359, 3685, 102, 27, "ProofInfo",
   CellTags->{"Proof|ID:2071551441", "Proof|ID:2071551441-1"},
   CellID->1663975332]},
 "Proof|ID:2071551441-1"->{
  Cell[497121, 14359, 3685, 102, 27, "ProofInfo",
   CellTags->{"Proof|ID:2071551441", "Proof|ID:2071551441-1"},
   CellID->1663975332]},
 "ID:48233088"->{
  Cell[503790, 14546, 1410, 39, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:48233088", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\[LessEqual]"},
   CellID->48233088]},
 "min\[LessEqual]"->{
  Cell[503790, 14546, 1410, 39, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:48233088", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\[LessEqual]"},
   CellID->48233088]},
 "ID:1034128987"->{
  Cell[505203, 14587, 1466, 39, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1034128987", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]min"},
   CellID->1034128987]},
 "\[LessEqual]min"->{
  Cell[505203, 14587, 1466, 39, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1034128987", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]min"},
   CellID->1034128987]},
 "ID:1323776552"->{
  Cell[506672, 14628, 1369, 38, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1323776552", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\[LessEqual]"},
   CellID->1323776552]},
 "max\[LessEqual]"->{
  Cell[506672, 14628, 1369, 38, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1323776552", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\[LessEqual]"},
   CellID->1323776552]},
 "ID:1520079070"->{
  Cell[508044, 14668, 1368, 37, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1520079070", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]max"},
   CellID->1520079070]},
 "\[LessEqual]max"->{
  Cell[508044, 14668, 1368, 37, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1520079070", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]max"},
   CellID->1520079070]},
 "ID:1860476328"->{
  Cell[509415, 14707, 1229, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1860476328", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min<"},
   CellID->1860476328]},
 "min<"->{
  Cell[509415, 14707, 1229, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1860476328", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min<"},
   CellID->1860476328]},
 "ID:1907558186"->{
  Cell[510647, 14745, 1230, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1907558186", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<min"},
   CellID->1907558186]},
 "<min"->{
  Cell[510647, 14745, 1230, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1907558186", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<min"},
   CellID->1907558186]},
 "ID:2008421481"->{
  Cell[511880, 14783, 1230, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:2008421481", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max<"},
   CellID->2008421481]},
 "max<"->{
  Cell[511880, 14783, 1230, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:2008421481", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max<"},
   CellID->2008421481]},
 "ID:1141242040"->{
  Cell[513113, 14821, 1230, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1141242040", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<max"},
   CellID->1141242040]},
 "<max"->{
  Cell[513113, 14821, 1230, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1141242040", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<max"},
   CellID->1141242040]},
 "Proof|ID:48233088"->{
  Cell[514499, 14868, 5819, 150, 29, "ProofInfo",
   CellTags->{"Proof|ID:48233088", "Proof|ID:48233088-1"},
   CellID->2119117596]},
 "Proof|ID:48233088-1"->{
  Cell[514499, 14868, 5819, 150, 29, "ProofInfo",
   CellTags->{"Proof|ID:48233088", "Proof|ID:48233088-1"},
   CellID->2119117596]},
 "Proof|ID:1034128987"->{
  Cell[520840, 15037, 5833, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1034128987", "Proof|ID:1034128987-1"},
   CellID->1605499322]},
 "Proof|ID:1034128987-1"->{
  Cell[520840, 15037, 5833, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1034128987", "Proof|ID:1034128987-1"},
   CellID->1605499322]},
 "Proof|ID:1323776552"->{
  Cell[527196, 15206, 5827, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1323776552", "Proof|ID:1323776552-1"},
   CellID->87934740]},
 "Proof|ID:1323776552-1"->{
  Cell[527196, 15206, 5827, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1323776552", "Proof|ID:1323776552-1"},
   CellID->87934740]},
 "Proof|ID:1520079070"->{
  Cell[533547, 15375, 5832, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1520079070", "Proof|ID:1520079070-1"},
   CellID->650805455]},
 "Proof|ID:1520079070-1"->{
  Cell[533547, 15375, 5832, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1520079070", "Proof|ID:1520079070-1"},
   CellID->650805455]},
 "Proof|ID:1860476328"->{
  Cell[539903, 15544, 5762, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1860476328", "Proof|ID:1860476328-1"},
   CellID->489839113]},
 "Proof|ID:1860476328-1"->{
  Cell[539903, 15544, 5762, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1860476328", "Proof|ID:1860476328-1"},
   CellID->489839113]},
 "Proof|ID:1907558186"->{
  Cell[546187, 15713, 5766, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1907558186", "Proof|ID:1907558186-1"},
   CellID->846360522]},
 "Proof|ID:1907558186-1"->{
  Cell[546187, 15713, 5766, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1907558186", "Proof|ID:1907558186-1"},
   CellID->846360522]},
 "Proof|ID:2008421481"->{
  Cell[552477, 15882, 5763, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:2008421481", "Proof|ID:2008421481-1"},
   CellID->1454120988]},
 "Proof|ID:2008421481-1"->{
  Cell[552477, 15882, 5763, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:2008421481", "Proof|ID:2008421481-1"},
   CellID->1454120988]},
 "Proof|ID:1141242040"->{
  Cell[558764, 16051, 5766, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1141242040", "Proof|ID:1141242040-1"},
   CellID->155865952]},
 "Proof|ID:1141242040-1"->{
  Cell[558764, 16051, 5766, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1141242040", "Proof|ID:1141242040-1"},
   CellID->155865952]},
 "ID:572537117"->{
  Cell[566074, 16250, 1266, 38, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:572537117", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "19"},
   CellID->572537117]},
 "19"->{
  Cell[566074, 16250, 1266, 38, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:572537117", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "19"},
   CellID->572537117]},
 "ID:2073093776"->{
  Cell[567343, 16290, 1892, 51, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2073093776", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a 1"},
   CellID->2073093776]},
 "min\:293a 1"->{
  Cell[567343, 16290, 1892, 51, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2073093776", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a 1"},
   CellID->2073093776]},
 "ID:1632846240"->{
  Cell[569238, 16343, 1359, 42, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1632846240", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a 2"},
   CellID->1632846240]},
 "min\:293a 2"->{
  Cell[569238, 16343, 1359, 42, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1632846240", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a 2"},
   CellID->1632846240]},
 "ID:1146565830"->{
  Cell[570600, 16387, 1269, 38, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1146565830", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "20"},
   CellID->1146565830]},
 "20"->{
  Cell[570600, 16387, 1269, 38, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1146565830", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "20"},
   CellID->1146565830]},
 "ID:1172583037"->{
  Cell[571872, 16427, 1849, 50, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1172583037", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a 1"},
   CellID->1172583037]},
 "max\:293a 1"->{
  Cell[571872, 16427, 1849, 50, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1172583037", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a 1"},
   CellID->1172583037]},
 "ID:2142810348"->{
  Cell[573724, 16479, 1359, 42, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2142810348", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a 2"},
   CellID->2142810348]},
 "max\:293a 2"->{
  Cell[573724, 16479, 1359, 42, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2142810348", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a 2"},
   CellID->2142810348]},
 "ID:1460290836"->{
  Cell[575387, 16532, 1096, 33, 36, "FormalTextInputFormula",
   CellTags->{
    "ID:1460290836", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a"},
   CellID->1460290836]},
 "min\:293a"->{
  Cell[575387, 16532, 1096, 33, 36, "FormalTextInputFormula",
   CellTags->{
    "ID:1460290836", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a"},
   CellID->1460290836]},
 "ID:1931124773"->{
  Cell[576486, 16567, 1098, 33, 36, "FormalTextInputFormula",
   CellTags->{
    "ID:1931124773", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a"},
   CellID->1931124773]},
 "max\:293a"->{
  Cell[576486, 16567, 1098, 33, 36, "FormalTextInputFormula",
   CellTags->{
    "ID:1931124773", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a"},
   CellID->1931124773]},
 "ID:978507648"->{
  Cell[577587, 16602, 990, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:978507648", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a decreases"},
   CellID->978507648]},
 "min\:293a decreases"->{
  Cell[577587, 16602, 990, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:978507648", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a decreases"},
   CellID->978507648]},
 "ID:2097939500"->{
  Cell[578580, 16633, 994, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2097939500", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a increases"},
   CellID->2097939500]},
 "max\:293a increases"->{
  Cell[578580, 16633, 994, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2097939500", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a increases"},
   CellID->2097939500]},
 "Proof|ID:572537117"->{
  Cell[579732, 16673, 5085, 134, 29, "ProofInfo",
   CellTags->{"Proof|ID:572537117", "Proof|ID:572537117-1"},
   CellID->1768246037]},
 "Proof|ID:572537117-1"->{
  Cell[579732, 16673, 5085, 134, 29, "ProofInfo",
   CellTags->{"Proof|ID:572537117", "Proof|ID:572537117-1"},
   CellID->1768246037]},
 "Proof|ID:2073093776"->{
  Cell[585340, 16826, 5136, 134, 27, "ProofInfo",
   CellTags->{"Proof|ID:2073093776", "Proof|ID:2073093776-1"},
   CellID->470202616]},
 "Proof|ID:2073093776-1"->{
  Cell[585340, 16826, 5136, 134, 27, "ProofInfo",
   CellTags->{"Proof|ID:2073093776", "Proof|ID:2073093776-1"},
   CellID->470202616]},
 "Proof|ID:1632846240"->{
  Cell[590999, 16979, 4985, 132, 27, "ProofInfo",
   CellTags->{"Proof|ID:1632846240", "Proof|ID:1632846240-1"},
   CellID->163303679]},
 "Proof|ID:1632846240-1"->{
  Cell[590999, 16979, 4985, 132, 27, "ProofInfo",
   CellTags->{"Proof|ID:1632846240", "Proof|ID:1632846240-1"},
   CellID->163303679]},
 "Proof|ID:1146565830"->{
  Cell[596508, 17130, 5090, 134, 27, "ProofInfo",
   CellTags->{"Proof|ID:1146565830", "Proof|ID:1146565830-1"},
   CellID->1152786378]},
 "Proof|ID:1146565830-1"->{
  Cell[596508, 17130, 5090, 134, 27, "ProofInfo",
   CellTags->{"Proof|ID:1146565830", "Proof|ID:1146565830-1"},
   CellID->1152786378]},
 "Proof|ID:1172583037"->{
  Cell[602121, 17283, 5136, 134, 27, "ProofInfo",
   CellTags->{"Proof|ID:1172583037", "Proof|ID:1172583037-1"},
   CellID->222384166]},
 "Proof|ID:1172583037-1"->{
  Cell[602121, 17283, 5136, 134, 27, "ProofInfo",
   CellTags->{"Proof|ID:1172583037", "Proof|ID:1172583037-1"},
   CellID->222384166]},
 "Proof|ID:2142810348"->{
  Cell[607780, 17436, 4986, 132, 27, "ProofInfo",
   CellTags->{"Proof|ID:2142810348", "Proof|ID:2142810348-1"},
   CellID->1174101328]},
 "Proof|ID:2142810348-1"->{
  Cell[607780, 17436, 4986, 132, 27, "ProofInfo",
   CellTags->{"Proof|ID:2142810348", "Proof|ID:2142810348-1"},
   CellID->1174101328]},
 "Proof|ID:1460290836"->{
  Cell[613289, 17587, 4501, 118, 27, "ProofInfo",
   CellTags->{"Proof|ID:1460290836", "Proof|ID:1460290836-1"},
   CellID->1313465390]},
 "Proof|ID:1460290836-1"->{
  Cell[613289, 17587, 4501, 118, 27, "ProofInfo",
   CellTags->{"Proof|ID:1460290836", "Proof|ID:1460290836-1"},
   CellID->1313465390]},
 "Proof|ID:1931124773"->{
  Cell[618379, 17725, 4499, 118, 27, "ProofInfo",
   CellTags->{"Proof|ID:1931124773", "Proof|ID:1931124773-1"},
   CellID->70708585]},
 "Proof|ID:1931124773-1"->{
  Cell[618379, 17725, 4499, 118, 27, "ProofInfo",
   CellTags->{"Proof|ID:1931124773", "Proof|ID:1931124773-1"},
   CellID->70708585]},
 "Proof|ID:978507648"->{
  Cell[623402, 17862, 4044, 108, 27, "ProofInfo",
   CellTags->{"Proof|ID:978507648", "Proof|ID:978507648-1"},
   CellID->1954389032]},
 "Proof|ID:978507648-1"->{
  Cell[623402, 17862, 4044, 108, 27, "ProofInfo",
   CellTags->{"Proof|ID:978507648", "Proof|ID:978507648-1"},
   CellID->1954389032]},
 "Proof|ID:2097939500"->{
  Cell[627969, 17989, 4048, 108, 27, "ProofInfo",
   CellTags->{"Proof|ID:2097939500", "Proof|ID:2097939500-1"},
   CellID->910681388]},
 "Proof|ID:2097939500-1"->{
  Cell[627969, 17989, 4048, 108, 27, "ProofInfo",
   CellTags->{"Proof|ID:2097939500", "Proof|ID:2097939500-1"},
   CellID->910681388]},
 "ID:900399941"->{
  Cell[634688, 18183, 1661, 47, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:900399941", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "21"},
   CellID->900399941]},
 "21"->{
  Cell[634688, 18183, 1661, 47, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:900399941", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "21"},
   CellID->900399941]},
 "ID:1833484473"->{
  Cell[636352, 18232, 1523, 44, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1833484473", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 1"},
   CellID->1833484473]},
 "min+ 1"->{
  Cell[636352, 18232, 1523, 44, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1833484473", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 1"},
   CellID->1833484473]},
 "ID:1976058899"->{
  Cell[637878, 18278, 1271, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1976058899", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 2"},
   CellID->1976058899]},
 "min+ 2"->{
  Cell[637878, 18278, 1271, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1976058899", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 2"},
   CellID->1976058899]},
 "ID:2067845318"->{
  Cell[639152, 18317, 1272, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:2067845318", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 3"},
   CellID->2067845318]},
 "min+ 3"->{
  Cell[639152, 18317, 1272, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:2067845318", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 3"},
   CellID->2067845318]},
 "ID:446229453"->{
  Cell[640427, 18356, 1430, 39, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:446229453", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+"},
   CellID->446229453]},
 "min+"->{
  Cell[640427, 18356, 1430, 39, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:446229453", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+"},
   CellID->446229453]},
 "ID:437038980"->{
  Cell[641860, 18397, 1472, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:437038980", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 1"},
   CellID->437038980]},
 "max+ 1"->{
  Cell[641860, 18397, 1472, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:437038980", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 1"},
   CellID->437038980]},
 "ID:367539508"->{
  Cell[643335, 18442, 1268, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:367539508", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 2"},
   CellID->367539508]},
 "max+ 2"->{
  Cell[643335, 18442, 1268, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:367539508", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 2"},
   CellID->367539508]},
 "ID:1689218301"->{
  Cell[644606, 18481, 1272, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1689218301", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 3"},
   CellID->1689218301]},
 "max+ 3"->{
  Cell[644606, 18481, 1272, 37, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1689218301", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 3"},
   CellID->1689218301]},
 "ID:513276857"->{
  Cell[645881, 18520, 1380, 38, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:513276857", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+"},
   CellID->513276857]},
 "max+"->{
  Cell[645881, 18520, 1380, 38, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:513276857", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+"},
   CellID->513276857]},
 "Proof|ID:900399941"->{
  Cell[647419, 18569, 8666, 210, 29, "ProofInfo",
   CellTags->{"Proof|ID:900399941", "Proof|ID:900399941-1"},
   CellID->1837094901]},
 "Proof|ID:900399941-1"->{
  Cell[647419, 18569, 8666, 210, 29, "ProofInfo",
   CellTags->{"Proof|ID:900399941", "Proof|ID:900399941-1"},
   CellID->1837094901]},
 "Proof|ID:1833484473"->{
  Cell[656608, 18798, 6464, 162, 27, "ProofInfo",
   CellTags->{"Proof|ID:1833484473", "Proof|ID:1833484473-1"},
   CellID->209997353]},
 "Proof|ID:1833484473-1"->{
  Cell[656608, 18798, 6464, 162, 27, "ProofInfo",
   CellTags->{"Proof|ID:1833484473", "Proof|ID:1833484473-1"},
   CellID->209997353]},
 "Proof|ID:1976058899"->{
  Cell[663662, 18980, 6159, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:1976058899", "Proof|ID:1976058899-1"},
   CellID->1890751220]},
 "Proof|ID:1976058899-1"->{
  Cell[663662, 18980, 6159, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:1976058899", "Proof|ID:1976058899-1"},
   CellID->1890751220]},
 "Proof|ID:2067845318"->{
  Cell[670410, 19156, 6168, 158, 27, "ProofInfo",
   CellTags->{"Proof|ID:2067845318", "Proof|ID:2067845318-1"},
   CellID->1653738850]},
 "Proof|ID:2067845318-1"->{
  Cell[670410, 19156, 6168, 158, 27, "ProofInfo",
   CellTags->{"Proof|ID:2067845318", "Proof|ID:2067845318-1"},
   CellID->1653738850]},
 "Proof|ID:446229453"->{
  Cell[677102, 19333, 6071, 154, 27, "ProofInfo",
   CellTags->{"Proof|ID:446229453", "Proof|ID:446229453-1"},
   CellID->357185825]},
 "Proof|ID:446229453-1"->{
  Cell[677102, 19333, 6071, 154, 27, "ProofInfo",
   CellTags->{"Proof|ID:446229453", "Proof|ID:446229453-1"},
   CellID->357185825]},
 "Proof|ID:437038980"->{
  Cell[683736, 19507, 6459, 162, 27, "ProofInfo",
   CellTags->{"Proof|ID:437038980", "Proof|ID:437038980-1"},
   CellID->132391776]},
 "Proof|ID:437038980-1"->{
  Cell[683736, 19507, 6459, 162, 27, "ProofInfo",
   CellTags->{"Proof|ID:437038980", "Proof|ID:437038980-1"},
   CellID->132391776]},
 "Proof|ID:367539508"->{
  Cell[690718, 19688, 6154, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:367539508", "Proof|ID:367539508-1"},
   CellID->1111387014]},
 "Proof|ID:367539508-1"->{
  Cell[690718, 19688, 6154, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:367539508", "Proof|ID:367539508-1"},
   CellID->1111387014]},
 "Proof|ID:1689218301"->{
  Cell[697394, 19863, 6167, 158, 27, "ProofInfo",
   CellTags->{"Proof|ID:1689218301", "Proof|ID:1689218301-1"},
   CellID->434528939]},
 "Proof|ID:1689218301-1"->{
  Cell[697394, 19863, 6167, 158, 27, "ProofInfo",
   CellTags->{"Proof|ID:1689218301", "Proof|ID:1689218301-1"},
   CellID->434528939]},
 "Proof|ID:513276857"->{
  Cell[704084, 20040, 6072, 154, 27, "ProofInfo",
   CellTags->{"Proof|ID:513276857", "Proof|ID:513276857-1"},
   CellID->1335154081]},
 "Proof|ID:513276857-1"->{
  Cell[704084, 20040, 6072, 154, 27, "ProofInfo",
   CellTags->{"Proof|ID:513276857", "Proof|ID:513276857-1"},
   CellID->1335154081]},
 "sect::SpecificNotions"->{
  Cell[711185, 20228, 279, 5, 43, "Section",
   CellTags->"sect::SpecificNotions",
   CellID->1752495898]},
 "ID:1076035631"->{
  Cell[712804, 20275, 718, 15, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1076035631", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->1076035631]},
 "ID:610987882"->{
  Cell[713525, 20292, 1777, 49, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:610987882", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExponentVector"},
   CellID->610987882]},
 "isExponentVector"->{
  Cell[713525, 20292, 1777, 49, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:610987882", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExponentVector"},
   CellID->610987882]},
 "ID:468249611"->{
  Cell[715305, 20343, 1542, 47, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:468249611", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExpVectorTuple"},
   CellID->468249611]},
 "isExpVectorTuple"->{
  Cell[715305, 20343, 1542, 47, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:468249611", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExpVectorTuple"},
   CellID->468249611]},
 "ID:444732983"->{
  Cell[716850, 20392, 1581, 41, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:444732983", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "divisibility"},
   CellID->444732983]},
 "divisibility"->{
  Cell[716850, 20392, 1581, 41, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:444732983", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "divisibility"},
   CellID->444732983]},
 "ID:1111389324"->{
  Cell[718434, 20435, 1154, 30, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1111389324", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "degree"},
   CellID->1111389324]},
 "degree"->{
  Cell[718434, 20435, 1154, 30, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1111389324", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "degree"},
   CellID->1111389324]},
 "ID:1699354265"->{
  Cell[719591, 20467, 1384, 36, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1699354265", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maxdeg"},
   CellID->1699354265]},
 "maxdeg"->{
  Cell[719591, 20467, 1384, 36, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1699354265", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maxdeg"},
   CellID->1699354265]},
 "ID:1614921088"->{
  Cell[720978, 20505, 1538, 39, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:1614921088", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "lcm"},
   CellID->1614921088]},
 "lcm"->{
  Cell[720978, 20505, 1538, 39, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:1614921088", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "lcm"},
   CellID->1614921088]},
 "ID:1784038925"->{
  Cell[722519, 20546, 2791, 78, 77, "FormalTextInputFormula",
   CellTags->{
    "ID:1784038925", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chain criterion"},
   CellID->1784038925]},
 "chain criterion"->{
  Cell[722519, 20546, 2791, 78, 77, "FormalTextInputFormula",
   CellTags->{
    "ID:1784038925", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chain criterion"},
   CellID->1784038925]},
 "ID:312341806"->{
  Cell[725313, 20626, 1856, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:312341806", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "width w.r.t. k-th coordinate"},
   CellID->312341806]},
 "width w.r.t. k-th coordinate"->{
  Cell[725313, 20626, 1856, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:312341806", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "width w.r.t. k-th coordinate"},
   CellID->312341806]},
 "ID:282925385"->{
  Cell[727172, 20671, 1398, 34, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:282925385", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "width"},
   CellID->282925385]},
 "width"->{
  Cell[727172, 20671, 1398, 34, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:282925385", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "width"},
   CellID->282925385]},
 "ID:1664335746"->{
  Cell[728573, 20707, 1913, 47, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1664335746", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maximum degree of admissible pairs of one object"},
   CellID->1664335746]},
 "maximum degree of admissible pairs of one object"->{
  Cell[728573, 20707, 1913, 47, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1664335746", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maximum degree of admissible pairs of one object"},
   CellID->1664335746]},
 "ID:640444143"->{
  Cell[730489, 20756, 1712, 41, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:640444143", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maximum degree of admissible pairs"},
   CellID->640444143]},
 "maximum degree of admissible pairs"->{
  Cell[730489, 20756, 1712, 41, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:640444143", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maximum degree of admissible pairs"},
   CellID->640444143]},
 "ID:1923113076"->{
  Cell[732204, 20799, 1697, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1923113076", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "corner of rectangular region"},
   CellID->1923113076]},
 "corner of rectangular region"->{
  Cell[732204, 20799, 1697, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1923113076", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "corner of rectangular region"},
   CellID->1923113076]},
 "ID:1774661003"->{
  Cell[734674, 20863, 1447, 40, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1774661003", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "above"},
   CellID->1774661003]},
 "above"->{
  Cell[734674, 20863, 1447, 40, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1774661003", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "above"},
   CellID->1774661003]},
 "ID:285552196"->{
  Cell[736124, 20905, 2246, 63, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:285552196", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "far exterior w.r.t. k-th coordinate"},
   CellID->285552196]},
 "far exterior w.r.t. k-th coordinate"->{
  Cell[736124, 20905, 2246, 63, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:285552196", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "far exterior w.r.t. k-th coordinate"},
   CellID->285552196]},
 "ID:494773971"->{
  Cell[738373, 20970, 1894, 50, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:494773971", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "rectangular region"},
   CellID->494773971]},
 "rectangular region"->{
  Cell[738373, 20970, 1894, 50, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:494773971", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "rectangular region"},
   CellID->494773971]},
 "ID:1278822860"->{
  Cell[750499, 21295, 2667, 81, 112, "FormalTextInputFormula",
   CellTags->{
    "ID:1278822860", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal w.r.t. k-th coordinate"},
   CellID->1278822860]},
 "chainCritDiagonal w.r.t. k-th coordinate"->{
  Cell[750499, 21295, 2667, 81, 112, "FormalTextInputFormula",
   CellTags->{
    "ID:1278822860", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal w.r.t. k-th coordinate"},
   CellID->1278822860]},
 "ID:1458824241"->{
  Cell[753169, 21378, 1356, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1458824241", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal"},
   CellID->1458824241]},
 "chainCritDiagonal"->{
  Cell[753169, 21378, 1356, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1458824241", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal"},
   CellID->1458824241]},
 "sect::Computations"->{
  Cell[774544, 21940, 323, 6, 43, "Section",
   CellTags->"sect::Computations",
   CellID->1485320096]},
 "ID:1559569657"->{
  Cell[776188, 21987, 809, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1559569657", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D=N0"},
   CellID->1559569657]},
 "D=N0"->{
  Cell[776188, 21987, 809, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1559569657", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D=N0"},
   CellID->1559569657]},
 "ID:977095909"->{
  Cell[778561, 22070, 296, 8, 30, "GlobalDeclaration",
   CellTags->{
    "ID:977095909", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->977095909]},
 "ID:615519889"->{
  Cell[778860, 22080, 975, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:615519889", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min in N0"},
   CellID->615519889]},
 "min in N0"->{
  Cell[778860, 22080, 975, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:615519889", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min in N0"},
   CellID->615519889]},
 "ID:1245684717"->{
  Cell[779838, 22108, 977, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1245684717", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max in N0"},
   CellID->1245684717]},
 "max in N0"->{
  Cell[779838, 22108, 977, 26, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1245684717", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max in N0"},
   CellID->1245684717]},
 "ID:1526008058"->{
  Cell[782089, 22185, 737, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1526008058", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n=2"},
   CellID->1526008058]},
 "n=2"->{
  Cell[782089, 22185, 737, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1526008058", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n=2"},
   CellID->1526008058]},
 "ID:1544493310"->{
  Cell[806526, 22925, 786, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1544493310", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n=3"},
   CellID->1544493310]},
 "n=3"->{
  Cell[806526, 22925, 786, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1544493310", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n=3"},
   CellID->1544493310]},
 "sect::LemmataSpecific"->{
  Cell[833639, 23743, 289, 5, 43, "Section",
   CellTags->"sect::LemmataSpecific",
   CellID->1812043342]},
 "sect::LemmataSpecificGeneral"->{
  Cell[836976, 23830, 236, 4, 38, "Subsection",
   CellTags->"sect::LemmataSpecificGeneral",
   CellID->1771021382]},
 "ID:18996081"->{
  Cell[839862, 23909, 1096, 31, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:18996081", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "22"},
   CellID->18996081]},
 "22"->{
  Cell[839862, 23909, 1096, 31, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:18996081", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "22"},
   CellID->18996081]},
 "Proof|ID:18996081"->{
  Cell[841114, 23951, 2947, 86, 29, "ProofInfo",
   CellTags->{"Proof|ID:18996081", "Proof|ID:18996081-1"},
   CellID->1450282884]},
 "Proof|ID:18996081-1"->{
  Cell[841114, 23951, 2947, 86, 29, "ProofInfo",
   CellTags->{"Proof|ID:18996081", "Proof|ID:18996081-1"},
   CellID->1450282884]},
 "ID:313023917"->{
  Cell[845263, 24078, 868, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:313023917", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExpVectorTuple\:293a"},
   CellID->313023917]},
 "isExpVectorTuple\:293a"->{
  Cell[845263, 24078, 868, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:313023917", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExpVectorTuple\:293a"},
   CellID->313023917]},
 "ID:1697603634"->{
  Cell[846134, 24104, 745, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1697603634", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "23"},
   CellID->1697603634]},
 "23"->{
  Cell[846134, 24104, 745, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1697603634", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "23"},
   CellID->1697603634]},
 "Proof|ID:313023917|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[847036, 24135, 2562, 73, 29, "ProofInfo",
   CellTags->
    "Proof|ID:313023917|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->360389676]},
 "Proof|ID:1697603634"->{
  Cell[849953, 24224, 2165, 66, 27, "ProofInfo",
   CellTags->{"Proof|ID:1697603634", "Proof|ID:1697603634-1"},
   CellID->1339747217]},
 "Proof|ID:1697603634-1"->{
  Cell[849953, 24224, 2165, 66, 27, "ProofInfo",
   CellTags->{"Proof|ID:1697603634", "Proof|ID:1697603634-1"},
   CellID->1339747217]},
 "ID:1313148627"->{
  Cell[854797, 24365, 1119, 31, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1313148627", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "antisymmetry of divisibility"},
   CellID->1313148627]},
 "antisymmetry of divisibility"->{
  Cell[854797, 24365, 1119, 31, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1313148627", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "antisymmetry of divisibility"},
   CellID->1313148627]},
 "ID:785542046"->{
  Cell[855919, 24398, 794, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:785542046", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of divisibility"},
   CellID->785542046]},
 "reflexivity of divisibility"->{
  Cell[855919, 24398, 794, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:785542046", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of divisibility"},
   CellID->785542046]},
 "ID:1055765423"->{
  Cell[856716, 24420, 1029, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1055765423", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity of divisibility"},
   CellID->1055765423]},
 "transitivity of divisibility"->{
  Cell[856716, 24420, 1029, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1055765423", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity of divisibility"},
   CellID->1055765423]},
 "Proof|ID:1313148627"->{
  Cell[857903, 24460, 3275, 91, 29, "ProofInfo",
   CellTags->{"Proof|ID:1313148627", "Proof|ID:1313148627-1"},
   CellID->513842368]},
 "Proof|ID:1313148627-1"->{
  Cell[857903, 24460, 3275, 91, 29, "ProofInfo",
   CellTags->{"Proof|ID:1313148627", "Proof|ID:1313148627-1"},
   CellID->513842368]},
 "Proof|ID:785542046"->{
  Cell[861701, 24570, 2314, 70, 27, "ProofInfo",
   CellTags->{"Proof|ID:785542046", "Proof|ID:785542046-1"},
   CellID->530754490]},
 "Proof|ID:785542046-1"->{
  Cell[861701, 24570, 2314, 70, 27, "ProofInfo",
   CellTags->{"Proof|ID:785542046", "Proof|ID:785542046-1"},
   CellID->530754490]},
 "Proof|ID:1055765423"->{
  Cell[864537, 24659, 3505, 95, 27, "ProofInfo",
   CellTags->{"Proof|ID:1055765423", "Proof|ID:1055765423-1"},
   CellID->985028352]},
 "Proof|ID:1055765423-1"->{
  Cell[864537, 24659, 3505, 95, 27, "ProofInfo",
   CellTags->{"Proof|ID:1055765423", "Proof|ID:1055765423-1"},
   CellID->985028352]},
 "ID:1918380148"->{
  Cell[870776, 24838, 893, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1918380148", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of lcm"},
   CellID->1918380148]},
 "type of lcm"->{
  Cell[870776, 24838, 893, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1918380148", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of lcm"},
   CellID->1918380148]},
 "ID:1065951991"->{
  Cell[871672, 24862, 844, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1065951991", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 1"},
   CellID->1065951991]},
 "defining property of lcm 1"->{
  Cell[871672, 24862, 844, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1065951991", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 1"},
   CellID->1065951991]},
 "ID:1071215803"->{
  Cell[872519, 24885, 844, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1071215803", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 2"},
   CellID->1071215803]},
 "defining property of lcm 2"->{
  Cell[872519, 24885, 844, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1071215803", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 2"},
   CellID->1071215803]},
 "ID:1028467786"->{
  Cell[873366, 24908, 1115, 29, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:1028467786", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 3"},
   CellID->1028467786]},
 "defining property of lcm 3"->{
  Cell[873366, 24908, 1115, 29, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:1028467786", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 3"},
   CellID->1028467786]},
 "ID:98270815"->{
  Cell[874484, 24939, 896, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:98270815", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "commutativity of lcm"},
   CellID->98270815]},
 "commutativity of lcm"->{
  Cell[874484, 24939, 896, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:98270815", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "commutativity of lcm"},
   CellID->98270815]},
 "ID:622866956"->{
  Cell[875383, 24964, 904, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:622866956", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "24"},
   CellID->622866956]},
 "24"->{
  Cell[875383, 24964, 904, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:622866956", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "24"},
   CellID->622866956]},
 "ID:810936059"->{
  Cell[876290, 24990, 925, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:810936059", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "idempotency of lcm"},
   CellID->810936059]},
 "idempotency of lcm"->{
  Cell[876290, 24990, 925, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:810936059", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "idempotency of lcm"},
   CellID->810936059]},
 "Proof|ID:1918380148"->{
  Cell[877373, 25026, 2669, 76, 29, "ProofInfo",
   CellTags->{"Proof|ID:1918380148", "Proof|ID:1918380148-1"},
   CellID->1760699056]},
 "Proof|ID:1918380148-1"->{
  Cell[877373, 25026, 2669, 76, 29, "ProofInfo",
   CellTags->{"Proof|ID:1918380148", "Proof|ID:1918380148-1"},
   CellID->1760699056]},
 "Proof|ID:1065951991"->{
  Cell[880565, 25121, 2777, 78, 27, "ProofInfo",
   CellTags->{"Proof|ID:1065951991", "Proof|ID:1065951991-1"},
   CellID->1695070992]},
 "Proof|ID:1065951991-1"->{
  Cell[880565, 25121, 2777, 78, 27, "ProofInfo",
   CellTags->{"Proof|ID:1065951991", "Proof|ID:1065951991-1"},
   CellID->1695070992]},
 "Proof|ID:1071215803"->{
  Cell[883865, 25218, 2777, 78, 27, "ProofInfo",
   CellTags->{"Proof|ID:1071215803", "Proof|ID:1071215803-1"},
   CellID->1919630296]},
 "Proof|ID:1071215803-1"->{
  Cell[883865, 25218, 2777, 78, 27, "ProofInfo",
   CellTags->{"Proof|ID:1071215803", "Proof|ID:1071215803-1"},
   CellID->1919630296]},
 "Proof|ID:1028467786"->{
  Cell[887164, 25315, 3783, 102, 27, "ProofInfo",
   CellTags->{"Proof|ID:1028467786", "Proof|ID:1028467786-1"},
   CellID->488707796]},
 "Proof|ID:1028467786-1"->{
  Cell[887164, 25315, 3783, 102, 27, "ProofInfo",
   CellTags->{"Proof|ID:1028467786", "Proof|ID:1028467786-1"},
   CellID->488707796]},
 "Proof|ID:98270815"->{
  Cell[891471, 25436, 2992, 84, 27, "ProofInfo",
   CellTags->{"Proof|ID:98270815", "Proof|ID:98270815-1"},
   CellID->615308222]},
 "Proof|ID:98270815-1"->{
  Cell[891471, 25436, 2992, 84, 27, "ProofInfo",
   CellTags->{"Proof|ID:98270815", "Proof|ID:98270815-1"},
   CellID->615308222]},
 "Proof|ID:622866956"->{
  Cell[894983, 25539, 3100, 88, 27, "ProofInfo",
   CellTags->{"Proof|ID:622866956", "Proof|ID:622866956-1"},
   CellID->947602302]},
 "Proof|ID:622866956-1"->{
  Cell[894983, 25539, 3100, 88, 27, "ProofInfo",
   CellTags->{"Proof|ID:622866956", "Proof|ID:622866956-1"},
   CellID->947602302]},
 "Proof|ID:810936059"->{
  Cell[898606, 25646, 2534, 76, 27, "ProofInfo",
   CellTags->{"Proof|ID:810936059", "Proof|ID:810936059-1"},
   CellID->898800181]},
 "Proof|ID:810936059-1"->{
  Cell[898606, 25646, 2534, 76, 27, "ProofInfo",
   CellTags->{"Proof|ID:810936059", "Proof|ID:810936059-1"},
   CellID->898800181]},
 "ID:1822207066"->{
  Cell[903499, 25794, 956, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1822207066", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of deg"},
   CellID->1822207066]},
 "type of deg"->{
  Cell[903499, 25794, 956, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1822207066", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of deg"},
   CellID->1822207066]},
 "ID:455032720"->{
  Cell[904458, 25821, 972, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:455032720", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "25"},
   CellID->455032720]},
 "25"->{
  Cell[904458, 25821, 972, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:455032720", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "25"},
   CellID->455032720]},
 "ID:839534060"->{
  Cell[905433, 25850, 980, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:839534060", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "26"},
   CellID->839534060]},
 "26"->{
  Cell[905433, 25850, 980, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:839534060", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "26"},
   CellID->839534060]},
 "Proof|ID:1822207066"->{
  Cell[906570, 25888, 2336, 70, 29, "ProofInfo",
   CellTags->{"Proof|ID:1822207066", "Proof|ID:1822207066-1"},
   CellID->694957922]},
 "Proof|ID:1822207066-1"->{
  Cell[906570, 25888, 2336, 70, 29, "ProofInfo",
   CellTags->{"Proof|ID:1822207066", "Proof|ID:1822207066-1"},
   CellID->694957922]},
 "Proof|ID:455032720|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[909430, 25977, 3037, 85, 27, "ProofInfo",
   CellTags->
    "Proof|ID:455032720|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->1987093019]},
 "Proof|ID:839534060|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[912885, 26079, 2918, 83, 27, "ProofInfo",
   CellTags->
    "Proof|ID:839534060|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->849905264]},
 "ID:2123737964"->{
  Cell[919004, 26260, 892, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2123737964", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of maxdeg"},
   CellID->2123737964]},
 "type of maxdeg"->{
  Cell[919004, 26260, 892, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2123737964", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of maxdeg"},
   CellID->2123737964]},
 "ID:647995479"->{
  Cell[919899, 26286, 1088, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:647995479", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "27"},
   CellID->647995479]},
 "27"->{
  Cell[919899, 26286, 1088, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:647995479", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "27"},
   CellID->647995479]},
 "Proof|ID:2123737964|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[921145, 26327, 3128, 87, 29, "ProofInfo",
   CellTags->
    "Proof|ID:2123737964|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->1501203891]},
 "Proof|ID:647995479|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[924671, 26431, 4876, 127, 27, "ProofInfo",
   CellTags->
    "Proof|ID:647995479|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->878198538]},
 "sect::LemmataSpecificGeneralM"->{
  Cell[929979, 26576, 284, 5, 22, "Subsubsection",
   CellTags->"sect::LemmataSpecificGeneralM",
   CellID->940386836]},
 "ID:1452963633"->{
  Cell[932058, 26632, 849, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1452963633", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "28"},
   CellID->1452963633]},
 "28"->{
  Cell[932058, 26632, 849, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1452963633", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "28"},
   CellID->1452963633]},
 "ID:1500097429"->{
  Cell[932910, 26657, 932, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1500097429", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "29"},
   CellID->1500097429]},
 "29"->{
  Cell[932910, 26657, 932, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1500097429", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "29"},
   CellID->1500097429]},
 "ID:279915377"->{
  Cell[934329, 26699, 1084, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:279915377", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCrit"},
   CellID->279915377]},
 "symmetry of chainCrit"->{
  Cell[934329, 26699, 1084, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:279915377", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCrit"},
   CellID->279915377]},
 "ID:1391995733"->{
  Cell[935416, 26730, 1291, 35, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1391995733", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "divisibility\[DoubleRightArrow]chainCrit"},
   CellID->1391995733]},
 "divisibility\[DoubleRightArrow]chainCrit"->{
  Cell[935416, 26730, 1291, 35, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1391995733", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "divisibility\[DoubleRightArrow]chainCrit"},
   CellID->1391995733]},
 "ID:1271079949"->{
  Cell[936710, 26767, 823, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1271079949", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of chainCrit"},
   CellID->1271079949]},
 "reflexivity of chainCrit"->{
  Cell[936710, 26767, 823, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1271079949", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of chainCrit"},
   CellID->1271079949]},
 "ID:488309194"->{
  Cell[937536, 26790, 2164, 64, 72, "FormalTextInputFormula",
   CellTags->{
    "ID:488309194", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit w.r.t. singleton"},
   CellID->488309194]},
 "chainCrit w.r.t. singleton"->{
  Cell[937536, 26790, 2164, 64, 72, "FormalTextInputFormula",
   CellTags->{
    "ID:488309194", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit w.r.t. singleton"},
   CellID->488309194]},
 "ID:1322351588"->{
  Cell[939703, 26856, 1261, 37, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1322351588", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a 1"},
   CellID->1322351588]},
 "chainCrit\:293a 1"->{
  Cell[939703, 26856, 1261, 37, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1322351588", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a 1"},
   CellID->1322351588]},
 "ID:1269486125"->{
  Cell[940967, 26895, 1262, 37, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1269486125", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a 2"},
   CellID->1269486125]},
 "chainCrit\:293a 2"->{
  Cell[940967, 26895, 1262, 37, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1269486125", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a 2"},
   CellID->1269486125]},
 "ID:941129646"->{
  Cell[942232, 26934, 1467, 42, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:941129646", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a"},
   CellID->941129646]},
 "chainCrit\:293a"->{
  Cell[942232, 26934, 1467, 42, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:941129646", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a"},
   CellID->941129646]},
 "ID:1657568255"->{
  Cell[943702, 26978, 1164, 33, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1657568255", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a I"},
   CellID->1657568255]},
 "chainCrit\:293a I"->{
  Cell[943702, 26978, 1164, 33, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1657568255", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a I"},
   CellID->1657568255]},
 "Proof|ID:1452963633|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[945022, 27022, 3208, 87, 29, "ProofInfo",
   CellTags->
    "Proof|ID:1452963633|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->1783184226]},
 "Proof|ID:1500097429|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[948651, 27126, 3400, 91, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1500097429|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->1859160805]},
 "Proof|ID:279915377|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[952472, 27234, 3485, 93, 27, "ProofInfo",
   CellTags->
    "Proof|ID:279915377|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->1592316742]},
 "Proof|ID:1391995733|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[956376, 27344, 3669, 97, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1391995733|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->218343730]},
 "Proof|ID:1271079949"->{
  Cell[960464, 27458, 2703, 76, 27, "ProofInfo",
   CellTags->{"Proof|ID:1271079949", "Proof|ID:1271079949-1"},
   CellID->1431739973]},
 "Proof|ID:1271079949-1"->{
  Cell[960464, 27458, 2703, 76, 27, "ProofInfo",
   CellTags->{"Proof|ID:1271079949", "Proof|ID:1271079949-1"},
   CellID->1431739973]},
 "Proof|ID:488309194|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[963733, 27554, 6438, 158, 27, "ProofInfo",
   CellTags->
    "Proof|ID:488309194|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->1697069805]},
 "Proof|ID:1322351588|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[970591, 27729, 4612, 115, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1322351588|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->1376405718]},
 "Proof|ID:1269486125|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[975624, 27861, 4607, 115, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1269486125|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->779731268]},
 "Proof|ID:941129646|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[980652, 27993, 4762, 119, 27, "ProofInfo",
   CellTags->
    "Proof|ID:941129646|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->581146426]},
 "Proof|ID:1657568255|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[985833, 28129, 3666, 97, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1657568255|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->1717614685]},
 "ID:469812703"->{
  Cell[993412, 28333, 1538, 43, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:469812703", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "30"},
   CellID->469812703]},
 "30"->{
  Cell[993412, 28333, 1538, 43, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:469812703", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "30"},
   CellID->469812703]},
 "ID:1053941118"->{
  Cell[994953, 28378, 1240, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1053941118", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit holds for some element"},
   CellID->1053941118]},
 "chainCrit holds for some element"->{
  Cell[994953, 28378, 1240, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1053941118", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit holds for some element"},
   CellID->1053941118]},
 "Proof|ID:469812703|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[996435, 28421, 6203, 155, 29, "ProofInfo",
   CellTags->
    "Proof|ID:469812703|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->1066023196]},
 "Proof|ID:1053941118|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[1003108, 28595, 5355, 136, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1053941118|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->875669745]},
 "ID:1635347335"->{
  Cell[1013396, 28870, 1529, 39, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1635347335", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "31"},
   CellID->1635347335]},
 "31"->{
  Cell[1013396, 28870, 1529, 39, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1635347335", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "31"},
   CellID->1635347335]},
 "ID:1901113035"->{
  Cell[1014928, 28911, 1279, 36, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1901113035", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "32"},
   CellID->1901113035]},
 "32"->{
  Cell[1014928, 28911, 1279, 36, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1901113035", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "32"},
   CellID->1901113035]},
 "ID:1760547878"->{
  Cell[1016210, 28949, 919, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1760547878", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M (binary)"},
   CellID->1760547878]},
 "type of M (binary)"->{
  Cell[1016210, 28949, 919, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1760547878", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M (binary)"},
   CellID->1760547878]},
 "ID:1896797254"->{
  Cell[1017132, 28976, 892, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1896797254", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M (unary)"},
   CellID->1896797254]},
 "type of M (unary)"->{
  Cell[1017132, 28976, 892, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1896797254", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M (unary)"},
   CellID->1896797254]},
 "ID:1611647919"->{
  Cell[1018027, 29002, 1131, 31, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1611647919", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "33"},
   CellID->1611647919]},
 "33"->{
  Cell[1018027, 29002, 1131, 31, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1611647919", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "33"},
   CellID->1611647919]},
 "Proof|ID:1635347335|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[1019315, 29044, 7097, 173, 29, "ProofInfo",
   CellTags->
    "Proof|ID:1635347335|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->786500781]},
 "Proof|ID:1901113035|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[1026810, 29234, 6895, 169, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1901113035|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->573625489]},
 "Proof|ID:1760547878|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[1034104, 29420, 3745, 101, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1760547878|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->177141490]},
 "Proof|ID:1896797254|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[1038247, 29538, 3212, 89, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1896797254|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->367799976]},
 "Proof|ID:1611647919|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[1041906, 29645, 5056, 131, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1611647919|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->2107085874]},
 "ID:1545946275"->{
  Cell[1049345, 29860, 1589, 42, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1545946275", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M (unary)"},
   CellID->1545946275]},
 "\[LessEqual]M (unary)"->{
  Cell[1049345, 29860, 1589, 42, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1545946275", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M (unary)"},
   CellID->1545946275]},
 "ID:1018591902"->{
  Cell[1050937, 29904, 1654, 43, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1018591902", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual] (unary)"},
   CellID->1018591902]},
 "M\[LessEqual] (unary)"->{
  Cell[1050937, 29904, 1654, 43, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1018591902", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual] (unary)"},
   CellID->1018591902]},
 "ID:382103250"->{
  Cell[1052594, 29949, 1747, 47, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:382103250", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M (binary)"},
   CellID->382103250]},
 "\[LessEqual]M (binary)"->{
  Cell[1052594, 29949, 1747, 47, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:382103250", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M (binary)"},
   CellID->382103250]},
 "ID:450457679"->{
  Cell[1054344, 29998, 1507, 44, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:450457679", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual] (binary)"},
   CellID->450457679]},
 "M\[LessEqual] (binary)"->{
  Cell[1054344, 29998, 1507, 44, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:450457679", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual] (binary)"},
   CellID->450457679]},
 "ID:643714041"->{
  Cell[1055854, 30044, 1966, 61, 57, "FormalTextInputFormula",
   CellTags->{
    "ID:643714041", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M+\[LessEqual]"},
   CellID->643714041]},
 "M+\[LessEqual]"->{
  Cell[1055854, 30044, 1966, 61, 57, "FormalTextInputFormula",
   CellTags->{
    "ID:643714041", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M+\[LessEqual]"},
   CellID->643714041]},
 "Proof|ID:1545946275"->{
  Cell[1057977, 30116, 6242, 158, 29, "ProofInfo",
   CellTags->{"Proof|ID:1545946275", "Proof|ID:1545946275-1"},
   CellID->1890851245]},
 "Proof|ID:1545946275-1"->{
  Cell[1057977, 30116, 6242, 158, 29, "ProofInfo",
   CellTags->{"Proof|ID:1545946275", "Proof|ID:1545946275-1"},
   CellID->1890851245]},
 "Proof|ID:1018591902"->{
  Cell[1064806, 30294, 6231, 158, 27, "ProofInfo",
   CellTags->{"Proof|ID:1018591902", "Proof|ID:1018591902-1"},
   CellID->1462209435]},
 "Proof|ID:1018591902-1"->{
  Cell[1064806, 30294, 6231, 158, 27, "ProofInfo",
   CellTags->{"Proof|ID:1018591902", "Proof|ID:1018591902-1"},
   CellID->1462209435]},
 "Proof|ID:382103250"->{
  Cell[1071602, 30472, 8080, 196, 27, "ProofInfo",
   CellTags->{"Proof|ID:382103250", "Proof|ID:382103250-1"},
   CellID->280851943]},
 "Proof|ID:382103250-1"->{
  Cell[1071602, 30472, 8080, 196, 27, "ProofInfo",
   CellTags->{"Proof|ID:382103250", "Proof|ID:382103250-1"},
   CellID->280851943]},
 "Proof|ID:450457679"->{
  Cell[1080269, 30688, 8071, 196, 27, "ProofInfo",
   CellTags->{"Proof|ID:450457679", "Proof|ID:450457679-1"},
   CellID->478833348]},
 "Proof|ID:450457679-1"->{
  Cell[1080269, 30688, 8071, 196, 27, "ProofInfo",
   CellTags->{"Proof|ID:450457679", "Proof|ID:450457679-1"},
   CellID->478833348]},
 "Proof|ID:643714041"->{
  Cell[1088926, 30904, 8841, 212, 27, "ProofInfo",
   CellTags->{"Proof|ID:643714041", "Proof|ID:643714041-1"},
   CellID->105955449]},
 "Proof|ID:643714041-1"->{
  Cell[1088926, 30904, 8841, 212, 27, "ProofInfo",
   CellTags->{"Proof|ID:643714041", "Proof|ID:643714041-1"},
   CellID->105955449]},
 "ID:1934562955"->{
  Cell[1100282, 31190, 1074, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1934562955", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "34"},
   CellID->1934562955]},
 "34"->{
  Cell[1100282, 31190, 1074, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1934562955", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "34"},
   CellID->1934562955]},
 "ID:1993966556"->{
  Cell[1101359, 31222, 1171, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1993966556", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "35"},
   CellID->1993966556]},
 "35"->{
  Cell[1101359, 31222, 1171, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1993966556", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "35"},
   CellID->1993966556]},
 "ID:1265284403"->{
  Cell[1102533, 31256, 1095, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1265284403", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "36"},
   CellID->1265284403]},
 "36"->{
  Cell[1102533, 31256, 1095, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1265284403", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "36"},
   CellID->1265284403]},
 "ID:934408883"->{
  Cell[1103631, 31288, 1063, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:934408883", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "37"},
   CellID->934408883]},
 "37"->{
  Cell[1103631, 31288, 1063, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:934408883", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "37"},
   CellID->934408883]},
 "ID:1342720412"->{
  Cell[1104697, 31320, 1826, 53, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:1342720412", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "38"},
   CellID->1342720412]},
 "38"->{
  Cell[1104697, 31320, 1826, 53, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:1342720412", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "38"},
   CellID->1342720412]},
 "ID:192359742"->{
  Cell[1106873, 31383, 1204, 33, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:192359742", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "39"},
   CellID->192359742]},
 "39"->{
  Cell[1106873, 31383, 1204, 33, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:192359742", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "39"},
   CellID->192359742]},
 "ID:2005322529"->{
  Cell[1108080, 31418, 1677, 49, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:2005322529", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "40"},
   CellID->2005322529]},
 "40"->{
  Cell[1108080, 31418, 1677, 49, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:2005322529", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "40"},
   CellID->2005322529]},
 "Proof|ID:1934562955|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[1109914, 31478, 4465, 117, 29, "ProofInfo",
   CellTags->
    "Proof|ID:1934562955|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->721371033]},
 "Proof|ID:1993966556|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[1114800, 31612, 4369, 115, 27, "ProofInfo",
   CellTags->
    "Proof|ID:1993966556|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->899634188]},
 "Proof|ID:1265284403"->{
  Cell[1119567, 31744, 4928, 126, 27, "ProofInfo",
   CellTags->{"Proof|ID:1265284403", "Proof|ID:1265284403-1"},
   CellID->612613216]},
 "Proof|ID:1265284403-1"->{
  Cell[1119567, 31744, 4928, 126, 27, "ProofInfo",
   CellTags->{"Proof|ID:1265284403", "Proof|ID:1265284403-1"},
   CellID->612613216]},
 "Proof|ID:934408883"->{
  Cell[1125019, 31889, 4721, 122, 27, "ProofInfo",
   CellTags->{"Proof|ID:934408883", "Proof|ID:934408883-1"},
   CellID->716604140]},
 "Proof|ID:934408883-1"->{
  Cell[1125019, 31889, 4721, 122, 27, "ProofInfo",
   CellTags->{"Proof|ID:934408883", "Proof|ID:934408883-1"},
   CellID->716604140]},
 "Proof|ID:192359742"->{
  Cell[1130263, 32030, 5838, 148, 27, "ProofInfo",
   CellTags->{"Proof|ID:192359742", "Proof|ID:192359742-1"},
   CellID->89073571]},
 "Proof|ID:192359742-1"->{
  Cell[1130263, 32030, 5838, 148, 27, "ProofInfo",
   CellTags->{"Proof|ID:192359742", "Proof|ID:192359742-1"},
   CellID->89073571]},
 "Proof|ID:2005322529"->{
  Cell[1136624, 32197, 7086, 174, 27, "ProofInfo",
   CellTags->{"Proof|ID:2005322529", "Proof|ID:2005322529-1"},
   CellID->592718415]},
 "Proof|ID:2005322529-1"->{
  Cell[1136624, 32197, 7086, 174, 27, "ProofInfo",
   CellTags->{"Proof|ID:2005322529", "Proof|ID:2005322529-1"},
   CellID->592718415]},
 "ID:984666222"->{
  Cell[1145936, 32447, 1160, 33, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:984666222", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "41"},
   CellID->984666222]},
 "41"->{
  Cell[1145936, 32447, 1160, 33, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:984666222", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "41"},
   CellID->984666222]},
 "ID:1616352426"->{
  Cell[1147099, 32482, 1321, 38, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:1616352426", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "42"},
   CellID->1616352426]},
 "42"->{
  Cell[1147099, 32482, 1321, 38, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:1616352426", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "42"},
   CellID->1616352426]},
 "Proof|ID:984666222"->{
  Cell[1148576, 32531, 6994, 174, 29, "ProofInfo",
   CellTags->{"Proof|ID:984666222", "Proof|ID:984666222-1"},
   CellID->740590150]},
 "Proof|ID:984666222-1"->{
  Cell[1148576, 32531, 6994, 174, 29, "ProofInfo",
   CellTags->{"Proof|ID:984666222", "Proof|ID:984666222-1"},
   CellID->740590150]},
 "Proof|ID:1616352426"->{
  Cell[1156092, 32724, 8354, 203, 27, "ProofInfo",
   CellTags->{"Proof|ID:1616352426", "Proof|ID:1616352426-1"},
   CellID->1148420320]},
 "Proof|ID:1616352426-1"->{
  Cell[1156092, 32724, 8354, 203, 27, "ProofInfo",
   CellTags->{"Proof|ID:1616352426", "Proof|ID:1616352426-1"},
   CellID->1148420320]},
 "ID:1399613937"->{
  Cell[1166853, 33005, 1768, 49, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1399613937", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 1"},
   CellID->1399613937]},
 "M\:293a 1"->{
  Cell[1166853, 33005, 1768, 49, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1399613937", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 1"},
   CellID->1399613937]},
 "ID:1210171526"->{
  Cell[1168624, 33056, 1552, 46, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1210171526", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 2"},
   CellID->1210171526]},
 "M\:293a 2"->{
  Cell[1168624, 33056, 1552, 46, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1210171526", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 2"},
   CellID->1210171526]},
 "ID:1251355025"->{
  Cell[1170179, 33104, 1104, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1251355025", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 3"},
   CellID->1251355025]},
 "M\:293a 3"->{
  Cell[1170179, 33104, 1104, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1251355025", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 3"},
   CellID->1251355025]},
 "ID:1093640190"->{
  Cell[1171286, 33140, 1223, 37, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1093640190", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 4"},
   CellID->1093640190]},
 "M\:293a 4"->{
  Cell[1171286, 33140, 1223, 37, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1093640190", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 4"},
   CellID->1093640190]},
 "ID:606212147"->{
  Cell[1172512, 33179, 1700, 52, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:606212147", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 5"},
   CellID->606212147]},
 "M\:293a 5"->{
  Cell[1172512, 33179, 1700, 52, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:606212147", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 5"},
   CellID->606212147]},
 "ID:1628665526"->{
  Cell[1174215, 33233, 1521, 43, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1628665526", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a"},
   CellID->1628665526]},
 "M\:293a"->{
  Cell[1174215, 33233, 1521, 43, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1628665526", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a"},
   CellID->1628665526]},
 "Proof|ID:1399613937"->{
  Cell[1175892, 33287, 6632, 165, 29, "ProofInfo",
   CellTags->{"Proof|ID:1399613937", "Proof|ID:1399613937-1"},
   CellID->1391233604]},
 "Proof|ID:1399613937-1"->{
  Cell[1175892, 33287, 6632, 165, 29, "ProofInfo",
   CellTags->{"Proof|ID:1399613937", "Proof|ID:1399613937-1"},
   CellID->1391233604]},
 "Proof|ID:1210171526"->{
  Cell[1183047, 33471, 7199, 176, 27, "ProofInfo",
   CellTags->{"Proof|ID:1210171526", "Proof|ID:1210171526-1"},
   CellID->2096401104]},
 "Proof|ID:1210171526-1"->{
  Cell[1183047, 33471, 7199, 176, 27, "ProofInfo",
   CellTags->{"Proof|ID:1210171526", "Proof|ID:1210171526-1"},
   CellID->2096401104]},
 "Proof|ID:1251355025"->{
  Cell[1190769, 33666, 4940, 128, 27, "ProofInfo",
   CellTags->{"Proof|ID:1251355025", "Proof|ID:1251355025-1"},
   CellID->185930261]},
 "Proof|ID:1251355025-1"->{
  Cell[1190769, 33666, 4940, 128, 27, "ProofInfo",
   CellTags->{"Proof|ID:1251355025", "Proof|ID:1251355025-1"},
   CellID->185930261]},
 "Proof|ID:1093640190"->{
  Cell[1196233, 33813, 5208, 132, 27, "ProofInfo",
   CellTags->{"Proof|ID:1093640190", "Proof|ID:1093640190-1"},
   CellID->808217224]},
 "Proof|ID:1093640190-1"->{
  Cell[1196233, 33813, 5208, 132, 27, "ProofInfo",
   CellTags->{"Proof|ID:1093640190", "Proof|ID:1093640190-1"},
   CellID->808217224]},
 "Proof|ID:606212147"->{
  Cell[1201964, 33964, 7292, 178, 27, "ProofInfo",
   CellTags->{"Proof|ID:606212147", "Proof|ID:606212147-1"},
   CellID->1011788619]},
 "Proof|ID:606212147-1"->{
  Cell[1201964, 33964, 7292, 178, 27, "ProofInfo",
   CellTags->{"Proof|ID:606212147", "Proof|ID:606212147-1"},
   CellID->1011788619]},
 "Proof|ID:1628665526"->{
  Cell[1209778, 34161, 5016, 128, 27, "ProofInfo",
   CellTags->{"Proof|ID:1628665526", "Proof|ID:1628665526-1"},
   CellID->975115047]},
 "Proof|ID:1628665526-1"->{
  Cell[1209778, 34161, 5016, 128, 27, "ProofInfo",
   CellTags->{"Proof|ID:1628665526", "Proof|ID:1628665526-1"},
   CellID->975115047]},
 "ID:1351181837"->{
  Cell[1216960, 34365, 1412, 39, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1351181837", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a I"},
   CellID->1351181837]},
 "M\:293a I"->{
  Cell[1216960, 34365, 1412, 39, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1351181837", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a I"},
   CellID->1351181837]},
 "ID:2057269766"->{
  Cell[1218375, 34406, 1446, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:2057269766", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a\[LessEqual]"},
   CellID->2057269766]},
 "M\:293a\[LessEqual]"->{
  Cell[1218375, 34406, 1446, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:2057269766", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a\[LessEqual]"},
   CellID->2057269766]},
 "ID:94917498"->{
  Cell[1219824, 34452, 1427, 42, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:94917498", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a II"},
   CellID->94917498]},
 "M\:293a II"->{
  Cell[1219824, 34452, 1427, 42, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:94917498", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a II"},
   CellID->94917498]},
 "ID:1065272829"->{
  Cell[1221254, 34496, 1596, 51, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1065272829", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a+\[LessEqual]"},
   CellID->1065272829]},
 "M\:293a+\[LessEqual]"->{
  Cell[1221254, 34496, 1596, 51, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1065272829", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a+\[LessEqual]"},
   CellID->1065272829]},
 "Proof|ID:1351181837"->{
  Cell[1223007, 34558, 6071, 152, 29, "ProofInfo",
   CellTags->{"Proof|ID:1351181837", "Proof|ID:1351181837-1"},
   CellID->1829706754]},
 "Proof|ID:1351181837-1"->{
  Cell[1223007, 34558, 6071, 152, 29, "ProofInfo",
   CellTags->{"Proof|ID:1351181837", "Proof|ID:1351181837-1"},
   CellID->1829706754]},
 "Proof|ID:2057269766"->{
  Cell[1229643, 34730, 6320, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:2057269766", "Proof|ID:2057269766-1"},
   CellID->706843964]},
 "Proof|ID:2057269766-1"->{
  Cell[1229643, 34730, 6320, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:2057269766", "Proof|ID:2057269766-1"},
   CellID->706843964]},
 "Proof|ID:94917498"->{
  Cell[1236550, 34906, 6919, 170, 27, "ProofInfo",
   CellTags->{"Proof|ID:94917498", "Proof|ID:94917498-1"},
   CellID->1107419424]},
 "Proof|ID:94917498-1"->{
  Cell[1236550, 34906, 6919, 170, 27, "ProofInfo",
   CellTags->{"Proof|ID:94917498", "Proof|ID:94917498-1"},
   CellID->1107419424]},
 "Proof|ID:1065272829"->{
  Cell[1243990, 35095, 7449, 180, 27, "ProofInfo",
   CellTags->{"Proof|ID:1065272829", "Proof|ID:1065272829-1"},
   CellID->1792160777]},
 "Proof|ID:1065272829-1"->{
  Cell[1243990, 35095, 7449, 180, 27, "ProofInfo",
   CellTags->{"Proof|ID:1065272829", "Proof|ID:1065272829-1"},
   CellID->1792160777]},
 "ID:1824645721"->{
  Cell[1254008, 35360, 943, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1824645721", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of W (binary)"},
   CellID->1824645721]},
 "type of W (binary)"->{
  Cell[1254008, 35360, 943, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1824645721", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of W (binary)"},
   CellID->1824645721]},
 "ID:1345307293"->{
  Cell[1254954, 35387, 893, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1345307293", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of W (unary)"},
   CellID->1345307293]},
 "type of W (unary)"->{
  Cell[1254954, 35387, 893, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1345307293", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of W (unary)"},
   CellID->1345307293]},
 "ID:1185705395"->{
  Cell[1255850, 35413, 1201, 33, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1185705395", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "43"},
   CellID->1185705395]},
 "43"->{
  Cell[1255850, 35413, 1201, 33, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1185705395", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "43"},
   CellID->1185705395]},
 "Proof|ID:1824645721"->{
  Cell[1257209, 35457, 3912, 106, 29, "ProofInfo",
   CellTags->{"Proof|ID:1824645721", "Proof|ID:1824645721-1"},
   CellID->286621138]},
 "Proof|ID:1824645721-1"->{
  Cell[1257209, 35457, 3912, 106, 29, "ProofInfo",
   CellTags->{"Proof|ID:1824645721", "Proof|ID:1824645721-1"},
   CellID->286621138]},
 "Proof|ID:1345307293"->{
  Cell[1261645, 35582, 3167, 88, 27, "ProofInfo",
   CellTags->{"Proof|ID:1345307293", "Proof|ID:1345307293-1"},
   CellID->28034114]},
 "Proof|ID:1345307293-1"->{
  Cell[1261645, 35582, 3167, 88, 27, "ProofInfo",
   CellTags->{"Proof|ID:1345307293", "Proof|ID:1345307293-1"},
   CellID->28034114]},
 "Proof|ID:1185705395"->{
  Cell[1265336, 35689, 6061, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:1185705395", "Proof|ID:1185705395-1"},
   CellID->2067392509]},
 "Proof|ID:1185705395-1"->{
  Cell[1265336, 35689, 6061, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:1185705395", "Proof|ID:1185705395-1"},
   CellID->2067392509]},
 "ID:354042430"->{
  Cell[1274630, 35950, 1549, 46, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:354042430", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]W"},
   CellID->354042430]},
 "\[LessEqual]W"->{
  Cell[1274630, 35950, 1549, 46, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:354042430", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]W"},
   CellID->354042430]},
 "ID:551974343"->{
  Cell[1276182, 35998, 1538, 45, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:551974343", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\[LessEqual]"},
   CellID->551974343]},
 "W\[LessEqual]"->{
  Cell[1276182, 35998, 1538, 45, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:551974343", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\[LessEqual]"},
   CellID->551974343]},
 "Proof|ID:354042430"->{
  Cell[1277877, 36054, 7186, 180, 29, "ProofInfo",
   CellTags->{"Proof|ID:354042430", "Proof|ID:354042430-1"},
   CellID->1517092650]},
 "Proof|ID:354042430-1"->{
  Cell[1277877, 36054, 7186, 180, 29, "ProofInfo",
   CellTags->{"Proof|ID:354042430", "Proof|ID:354042430-1"},
   CellID->1517092650]},
 "Proof|ID:551974343"->{
  Cell[1285627, 36254, 7185, 180, 27, "ProofInfo",
   CellTags->{"Proof|ID:551974343", "Proof|ID:551974343-1"},
   CellID->766317937]},
 "Proof|ID:551974343-1"->{
  Cell[1285627, 36254, 7185, 180, 27, "ProofInfo",
   CellTags->{"Proof|ID:551974343", "Proof|ID:551974343-1"},
   CellID->766317937]},
 "ID:894623443"->{
  Cell[1295731, 36531, 965, 28, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:894623443", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "44"},
   CellID->894623443]},
 "44"->{
  Cell[1295731, 36531, 965, 28, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:894623443", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "44"},
   CellID->894623443]},
 "ID:1995654247"->{
  Cell[1296699, 36561, 1128, 32, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1995654247", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "45"},
   CellID->1995654247]},
 "45"->{
  Cell[1296699, 36561, 1128, 32, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1995654247", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "45"},
   CellID->1995654247]},
 "ID:1916087968"->{
  Cell[1297830, 36595, 1240, 34, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1916087968", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "46"},
   CellID->1916087968]},
 "46"->{
  Cell[1297830, 36595, 1240, 34, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1916087968", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "46"},
   CellID->1916087968]},
 "Proof|ID:894623443"->{
  Cell[1299227, 36640, 5465, 140, 29, "ProofInfo",
   CellTags->{"Proof|ID:894623443", "Proof|ID:894623443-1"},
   CellID->456641578]},
 "Proof|ID:894623443-1"->{
  Cell[1299227, 36640, 5465, 140, 29, "ProofInfo",
   CellTags->{"Proof|ID:894623443", "Proof|ID:894623443-1"},
   CellID->456641578]},
 "Proof|ID:1995654247"->{
  Cell[1305256, 36800, 6147, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:1995654247", "Proof|ID:1995654247-1"},
   CellID->1030159721]},
 "Proof|ID:1995654247-1"->{
  Cell[1305256, 36800, 6147, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:1995654247", "Proof|ID:1995654247-1"},
   CellID->1030159721]},
 "Proof|ID:1916087968"->{
  Cell[1311968, 36976, 6347, 160, 27, "ProofInfo",
   CellTags->{"Proof|ID:1916087968", "Proof|ID:1916087968-1"},
   CellID->1865547860]},
 "Proof|ID:1916087968-1"->{
  Cell[1311968, 36976, 6347, 160, 27, "ProofInfo",
   CellTags->{"Proof|ID:1916087968", "Proof|ID:1916087968-1"},
   CellID->1865547860]},
 "ID:2064198302"->{
  Cell[1321436, 37240, 1577, 41, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2064198302", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a"},
   CellID->2064198302]},
 "W\:293a"->{
  Cell[1321436, 37240, 1577, 41, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2064198302", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a"},
   CellID->2064198302]},
 "ID:2144277606"->{
  Cell[1323016, 37283, 948, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2144277606", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a I"},
   CellID->2144277606]},
 "W\:293a I"->{
  Cell[1323016, 37283, 948, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2144277606", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a I"},
   CellID->2144277606]},
 "ID:388422476"->{
  Cell[1323967, 37314, 1038, 30, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:388422476", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a II"},
   CellID->388422476]},
 "W\:293a II"->{
  Cell[1323967, 37314, 1038, 30, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:388422476", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a II"},
   CellID->388422476]},
 "ID:144808276"->{
  Cell[1325008, 37346, 1030, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:144808276", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a decreases"},
   CellID->144808276]},
 "W\:293a decreases"->{
  Cell[1325008, 37346, 1030, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:144808276", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a decreases"},
   CellID->144808276]},
 "Proof|ID:2064198302"->{
  Cell[1326194, 37386, 5903, 151, 29, "ProofInfo",
   CellTags->{"Proof|ID:2064198302", "Proof|ID:2064198302-1"},
   CellID->595904185]},
 "Proof|ID:2064198302-1"->{
  Cell[1326194, 37386, 5903, 151, 29, "ProofInfo",
   CellTags->{"Proof|ID:2064198302", "Proof|ID:2064198302-1"},
   CellID->595904185]},
 "Proof|ID:2144277606"->{
  Cell[1332620, 37556, 4933, 129, 27, "ProofInfo",
   CellTags->{"Proof|ID:2144277606", "Proof|ID:2144277606-1"},
   CellID->391021459]},
 "Proof|ID:2144277606-1"->{
  Cell[1332620, 37556, 4933, 129, 27, "ProofInfo",
   CellTags->{"Proof|ID:2144277606", "Proof|ID:2144277606-1"},
   CellID->391021459]},
 "Proof|ID:388422476"->{
  Cell[1338077, 37704, 5105, 134, 27, "ProofInfo",
   CellTags->{"Proof|ID:388422476", "Proof|ID:388422476-1"},
   CellID->743656091]},
 "Proof|ID:388422476-1"->{
  Cell[1338077, 37704, 5105, 134, 27, "ProofInfo",
   CellTags->{"Proof|ID:388422476", "Proof|ID:388422476-1"},
   CellID->743656091]},
 "Proof|ID:144808276"->{
  Cell[1343705, 37857, 3974, 106, 27, "ProofInfo",
   CellTags->{"Proof|ID:144808276", "Proof|ID:144808276-1"},
   CellID->1758271591]},
 "Proof|ID:144808276-1"->{
  Cell[1343705, 37857, 3974, 106, 27, "ProofInfo",
   CellTags->{"Proof|ID:144808276", "Proof|ID:144808276-1"},
   CellID->1758271591]},
 "ID:1149216477"->{
  Cell[1349301, 38020, 1074, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1149216477", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of L"},
   CellID->1149216477]},
 "type of L"->{
  Cell[1349301, 38020, 1074, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1149216477", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of L"},
   CellID->1149216477]},
 "Proof|ID:1149216477|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb"->{
  Cell[1350531, 38059, 3049, 85, 29, "ProofInfo",
   CellTags->
    "Proof|ID:1149216477|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
   CellID->187961443]},
 "sect::LemmataSpecificSpecial"->{
  Cell[1354001, 38163, 253, 8, 30, "Subsection",
   CellTags->"sect::LemmataSpecificSpecial",
   CellID->1466386606]},
 "ID:504163160"->{
  Cell[1355455, 38202, 322, 9, 21, "GlobalDeclaration",
   CellTags->{
    "ID:504163160", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->504163160]},
 "ID:649536122"->{
  Cell[1355819, 38216, 691, 18, 57, "GlobalDeclaration",
   CellTags->{
    "ID:649536122", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->649536122]},
 "sect::Cover"->{
  Cell[1356535, 38238, 363, 6, 28, "Subsubsection",
   CellTags->"sect::Cover",
   CellID->676471010]},
 "ID:465693853"->{
  Cell[1359150, 38314, 1011, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:465693853", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "47"},
   CellID->465693853]},
 "47"->{
  Cell[1359150, 38314, 1011, 30, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:465693853", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "47"},
   CellID->465693853]},
 "ID:777205621"->{
  Cell[1360164, 38346, 2424, 78, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:777205621", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "48"},
   CellID->777205621]},
 "48"->{
  Cell[1360164, 38346, 2424, 78, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:777205621", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "48"},
   CellID->777205621]},
 "ID:42909295"->{
  Cell[1362591, 38426, 1737, 51, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:42909295", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "49"},
   CellID->42909295]},
 "49"->{
  Cell[1362591, 38426, 1737, 51, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:42909295", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "49"},
   CellID->42909295]},
 "ID:112873953"->{
  Cell[1364331, 38479, 1942, 62, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:112873953", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "50"},
   CellID->112873953]},
 "50"->{
  Cell[1364331, 38479, 1942, 62, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:112873953", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "50"},
   CellID->112873953]},
 "Proof|ID:465693853"->{
  Cell[1366430, 38552, 4608, 122, 29, "ProofInfo",
   CellTags->{"Proof|ID:465693853", "Proof|ID:465693853-1"},
   CellID->2114883905]},
 "Proof|ID:465693853-1"->{
  Cell[1366430, 38552, 4608, 122, 29, "ProofInfo",
   CellTags->{"Proof|ID:465693853", "Proof|ID:465693853-1"},
   CellID->2114883905]},
 "Proof|ID:777205621"->{
  Cell[1371603, 38694, 15311, 356, 27, "ProofInfo",
   CellTags->{"Proof|ID:777205621", "Proof|ID:777205621-1"},
   CellID->118523572]},
 "Proof|ID:777205621-1"->{
  Cell[1371603, 38694, 15311, 356, 27, "ProofInfo",
   CellTags->{"Proof|ID:777205621", "Proof|ID:777205621-1"},
   CellID->118523572]},
 "Proof|ID:42909295"->{
  Cell[1387437, 39069, 10171, 244, 27, "ProofInfo",
   CellTags->{"Proof|ID:42909295", "Proof|ID:42909295-1"},
   CellID->234727221]},
 "Proof|ID:42909295-1"->{
  Cell[1387437, 39069, 10171, 244, 27, "ProofInfo",
   CellTags->{"Proof|ID:42909295", "Proof|ID:42909295-1"},
   CellID->234727221]},
 "Proof|ID:112873953"->{
  Cell[1398193, 39333, 11274, 268, 27, "ProofInfo",
   CellTags->{"Proof|ID:112873953", "Proof|ID:112873953-1"},
   CellID->334296225]},
 "Proof|ID:112873953-1"->{
  Cell[1398193, 39333, 11274, 268, 27, "ProofInfo",
   CellTags->{"Proof|ID:112873953", "Proof|ID:112873953-1"},
   CellID->334296225]},
 "ID:315092621"->{
  Cell[1410709, 39640, 1579, 46, 66, "FormalTextInputFormula",
   CellTags->{
    "ID:315092621", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "cover"},
   CellID->315092621]},
 "cover"->{
  Cell[1410709, 39640, 1579, 46, 66, "FormalTextInputFormula",
   CellTags->{
    "ID:315092621", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "cover"},
   CellID->315092621]},
 "Proof|ID:315092621"->{
  Cell[1412445, 39697, 5482, 138, 29, "ProofInfo",
   CellTags->{"Proof|ID:315092621", "Proof|ID:315092621-1"},
   CellID->875989985]},
 "Proof|ID:315092621-1"->{
  Cell[1412445, 39697, 5482, 138, 29, "ProofInfo",
   CellTags->{"Proof|ID:315092621", "Proof|ID:315092621-1"},
   CellID->875989985]},
 "ID:744007071"->{
  Cell[1424417, 40031, 1434, 44, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:744007071", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "51"},
   CellID->744007071]},
 "51"->{
  Cell[1424417, 40031, 1434, 44, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:744007071", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "51"},
   CellID->744007071]},
 "ID:1720661645"->{
  Cell[1425854, 40077, 1368, 43, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1720661645", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "52"},
   CellID->1720661645]},
 "52"->{
  Cell[1425854, 40077, 1368, 43, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1720661645", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "52"},
   CellID->1720661645]},
 "ID:1022330126"->{
  Cell[1427225, 40122, 1938, 62, 55, "FormalTextInputFormula",
   CellTags->{
    "ID:1022330126", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[VerticalBar]lcm"},
   CellID->1022330126]},
 "\[VerticalBar]lcm"->{
  Cell[1427225, 40122, 1938, 62, 55, "FormalTextInputFormula",
   CellTags->{
    "ID:1022330126", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[VerticalBar]lcm"},
   CellID->1022330126]},
 "Proof|ID:744007071"->{
  Cell[1429317, 40195, 6247, 158, 29, "ProofInfo",
   CellTags->{"Proof|ID:744007071", "Proof|ID:744007071-1"},
   CellID->1991259271]},
 "Proof|ID:744007071-1"->{
  Cell[1429317, 40195, 6247, 158, 29, "ProofInfo",
   CellTags->{"Proof|ID:744007071", "Proof|ID:744007071-1"},
   CellID->1991259271]},
 "Proof|ID:1720661645"->{
  Cell[1436129, 40373, 6251, 158, 27, "ProofInfo",
   CellTags->{"Proof|ID:1720661645", "Proof|ID:1720661645-1"},
   CellID->491886609]},
 "Proof|ID:1720661645-1"->{
  Cell[1436129, 40373, 6251, 158, 27, "ProofInfo",
   CellTags->{"Proof|ID:1720661645", "Proof|ID:1720661645-1"},
   CellID->491886609]},
 "Proof|ID:1022330126"->{
  Cell[1442945, 40551, 6872, 170, 27, "ProofInfo",
   CellTags->{"Proof|ID:1022330126", "Proof|ID:1022330126-1"},
   CellID->448824604]},
 "Proof|ID:1022330126-1"->{
  Cell[1442945, 40551, 6872, 170, 27, "ProofInfo",
   CellTags->{"Proof|ID:1022330126", "Proof|ID:1022330126-1"},
   CellID->448824604]},
 "ID:528222346"->{
  Cell[1451346, 40772, 1221, 38, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:528222346", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "53"},
   CellID->528222346]},
 "53"->{
  Cell[1451346, 40772, 1221, 38, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:528222346", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "53"},
   CellID->528222346]},
 "ID:1598627200"->{
  Cell[1452570, 40812, 1083, 33, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1598627200", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "54"},
   CellID->1598627200]},
 "54"->{
  Cell[1452570, 40812, 1083, 33, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1598627200", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "54"},
   CellID->1598627200]},
 "Proof|ID:528222346"->{
  Cell[1453809, 40856, 4594, 124, 29, "ProofInfo",
   CellTags->{"Proof|ID:528222346", "Proof|ID:528222346-1"},
   CellID->1476471173]},
 "Proof|ID:528222346-1"->{
  Cell[1453809, 40856, 4594, 124, 29, "ProofInfo",
   CellTags->{"Proof|ID:528222346", "Proof|ID:528222346-1"},
   CellID->1476471173]},
 "Proof|ID:1598627200"->{
  Cell[1459041, 41002, 3129, 90, 27, "ProofInfo",
   CellTags->{"Proof|ID:1598627200", "Proof|ID:1598627200-1"},
   CellID->605461210]},
 "Proof|ID:1598627200-1"->{
  Cell[1459041, 41002, 3129, 90, 27, "ProofInfo",
   CellTags->{"Proof|ID:1598627200", "Proof|ID:1598627200-1"},
   CellID->605461210]},
 "ID:1537560967"->{
  Cell[1463450, 41134, 1434, 46, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1537560967", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "55"},
   CellID->1537560967]},
 "55"->{
  Cell[1463450, 41134, 1434, 46, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1537560967", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "55"},
   CellID->1537560967]},
 "ID:507460317"->{
  Cell[1464887, 41182, 1431, 46, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:507460317", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "56"},
   CellID->507460317]},
 "56"->{
  Cell[1464887, 41182, 1431, 46, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:507460317", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "56"},
   CellID->507460317]},
 "Proof|ID:1537560967"->{
  Cell[1466475, 41239, 7077, 174, 29, "ProofInfo",
   CellTags->{"Proof|ID:1537560967", "Proof|ID:1537560967-1"},
   CellID->1213329230]},
 "Proof|ID:1537560967-1"->{
  Cell[1466475, 41239, 7077, 174, 29, "ProofInfo",
   CellTags->{"Proof|ID:1537560967", "Proof|ID:1537560967-1"},
   CellID->1213329230]},
 "Proof|ID:507460317"->{
  Cell[1474118, 41433, 7071, 174, 27, "ProofInfo",
   CellTags->{"Proof|ID:507460317", "Proof|ID:507460317-1"},
   CellID->688346270]},
 "Proof|ID:507460317-1"->{
  Cell[1474118, 41433, 7071, 174, 27, "ProofInfo",
   CellTags->{"Proof|ID:507460317", "Proof|ID:507460317-1"},
   CellID->688346270]},
 "sect::LemmataSpecificSpecialW"->{
  Cell[1481766, 41628, 354, 6, 22, "Subsubsection",
   CellTags->"sect::LemmataSpecificSpecialW",
   CellID->1209432111]},
 "Lem::W[A<-x]"->{
  Cell[1483631, 41683, 869, 27, 41, "EnvironmentHeader",
   CellTags->"Lem::W[A<-x]",
   CellID->1283608134]},
 "ID:1442383873"->{
  Cell[1485178, 41731, 859, 26, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1442383873", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "57"},
   CellID->1442383873]},
 "57"->{
  Cell[1485178, 41731, 859, 26, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1442383873", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "57"},
   CellID->1442383873]},
 "ID:523824248"->{
  Cell[1486040, 41759, 1384, 33, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:523824248", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (binary) FarExterior 1"},
   CellID->523824248]},
 "W\:293a (binary) FarExterior 1"->{
  Cell[1486040, 41759, 1384, 33, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:523824248", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (binary) FarExterior 1"},
   CellID->523824248]},
 "ID:1663623856"->{
  Cell[1487427, 41794, 1326, 34, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1663623856", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (binary) FarExterior 2"},
   CellID->1663623856]},
 "W\:293a (binary) FarExterior 2"->{
  Cell[1487427, 41794, 1326, 34, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1663623856", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (binary) FarExterior 2"},
   CellID->1663623856]},
 "Proof|ID:1442383873"->{
  Cell[1488909, 41839, 6373, 162, 29, "ProofInfo",
   CellTags->{"Proof|ID:1442383873", "Proof|ID:1442383873-1"},
   CellID->1192136911]},
 "Proof|ID:1442383873-1"->{
  Cell[1488909, 41839, 6373, 162, 29, "ProofInfo",
   CellTags->{"Proof|ID:1442383873", "Proof|ID:1442383873-1"},
   CellID->1192136911]},
 "Proof|ID:523824248"->{
  Cell[1495869, 42021, 5885, 151, 27, "ProofInfo",
   CellTags->{"Proof|ID:523824248", "Proof|ID:523824248-1"},
   CellID->842417444]},
 "Proof|ID:523824248-1"->{
  Cell[1495869, 42021, 5885, 151, 27, "ProofInfo",
   CellTags->{"Proof|ID:523824248", "Proof|ID:523824248-1"},
   CellID->842417444]},
 "Proof|ID:1663623856"->{
  Cell[1502276, 42191, 6395, 160, 27, "ProofInfo",
   CellTags->{"Proof|ID:1663623856", "Proof|ID:1663623856-1"},
   CellID->658477440]},
 "Proof|ID:1663623856-1"->{
  Cell[1502276, 42191, 6395, 160, 27, "ProofInfo",
   CellTags->{"Proof|ID:1663623856", "Proof|ID:1663623856-1"},
   CellID->658477440]},
 "ID:1829996465"->{
  Cell[1510194, 42406, 1579, 39, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1829996465", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (unary) FarExterior 1"},
   CellID->1829996465]},
 "W\:293a (unary) FarExterior 1"->{
  Cell[1510194, 42406, 1579, 39, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1829996465", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (unary) FarExterior 1"},
   CellID->1829996465]},
 "ID:842592091"->{
  Cell[1511776, 42447, 1598, 40, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:842592091", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (unary) FarExterior 2"},
   CellID->842592091]},
 "W\:293a (unary) FarExterior 2"->{
  Cell[1511776, 42447, 1598, 40, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:842592091", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (unary) FarExterior 2"},
   CellID->842592091]},
 "ID:1051080023"->{
  Cell[1513377, 42489, 1269, 38, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1051080023", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a FarExterior"},
   CellID->1051080023]},
 "W\:293a FarExterior"->{
  Cell[1513377, 42489, 1269, 38, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1051080023", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a FarExterior"},
   CellID->1051080023]},
 "Proof|ID:1829996465"->{
  Cell[1514802, 42538, 5411, 140, 29, "ProofInfo",
   CellTags->{"Proof|ID:1829996465", "Proof|ID:1829996465-1"},
   CellID->590566551]},
 "Proof|ID:1829996465-1"->{
  Cell[1514802, 42538, 5411, 140, 29, "ProofInfo",
   CellTags->{"Proof|ID:1829996465", "Proof|ID:1829996465-1"},
   CellID->590566551]},
 "Proof|ID:842592091"->{
  Cell[1520779, 42698, 5397, 140, 27, "ProofInfo",
   CellTags->{"Proof|ID:842592091", "Proof|ID:842592091-1"},
   CellID->917529032]},
 "Proof|ID:842592091-1"->{
  Cell[1520779, 42698, 5397, 140, 27, "ProofInfo",
   CellTags->{"Proof|ID:842592091", "Proof|ID:842592091-1"},
   CellID->917529032]},
 "Proof|ID:1051080023"->{
  Cell[1526740, 42858, 6595, 166, 27, "ProofInfo",
   CellTags->{"Proof|ID:1051080023", "Proof|ID:1051080023-1"},
   CellID->148396000]},
 "Proof|ID:1051080023-1"->{
  Cell[1526740, 42858, 6595, 166, 27, "ProofInfo",
   CellTags->{"Proof|ID:1051080023", "Proof|ID:1051080023-1"},
   CellID->148396000]},
 "ID:631716597"->{
  Cell[1535244, 43091, 997, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:631716597", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "58"},
   CellID->631716597]},
 "58"->{
  Cell[1535244, 43091, 997, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:631716597", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "58"},
   CellID->631716597]},
 "ID:1988188117"->{
  Cell[1536244, 43122, 980, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1988188117", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "59"},
   CellID->1988188117]},
 "59"->{
  Cell[1536244, 43122, 980, 29, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1988188117", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "59"},
   CellID->1988188117]},
 "Proof|ID:631716597"->{
  Cell[1537380, 43162, 5061, 133, 29, "ProofInfo",
   CellTags->{"Proof|ID:631716597", "Proof|ID:631716597-1"},
   CellID->1960584677]},
 "Proof|ID:631716597-1"->{
  Cell[1537380, 43162, 5061, 133, 29, "ProofInfo",
   CellTags->{"Proof|ID:631716597", "Proof|ID:631716597-1"},
   CellID->1960584677]},
 "Proof|ID:1988188117"->{
  Cell[1543006, 43315, 4899, 128, 27, "ProofInfo",
   CellTags->{"Proof|ID:1988188117", "Proof|ID:1988188117-1"},
   CellID->1726608150]},
 "Proof|ID:1988188117-1"->{
  Cell[1543006, 43315, 4899, 128, 27, "ProofInfo",
   CellTags->{"Proof|ID:1988188117", "Proof|ID:1988188117-1"},
   CellID->1726608150]},
 "ID:590870546"->{
  Cell[1549691, 43508, 399, 12, 41, "GlobalDeclaration",
   CellTags->{
    "ID:590870546", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
   CellID->590870546]},
 "ID:501569980"->{
  Cell[1550093, 43522, 1281, 38, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:501569980", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "60"},
   CellID->501569980]},
 "60"->{
  Cell[1550093, 43522, 1281, 38, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:501569980", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "60"},
   CellID->501569980]},
 "ID:914554779"->{
  Cell[1551377, 43562, 1614, 46, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:914554779", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "61"},
   CellID->914554779]},
 "61"->{
  Cell[1551377, 43562, 1614, 46, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:914554779", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "61"},
   CellID->914554779]},
 "ID:1336064727"->{
  Cell[1552994, 43610, 1748, 50, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1336064727", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "62"},
   CellID->1336064727]},
 "62"->{
  Cell[1552994, 43610, 1748, 50, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1336064727", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "62"},
   CellID->1336064727]},
 "ID:1651077201"->{
  Cell[1554745, 43662, 1710, 51, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1651077201", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "L+\[LessEqual]"},
   CellID->1651077201]},
 "L+\[LessEqual]"->{
  Cell[1554745, 43662, 1710, 51, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1651077201", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "L+\[LessEqual]"},
   CellID->1651077201]},
 "Proof|ID:501569980"->{
  Cell[1556612, 43724, 7748, 190, 29, "ProofInfo",
   CellTags->{"Proof|ID:501569980", "Proof|ID:501569980-1"},
   CellID->1185891342]},
 "Proof|ID:501569980-1"->{
  Cell[1556612, 43724, 7748, 190, 29, "ProofInfo",
   CellTags->{"Proof|ID:501569980", "Proof|ID:501569980-1"},
   CellID->1185891342]},
 "Proof|ID:914554779"->{
  Cell[1564883, 43933, 7869, 194, 27, "ProofInfo",
   CellTags->{"Proof|ID:914554779", "Proof|ID:914554779-1"},
   CellID->5884268]},
 "Proof|ID:914554779-1"->{
  Cell[1564883, 43933, 7869, 194, 27, "ProofInfo",
   CellTags->{"Proof|ID:914554779", "Proof|ID:914554779-1"},
   CellID->5884268]},
 "Proof|ID:1336064727"->{
  Cell[1573315, 44147, 8362, 206, 27, "ProofInfo",
   CellTags->{"Proof|ID:1336064727", "Proof|ID:1336064727-1"},
   CellID->970257786]},
 "Proof|ID:1336064727-1"->{
  Cell[1573315, 44147, 8362, 206, 27, "ProofInfo",
   CellTags->{"Proof|ID:1336064727", "Proof|ID:1336064727-1"},
   CellID->970257786]},
 "Proof|ID:1651077201"->{
  Cell[1582243, 44373, 9382, 226, 27, "ProofInfo",
   CellTags->{"Proof|ID:1651077201", "Proof|ID:1651077201-1"},
   CellID->451306320]},
 "Proof|ID:1651077201-1"->{
  Cell[1582243, 44373, 9382, 226, 27, "ProofInfo",
   CellTags->{"Proof|ID:1651077201", "Proof|ID:1651077201-1"},
   CellID->451306320]},
 "sect::chainCritSpecial"->{
  Cell[1592202, 44620, 229, 4, 22, "Subsubsection",
   CellTags->"sect::chainCritSpecial",
   CellID->226754043]},
 "ID:53741691"->{
  Cell[1595418, 44724, 1054, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:53741691", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "63"},
   CellID->53741691]},
 "63"->{
  Cell[1595418, 44724, 1054, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:53741691", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "63"},
   CellID->53741691]},
 "ID:1045358942"->{
  Cell[1596475, 44756, 1014, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1045358942", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "64"},
   CellID->1045358942]},
 "64"->{
  Cell[1596475, 44756, 1014, 30, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1045358942", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "64"},
   CellID->1045358942]},
 "ID:1132937314"->{
  Cell[1597492, 44788, 1068, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1132937314", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "65"},
   CellID->1132937314]},
 "65"->{
  Cell[1597492, 44788, 1068, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1132937314", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "65"},
   CellID->1132937314]},
 "ID:152254973"->{
  Cell[1598563, 44822, 1066, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:152254973", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "66"},
   CellID->152254973]},
 "66"->{
  Cell[1598563, 44822, 1066, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:152254973", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "66"},
   CellID->152254973]},
 "Proof|ID:53741691"->{
  Cell[1599786, 44865, 6652, 167, 29, "ProofInfo",
   CellTags->{"Proof|ID:53741691", "Proof|ID:53741691-1"},
   CellID->1005730830]},
 "Proof|ID:53741691-1"->{
  Cell[1599786, 44865, 6652, 167, 29, "ProofInfo",
   CellTags->{"Proof|ID:53741691", "Proof|ID:53741691-1"},
   CellID->1005730830]},
 "Proof|ID:1045358942"->{
  Cell[1607002, 45052, 6662, 167, 27, "ProofInfo",
   CellTags->{"Proof|ID:1045358942", "Proof|ID:1045358942-1"},
   CellID->1783727900]},
 "Proof|ID:1045358942-1"->{
  Cell[1607002, 45052, 6662, 167, 27, "ProofInfo",
   CellTags->{"Proof|ID:1045358942", "Proof|ID:1045358942-1"},
   CellID->1783727900]},
 "Proof|ID:1132937314"->{
  Cell[1614277, 45240, 6547, 165, 27, "ProofInfo",
   CellTags->{"Proof|ID:1132937314", "Proof|ID:1132937314-1"},
   CellID->279556406]},
 "Proof|ID:1132937314-1"->{
  Cell[1614277, 45240, 6547, 165, 27, "ProofInfo",
   CellTags->{"Proof|ID:1132937314", "Proof|ID:1132937314-1"},
   CellID->279556406]},
 "Proof|ID:152254973"->{
  Cell[1621389, 45425, 6543, 165, 27, "ProofInfo",
   CellTags->{"Proof|ID:152254973", "Proof|ID:152254973-1"},
   CellID->1134289223]},
 "Proof|ID:152254973-1"->{
  Cell[1621389, 45425, 6543, 165, 27, "ProofInfo",
   CellTags->{"Proof|ID:152254973", "Proof|ID:152254973-1"},
   CellID->1134289223]},
 "ID:836877533"->{
  Cell[1630384, 45669, 1000, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:836877533", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 1"},
   CellID->836877533]},
 "symmetry of chainCritDiagonal 1"->{
  Cell[1630384, 45669, 1000, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:836877533", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 1"},
   CellID->836877533]},
 "ID:1226632396"->{
  Cell[1631387, 45696, 1071, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1226632396", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 2"},
   CellID->1226632396]},
 "symmetry of chainCritDiagonal 2"->{
  Cell[1631387, 45696, 1071, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1226632396", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 2"},
   CellID->1226632396]},
 "ID:457131364"->{
  Cell[1632461, 45726, 1264, 32, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:457131364", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 3"},
   CellID->457131364]},
 "symmetry of chainCritDiagonal 3"->{
  Cell[1632461, 45726, 1264, 32, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:457131364", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 3"},
   CellID->457131364]},
 "ID:1490326930"->{
  Cell[1633728, 45760, 1020, 27, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1490326930", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal"},
   CellID->1490326930]},
 "symmetry of chainCritDiagonal"->{
  Cell[1633728, 45760, 1020, 27, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1490326930", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal"},
   CellID->1490326930]},
 "ID:749802963"->{
  Cell[1634751, 45789, 2256, 68, 112, "FormalTextInputFormula",
   CellTags->{
    "ID:749802963", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal w.r.t. first coordinate"},
   CellID->749802963]},
 "chainCritDiagonal w.r.t. first coordinate"->{
  Cell[1634751, 45789, 2256, 68, 112, "FormalTextInputFormula",
   CellTags->{
    "ID:749802963", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal w.r.t. first coordinate"},
   CellID->749802963]},
 "ID:505706706"->{
  Cell[1637256, 45867, 1026, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:505706706", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "67"},
   CellID->505706706]},
 "67"->{
  Cell[1637256, 45867, 1026, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:505706706", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "67"},
   CellID->505706706]},
 "ID:1262287499"->{
  Cell[1638285, 45898, 2989, 83, 71, "FormalTextInputFormula",
   CellTags->{
    "ID:1262287499", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "68"},
   CellID->1262287499]},
 "68"->{
  Cell[1638285, 45898, 2989, 83, 71, "FormalTextInputFormula",
   CellTags->{
    "ID:1262287499", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "68"},
   CellID->1262287499]},
 "Proof|ID:836877533"->{
  Cell[1641430, 45992, 6076, 150, 29, "ProofInfo",
   CellTags->{"Proof|ID:836877533", "Proof|ID:836877533-1"},
   CellID->313941599]},
 "Proof|ID:836877533-1"->{
  Cell[1641430, 45992, 6076, 150, 29, "ProofInfo",
   CellTags->{"Proof|ID:836877533", "Proof|ID:836877533-1"},
   CellID->313941599]},
 "Proof|ID:1226632396"->{
  Cell[1648071, 46162, 6107, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1226632396", "Proof|ID:1226632396-1"},
   CellID->346366551]},
 "Proof|ID:1226632396-1"->{
  Cell[1648071, 46162, 6107, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:1226632396", "Proof|ID:1226632396-1"},
   CellID->346366551]},
 "Proof|ID:457131364"->{
  Cell[1654742, 46332, 6103, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:457131364", "Proof|ID:457131364-1"},
   CellID->1718476676]},
 "Proof|ID:457131364-1"->{
  Cell[1654742, 46332, 6103, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:457131364", "Proof|ID:457131364-1"},
   CellID->1718476676]},
 "Proof|ID:1490326930"->{
  Cell[1661408, 46502, 4882, 122, 27, "ProofInfo",
   CellTags->{"Proof|ID:1490326930", "Proof|ID:1490326930-1"},
   CellID->758177161]},
 "Proof|ID:1490326930-1"->{
  Cell[1661408, 46502, 4882, 122, 27, "ProofInfo",
   CellTags->{"Proof|ID:1490326930", "Proof|ID:1490326930-1"},
   CellID->758177161]},
 "Proof|ID:749802963"->{
  Cell[1666856, 46644, 11170, 262, 27, "ProofInfo",
   CellTags->{"Proof|ID:749802963", "Proof|ID:749802963-1"},
   CellID->1283930748]},
 "Proof|ID:749802963-1"->{
  Cell[1666856, 46644, 11170, 262, 27, "ProofInfo",
   CellTags->{"Proof|ID:749802963", "Proof|ID:749802963-1"},
   CellID->1283930748]},
 "Proof|ID:505706706"->{
  Cell[1678590, 46926, 5957, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:505706706", "Proof|ID:505706706-1"},
   CellID->292146520]},
 "Proof|ID:505706706-1"->{
  Cell[1678590, 46926, 5957, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:505706706", "Proof|ID:505706706-1"},
   CellID->292146520]},
 "Proof|ID:1262287499"->{
  Cell[1685111, 47096, 13803, 320, 27, "ProofInfo",
   CellTags->{"Proof|ID:1262287499", "Proof|ID:1262287499-1"},
   CellID->1935805160]},
 "Proof|ID:1262287499-1"->{
  Cell[1685111, 47096, 13803, 320, 27, "ProofInfo",
   CellTags->{"Proof|ID:1262287499", "Proof|ID:1262287499-1"},
   CellID->1935805160]},
 "ID:1909233179"->{
  Cell[1700637, 47472, 1576, 50, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1909233179", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "69"},
   CellID->1909233179]},
 "69"->{
  Cell[1700637, 47472, 1576, 50, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1909233179", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "69"},
   CellID->1909233179]},
 "ID:2125747942"->{
  Cell[1702216, 47524, 1682, 54, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:2125747942", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "70"},
   CellID->2125747942]},
 "70"->{
  Cell[1702216, 47524, 1682, 54, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:2125747942", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "70"},
   CellID->2125747942]},
 "ID:1211430036"->{
  Cell[1703901, 47580, 1776, 55, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1211430036", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "71"},
   CellID->1211430036]},
 "71"->{
  Cell[1703901, 47580, 1776, 55, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1211430036", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "71"},
   CellID->1211430036]},
 "ID:2061027934"->{
  Cell[1706305, 47655, 900, 25, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:2061027934", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "72"},
   CellID->2061027934]},
 "72"->{
  Cell[1706305, 47655, 900, 25, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:2061027934", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "72"},
   CellID->2061027934]},
 "ID:1653917255"->{
  Cell[1707208, 47682, 778, 23, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1653917255", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "73"},
   CellID->1653917255]},
 "73"->{
  Cell[1707208, 47682, 778, 23, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1653917255", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "73"},
   CellID->1653917255]},
 "Proof|ID:1909233179"->{
  Cell[1708143, 47716, 11159, 264, 29, "ProofInfo",
   CellTags->{"Proof|ID:1909233179", "Proof|ID:1909233179-1"},
   CellID->1389502986]},
 "Proof|ID:1909233179-1"->{
  Cell[1708143, 47716, 11159, 264, 29, "ProofInfo",
   CellTags->{"Proof|ID:1909233179", "Proof|ID:1909233179-1"},
   CellID->1389502986]},
 "Proof|ID:2125747942"->{
  Cell[1719867, 48000, 11434, 272, 27, "ProofInfo",
   CellTags->{"Proof|ID:2125747942", "Proof|ID:2125747942-1"},
   CellID->673958282]},
 "Proof|ID:2125747942-1"->{
  Cell[1719867, 48000, 11434, 272, 27, "ProofInfo",
   CellTags->{"Proof|ID:2125747942", "Proof|ID:2125747942-1"},
   CellID->673958282]},
 "Proof|ID:1211430036"->{
  Cell[1731865, 48292, 11435, 272, 27, "ProofInfo",
   CellTags->{"Proof|ID:1211430036", "Proof|ID:1211430036-1"},
   CellID->1679514424]},
 "Proof|ID:1211430036-1"->{
  Cell[1731865, 48292, 11435, 272, 27, "ProofInfo",
   CellTags->{"Proof|ID:1211430036", "Proof|ID:1211430036-1"},
   CellID->1679514424]},
 "Proof|ID:2061027934"->{
  Cell[1743864, 48584, 7690, 184, 27, "ProofInfo",
   CellTags->{"Proof|ID:2061027934", "Proof|ID:2061027934-1"},
   CellID->381202233]},
 "Proof|ID:2061027934-1"->{
  Cell[1743864, 48584, 7690, 184, 27, "ProofInfo",
   CellTags->{"Proof|ID:2061027934", "Proof|ID:2061027934-1"},
   CellID->381202233]},
 "Proof|ID:1653917255"->{
  Cell[1752120, 48788, 7690, 184, 27, "ProofInfo",
   CellTags->{"Proof|ID:1653917255", "Proof|ID:1653917255-1"},
   CellID->929231870]},
 "Proof|ID:1653917255-1"->{
  Cell[1752120, 48788, 7690, 184, 27, "ProofInfo",
   CellTags->{"Proof|ID:1653917255", "Proof|ID:1653917255-1"},
   CellID->929231870]},
 "ID:1690170957"->{
  Cell[1762262, 49041, 1132, 27, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1690170957", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 1"},
   CellID->1690170957]},
 "chainCrit bivariate 1"->{
  Cell[1762262, 49041, 1132, 27, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1690170957", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 1"},
   CellID->1690170957]},
 "ID:271680825"->{
  Cell[1763397, 49070, 926, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:271680825", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 2"},
   CellID->271680825]},
 "chainCrit bivariate 2"->{
  Cell[1763397, 49070, 926, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:271680825", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 2"},
   CellID->271680825]},
 "ID:1572461042"->{
  Cell[1764326, 49096, 3340, 100, 111, "FormalTextInputFormula",
   CellTags->{
    "ID:1572461042", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 3"},
   CellID->1572461042]},
 "chainCrit bivariate 3"->{
  Cell[1764326, 49096, 3340, 100, 111, "FormalTextInputFormula",
   CellTags->{
    "ID:1572461042", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 3"},
   CellID->1572461042]},
 "ID:2046865956"->{
  Cell[1767669, 49198, 1385, 41, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:2046865956", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 4"},
   CellID->2046865956]},
 "chainCrit bivariate 4"->{
  Cell[1767669, 49198, 1385, 41, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:2046865956", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 4"},
   CellID->2046865956]},
 "ID:917844397"->{
  Cell[1769057, 49241, 1244, 35, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:917844397", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 5"},
   CellID->917844397]},
 "chainCrit bivariate 5"->{
  Cell[1769057, 49241, 1244, 35, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:917844397", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 5"},
   CellID->917844397]},
 "ID:810037327"->{
  Cell[1770304, 49278, 2181, 49, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:810037327", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate"},
   CellID->810037327]},
 "chainCrit bivariate"->{
  Cell[1770304, 49278, 2181, 49, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:810037327", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate"},
   CellID->810037327]},
 "ID:2111681775"->{
  Cell[1772488, 49329, 1642, 50, 71, "FormalTextInputFormula",
   CellTags->{
    "ID:2111681775", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate I"},
   CellID->2111681775]},
 "chainCrit bivariate I"->{
  Cell[1772488, 49329, 1642, 50, 71, "FormalTextInputFormula",
   CellTags->{
    "ID:2111681775", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate I"},
   CellID->2111681775]},
 "Proof|ID:1690170957"->{
  Cell[1774287, 49390, 4845, 122, 29, "ProofInfo",
   CellTags->{"Proof|ID:1690170957", "Proof|ID:1690170957-1"},
   CellID->2085296512]},
 "Proof|ID:1690170957-1"->{
  Cell[1774287, 49390, 4845, 122, 29, "ProofInfo",
   CellTags->{"Proof|ID:1690170957", "Proof|ID:1690170957-1"},
   CellID->2085296512]},
 "Proof|ID:271680825"->{
  Cell[1779697, 49532, 4788, 120, 27, "ProofInfo",
   CellTags->{"Proof|ID:271680825", "Proof|ID:271680825-1"},
   CellID->1844198183]},
 "Proof|ID:271680825-1"->{
  Cell[1779697, 49532, 4788, 120, 27, "ProofInfo",
   CellTags->{"Proof|ID:271680825", "Proof|ID:271680825-1"},
   CellID->1844198183]},
 "Proof|ID:1572461042"->{
  Cell[1785050, 49672, 15570, 360, 27, "ProofInfo",
   CellTags->{"Proof|ID:1572461042", "Proof|ID:1572461042-1"},
   CellID->1050587621]},
 "Proof|ID:1572461042-1"->{
  Cell[1785050, 49672, 15570, 360, 27, "ProofInfo",
   CellTags->{"Proof|ID:1572461042", "Proof|ID:1572461042-1"},
   CellID->1050587621]},
 "Proof|ID:2046865956"->{
  Cell[1801143, 50051, 7087, 170, 27, "ProofInfo",
   CellTags->{"Proof|ID:2046865956", "Proof|ID:2046865956-1"},
   CellID->1535230662]},
 "Proof|ID:2046865956-1"->{
  Cell[1801143, 50051, 7087, 170, 27, "ProofInfo",
   CellTags->{"Proof|ID:2046865956", "Proof|ID:2046865956-1"},
   CellID->1535230662]},
 "Proof|ID:917844397"->{
  Cell[1808795, 50241, 6249, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:917844397", "Proof|ID:917844397-1"},
   CellID->665088749]},
 "Proof|ID:917844397-1"->{
  Cell[1808795, 50241, 6249, 150, 27, "ProofInfo",
   CellTags->{"Proof|ID:917844397", "Proof|ID:917844397-1"},
   CellID->665088749]},
 "Proof|ID:810037327"->{
  Cell[1815608, 50411, 5682, 138, 27, "ProofInfo",
   CellTags->{"Proof|ID:810037327", "Proof|ID:810037327-1"},
   CellID->1607548979]},
 "Proof|ID:810037327-1"->{
  Cell[1815608, 50411, 5682, 138, 27, "ProofInfo",
   CellTags->{"Proof|ID:810037327", "Proof|ID:810037327-1"},
   CellID->1607548979]},
 "Proof|ID:2111681775"->{
  Cell[1821855, 50569, 7475, 175, 27, "ProofInfo",
   CellTags->{"Proof|ID:2111681775", "Proof|ID:2111681775-1"},
   CellID->215040822]},
 "Proof|ID:2111681775-1"->{
  Cell[1821855, 50569, 7475, 175, 27, "ProofInfo",
   CellTags->{"Proof|ID:2111681775", "Proof|ID:2111681775-1"},
   CellID->215040822]},
 "sect::Theorems"->{
  Cell[1829932, 50767, 289, 5, 43, "Section",
   CellTags->"sect::Theorems",
   CellID->1928916681]},
 "sect::M\[RawStar]"->{
  Cell[1832550, 50834, 363, 9, 38, "Subsection",
   CellTags->"sect::M\[RawStar]",
   CellID->391063855]},
 "ID:2973275"->{
  Cell[1836634, 50957, 2387, 65, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:2973275", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar] binary"},
   CellID->2973275]},
 "M\[RawStar] binary"->{
  Cell[1836634, 50957, 2387, 65, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:2973275", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar] binary"},
   CellID->2973275]},
 "ID:1258784171"->{
  Cell[1839024, 51024, 1541, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1258784171", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar] unary"},
   CellID->1258784171]},
 "M\[RawStar] unary"->{
  Cell[1839024, 51024, 1541, 43, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:1258784171", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar] unary"},
   CellID->1258784171]},
 "ID:1549582561"->{
  Cell[1842616, 51136, 1393, 35, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1549582561", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit holds for distinct indices"},
   CellID->1549582561]},
 "chainCrit holds for distinct indices"->{
  Cell[1842616, 51136, 1393, 35, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1549582561", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit holds for distinct indices"},
   CellID->1549582561]},
 "ID:1966722449"->{
  Cell[1844012, 51173, 1470, 41, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1966722449", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "74"},
   CellID->1966722449]},
 "74"->{
  Cell[1844012, 51173, 1470, 41, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1966722449", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "74"},
   CellID->1966722449]},
 "ID:164254989"->{
  Cell[1845485, 51216, 1405, 40, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:164254989", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "75"},
   CellID->164254989]},
 "75"->{
  Cell[1845485, 51216, 1405, 40, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:164254989", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "75"},
   CellID->164254989]},
 "ID:290364666"->{
  Cell[1846893, 51258, 1018, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:290364666", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M\[RawStar] (binary)"},
   CellID->290364666]},
 "type of M\[RawStar] (binary)"->{
  Cell[1846893, 51258, 1018, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:290364666", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M\[RawStar] (binary)"},
   CellID->290364666]},
 "ID:1063863080"->{
  Cell[1847914, 51287, 974, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1063863080", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M\[RawStar] (unary)"},
   CellID->1063863080]},
 "type of M\[RawStar] (unary)"->{
  Cell[1847914, 51287, 974, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1063863080", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M\[RawStar] (unary)"},
   CellID->1063863080]},
 "Proof|ID:1549582561"->{
  Cell[1849046, 51325, 6957, 172, 29, "ProofInfo",
   CellTags->{"Proof|ID:1549582561", "Proof|ID:1549582561-1"},
   CellID->1962722847]},
 "Proof|ID:1549582561-1"->{
  Cell[1849046, 51325, 6957, 172, 29, "ProofInfo",
   CellTags->{"Proof|ID:1549582561", "Proof|ID:1549582561-1"},
   CellID->1962722847]},
 "Proof|ID:1966722449"->{
  Cell[1856569, 51517, 9170, 218, 27, "ProofInfo",
   CellTags->{"Proof|ID:1966722449", "Proof|ID:1966722449-1"},
   CellID->938794098]},
 "Proof|ID:1966722449-1"->{
  Cell[1856569, 51517, 9170, 218, 27, "ProofInfo",
   CellTags->{"Proof|ID:1966722449", "Proof|ID:1966722449-1"},
   CellID->938794098]},
 "Proof|ID:164254989"->{
  Cell[1866303, 51755, 8963, 214, 27, "ProofInfo",
   CellTags->{"Proof|ID:164254989", "Proof|ID:164254989-1"},
   CellID->895811688]},
 "Proof|ID:164254989-1"->{
  Cell[1866303, 51755, 8963, 214, 27, "ProofInfo",
   CellTags->{"Proof|ID:164254989", "Proof|ID:164254989-1"},
   CellID->895811688]},
 "Proof|ID:290364666"->{
  Cell[1875830, 51989, 4836, 126, 27, "ProofInfo",
   CellTags->{"Proof|ID:290364666", "Proof|ID:290364666-1"},
   CellID->1493869216]},
 "Proof|ID:290364666-1"->{
  Cell[1875830, 51989, 4836, 126, 27, "ProofInfo",
   CellTags->{"Proof|ID:290364666", "Proof|ID:290364666-1"},
   CellID->1493869216]},
 "Proof|ID:1063863080"->{
  Cell[1881231, 52135, 3500, 96, 27, "ProofInfo",
   CellTags->{"Proof|ID:1063863080", "Proof|ID:1063863080-1"},
   CellID->391529240]},
 "Proof|ID:1063863080-1"->{
  Cell[1881231, 52135, 3500, 96, 27, "ProofInfo",
   CellTags->{"Proof|ID:1063863080", "Proof|ID:1063863080-1"},
   CellID->391529240]},
 "ID:1680577066"->{
  Cell[1890170, 52400, 1658, 42, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1680577066", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M\[RawStar] (unary)"},
   CellID->1680577066]},
 "\[LessEqual]M\[RawStar] (unary)"->{
  Cell[1890170, 52400, 1658, 42, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1680577066", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M\[RawStar] (unary)"},
   CellID->1680577066]},
 "ID:1866021817"->{
  Cell[1891831, 52444, 1729, 44, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1866021817", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual] (unary)"},
   CellID->1866021817]},
 "M\[RawStar]\[LessEqual] (unary)"->{
  Cell[1891831, 52444, 1729, 44, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1866021817", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual] (unary)"},
   CellID->1866021817]},
 "ID:1272742036"->{
  Cell[1893563, 52490, 2109, 58, 54, "FormalTextInputFormula",
   CellTags->{
    "ID:1272742036", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M\[RawStar] (binary)"},
   CellID->1272742036]},
 "\[LessEqual]M\[RawStar] (binary)"->{
  Cell[1893563, 52490, 2109, 58, 54, "FormalTextInputFormula",
   CellTags->{
    "ID:1272742036", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M\[RawStar] (binary)"},
   CellID->1272742036]},
 "ID:462890005"->{
  Cell[1895675, 52550, 1913, 55, 54, "FormalTextInputFormula",
   CellTags->{
    "ID:462890005", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual] (binary)"},
   CellID->462890005]},
 "M\[RawStar]\[LessEqual] (binary)"->{
  Cell[1895675, 52550, 1913, 55, 54, "FormalTextInputFormula",
   CellTags->{
    "ID:462890005", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual] (binary)"},
   CellID->462890005]},
 "Proof|ID:1680577066"->{
  Cell[1897744, 52616, 6497, 164, 29, "ProofInfo",
   CellTags->{"Proof|ID:1680577066", "Proof|ID:1680577066-1"},
   CellID->631558205]},
 "Proof|ID:1680577066-1"->{
  Cell[1897744, 52616, 6497, 164, 29, "ProofInfo",
   CellTags->{"Proof|ID:1680577066", "Proof|ID:1680577066-1"},
   CellID->631558205]},
 "Proof|ID:1866021817"->{
  Cell[1904806, 52800, 6486, 164, 27, "ProofInfo",
   CellTags->{"Proof|ID:1866021817", "Proof|ID:1866021817-1"},
   CellID->860652560]},
 "Proof|ID:1866021817-1"->{
  Cell[1904806, 52800, 6486, 164, 27, "ProofInfo",
   CellTags->{"Proof|ID:1866021817", "Proof|ID:1866021817-1"},
   CellID->860652560]},
 "Proof|ID:1272742036"->{
  Cell[1911857, 52984, 10477, 248, 27, "ProofInfo",
   CellTags->{"Proof|ID:1272742036", "Proof|ID:1272742036-1"},
   CellID->135701756]},
 "Proof|ID:1272742036-1"->{
  Cell[1911857, 52984, 10477, 248, 27, "ProofInfo",
   CellTags->{"Proof|ID:1272742036", "Proof|ID:1272742036-1"},
   CellID->135701756]},
 "Proof|ID:462890005"->{
  Cell[1922900, 53252, 10513, 250, 27, "ProofInfo",
   CellTags->{"Proof|ID:462890005", "Proof|ID:462890005-1"},
   CellID->829755125]},
 "Proof|ID:462890005-1"->{
  Cell[1922900, 53252, 10513, 250, 27, "ProofInfo",
   CellTags->{"Proof|ID:462890005", "Proof|ID:462890005-1"},
   CellID->829755125]},
 "ID:929561290"->{
  Cell[1936278, 53591, 873, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:929561290", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual]M\[RawStar]"},
   CellID->929561290]},
 "M\[LessEqual]M\[RawStar]"->{
  Cell[1936278, 53591, 873, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:929561290", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual]M\[RawStar]"},
   CellID->929561290]},
 "ID:795795393"->{
  Cell[1937154, 53617, 872, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:795795393", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual]M"},
   CellID->795795393]},
 "M\[RawStar]\[LessEqual]M"->{
  Cell[1937154, 53617, 872, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:795795393", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual]M"},
   CellID->795795393]},
 "ID:872843574"->{
  Cell[1938029, 53643, 1234, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:872843574", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M=M\[RawStar]"},
   CellID->872843574]},
 "M=M\[RawStar]"->{
  Cell[1938029, 53643, 1234, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:872843574", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M=M\[RawStar]"},
   CellID->872843574]},
 "Proof|ID:929561290"->{
  Cell[1939503, 53685, 3661, 100, 29, "ProofInfo",
   CellTags->{"Proof|ID:929561290", "Proof|ID:929561290-1"},
   CellID->286309708]},
 "Proof|ID:929561290-1"->{
  Cell[1939503, 53685, 3661, 100, 29, "ProofInfo",
   CellTags->{"Proof|ID:929561290", "Proof|ID:929561290-1"},
   CellID->286309708]},
 "Proof|ID:795795393"->{
  Cell[1943729, 53805, 3662, 100, 27, "ProofInfo",
   CellTags->{"Proof|ID:795795393", "Proof|ID:795795393-1"},
   CellID->1883422493]},
 "Proof|ID:795795393-1"->{
  Cell[1943729, 53805, 3662, 100, 27, "ProofInfo",
   CellTags->{"Proof|ID:795795393", "Proof|ID:795795393-1"},
   CellID->1883422493]},
 "Proof|ID:872843574"->{
  Cell[1947956, 53925, 3562, 98, 27, "ProofInfo",
   CellTags->{"Proof|ID:872843574", "Proof|ID:872843574-1"},
   CellID->1077588120]},
 "Proof|ID:872843574-1"->{
  Cell[1947956, 53925, 3562, 98, 27, "ProofInfo",
   CellTags->{"Proof|ID:872843574", "Proof|ID:872843574-1"},
   CellID->1077588120]},
 "sect::TheoremsGeneral"->{
  Cell[1952095, 54044, 229, 4, 30, "Subsection",
   CellTags->"sect::TheoremsGeneral",
   CellID->1847669776]},
 "???"->{
  Cell[1952954, 54069, 757, 16, 57, "GlobalDeclaration",
   CellTags->"???",
   CellID->560938714],
  Cell[1990474, 55139, 757, 16, 57, "GlobalDeclaration",
   CellTags->"???",
   CellID->299478056],
  Cell[2266899, 62298, 755, 16, 57, "GlobalDeclaration",
   CellTags->"???",
   CellID->1004739293],
  Cell[2267657, 62316, 892, 20, 61, "GlobalDeclaration",
   CellTags->"???",
   CellID->1633843438],
  Cell[2287935, 62904, 755, 16, 57, "GlobalDeclaration",
   CellTags->"???",
   CellID->954349703]},
 "ID:642765392"->{
  Cell[1956073, 54152, 1033, 29, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:642765392", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "76"},
   CellID->642765392]},
 "76"->{
  Cell[1956073, 54152, 1033, 29, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:642765392", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "76"},
   CellID->642765392]},
 "ID:592779258"->{
  Cell[1957109, 54183, 932, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:592779258", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maxdeg\[LessEqual]M"},
   CellID->592779258]},
 "maxdeg\[LessEqual]M"->{
  Cell[1957109, 54183, 932, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:592779258", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maxdeg\[LessEqual]M"},
   CellID->592779258]},
 "Proof|ID:642765392"->{
  Cell[1958280, 54220, 4620, 122, 29, "ProofInfo",
   CellTags->{"Proof|ID:642765392", "Proof|ID:642765392-1"},
   CellID->1075234758]},
 "Proof|ID:642765392-1"->{
  Cell[1958280, 54220, 4620, 122, 29, "ProofInfo",
   CellTags->{"Proof|ID:642765392", "Proof|ID:642765392-1"},
   CellID->1075234758]},
 "Proof|ID:592779258"->{
  Cell[1963465, 54362, 3349, 92, 27, "ProofInfo",
   CellTags->{"Proof|ID:592779258", "Proof|ID:592779258-1"},
   CellID->582763895]},
 "Proof|ID:592779258-1"->{
  Cell[1963465, 54362, 3349, 92, 27, "ProofInfo",
   CellTags->{"Proof|ID:592779258", "Proof|ID:592779258-1"},
   CellID->582763895]},
 "ID:691015215"->{
  Cell[1968850, 54524, 890, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:691015215", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "77"},
   CellID->691015215]},
 "77"->{
  Cell[1968850, 54524, 890, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:691015215", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "77"},
   CellID->691015215]},
 "ID:402511110"->{
  Cell[1969743, 54550, 1254, 33, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:402511110", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "78"},
   CellID->402511110]},
 "78"->{
  Cell[1969743, 54550, 1254, 33, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:402511110", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "78"},
   CellID->402511110]},
 "Proof|ID:691015215"->{
  Cell[1971154, 54594, 4446, 118, 29, "ProofInfo",
   CellTags->{"Proof|ID:691015215", "Proof|ID:691015215-1"},
   CellID->31238831]},
 "Proof|ID:691015215-1"->{
  Cell[1971154, 54594, 4446, 118, 29, "ProofInfo",
   CellTags->{"Proof|ID:691015215", "Proof|ID:691015215-1"},
   CellID->31238831]},
 "Proof|ID:402511110"->{
  Cell[1976165, 54732, 6108, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:402511110", "Proof|ID:402511110-1"},
   CellID->1203109870]},
 "Proof|ID:402511110-1"->{
  Cell[1976165, 54732, 6108, 156, 27, "ProofInfo",
   CellTags->{"Proof|ID:402511110", "Proof|ID:402511110-1"},
   CellID->1203109870]},
 "ID:341822661"->{
  Cell[1983724, 54935, 973, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:341822661", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual]L"},
   CellID->341822661]},
 "M\[LessEqual]L"->{
  Cell[1983724, 54935, 973, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:341822661", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual]L"},
   CellID->341822661]},
 "Proof|ID:341822661"->{
  Cell[1984853, 54972, 3420, 94, 29, "ProofInfo",
   CellTags->{"Proof|ID:341822661", "Proof|ID:341822661-1"},
   CellID->1958581994]},
 "Proof|ID:341822661-1"->{
  Cell[1984853, 54972, 3420, 94, 29, "ProofInfo",
   CellTags->{"Proof|ID:341822661", "Proof|ID:341822661-1"},
   CellID->1958581994]},
 "sect::TheoremsSpecial"->{
  Cell[1989199, 55097, 246, 8, 30, "Subsection",
   CellTags->"sect::TheoremsSpecial",
   CellID->240164494]},
 "ID:245124153"->{
  Cell[1994463, 55247, 1421, 40, 56, "FormalTextInputFormula",
   CellTags->{
    "ID:245124153", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "79"},
   CellID->245124153]},
 "79"->{
  Cell[1994463, 55247, 1421, 40, 56, "FormalTextInputFormula",
   CellTags->{
    "ID:245124153", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "79"},
   CellID->245124153]},
 "ID:952627996"->{
  Cell[1995887, 55289, 1437, 47, 63, "FormalTextInputFormula",
   CellTags->{
    "ID:952627996", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "80"},
   CellID->952627996]},
 "80"->{
  Cell[1995887, 55289, 1437, 47, 63, "FormalTextInputFormula",
   CellTags->{
    "ID:952627996", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "80"},
   CellID->952627996]},
 "ID:2013760956"->{
  Cell[1998901, 55390, 1453, 31, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2013760956", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "81"},
   CellID->2013760956]},
 "81"->{
  Cell[1998901, 55390, 1453, 31, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2013760956", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "81"},
   CellID->2013760956]},
 "ID:1011330545"->{
  Cell[2000357, 55423, 964, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1011330545", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "82"},
   CellID->1011330545]},
 "82"->{
  Cell[2000357, 55423, 964, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1011330545", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "82"},
   CellID->1011330545]},
 "ID:718586393"->{
  Cell[2001324, 55453, 1278, 37, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:718586393", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "83"},
   CellID->718586393]},
 "83"->{
  Cell[2001324, 55453, 1278, 37, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:718586393", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "83"},
   CellID->718586393]},
 "Proof|ID:245124153"->{
  Cell[2002759, 55501, 8975, 216, 29, "ProofInfo",
   CellTags->{"Proof|ID:245124153", "Proof|ID:245124153-1"},
   CellID->1061920785]},
 "Proof|ID:245124153-1"->{
  Cell[2002759, 55501, 8975, 216, 29, "ProofInfo",
   CellTags->{"Proof|ID:245124153", "Proof|ID:245124153-1"},
   CellID->1061920785]},
 "Proof|ID:952627996"->{
  Cell[2012299, 55737, 9879, 236, 27, "ProofInfo",
   CellTags->{"Proof|ID:952627996", "Proof|ID:952627996-1"},
   CellID->921465591]},
 "Proof|ID:952627996-1"->{
  Cell[2012299, 55737, 9879, 236, 27, "ProofInfo",
   CellTags->{"Proof|ID:952627996", "Proof|ID:952627996-1"},
   CellID->921465591]},
 "Proof|ID:2013760956"->{
  Cell[2022742, 55993, 15095, 350, 27, "ProofInfo",
   CellTags->{"Proof|ID:2013760956", "Proof|ID:2013760956-1"},
   CellID->605609267]},
 "Proof|ID:2013760956-1"->{
  Cell[2022742, 55993, 15095, 350, 27, "ProofInfo",
   CellTags->{"Proof|ID:2013760956", "Proof|ID:2013760956-1"},
   CellID->605609267]},
 "Proof|ID:1011330545"->{
  Cell[2038401, 56363, 15728, 364, 27, "ProofInfo",
   CellTags->{"Proof|ID:1011330545", "Proof|ID:1011330545-1"},
   CellID->1676671824]},
 "Proof|ID:1011330545-1"->{
  Cell[2038401, 56363, 15728, 364, 27, "ProofInfo",
   CellTags->{"Proof|ID:1011330545", "Proof|ID:1011330545-1"},
   CellID->1676671824]},
 "Proof|ID:718586393"->{
  Cell[2054695, 56747, 12863, 300, 27, "ProofInfo",
   CellTags->{"Proof|ID:718586393", "Proof|ID:718586393-1"},
   CellID->1047095018]},
 "Proof|ID:718586393-1"->{
  Cell[2054695, 56747, 12863, 300, 27, "ProofInfo",
   CellTags->{"Proof|ID:718586393", "Proof|ID:718586393-1"},
   CellID->1047095018]},
 "ID:696507084"->{
  Cell[2079705, 57400, 1342, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:696507084", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "84"},
   CellID->696507084]},
 "84"->{
  Cell[2079705, 57400, 1342, 40, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:696507084", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "84"},
   CellID->696507084]},
 "ID:626970451"->{
  Cell[2081050, 57442, 895, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:626970451", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "85"},
   CellID->626970451]},
 "85"->{
  Cell[2081050, 57442, 895, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:626970451", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "85"},
   CellID->626970451]},
 "ID:789654176"->{
  Cell[2081948, 57469, 1525, 36, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:789654176", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "86"},
   CellID->789654176]},
 "86"->{
  Cell[2081948, 57469, 1525, 36, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:789654176", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "86"},
   CellID->789654176]},
 "Proof|ID:696507084"->{
  Cell[2083629, 57516, 11452, 270, 29, "ProofInfo",
   CellTags->{"Proof|ID:696507084", "Proof|ID:696507084-1"},
   CellID->785497368]},
 "Proof|ID:696507084-1"->{
  Cell[2083629, 57516, 11452, 270, 29, "ProofInfo",
   CellTags->{"Proof|ID:696507084", "Proof|ID:696507084-1"},
   CellID->785497368]},
 "Proof|ID:626970451"->{
  Cell[2095646, 57806, 8563, 207, 27, "ProofInfo",
   CellTags->{"Proof|ID:626970451", "Proof|ID:626970451-1"},
   CellID->1008962236]},
 "Proof|ID:626970451-1"->{
  Cell[2095646, 57806, 8563, 207, 27, "ProofInfo",
   CellTags->{"Proof|ID:626970451", "Proof|ID:626970451-1"},
   CellID->1008962236]},
 "Proof|ID:789654176"->{
  Cell[2104773, 58033, 8572, 207, 27, "ProofInfo",
   CellTags->{"Proof|ID:789654176", "Proof|ID:789654176-1"},
   CellID->543282043]},
 "Proof|ID:789654176-1"->{
  Cell[2104773, 58033, 8572, 207, 27, "ProofInfo",
   CellTags->{"Proof|ID:789654176", "Proof|ID:789654176-1"},
   CellID->543282043]},
 "ID:1437590439"->{
  Cell[2119489, 58422, 1706, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1437590439", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "87"},
   CellID->1437590439]},
 "87"->{
  Cell[2119489, 58422, 1706, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1437590439", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "87"},
   CellID->1437590439]},
 "ID:477242065"->{
  Cell[2121198, 58472, 1572, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:477242065", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "88"},
   CellID->477242065]},
 "88"->{
  Cell[2121198, 58472, 1572, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:477242065", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "88"},
   CellID->477242065]},
 "Proof|ID:1437590439"->{
  Cell[2122927, 58528, 7513, 184, 29, "ProofInfo",
   CellTags->{"Proof|ID:1437590439", "Proof|ID:1437590439-1"},
   CellID->773586293]},
 "Proof|ID:1437590439-1"->{
  Cell[2122927, 58528, 7513, 184, 29, "ProofInfo",
   CellTags->{"Proof|ID:1437590439", "Proof|ID:1437590439-1"},
   CellID->773586293]},
 "Proof|ID:477242065"->{
  Cell[2131004, 58732, 5334, 137, 27, "ProofInfo",
   CellTags->{"Proof|ID:477242065", "Proof|ID:477242065-1"},
   CellID->700215890]},
 "Proof|ID:477242065-1"->{
  Cell[2131004, 58732, 5334, 137, 27, "ProofInfo",
   CellTags->{"Proof|ID:477242065", "Proof|ID:477242065-1"},
   CellID->700215890]},
 "sect::TheoremsFarExterior"->{
  Cell[2136915, 58890, 213, 4, 22, "Subsubsection",
   CellTags->"sect::TheoremsFarExterior",
   CellID->1174010248]},
 "ID:1041359623"->{
  Cell[2140991, 59012, 793, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1041359623", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "89"},
   CellID->1041359623]},
 "89"->{
  Cell[2140991, 59012, 793, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1041359623", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "89"},
   CellID->1041359623]},
 "ID:2130272336"->{
  Cell[2141787, 59036, 797, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2130272336", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "90"},
   CellID->2130272336]},
 "90"->{
  Cell[2141787, 59036, 797, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2130272336", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "90"},
   CellID->2130272336]},
 "ID:1584641667"->{
  Cell[2142587, 59061, 872, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1584641667", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "91"},
   CellID->1584641667]},
 "91"->{
  Cell[2142587, 59061, 872, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1584641667", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "91"},
   CellID->1584641667]},
 "ID:1175076282"->{
  Cell[2143462, 59087, 1139, 32, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1175076282", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "92"},
   CellID->1175076282]},
 "92"->{
  Cell[2143462, 59087, 1139, 32, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1175076282", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "92"},
   CellID->1175076282]},
 "ID:333980150"->{
  Cell[2144856, 59130, 1215, 33, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:333980150", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "93"},
   CellID->333980150]},
 "93"->{
  Cell[2144856, 59130, 1215, 33, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:333980150", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "93"},
   CellID->333980150]},
 "ID:1370308823"->{
  Cell[2146074, 59165, 863, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1370308823", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "94"},
   CellID->1370308823]},
 "94"->{
  Cell[2146074, 59165, 863, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1370308823", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "94"},
   CellID->1370308823]},
 "Proof|ID:1041359623"->{
  Cell[2147095, 59201, 6928, 174, 29, "ProofInfo",
   CellTags->{"Proof|ID:1041359623", "Proof|ID:1041359623-1"},
   CellID->650148495]},
 "Proof|ID:1041359623-1"->{
  Cell[2147095, 59201, 6928, 174, 29, "ProofInfo",
   CellTags->{"Proof|ID:1041359623", "Proof|ID:1041359623-1"},
   CellID->650148495]},
 "Proof|ID:2130272336"->{
  Cell[2154588, 59395, 6928, 174, 27, "ProofInfo",
   CellTags->{"Proof|ID:2130272336", "Proof|ID:2130272336-1"},
   CellID->874306688]},
 "Proof|ID:2130272336-1"->{
  Cell[2154588, 59395, 6928, 174, 27, "ProofInfo",
   CellTags->{"Proof|ID:2130272336", "Proof|ID:2130272336-1"},
   CellID->874306688]},
 "Proof|ID:1584641667"->{
  Cell[2162082, 59589, 7636, 188, 27, "ProofInfo",
   CellTags->{"Proof|ID:1584641667", "Proof|ID:1584641667-1"},
   CellID->813414581]},
 "Proof|ID:1584641667-1"->{
  Cell[2162082, 59589, 7636, 188, 27, "ProofInfo",
   CellTags->{"Proof|ID:1584641667", "Proof|ID:1584641667-1"},
   CellID->813414581]},
 "Proof|ID:1175076282"->{
  Cell[2170284, 59797, 8672, 210, 27, "ProofInfo",
   CellTags->{"Proof|ID:1175076282", "Proof|ID:1175076282-1"},
   CellID->54321453]},
 "Proof|ID:1175076282-1"->{
  Cell[2170284, 59797, 8672, 210, 27, "ProofInfo",
   CellTags->{"Proof|ID:1175076282", "Proof|ID:1175076282-1"},
   CellID->54321453]},
 "Proof|ID:333980150"->{
  Cell[2179522, 60027, 8526, 208, 27, "ProofInfo",
   CellTags->{"Proof|ID:333980150", "Proof|ID:333980150-1"},
   CellID->749915393]},
 "Proof|ID:333980150-1"->{
  Cell[2179522, 60027, 8526, 208, 27, "ProofInfo",
   CellTags->{"Proof|ID:333980150", "Proof|ID:333980150-1"},
   CellID->749915393]},
 "Proof|ID:1370308823"->{
  Cell[2188613, 60255, 8440, 206, 27, "ProofInfo",
   CellTags->{"Proof|ID:1370308823", "Proof|ID:1370308823-1"},
   CellID->1775388908]},
 "Proof|ID:1370308823-1"->{
  Cell[2188613, 60255, 8440, 206, 27, "ProofInfo",
   CellTags->{"Proof|ID:1370308823", "Proof|ID:1370308823-1"},
   CellID->1775388908]},
 "ID:1434774204"->{
  Cell[2208836, 60784, 1216, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1434774204", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "95"},
   CellID->1434774204]},
 "95"->{
  Cell[2208836, 60784, 1216, 36, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1434774204", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "95"},
   CellID->1434774204]},
 "ID:1867780171"->{
  Cell[2210055, 60822, 976, 30, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1867780171", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "96"},
   CellID->1867780171]},
 "96"->{
  Cell[2210055, 60822, 976, 30, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1867780171", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "96"},
   CellID->1867780171]},
 "ID:786584693"->{
  Cell[2211307, 60863, 1048, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:786584693", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "97"},
   CellID->786584693]},
 "97"->{
  Cell[2211307, 60863, 1048, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:786584693", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "97"},
   CellID->786584693]},
 "ID:183897269"->{
  Cell[2212358, 60896, 1022, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:183897269", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "98"},
   CellID->183897269]},
 "98"->{
  Cell[2212358, 60896, 1022, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:183897269", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "98"},
   CellID->183897269]},
 "ID:1246401530"->{
  Cell[2213383, 60929, 1403, 39, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1246401530", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "99"},
   CellID->1246401530]},
 "99"->{
  Cell[2213383, 60929, 1403, 39, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1246401530", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "99"},
   CellID->1246401530]},
 "Proof|ID:1434774204"->{
  Cell[2214942, 60979, 8990, 220, 29, "ProofInfo",
   CellTags->{"Proof|ID:1434774204", "Proof|ID:1434774204-1"},
   CellID->4587845]},
 "Proof|ID:1434774204-1"->{
  Cell[2214942, 60979, 8990, 220, 29, "ProofInfo",
   CellTags->{"Proof|ID:1434774204", "Proof|ID:1434774204-1"},
   CellID->4587845]},
 "Proof|ID:1867780171"->{
  Cell[2224496, 61219, 6493, 164, 27, "ProofInfo",
   CellTags->{"Proof|ID:1867780171", "Proof|ID:1867780171-1"},
   CellID->975302909]},
 "Proof|ID:1867780171-1"->{
  Cell[2224496, 61219, 6493, 164, 27, "ProofInfo",
   CellTags->{"Proof|ID:1867780171", "Proof|ID:1867780171-1"},
   CellID->975302909]},
 "Proof|ID:786584693"->{
  Cell[2231554, 61403, 7223, 180, 27, "ProofInfo",
   CellTags->{"Proof|ID:786584693", "Proof|ID:786584693-1"},
   CellID->1192023621]},
 "Proof|ID:786584693-1"->{
  Cell[2231554, 61403, 7223, 180, 27, "ProofInfo",
   CellTags->{"Proof|ID:786584693", "Proof|ID:786584693-1"},
   CellID->1192023621]},
 "Proof|ID:183897269"->{
  Cell[2239342, 61603, 6819, 168, 27, "ProofInfo",
   CellTags->{"Proof|ID:183897269", "Proof|ID:183897269-1"},
   CellID->1230628713]},
 "Proof|ID:183897269-1"->{
  Cell[2239342, 61603, 6819, 168, 27, "ProofInfo",
   CellTags->{"Proof|ID:183897269", "Proof|ID:183897269-1"},
   CellID->1230628713]},
 "Proof|ID:1246401530"->{
  Cell[2246682, 61790, 6833, 168, 27, "ProofInfo",
   CellTags->{"Proof|ID:1246401530", "Proof|ID:1246401530-1"},
   CellID->879169295]},
 "Proof|ID:1246401530-1"->{
  Cell[2246682, 61790, 6833, 168, 27, "ProofInfo",
   CellTags->{"Proof|ID:1246401530", "Proof|ID:1246401530-1"},
   CellID->879169295]},
 "sect::MainResults"->{
  Cell[2261928, 62162, 149, 3, 43, "Section",
   CellTags->"sect::MainResults",
   CellID->356961332]},
 "sect::Invariant"->{
  Cell[2263274, 62199, 523, 13, 38, "Subsection",
   CellTags->"sect::Invariant",
   CellID->1255092584]},
 "ID:2102726322"->{
  Cell[2268552, 62338, 1549, 43, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2102726322", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "invariant"},
   CellID->2102726322]},
 "invariant"->{
  Cell[2268552, 62338, 1549, 43, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:2102726322", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "invariant"},
   CellID->2102726322]},
 "Proof|ID:2102726322"->{
  Cell[2270258, 62392, 6517, 163, 29, "ProofInfo",
   CellTags->{"Proof|ID:2102726322", "Proof|ID:2102726322-1"},
   CellID->987040057]},
 "Proof|ID:2102726322-1"->{
  Cell[2270258, 62392, 6517, 163, 29, "ProofInfo",
   CellTags->{"Proof|ID:2102726322", "Proof|ID:2102726322-1"},
   CellID->987040057]},
 "sect::Bound"->{
  Cell[2283793, 62773, 557, 13, 30, "Subsection",
   CellTags->"sect::Bound",
   CellID->485945541]},
 "ID:724101256"->{
  Cell[2290623, 62988, 1120, 32, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:724101256", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "100"},
   CellID->724101256]},
 "100"->{
  Cell[2290623, 62988, 1120, 32, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:724101256", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "100"},
   CellID->724101256]},
 "Proof|ID:724101256"->{
  Cell[2291899, 63031, 4269, 114, 29, "ProofInfo",
   CellTags->{"Proof|ID:724101256", "Proof|ID:724101256-1"},
   CellID->205519987]},
 "Proof|ID:724101256-1"->{
  Cell[2291899, 63031, 4269, 114, 29, "ProofInfo",
   CellTags->{"Proof|ID:724101256", "Proof|ID:724101256-1"},
   CellID->205519987]},
 "ID:1990359651"->{
  Cell[2297499, 63192, 1140, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1990359651", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "bound"},
   CellID->1990359651]},
 "bound"->{
  Cell[2297499, 63192, 1140, 31, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1990359651", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "bound"},
   CellID->1990359651]},
 "Proof|ID:1990359651"->{
  Cell[2298796, 63234, 4185, 112, 29, "ProofInfo",
   CellTags->{"Proof|ID:1990359651", "Proof|ID:1990359651-1"},
   CellID->2132280415]},
 "Proof|ID:1990359651-1"->{
  Cell[2298796, 63234, 4185, 112, 29, "ProofInfo",
   CellTags->{"Proof|ID:1990359651", "Proof|ID:1990359651-1"},
   CellID->2132280415]},
 "sect::MainResult"->{
  Cell[2307390, 63481, 316, 8, 30, "Subsection",
   CellTags->"sect::MainResult",
   CellID->1121720170]},
 "ID:1891911403"->{
  Cell[2310387, 63573, 987, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1891911403", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "main theorem"},
   CellID->1891911403]},
 "main theorem"->{
  Cell[2310387, 63573, 987, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1891911403", 
     "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "main theorem"},
   CellID->1891911403]},
 "Proof|ID:1891911403"->{
  Cell[2311531, 63611, 6867, 167, 29, "ProofInfo",
   CellTags->{"Proof|ID:1891911403", "Proof|ID:1891911403-1"},
   CellID->233183322]},
 "Proof|ID:1891911403-1"->{
  Cell[2311531, 63611, 6867, 167, 29, "ProofInfo",
   CellTags->{"Proof|ID:1891911403", "Proof|ID:1891911403-1"},
   CellID->233183322]},
 "sect::Prover"->{
  Cell[2322377, 63919, 210, 4, 43, "Section",
   CellTags->"sect::Prover",
   CellID->1903011022]},
 "sect::ProverTuples"->{
  Cell[2349301, 64713, 416, 7, 30, "Subsection",
   CellTags->"sect::ProverTuples",
   CellID->1679395489]},
 "sect::ProverAddition"->{
  Cell[2378155, 65501, 379, 6, 30, "Subsection",
   CellTags->"sect::ProverAddition",
   CellID->285731009]},
 "sect::ProverOrdering"->{
  Cell[2385700, 65713, 432, 7, 30, "Subsection",
   CellTags->"sect::ProverOrdering",
   CellID->915360441]},
 "sect::ProverMinMax"->{
  Cell[2408564, 66364, 404, 7, 30, "Subsection",
   CellTags->"sect::ProverMinMax",
   CellID->200612129]},
 "sect::ProverQuantifiers"->{
  Cell[2432301, 66978, 385, 6, 30, "Subsection",
   CellTags->"sect::ProverQuantifiers",
   CellID->141176790]},
 "sect::ProverRewriting"->{
  Cell[2458290, 67708, 329, 6, 30, "Subsection",
   CellTags->"sect::ProverRewriting",
   CellID->1387546506]},
 "sect::Conclusion"->{
  Cell[2465779, 67898, 147, 3, 43, "Section",
   CellTags->"sect::Conclusion",
   CellID->1304797350]},
 "sect::References"->{
  Cell[2475277, 68160, 146, 3, 43, "Section",
   CellTags->"sect::References",
   CellID->664836793]},
 "Buchberger1965"->{
  Cell[2475448, 68167, 289, 6, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10000.},
   CellTags->"Buchberger1965",
   CellID->1095181227]},
 "Buchberger1979"->{
  Cell[2476512, 68195, 340, 7, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10001.},
   CellTags->"Buchberger1979",
   CellID->1664980398]},
 "Buchberger1979a"->{
  Cell[2477625, 68226, 388, 7, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10002.},
   CellTags->"Buchberger1979a",
   CellID->1739310831]},
 "Buchberger1983a"->{
  Cell[2478708, 68254, 314, 6, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10003.},
   CellTags->"Buchberger1983a",
   CellID->1873989285]},
 "Buchberger1983b"->{
  Cell[2479682, 68280, 364, 7, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10004.},
   CellTags->"Buchberger1983b",
   CellID->212964405]},
 "Buchberger1998"->{
  Cell[2480539, 68303, 287, 6, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10005.},
   CellTags->"Buchberger1998",
   CellID->748631075]},
 "Buchberger2003"->{
  Cell[2481258, 68326, 290, 6, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10006.},
   CellTags->"Buchberger2003",
   CellID->1719245896]},
 "Buchberger2006"->{
  Cell[2481968, 68347, 290, 6, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10007.},
   CellTags->"Buchberger2006",
   CellID->1062875853]},
 "Buchberger2014"->{
  Cell[2482743, 68371, 267, 6, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10008.},
   CellTags->"Buchberger2014",
   CellID->473088647]},
 "Giese2007"->{
  Cell[2483456, 68392, 302, 6, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10009.},
   CellTags->"Giese2007",
   CellID->723325741]},
 "Mayr-Meyer1982"->{
  Cell[2484255, 68414, 268, 6, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10011.},
   CellTags->"Mayr-Meyer1982",
   CellID->1957558217]},
 "Theorema"->{
  Cell[2484947, 68436, 277, 6, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10010.},
   CellTags->"Theorema",
   CellID->367095808]},
 "Windsteiger1999"->{
  Cell[2485925, 68465, 291, 6, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10011.},
   CellTags->"Windsteiger1999",
   CellID->470851853]},
 "Winkler1984"->{
  Cell[2486867, 68492, 283, 6, 24, "Text",
   CellGroupingRules->{GroupTogetherGrouping, 10012.},
   CellTags->"Winkler1984",
   CellID->805682494]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"sect::FormalMathematics", 2525467, 69282},
 {"sect::Conditional", 2525598, 69286},
 {"sect::ProblemSetting", 2525730, 69290},
 {"Alg::BB", 2525848, 69294},
 {"Eqn::ChainCrit", 2526027, 69299},
 {"Eqn::M", 2526138, 69303},
 {"Eqn::W", 2526239, 69307},
 {"sect::BasicNotions", 2526353, 69311},
 {"sect::DefinitionTuples", 2526485, 69315},
 {"ID:1969097229", 2526617, 69319},
 {"Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", 2526938, 69327},
 {"length is natural", 2594770, 71116},
 {"ID:1508742051", 2595024, 71123},
 {"tuple equality", 2595276, 71130},
 {"ID:1264372773", 2595527, 71137},
 {"isTuple\:293a", 2595777, 71144},
 {"ID:2086441677", 2596027, 71151},
 {"length\:293a", 2596275, 71158},
 {"ID:585712945", 2596523, 71165},
 {"element\:293a 1", 2596775, 71172},
 {"ID:1075236060", 2597025, 71179},
 {"element\:293a 2", 2597279, 71186},
 {"ID:1405024037", 2597531, 71193},
 {"ID:379346204", 2597757, 71200},
 {"domain tuple", 2598003, 71207},
 {"sect::DefinitionOC", 2598255, 71214},
 {"ID:1473378407", 2598384, 71218},
 {"other coordinate 1", 2598643, 71225},
 {"ID:406630295", 2598896, 71232},
 {"other coordinate 2", 2599153, 71239},
 {"sect::DefinitionZ", 2599409, 71246},
 {"ID:1091916109", 2599538, 71250},
 {"Z is TOM", 2599777, 71257},
 {"ID:1845867532", 2600021, 71264},
 {"Z is totally ordered", 2600284, 71271},
 {"ID:492790093", 2600539, 71278},
 {"some integers", 2600787, 71285},
 {"ID:1907318616", 2601035, 71292},
 {"1\[LessEqual]2", 2601286, 71299},
 {"ID:1708561859", 2601536, 71306},
 {"1", 2601762, 71313},
 {"ID:49894867", 2601998, 71320},
 {"2", 2602220, 71327},
 {"sect::DefinitionCategories", 2602467, 71334},
 {"sect::DefinitionAbelianMonoid", 2602618, 71338},
 {"ID:349901806", 2602758, 71342},
 {"closure", 2602994, 71349},
 {"ID:1476372605", 2603236, 71356},
 {"0 in domain", 2603482, 71363},
 {"ID:1359514637", 2603730, 71370},
 {"0 is neutral", 2603978, 71377},
 {"ID:1215514258", 2604227, 71384},
 {"associativity", 2604477, 71391},
 {"ID:1531751179", 2604727, 71398},
 {"commutativity", 2604977, 71405},
 {"ID:929558765", 2605226, 71412},
 {"cancellativity", 2605476, 71419},
 {"ID:1003575752", 2605725, 71426},
 {"Abelian monoid", 2605978, 71433},
 {"sect::DefinitionTotallyOrderedDomain", 2606253, 71440},
 {"ID:1939013113", 2606400, 71444},
 {"total ordering", 2606654, 71451},
 {"ID:1585090210", 2606907, 71458},
 {"irreflexive relation", 2607171, 71465},
 {"ID:1535164863", 2607428, 71472},
 {"strict version of ordering", 2607704, 71479},
 {"ID:795999034", 2607966, 71486},
 {"totally-ordered domain", 2608232, 71493},
 {"sect::DefinitionminmaxDomain", 2608504, 71500},
 {"ID:616702295", 2608643, 71504},
 {"min-max domain", 2608893, 71511},
 {"ID:1204372660", 2609142, 71518},
 {"min domain", 2609386, 71525},
 {"ID:1875462394", 2609633, 71532},
 {"max domain", 2609877, 71539},
 {"sect::DefinitionTOM", 2610130, 71546},
 {"ID:743220188", 2610259, 71550},
 {"totally-ordered Abelian monoid", 2610541, 71557},
 {"ID:1329821571", 2610806, 71564},
 {"monotonicity", 2611054, 71571},
 {"ID:2137457113", 2611303, 71578},
 {"D is TOM", 2611542, 71585},
 {"ID:1975341048", 2611786, 71592},
 {"D is Abelian monoid", 2612047, 71599},
 {"ID:275749992", 2612301, 71606},
 {"D is totally ordered", 2612562, 71613},
 {"ID:1904592343", 2612816, 71620},
 {"D is min-max domain", 2613077, 71627},
 {"ID:589296925", 2613331, 71634},
 {"n is integer", 2613576, 71641},
 {"ID:54512703", 2613820, 71648},
 {"n is positive", 2614065, 71655},
 {"sect::LemmataBasic", 2614315, 71662},
 {"ID:1538785027", 2614439, 71666},
 {"3", 2614665, 71673},
 {"Proof|ID:1538785027", 2614909, 71680},
 {"Proof|ID:1538785027-1", 2615073, 71684},
 {"ID:968428080", 2615228, 71688},
 {"4", 2615451, 71695},
 {"ID:680012314", 2615685, 71702},
 {"5", 2615908, 71709},
 {"ID:655751250", 2616142, 71716},
 {"6", 2616366, 71723},
 {"ID:1963081417", 2616602, 71730},
 {"7", 2616828, 71737},
 {"Proof|ID:968428080", 2617071, 71744},
 {"Proof|ID:968428080-1", 2617231, 71748},
 {"Proof|ID:680012314", 2617389, 71752},
 {"Proof|ID:680012314-1", 2617549, 71756},
 {"Proof|ID:655751250", 2617707, 71760},
 {"Proof|ID:655751250-1", 2617867, 71764},
 {"Proof|ID:1963081417", 2618026, 71768},
 {"Proof|ID:1963081417-1", 2618190, 71772},
 {"ID:1457798778", 2618346, 71776},
 {"8", 2618571, 71783},
 {"ID:1044840887", 2618808, 71790},
 {"9", 2619034, 71797},
 {"ID:2017090463", 2619272, 71804},
 {"10", 2619499, 71811},
 {"Proof|ID:1457798778", 2619743, 71818},
 {"Proof|ID:1457798778-1", 2619906, 71822},
 {"Proof|ID:1044840887", 2620067, 71826},
 {"Proof|ID:1044840887-1", 2620231, 71830},
 {"Proof|ID:2017090463", 2620393, 71834},
 {"Proof|ID:2017090463-1", 2620557, 71838},
 {"sect::LemmataOrderings", 2620722, 71842},
 {"ID:1308600401", 2620853, 71846},
 {"irreflexivity of <", 2621112, 71853},
 {"ID:914659821", 2621365, 71860},
 {"reflexivity of \[LessEqual]", 2621641, 71867},
 {"ID:966362506", 2621902, 71874},
 {"<\[Or]=", 2622138, 71881},
 {"ID:1288924674", 2622380, 71888},
 {"\[Not]\[LessEqual]", 2622640, 71895},
 {"ID:1962569111", 2622895, 71902},
 {"\[Not]<", 2623133, 71909},
 {"ID:243913977", 2623376, 71916},
 {"transitivity <\[LessEqual]", 2623650, 71923},
 {"ID:1623404218", 2623911, 71930},
 {"transitivity \[LessEqual]<", 2624187, 71937},
 {"ID:1509384446", 2624450, 71944},
 {"transitivity <<", 2624704, 71951},
 {"ID:972758509", 2624955, 71958},
 {"\[NotEqual] in terms of <", 2625227, 71965},
 {"Proof|ID:1308600401", 2625493, 71972},
 {"Proof|ID:1308600401-1", 2625655, 71976},
 {"Proof|ID:914659821", 2625814, 71980},
 {"Proof|ID:914659821-1", 2625974, 71984},
 {"Proof|ID:966362506", 2626132, 71988},
 {"Proof|ID:966362506-1", 2626293, 71992},
 {"Proof|ID:1288924674", 2626453, 71996},
 {"Proof|ID:1288924674-1", 2626618, 72000},
 {"Proof|ID:1962569111", 2626781, 72004},
 {"Proof|ID:1962569111-1", 2626946, 72008},
 {"Proof|ID:243913977", 2627108, 72012},
 {"Proof|ID:243913977-1", 2627270, 72016},
 {"Proof|ID:1623404218", 2627431, 72020},
 {"Proof|ID:1623404218-1", 2627596, 72024},
 {"Proof|ID:1509384446", 2627759, 72028},
 {"Proof|ID:1509384446-1", 2627923, 72032},
 {"Proof|ID:972758509", 2628084, 72036},
 {"Proof|ID:972758509-1", 2628246, 72040},
 {"ID:786186036", 2628400, 72044},
 {"\[LessEqual]+", 2628648, 72051},
 {"ID:1385901945", 2628896, 72058},
 {"<+", 2629124, 72065},
 {"ID:1581648423", 2629363, 72072},
 {"\[LessEqual]\[LessEqual]+", 2629637, 72079},
 {"ID:1815472507", 2629899, 72086},
 {"\[LessEqual]<+", 2630151, 72093},
 {"ID:443432455", 2630401, 72100},
 {"<<+", 2630629, 72107},
 {"Proof|ID:786186036", 2630872, 72114},
 {"Proof|ID:786186036-1", 2631034, 72118},
 {"Proof|ID:1385901945", 2631195, 72122},
 {"Proof|ID:1385901945-1", 2631360, 72126},
 {"Proof|ID:1581648423", 2631523, 72130},
 {"Proof|ID:1581648423-1", 2631688, 72134},
 {"Proof|ID:1815472507", 2631851, 72138},
 {"Proof|ID:1815472507-1", 2632017, 72142},
 {"Proof|ID:443432455", 2632180, 72146},
 {"Proof|ID:443432455-1", 2632342, 72150},
 {"ID:1546357463", 2632497, 72154},
 {"1\[LessEqual]length", 2632760, 72161},
 {"ID:409966668", 2633016, 72168},
 {"11", 2633243, 72175},
 {"ID:1768549223", 2633481, 72182},
 {"2\[LessEqual]length", 2633744, 72189},
 {"ID:1399858996", 2634001, 72196},
 {"length\:293a is integer", 2634272, 72203},
 {"ID:1353010357", 2634533, 72210},
 {"length\:293a is positive", 2634805, 72217},
 {"ID:763884008", 2635065, 72224},
 {"12", 2635292, 72231},
 {"ID:1442641059", 2635530, 72238},
 {"13", 2635759, 72245},
 {"ID:215173194", 2635998, 72252},
 {"14", 2636225, 72259},
 {"Proof|ID:1546357463", 2636469, 72266},
 {"Proof|ID:1546357463-1", 2636635, 72270},
 {"Proof|ID:409966668", 2636798, 72274},
 {"Proof|ID:409966668-1", 2636959, 72278},
 {"Proof|ID:1768549223", 2637119, 72282},
 {"Proof|ID:1768549223-1", 2637285, 72286},
 {"Proof|ID:1399858996", 2637449, 72290},
 {"Proof|ID:1399858996-1", 2637612, 72294},
 {"Proof|ID:1353010357", 2637773, 72298},
 {"Proof|ID:1353010357-1", 2637938, 72302},
 {"Proof|ID:763884008", 2638100, 72306},
 {"Proof|ID:763884008-1", 2638263, 72310},
 {"Proof|ID:1442641059", 2638425, 72314},
 {"Proof|ID:1442641059-1", 2638591, 72318},
 {"Proof|ID:215173194", 2638754, 72322},
 {"Proof|ID:215173194-1", 2638917, 72326},
 {"ID:8303793", 2639070, 72330},
 {"15", 2639293, 72337},
 {"ID:1339530720", 2639527, 72344},
 {"16", 2639756, 72351},
 {"ID:759960416", 2639995, 72358},
 {"\[SquareSubsetEqual]\:293a", 2640270, 72365},
 {"Proof|ID:8303793|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2640616, 72373},
 {"Proof|ID:1339530720|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2640916, 72380},
 {"Proof|ID:759960416|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2641218, 72387},
 {"sect::Lemmataminmax", 2641440, 72393},
 {"ID:1448857518", 2641568, 72397},
 {"defining property of min", 2641841, 72404},
 {"ID:1578657786", 2642103, 72411},
 {"defining property of max", 2642376, 72418},
 {"ID:1968885164", 2642638, 72425},
 {"type of min", 2642884, 72432},
 {"ID:1151569277", 2643132, 72439},
 {"type of max", 2643378, 72446},
 {"Proof|ID:1448857518", 2643632, 72453},
 {"Proof|ID:1448857518-1", 2643798, 72457},
 {"Proof|ID:1578657786", 2643962, 72461},
 {"Proof|ID:1578657786-1", 2644128, 72465},
 {"Proof|ID:1968885164", 2644292, 72469},
 {"Proof|ID:1968885164-1", 2644457, 72473},
 {"Proof|ID:1151569277", 2644620, 72477},
 {"Proof|ID:1151569277-1", 2644785, 72481},
 {"ID:232133128", 2644941, 72485},
 {"17", 2645168, 72492},
 {"ID:1533503972", 2645406, 72499},
 {"18", 2645635, 72506},
 {"ID:881325315", 2645874, 72513},
 {"17'", 2646103, 72520},
 {"ID:1667609383", 2646342, 72527},
 {"18'", 2646573, 72534},
 {"ID:1257756255", 2646814, 72541},
 {"min is invariant under permutations", 2647109, 72548},
 {"ID:2071551441", 2647382, 72555},
 {"max is invariant under permutations", 2647677, 72562},
 {"Proof|ID:232133128", 2647955, 72569},
 {"Proof|ID:232133128-1", 2648117, 72573},
 {"Proof|ID:1533503972", 2648278, 72577},
 {"Proof|ID:1533503972-1", 2648444, 72581},
 {"Proof|ID:881325315", 2648607, 72585},
 {"Proof|ID:881325315-1", 2648769, 72589},
 {"Proof|ID:1667609383", 2648930, 72593},
 {"Proof|ID:1667609383-1", 2649096, 72597},
 {"Proof|ID:1257756255", 2649260, 72601},
 {"Proof|ID:1257756255-1", 2649425, 72605},
 {"Proof|ID:2071551441", 2649588, 72609},
 {"Proof|ID:2071551441-1", 2649754, 72613},
 {"ID:48233088", 2649910, 72617},
 {"min\[LessEqual]", 2650161, 72624},
 {"ID:1034128987", 2650410, 72631},
 {"\[LessEqual]min", 2650665, 72638},
 {"ID:1323776552", 2650918, 72645},
 {"max\[LessEqual]", 2651173, 72652},
 {"ID:1520079070", 2651426, 72659},
 {"\[LessEqual]max", 2651681, 72666},
 {"ID:1860476328", 2651934, 72673},
 {"min<", 2652167, 72680},
 {"ID:1907558186", 2652409, 72687},
 {"<min", 2652642, 72694},
 {"ID:2008421481", 2652884, 72701},
 {"max<", 2653117, 72708},
 {"ID:1141242040", 2653359, 72715},
 {"<max", 2653592, 72722},
 {"Proof|ID:48233088", 2653838, 72729},
 {"Proof|ID:48233088-1", 2653998, 72733},
 {"Proof|ID:1034128987", 2654158, 72737},
 {"Proof|ID:1034128987-1", 2654324, 72741},
 {"Proof|ID:1323776552", 2654488, 72745},
 {"Proof|ID:1323776552-1", 2654652, 72749},
 {"Proof|ID:1520079070", 2654814, 72753},
 {"Proof|ID:1520079070-1", 2654979, 72757},
 {"Proof|ID:1860476328", 2655142, 72761},
 {"Proof|ID:1860476328-1", 2655307, 72765},
 {"Proof|ID:1907558186", 2655470, 72769},
 {"Proof|ID:1907558186-1", 2655635, 72773},
 {"Proof|ID:2008421481", 2655798, 72777},
 {"Proof|ID:2008421481-1", 2655964, 72781},
 {"Proof|ID:1141242040", 2656128, 72785},
 {"Proof|ID:1141242040-1", 2656293, 72789},
 {"ID:572537117", 2656449, 72793},
 {"19", 2656676, 72800},
 {"ID:2073093776", 2656914, 72807},
 {"min\:293a 1", 2657161, 72814},
 {"ID:1632846240", 2657410, 72821},
 {"min\:293a 2", 2657657, 72828},
 {"ID:1146565830", 2657906, 72835},
 {"20", 2658135, 72842},
 {"ID:1172583037", 2658375, 72849},
 {"max\:293a 1", 2658622, 72856},
 {"ID:2142810348", 2658871, 72863},
 {"max\:293a 2", 2659118, 72870},
 {"ID:1460290836", 2659367, 72877},
 {"min\:293a", 2659610, 72884},
 {"ID:1931124773", 2659857, 72891},
 {"max\:293a", 2660100, 72898},
 {"ID:978507648", 2660346, 72905},
 {"min\:293a decreases", 2660606, 72912},
 {"ID:2097939500", 2660860, 72919},
 {"max\:293a increases", 2661122, 72926},
 {"Proof|ID:572537117", 2661383, 72933},
 {"Proof|ID:572537117-1", 2661546, 72937},
 {"Proof|ID:2073093776", 2661708, 72941},
 {"Proof|ID:2073093776-1", 2661873, 72945},
 {"Proof|ID:1632846240", 2662036, 72949},
 {"Proof|ID:1632846240-1", 2662201, 72953},
 {"Proof|ID:1146565830", 2662364, 72957},
 {"Proof|ID:1146565830-1", 2662530, 72961},
 {"Proof|ID:1172583037", 2662694, 72965},
 {"Proof|ID:1172583037-1", 2662859, 72969},
 {"Proof|ID:2142810348", 2663022, 72973},
 {"Proof|ID:2142810348-1", 2663188, 72977},
 {"Proof|ID:1460290836", 2663352, 72981},
 {"Proof|ID:1460290836-1", 2663518, 72985},
 {"Proof|ID:1931124773", 2663682, 72989},
 {"Proof|ID:1931124773-1", 2663846, 72993},
 {"Proof|ID:978507648", 2664007, 72997},
 {"Proof|ID:978507648-1", 2664170, 73001},
 {"Proof|ID:2097939500", 2664332, 73005},
 {"Proof|ID:2097939500-1", 2664497, 73009},
 {"ID:900399941", 2664653, 73013},
 {"21", 2664880, 73020},
 {"ID:1833484473", 2665118, 73027},
 {"min+ 1", 2665355, 73034},
 {"ID:1976058899", 2665599, 73041},
 {"min+ 2", 2665836, 73048},
 {"ID:2067845318", 2666080, 73055},
 {"min+ 3", 2666317, 73062},
 {"ID:446229453", 2666560, 73069},
 {"min+", 2666791, 73076},
 {"ID:437038980", 2667030, 73083},
 {"max+ 1", 2667265, 73090},
 {"ID:367539508", 2667506, 73097},
 {"max+ 2", 2667741, 73104},
 {"ID:1689218301", 2667983, 73111},
 {"max+ 3", 2668220, 73118},
 {"ID:513276857", 2668463, 73125},
 {"max+", 2668694, 73132},
 {"Proof|ID:900399941", 2668939, 73139},
 {"Proof|ID:900399941-1", 2669102, 73143},
 {"Proof|ID:1833484473", 2669264, 73147},
 {"Proof|ID:1833484473-1", 2669429, 73151},
 {"Proof|ID:1976058899", 2669592, 73155},
 {"Proof|ID:1976058899-1", 2669758, 73159},
 {"Proof|ID:2067845318", 2669922, 73163},
 {"Proof|ID:2067845318-1", 2670088, 73167},
 {"Proof|ID:446229453", 2670251, 73171},
 {"Proof|ID:446229453-1", 2670413, 73175},
 {"Proof|ID:437038980", 2670573, 73179},
 {"Proof|ID:437038980-1", 2670735, 73183},
 {"Proof|ID:367539508", 2670895, 73187},
 {"Proof|ID:367539508-1", 2671058, 73191},
 {"Proof|ID:1689218301", 2671220, 73195},
 {"Proof|ID:1689218301-1", 2671385, 73199},
 {"Proof|ID:513276857", 2671547, 73203},
 {"Proof|ID:513276857-1", 2671710, 73207},
 {"sect::SpecificNotions", 2671874, 73211},
 {"ID:1076035631", 2672002, 73215},
 {"ID:610987882", 2672229, 73222},
 {"isExponentVector", 2672484, 73229},
 {"ID:468249611", 2672735, 73236},
 {"isExpVectorTuple", 2672990, 73243},
 {"ID:444732983", 2673241, 73250},
 {"divisibility", 2673488, 73257},
 {"ID:1111389324", 2673736, 73264},
 {"degree", 2673973, 73271},
 {"ID:1699354265", 2674217, 73278},
 {"maxdeg", 2674454, 73285},
 {"ID:1614921088", 2674698, 73292},
 {"lcm", 2674929, 73299},
 {"ID:1784038925", 2675170, 73306},
 {"chain criterion", 2675425, 73313},
 {"ID:312341806", 2675677, 73320},
 {"width w.r.t. k-th coordinate", 2675956, 73327},
 {"ID:282925385", 2676219, 73334},
 {"width", 2676452, 73341},
 {"ID:1664335746", 2676693, 73348},
 {"maximum degree of admissible pairs of one object", 2677014, 73355},
 {"ID:640444143", 2677299, 73362},
 {"maximum degree of admissible pairs", 2677590, 73369},
 {"ID:1923113076", 2677860, 73376},
 {"corner of rectangular region", 2678141, 73383},
 {"ID:1774661003", 2678407, 73390},
 {"above", 2678642, 73397},
 {"ID:285552196", 2678884, 73404},
 {"far exterior w.r.t. k-th coordinate", 2679177, 73411},
 {"ID:494773971", 2679447, 73418},
 {"rectangular region", 2679706, 73425},
 {"ID:1278822860", 2679960, 73432},
 {"chainCritDiagonal w.r.t. k-th coordinate", 2680266, 73439},
 {"ID:1458824241", 2680545, 73446},
 {"chainCritDiagonal", 2680804, 73453},
 {"sect::Computations", 2681064, 73460},
 {"ID:1559569657", 2681189, 73464},
 {"D=N0", 2681421, 73471},
 {"ID:977095909", 2681661, 73478},
 {"ID:615519889", 2681885, 73485},
 {"min in N0", 2682125, 73492},
 {"ID:1245684717", 2682369, 73499},
 {"max in N0", 2682611, 73506},
 {"ID:1526008058", 2682857, 73513},
 {"n=2", 2683087, 73520},
 {"ID:1544493310", 2683327, 73527},
 {"n=3", 2683557, 73534},
 {"sect::LemmataSpecific", 2683805, 73541},
 {"sect::LemmataSpecificGeneral", 2683948, 73545},
 {"ID:18996081", 2684084, 73549},
 {"22", 2684309, 73556},
 {"Proof|ID:18996081", 2684549, 73563},
 {"Proof|ID:18996081-1", 2684708, 73567},
 {"ID:313023917", 2684860, 73571},
 {"isExpVectorTuple\:293a", 2685126, 73578},
 {"ID:1697603634", 2685383, 73585},
 {"23", 2685611, 73592},
 {"Proof|ID:313023917|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2685936, 73600},
 {"Proof|ID:1697603634", 2686156, 73606},
 {"Proof|ID:1697603634-1", 2686321, 73610},
 {"ID:1313148627", 2686478, 73614},
 {"antisymmetry of divisibility", 2686759, 73621},
 {"ID:785542046", 2687024, 73628},
 {"reflexivity of divisibility", 2687300, 73635},
 {"ID:1055765423", 2687562, 73642},
 {"transitivity of divisibility", 2687843, 73649},
 {"Proof|ID:1313148627", 2688115, 73656},
 {"Proof|ID:1313148627-1", 2688279, 73660},
 {"Proof|ID:785542046", 2688440, 73664},
 {"Proof|ID:785542046-1", 2688601, 73668},
 {"Proof|ID:1055765423", 2688761, 73672},
 {"Proof|ID:1055765423-1", 2688925, 73676},
 {"ID:1918380148", 2689081, 73680},
 {"type of lcm", 2689327, 73687},
 {"ID:1065951991", 2689575, 73694},
 {"defining property of lcm 1", 2689851, 73701},
 {"ID:1071215803", 2690114, 73708},
 {"defining property of lcm 2", 2690390, 73715},
 {"ID:1028467786", 2690653, 73722},
 {"defining property of lcm 3", 2690930, 73729},
 {"ID:98270815", 2691192, 73736},
 {"commutativity of lcm", 2691452, 73743},
 {"ID:622866956", 2691704, 73750},
 {"24", 2691930, 73757},
 {"ID:810936059", 2692166, 73764},
 {"idempotency of lcm", 2692424, 73771},
 {"Proof|ID:1918380148", 2692683, 73778},
 {"Proof|ID:1918380148-1", 2692848, 73782},
 {"Proof|ID:1065951991", 2693011, 73786},
 {"Proof|ID:1065951991-1", 2693176, 73790},
 {"Proof|ID:1071215803", 2693339, 73794},
 {"Proof|ID:1071215803-1", 2693504, 73798},
 {"Proof|ID:1028467786", 2693667, 73802},
 {"Proof|ID:1028467786-1", 2693832, 73806},
 {"Proof|ID:98270815", 2693993, 73810},
 {"Proof|ID:98270815-1", 2694151, 73814},
 {"Proof|ID:622866956", 2694308, 73818},
 {"Proof|ID:622866956-1", 2694469, 73822},
 {"Proof|ID:810936059", 2694628, 73826},
 {"Proof|ID:810936059-1", 2694789, 73830},
 {"ID:1822207066", 2694943, 73834},
 {"type of deg", 2695189, 73841},
 {"ID:455032720", 2695436, 73848},
 {"25", 2695662, 73855},
 {"ID:839534060", 2695898, 73862},
 {"26", 2696124, 73869},
 {"Proof|ID:1822207066", 2696367, 73876},
 {"Proof|ID:1822207066-1", 2696531, 73880},
 {"Proof|ID:455032720|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2696773, 73885},
 {"Proof|ID:839534060|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2697074, 73892},
 {"ID:2123737964", 2697288, 73898},
 {"type of maxdeg", 2697540, 73905},
 {"ID:647995479", 2697790, 73912},
 {"27", 2698017, 73919},
 {"Proof|ID:2123737964|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2698342, 73927},
 {"Proof|ID:647995479|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2698644, 73934},
 {"sect::LemmataSpecificGeneralM", 2698875, 73940},
 {"ID:1452963633", 2699016, 73944},
 {"28", 2699244, 73951},
 {"ID:1500097429", 2699483, 73958},
 {"29", 2699711, 73965},
 {"ID:279915377", 2699949, 73972},
 {"symmetry of chainCrit", 2700214, 73979},
 {"ID:1391995733", 2700471, 73986},
 {"divisibility\[DoubleRightArrow]chainCrit", 2700776, 73993},
 {"ID:1271079949", 2701054, 74000},
 {"reflexivity of chainCrit", 2701326, 74007},
 {"ID:488309194", 2701586, 74014},
 {"chainCrit w.r.t. singleton", 2701861, 74021},
 {"ID:1322351588", 2702123, 74028},
 {"chainCrit\:293a 1", 2702382, 74035},
 {"ID:1269486125", 2702637, 74042},
 {"chainCrit\:293a 2", 2702896, 74049},
 {"ID:941129646", 2703150, 74056},
 {"chainCrit\:293a", 2703403, 74063},
 {"ID:1657568255", 2703654, 74070},
 {"chainCrit\:293a I", 2703913, 74077},
 {"Proof|ID:1452963633|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2704255, 74085},
 {"Proof|ID:1500097429|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2704558, 74092},
 {"Proof|ID:279915377|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2704860, 74099},
 {"Proof|ID:1391995733|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2705162, 74106},
 {"Proof|ID:1271079949", 2705383, 74112},
 {"Proof|ID:1271079949-1", 2705548, 74116},
 {"Proof|ID:488309194|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2705791, 74121},
 {"Proof|ID:1322351588|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2706094, 74128},
 {"Proof|ID:1269486125|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2706398, 74135},
 {"Proof|ID:941129646|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2706700, 74142},
 {"Proof|ID:1657568255|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2707002, 74149},
 {"ID:469812703", 2707217, 74155},
 {"30", 2707444, 74162},
 {"ID:1053941118", 2707682, 74169},
 {"chainCrit holds for some element", 2707971, 74176},
 {"Proof|ID:469812703|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2708327, 74184},
 {"Proof|ID:1053941118|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2708630, 74191},
 {"ID:1635347335", 2708847, 74197},
 {"31", 2709077, 74204},
 {"ID:1901113035", 2709318, 74211},
 {"32", 2709548, 74218},
 {"ID:1760547878", 2709789, 74225},
 {"type of M (binary)", 2710050, 74232},
 {"ID:1896797254", 2710306, 74239},
 {"type of M (unary)", 2710565, 74246},
 {"ID:1611647919", 2710820, 74253},
 {"33", 2711050, 74260},
 {"Proof|ID:1635347335|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2711378, 74268},
 {"Proof|ID:1901113035|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2711682, 74275},
 {"Proof|ID:1760547878|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2711986, 74282},
 {"Proof|ID:1896797254|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2712290, 74289},
 {"Proof|ID:1611647919|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2712593, 74296},
 {"ID:1545946275", 2712811, 74302},
 {"\[LessEqual]M (unary)", 2713079, 74309},
 {"ID:1018591902", 2713339, 74316},
 {"M\[LessEqual] (unary)", 2713607, 74323},
 {"ID:382103250", 2713866, 74330},
 {"\[LessEqual]M (binary)", 2714134, 74337},
 {"ID:450457679", 2714392, 74344},
 {"M\[LessEqual] (binary)", 2714660, 74351},
 {"ID:643714041", 2714918, 74358},
 {"M+\[LessEqual]", 2715170, 74365},
 {"Proof|ID:1545946275", 2715427, 74372},
 {"Proof|ID:1545946275-1", 2715594, 74376},
 {"Proof|ID:1018591902", 2715759, 74380},
 {"Proof|ID:1018591902-1", 2715926, 74384},
 {"Proof|ID:382103250", 2716090, 74388},
 {"Proof|ID:382103250-1", 2716253, 74392},
 {"Proof|ID:450457679", 2716414, 74396},
 {"Proof|ID:450457679-1", 2716577, 74400},
 {"Proof|ID:643714041", 2716738, 74404},
 {"Proof|ID:643714041-1", 2716901, 74408},
 {"ID:1934562955", 2717057, 74412},
 {"34", 2717287, 74419},
 {"ID:1993966556", 2717528, 74426},
 {"35", 2717758, 74433},
 {"ID:1265284403", 2717999, 74440},
 {"36", 2718229, 74447},
 {"ID:934408883", 2718469, 74454},
 {"37", 2718697, 74461},
 {"ID:1342720412", 2718936, 74468},
 {"38", 2719166, 74475},
 {"ID:192359742", 2719406, 74482},
 {"39", 2719634, 74489},
 {"ID:2005322529", 2719873, 74496},
 {"40", 2720103, 74503},
 {"Proof|ID:1934562955|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2720431, 74511},
 {"Proof|ID:1993966556|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2720735, 74518},
 {"Proof|ID:1265284403", 2720958, 74524},
 {"Proof|ID:1265284403-1", 2721124, 74528},
 {"Proof|ID:934408883", 2721287, 74532},
 {"Proof|ID:934408883-1", 2721450, 74536},
 {"Proof|ID:192359742", 2721611, 74540},
 {"Proof|ID:192359742-1", 2721773, 74544},
 {"Proof|ID:2005322529", 2721934, 74548},
 {"Proof|ID:2005322529-1", 2722100, 74552},
 {"ID:984666222", 2722257, 74556},
 {"41", 2722485, 74563},
 {"ID:1616352426", 2722724, 74570},
 {"42", 2722954, 74577},
 {"Proof|ID:984666222", 2723200, 74584},
 {"Proof|ID:984666222-1", 2723363, 74588},
 {"Proof|ID:1616352426", 2723525, 74592},
 {"Proof|ID:1616352426-1", 2723692, 74596},
 {"ID:1399613937", 2723851, 74600},
 {"M\:293a 1", 2724095, 74607},
 {"ID:1210171526", 2724343, 74614},
 {"M\:293a 2", 2724587, 74621},
 {"ID:1251355025", 2724835, 74628},
 {"M\:293a 3", 2725079, 74635},
 {"ID:1093640190", 2725327, 74642},
 {"M\:293a 4", 2725571, 74649},
 {"ID:606212147", 2725818, 74656},
 {"M\:293a 5", 2726060, 74663},
 {"ID:1628665526", 2726306, 74670},
 {"M\:293a", 2726546, 74677},
 {"Proof|ID:1399613937", 2726798, 74684},
 {"Proof|ID:1399613937-1", 2726965, 74688},
 {"Proof|ID:1210171526", 2727130, 74692},
 {"Proof|ID:1210171526-1", 2727297, 74696},
 {"Proof|ID:1251355025", 2727462, 74700},
 {"Proof|ID:1251355025-1", 2727628, 74704},
 {"Proof|ID:1093640190", 2727792, 74708},
 {"Proof|ID:1093640190-1", 2727958, 74712},
 {"Proof|ID:606212147", 2728121, 74716},
 {"Proof|ID:606212147-1", 2728285, 74720},
 {"Proof|ID:1628665526", 2728448, 74724},
 {"Proof|ID:1628665526-1", 2728614, 74728},
 {"ID:1351181837", 2728772, 74732},
 {"M\:293a I", 2729016, 74739},
 {"ID:2057269766", 2729264, 74746},
 {"M\:293a\[LessEqual]", 2729528, 74753},
 {"ID:94917498", 2729784, 74760},
 {"M\:293a II", 2730026, 74767},
 {"ID:1065272829", 2730271, 74774},
 {"M\:293a+\[LessEqual]", 2730537, 74781},
 {"Proof|ID:1351181837", 2730802, 74788},
 {"Proof|ID:1351181837-1", 2730969, 74792},
 {"Proof|ID:2057269766", 2731134, 74796},
 {"Proof|ID:2057269766-1", 2731300, 74800},
 {"Proof|ID:94917498", 2731462, 74804},
 {"Proof|ID:94917498-1", 2731623, 74808},
 {"Proof|ID:1065272829", 2731784, 74812},
 {"Proof|ID:1065272829-1", 2731951, 74816},
 {"ID:1824645721", 2732110, 74820},
 {"type of W (binary)", 2732371, 74827},
 {"ID:1345307293", 2732627, 74834},
 {"type of W (unary)", 2732886, 74841},
 {"ID:1185705395", 2733141, 74848},
 {"43", 2733371, 74855},
 {"Proof|ID:1824645721", 2733618, 74862},
 {"Proof|ID:1824645721-1", 2733784, 74866},
 {"Proof|ID:1345307293", 2733948, 74870},
 {"Proof|ID:1345307293-1", 2734112, 74874},
 {"Proof|ID:1185705395", 2734274, 74878},
 {"Proof|ID:1185705395-1", 2734441, 74882},
 {"ID:354042430", 2734599, 74886},
 {"\[LessEqual]W", 2734849, 74893},
 {"ID:551974343", 2735098, 74900},
 {"W\[LessEqual]", 2735348, 74907},
 {"Proof|ID:354042430", 2735603, 74914},
 {"Proof|ID:354042430-1", 2735767, 74918},
 {"Proof|ID:551974343", 2735929, 74922},
 {"Proof|ID:551974343-1", 2736092, 74926},
 {"ID:894623443", 2736247, 74930},
 {"44", 2736474, 74937},
 {"ID:1995654247", 2736712, 74944},
 {"45", 2736942, 74951},
 {"ID:1916087968", 2737183, 74958},
 {"46", 2737413, 74965},
 {"Proof|ID:894623443", 2737659, 74972},
 {"Proof|ID:894623443-1", 2737822, 74976},
 {"Proof|ID:1995654247", 2737984, 74980},
 {"Proof|ID:1995654247-1", 2738151, 74984},
 {"Proof|ID:1916087968", 2738316, 74988},
 {"Proof|ID:1916087968-1", 2738483, 74992},
 {"ID:2064198302", 2738642, 74996},
 {"W\:293a", 2738882, 75003},
 {"ID:2144277606", 2739128, 75010},
 {"W\:293a I", 2739371, 75017},
 {"ID:388422476", 2739617, 75024},
 {"W\:293a II", 2739861, 75031},
 {"ID:144808276", 2740107, 75038},
 {"W\:293a decreases", 2740365, 75045},
 {"Proof|ID:2064198302", 2740625, 75052},
 {"Proof|ID:2064198302-1", 2740791, 75056},
 {"Proof|ID:2144277606", 2740955, 75060},
 {"Proof|ID:2144277606-1", 2741121, 75064},
 {"Proof|ID:388422476", 2741284, 75068},
 {"Proof|ID:388422476-1", 2741447, 75072},
 {"Proof|ID:144808276", 2741608, 75076},
 {"Proof|ID:144808276-1", 2741772, 75080},
 {"ID:1149216477", 2741929, 75084},
 {"type of L", 2742173, 75091},
 {"Proof|ID:1149216477|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb", 2742508, 75099},
 {"sect::LemmataSpecificSpecial", 2742739, 75105},
 {"ID:504163160", 2742877, 75109},
 {"ID:649536122", 2743102, 75116},
 {"sect::Cover", 2743327, 75123},
 {"ID:465693853", 2743450, 75127},
 {"47", 2743678, 75134},
 {"ID:777205621", 2743916, 75141},
 {"48", 2744144, 75148},
 {"ID:42909295", 2744381, 75155},
 {"49", 2744607, 75162},
 {"ID:112873953", 2744843, 75169},
 {"50", 2745071, 75176},
 {"Proof|ID:465693853", 2745315, 75183},
 {"Proof|ID:465693853-1", 2745479, 75187},
 {"Proof|ID:777205621", 2745641, 75191},
 {"Proof|ID:777205621-1", 2745805, 75195},
 {"Proof|ID:42909295", 2745966, 75199},
 {"Proof|ID:42909295-1", 2746127, 75203},
 {"Proof|ID:112873953", 2746287, 75207},
 {"Proof|ID:112873953-1", 2746451, 75211},
 {"ID:315092621", 2746607, 75215},
 {"cover", 2746841, 75222},
 {"Proof|ID:315092621", 2747088, 75229},
 {"Proof|ID:315092621-1", 2747251, 75233},
 {"ID:744007071", 2747406, 75237},
 {"51", 2747634, 75244},
 {"ID:1720661645", 2747873, 75251},
 {"52", 2748103, 75258},
 {"ID:1022330126", 2748344, 75265},
 {"\[VerticalBar]lcm", 2748604, 75272},
 {"Proof|ID:744007071", 2748865, 75279},
 {"Proof|ID:744007071-1", 2749029, 75283},
 {"Proof|ID:1720661645", 2749192, 75287},
 {"Proof|ID:1720661645-1", 2749358, 75291},
 {"Proof|ID:1022330126", 2749522, 75295},
 {"Proof|ID:1022330126-1", 2749688, 75299},
 {"ID:528222346", 2749845, 75303},
 {"53", 2750073, 75310},
 {"ID:1598627200", 2750312, 75317},
 {"54", 2750542, 75324},
 {"Proof|ID:528222346", 2750788, 75331},
 {"Proof|ID:528222346-1", 2750952, 75335},
 {"Proof|ID:1598627200", 2751115, 75339},
 {"Proof|ID:1598627200-1", 2751280, 75343},
 {"ID:1537560967", 2751437, 75347},
 {"55", 2751667, 75354},
 {"ID:507460317", 2751907, 75361},
 {"56", 2752135, 75368},
 {"Proof|ID:1537560967", 2752380, 75375},
 {"Proof|ID:1537560967-1", 2752547, 75379},
 {"Proof|ID:507460317", 2752711, 75383},
 {"Proof|ID:507460317-1", 2752874, 75387},
 {"sect::LemmataSpecificSpecialW", 2753046, 75391},
 {"Lem::W[A<-x]", 2753188, 75395},
 {"ID:1442383873", 2753319, 75399},
 {"57", 2753548, 75406},
 {"ID:523824248", 2753787, 75413},
 {"W\:293a (binary) FarExterior 1", 2754071, 75420},
 {"ID:1663623856", 2754338, 75427},
 {"W\:293a (binary) FarExterior 2", 2754624, 75434},
 {"Proof|ID:1442383873", 2754899, 75441},
 {"Proof|ID:1442383873-1", 2755066, 75445},
 {"Proof|ID:523824248", 2755230, 75449},
 {"Proof|ID:523824248-1", 2755393, 75453},
 {"Proof|ID:1663623856", 2755555, 75457},
 {"Proof|ID:1663623856-1", 2755721, 75461},
 {"ID:1829996465", 2755879, 75465},
 {"W\:293a (unary) FarExterior 1", 2756163, 75472},
 {"ID:842592091", 2756430, 75479},
 {"W\:293a (unary) FarExterior 2", 2756712, 75486},
 {"ID:1051080023", 2756978, 75493},
 {"W\:293a FarExterior", 2757242, 75500},
 {"Proof|ID:1829996465", 2757506, 75507},
 {"Proof|ID:1829996465-1", 2757672, 75511},
 {"Proof|ID:842592091", 2757835, 75515},
 {"Proof|ID:842592091-1", 2757998, 75519},
 {"Proof|ID:1051080023", 2758160, 75523},
 {"Proof|ID:1051080023-1", 2758326, 75527},
 {"ID:631716597", 2758483, 75531},
 {"58", 2758710, 75538},
 {"ID:1988188117", 2758948, 75545},
 {"59", 2759177, 75552},
 {"Proof|ID:631716597", 2759422, 75559},
 {"Proof|ID:631716597-1", 2759586, 75563},
 {"Proof|ID:1988188117", 2759749, 75567},
 {"Proof|ID:1988188117-1", 2759916, 75571},
 {"ID:590870546", 2760074, 75575},
 {"ID:501569980", 2760300, 75582},
 {"60", 2760528, 75589},
 {"ID:914554779", 2760766, 75596},
 {"61", 2760994, 75603},
 {"ID:1336064727", 2761233, 75610},
 {"62", 2761463, 75617},
 {"ID:1651077201", 2761704, 75624},
 {"L+\[LessEqual]", 2761958, 75631},
 {"Proof|ID:501569980", 2762216, 75638},
 {"Proof|ID:501569980-1", 2762380, 75642},
 {"Proof|ID:914554779", 2762542, 75646},
 {"Proof|ID:914554779-1", 2762703, 75650},
 {"Proof|ID:1336064727", 2762863, 75654},
 {"Proof|ID:1336064727-1", 2763029, 75658},
 {"Proof|ID:1651077201", 2763193, 75662},
 {"Proof|ID:1651077201-1", 2763359, 75666},
 {"sect::chainCritSpecial", 2763526, 75670},
 {"ID:53741691", 2763659, 75674},
 {"63", 2763885, 75681},
 {"ID:1045358942", 2764122, 75688},
 {"64", 2764352, 75695},
 {"ID:1132937314", 2764593, 75702},
 {"65", 2764823, 75709},
 {"ID:152254973", 2765063, 75716},
 {"66", 2765291, 75723},
 {"Proof|ID:53741691", 2765534, 75730},
 {"Proof|ID:53741691-1", 2765695, 75734},
 {"Proof|ID:1045358942", 2765856, 75738},
 {"Proof|ID:1045358942-1", 2766023, 75742},
 {"Proof|ID:1132937314", 2766188, 75746},
 {"Proof|ID:1132937314-1", 2766354, 75750},
 {"Proof|ID:152254973", 2766517, 75754},
 {"Proof|ID:152254973-1", 2766681, 75758},
 {"ID:836877533", 2766837, 75762},
 {"symmetry of chainCritDiagonal 1", 2767123, 75769},
 {"ID:1226632396", 2767391, 75776},
 {"symmetry of chainCritDiagonal 2", 2767679, 75783},
 {"ID:457131364", 2767948, 75790},
 {"symmetry of chainCritDiagonal 3", 2768234, 75797},
 {"ID:1490326930", 2768502, 75804},
 {"symmetry of chainCritDiagonal", 2768786, 75811},
 {"ID:749802963", 2769053, 75818},
 {"chainCritDiagonal w.r.t. first coordinate", 2769360, 75825},
 {"ID:505706706", 2769638, 75832},
 {"67", 2769866, 75839},
 {"ID:1262287499", 2770105, 75846},
 {"68", 2770335, 75853},
 {"Proof|ID:836877533", 2770581, 75860},
 {"Proof|ID:836877533-1", 2770744, 75864},
 {"Proof|ID:1226632396", 2770906, 75868},
 {"Proof|ID:1226632396-1", 2771072, 75872},
 {"Proof|ID:457131364", 2771235, 75876},
 {"Proof|ID:457131364-1", 2771399, 75880},
 {"Proof|ID:1490326930", 2771562, 75884},
 {"Proof|ID:1490326930-1", 2771728, 75888},
 {"Proof|ID:749802963", 2771891, 75892},
 {"Proof|ID:749802963-1", 2772056, 75896},
 {"Proof|ID:505706706", 2772219, 75900},
 {"Proof|ID:505706706-1", 2772382, 75904},
 {"Proof|ID:1262287499", 2772544, 75908},
 {"Proof|ID:1262287499-1", 2772712, 75912},
 {"ID:1909233179", 2772872, 75916},
 {"69", 2773102, 75923},
 {"ID:2125747942", 2773343, 75930},
 {"70", 2773573, 75937},
 {"ID:1211430036", 2773814, 75944},
 {"71", 2774044, 75951},
 {"ID:2061027934", 2774285, 75958},
 {"72", 2774514, 75965},
 {"ID:1653917255", 2774754, 75972},
 {"73", 2774983, 75979},
 {"Proof|ID:1909233179", 2775229, 75986},
 {"Proof|ID:1909233179-1", 2775397, 75990},
 {"Proof|ID:2125747942", 2775563, 75994},
 {"Proof|ID:2125747942-1", 2775730, 75998},
 {"Proof|ID:1211430036", 2775895, 76002},
 {"Proof|ID:1211430036-1", 2776063, 76006},
 {"Proof|ID:2061027934", 2776229, 76010},
 {"Proof|ID:2061027934-1", 2776395, 76014},
 {"Proof|ID:1653917255", 2776559, 76018},
 {"Proof|ID:1653917255-1", 2776725, 76022},
 {"ID:1690170957", 2776883, 76026},
 {"chainCrit bivariate 1", 2777151, 76033},
 {"ID:271680825", 2777410, 76040},
 {"chainCrit bivariate 2", 2777675, 76047},
 {"ID:1572461042", 2777932, 76054},
 {"chainCrit bivariate 3", 2778202, 76061},
 {"ID:2046865956", 2778464, 76068},
 {"chainCrit bivariate 4", 2778732, 76075},
 {"ID:917844397", 2778991, 76082},
 {"chainCrit bivariate 5", 2779257, 76089},
 {"ID:810037327", 2779514, 76096},
 {"chainCrit bivariate", 2779776, 76103},
 {"ID:2111681775", 2780032, 76110},
 {"chainCrit bivariate I", 2780300, 76117},
 {"Proof|ID:1690170957", 2780566, 76124},
 {"Proof|ID:1690170957-1", 2780733, 76128},
 {"Proof|ID:271680825", 2780897, 76132},
 {"Proof|ID:271680825-1", 2781061, 76136},
 {"Proof|ID:1572461042", 2781224, 76140},
 {"Proof|ID:1572461042-1", 2781392, 76144},
 {"Proof|ID:2046865956", 2781558, 76148},
 {"Proof|ID:2046865956-1", 2781725, 76152},
 {"Proof|ID:917844397", 2781889, 76156},
 {"Proof|ID:917844397-1", 2782052, 76160},
 {"Proof|ID:810037327", 2782213, 76164},
 {"Proof|ID:810037327-1", 2782377, 76168},
 {"Proof|ID:2111681775", 2782540, 76172},
 {"Proof|ID:2111681775-1", 2782706, 76176},
 {"sect::Theorems", 2782865, 76180},
 {"sect::M\[RawStar]", 2782991, 76184},
 {"ID:2973275", 2783115, 76188},
 {"M\[RawStar] binary", 2783371, 76195},
 {"ID:1258784171", 2783622, 76202},
 {"M\[RawStar] unary", 2783882, 76209},
 {"ID:1549582561", 2784138, 76216},
 {"chainCrit holds for distinct indices", 2784436, 76223},
 {"ID:1966722449", 2784711, 76230},
 {"74", 2784941, 76237},
 {"ID:164254989", 2785181, 76244},
 {"75", 2785409, 76251},
 {"ID:290364666", 2785647, 76258},
 {"type of M\[RawStar] (binary)", 2785927, 76265},
 {"ID:1063863080", 2786192, 76272},
 {"type of M\[RawStar] (unary)", 2786471, 76279},
 {"Proof|ID:1549582561", 2786742, 76286},
 {"Proof|ID:1549582561-1", 2786909, 76290},
 {"Proof|ID:1966722449", 2787074, 76294},
 {"Proof|ID:1966722449-1", 2787240, 76298},
 {"Proof|ID:164254989", 2787403, 76302},
 {"Proof|ID:164254989-1", 2787566, 76306},
 {"Proof|ID:290364666", 2787727, 76310},
 {"Proof|ID:290364666-1", 2787891, 76314},
 {"Proof|ID:1063863080", 2788054, 76318},
 {"Proof|ID:1063863080-1", 2788219, 76322},
 {"ID:1680577066", 2788376, 76326},
 {"\[LessEqual]M\[RawStar] (unary)", 2788664, 76333},
 {"ID:1866021817", 2788934, 76340},
 {"M\[RawStar]\[LessEqual] (unary)", 2789222, 76347},
 {"ID:1272742036", 2789492, 76354},
 {"\[LessEqual]M\[RawStar] (binary)", 2789782, 76361},
 {"ID:462890005", 2790052, 76368},
 {"M\[RawStar]\[LessEqual] (binary)", 2790340, 76375},
 {"Proof|ID:1680577066", 2790615, 76382},
 {"Proof|ID:1680577066-1", 2790781, 76386},
 {"Proof|ID:1866021817", 2790945, 76390},
 {"Proof|ID:1866021817-1", 2791111, 76394},
 {"Proof|ID:1272742036", 2791275, 76398},
 {"Proof|ID:1272742036-1", 2791442, 76402},
 {"Proof|ID:462890005", 2791606, 76406},
 {"Proof|ID:462890005-1", 2791770, 76410},
 {"ID:929561290", 2791926, 76414},
 {"M\[LessEqual]M\[RawStar]", 2792197, 76421},
 {"ID:795795393", 2792456, 76428},
 {"M\[RawStar]\[LessEqual]M", 2792727, 76435},
 {"ID:872843574", 2792986, 76442},
 {"M=M\[RawStar]", 2793236, 76449},
 {"Proof|ID:929561290", 2793491, 76456},
 {"Proof|ID:929561290-1", 2793654, 76460},
 {"Proof|ID:795795393", 2793815, 76464},
 {"Proof|ID:795795393-1", 2793979, 76468},
 {"Proof|ID:872843574", 2794141, 76472},
 {"Proof|ID:872843574-1", 2794304, 76476},
 {"sect::TheoremsGeneral", 2794468, 76480},
 {"???", 2794590, 76484},
 {"ID:642765392", 2795112, 76500},
 {"76", 2795340, 76507},
 {"ID:592779258", 2795578, 76514},
 {"maxdeg\[LessEqual]M", 2795839, 76521},
 {"Proof|ID:642765392", 2796099, 76528},
 {"Proof|ID:642765392-1", 2796263, 76532},
 {"Proof|ID:592779258", 2796425, 76536},
 {"Proof|ID:592779258-1", 2796587, 76540},
 {"ID:691015215", 2796741, 76544},
 {"77", 2796968, 76551},
 {"ID:402511110", 2797205, 76558},
 {"78", 2797433, 76565},
 {"Proof|ID:691015215", 2797677, 76572},
 {"Proof|ID:691015215-1", 2797839, 76576},
 {"Proof|ID:402511110", 2797999, 76580},
 {"Proof|ID:402511110-1", 2798163, 76584},
 {"ID:341822661", 2798319, 76588},
 {"M\[LessEqual]L", 2798570, 76595},
 {"Proof|ID:341822661", 2798825, 76602},
 {"Proof|ID:341822661-1", 2798988, 76606},
 {"sect::TheoremsSpecial", 2799152, 76610},
 {"ID:245124153", 2799282, 76614},
 {"79", 2799510, 76621},
 {"ID:952627996", 2799748, 76628},
 {"80", 2799976, 76635},
 {"ID:2013760956", 2800215, 76642},
 {"81", 2800445, 76649},
 {"ID:1011330545", 2800686, 76656},
 {"82", 2800915, 76663},
 {"ID:718586393", 2801154, 76670},
 {"83", 2801382, 76677},
 {"Proof|ID:245124153", 2801626, 76684},
 {"Proof|ID:245124153-1", 2801790, 76688},
 {"Proof|ID:952627996", 2801952, 76692},
 {"Proof|ID:952627996-1", 2802115, 76696},
 {"Proof|ID:2013760956", 2802277, 76700},
 {"Proof|ID:2013760956-1", 2802444, 76704},
 {"Proof|ID:1011330545", 2802609, 76708},
 {"Proof|ID:1011330545-1", 2802777, 76712},
 {"Proof|ID:718586393", 2802942, 76716},
 {"Proof|ID:718586393-1", 2803107, 76720},
 {"ID:696507084", 2803264, 76724},
 {"84", 2803492, 76731},
 {"ID:626970451", 2803730, 76738},
 {"85", 2803957, 76745},
 {"ID:789654176", 2804194, 76752},
 {"86", 2804422, 76759},
 {"Proof|ID:696507084", 2804666, 76766},
 {"Proof|ID:696507084-1", 2804830, 76770},
 {"Proof|ID:626970451", 2804992, 76774},
 {"Proof|ID:626970451-1", 2805156, 76778},
 {"Proof|ID:789654176", 2805318, 76782},
 {"Proof|ID:789654176-1", 2805481, 76786},
 {"ID:1437590439", 2805637, 76790},
 {"87", 2805867, 76797},
 {"ID:477242065", 2806107, 76804},
 {"88", 2806335, 76811},
 {"Proof|ID:1437590439", 2806580, 76818},
 {"Proof|ID:1437590439-1", 2806746, 76822},
 {"Proof|ID:477242065", 2806909, 76826},
 {"Proof|ID:477242065-1", 2807072, 76830},
 {"sect::TheoremsFarExterior", 2807240, 76834},
 {"ID:1041359623", 2807379, 76838},
 {"89", 2807608, 76845},
 {"ID:2130272336", 2807848, 76852},
 {"90", 2808077, 76859},
 {"ID:1584641667", 2808317, 76866},
 {"91", 2808546, 76873},
 {"ID:1175076282", 2808786, 76880},
 {"92", 2809016, 76887},
 {"ID:333980150", 2809256, 76894},
 {"93", 2809484, 76901},
 {"ID:1370308823", 2809723, 76908},
 {"94", 2809952, 76915},
 {"Proof|ID:1041359623", 2810198, 76922},
 {"Proof|ID:1041359623-1", 2810364, 76926},
 {"Proof|ID:2130272336", 2810528, 76930},
 {"Proof|ID:2130272336-1", 2810694, 76934},
 {"Proof|ID:1584641667", 2810858, 76938},
 {"Proof|ID:1584641667-1", 2811024, 76942},
 {"Proof|ID:1175076282", 2811188, 76946},
 {"Proof|ID:1175076282-1", 2811353, 76950},
 {"Proof|ID:333980150", 2811515, 76954},
 {"Proof|ID:333980150-1", 2811678, 76958},
 {"Proof|ID:1370308823", 2811840, 76962},
 {"Proof|ID:1370308823-1", 2812007, 76966},
 {"ID:1434774204", 2812166, 76970},
 {"95", 2812396, 76977},
 {"ID:1867780171", 2812637, 76984},
 {"96", 2812866, 76991},
 {"ID:786584693", 2813105, 76998},
 {"97", 2813333, 77005},
 {"ID:183897269", 2813571, 77012},
 {"98", 2813799, 77019},
 {"ID:1246401530", 2814038, 77026},
 {"99", 2814268, 77033},
 {"Proof|ID:1434774204", 2814515, 77040},
 {"Proof|ID:1434774204-1", 2814679, 77044},
 {"Proof|ID:1867780171", 2814841, 77048},
 {"Proof|ID:1867780171-1", 2815007, 77052},
 {"Proof|ID:786584693", 2815170, 77056},
 {"Proof|ID:786584693-1", 2815334, 77060},
 {"Proof|ID:183897269", 2815496, 77064},
 {"Proof|ID:183897269-1", 2815660, 77068},
 {"Proof|ID:1246401530", 2815823, 77072},
 {"Proof|ID:1246401530-1", 2815989, 77076},
 {"sect::MainResults", 2816151, 77080},
 {"sect::Invariant", 2816277, 77084},
 {"ID:2102726322", 2816404, 77088},
 {"invariant", 2816648, 77095},
 {"Proof|ID:2102726322", 2816902, 77102},
 {"Proof|ID:2102726322-1", 2817068, 77106},
 {"sect::Bound", 2817224, 77110},
 {"ID:724101256", 2817345, 77114},
 {"100", 2817575, 77121},
 {"Proof|ID:724101256", 2817820, 77128},
 {"Proof|ID:724101256-1", 2817983, 77132},
 {"ID:1990359651", 2818139, 77136},
 {"bound", 2818375, 77143},
 {"Proof|ID:1990359651", 2818625, 77150},
 {"Proof|ID:1990359651-1", 2818792, 77154},
 {"sect::MainResult", 2818954, 77158},
 {"ID:1891911403", 2819081, 77162},
 {"main theorem", 2819330, 77169},
 {"Proof|ID:1891911403", 2819586, 77176},
 {"Proof|ID:1891911403-1", 2819752, 77180},
 {"sect::Prover", 2819909, 77184},
 {"sect::ProverTuples", 2820034, 77188},
 {"sect::ProverAddition", 2820170, 77192},
 {"sect::ProverOrdering", 2820307, 77196},
 {"sect::ProverMinMax", 2820442, 77200},
 {"sect::ProverQuantifiers", 2820580, 77204},
 {"sect::ProverRewriting", 2820721, 77208},
 {"sect::Conclusion", 2820856, 77212},
 {"sect::References", 2820983, 77216},
 {"Buchberger1965", 2821107, 77220},
 {"Buchberger1979", 2821282, 77225},
 {"Buchberger1979a", 2821458, 77230},
 {"Buchberger1983a", 2821635, 77235},
 {"Buchberger1983b", 2821812, 77240},
 {"Buchberger1998", 2821987, 77245},
 {"Buchberger2003", 2822161, 77250},
 {"Buchberger2006", 2822336, 77255},
 {"Buchberger2014", 2822511, 77260},
 {"Giese2007", 2822680, 77265},
 {"Mayr-Meyer1982", 2822849, 77270},
 {"Theorema", 2823018, 77275},
 {"Windsteiger1999", 2823187, 77280},
 {"Winkler1984", 2823359, 77285}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 518, 11, 133, "Title",
 CellID->1633829306],
Cell[1100, 35, 585, 17, 218, "Subsubtitle",
 CellID->777484366],
Cell[1688, 54, 269, 6, 34, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->446898962],
Cell[1960, 62, 1051, 17, 110, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->644531705],
Cell[3014, 81, 407, 9, 24, "Text",
 CellID->95312927],
Cell[CellGroupData[{
Cell[3446, 94, 125, 2, 73, "Section",
 CellID->730227987],
Cell[3574, 98, 1468, 36, 84, "Text",
 CellID->1185065486],
Cell[5045, 136, 565, 14, 45, "Text",
 CellID->1524961043],
Cell[5613, 152, 304, 8, 25, "Text",
 CellID->1630144601],
Cell[CellGroupData[{
Cell[5942, 164, 263, 6, 24, "Item",
 CellID->2091470436],
Cell[6208, 172, 349, 9, 25, "Item",
 CellID->73938940],
Cell[6560, 183, 675, 13, 59, "Item",
 CellID->1984099723],
Cell[7238, 198, 832, 16, 76, "Item",
 CellID->1184009133]
}, Open  ]],
Cell[8085, 217, 432, 11, 24, "Text",
 CellID->2119521717]
}, Open  ]],
Cell[CellGroupData[{
Cell[8554, 233, 117, 2, 73, "Section",
 CellID->590276341],
Cell[8674, 237, 2984, 64, 237, "Text",
 CellID->831646513],
Cell[11661, 303, 1288, 33, 65, "Text",
 CellID->1530781958],
Cell[12952, 338, 5191, 115, 331, "Text",
 CellID->1271320345],
Cell[CellGroupData[{
Cell[18168, 457, 163, 3, 38, "Subsection",
 CellID->448627247],
Cell[18334, 462, 499, 13, 44, "Text",
 CellID->728060281],
Cell[18836, 477, 963, 20, 62, "Text",
 CellID->1524435427],
Cell[19802, 499, 2708, 78, 103, "Text",
 CellID->380145561],
Cell[22513, 579, 456, 9, 43, "Text",
 CellID->83311740]
}, Closed]],
Cell[CellGroupData[{
Cell[23006, 593, 137, 2, 30, "Subsection",
 CellID->1413123984],
Cell[23146, 597, 141, 2, 24, "Text",
 CellID->1937745430],
Cell[CellGroupData[{
Cell[23312, 603, 802, 12, 32, "Input",
 CellID->1715249823],
Cell[24117, 617, 307, 5, 32, "Output",
 CellID->1104335559]
}, Open  ]],
Cell[24439, 625, 260, 4, 24, "Text",
 CellID->2105302388],
Cell[CellGroupData[{
Cell[24724, 633, 432, 7, 32, "Input",
 CellID->503539654],
Cell[25159, 642, 523, 9, 32, "Output",
 CellID->1004345167]
}, Open  ]],
Cell[25697, 654, 190, 3, 24, "Text",
 CellID->675589372],
Cell[CellGroupData[{
Cell[25912, 661, 135, 3, 32, "Input",
 CellID->834113638],
Cell[26050, 666, 114, 2, 32, "Output",
 CellID->1599837690]
}, Open  ]],
Cell[26179, 671, 206, 3, 24, "Text",
 CellID->352834493],
Cell[CellGroupData[{
Cell[26410, 678, 251, 7, 28, "Subsubsection",
 CellID->872582327],
Cell[26664, 687, 860, 23, 99, "Input",
 CellID->458961238],
Cell[27527, 712, 1472, 35, 121, "Input",
 CellID->1459432484],
Cell[29002, 749, 731, 19, 55, "Input",
 CellID->251027589]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[29794, 775, 263, 5, 43, "Section",
 CellTags->"sect::FormalMathematics",
 CellID->1882786356],
Cell[30060, 782, 610, 17, 45, "Text",
 CellID->723171646],
Cell[30673, 801, 487, 12, 43, "Text",
 CellID->228411370],
Cell[31163, 815, 1429, 25, 100, "Text",
 CellID->1933341333],
Cell[32595, 842, 1656, 31, 82, "Text",
 CellID->547946198],
Cell[34254, 875, 1508, 27, 62, "Text",
 CellID->1462515714],
Cell[35765, 904, 4937, 106, 217, "Text",
 CellID->1090519763],
Cell[40705, 1012, 1442, 29, 119, "Text",
 CellID->651312377],
Cell[42150, 1043, 204, 5, 24, "Text",
 CellID->887402679],
Cell[42357, 1050, 9955, 222, 363, "Input",
 Evaluatable->False,
 CellID->985362478],
Cell[CellGroupData[{
Cell[52337, 1276, 175, 3, 38, "Subsection",
 CellTags->"sect::Conditional",
 CellID->1621317523],
Cell[52515, 1281, 207, 5, 24, "Text",
 CellID->2117311209],
Cell[52725, 1288, 224, 5, 24, "Text",
 CellID->1740035160],
Cell[52952, 1295, 1023, 28, 82, "Text",
 CellID->62928363],
Cell[53978, 1325, 654, 18, 25, "Text",
 CellID->1867399679],
Cell[54635, 1345, 138, 2, 24, "Text",
 CellID->1242073734],
Cell[CellGroupData[{
Cell[54798, 1351, 767, 20, 42, "Item",
 CellID->1092788260],
Cell[55568, 1373, 439, 12, 24, "Item",
 CellID->1560604785]
}, Open  ]],
Cell[56022, 1388, 787, 18, 63, "Text",
 CellID->1630050989],
Cell[CellGroupData[{
Cell[56834, 1410, 506, 9, 58, "Item",
 CellID->2003563700],
Cell[57343, 1421, 493, 9, 41, "Item",
 CellID->556412106]
}, Open  ]],
Cell[57851, 1433, 1547, 35, 102, "Text",
 CellID->1546318573],
Cell[59401, 1470, 477, 12, 44, "Text",
 CellID->2094289943],
Cell[59881, 1484, 1961, 40, 139, "Text",
 CellID->8212415],
Cell[61845, 1526, 1445, 23, 139, "Text",
 CellID->1516906979],
Cell[63293, 1551, 866, 25, 44, "Text",
 CellID->380971111],
Cell[64162, 1578, 1265, 35, 82, "Text",
 CellID->2067379196],
Cell[65430, 1615, 561, 12, 62, "Text",
 CellID->737651657],
Cell[65994, 1629, 370, 7, 43, "Text",
 CellID->1982664360]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[66413, 1642, 155, 3, 43, "Section",
 CellTags->"sect::ProblemSetting",
 CellID->264353757],
Cell[66571, 1647, 1004, 27, 62, "Text",
 CellID->229041570],
Cell[CellGroupData[{
Cell[67600, 1678, 122, 2, 38, "Subsection",
 CellID->1990765324],
Cell[67725, 1682, 1091, 33, 84, "Text",
 CellID->453193865],
Cell[CellGroupData[{
Cell[68841, 1719, 816, 22, 59, "Item",
 CellID->708458737],
Cell[69660, 1743, 518, 11, 59, "Item",
 CellID->709660607]
}, Open  ]],
Cell[70193, 1757, 887, 19, 81, "Text",
 CellID->808711520],
Cell[71083, 1778, 156, 2, 24, "Text",
 CellID->1150089532],
Cell[71242, 1782, 473, 13, 41, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->"Alg::BB",
 CellID->1191384728],
Cell[71718, 1797, 1008, 34, 44, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->103682965],
Cell[72729, 1833, 590, 17, 27, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1180081128],
Cell[73322, 1852, 406, 9, 24, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->772744986],
Cell[73731, 1863, 515, 17, 24, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->789122655],
Cell[74249, 1882, 482, 14, 24, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1759507030],
Cell[74734, 1898, 538, 16, 24, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->2063177316],
Cell[75275, 1916, 602, 21, 70, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->2110989409],
Cell[75880, 1939, 537, 16, 70, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->2023860787],
Cell[76420, 1957, 541, 16, 70, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->144735239],
Cell[76964, 1975, 514, 18, 70, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1954271085],
Cell[77481, 1995, 604, 18, 70, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->329571004],
Cell[78088, 2015, 483, 14, 70, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->853193976],
Cell[78574, 2031, 381, 11, 70, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->516448044],
Cell[78958, 2044, 380, 11, 70, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1138282524],
Cell[79341, 2057, 370, 8, 70, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->2003283509],
Cell[79714, 2067, 461, 14, 70, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1764664231],
Cell[80178, 2083, 1115, 28, 70, "Text",
 CellID->764371851],
Cell[81296, 2113, 3499, 96, 70, "Text",
 CellTags->"Eqn::ChainCrit",
 CellID->1279890854],
Cell[84798, 2211, 325, 13, 70, "Text",
 CellID->1752125308],
Cell[85126, 2226, 430, 15, 70, "Text",
 CellID->471797467],
Cell[CellGroupData[{
Cell[85581, 2245, 558, 16, 70, "Item",
 CellID->1506065214],
Cell[86142, 2263, 329, 10, 70, "Item",
 CellID->684093844],
Cell[86474, 2275, 392, 13, 70, "Item",
 CellID->417957227],
Cell[86869, 2290, 429, 14, 70, "Item",
 CellID->1858807165]
}, Open  ]],
Cell[87313, 2307, 969, 26, 70, "Text",
 CellID->1268377962]
}, Closed]],
Cell[CellGroupData[{
Cell[88319, 2338, 132, 2, 30, "Subsection",
 CellID->1943935944],
Cell[88454, 2342, 866, 30, 82, "Text",
 CellID->2505218],
Cell[89323, 2374, 807, 27, 44, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1802962063],
Cell[90133, 2403, 707, 21, 44, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1700275030],
Cell[90843, 2426, 526, 16, 24, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1831213623],
Cell[91372, 2444, 1366, 46, 66, "Text",
 CellID->1948833634],
Cell[92741, 2492, 181, 3, 24, "Text",
 CellID->827429947],
Cell[92925, 2497, 1004, 31, 25, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1480049348],
Cell[93932, 2530, 720, 23, 24, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->200043405],
Cell[94655, 2555, 299, 10, 25, "Text",
 CellID->1617422914],
Cell[94957, 2567, 658, 20, 27, "Text",
 CellTags->"Eqn::M",
 CellID->862257598],
Cell[95618, 2589, 671, 21, 28, "Text",
 CellTags->"Eqn::W",
 CellID->2012096862],
Cell[96292, 2612, 2490, 81, 128, "Text",
 CellID->44376836],
Cell[98785, 2695, 285, 8, 70, "Text",
 CellID->584987536]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[99119, 2709, 200, 4, 43, "Section",
 CellTags->"sect::BasicNotions",
 CellID->2131707921],
Cell[99322, 2715, 826, 24, 63, "Text",
 CellID->2115353011],
Cell[CellGroupData[{
Cell[100173, 2743, 171, 3, 38, "Subsection",
 CellID->1121114900],
Cell[CellGroupData[{
Cell[100369, 2750, 179, 4, 28, "Subsubsection",
 CellTags->"sect::DefinitionTuples",
 CellID->25850532],
Cell[100551, 2756, 853, 14, 62, "Text",
 CellID->849860725],
Cell[101407, 2772, 1464, 38, 103, "Text",
 CellID->2133492005],
Cell[102874, 2812, 48, 1, 12, "OpenEnvironment",
 CellID->1844672668],
Cell[CellGroupData[{
Cell[102947, 2817, 396, 10, 40, "EnvironmentHeader",
 CellID->796541026],
Cell[103346, 2829, 349, 8, 33, "GlobalDeclaration",
 CellID->1752837732],
Cell[103698, 2839, 1563, 37, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1969097229", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length is natural"},
 CellID->1969097229],
Cell[105264, 2878, 2068, 66, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1508742051", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "tuple equality"},
 CellID->1508742051],
Cell[107335, 2946, 1125, 31, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1264372773", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isTuple\:293a"},
 CellID->1264372773],
Cell[108463, 2979, 1139, 31, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:2086441677", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a"},
 CellID->2086441677],
Cell[109605, 3012, 1431, 43, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:585712945", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "element\:293a 1"},
 CellID->585712945],
Cell[111039, 3057, 1123, 36, 24, "FormalTextInputFormula",
 CellTags->{
  "ID:1075236060", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "element\:293a 2"},
 CellID->1075236060],
Cell[112165, 3095, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1505428581]
}, Open  ]],
Cell[112246, 3099, 47, 1, 16, "CloseEnvironment",
 CellID->32281044],
Cell[112296, 3102, 1069, 30, 63, "Text",
 CellID->1869820165],
Cell[113368, 3134, 934, 22, 54, "Text",
 CellID->1183041550],
Cell[114305, 3158, 47, 1, 12, "OpenEnvironment",
 CellID->822989180],
Cell[CellGroupData[{
Cell[114377, 3163, 793, 15, 40, "EnvironmentHeader",
 CellID->1879793312],
Cell[115173, 3180, 446, 11, 32, "GlobalDeclaration",
 CellTags->{
  "ID:1405024037", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->1405024037],
Cell[115622, 3193, 1703, 48, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:379346204", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "domain tuple"},
 CellID->379346204],
Cell[117328, 3243, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1742211894]
}, Open  ]],
Cell[117409, 3247, 48, 1, 16, "CloseEnvironment",
 CellID->615285686]
}, Closed]],
Cell[CellGroupData[{
Cell[117494, 3253, 160, 3, 22, "Subsubsection",
 CellTags->"sect::DefinitionOC",
 CellID->157546407],
Cell[117657, 3258, 611, 13, 43, "Text",
 CellID->2005240008],
Cell[118271, 3273, 47, 1, 12, "OpenEnvironment",
 CellID->205551855],
Cell[CellGroupData[{
Cell[118343, 3278, 430, 10, 40, "EnvironmentHeader",
 CellID->868169721],
Cell[118776, 3290, 791, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1473378407", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "other coordinate 1"},
 CellID->1473378407],
Cell[119570, 3313, 788, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:406630295", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "other coordinate 2"},
 CellID->406630295],
Cell[120361, 3336, 65, 1, 24, "EndEnvironmentMarker",
 CellID->882180403]
}, Open  ]],
Cell[120441, 3340, 49, 1, 16, "CloseEnvironment",
 CellID->1782370292],
Cell[120493, 3343, 924, 19, 62, "Text",
 CellID->388142211]
}, Closed]],
Cell[CellGroupData[{
Cell[121454, 3367, 167, 3, 22, "Subsubsection",
 CellTags->"sect::DefinitionZ",
 CellID->1036621313],
Cell[121624, 3372, 1332, 32, 83, "Text",
 CellID->512342355],
Cell[122959, 3406, 1375, 36, 103, "Text",
 CellID->1266703458],
Cell[124337, 3444, 88, 2, 12, "OpenEnvironment",
 CellID->51464841],
Cell[CellGroupData[{
Cell[124450, 3450, 485, 11, 40, "EnvironmentHeader",
 CellID->2104076581],
Cell[124938, 3463, 950, 24, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1091916109", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Z is TOM"},
 CellID->1091916109],
Cell[125891, 3489, 841, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1845867532", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Z is totally ordered"},
 CellID->1845867532],
Cell[126735, 3513, 1282, 32, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:492790093", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "some integers"},
 CellID->492790093],
Cell[128020, 3547, 844, 23, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1907318616", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1\[LessEqual]2"},
 CellID->1907318616],
Cell[128867, 3572, 496, 10, 41, "GlobalDeclaration",
 CellID->529342921],
Cell[129366, 3584, 1402, 41, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:1708561859", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1"},
 CellID->1708561859],
Cell[130771, 3627, 1152, 32, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:49894867", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "2"},
 CellID->49894867],
Cell[131926, 3661, 107, 2, 24, "EndEnvironmentMarker",
 CellID->786201735]
}, Open  ]],
Cell[132048, 3666, 90, 2, 16, "CloseEnvironment",
 CellID->712329467]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[132187, 3674, 159, 3, 30, "Subsection",
 CellTags->"sect::DefinitionCategories",
 CellID->1469832235],
Cell[132349, 3679, 277, 6, 24, "Text",
 CellID->1067846941],
Cell[132629, 3687, 1397, 32, 102, "Text",
 CellID->1228146070],
Cell[CellGroupData[{
Cell[134051, 3723, 266, 5, 28, "Subsubsection",
 CellTags->"sect::DefinitionAbelianMonoid",
 CellID->1828699765],
Cell[134320, 3730, 762, 16, 62, "Text",
 CellID->337701861],
Cell[135085, 3748, 48, 1, 12, "OpenEnvironment",
 CellID->1763006174],
Cell[CellGroupData[{
Cell[135158, 3753, 404, 9, 40, "EnvironmentHeader",
 CellID->1874281841],
Cell[135565, 3764, 159, 3, 31, "GlobalDeclaration",
 CellID->1405235777],
Cell[135727, 3769, 1253, 39, 45, "FormalTextInputFormula",
 CellTags->{
  "ID:349901806", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "closure"},
 CellID->349901806],
Cell[136983, 3810, 1145, 33, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:1476372605", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "0 in domain"},
 CellID->1476372605],
Cell[138131, 3845, 1343, 42, 49, "FormalTextInputFormula",
 CellTags->{
  "ID:1359514637", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "0 is neutral"},
 CellID->1359514637],
Cell[139477, 3889, 1589, 50, 49, "FormalTextInputFormula",
 CellTags->{
  "ID:1215514258", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "associativity"},
 CellID->1215514258],
Cell[141069, 3941, 1403, 44, 49, "FormalTextInputFormula",
 CellTags->{
  "ID:1531751179", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "commutativity"},
 CellID->1531751179],
Cell[142475, 3987, 1532, 48, 49, "FormalTextInputFormula",
 CellTags->{
  "ID:929558765", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "cancellativity"},
 CellID->929558765],
Cell[144010, 4037, 64, 1, 24, "EndEnvironmentMarker",
 CellID->42267126]
}, Open  ]],
Cell[144089, 4041, 49, 1, 16, "CloseEnvironment",
 CellID->1814682497],
Cell[144141, 4044, 47, 1, 12, "OpenEnvironment",
 CellID->617751546],
Cell[CellGroupData[{
Cell[144213, 4049, 546, 11, 40, "EnvironmentHeader",
 CellID->1751978973],
Cell[144762, 4062, 209, 4, 31, "GlobalDeclaration",
 CellID->996748630],
Cell[144974, 4068, 2225, 54, 109, "FormalTextInputFormula",
 CellTags->{
  "ID:1003575752", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "Abelian monoid"},
 CellID->1003575752],
Cell[147202, 4124, 64, 1, 24, "EndEnvironmentMarker",
 CellID->36024838]
}, Open  ]],
Cell[147281, 4128, 49, 1, 16, "CloseEnvironment",
 CellID->1667893272],
Cell[147333, 4131, 421, 8, 43, "Text",
 CellID->1636890686],
Cell[147757, 4141, 393, 8, 26, "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellID->1203490576],
Cell[148153, 4151, 1066, 17, 43, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellID->268251286]
}, Closed]],
Cell[CellGroupData[{
Cell[149256, 4173, 184, 3, 22, "Subsubsection",
 CellTags->"sect::DefinitionTotallyOrderedDomain",
 CellID->898592553],
Cell[149443, 4178, 668, 17, 44, "Text",
 CellID->1503249513],
Cell[150114, 4197, 47, 1, 12, "OpenEnvironment",
 CellID->833748108],
Cell[CellGroupData[{
Cell[150186, 4202, 441, 10, 40, "EnvironmentHeader",
 CellID->567790588],
Cell[150630, 4214, 518, 19, 33, "GlobalDeclaration",
 CellID->1996046852],
Cell[151151, 4235, 3526, 109, 145, "FormalTextInputFormula",
 CellTags->{
  "ID:1939013113", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "total ordering"},
 CellID->1939013113],
Cell[154680, 4346, 1624, 51, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:1585090210", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "irreflexive relation"},
 CellID->1585090210],
Cell[156307, 4399, 2287, 74, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:1535164863", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "strict version of ordering"},
 CellID->1535164863],
Cell[158597, 4475, 65, 1, 24, "EndEnvironmentMarker",
 CellID->251656430]
}, Open  ]],
Cell[158677, 4479, 49, 1, 16, "CloseEnvironment",
 CellID->1535644204],
Cell[158729, 4482, 629, 16, 63, "Text",
 CellID->672448588],
Cell[159361, 4500, 48, 1, 12, "OpenEnvironment",
 CellID->2106479689],
Cell[CellGroupData[{
Cell[159434, 4505, 403, 9, 40, "EnvironmentHeader",
 CellID->341504943],
Cell[159840, 4516, 207, 4, 31, "GlobalDeclaration",
 CellID->1430533020],
Cell[160050, 4522, 1695, 53, 72, "FormalTextInputFormula",
 CellTags->{
  "ID:795999034", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "totally-ordered domain"},
 CellID->795999034],
Cell[161748, 4577, 65, 1, 24, "EndEnvironmentMarker",
 CellID->461069387]
}, Open  ]],
Cell[161828, 4581, 49, 1, 16, "CloseEnvironment",
 CellID->1286085183]
}, Closed]],
Cell[CellGroupData[{
Cell[161914, 4587, 169, 3, 22, "Subsubsection",
 CellTags->"sect::DefinitionminmaxDomain",
 CellID->2071321112],
Cell[162086, 4592, 1726, 46, 113, "Text",
 CellID->1673907690],
Cell[163815, 4640, 47, 1, 12, "OpenEnvironment",
 CellID->658511437],
Cell[CellGroupData[{
Cell[163887, 4645, 538, 11, 40, "EnvironmentHeader",
 CellID->807480567],
Cell[164428, 4658, 255, 5, 31, "GlobalDeclaration",
 CellID->1209270974],
Cell[164686, 4665, 1780, 44, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:616702295", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min-max domain"},
 CellID->616702295],
Cell[166469, 4711, 1888, 54, 75, "FormalTextInputFormula",
 CellTags->{
  "ID:1204372660", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min domain"},
 CellID->1204372660],
Cell[168360, 4767, 1940, 55, 75, "FormalTextInputFormula",
 CellTags->{
  "ID:1875462394", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max domain"},
 CellID->1875462394],
Cell[170303, 4824, 65, 1, 24, "EndEnvironmentMarker",
 CellID->856781492]
}, Open  ]],
Cell[170383, 4828, 49, 1, 16, "CloseEnvironment",
 CellID->1692376529],
Cell[170435, 4831, 548, 14, 43, "Text",
 CellID->413114653],
Cell[170986, 4847, 321, 7, 26, "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellID->1691893048],
Cell[171310, 4856, 2336, 58, 95, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellID->1732619610]
}, Closed]],
Cell[CellGroupData[{
Cell[173683, 4919, 224, 4, 22, "Subsubsection",
 CellTags->"sect::DefinitionTOM",
 CellID->991805917],
Cell[173910, 4925, 1550, 42, 91, "Text",
 CellID->1505273851],
Cell[175463, 4969, 48, 1, 12, "OpenEnvironment",
 CellID->2123192141],
Cell[CellGroupData[{
Cell[175536, 4974, 605, 12, 40, "EnvironmentHeader",
 CellID->1445007926],
Cell[176144, 4988, 182, 4, 31, "GlobalDeclaration",
 CellID->432239872],
Cell[176329, 4994, 3839, 73, 73, "FormalTextInputFormula",
 CellTags->{
  "ID:743220188", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "totally-ordered Abelian monoid"},
 CellID->743220188],
Cell[180171, 5069, 1577, 48, 49, "FormalTextInputFormula",
 CellTags->{
  "ID:1329821571", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "monotonicity"},
 CellID->1329821571],
Cell[181751, 5119, 65, 1, 24, "EndEnvironmentMarker",
 CellID->270602909]
}, Open  ]],
Cell[181831, 5123, 49, 1, 16, "CloseEnvironment",
 CellID->1275160981],
Cell[181883, 5126, 543, 12, 44, "Text",
 CellID->733533220],
Cell[182429, 5140, 893, 27, 43, "Text",
 CellID->1975846358],
Cell[183325, 5169, 3453, 104, 181, "Text",
 CellID->1496861707],
Cell[186781, 5275, 217, 4, 25, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1979602345],
Cell[187001, 5281, 5677, 205, 140, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1924228650],
Cell[192681, 5488, 4815, 170, 166, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->666720646],
Cell[197499, 5660, 593, 19, 24, "Text",
 CellID->1515751530],
Cell[198095, 5681, 2754, 93, 113, "Text",
 CellID->1432108623],
Cell[200852, 5776, 1340, 41, 44, "Text",
 CellID->1887306109],
Cell[202195, 5819, 2139, 69, 84, "Text",
 CellID->1082635799],
Cell[204337, 5890, 242, 5, 25, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->593477305],
Cell[204582, 5897, 2240, 76, 84, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->909437173],
Cell[206825, 5975, 1426, 43, 43, "Text",
 CellID->423204886],
Cell[208254, 6020, 2030, 51, 125, "Text",
 CellID->174363722],
Cell[210287, 6073, 2433, 75, 105, "Text",
 CellID->136010290],
Cell[212723, 6150, 1494, 42, 85, "Text",
 CellID->1956243226],
Cell[214220, 6194, 5938, 197, 191, "Text",
 CellID->72224179],
Cell[220161, 6393, 2037, 62, 105, "Text",
 CellID->2079853062],
Cell[222201, 6457, 701, 13, 43, "Text",
 CellID->181517487]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[222951, 6476, 127, 2, 30, "Subsection",
 CellID->362214253],
Cell[223081, 6480, 522, 11, 25, "Text",
 CellID->1764685072],
Cell[223606, 6493, 48, 1, 12, "OpenEnvironment",
 CellID->1607947619],
Cell[CellGroupData[{
Cell[223679, 6498, 487, 14, 41, "EnvironmentHeader",
 CellID->1463782225],
Cell[224169, 6514, 755, 20, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:2137457113", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is TOM"},
 CellID->2137457113],
Cell[224927, 6536, 787, 20, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1975341048", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is Abelian monoid"},
 CellID->1975341048],
Cell[225717, 6558, 787, 20, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:275749992", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is totally ordered"},
 CellID->275749992],
Cell[226507, 6580, 785, 20, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1904592343", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D is min-max domain"},
 CellID->1904592343],
Cell[227295, 6602, 65, 1, 24, "EndEnvironmentMarker",
 CellID->763770611]
}, Open  ]],
Cell[227375, 6606, 49, 1, 16, "CloseEnvironment",
 CellID->1609998906],
Cell[227427, 6609, 665, 15, 63, "Text",
 CellID->1977697537],
Cell[228095, 6626, 387, 12, 25, "Text",
 CellID->981023549],
Cell[228485, 6640, 48, 1, 12, "OpenEnvironment",
 CellID->1810710293],
Cell[CellGroupData[{
Cell[228558, 6645, 438, 13, 41, "EnvironmentHeader",
 CellID->1016833517],
Cell[228999, 6660, 919, 24, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:589296925", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n is integer"},
 CellID->589296925],
Cell[229921, 6686, 810, 22, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:54512703", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n is positive"},
 CellID->54512703],
Cell[230734, 6710, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1567750413]
}, Open  ]],
Cell[230815, 6714, 48, 1, 16, "CloseEnvironment",
 CellID->307153015]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[230912, 6721, 305, 5, 43, "Section",
 CellTags->"sect::LemmataBasic",
 CellID->1171017268],
Cell[231220, 6728, 1442, 29, 83, "Text",
 CellID->615481853],
Cell[232665, 6759, 3334, 77, 241, "Text",
 CellID->1795547704],
Cell[236002, 6838, 980, 22, 101, "Text",
 CellID->541338696],
Cell[236985, 6862, 439, 8, 62, "Text",
 CellID->765415560],
Cell[CellGroupData[{
Cell[237449, 6874, 115, 2, 38, "Subsection",
 CellID->497241048],
Cell[237567, 6878, 419, 10, 43, "Text",
 CellID->627429447],
Cell[237989, 6890, 47, 1, 12, "OpenEnvironment",
 CellID->518018031],
Cell[CellGroupData[{
Cell[238061, 6895, 401, 9, 40, "EnvironmentHeader",
 CellID->1477491869],
Cell[238465, 6906, 1019, 31, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:1538785027", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "3"},
 CellID->1538785027],
Cell[239487, 6939, 65, 1, 24, "EndEnvironmentMarker",
 CellID->238945331]
}, Open  ]],
Cell[239567, 6943, 49, 1, 16, "CloseEnvironment",
 CellID->1212198655],
Cell[CellGroupData[{
Cell[239641, 6948, 2604, 79, 29, "ProofInfo",
 CellTags->{"Proof|ID:1538785027", "Proof|ID:1538785027-1"},
 CellID->1800356526],
Cell[242248, 7029, 482, 12, 70, "ProofInfoBody",
 CellID->1169871723]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[242779, 7047, 252, 4, 30, "Subsection",
 CellID->777956154],
Cell[243034, 7053, 592, 13, 24, "Text",
 CellID->1725363192],
Cell[243629, 7068, 48, 1, 12, "OpenEnvironment",
 CellID->1655989806],
Cell[CellGroupData[{
Cell[243702, 7073, 502, 12, 40, "EnvironmentHeader",
 CellID->276004122],
Cell[244207, 7087, 254, 6, 41, "GlobalDeclaration",
 CellID->2116038141],
Cell[244464, 7095, 919, 25, 34, "FormalTextInputFormula",
 CellTags->{
  "ID:968428080", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "4"},
 CellID->968428080],
Cell[245386, 7122, 846, 24, 34, "FormalTextInputFormula",
 CellTags->{
  "ID:680012314", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "5"},
 CellID->680012314],
Cell[246235, 7148, 1182, 31, 34, "FormalTextInputFormula",
 CellTags->{
  "ID:655751250", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "6"},
 CellID->655751250],
Cell[247420, 7181, 1076, 30, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:1963081417", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "7"},
 CellID->1963081417],
Cell[248499, 7213, 66, 1, 24, "EndEnvironmentMarker",
 CellID->2087479351]
}, Open  ]],
Cell[248580, 7217, 49, 1, 16, "CloseEnvironment",
 CellID->1363114882],
Cell[CellGroupData[{
Cell[248654, 7222, 1963, 60, 29, "ProofInfo",
 CellTags->{"Proof|ID:968428080", "Proof|ID:968428080-1"},
 CellID->665537877],
Cell[250620, 7284, 481, 12, 70, "ProofInfoBody",
 CellID->1093697245]
}, Closed]],
Cell[CellGroupData[{
Cell[251138, 7301, 1963, 60, 27, "ProofInfo",
 CellTags->{"Proof|ID:680012314", "Proof|ID:680012314-1"},
 CellID->134690164],
Cell[253104, 7363, 480, 12, 70, "ProofInfoBody",
 CellID->633008601]
}, Closed]],
Cell[CellGroupData[{
Cell[253621, 7380, 3047, 84, 27, "ProofInfo",
 CellTags->{"Proof|ID:655751250", "Proof|ID:655751250-1"},
 CellID->234671713],
Cell[256671, 7466, 480, 12, 70, "ProofInfoBody",
 CellID->319014567]
}, Closed]],
Cell[CellGroupData[{
Cell[257188, 7483, 3196, 90, 27, "ProofInfo",
 CellTags->{"Proof|ID:1963081417", "Proof|ID:1963081417-1"},
 CellID->1541015864],
Cell[260387, 7575, 481, 12, 70, "ProofInfoBody",
 CellID->845366744]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[260917, 7593, 171, 3, 30, "Subsection",
 CellID->58313402],
Cell[261091, 7598, 751, 18, 44, "Text",
 CellID->732300639],
Cell[261845, 7618, 47, 1, 12, "OpenEnvironment",
 CellID->885155644],
Cell[CellGroupData[{
Cell[261917, 7623, 464, 14, 40, "EnvironmentHeader",
 CellID->1272852245],
Cell[262384, 7639, 227, 6, 32, "GlobalDeclaration",
 CellID->405924544],
Cell[262614, 7647, 898, 25, 34, "FormalTextInputFormula",
 CellTags->{
  "ID:1457798778", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "8"},
 CellID->1457798778],
Cell[263515, 7674, 1264, 37, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:1044840887", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "9"},
 CellID->1044840887],
Cell[264782, 7713, 920, 30, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:2017090463", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "10"},
 CellID->2017090463],
Cell[265705, 7745, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1459353343]
}, Open  ]],
Cell[265786, 7749, 48, 1, 16, "CloseEnvironment",
 CellID->384181077],
Cell[CellGroupData[{
Cell[265859, 7754, 2868, 85, 29, "ProofInfo",
 CellTags->{"Proof|ID:1457798778", "Proof|ID:1457798778-1"},
 CellID->122079454],
Cell[268730, 7841, 481, 12, 70, "ProofInfoBody",
 CellID->717730248]
}, Closed]],
Cell[CellGroupData[{
Cell[269248, 7858, 3408, 99, 27, "ProofInfo",
 CellTags->{"Proof|ID:1044840887", "Proof|ID:1044840887-1"},
 CellID->1585772568],
Cell[272659, 7959, 481, 12, 70, "ProofInfoBody",
 CellID->240181621]
}, Closed]],
Cell[CellGroupData[{
Cell[273177, 7976, 3004, 91, 27, "ProofInfo",
 CellTags->{"Proof|ID:2017090463", "Proof|ID:2017090463-1"},
 CellID->1023157275],
Cell[276184, 8069, 482, 12, 70, "ProofInfoBody",
 CellID->1765168555]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[276715, 8087, 275, 5, 30, "Subsection",
 CellTags->"sect::LemmataOrderings",
 CellID->1103427178],
Cell[276993, 8094, 895, 18, 63, "Text",
 CellID->833726611],
Cell[277891, 8114, 740, 18, 44, "Text",
 CellID->591871091],
Cell[278634, 8134, 48, 1, 12, "OpenEnvironment",
 CellID->1950877292],
Cell[CellGroupData[{
Cell[278707, 8139, 465, 10, 40, "EnvironmentHeader",
 CellID->1909682137],
Cell[279175, 8151, 251, 5, 33, "GlobalDeclaration",
 CellID->1236064582],
Cell[279429, 8158, 931, 23, 35, "FormalTextInputFormula",
 CellTags->{
  "ID:1308600401", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "irreflexivity of <"},
 CellID->1308600401],
Cell[280363, 8183, 368, 9, 42, "GlobalDeclaration",
 CellID->1862701560],
Cell[280734, 8194, 1009, 24, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:914659821", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of \[LessEqual]"},
 CellID->914659821],
Cell[281746, 8220, 1422, 39, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:966362506", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<\[Or]="},
 CellID->966362506],
Cell[283171, 8261, 1285, 34, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:1288924674", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[Not]\[LessEqual]"},
 CellID->1288924674],
Cell[284459, 8297, 1265, 34, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:1962569111", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[Not]<"},
 CellID->1962569111],
Cell[285727, 8333, 1296, 35, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:243913977", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity <\[LessEqual]"},
 CellID->243913977],
Cell[287026, 8370, 1300, 35, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:1623404218", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity \[LessEqual]<"},
 CellID->1623404218],
Cell[288329, 8407, 1098, 33, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:1509384446", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity <<"},
 CellID->1509384446],
Cell[289430, 8442, 1215, 36, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:972758509", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[NotEqual] in terms of <"},
 CellID->972758509],
Cell[290648, 8480, 65, 1, 24, "EndEnvironmentMarker",
 CellID->931362008]
}, Open  ]],
Cell[290728, 8484, 49, 1, 16, "CloseEnvironment",
 CellID->1684574786],
Cell[CellGroupData[{
Cell[290802, 8489, 2289, 69, 29, "ProofInfo",
 CellTags->{"Proof|ID:1308600401", "Proof|ID:1308600401-1"},
 CellID->21123431],
Cell[293094, 8560, 482, 12, 70, "ProofInfoBody",
 CellID->1409828609]
}, Closed]],
Cell[CellGroupData[{
Cell[293613, 8577, 2858, 82, 27, "ProofInfo",
 CellTags->{"Proof|ID:914659821", "Proof|ID:914659821-1"},
 CellID->388106339],
Cell[296474, 8661, 480, 12, 70, "ProofInfoBody",
 CellID->496671473]
}, Closed]],
Cell[CellGroupData[{
Cell[296991, 8678, 4234, 114, 27, "ProofInfo",
 CellTags->{"Proof|ID:966362506", "Proof|ID:966362506-1"},
 CellID->831539329],
Cell[301228, 8794, 545, 13, 70, "ProofInfoBody",
 CellID->440020905]
}, Closed]],
Cell[CellGroupData[{
Cell[301810, 8812, 4244, 115, 27, "ProofInfo",
 CellTags->{"Proof|ID:1288924674", "Proof|ID:1288924674-1"},
 CellID->1029351046],
Cell[306057, 8929, 482, 12, 70, "ProofInfoBody",
 CellID->1669987450]
}, Closed]],
Cell[CellGroupData[{
Cell[306576, 8946, 4222, 115, 27, "ProofInfo",
 CellTags->{"Proof|ID:1962569111", "Proof|ID:1962569111-1"},
 CellID->1777374244],
Cell[310801, 9063, 482, 12, 70, "ProofInfoBody",
 CellID->1678900439]
}, Closed]],
Cell[CellGroupData[{
Cell[311320, 9080, 4669, 122, 27, "ProofInfo",
 CellTags->{"Proof|ID:243913977", "Proof|ID:243913977-1"},
 CellID->1987732976],
Cell[315992, 9204, 481, 12, 70, "ProofInfoBody",
 CellID->1471369452]
}, Closed]],
Cell[CellGroupData[{
Cell[316510, 9221, 4678, 123, 27, "ProofInfo",
 CellTags->{"Proof|ID:1623404218", "Proof|ID:1623404218-1"},
 CellID->1964215912],
Cell[321191, 9346, 482, 12, 70, "ProofInfoBody",
 CellID->1573508644]
}, Closed]],
Cell[CellGroupData[{
Cell[321710, 9363, 4633, 123, 27, "ProofInfo",
 CellTags->{"Proof|ID:1509384446", "Proof|ID:1509384446-1"},
 CellID->295324759],
Cell[326346, 9488, 481, 12, 70, "ProofInfoBody",
 CellID->572410148]
}, Closed]],
Cell[CellGroupData[{
Cell[326864, 9505, 4234, 110, 27, "ProofInfo",
 CellTags->{"Proof|ID:972758509", "Proof|ID:972758509-1"},
 CellID->1361125711],
Cell[331101, 9617, 480, 12, 70, "ProofInfoBody",
 CellID->436371387]
}, Closed]],
Cell[331596, 9632, 824, 22, 53, "Text",
 CellID->2080376934],
Cell[332423, 9656, 47, 1, 12, "OpenEnvironment",
 CellID->652155032],
Cell[CellGroupData[{
Cell[332495, 9661, 461, 11, 40, "EnvironmentHeader",
 CellID->1674491660],
Cell[332959, 9674, 191, 4, 33, "GlobalDeclaration",
 CellID->1774773558],
Cell[333153, 9680, 357, 9, 42, "GlobalDeclaration",
 CellID->1799285719],
Cell[333513, 9691, 1050, 29, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:786186036", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]+"},
 CellID->786186036],
Cell[334566, 9722, 1111, 33, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:1385901945", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<+"},
 CellID->1385901945],
Cell[335680, 9757, 1253, 37, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:1581648423", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]\[LessEqual]+"},
 CellID->1581648423],
Cell[336936, 9796, 1211, 37, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:1815472507", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]<+"},
 CellID->1815472507],
Cell[338150, 9835, 1175, 37, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:443432455", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<<+"},
 CellID->443432455],
Cell[339328, 9874, 66, 1, 24, "EndEnvironmentMarker",
 CellID->2069740564]
}, Open  ]],
Cell[339409, 9878, 49, 1, 16, "CloseEnvironment",
 CellID->2028534090],
Cell[CellGroupData[{
Cell[339483, 9883, 4851, 128, 29, "ProofInfo",
 CellTags->{"Proof|ID:786186036", "Proof|ID:786186036-1"},
 CellID->1923006109],
Cell[344337, 10013, 481, 12, 70, "ProofInfoBody",
 CellID->1549560626]
}, Closed]],
Cell[CellGroupData[{
Cell[344855, 10030, 4793, 129, 27, "ProofInfo",
 CellTags->{"Proof|ID:1385901945", "Proof|ID:1385901945-1"},
 CellID->263629365],
Cell[349651, 10161, 481, 12, 70, "ProofInfoBody",
 CellID->701073402]
}, Closed]],
Cell[CellGroupData[{
Cell[350169, 10178, 5767, 147, 27, "ProofInfo",
 CellTags->{"Proof|ID:1581648423", "Proof|ID:1581648423-1"},
 CellID->422549508],
Cell[355939, 10327, 482, 12, 70, "ProofInfoBody",
 CellID->1666906397]
}, Closed]],
Cell[CellGroupData[{
Cell[356458, 10344, 5702, 147, 27, "ProofInfo",
 CellTags->{"Proof|ID:1815472507", "Proof|ID:1815472507-1"},
 CellID->1818000638],
Cell[362163, 10493, 481, 12, 70, "ProofInfoBody",
 CellID->743904775]
}, Closed]],
Cell[CellGroupData[{
Cell[362681, 10510, 5648, 146, 27, "ProofInfo",
 CellTags->{"Proof|ID:443432455", "Proof|ID:443432455-1"},
 CellID->633496152],
Cell[368332, 10658, 481, 12, 70, "ProofInfoBody",
 CellID->1736685396]
}, Closed]],
Cell[368828, 10673, 810, 19, 62, "Text",
 CellID->2067863327]
}, Closed]],
Cell[CellGroupData[{
Cell[369675, 10697, 162, 3, 30, "Subsection",
 CellID->253205562],
Cell[369840, 10702, 656, 12, 62, "Text",
 CellID->859597688],
Cell[370499, 10716, 47, 1, 12, "OpenEnvironment",
 CellID->998146153],
Cell[CellGroupData[{
Cell[370571, 10721, 425, 10, 40, "EnvironmentHeader",
 CellID->1763958272],
Cell[370999, 10733, 440, 9, 33, "GlobalDeclaration",
 CellID->224379101],
Cell[371442, 10744, 1421, 39, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1546357463", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "1\[LessEqual]length"},
 CellID->1546357463],
Cell[372866, 10785, 1184, 31, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:409966668", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "11"},
 CellID->409966668],
Cell[374053, 10818, 1710, 45, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1768549223", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "2\[LessEqual]length"},
 CellID->1768549223],
Cell[375766, 10865, 1001, 28, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1399858996", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a is integer"},
 CellID->1399858996],
Cell[376770, 10895, 982, 27, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1353010357", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "length\:293a is positive"},
 CellID->1353010357],
Cell[377755, 10924, 1769, 47, 56, "FormalTextInputFormula",
 CellTags->{
  "ID:763884008", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "12"},
 CellID->763884008],
Cell[379527, 10973, 1175, 33, 35, "FormalTextInputFormula",
 CellTags->{
  "ID:1442641059", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "13"},
 CellID->1442641059],
Cell[380705, 11008, 1294, 41, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:215173194", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "14"},
 CellID->215173194],
Cell[382002, 11051, 65, 1, 24, "EndEnvironmentMarker",
 CellID->274743996]
}, Open  ]],
Cell[382082, 11055, 48, 1, 16, "CloseEnvironment",
 CellID->183782654],
Cell[CellGroupData[{
Cell[382155, 11060, 3999, 109, 29, "ProofInfo",
 CellTags->{"Proof|ID:1546357463", "Proof|ID:1546357463-1"},
 CellID->1202974989],
Cell[386157, 11171, 482, 12, 70, "ProofInfoBody",
 CellID->1207870308]
}, Closed]],
Cell[CellGroupData[{
Cell[386676, 11188, 3542, 98, 27, "ProofInfo",
 CellTags->{"Proof|ID:409966668", "Proof|ID:409966668-1"},
 CellID->577682993],
Cell[390221, 11288, 480, 12, 70, "ProofInfoBody",
 CellID->924364061]
}, Closed]],
Cell[CellGroupData[{
Cell[390738, 11305, 5387, 141, 27, "ProofInfo",
 CellTags->{"Proof|ID:1768549223", "Proof|ID:1768549223-1"},
 CellID->1717553837],
Cell[396128, 11448, 481, 12, 70, "ProofInfoBody",
 CellID->499654441]
}, Closed]],
Cell[CellGroupData[{
Cell[396646, 11465, 3015, 85, 27, "ProofInfo",
 CellTags->{"Proof|ID:1399858996", "Proof|ID:1399858996-1"},
 CellID->86355476],
Cell[399664, 11552, 482, 12, 70, "ProofInfoBody",
 CellID->1825116772]
}, Closed]],
Cell[CellGroupData[{
Cell[400183, 11569, 3016, 86, 27, "ProofInfo",
 CellTags->{"Proof|ID:1353010357", "Proof|ID:1353010357-1"},
 CellID->1804120756],
Cell[403202, 11657, 482, 12, 70, "ProofInfoBody",
 CellID->1951723367]
}, Closed]],
Cell[CellGroupData[{
Cell[403721, 11674, 4378, 112, 27, "ProofInfo",
 CellTags->{"Proof|ID:763884008", "Proof|ID:763884008-1"},
 CellID->1513904826],
Cell[408102, 11788, 481, 12, 70, "ProofInfoBody",
 CellID->1881258437]
}, Closed]],
Cell[CellGroupData[{
Cell[408620, 11805, 4110, 110, 27, "ProofInfo",
 CellTags->{"Proof|ID:1442641059", "Proof|ID:1442641059-1"},
 CellID->1528692265],
Cell[412733, 11917, 482, 12, 70, "ProofInfoBody",
 CellID->1729488571]
}, Closed]],
Cell[CellGroupData[{
Cell[413252, 11934, 3994, 106, 27, "ProofInfo",
 CellTags->{"Proof|ID:215173194", "Proof|ID:215173194-1"},
 CellID->1068613755],
Cell[417249, 12042, 481, 12, 70, "ProofInfoBody",
 CellID->1165482879]
}, Closed]],
Cell[417745, 12057, 877, 26, 54, "Text",
 CellID->1820728087],
Cell[418625, 12085, 47, 1, 12, "OpenEnvironment",
 CellID->131337638],
Cell[CellGroupData[{
Cell[418697, 12090, 458, 13, 40, "EnvironmentHeader",
 CellID->210695348],
Cell[419158, 12105, 239, 5, 32, "GlobalDeclaration",
 CellID->924384146],
Cell[419400, 12112, 1256, 40, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:8303793", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "15"},
 CellID->8303793],
Cell[420659, 12154, 1265, 40, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:1339530720", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "16"},
 CellID->1339530720],
Cell[421927, 12196, 1378, 43, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:759960416", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[SquareSubsetEqual]\:293a"},
 CellID->759960416],
Cell[423308, 12241, 65, 1, 24, "EndEnvironmentMarker",
 CellID->730215235]
}, Open  ]],
Cell[423388, 12245, 49, 1, 16, "CloseEnvironment",
 CellID->1781807047],
Cell[CellGroupData[{
Cell[423462, 12250, 4020, 107, 29, "ProofInfo",
 CellTags->
  "Proof|ID:8303793|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->222527391],
Cell[427485, 12359, 317, 9, 70, "ProofInfoBody",
 CellID->1226520191]
}, Closed]],
Cell[CellGroupData[{
Cell[427839, 12373, 4025, 107, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1339530720|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->980530952],
Cell[431867, 12482, 316, 9, 70, "ProofInfoBody",
 CellID->447677394]
}, Closed]],
Cell[CellGroupData[{
Cell[432220, 12496, 3839, 105, 27, "ProofInfo",
 CellTags->
  "Proof|ID:759960416|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1092143830],
Cell[436062, 12603, 315, 9, 70, "ProofInfoBody",
 CellID->790789736]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[436426, 12618, 272, 5, 30, "Subsection",
 CellTags->"sect::Lemmataminmax",
 CellID->473572412],
Cell[436701, 12625, 2167, 63, 83, "Text",
 CellID->1049858421],
Cell[438871, 12690, 291, 7, 43, "GlobalDeclaration",
 CellID->931803829],
Cell[439165, 12699, 48, 1, 12, "OpenEnvironment",
 CellID->1523217451],
Cell[CellGroupData[{
Cell[439238, 12704, 391, 9, 40, "EnvironmentHeader",
 CellID->1515714911],
Cell[439632, 12715, 297, 7, 30, "GlobalDeclaration",
 CellID->656269750],
Cell[439932, 12724, 1125, 30, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1448857518", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of min"},
 CellID->1448857518],
Cell[441060, 12756, 1128, 31, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1578657786", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of max"},
 CellID->1578657786],
Cell[442191, 12789, 913, 26, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1968885164", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of min"},
 CellID->1968885164],
Cell[443107, 12817, 913, 26, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1151569277", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of max"},
 CellID->1151569277],
Cell[444023, 12845, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1603322875]
}, Open  ]],
Cell[444104, 12849, 49, 1, 16, "CloseEnvironment",
 CellID->1913696948],
Cell[CellGroupData[{
Cell[444178, 12854, 5017, 132, 29, "ProofInfo",
 CellTags->{"Proof|ID:1448857518", "Proof|ID:1448857518-1"},
 CellID->1149489714],
Cell[449198, 12988, 484, 12, 74, "ProofInfoBody",
 CellID->1171443524]
}, Closed]],
Cell[CellGroupData[{
Cell[449719, 13005, 5013, 130, 27, "ProofInfo",
 CellTags->{"Proof|ID:1578657786", "Proof|ID:1578657786-1"},
 CellID->1807688628],
Cell[454735, 13137, 484, 12, 70, "ProofInfoBody",
 CellID->1889510511]
}, Closed]],
Cell[CellGroupData[{
Cell[455256, 13154, 3317, 92, 27, "ProofInfo",
 CellTags->{"Proof|ID:1968885164", "Proof|ID:1968885164-1"},
 CellID->1971047889],
Cell[458576, 13248, 483, 12, 70, "ProofInfoBody",
 CellID->773286761]
}, Closed]],
Cell[CellGroupData[{
Cell[459096, 13265, 3317, 92, 27, "ProofInfo",
 CellTags->{"Proof|ID:1151569277", "Proof|ID:1151569277-1"},
 CellID->2058731697],
Cell[462416, 13359, 483, 12, 70, "ProofInfoBody",
 CellID->158595008]
}, Closed]],
Cell[462914, 13374, 352, 13, 25, "Text",
 CellID->1149895741],
Cell[463269, 13389, 642, 13, 41, "GlobalDeclaration",
 CellID->1335750375],
Cell[463914, 13404, 379, 13, 24, "Text",
 CellID->744956650],
Cell[464296, 13419, 48, 1, 12, "OpenEnvironment",
 CellID->1583106624],
Cell[CellGroupData[{
Cell[464369, 13424, 391, 9, 40, "EnvironmentHeader",
 CellID->345801648],
Cell[464763, 13435, 1591, 44, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:232133128", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "17"},
 CellID->232133128],
Cell[466357, 13481, 1593, 44, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:1533503972", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "18"},
 CellID->1533503972],
Cell[467953, 13527, 1294, 40, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:881325315", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "17'"},
 CellID->881325315],
Cell[469250, 13569, 1297, 40, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:1667609383", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "18'"},
 CellID->1667609383],
Cell[470550, 13611, 1093, 29, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1257756255", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min is invariant under permutations"},
 CellID->1257756255],
Cell[471646, 13642, 1093, 29, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:2071551441", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max is invariant under permutations"},
 CellID->2071551441],
Cell[472742, 13673, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1103232359]
}, Open  ]],
Cell[472823, 13677, 48, 1, 16, "CloseEnvironment",
 CellID->678471395],
Cell[CellGroupData[{
Cell[472896, 13682, 4488, 120, 29, "ProofInfo",
 CellTags->{"Proof|ID:232133128", "Proof|ID:232133128-1"},
 CellID->557201343],
Cell[477387, 13804, 483, 12, 70, "ProofInfoBody",
 CellID->1057192124]
}, Closed]],
Cell[CellGroupData[{
Cell[477907, 13821, 4494, 120, 27, "ProofInfo",
 CellTags->{"Proof|ID:1533503972", "Proof|ID:1533503972-1"},
 CellID->1508553184],
Cell[482404, 13943, 483, 12, 70, "ProofInfoBody",
 CellID->849429518]
}, Closed]],
Cell[CellGroupData[{
Cell[482924, 13960, 4468, 120, 27, "ProofInfo",
 CellTags->{"Proof|ID:881325315", "Proof|ID:881325315-1"},
 CellID->803466674],
Cell[487395, 14082, 483, 12, 70, "ProofInfoBody",
 CellID->1693750335]
}, Closed]],
Cell[CellGroupData[{
Cell[487915, 14099, 4474, 120, 27, "ProofInfo",
 CellTags->{"Proof|ID:1667609383", "Proof|ID:1667609383-1"},
 CellID->1598215446],
Cell[492392, 14221, 484, 12, 70, "ProofInfoBody",
 CellID->2061132157]
}, Closed]],
Cell[CellGroupData[{
Cell[492913, 14238, 3684, 102, 27, "ProofInfo",
 CellTags->{"Proof|ID:1257756255", "Proof|ID:1257756255-1"},
 CellID->725596050],
Cell[496600, 14342, 484, 12, 70, "ProofInfoBody",
 CellID->1826872359]
}, Closed]],
Cell[CellGroupData[{
Cell[497121, 14359, 3685, 102, 27, "ProofInfo",
 CellTags->{"Proof|ID:2071551441", "Proof|ID:2071551441-1"},
 CellID->1663975332],
Cell[500809, 14463, 484, 12, 70, "ProofInfoBody",
 CellID->1814758527]
}, Closed]],
Cell[501308, 14478, 1425, 36, 72, "Text",
 CellID->304451747],
Cell[502736, 14516, 45, 1, 12, "OpenEnvironment",
 CellID->3199803],
Cell[CellGroupData[{
Cell[502806, 14521, 581, 12, 40, "EnvironmentHeader",
 CellID->1111682865],
Cell[503390, 14535, 397, 9, 30, "GlobalDeclaration",
 CellID->1035719112],
Cell[503790, 14546, 1410, 39, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:48233088", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\[LessEqual]"},
 CellID->48233088],
Cell[505203, 14587, 1466, 39, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1034128987", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]min"},
 CellID->1034128987],
Cell[506672, 14628, 1369, 38, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1323776552", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\[LessEqual]"},
 CellID->1323776552],
Cell[508044, 14668, 1368, 37, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1520079070", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]max"},
 CellID->1520079070],
Cell[509415, 14707, 1229, 36, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1860476328", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min<"},
 CellID->1860476328],
Cell[510647, 14745, 1230, 36, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1907558186", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<min"},
 CellID->1907558186],
Cell[511880, 14783, 1230, 36, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:2008421481", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max<"},
 CellID->2008421481],
Cell[513113, 14821, 1230, 36, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1141242040", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "<max"},
 CellID->1141242040],
Cell[514346, 14859, 65, 1, 24, "EndEnvironmentMarker",
 CellID->562397624]
}, Open  ]],
Cell[514426, 14863, 48, 1, 16, "CloseEnvironment",
 CellID->954132810],
Cell[CellGroupData[{
Cell[514499, 14868, 5819, 150, 29, "ProofInfo",
 CellTags->{"Proof|ID:48233088", "Proof|ID:48233088-1"},
 CellID->2119117596],
Cell[520321, 15020, 482, 12, 70, "ProofInfoBody",
 CellID->1487833047]
}, Closed]],
Cell[CellGroupData[{
Cell[520840, 15037, 5833, 150, 27, "ProofInfo",
 CellTags->{"Proof|ID:1034128987", "Proof|ID:1034128987-1"},
 CellID->1605499322],
Cell[526676, 15189, 483, 12, 70, "ProofInfoBody",
 CellID->980083213]
}, Closed]],
Cell[CellGroupData[{
Cell[527196, 15206, 5827, 150, 27, "ProofInfo",
 CellTags->{"Proof|ID:1323776552", "Proof|ID:1323776552-1"},
 CellID->87934740],
Cell[533026, 15358, 484, 12, 70, "ProofInfoBody",
 CellID->1396094928]
}, Closed]],
Cell[CellGroupData[{
Cell[533547, 15375, 5832, 150, 27, "ProofInfo",
 CellTags->{"Proof|ID:1520079070", "Proof|ID:1520079070-1"},
 CellID->650805455],
Cell[539382, 15527, 484, 12, 70, "ProofInfoBody",
 CellID->1652314721]
}, Closed]],
Cell[CellGroupData[{
Cell[539903, 15544, 5762, 150, 27, "ProofInfo",
 CellTags->{"Proof|ID:1860476328", "Proof|ID:1860476328-1"},
 CellID->489839113],
Cell[545668, 15696, 482, 12, 70, "ProofInfoBody",
 CellID->58718705]
}, Closed]],
Cell[CellGroupData[{
Cell[546187, 15713, 5766, 150, 27, "ProofInfo",
 CellTags->{"Proof|ID:1907558186", "Proof|ID:1907558186-1"},
 CellID->846360522],
Cell[551956, 15865, 484, 12, 70, "ProofInfoBody",
 CellID->1661282637]
}, Closed]],
Cell[CellGroupData[{
Cell[552477, 15882, 5763, 150, 27, "ProofInfo",
 CellTags->{"Proof|ID:2008421481", "Proof|ID:2008421481-1"},
 CellID->1454120988],
Cell[558243, 16034, 484, 12, 70, "ProofInfoBody",
 CellID->2055319065]
}, Closed]],
Cell[CellGroupData[{
Cell[558764, 16051, 5766, 150, 27, "ProofInfo",
 CellTags->{"Proof|ID:1141242040", "Proof|ID:1141242040-1"},
 CellID->155865952],
Cell[564533, 16203, 484, 12, 70, "ProofInfoBody",
 CellID->1159855003]
}, Closed]],
Cell[565032, 16218, 523, 13, 43, "Text",
 CellID->445509307],
Cell[565558, 16233, 46, 1, 12, "OpenEnvironment",
 CellID->50622437],
Cell[CellGroupData[{
Cell[565629, 16238, 442, 10, 40, "EnvironmentHeader",
 CellID->409867065],
Cell[566074, 16250, 1266, 38, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:572537117", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "19"},
 CellID->572537117],
Cell[567343, 16290, 1892, 51, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:2073093776", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a 1"},
 CellID->2073093776],
Cell[569238, 16343, 1359, 42, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1632846240", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a 2"},
 CellID->1632846240],
Cell[570600, 16387, 1269, 38, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1146565830", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "20"},
 CellID->1146565830],
Cell[571872, 16427, 1849, 50, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1172583037", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a 1"},
 CellID->1172583037],
Cell[573724, 16479, 1359, 42, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:2142810348", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a 2"},
 CellID->2142810348],
Cell[575086, 16523, 298, 7, 30, "GlobalDeclaration",
 CellID->419566097],
Cell[575387, 16532, 1096, 33, 36, "FormalTextInputFormula",
 CellTags->{
  "ID:1460290836", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a"},
 CellID->1460290836],
Cell[576486, 16567, 1098, 33, 36, "FormalTextInputFormula",
 CellTags->{
  "ID:1931124773", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a"},
 CellID->1931124773],
Cell[577587, 16602, 990, 29, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:978507648", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min\:293a decreases"},
 CellID->978507648],
Cell[578580, 16633, 994, 29, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:2097939500", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max\:293a increases"},
 CellID->2097939500],
Cell[579577, 16664, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1315514221]
}, Open  ]],
Cell[579658, 16668, 49, 1, 16, "CloseEnvironment",
 CellID->1827651235],
Cell[CellGroupData[{
Cell[579732, 16673, 5085, 134, 29, "ProofInfo",
 CellTags->{"Proof|ID:572537117", "Proof|ID:572537117-1"},
 CellID->1768246037],
Cell[584820, 16809, 483, 12, 70, "ProofInfoBody",
 CellID->1139235207]
}, Closed]],
Cell[CellGroupData[{
Cell[585340, 16826, 5136, 134, 27, "ProofInfo",
 CellTags->{"Proof|ID:2073093776", "Proof|ID:2073093776-1"},
 CellID->470202616],
Cell[590479, 16962, 483, 12, 70, "ProofInfoBody",
 CellID->986447845]
}, Closed]],
Cell[CellGroupData[{
Cell[590999, 16979, 4985, 132, 27, "ProofInfo",
 CellTags->{"Proof|ID:1632846240", "Proof|ID:1632846240-1"},
 CellID->163303679],
Cell[595987, 17113, 484, 12, 70, "ProofInfoBody",
 CellID->1508519340]
}, Closed]],
Cell[CellGroupData[{
Cell[596508, 17130, 5090, 134, 27, "ProofInfo",
 CellTags->{"Proof|ID:1146565830", "Proof|ID:1146565830-1"},
 CellID->1152786378],
Cell[601601, 17266, 483, 12, 70, "ProofInfoBody",
 CellID->880629379]
}, Closed]],
Cell[CellGroupData[{
Cell[602121, 17283, 5136, 134, 27, "ProofInfo",
 CellTags->{"Proof|ID:1172583037", "Proof|ID:1172583037-1"},
 CellID->222384166],
Cell[607260, 17419, 483, 12, 70, "ProofInfoBody",
 CellID->822954280]
}, Closed]],
Cell[CellGroupData[{
Cell[607780, 17436, 4986, 132, 27, "ProofInfo",
 CellTags->{"Proof|ID:2142810348", "Proof|ID:2142810348-1"},
 CellID->1174101328],
Cell[612769, 17570, 483, 12, 70, "ProofInfoBody",
 CellID->653142793]
}, Closed]],
Cell[CellGroupData[{
Cell[613289, 17587, 4501, 118, 27, "ProofInfo",
 CellTags->{"Proof|ID:1460290836", "Proof|ID:1460290836-1"},
 CellID->1313465390],
Cell[617793, 17707, 549, 13, 70, "ProofInfoBody",
 CellID->1745865902]
}, Closed]],
Cell[CellGroupData[{
Cell[618379, 17725, 4499, 118, 27, "ProofInfo",
 CellTags->{"Proof|ID:1931124773", "Proof|ID:1931124773-1"},
 CellID->70708585],
Cell[622881, 17845, 484, 12, 70, "ProofInfoBody",
 CellID->1543615849]
}, Closed]],
Cell[CellGroupData[{
Cell[623402, 17862, 4044, 108, 27, "ProofInfo",
 CellTags->{"Proof|ID:978507648", "Proof|ID:978507648-1"},
 CellID->1954389032],
Cell[627449, 17972, 483, 12, 70, "ProofInfoBody",
 CellID->1594840844]
}, Closed]],
Cell[CellGroupData[{
Cell[627969, 17989, 4048, 108, 27, "ProofInfo",
 CellTags->{"Proof|ID:2097939500", "Proof|ID:2097939500-1"},
 CellID->910681388],
Cell[632020, 18099, 484, 12, 70, "ProofInfoBody",
 CellID->1555785782]
}, Closed]],
Cell[632519, 18114, 1156, 31, 72, "Text",
 CellID->755189501],
Cell[633678, 18147, 48, 1, 12, "OpenEnvironment",
 CellID->1830383482],
Cell[CellGroupData[{
Cell[633751, 18152, 632, 20, 40, "EnvironmentHeader",
 CellID->524486684],
Cell[634386, 18174, 299, 7, 30, "GlobalDeclaration",
 CellID->1418774134],
Cell[634688, 18183, 1661, 47, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:900399941", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "21"},
 CellID->900399941],
Cell[636352, 18232, 1523, 44, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1833484473", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 1"},
 CellID->1833484473],
Cell[637878, 18278, 1271, 37, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1976058899", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 2"},
 CellID->1976058899],
Cell[639152, 18317, 1272, 37, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:2067845318", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+ 3"},
 CellID->2067845318],
Cell[640427, 18356, 1430, 39, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:446229453", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min+"},
 CellID->446229453],
Cell[641860, 18397, 1472, 43, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:437038980", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 1"},
 CellID->437038980],
Cell[643335, 18442, 1268, 37, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:367539508", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 2"},
 CellID->367539508],
Cell[644606, 18481, 1272, 37, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1689218301", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+ 3"},
 CellID->1689218301],
Cell[645881, 18520, 1380, 38, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:513276857", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max+"},
 CellID->513276857],
Cell[647264, 18560, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1359585033]
}, Open  ]],
Cell[647345, 18564, 49, 1, 16, "CloseEnvironment",
 CellID->1284690125],
Cell[CellGroupData[{
Cell[647419, 18569, 8666, 210, 29, "ProofInfo",
 CellTags->{"Proof|ID:900399941", "Proof|ID:900399941-1"},
 CellID->1837094901],
Cell[656088, 18781, 483, 12, 70, "ProofInfoBody",
 CellID->1605223748]
}, Closed]],
Cell[CellGroupData[{
Cell[656608, 18798, 6464, 162, 27, "ProofInfo",
 CellTags->{"Proof|ID:1833484473", "Proof|ID:1833484473-1"},
 CellID->209997353],
Cell[663075, 18962, 550, 13, 70, "ProofInfoBody",
 CellID->1366892285]
}, Closed]],
Cell[CellGroupData[{
Cell[663662, 18980, 6159, 156, 27, "ProofInfo",
 CellTags->{"Proof|ID:1976058899", "Proof|ID:1976058899-1"},
 CellID->1890751220],
Cell[669824, 19138, 549, 13, 70, "ProofInfoBody",
 CellID->2116571744]
}, Closed]],
Cell[CellGroupData[{
Cell[670410, 19156, 6168, 158, 27, "ProofInfo",
 CellTags->{"Proof|ID:2067845318", "Proof|ID:2067845318-1"},
 CellID->1653738850],
Cell[676581, 19316, 484, 12, 70, "ProofInfoBody",
 CellID->2084769334]
}, Closed]],
Cell[CellGroupData[{
Cell[677102, 19333, 6071, 154, 27, "ProofInfo",
 CellTags->{"Proof|ID:446229453", "Proof|ID:446229453-1"},
 CellID->357185825],
Cell[683176, 19489, 523, 13, 70, "ProofInfoBody",
 CellID->22513353]
}, Closed]],
Cell[CellGroupData[{
Cell[683736, 19507, 6459, 162, 27, "ProofInfo",
 CellTags->{"Proof|ID:437038980", "Proof|ID:437038980-1"},
 CellID->132391776],
Cell[690198, 19671, 483, 12, 70, "ProofInfoBody",
 CellID->1014774542]
}, Closed]],
Cell[CellGroupData[{
Cell[690718, 19688, 6154, 156, 27, "ProofInfo",
 CellTags->{"Proof|ID:367539508", "Proof|ID:367539508-1"},
 CellID->1111387014],
Cell[696875, 19846, 482, 12, 70, "ProofInfoBody",
 CellID->785534569]
}, Closed]],
Cell[CellGroupData[{
Cell[697394, 19863, 6167, 158, 27, "ProofInfo",
 CellTags->{"Proof|ID:1689218301", "Proof|ID:1689218301-1"},
 CellID->434528939],
Cell[703564, 20023, 483, 12, 70, "ProofInfoBody",
 CellID->984365216]
}, Closed]],
Cell[CellGroupData[{
Cell[704084, 20040, 6072, 154, 27, "ProofInfo",
 CellTags->{"Proof|ID:513276857", "Proof|ID:513276857-1"},
 CellID->1335154081],
Cell[710159, 20196, 481, 12, 70, "ProofInfoBody",
 CellID->15476564]
}, Closed]],
Cell[710655, 20211, 481, 11, 43, "Text",
 CellID->2110478468]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[711185, 20228, 279, 5, 43, "Section",
 CellTags->"sect::SpecificNotions",
 CellID->1752495898],
Cell[711467, 20235, 699, 20, 44, "Text",
 CellID->96197413],
Cell[712169, 20257, 48, 1, 12, "OpenEnvironment",
 CellID->1718449027],
Cell[CellGroupData[{
Cell[712242, 20262, 559, 11, 40, "EnvironmentHeader",
 CellID->1247544428],
Cell[712804, 20275, 718, 15, 33, "GlobalDeclaration",
 CellTags->{
  "ID:1076035631", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->1076035631],
Cell[713525, 20292, 1777, 49, 47, "FormalTextInputFormula",
 CellTags->{
  "ID:610987882", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExponentVector"},
 CellID->610987882],
Cell[715305, 20343, 1542, 47, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:468249611", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExpVectorTuple"},
 CellID->468249611],
Cell[716850, 20392, 1581, 41, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:444732983", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "divisibility"},
 CellID->444732983],
Cell[718434, 20435, 1154, 30, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1111389324", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "degree"},
 CellID->1111389324],
Cell[719591, 20467, 1384, 36, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1699354265", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maxdeg"},
 CellID->1699354265],
Cell[720978, 20505, 1538, 39, 43, "FormalTextInputFormula",
 CellTags->{
  "ID:1614921088", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "lcm"},
 CellID->1614921088],
Cell[722519, 20546, 2791, 78, 77, "FormalTextInputFormula",
 CellTags->{
  "ID:1784038925", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chain criterion"},
 CellID->1784038925],
Cell[725313, 20626, 1856, 43, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:312341806", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "width w.r.t. k-th coordinate"},
 CellID->312341806],
Cell[727172, 20671, 1398, 34, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:282925385", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "width"},
 CellID->282925385],
Cell[728573, 20707, 1913, 47, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1664335746", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maximum degree of admissible pairs of one object"},
 CellID->1664335746],
Cell[730489, 20756, 1712, 41, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:640444143", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maximum degree of admissible pairs"},
 CellID->640444143],
Cell[732204, 20799, 1697, 43, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1923113076", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "corner of rectangular region"},
 CellID->1923113076],
Cell[733904, 20844, 767, 17, 44, "EnvironmentText",
 CellID->2042263002],
Cell[734674, 20863, 1447, 40, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1774661003", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "above"},
 CellID->1774661003],
Cell[736124, 20905, 2246, 63, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:285552196", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "far exterior w.r.t. k-th coordinate"},
 CellID->285552196],
Cell[738373, 20970, 1894, 50, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:494773971", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "rectangular region"},
 CellID->494773971],
Cell[740270, 21022, 65, 1, 24, "EndEnvironmentMarker",
 CellID->544969917]
}, Open  ]],
Cell[740350, 21026, 49, 1, 16, "CloseEnvironment",
 CellID->1228004538],
Cell[740402, 21029, 661, 15, 25, "Text",
 CellID->1229204291],
Cell[741066, 21046, 909, 24, 44, "Text",
 CellID->955050321],
Cell[CellGroupData[{
Cell[742000, 21074, 575, 21, 24, "Item",
 CellID->1782480268],
Cell[742578, 21097, 1065, 25, 76, "Item",
 CellID->1033322772]
}, Open  ]],
Cell[743658, 21125, 420, 8, 26, "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellID->121149513],
Cell[744081, 21135, 2383, 56, 140, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellID->787517132],
Cell[746467, 21193, 423, 8, 26, "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15002.},
 CellID->2122574584],
Cell[746893, 21203, 1877, 39, 122, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15002.},
 CellID->365888472],
Cell[748773, 21244, 891, 24, 64, "Text",
 CellID->408039679],
Cell[749667, 21270, 48, 1, 12, "OpenEnvironment",
 CellID->1453057307],
Cell[CellGroupData[{
Cell[749740, 21275, 554, 12, 40, "EnvironmentHeader",
 CellID->2050713516],
Cell[750297, 21289, 199, 4, 33, "GlobalDeclaration",
 CellID->852694876],
Cell[750499, 21295, 2667, 81, 112, "FormalTextInputFormula",
 CellTags->{
  "ID:1278822860", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal w.r.t. k-th coordinate"},
 CellID->1278822860],
Cell[753169, 21378, 1356, 40, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:1458824241", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal"},
 CellID->1458824241],
Cell[754528, 21420, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1156889619]
}, Open  ]],
Cell[754609, 21424, 48, 1, 16, "CloseEnvironment",
 CellID->568715675],
Cell[CellGroupData[{
Cell[754682, 21429, 170, 3, 38, "Subsection",
 CellID->1231732349],
Cell[754855, 21434, 2192, 60, 123, "Text",
 CellID->1857632321],
Cell[757050, 21496, 3672, 84, 268, "Text",
 CellID->1503678106],
Cell[760725, 21582, 511, 14, 44, "Text",
 CellID->342302046],
Cell[761239, 21598, 7117, 150, 368, "Text",
 CellID->652519512],
Cell[CellGroupData[{
Cell[768381, 21752, 512, 14, 43, "Item",
 CellID->457187766],
Cell[768896, 21768, 472, 10, 59, "Item",
 CellID->574066102],
Cell[769371, 21780, 301, 8, 25, "Item",
 CellID->438806075],
Cell[769675, 21790, 431, 13, 41, "Item",
 CellID->665266488],
Cell[770109, 21805, 377, 10, 41, "Item",
 CellID->914798491],
Cell[770489, 21817, 770, 22, 60, "Item",
 CellID->872492695],
Cell[771262, 21841, 697, 19, 41, "Item",
 CellID->41635902],
Cell[771962, 21862, 528, 16, 24, "Item",
 CellID->484929528],
Cell[772493, 21880, 734, 21, 42, "Item",
 CellID->1357948457],
Cell[773230, 21903, 511, 10, 59, "Item",
 CellID->545799062],
Cell[773744, 21915, 739, 18, 61, "Item",
 CellID->1522152467]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[774544, 21940, 323, 6, 43, "Section",
 CellTags->"sect::Computations",
 CellID->1485320096],
Cell[774870, 21948, 743, 16, 63, "Text",
 CellID->472378341],
Cell[775616, 21966, 47, 1, 12, "OpenEnvironment",
 CellID->871319643],
Cell[CellGroupData[{
Cell[775688, 21971, 497, 14, 40, "EnvironmentHeader",
 CellID->1397499305],
Cell[776188, 21987, 809, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1559569657", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "D=N0"},
 CellID->1559569657],
Cell[777000, 22011, 65, 1, 24, "EndEnvironmentMarker",
 CellID->156157206]
}, Open  ]],
Cell[777080, 22015, 49, 1, 16, "CloseEnvironment",
 CellID->1451763118],
Cell[CellGroupData[{
Cell[777154, 22020, 178, 3, 38, "Subsection",
 CellID->589691952],
Cell[777335, 22025, 662, 22, 24, "Text",
 CellID->1351740566],
Cell[778000, 22049, 47, 1, 12, "OpenEnvironment",
 CellID->256974579],
Cell[CellGroupData[{
Cell[778072, 22054, 486, 14, 40, "EnvironmentHeader",
 CellID->149993638],
Cell[778561, 22070, 296, 8, 30, "GlobalDeclaration",
 CellTags->{
  "ID:977095909", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->977095909],
Cell[778860, 22080, 975, 26, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:615519889", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "min in N0"},
 CellID->615519889],
Cell[779838, 22108, 977, 26, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1245684717", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "max in N0"},
 CellID->1245684717],
Cell[780818, 22136, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1875602967]
}, Open  ]],
Cell[780899, 22140, 49, 1, 16, "CloseEnvironment",
 CellID->1548667982]
}, Closed]],
Cell[CellGroupData[{
Cell[780985, 22146, 185, 4, 30, "Subsection",
 CellID->586087969],
Cell[781173, 22152, 385, 10, 24, "Text",
 CellID->1272854071],
Cell[781561, 22164, 48, 1, 12, "OpenEnvironment",
 CellID->1322766950],
Cell[CellGroupData[{
Cell[781634, 22169, 452, 14, 40, "EnvironmentHeader",
 CellID->540238681],
Cell[782089, 22185, 737, 20, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1526008058", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n=2"},
 CellID->1526008058],
Cell[782829, 22207, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1365355815]
}, Open  ]],
Cell[782910, 22211, 49, 1, 16, "CloseEnvironment",
 CellID->2067681119],
Cell[782962, 22214, 251, 5, 24, "Text",
 CellID->2004358328],
Cell[CellGroupData[{
Cell[783238, 22223, 294, 6, 31, "Computation",
 CellID->363465800],
Cell[783535, 22231, 525, 8, 25, "ComputationResult",
 CellID->1355610580]
}, Open  ]],
Cell[CellGroupData[{
Cell[784097, 22244, 377, 10, 28, "ComputationInfo",
 CellID->1483357400],
Cell[784477, 22256, 673, 17, 70, "ComputationInfoBody",
 CellID->483700261]
}, Closed]],
Cell[CellGroupData[{
Cell[785187, 22278, 307, 7, 28, "Computation",
 CellID->301274866],
Cell[785497, 22287, 370, 6, 25, "ComputationResult",
 CellID->43157872]
}, Open  ]],
Cell[CellGroupData[{
Cell[785904, 22298, 377, 10, 28, "ComputationInfo",
 CellID->1539369667],
Cell[786284, 22310, 650, 17, 70, "ComputationInfoBody",
 CellID->2109196125]
}, Closed]],
Cell[CellGroupData[{
Cell[786971, 22332, 315, 7, 28, "Computation",
 CellID->1843740659],
Cell[787289, 22341, 350, 6, 25, "ComputationResult",
 CellID->494166568]
}, Open  ]],
Cell[CellGroupData[{
Cell[787676, 22352, 376, 10, 28, "ComputationInfo",
 CellID->77268719],
Cell[788055, 22364, 650, 17, 70, "ComputationInfoBody",
 CellID->699968735]
}, Closed]],
Cell[CellGroupData[{
Cell[788742, 22386, 393, 10, 28, "Computation",
 CellID->494647160],
Cell[789138, 22398, 459, 10, 70, "ComputationResult",
 CellID->1016883962]
}, Open  ]],
Cell[CellGroupData[{
Cell[789634, 22413, 377, 10, 70, "ComputationInfo",
 CellID->1918883531],
Cell[790014, 22425, 650, 17, 70, "ComputationInfoBody",
 CellID->1375661780]
}, Closed]],
Cell[CellGroupData[{
Cell[790701, 22447, 282, 6, 70, "Computation",
 CellID->1489184765],
Cell[790986, 22455, 298, 5, 70, "ComputationResult",
 CellID->965057731]
}, Open  ]],
Cell[CellGroupData[{
Cell[791321, 22465, 378, 10, 70, "ComputationInfo",
 CellID->1070988792],
Cell[791702, 22477, 651, 17, 70, "ComputationInfoBody",
 CellID->1857348292]
}, Closed]],
Cell[CellGroupData[{
Cell[792390, 22499, 835, 19, 70, "Computation",
 CellID->379223532],
Cell[793228, 22520, 302, 5, 70, "ComputationResult",
 CellID->174076915]
}, Open  ]],
Cell[CellGroupData[{
Cell[793567, 22530, 376, 10, 70, "ComputationInfo",
 CellID->612957880],
Cell[793946, 22542, 650, 17, 70, "ComputationInfoBody",
 CellID->1453820005]
}, Closed]],
Cell[CellGroupData[{
Cell[794633, 22564, 859, 19, 70, "Computation",
 CellID->1396655990],
Cell[795495, 22585, 332, 6, 70, "ComputationResult",
 CellID->1382976801]
}, Open  ]],
Cell[CellGroupData[{
Cell[795864, 22596, 377, 10, 70, "ComputationInfo",
 CellID->488228445],
Cell[796244, 22608, 651, 17, 70, "ComputationInfoBody",
 CellID->1091772466]
}, Closed]],
Cell[CellGroupData[{
Cell[796932, 22630, 655, 15, 70, "Computation",
 CellID->164423730],
Cell[797590, 22647, 278, 5, 70, "ComputationResult",
 CellID->782644456]
}, Open  ]],
Cell[CellGroupData[{
Cell[797905, 22657, 376, 10, 70, "ComputationInfo",
 CellID->252386576],
Cell[798284, 22669, 648, 17, 70, "ComputationInfoBody",
 CellID->38275912]
}, Closed]],
Cell[CellGroupData[{
Cell[798969, 22691, 631, 15, 70, "Computation",
 CellID->752433928],
Cell[799603, 22708, 368, 9, 70, "ComputationResult",
 CellID->1528747130]
}, Open  ]],
Cell[CellGroupData[{
Cell[800008, 22722, 376, 10, 70, "ComputationInfo",
 CellID->417165553],
Cell[800387, 22734, 650, 17, 70, "ComputationInfoBody",
 CellID->1567625520]
}, Closed]],
Cell[CellGroupData[{
Cell[801074, 22756, 630, 15, 70, "Computation",
 CellID->20329479],
Cell[801707, 22773, 276, 5, 70, "ComputationResult",
 CellID->1193416390]
}, Open  ]],
Cell[CellGroupData[{
Cell[802020, 22783, 375, 10, 70, "ComputationInfo",
 CellID->837187957],
Cell[802398, 22795, 649, 17, 70, "ComputationInfoBody",
 CellID->1896946096]
}, Closed]],
Cell[CellGroupData[{
Cell[803084, 22817, 659, 15, 70, "Computation",
 CellID->285678123],
Cell[803746, 22834, 277, 5, 70, "ComputationResult",
 CellID->1225552093]
}, Open  ]],
Cell[CellGroupData[{
Cell[804060, 22844, 375, 10, 70, "ComputationInfo",
 CellID->45314890],
Cell[804438, 22856, 650, 17, 70, "ComputationInfoBody",
 CellID->1232819849]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[805137, 22879, 235, 5, 30, "Subsection",
 CellID->1098302827],
Cell[805375, 22886, 520, 14, 24, "Text",
 CellID->1778373308],
Cell[805898, 22902, 48, 1, 12, "OpenEnvironment",
 CellID->2113767064],
Cell[CellGroupData[{
Cell[805971, 22907, 552, 16, 40, "EnvironmentHeader",
 CellID->1847813459],
Cell[806526, 22925, 786, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1544493310", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "n=3"},
 CellID->1544493310],
Cell[807315, 22948, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1619210422]
}, Open  ]],
Cell[807396, 22952, 48, 1, 16, "CloseEnvironment",
 CellID->766446153],
Cell[807447, 22955, 251, 5, 24, "Text",
 CellID->1985946244],
Cell[CellGroupData[{
Cell[807723, 22964, 307, 7, 31, "Computation",
 CellID->189844734],
Cell[808033, 22973, 430, 8, 25, "ComputationResult",
 CellID->1709252354]
}, Open  ]],
Cell[CellGroupData[{
Cell[808500, 22986, 375, 10, 28, "ComputationInfo",
 CellID->88472762],
Cell[808878, 22998, 649, 17, 70, "ComputationInfoBody",
 CellID->624705868]
}, Closed]],
Cell[CellGroupData[{
Cell[809564, 23020, 245, 5, 28, "Computation",
 CellID->491362362],
Cell[809812, 23027, 327, 6, 25, "ComputationResult",
 CellID->1670964831]
}, Open  ]],
Cell[CellGroupData[{
Cell[810176, 23038, 376, 10, 28, "ComputationInfo",
 CellID->582639331],
Cell[810555, 23050, 649, 17, 70, "ComputationInfoBody",
 CellID->701974587]
}, Closed]],
Cell[CellGroupData[{
Cell[811241, 23072, 276, 6, 28, "Computation",
 CellID->1181188516],
Cell[811520, 23080, 328, 6, 25, "ComputationResult",
 CellID->223449918]
}, Open  ]],
Cell[CellGroupData[{
Cell[811885, 23091, 378, 10, 28, "ComputationInfo",
 CellID->1599523293],
Cell[812266, 23103, 650, 17, 70, "ComputationInfoBody",
 CellID->473374470]
}, Closed]],
Cell[CellGroupData[{
Cell[812953, 23125, 370, 9, 28, "Computation",
 CellID->307478637],
Cell[813326, 23136, 402, 9, 25, "ComputationResult",
 CellID->1599111698]
}, Open  ]],
Cell[CellGroupData[{
Cell[813765, 23150, 375, 10, 28, "ComputationInfo",
 CellID->417097376],
Cell[814143, 23162, 650, 17, 70, "ComputationInfoBody",
 CellID->1544363263]
}, Closed]],
Cell[CellGroupData[{
Cell[814830, 23184, 297, 7, 28, "Computation",
 CellID->1510116252],
Cell[815130, 23193, 300, 5, 25, "ComputationResult",
 CellID->1308976343]
}, Open  ]],
Cell[CellGroupData[{
Cell[815467, 23203, 377, 10, 28, "ComputationInfo",
 CellID->591174291],
Cell[815847, 23215, 648, 17, 70, "ComputationInfoBody",
 CellID->9837495]
}, Closed]],
Cell[CellGroupData[{
Cell[816532, 23237, 839, 18, 51, "Computation",
 CellID->1391592445],
Cell[817374, 23257, 303, 5, 25, "ComputationResult",
 CellID->615312700]
}, Open  ]],
Cell[CellGroupData[{
Cell[817714, 23267, 378, 10, 28, "ComputationInfo",
 CellID->1974151093],
Cell[818095, 23279, 650, 17, 70, "ComputationInfoBody",
 CellID->498065940]
}, Closed]],
Cell[CellGroupData[{
Cell[818782, 23301, 836, 18, 51, "Computation",
 CellID->1172710914],
Cell[819621, 23321, 302, 5, 25, "ComputationResult",
 CellID->1707085166]
}, Open  ]],
Cell[CellGroupData[{
Cell[819960, 23331, 377, 10, 28, "ComputationInfo",
 CellID->609311901],
Cell[820340, 23343, 650, 17, 70, "ComputationInfoBody",
 CellID->750452517]
}, Closed]],
Cell[CellGroupData[{
Cell[821027, 23365, 453, 11, 28, "Computation",
 CellID->823265960],
Cell[821483, 23378, 300, 5, 25, "ComputationResult",
 CellID->1745361079]
}, Open  ]],
Cell[CellGroupData[{
Cell[821820, 23388, 376, 10, 28, "ComputationInfo",
 CellID->2138059031],
Cell[822199, 23400, 650, 17, 70, "ComputationInfoBody",
 CellID->1167618070]
}, Closed]],
Cell[822864, 23420, 445, 14, 24, "Text",
 CellID->1360124193],
Cell[CellGroupData[{
Cell[823334, 23438, 539, 12, 31, "Computation",
 CellID->1205577273],
Cell[823876, 23452, 301, 5, 25, "ComputationResult",
 CellID->1165502951]
}, Open  ]],
Cell[CellGroupData[{
Cell[824214, 23462, 378, 10, 28, "ComputationInfo",
 CellID->1183991773],
Cell[824595, 23474, 651, 17, 70, "ComputationInfoBody",
 CellID->2004806027]
}, Closed]],
Cell[CellGroupData[{
Cell[825283, 23496, 647, 14, 28, "Computation",
 CellID->1005372790],
Cell[825933, 23512, 324, 6, 25, "ComputationResult",
 CellID->914965399]
}, Open  ]],
Cell[CellGroupData[{
Cell[826294, 23523, 377, 10, 28, "ComputationInfo",
 CellID->262060218],
Cell[826674, 23535, 651, 17, 70, "ComputationInfoBody",
 CellID->2050120917]
}, Closed]],
Cell[CellGroupData[{
Cell[827362, 23557, 646, 14, 28, "Computation",
 CellID->317313208],
Cell[828011, 23573, 294, 5, 25, "ComputationResult",
 CellID->301600]
}, Open  ]],
Cell[CellGroupData[{
Cell[828342, 23583, 377, 10, 28, "ComputationInfo",
 CellID->1971312572],
Cell[828722, 23595, 650, 17, 70, "ComputationInfoBody",
 CellID->2138593680]
}, Closed]],
Cell[CellGroupData[{
Cell[829409, 23617, 645, 14, 28, "Computation",
 CellID->176496894],
Cell[830057, 23633, 398, 9, 25, "ComputationResult",
 CellID->625007468]
}, Open  ]],
Cell[CellGroupData[{
Cell[830492, 23647, 377, 10, 28, "ComputationInfo",
 CellID->1494793755],
Cell[830872, 23659, 649, 17, 70, "ComputationInfoBody",
 CellID->573749363]
}, Closed]],
Cell[CellGroupData[{
Cell[831558, 23681, 652, 14, 28, "Computation",
 CellID->1743340434],
Cell[832213, 23697, 299, 5, 25, "ComputationResult",
 CellID->1326982056]
}, Open  ]],
Cell[CellGroupData[{
Cell[832549, 23707, 377, 10, 28, "ComputationInfo",
 CellID->1718243673],
Cell[832929, 23719, 649, 17, 70, "ComputationInfoBody",
 CellID->25789008]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[833639, 23743, 289, 5, 43, "Section",
 CellTags->"sect::LemmataSpecific",
 CellID->1812043342],
Cell[833931, 23750, 1522, 39, 82, "Text",
 CellID->207929894],
Cell[835456, 23791, 1495, 35, 83, "Text",
 CellID->58811915],
Cell[CellGroupData[{
Cell[836976, 23830, 236, 4, 38, "Subsection",
 CellTags->"sect::LemmataSpecificGeneral",
 CellID->1771021382],
Cell[837215, 23836, 1001, 24, 83, "Text",
 CellID->506293920],
Cell[838219, 23862, 271, 5, 33, "GlobalDeclaration",
 CellID->1507641767],
Cell[CellGroupData[{
Cell[838515, 23871, 318, 5, 28, "Subsubsection",
 CellID->2076981791],
Cell[838836, 23878, 441, 10, 43, "Text",
 CellID->1783959097],
Cell[839280, 23890, 48, 1, 12, "OpenEnvironment",
 CellID->2089232889],
Cell[CellGroupData[{
Cell[839353, 23895, 506, 12, 40, "EnvironmentHeader",
 CellID->1181039367],
Cell[839862, 23909, 1096, 31, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:18996081", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "22"},
 CellID->18996081],
Cell[840961, 23942, 65, 1, 24, "EndEnvironmentMarker",
 CellID->489695439]
}, Open  ]],
Cell[841041, 23946, 48, 1, 16, "CloseEnvironment",
 CellID->496071433],
Cell[CellGroupData[{
Cell[841114, 23951, 2947, 86, 29, "ProofInfo",
 CellTags->{"Proof|ID:18996081", "Proof|ID:18996081-1"},
 CellID->1450282884],
Cell[844064, 24039, 481, 12, 70, "ProofInfoBody",
 CellID->517597611]
}, Closed]],
Cell[844560, 24054, 47, 1, 12, "OpenEnvironment",
 CellID->844735177],
Cell[CellGroupData[{
Cell[844632, 24059, 425, 11, 40, "EnvironmentHeader",
 CellID->1405903066],
Cell[845060, 24072, 200, 4, 33, "GlobalDeclaration",
 CellID->1825557675],
Cell[845263, 24078, 868, 24, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:313023917", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "isExpVectorTuple\:293a"},
 CellID->313023917],
Cell[846134, 24104, 745, 20, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1697603634", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "23"},
 CellID->1697603634],
Cell[846882, 24126, 65, 1, 24, "EndEnvironmentMarker",
 CellID->173042086]
}, Open  ]],
Cell[846962, 24130, 49, 1, 16, "CloseEnvironment",
 CellID->1570375643],
Cell[CellGroupData[{
Cell[847036, 24135, 2562, 73, 29, "ProofInfo",
 CellTags->
  "Proof|ID:313023917|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->360389676],
Cell[849601, 24210, 315, 9, 70, "ProofInfoBody",
 CellID->511935660]
}, Closed]],
Cell[CellGroupData[{
Cell[849953, 24224, 2165, 66, 27, "ProofInfo",
 CellTags->{"Proof|ID:1697603634", "Proof|ID:1697603634-1"},
 CellID->1339747217],
Cell[852121, 24292, 483, 12, 70, "ProofInfoBody",
 CellID->918500356]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[852653, 24310, 250, 4, 22, "Subsubsection",
 CellID->513543808],
Cell[852906, 24316, 1072, 20, 63, "Text",
 CellID->1070024737],
Cell[853981, 24338, 46, 1, 12, "OpenEnvironment",
 CellID->91834526],
Cell[CellGroupData[{
Cell[854052, 24343, 505, 13, 40, "EnvironmentHeader",
 CellID->1104374192],
Cell[854560, 24358, 234, 5, 33, "GlobalDeclaration",
 CellID->1959503054],
Cell[854797, 24365, 1119, 31, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1313148627", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "antisymmetry of divisibility"},
 CellID->1313148627],
Cell[855919, 24398, 794, 20, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:785542046", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of divisibility"},
 CellID->785542046],
Cell[856716, 24420, 1029, 29, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1055765423", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "transitivity of divisibility"},
 CellID->1055765423],
Cell[857748, 24451, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1453338991]
}, Open  ]],
Cell[857829, 24455, 49, 1, 16, "CloseEnvironment",
 CellID->2087869455],
Cell[CellGroupData[{
Cell[857903, 24460, 3275, 91, 29, "ProofInfo",
 CellTags->{"Proof|ID:1313148627", "Proof|ID:1313148627-1"},
 CellID->513842368],
Cell[861181, 24553, 483, 12, 70, "ProofInfoBody",
 CellID->847402406]
}, Closed]],
Cell[CellGroupData[{
Cell[861701, 24570, 2314, 70, 27, "ProofInfo",
 CellTags->{"Proof|ID:785542046", "Proof|ID:785542046-1"},
 CellID->530754490],
Cell[864018, 24642, 482, 12, 70, "ProofInfoBody",
 CellID->133907625]
}, Closed]],
Cell[CellGroupData[{
Cell[864537, 24659, 3505, 95, 27, "ProofInfo",
 CellTags->{"Proof|ID:1055765423", "Proof|ID:1055765423-1"},
 CellID->985028352],
Cell[868045, 24756, 483, 12, 70, "ProofInfoBody",
 CellID->794383855]
}, Closed]],
Cell[868543, 24771, 1154, 31, 62, "Text",
 CellID->326509238],
Cell[869700, 24804, 48, 1, 12, "OpenEnvironment",
 CellID->1361910085],
Cell[CellGroupData[{
Cell[869773, 24809, 444, 10, 40, "EnvironmentHeader",
 CellID->113793713],
Cell[870220, 24821, 272, 6, 33, "GlobalDeclaration",
 CellID->2077457411],
Cell[870495, 24829, 278, 7, 32, "GlobalDeclaration",
 CellID->1354533749],
Cell[870776, 24838, 893, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1918380148", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of lcm"},
 CellID->1918380148],
Cell[871672, 24862, 844, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1065951991", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 1"},
 CellID->1065951991],
Cell[872519, 24885, 844, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1071215803", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 2"},
 CellID->1071215803],
Cell[873366, 24908, 1115, 29, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:1028467786", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "defining property of lcm 3"},
 CellID->1028467786],
Cell[874484, 24939, 896, 23, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:98270815", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "commutativity of lcm"},
 CellID->98270815],
Cell[875383, 24964, 904, 24, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:622866956", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "24"},
 CellID->622866956],
Cell[876290, 24990, 925, 25, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:810936059", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "idempotency of lcm"},
 CellID->810936059],
Cell[877218, 25017, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1802578373]
}, Open  ]],
Cell[877299, 25021, 49, 1, 16, "CloseEnvironment",
 CellID->2058203328],
Cell[CellGroupData[{
Cell[877373, 25026, 2669, 76, 29, "ProofInfo",
 CellTags->{"Proof|ID:1918380148", "Proof|ID:1918380148-1"},
 CellID->1760699056],
Cell[880045, 25104, 483, 12, 70, "ProofInfoBody",
 CellID->524153589]
}, Closed]],
Cell[CellGroupData[{
Cell[880565, 25121, 2777, 78, 27, "ProofInfo",
 CellTags->{"Proof|ID:1065951991", "Proof|ID:1065951991-1"},
 CellID->1695070992],
Cell[883345, 25201, 483, 12, 70, "ProofInfoBody",
 CellID->542892896]
}, Closed]],
Cell[CellGroupData[{
Cell[883865, 25218, 2777, 78, 27, "ProofInfo",
 CellTags->{"Proof|ID:1071215803", "Proof|ID:1071215803-1"},
 CellID->1919630296],
Cell[886645, 25298, 482, 12, 70, "ProofInfoBody",
 CellID->59184554]
}, Closed]],
Cell[CellGroupData[{
Cell[887164, 25315, 3783, 102, 27, "ProofInfo",
 CellTags->{"Proof|ID:1028467786", "Proof|ID:1028467786-1"},
 CellID->488707796],
Cell[890950, 25419, 484, 12, 70, "ProofInfoBody",
 CellID->1260793235]
}, Closed]],
Cell[CellGroupData[{
Cell[891471, 25436, 2992, 84, 27, "ProofInfo",
 CellTags->{"Proof|ID:98270815", "Proof|ID:98270815-1"},
 CellID->615308222],
Cell[894466, 25522, 480, 12, 70, "ProofInfoBody",
 CellID->65135107]
}, Closed]],
Cell[CellGroupData[{
Cell[894983, 25539, 3100, 88, 27, "ProofInfo",
 CellTags->{"Proof|ID:622866956", "Proof|ID:622866956-1"},
 CellID->947602302],
Cell[898086, 25629, 483, 12, 70, "ProofInfoBody",
 CellID->1486588817]
}, Closed]],
Cell[CellGroupData[{
Cell[898606, 25646, 2534, 76, 27, "ProofInfo",
 CellTags->{"Proof|ID:810936059", "Proof|ID:810936059-1"},
 CellID->898800181],
Cell[901143, 25724, 483, 12, 70, "ProofInfoBody",
 CellID->1345421969]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[901675, 25742, 483, 7, 22, "Subsubsection",
 CellID->222564893],
Cell[902161, 25751, 510, 15, 45, "Text",
 CellID->1443198008],
Cell[902674, 25768, 48, 1, 12, "OpenEnvironment",
 CellID->1609431449],
Cell[CellGroupData[{
Cell[902747, 25773, 426, 10, 40, "EnvironmentHeader",
 CellID->483808494],
Cell[903176, 25785, 320, 7, 33, "GlobalDeclaration",
 CellID->1370144613],
Cell[903499, 25794, 956, 25, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1822207066", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of deg"},
 CellID->1822207066],
Cell[904458, 25821, 972, 27, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:455032720", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "25"},
 CellID->455032720],
Cell[905433, 25850, 980, 27, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:839534060", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "26"},
 CellID->839534060],
Cell[906416, 25879, 65, 1, 24, "EndEnvironmentMarker",
 CellID->994235688]
}, Open  ]],
Cell[906496, 25883, 49, 1, 16, "CloseEnvironment",
 CellID->1479111596],
Cell[CellGroupData[{
Cell[906570, 25888, 2336, 70, 29, "ProofInfo",
 CellTags->{"Proof|ID:1822207066", "Proof|ID:1822207066-1"},
 CellID->694957922],
Cell[908909, 25960, 484, 12, 70, "ProofInfoBody",
 CellID->2069252785]
}, Closed]],
Cell[CellGroupData[{
Cell[909430, 25977, 3037, 85, 27, "ProofInfo",
 CellTags->
  "Proof|ID:455032720|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1987093019],
Cell[912470, 26064, 378, 10, 70, "ProofInfoBody",
 CellID->841372092]
}, Closed]],
Cell[CellGroupData[{
Cell[912885, 26079, 2918, 83, 27, "ProofInfo",
 CellTags->
  "Proof|ID:839534060|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->849905264],
Cell[915806, 26164, 379, 10, 70, "ProofInfoBody",
 CellID->135010444]
}, Closed]],
Cell[916200, 26177, 1487, 40, 62, "Text",
 CellID->1710837182],
Cell[917690, 26219, 457, 12, 24, "Text",
 CellID->210483232],
Cell[918150, 26233, 47, 1, 12, "OpenEnvironment",
 CellID->613968158],
Cell[CellGroupData[{
Cell[918222, 26238, 429, 10, 40, "EnvironmentHeader",
 CellID->151723375],
Cell[918654, 26250, 347, 8, 30, "GlobalDeclaration",
 CellID->717930824],
Cell[919004, 26260, 892, 24, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:2123737964", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of maxdeg"},
 CellID->2123737964],
Cell[919899, 26286, 1088, 30, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:647995479", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "27"},
 CellID->647995479],
Cell[920990, 26318, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1382258393]
}, Open  ]],
Cell[921071, 26322, 49, 1, 16, "CloseEnvironment",
 CellID->1369097639],
Cell[CellGroupData[{
Cell[921145, 26327, 3128, 87, 29, "ProofInfo",
 CellTags->
  "Proof|ID:2123737964|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1501203891],
Cell[924276, 26416, 358, 10, 70, "ProofInfoBody",
 CellID->1309062189]
}, Closed]],
Cell[CellGroupData[{
Cell[924671, 26431, 4876, 127, 27, "ProofInfo",
 CellTags->
  "Proof|ID:647995479|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->878198538],
Cell[929550, 26560, 380, 10, 70, "ProofInfoBody",
 CellID->1033778194]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[929979, 26576, 284, 5, 22, "Subsubsection",
 CellTags->"sect::LemmataSpecificGeneralM",
 CellID->940386836],
Cell[930266, 26583, 764, 16, 44, "Text",
 CellID->78870630],
Cell[931033, 26601, 47, 1, 12, "OpenEnvironment",
 CellID->446526914],
Cell[CellGroupData[{
Cell[931105, 26606, 664, 16, 40, "EnvironmentHeader",
 CellID->915159816],
Cell[931772, 26624, 283, 6, 33, "GlobalDeclaration",
 CellID->2015253534],
Cell[932058, 26632, 849, 23, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1452963633", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "28"},
 CellID->1452963633],
Cell[932910, 26657, 932, 25, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1500097429", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "29"},
 CellID->1500097429],
Cell[933845, 26684, 481, 13, 24, "EnvironmentText",
 CellID->1103072691],
Cell[934329, 26699, 1084, 29, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:279915377", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCrit"},
 CellID->279915377],
Cell[935416, 26730, 1291, 35, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1391995733", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "divisibility\[DoubleRightArrow]chainCrit"},
 CellID->1391995733],
Cell[936710, 26767, 823, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1271079949", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "reflexivity of chainCrit"},
 CellID->1271079949],
Cell[937536, 26790, 2164, 64, 72, "FormalTextInputFormula",
 CellTags->{
  "ID:488309194", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit w.r.t. singleton"},
 CellID->488309194],
Cell[939703, 26856, 1261, 37, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1322351588", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a 1"},
 CellID->1322351588],
Cell[940967, 26895, 1262, 37, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1269486125", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a 2"},
 CellID->1269486125],
Cell[942232, 26934, 1467, 42, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:941129646", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a"},
 CellID->941129646],
Cell[943702, 26978, 1164, 33, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1657568255", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit\:293a I"},
 CellID->1657568255],
Cell[944869, 27013, 64, 1, 24, "EndEnvironmentMarker",
 CellID->41255160]
}, Open  ]],
Cell[944948, 27017, 49, 1, 16, "CloseEnvironment",
 CellID->1530445502],
Cell[CellGroupData[{
Cell[945022, 27022, 3208, 87, 29, "ProofInfo",
 CellTags->
  "Proof|ID:1452963633|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1783184226],
Cell[948233, 27111, 381, 10, 70, "ProofInfoBody",
 CellID->1043672466]
}, Closed]],
Cell[CellGroupData[{
Cell[948651, 27126, 3400, 91, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1500097429|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1859160805],
Cell[952054, 27219, 381, 10, 70, "ProofInfoBody",
 CellID->1975569240]
}, Closed]],
Cell[CellGroupData[{
Cell[952472, 27234, 3485, 93, 27, "ProofInfo",
 CellTags->
  "Proof|ID:279915377|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1592316742],
Cell[955960, 27329, 379, 10, 70, "ProofInfoBody",
 CellID->1362235388]
}, Closed]],
Cell[CellGroupData[{
Cell[956376, 27344, 3669, 97, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1391995733|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->218343730],
Cell[960048, 27443, 379, 10, 70, "ProofInfoBody",
 CellID->84993298]
}, Closed]],
Cell[CellGroupData[{
Cell[960464, 27458, 2703, 76, 27, "ProofInfo",
 CellTags->{"Proof|ID:1271079949", "Proof|ID:1271079949-1"},
 CellID->1431739973],
Cell[963170, 27536, 526, 13, 70, "ProofInfoBody",
 CellID->1498238151]
}, Closed]],
Cell[CellGroupData[{
Cell[963733, 27554, 6438, 158, 27, "ProofInfo",
 CellTags->
  "Proof|ID:488309194|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1697069805],
Cell[970174, 27714, 380, 10, 70, "ProofInfoBody",
 CellID->1563635678]
}, Closed]],
Cell[CellGroupData[{
Cell[970591, 27729, 4612, 115, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1322351588|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1376405718],
Cell[975206, 27846, 381, 10, 70, "ProofInfoBody",
 CellID->1107600169]
}, Closed]],
Cell[CellGroupData[{
Cell[975624, 27861, 4607, 115, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1269486125|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->779731268],
Cell[980234, 27978, 381, 10, 70, "ProofInfoBody",
 CellID->1608520609]
}, Closed]],
Cell[CellGroupData[{
Cell[980652, 27993, 4762, 119, 27, "ProofInfo",
 CellTags->
  "Proof|ID:941129646|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->581146426],
Cell[985417, 28114, 379, 10, 70, "ProofInfoBody",
 CellID->471861581]
}, Closed]],
Cell[CellGroupData[{
Cell[985833, 28129, 3666, 97, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1657568255|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1717614685],
Cell[989502, 28228, 380, 10, 70, "ProofInfoBody",
 CellID->485455762]
}, Closed]],
Cell[989897, 28241, 1458, 35, 101, "Text",
 CellID->633826383],
Cell[991358, 28278, 684, 14, 24, "Text",
 CellID->897047203],
Cell[992045, 28294, 90, 2, 12, "OpenEnvironment",
 CellID->1158939693],
Cell[CellGroupData[{
Cell[992160, 28300, 695, 15, 40, "EnvironmentHeader",
 CellID->1661037347],
Cell[992858, 28317, 299, 7, 30, "GlobalDeclaration",
 CellID->1227578886],
Cell[993160, 28326, 249, 5, 33, "GlobalDeclaration",
 CellID->1399902871],
Cell[993412, 28333, 1538, 43, 49, "FormalTextInputFormula",
 CellTags->{
  "ID:469812703", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "30"},
 CellID->469812703],
Cell[994953, 28378, 1240, 30, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1053941118", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit holds for some element"},
 CellID->1053941118],
Cell[996196, 28410, 108, 2, 24, "EndEnvironmentMarker",
 CellID->1542713411]
}, Open  ]],
Cell[996319, 28415, 91, 2, 16, "CloseEnvironment",
 CellID->1684866112],
Cell[CellGroupData[{
Cell[996435, 28421, 6203, 155, 29, "ProofInfo",
 CellTags->
  "Proof|ID:469812703|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->1066023196],
Cell[1002641, 28578, 430, 12, 70, "ProofInfoBody",
 CellID->370971114]
}, Closed]],
Cell[CellGroupData[{
Cell[1003108, 28595, 5355, 136, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1053941118|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->875669745],
Cell[1008466, 28733, 379, 10, 70, "ProofInfoBody",
 CellID->419994512]
}, Closed]],
Cell[1008860, 28746, 1493, 35, 81, "Text",
 CellID->499366557]
}, Closed]],
Cell[CellGroupData[{
Cell[1010390, 28786, 528, 9, 22, "Subsubsection",
 CellID->1911387542],
Cell[1010921, 28797, 397, 9, 30, "GlobalDeclaration",
 CellID->1584445324],
Cell[1011321, 28808, 1135, 30, 43, "Text",
 CellID->855235021],
Cell[1012459, 28840, 47, 1, 12, "OpenEnvironment",
 CellID->615718314],
Cell[CellGroupData[{
Cell[1012531, 28845, 537, 15, 41, "EnvironmentHeader",
 CellID->529250115],
Cell[1013071, 28862, 322, 6, 33, "GlobalDeclaration",
 CellID->1727832338],
Cell[1013396, 28870, 1529, 39, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:1635347335", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "31"},
 CellID->1635347335],
Cell[1014928, 28911, 1279, 36, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1901113035", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "32"},
 CellID->1901113035],
Cell[1016210, 28949, 919, 25, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1760547878", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M (binary)"},
 CellID->1760547878],
Cell[1017132, 28976, 892, 24, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1896797254", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M (unary)"},
 CellID->1896797254],
Cell[1018027, 29002, 1131, 31, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1611647919", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "33"},
 CellID->1611647919],
Cell[1019161, 29035, 65, 1, 24, "EndEnvironmentMarker",
 CellID->481543849]
}, Open  ]],
Cell[1019241, 29039, 49, 1, 16, "CloseEnvironment",
 CellID->1953187198],
Cell[CellGroupData[{
Cell[1019315, 29044, 7097, 173, 29, "ProofInfo",
 CellTags->
  "Proof|ID:1635347335|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->786500781],
Cell[1026415, 29219, 358, 10, 70, "ProofInfoBody",
 CellID->937118081]
}, Closed]],
Cell[CellGroupData[{
Cell[1026810, 29234, 6895, 169, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1901113035|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->573625489],
Cell[1033708, 29405, 359, 10, 70, "ProofInfoBody",
 CellID->1065390053]
}, Closed]],
Cell[CellGroupData[{
Cell[1034104, 29420, 3745, 101, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1760547878|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->177141490],
Cell[1037852, 29523, 358, 10, 70, "ProofInfoBody",
 CellID->1878691620]
}, Closed]],
Cell[CellGroupData[{
Cell[1038247, 29538, 3212, 89, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1896797254|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->367799976],
Cell[1041462, 29629, 407, 11, 70, "ProofInfoBody",
 CellID->2136307825]
}, Closed]],
Cell[CellGroupData[{
Cell[1041906, 29645, 5056, 131, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1611647919|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->2107085874],
Cell[1046965, 29778, 359, 10, 70, "ProofInfoBody",
 CellID->1300790402]
}, Closed]],
Cell[1047339, 29791, 864, 26, 54, "Text",
 CellID->1704242160],
Cell[1048206, 29819, 48, 1, 12, "OpenEnvironment",
 CellID->1170519931],
Cell[CellGroupData[{
Cell[1048279, 29824, 601, 19, 40, "EnvironmentHeader",
 CellID->1280894414],
Cell[1048883, 29845, 256, 7, 41, "GlobalDeclaration",
 CellID->664584010],
Cell[1049142, 29854, 200, 4, 33, "GlobalDeclaration",
 CellID->1887894187],
Cell[1049345, 29860, 1589, 42, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:1545946275", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M (unary)"},
 CellID->1545946275],
Cell[1050937, 29904, 1654, 43, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:1018591902", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual] (unary)"},
 CellID->1018591902],
Cell[1052594, 29949, 1747, 47, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:382103250", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M (binary)"},
 CellID->382103250],
Cell[1054344, 29998, 1507, 44, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:450457679", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual] (binary)"},
 CellID->450457679],
Cell[1055854, 30044, 1966, 61, 57, "FormalTextInputFormula",
 CellTags->{
  "ID:643714041", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M+\[LessEqual]"},
 CellID->643714041],
Cell[1057823, 30107, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1721951835]
}, Open  ]],
Cell[1057904, 30111, 48, 1, 16, "CloseEnvironment",
 CellID->671143984],
Cell[CellGroupData[{
Cell[1057977, 30116, 6242, 158, 29, "ProofInfo",
 CellTags->{"Proof|ID:1545946275", "Proof|ID:1545946275-1"},
 CellID->1890851245],
Cell[1064222, 30276, 547, 13, 70, "ProofInfoBody",
 CellID->735815697]
}, Closed]],
Cell[CellGroupData[{
Cell[1064806, 30294, 6231, 158, 27, "ProofInfo",
 CellTags->{"Proof|ID:1018591902", "Proof|ID:1018591902-1"},
 CellID->1462209435],
Cell[1071040, 30454, 525, 13, 70, "ProofInfoBody",
 CellID->209766788]
}, Closed]],
Cell[CellGroupData[{
Cell[1071602, 30472, 8080, 196, 27, "ProofInfo",
 CellTags->{"Proof|ID:382103250", "Proof|ID:382103250-1"},
 CellID->280851943],
Cell[1079685, 30670, 547, 13, 70, "ProofInfoBody",
 CellID->1182486729]
}, Closed]],
Cell[CellGroupData[{
Cell[1080269, 30688, 8071, 196, 27, "ProofInfo",
 CellTags->{"Proof|ID:450457679", "Proof|ID:450457679-1"},
 CellID->478833348],
Cell[1088343, 30886, 546, 13, 70, "ProofInfoBody",
 CellID->967557673]
}, Closed]],
Cell[CellGroupData[{
Cell[1088926, 30904, 8841, 212, 27, "ProofInfo",
 CellTags->{"Proof|ID:643714041", "Proof|ID:643714041-1"},
 CellID->105955449],
Cell[1097770, 31118, 482, 12, 70, "ProofInfoBody",
 CellID->754929079]
}, Closed]],
Cell[1098267, 31133, 1058, 22, 63, "Text",
 CellID->512688545],
Cell[1099328, 31157, 47, 1, 12, "OpenEnvironment",
 CellID->791144920],
Cell[CellGroupData[{
Cell[1099400, 31162, 557, 18, 40, "EnvironmentHeader",
 CellID->1432366078],
Cell[1099960, 31182, 319, 6, 33, "GlobalDeclaration",
 CellID->140598250],
Cell[1100282, 31190, 1074, 30, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:1934562955", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "34"},
 CellID->1934562955],
Cell[1101359, 31222, 1171, 32, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1993966556", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "35"},
 CellID->1993966556],
Cell[1102533, 31256, 1095, 30, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:1265284403", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "36"},
 CellID->1265284403],
Cell[1103631, 31288, 1063, 30, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:934408883", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "37"},
 CellID->934408883],
Cell[1104697, 31320, 1826, 53, 92, "FormalTextInputFormula",
 CellTags->{
  "ID:1342720412", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "38"},
 CellID->1342720412],
Cell[1106526, 31375, 344, 6, 33, "GlobalDeclaration",
 CellID->432451001],
Cell[1106873, 31383, 1204, 33, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:192359742", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "39"},
 CellID->192359742],
Cell[1108080, 31418, 1677, 49, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:2005322529", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "40"},
 CellID->2005322529],
Cell[1109760, 31469, 65, 1, 24, "EndEnvironmentMarker",
 CellID->469218948]
}, Open  ]],
Cell[1109840, 31473, 49, 1, 16, "CloseEnvironment",
 CellID->1792755754],
Cell[CellGroupData[{
Cell[1109914, 31478, 4465, 117, 29, "ProofInfo",
 CellTags->
  "Proof|ID:1934562955|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->721371033],
Cell[1114382, 31597, 381, 10, 70, "ProofInfoBody",
 CellID->1356713102]
}, Closed]],
Cell[CellGroupData[{
Cell[1114800, 31612, 4369, 115, 27, "ProofInfo",
 CellTags->
  "Proof|ID:1993966556|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->899634188],
Cell[1119172, 31729, 358, 10, 70, "ProofInfoBody",
 CellID->599689410]
}, Closed]],
Cell[CellGroupData[{
Cell[1119567, 31744, 4928, 126, 27, "ProofInfo",
 CellTags->{"Proof|ID:1265284403", "Proof|ID:1265284403-1"},
 CellID->612613216],
Cell[1124498, 31872, 484, 12, 70, "ProofInfoBody",
 CellID->1433050870]
}, Closed]],
Cell[CellGroupData[{
Cell[1125019, 31889, 4721, 122, 27, "ProofInfo",
 CellTags->{"Proof|ID:934408883", "Proof|ID:934408883-1"},
 CellID->716604140],
Cell[1129743, 32013, 483, 12, 70, "ProofInfoBody",
 CellID->2015567188]
}, Closed]],
Cell[CellGroupData[{
Cell[1130263, 32030, 5838, 148, 27, "ProofInfo",
 CellTags->{"Proof|ID:192359742", "Proof|ID:192359742-1"},
 CellID->89073571],
Cell[1136104, 32180, 483, 12, 70, "ProofInfoBody",
 CellID->1022528965]
}, Closed]],
Cell[CellGroupData[{
Cell[1136624, 32197, 7086, 174, 27, "ProofInfo",
 CellTags->{"Proof|ID:2005322529", "Proof|ID:2005322529-1"},
 CellID->592718415],
Cell[1143713, 32373, 484, 12, 70, "ProofInfoBody",
 CellID->1252787252]
}, Closed]],
Cell[1144212, 32388, 622, 17, 44, "Text",
 CellID->1468024158],
Cell[1144837, 32407, 47, 1, 12, "OpenEnvironment",
 CellID->744151466],
Cell[CellGroupData[{
Cell[1144909, 32412, 524, 17, 40, "EnvironmentHeader",
 CellID->1514731226],
Cell[1145436, 32431, 198, 4, 33, "GlobalDeclaration",
 CellID->1447892216],
Cell[1145637, 32437, 296, 8, 33, "GlobalDeclaration",
 CellID->1051935310],
Cell[1145936, 32447, 1160, 33, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:984666222", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "41"},
 CellID->984666222],
Cell[1147099, 32482, 1321, 38, 53, "FormalTextInputFormula",
 CellTags->{
  "ID:1616352426", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "42"},
 CellID->1616352426],
Cell[1148423, 32522, 65, 1, 24, "EndEnvironmentMarker",
 CellID->850223971]
}, Open  ]],
Cell[1148503, 32526, 48, 1, 16, "CloseEnvironment",
 CellID->272111712],
Cell[CellGroupData[{
Cell[1148576, 32531, 6994, 174, 29, "ProofInfo",
 CellTags->{"Proof|ID:984666222", "Proof|ID:984666222-1"},
 CellID->740590150],
Cell[1155573, 32707, 482, 12, 70, "ProofInfoBody",
 CellID->238534644]
}, Closed]],
Cell[CellGroupData[{
Cell[1156092, 32724, 8354, 203, 27, "ProofInfo",
 CellTags->{"Proof|ID:1616352426", "Proof|ID:1616352426-1"},
 CellID->1148420320],
Cell[1164449, 32929, 484, 12, 70, "ProofInfoBody",
 CellID->1897583466]
}, Closed]],
Cell[1164948, 32944, 919, 25, 62, "Text",
 CellID->624927140],
Cell[1165870, 32971, 47, 1, 12, "OpenEnvironment",
 CellID->667675291],
Cell[CellGroupData[{
Cell[1165942, 32976, 657, 20, 40, "EnvironmentHeader",
 CellID->1051433415],
Cell[1166602, 32998, 248, 5, 33, "GlobalDeclaration",
 CellID->962205404],
Cell[1166853, 33005, 1768, 49, 40, "FormalTextInputFormula",
 CellTags->{
  "ID:1399613937", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 1"},
 CellID->1399613937],
Cell[1168624, 33056, 1552, 46, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1210171526", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 2"},
 CellID->1210171526],
Cell[1170179, 33104, 1104, 34, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1251355025", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 3"},
 CellID->1251355025],
Cell[1171286, 33140, 1223, 37, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1093640190", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 4"},
 CellID->1093640190],
Cell[1172512, 33179, 1700, 52, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:606212147", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a 5"},
 CellID->606212147],
Cell[1174215, 33233, 1521, 43, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1628665526", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a"},
 CellID->1628665526],
Cell[1175739, 33278, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1798898406]
}, Open  ]],
Cell[1175820, 33282, 47, 1, 16, "CloseEnvironment",
 CellID->39175054],
Cell[CellGroupData[{
Cell[1175892, 33287, 6632, 165, 29, "ProofInfo",
 CellTags->{"Proof|ID:1399613937", "Proof|ID:1399613937-1"},
 CellID->1391233604],
Cell[1182527, 33454, 483, 12, 70, "ProofInfoBody",
 CellID->713233342]
}, Closed]],
Cell[CellGroupData[{
Cell[1183047, 33471, 7199, 176, 27, "ProofInfo",
 CellTags->{"Proof|ID:1210171526", "Proof|ID:1210171526-1"},
 CellID->2096401104],
Cell[1190249, 33649, 483, 12, 70, "ProofInfoBody",
 CellID->848867484]
}, Closed]],
Cell[CellGroupData[{
Cell[1190769, 33666, 4940, 128, 27, "ProofInfo",
 CellTags->{"Proof|ID:1251355025", "Proof|ID:1251355025-1"},
 CellID->185930261],
Cell[1195712, 33796, 484, 12, 70, "ProofInfoBody",
 CellID->2087850823]
}, Closed]],
Cell[CellGroupData[{
Cell[1196233, 33813, 5208, 132, 27, "ProofInfo",
 CellTags->{"Proof|ID:1093640190", "Proof|ID:1093640190-1"},
 CellID->808217224],
Cell[1201444, 33947, 483, 12, 70, "ProofInfoBody",
 CellID->348117997]
}, Closed]],
Cell[CellGroupData[{
Cell[1201964, 33964, 7292, 178, 27, "ProofInfo",
 CellTags->{"Proof|ID:606212147", "Proof|ID:606212147-1"},
 CellID->1011788619],
Cell[1209259, 34144, 482, 12, 70, "ProofInfoBody",
 CellID->288491612]
}, Closed]],
Cell[CellGroupData[{
Cell[1209778, 34161, 5016, 128, 27, "ProofInfo",
 CellTags->{"Proof|ID:1628665526", "Proof|ID:1628665526-1"},
 CellID->975115047],
Cell[1214797, 34291, 484, 12, 70, "ProofInfoBody",
 CellID->1079430099]
}, Closed]],
Cell[1215296, 34306, 563, 17, 43, "Text",
 CellID->327532369],
Cell[1215862, 34325, 48, 1, 12, "OpenEnvironment",
 CellID->1255006385],
Cell[CellGroupData[{
Cell[1215935, 34330, 558, 18, 40, "EnvironmentHeader",
 CellID->542844259],
Cell[1216496, 34350, 200, 4, 33, "GlobalDeclaration",
 CellID->1379329847],
Cell[1216699, 34356, 258, 7, 41, "GlobalDeclaration",
 CellID->1704783354],
Cell[1216960, 34365, 1412, 39, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1351181837", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a I"},
 CellID->1351181837],
Cell[1218375, 34406, 1446, 44, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:2057269766", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a\[LessEqual]"},
 CellID->2057269766],
Cell[1219824, 34452, 1427, 42, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:94917498", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a II"},
 CellID->94917498],
Cell[1221254, 34496, 1596, 51, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:1065272829", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\:293a+\[LessEqual]"},
 CellID->1065272829],
Cell[1222853, 34549, 65, 1, 24, "EndEnvironmentMarker",
 CellID->659421005]
}, Open  ]],
Cell[1222933, 34553, 49, 1, 16, "CloseEnvironment",
 CellID->2130518472],
Cell[CellGroupData[{
Cell[1223007, 34558, 6071, 152, 29, "ProofInfo",
 CellTags->{"Proof|ID:1351181837", "Proof|ID:1351181837-1"},
 CellID->1829706754],
Cell[1229081, 34712, 525, 13, 70, "ProofInfoBody",
 CellID->468214080]
}, Closed]],
Cell[CellGroupData[{
Cell[1229643, 34730, 6320, 156, 27, "ProofInfo",
 CellTags->{"Proof|ID:2057269766", "Proof|ID:2057269766-1"},
 CellID->706843964],
Cell[1235966, 34888, 547, 13, 70, "ProofInfoBody",
 CellID->951036626]
}, Closed]],
Cell[CellGroupData[{
Cell[1236550, 34906, 6919, 170, 27, "ProofInfo",
 CellTags->{"Proof|ID:94917498", "Proof|ID:94917498-1"},
 CellID->1107419424],
Cell[1243472, 35078, 481, 12, 70, "ProofInfoBody",
 CellID->196846389]
}, Closed]],
Cell[CellGroupData[{
Cell[1243990, 35095, 7449, 180, 27, "ProofInfo",
 CellTags->{"Proof|ID:1065272829", "Proof|ID:1065272829-1"},
 CellID->1792160777],
Cell[1251442, 35277, 483, 12, 70, "ProofInfoBody",
 CellID->994864864]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1251974, 35295, 237, 4, 22, "Subsubsection",
 CellID->362221155],
Cell[1252214, 35301, 397, 9, 30, "GlobalDeclaration",
 CellID->2013493621],
Cell[1252614, 35312, 597, 17, 25, "Text",
 CellID->231154263],
Cell[1253214, 35331, 47, 1, 12, "OpenEnvironment",
 CellID->593345801],
Cell[CellGroupData[{
Cell[1253286, 35336, 488, 14, 41, "EnvironmentHeader",
 CellID->1877952819],
Cell[1253777, 35352, 228, 6, 32, "GlobalDeclaration",
 CellID->1922840150],
Cell[1254008, 35360, 943, 25, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1824645721", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of W (binary)"},
 CellID->1824645721],
Cell[1254954, 35387, 893, 24, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1345307293", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of W (unary)"},
 CellID->1345307293],
Cell[1255850, 35413, 1201, 33, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1185705395", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "43"},
 CellID->1185705395],
Cell[1257054, 35448, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1776977096]
}, Open  ]],
Cell[1257135, 35452, 49, 1, 16, "CloseEnvironment",
 CellID->2030887648],
Cell[CellGroupData[{
Cell[1257209, 35457, 3912, 106, 29, "ProofInfo",
 CellTags->{"Proof|ID:1824645721", "Proof|ID:1824645721-1"},
 CellID->286621138],
Cell[1261124, 35565, 484, 12, 70, "ProofInfoBody",
 CellID->1193162250]
}, Closed]],
Cell[CellGroupData[{
Cell[1261645, 35582, 3167, 88, 27, "ProofInfo",
 CellTags->{"Proof|ID:1345307293", "Proof|ID:1345307293-1"},
 CellID->28034114],
Cell[1264815, 35672, 484, 12, 70, "ProofInfoBody",
 CellID->1750002018]
}, Closed]],
Cell[CellGroupData[{
Cell[1265336, 35689, 6061, 156, 27, "ProofInfo",
 CellTags->{"Proof|ID:1185705395", "Proof|ID:1185705395-1"},
 CellID->2067392509],
Cell[1271400, 35847, 483, 12, 70, "ProofInfoBody",
 CellID->508522961]
}, Closed]],
Cell[1271898, 35862, 1195, 30, 44, "Text",
 CellID->2017434267],
Cell[1273096, 35894, 438, 13, 24, "Text",
 CellID->1656499058],
Cell[1273537, 35909, 48, 1, 12, "OpenEnvironment",
 CellID->1706918605],
Cell[CellGroupData[{
Cell[1273610, 35914, 552, 18, 40, "EnvironmentHeader",
 CellID->938515413],
Cell[1274165, 35934, 227, 6, 32, "GlobalDeclaration",
 CellID->1091363795],
Cell[1274395, 35942, 232, 6, 41, "GlobalDeclaration",
 CellID->110581283],
Cell[1274630, 35950, 1549, 46, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:354042430", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]W"},
 CellID->354042430],
Cell[1276182, 35998, 1538, 45, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:551974343", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\[LessEqual]"},
 CellID->551974343],
Cell[1277723, 36045, 65, 1, 24, "EndEnvironmentMarker",
 CellID->135917884]
}, Open  ]],
Cell[1277803, 36049, 49, 1, 16, "CloseEnvironment",
 CellID->1118462943],
Cell[CellGroupData[{
Cell[1277877, 36054, 7186, 180, 29, "ProofInfo",
 CellTags->{"Proof|ID:354042430", "Proof|ID:354042430-1"},
 CellID->1517092650],
Cell[1285066, 36236, 524, 13, 70, "ProofInfoBody",
 CellID->624328162]
}, Closed]],
Cell[CellGroupData[{
Cell[1285627, 36254, 7185, 180, 27, "ProofInfo",
 CellTags->{"Proof|ID:551974343", "Proof|ID:551974343-1"},
 CellID->766317937],
Cell[1292815, 36436, 524, 13, 70, "ProofInfoBody",
 CellID->604160603]
}, Closed]],
Cell[1293354, 36452, 1281, 37, 64, "Text",
 CellID->1123187194],
Cell[1294638, 36491, 47, 1, 12, "OpenEnvironment",
 CellID->622337541],
Cell[CellGroupData[{
Cell[1294710, 36496, 585, 19, 40, "EnvironmentHeader",
 CellID->416872943],
Cell[1295298, 36517, 228, 6, 32, "GlobalDeclaration",
 CellID->1770721675],
Cell[1295529, 36525, 199, 4, 33, "GlobalDeclaration",
 CellID->887909859],
Cell[1295731, 36531, 965, 28, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:894623443", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "44"},
 CellID->894623443],
Cell[1296699, 36561, 1128, 32, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1995654247", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "45"},
 CellID->1995654247],
Cell[1297830, 36595, 1240, 34, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:1916087968", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "46"},
 CellID->1916087968],
Cell[1299073, 36631, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1446445473]
}, Open  ]],
Cell[1299154, 36635, 48, 1, 16, "CloseEnvironment",
 CellID->523519669],
Cell[CellGroupData[{
Cell[1299227, 36640, 5465, 140, 29, "ProofInfo",
 CellTags->{"Proof|ID:894623443", "Proof|ID:894623443-1"},
 CellID->456641578],
Cell[1304695, 36782, 524, 13, 70, "ProofInfoBody",
 CellID->314453048]
}, Closed]],
Cell[CellGroupData[{
Cell[1305256, 36800, 6147, 156, 27, "ProofInfo",
 CellTags->{"Proof|ID:1995654247", "Proof|ID:1995654247-1"},
 CellID->1030159721],
Cell[1311406, 36958, 525, 13, 70, "ProofInfoBody",
 CellID->645849534]
}, Closed]],
Cell[CellGroupData[{
Cell[1311968, 36976, 6347, 160, 27, "ProofInfo",
 CellTags->{"Proof|ID:1916087968", "Proof|ID:1916087968-1"},
 CellID->1865547860],
Cell[1318318, 37138, 526, 13, 70, "ProofInfoBody",
 CellID->1576636564]
}, Closed]],
Cell[1318859, 37154, 1238, 40, 44, "Text",
 CellID->1668358382],
Cell[1320100, 37196, 48, 1, 12, "OpenEnvironment",
 CellID->1003278763],
Cell[CellGroupData[{
Cell[1320173, 37201, 778, 22, 40, "EnvironmentHeader",
 CellID->966932569],
Cell[1320954, 37225, 248, 5, 33, "GlobalDeclaration",
 CellID->930876469],
Cell[1321205, 37232, 228, 6, 32, "GlobalDeclaration",
 CellID->1247889975],
Cell[1321436, 37240, 1577, 41, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:2064198302", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a"},
 CellID->2064198302],
Cell[1323016, 37283, 948, 29, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:2144277606", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a I"},
 CellID->2144277606],
Cell[1323967, 37314, 1038, 30, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:388422476", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a II"},
 CellID->388422476],
Cell[1325008, 37346, 1030, 29, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:144808276", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a decreases"},
 CellID->144808276],
Cell[1326041, 37377, 65, 1, 24, "EndEnvironmentMarker",
 CellID->572158620]
}, Open  ]],
Cell[1326121, 37381, 48, 1, 16, "CloseEnvironment",
 CellID->944992805],
Cell[CellGroupData[{
Cell[1326194, 37386, 5903, 151, 29, "ProofInfo",
 CellTags->{"Proof|ID:2064198302", "Proof|ID:2064198302-1"},
 CellID->595904185],
Cell[1332100, 37539, 483, 12, 70, "ProofInfoBody",
 CellID->472964561]
}, Closed]],
Cell[CellGroupData[{
Cell[1332620, 37556, 4933, 129, 27, "ProofInfo",
 CellTags->{"Proof|ID:2144277606", "Proof|ID:2144277606-1"},
 CellID->391021459],
Cell[1337556, 37687, 484, 12, 70, "ProofInfoBody",
 CellID->1450565738]
}, Closed]],
Cell[CellGroupData[{
Cell[1338077, 37704, 5105, 134, 27, "ProofInfo",
 CellTags->{"Proof|ID:388422476", "Proof|ID:388422476-1"},
 CellID->743656091],
Cell[1343185, 37840, 483, 12, 70, "ProofInfoBody",
 CellID->1071526442]
}, Closed]],
Cell[CellGroupData[{
Cell[1343705, 37857, 3974, 106, 27, "ProofInfo",
 CellTags->{"Proof|ID:144808276", "Proof|ID:144808276-1"},
 CellID->1758271591],
Cell[1347682, 37965, 482, 12, 70, "ProofInfoBody",
 CellID->672863167]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1348213, 37983, 236, 4, 22, "Subsubsection",
 CellID->138333285],
Cell[1348452, 37989, 339, 9, 25, "Text",
 CellID->1953748933],
Cell[1348794, 38000, 47, 1, 12, "OpenEnvironment",
 CellID->474703228],
Cell[CellGroupData[{
Cell[1348866, 38005, 432, 13, 41, "EnvironmentHeader",
 CellID->925860718],
Cell[1349301, 38020, 1074, 28, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1149216477", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of L"},
 CellID->1149216477],
Cell[1350378, 38050, 65, 1, 24, "EndEnvironmentMarker",
 CellID->990555961]
}, Open  ]],
Cell[1350458, 38054, 48, 1, 16, "CloseEnvironment",
 CellID->880453912],
Cell[CellGroupData[{
Cell[1350531, 38059, 3049, 85, 29, "ProofInfo",
 CellTags->
  "Proof|ID:1149216477|Source:/home/amaletzk/Theorema \
Notebooks/Complexity/Complexity (no functor).nb",
 CellID->187961443],
Cell[1353583, 38146, 357, 10, 70, "ProofInfoBody",
 CellID->1995973371]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1354001, 38163, 253, 8, 30, "Subsection",
 CellTags->"sect::LemmataSpecificSpecial",
 CellID->1466386606],
Cell[1354257, 38173, 1195, 27, 82, "Text",
 CellID->765254624],
Cell[1355455, 38202, 322, 9, 21, "GlobalDeclaration",
 CellTags->{
  "ID:504163160", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->504163160],
Cell[1355780, 38213, 36, 1, 24, "Text",
 CellID->853094610],
Cell[1355819, 38216, 691, 18, 57, "GlobalDeclaration",
 CellTags->{
  "ID:649536122", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->649536122],
Cell[CellGroupData[{
Cell[1356535, 38238, 363, 6, 28, "Subsubsection",
 CellTags->"sect::Cover",
 CellID->676471010],
Cell[1356901, 38246, 1356, 36, 65, "Text",
 CellID->466566289],
Cell[1358260, 38284, 47, 1, 12, "OpenEnvironment",
 CellID->813984505],
Cell[CellGroupData[{
Cell[1358332, 38289, 382, 9, 40, "EnvironmentHeader",
 CellID->294236068],
Cell[1358717, 38300, 199, 4, 33, "GlobalDeclaration",
 CellID->432121352],
Cell[1358919, 38306, 228, 6, 32, "GlobalDeclaration",
 CellID->2038539178],
Cell[1359150, 38314, 1011, 30, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:465693853", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "47"},
 CellID->465693853],
Cell[1360164, 38346, 2424, 78, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:777205621", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "48"},
 CellID->777205621],
Cell[1362591, 38426, 1737, 51, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:42909295", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "49"},
 CellID->42909295],
Cell[1364331, 38479, 1942, 62, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:112873953", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "50"},
 CellID->112873953],
Cell[1366276, 38543, 65, 1, 24, "EndEnvironmentMarker",
 CellID->505349263]
}, Open  ]],
Cell[1366356, 38547, 49, 1, 16, "CloseEnvironment",
 CellID->1249260285],
Cell[CellGroupData[{
Cell[1366430, 38552, 4608, 122, 29, "ProofInfo",
 CellTags->{"Proof|ID:465693853", "Proof|ID:465693853-1"},
 CellID->2114883905],
Cell[1371041, 38676, 525, 13, 70, "ProofInfoBody",
 CellID->1227222626]
}, Closed]],
Cell[CellGroupData[{
Cell[1371603, 38694, 15311, 356, 27, "ProofInfo",
 CellTags->{"Proof|ID:777205621", "Proof|ID:777205621-1"},
 CellID->118523572],
Cell[1386917, 39052, 483, 12, 70, "ProofInfoBody",
 CellID->1502988398]
}, Closed]],
Cell[CellGroupData[{
Cell[1387437, 39069, 10171, 244, 27, "ProofInfo",
 CellTags->{"Proof|ID:42909295", "Proof|ID:42909295-1"},
 CellID->234727221],
Cell[1397611, 39315, 545, 13, 70, "ProofInfoBody",
 CellID->849552024]
}, Closed]],
Cell[CellGroupData[{
Cell[1398193, 39333, 11274, 268, 27, "ProofInfo",
 CellTags->{"Proof|ID:112873953", "Proof|ID:112873953-1"},
 CellID->334296225],
Cell[1409470, 39603, 545, 13, 70, "ProofInfoBody",
 CellID->1258513368]
}, Closed]],
Cell[1410030, 39619, 48, 1, 12, "OpenEnvironment",
 CellID->1565002200],
Cell[CellGroupData[{
Cell[1410103, 39624, 603, 14, 40, "EnvironmentHeader",
 CellID->1334741891],
Cell[1410709, 39640, 1579, 46, 66, "FormalTextInputFormula",
 CellTags->{
  "ID:315092621", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "cover"},
 CellID->315092621],
Cell[1412291, 39688, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1801799012]
}, Open  ]],
Cell[1412372, 39692, 48, 1, 16, "CloseEnvironment",
 CellID->768394677],
Cell[CellGroupData[{
Cell[1412445, 39697, 5482, 138, 29, "ProofInfo",
 CellTags->{"Proof|ID:315092621", "Proof|ID:315092621-1"},
 CellID->875989985],
Cell[1417930, 39837, 546, 13, 70, "ProofInfoBody",
 CellID->1116871946]
}, Closed]],
Cell[1418491, 39853, 807, 20, 44, "Text",
 CellID->1219107783],
Cell[CellGroupData[{
Cell[1419323, 39877, 823, 25, 25, "Item",
 CellID->1150682902],
Cell[1420149, 39904, 699, 19, 25, "Item",
 CellID->1824552351],
Cell[1420851, 39925, 412, 9, 25, "Item",
 CellID->1894858166]
}, Open  ]],
Cell[1421278, 39937, 1729, 45, 81, "Text",
 CellID->857096204]
}, Closed]],
Cell[CellGroupData[{
Cell[1423044, 39987, 177, 3, 22, "Subsubsection",
 CellID->526429181],
Cell[1423224, 39992, 268, 7, 24, "Text",
 CellID->656737055],
Cell[1423495, 40001, 48, 1, 12, "OpenEnvironment",
 CellID->1128068551],
Cell[CellGroupData[{
Cell[1423568, 40006, 620, 16, 40, "EnvironmentHeader",
 CellID->624744263],
Cell[1424191, 40024, 223, 5, 33, "GlobalDeclaration",
 CellID->413827810],
Cell[1424417, 40031, 1434, 44, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:744007071", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "51"},
 CellID->744007071],
Cell[1425854, 40077, 1368, 43, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:1720661645", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "52"},
 CellID->1720661645],
Cell[1427225, 40122, 1938, 62, 55, "FormalTextInputFormula",
 CellTags->{
  "ID:1022330126", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[VerticalBar]lcm"},
 CellID->1022330126],
Cell[1429166, 40186, 64, 1, 24, "EndEnvironmentMarker",
 CellID->73373163]
}, Open  ]],
Cell[1429245, 40190, 47, 1, 16, "CloseEnvironment",
 CellID->37865400],
Cell[CellGroupData[{
Cell[1429317, 40195, 6247, 158, 29, "ProofInfo",
 CellTags->{"Proof|ID:744007071", "Proof|ID:744007071-1"},
 CellID->1991259271],
Cell[1435567, 40355, 525, 13, 70, "ProofInfoBody",
 CellID->1421118479]
}, Closed]],
Cell[CellGroupData[{
Cell[1436129, 40373, 6251, 158, 27, "ProofInfo",
 CellTags->{"Proof|ID:1720661645", "Proof|ID:1720661645-1"},
 CellID->491886609],
Cell[1442383, 40533, 525, 13, 70, "ProofInfoBody",
 CellID->814496018]
}, Closed]],
Cell[CellGroupData[{
Cell[1442945, 40551, 6872, 170, 27, "ProofInfo",
 CellTags->{"Proof|ID:1022330126", "Proof|ID:1022330126-1"},
 CellID->448824604],
Cell[1449820, 40723, 526, 13, 70, "ProofInfoBody",
 CellID->1669274429]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1450395, 40742, 140, 3, 22, "Subsubsection",
 CellID->497056006],
Cell[1450538, 40747, 291, 6, 43, "Text",
 CellID->2048305540],
Cell[1450832, 40755, 47, 1, 12, "OpenEnvironment",
 CellID->537258638],
Cell[CellGroupData[{
Cell[1450904, 40760, 439, 10, 40, "EnvironmentHeader",
 CellID->575367155],
Cell[1451346, 40772, 1221, 38, 40, "FormalTextInputFormula",
 CellTags->{
  "ID:528222346", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "53"},
 CellID->528222346],
Cell[1452570, 40812, 1083, 33, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:1598627200", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "54"},
 CellID->1598627200],
Cell[1453656, 40847, 65, 1, 24, "EndEnvironmentMarker",
 CellID->728081085]
}, Open  ]],
Cell[1453736, 40851, 48, 1, 16, "CloseEnvironment",
 CellID->949543245],
Cell[CellGroupData[{
Cell[1453809, 40856, 4594, 124, 29, "ProofInfo",
 CellTags->{"Proof|ID:528222346", "Proof|ID:528222346-1"},
 CellID->1476471173],
Cell[1458406, 40982, 598, 15, 70, "ProofInfoBody",
 CellID->876298637]
}, Closed]],
Cell[CellGroupData[{
Cell[1459041, 41002, 3129, 90, 27, "ProofInfo",
 CellTags->{"Proof|ID:1598627200", "Proof|ID:1598627200-1"},
 CellID->605461210],
Cell[1462173, 41094, 525, 13, 70, "ProofInfoBody",
 CellID->162691003]
}, Closed]],
Cell[1462713, 41110, 48, 1, 12, "OpenEnvironment",
 CellID->1347133480],
Cell[CellGroupData[{
Cell[1462786, 41115, 377, 9, 40, "EnvironmentHeader",
 CellID->761168772],
Cell[1463166, 41126, 281, 6, 33, "GlobalDeclaration",
 CellID->2063990149],
Cell[1463450, 41134, 1434, 46, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:1537560967", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "55"},
 CellID->1537560967],
Cell[1464887, 41182, 1431, 46, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:507460317", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "56"},
 CellID->507460317],
Cell[1466321, 41230, 65, 1, 24, "EndEnvironmentMarker",
 CellID->565711075]
}, Open  ]],
Cell[1466401, 41234, 49, 1, 16, "CloseEnvironment",
 CellID->1713796769],
Cell[CellGroupData[{
Cell[1466475, 41239, 7077, 174, 29, "ProofInfo",
 CellTags->{"Proof|ID:1537560967", "Proof|ID:1537560967-1"},
 CellID->1213329230],
Cell[1473555, 41415, 526, 13, 70, "ProofInfoBody",
 CellID->1707761366]
}, Closed]],
Cell[CellGroupData[{
Cell[1474118, 41433, 7071, 174, 27, "ProofInfo",
 CellTags->{"Proof|ID:507460317", "Proof|ID:507460317-1"},
 CellID->688346270],
Cell[1481192, 41609, 525, 13, 70, "ProofInfoBody",
 CellID->1314245273]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1481766, 41628, 354, 6, 22, "Subsubsection",
 CellTags->"sect::LemmataSpecificSpecialW",
 CellID->1209432111],
Cell[1482123, 41636, 1436, 40, 84, "Text",
 CellID->434235030],
Cell[1483562, 41678, 44, 1, 12, "OpenEnvironment",
 CellID->551873],
Cell[CellGroupData[{
Cell[1483631, 41683, 869, 27, 41, "EnvironmentHeader",
 CellTags->"Lem::W[A<-x]",
 CellID->1283608134],
Cell[1484503, 41712, 228, 6, 32, "GlobalDeclaration",
 CellID->1093627130],
Cell[1484734, 41720, 441, 9, 48, "GlobalDeclaration",
 CellID->1464211085],
Cell[1485178, 41731, 859, 26, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:1442383873", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "57"},
 CellID->1442383873],
Cell[1486040, 41759, 1384, 33, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:523824248", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (binary) FarExterior 1"},
 CellID->523824248],
Cell[1487427, 41794, 1326, 34, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1663623856", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (binary) FarExterior 2"},
 CellID->1663623856],
Cell[1488756, 41830, 65, 1, 24, "EndEnvironmentMarker",
 CellID->249399937]
}, Open  ]],
Cell[1488836, 41834, 48, 1, 16, "CloseEnvironment",
 CellID->532566392],
Cell[CellGroupData[{
Cell[1488909, 41839, 6373, 162, 29, "ProofInfo",
 CellTags->{"Proof|ID:1442383873", "Proof|ID:1442383873-1"},
 CellID->1192136911],
Cell[1495285, 42003, 547, 13, 70, "ProofInfoBody",
 CellID->709403126]
}, Closed]],
Cell[CellGroupData[{
Cell[1495869, 42021, 5885, 151, 27, "ProofInfo",
 CellTags->{"Proof|ID:523824248", "Proof|ID:523824248-1"},
 CellID->842417444],
Cell[1501757, 42174, 482, 12, 70, "ProofInfoBody",
 CellID->960603845]
}, Closed]],
Cell[CellGroupData[{
Cell[1502276, 42191, 6395, 160, 27, "ProofInfo",
 CellTags->{"Proof|ID:1663623856", "Proof|ID:1663623856-1"},
 CellID->658477440],
Cell[1508674, 42353, 483, 12, 70, "ProofInfoBody",
 CellID->430830718]
}, Closed]],
Cell[1509172, 42368, 47, 1, 12, "OpenEnvironment",
 CellID->693346110],
Cell[CellGroupData[{
Cell[1509244, 42373, 746, 25, 41, "EnvironmentHeader",
 CellID->1916344980],
Cell[1509993, 42400, 198, 4, 33, "GlobalDeclaration",
 CellID->888020555],
Cell[1510194, 42406, 1579, 39, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1829996465", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (unary) FarExterior 1"},
 CellID->1829996465],
Cell[1511776, 42447, 1598, 40, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:842592091", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a (unary) FarExterior 2"},
 CellID->842592091],
Cell[1513377, 42489, 1269, 38, 47, "FormalTextInputFormula",
 CellTags->{
  "ID:1051080023", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "W\:293a FarExterior"},
 CellID->1051080023],
Cell[1514649, 42529, 65, 1, 24, "EndEnvironmentMarker",
 CellID->965026936]
}, Open  ]],
Cell[1514729, 42533, 48, 1, 16, "CloseEnvironment",
 CellID->686705729],
Cell[CellGroupData[{
Cell[1514802, 42538, 5411, 140, 29, "ProofInfo",
 CellTags->{"Proof|ID:1829996465", "Proof|ID:1829996465-1"},
 CellID->590566551],
Cell[1520216, 42680, 526, 13, 70, "ProofInfoBody",
 CellID->1758652648]
}, Closed]],
Cell[CellGroupData[{
Cell[1520779, 42698, 5397, 140, 27, "ProofInfo",
 CellTags->{"Proof|ID:842592091", "Proof|ID:842592091-1"},
 CellID->917529032],
Cell[1526179, 42840, 524, 13, 70, "ProofInfoBody",
 CellID->864478433]
}, Closed]],
Cell[CellGroupData[{
Cell[1526740, 42858, 6595, 166, 27, "ProofInfo",
 CellTags->{"Proof|ID:1051080023", "Proof|ID:1051080023-1"},
 CellID->148396000],
Cell[1533338, 43026, 526, 13, 70, "ProofInfoBody",
 CellID->1486101954]
}, Closed]],
Cell[1533879, 43042, 574, 18, 24, "Text",
 CellID->1033758105],
Cell[1534456, 43062, 48, 1, 12, "OpenEnvironment",
 CellID->1236240386],
Cell[CellGroupData[{
Cell[1534529, 43067, 481, 14, 40, "EnvironmentHeader",
 CellID->1017694636],
Cell[1535013, 43083, 228, 6, 32, "GlobalDeclaration",
 CellID->1667409472],
Cell[1535244, 43091, 997, 29, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:631716597", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "58"},
 CellID->631716597],
Cell[1536244, 43122, 980, 29, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1988188117", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "59"},
 CellID->1988188117],
Cell[1537227, 43153, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1087276993]
}, Open  ]],
Cell[1537308, 43157, 47, 1, 16, "CloseEnvironment",
 CellID->42773662],
Cell[CellGroupData[{
Cell[1537380, 43162, 5061, 133, 29, "ProofInfo",
 CellTags->{"Proof|ID:631716597", "Proof|ID:631716597-1"},
 CellID->1960584677],
Cell[1542444, 43297, 525, 13, 70, "ProofInfoBody",
 CellID->1246911685]
}, Closed]],
Cell[CellGroupData[{
Cell[1543006, 43315, 4899, 128, 27, "ProofInfo",
 CellTags->{"Proof|ID:1988188117", "Proof|ID:1988188117-1"},
 CellID->1726608150],
Cell[1547908, 43445, 483, 12, 70, "ProofInfoBody",
 CellID->737511693]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1548440, 43463, 165, 3, 22, "Subsubsection",
 CellID->414836815],
Cell[1548608, 43468, 712, 23, 43, "Text",
 CellID->648452100],
Cell[1549323, 43493, 47, 1, 12, "OpenEnvironment",
 CellID->641967087],
Cell[CellGroupData[{
Cell[1549395, 43498, 293, 8, 40, "EnvironmentHeader",
 CellID->1971506263],
Cell[1549691, 43508, 399, 12, 41, "GlobalDeclaration",
 CellTags->{
  "ID:590870546", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb"},
 CellID->590870546],
Cell[1550093, 43522, 1281, 38, 40, "FormalTextInputFormula",
 CellTags->{
  "ID:501569980", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "60"},
 CellID->501569980],
Cell[1551377, 43562, 1614, 46, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:914554779", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "61"},
 CellID->914554779],
Cell[1552994, 43610, 1748, 50, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1336064727", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "62"},
 CellID->1336064727],
Cell[1554745, 43662, 1710, 51, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1651077201", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "L+\[LessEqual]"},
 CellID->1651077201],
Cell[1556458, 43715, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1659661723]
}, Open  ]],
Cell[1556539, 43719, 48, 1, 16, "CloseEnvironment",
 CellID->455739212],
Cell[CellGroupData[{
Cell[1556612, 43724, 7748, 190, 29, "ProofInfo",
 CellTags->{"Proof|ID:501569980", "Proof|ID:501569980-1"},
 CellID->1185891342],
Cell[1564363, 43916, 483, 12, 70, "ProofInfoBody",
 CellID->1281666503]
}, Closed]],
Cell[CellGroupData[{
Cell[1564883, 43933, 7869, 194, 27, "ProofInfo",
 CellTags->{"Proof|ID:914554779", "Proof|ID:914554779-1"},
 CellID->5884268],
Cell[1572755, 44129, 523, 13, 70, "ProofInfoBody",
 CellID->830333471]
}, Closed]],
Cell[CellGroupData[{
Cell[1573315, 44147, 8362, 206, 27, "ProofInfo",
 CellTags->{"Proof|ID:1336064727", "Proof|ID:1336064727-1"},
 CellID->970257786],
Cell[1581680, 44355, 526, 13, 70, "ProofInfoBody",
 CellID->1589057661]
}, Closed]],
Cell[CellGroupData[{
Cell[1582243, 44373, 9382, 226, 27, "ProofInfo",
 CellTags->{"Proof|ID:1651077201", "Proof|ID:1651077201-1"},
 CellID->451306320],
Cell[1591628, 44601, 525, 13, 70, "ProofInfoBody",
 CellID->965210338]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1592202, 44620, 229, 4, 22, "Subsubsection",
 CellTags->"sect::chainCritSpecial",
 CellID->226754043],
Cell[1592434, 44626, 937, 26, 62, "Text",
 CellID->436231197],
Cell[1593374, 44654, 287, 10, 25, "Text",
 CellID->826985685],
Cell[1593664, 44666, 223, 5, 33, "GlobalDeclaration",
 CellID->905478503],
Cell[1593890, 44673, 175, 4, 24, "Text",
 CellID->1009550453],
Cell[1594068, 44679, 46, 1, 12, "OpenEnvironment",
 CellID->28139740],
Cell[CellGroupData[{
Cell[1594139, 44684, 518, 13, 40, "EnvironmentHeader",
 CellID->755849276],
Cell[1594660, 44699, 200, 4, 33, "GlobalDeclaration",
 CellID->2145956178],
Cell[1594863, 44705, 228, 6, 32, "GlobalDeclaration",
 CellID->1284341245],
Cell[1595094, 44713, 321, 9, 33, "GlobalDeclaration",
 CellID->325823042],
Cell[1595418, 44724, 1054, 30, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:53741691", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "63"},
 CellID->53741691],
Cell[1596475, 44756, 1014, 30, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:1045358942", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "64"},
 CellID->1045358942],
Cell[1597492, 44788, 1068, 32, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1132937314", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "65"},
 CellID->1132937314],
Cell[1598563, 44822, 1066, 32, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:152254973", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "66"},
 CellID->152254973],
Cell[1599632, 44856, 65, 1, 24, "EndEnvironmentMarker",
 CellID->204886373]
}, Open  ]],
Cell[1599712, 44860, 49, 1, 16, "CloseEnvironment",
 CellID->1003232302],
Cell[CellGroupData[{
Cell[1599786, 44865, 6652, 167, 29, "ProofInfo",
 CellTags->{"Proof|ID:53741691", "Proof|ID:53741691-1"},
 CellID->1005730830],
Cell[1606441, 45034, 524, 13, 70, "ProofInfoBody",
 CellID->1697566003]
}, Closed]],
Cell[CellGroupData[{
Cell[1607002, 45052, 6662, 167, 27, "ProofInfo",
 CellTags->{"Proof|ID:1045358942", "Proof|ID:1045358942-1"},
 CellID->1783727900],
Cell[1613667, 45221, 573, 14, 58, "ProofInfoBody",
 CellID->74922882]
}, Closed]],
Cell[CellGroupData[{
Cell[1614277, 45240, 6547, 165, 27, "ProofInfo",
 CellTags->{"Proof|ID:1132937314", "Proof|ID:1132937314-1"},
 CellID->279556406],
Cell[1620827, 45407, 525, 13, 70, "ProofInfoBody",
 CellID->982255660]
}, Closed]],
Cell[CellGroupData[{
Cell[1621389, 45425, 6543, 165, 27, "ProofInfo",
 CellTags->{"Proof|ID:152254973", "Proof|ID:152254973-1"},
 CellID->1134289223],
Cell[1627935, 45592, 523, 13, 70, "ProofInfoBody",
 CellID->39010603]
}, Closed]],
Cell[1628473, 45608, 1113, 33, 82, "Text",
 CellID->891029648],
Cell[1629589, 45643, 48, 1, 12, "OpenEnvironment",
 CellID->1820021401],
Cell[CellGroupData[{
Cell[1629662, 45648, 489, 11, 40, "EnvironmentHeader",
 CellID->739473411],
Cell[1630154, 45661, 227, 6, 32, "GlobalDeclaration",
 CellID->804277790],
Cell[1630384, 45669, 1000, 25, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:836877533", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 1"},
 CellID->836877533],
Cell[1631387, 45696, 1071, 28, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1226632396", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 2"},
 CellID->1226632396],
Cell[1632461, 45726, 1264, 32, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:457131364", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal 3"},
 CellID->457131364],
Cell[1633728, 45760, 1020, 27, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1490326930", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "symmetry of chainCritDiagonal"},
 CellID->1490326930],
Cell[1634751, 45789, 2256, 68, 112, "FormalTextInputFormula",
 CellTags->{
  "ID:749802963", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCritDiagonal w.r.t. first coordinate"},
 CellID->749802963],
Cell[1637010, 45859, 243, 6, 21, "GlobalDeclaration",
 CellID->2008961285],
Cell[1637256, 45867, 1026, 29, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:505706706", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "67"},
 CellID->505706706],
Cell[1638285, 45898, 2989, 83, 71, "FormalTextInputFormula",
 CellTags->{
  "ID:1262287499", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "68"},
 CellID->1262287499],
Cell[1641277, 45983, 65, 1, 24, "EndEnvironmentMarker",
 CellID->263104304]
}, Open  ]],
Cell[1641357, 45987, 48, 1, 16, "CloseEnvironment",
 CellID->350642411],
Cell[CellGroupData[{
Cell[1641430, 45992, 6076, 150, 29, "ProofInfo",
 CellTags->{"Proof|ID:836877533", "Proof|ID:836877533-1"},
 CellID->313941599],
Cell[1647509, 46144, 525, 13, 70, "ProofInfoBody",
 CellID->1689572409]
}, Closed]],
Cell[CellGroupData[{
Cell[1648071, 46162, 6107, 150, 27, "ProofInfo",
 CellTags->{"Proof|ID:1226632396", "Proof|ID:1226632396-1"},
 CellID->346366551],
Cell[1654181, 46314, 524, 13, 70, "ProofInfoBody",
 CellID->13987266]
}, Closed]],
Cell[CellGroupData[{
Cell[1654742, 46332, 6103, 150, 27, "ProofInfo",
 CellTags->{"Proof|ID:457131364", "Proof|ID:457131364-1"},
 CellID->1718476676],
Cell[1660848, 46484, 523, 13, 70, "ProofInfoBody",
 CellID->89734148]
}, Closed]],
Cell[CellGroupData[{
Cell[1661408, 46502, 4882, 122, 27, "ProofInfo",
 CellTags->{"Proof|ID:1490326930", "Proof|ID:1490326930-1"},
 CellID->758177161],
Cell[1666293, 46626, 526, 13, 70, "ProofInfoBody",
 CellID->1459213616]
}, Closed]],
Cell[CellGroupData[{
Cell[1666856, 46644, 11170, 262, 27, "ProofInfo",
 CellTags->{"Proof|ID:749802963", "Proof|ID:749802963-1"},
 CellID->1283930748],
Cell[1678029, 46908, 524, 13, 70, "ProofInfoBody",
 CellID->443166725]
}, Closed]],
Cell[CellGroupData[{
Cell[1678590, 46926, 5957, 150, 27, "ProofInfo",
 CellTags->{"Proof|ID:505706706", "Proof|ID:505706706-1"},
 CellID->292146520],
Cell[1684550, 47078, 524, 13, 70, "ProofInfoBody",
 CellID->389998305]
}, Closed]],
Cell[CellGroupData[{
Cell[1685111, 47096, 13803, 320, 27, "ProofInfo",
 CellTags->{"Proof|ID:1262287499", "Proof|ID:1262287499-1"},
 CellID->1935805160],
Cell[1698917, 47418, 525, 13, 70, "ProofInfoBody",
 CellID->1719653774]
}, Closed]],
Cell[1699457, 47434, 279, 6, 43, "Text",
 CellID->6932400],
Cell[1699739, 47442, 48, 1, 12, "OpenEnvironment",
 CellID->1921149736],
Cell[CellGroupData[{
Cell[1699812, 47447, 523, 13, 40, "EnvironmentHeader",
 CellID->436470413],
Cell[1700338, 47462, 296, 8, 33, "GlobalDeclaration",
 CellID->1928470162],
Cell[1700637, 47472, 1576, 50, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:1909233179", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "69"},
 CellID->1909233179],
Cell[1702216, 47524, 1682, 54, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:2125747942", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "70"},
 CellID->2125747942],
Cell[1703901, 47580, 1776, 55, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:1211430036", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "71"},
 CellID->1211430036],
Cell[1705680, 47637, 622, 16, 21, "GlobalDeclaration",
 CellID->643000006],
Cell[1706305, 47655, 900, 25, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:2061027934", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "72"},
 CellID->2061027934],
Cell[1707208, 47682, 778, 23, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:1653917255", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "73"},
 CellID->1653917255],
Cell[1707989, 47707, 65, 1, 24, "EndEnvironmentMarker",
 CellID->638144522]
}, Open  ]],
Cell[1708069, 47711, 49, 1, 16, "CloseEnvironment",
 CellID->1008755936],
Cell[CellGroupData[{
Cell[1708143, 47716, 11159, 264, 29, "ProofInfo",
 CellTags->{"Proof|ID:1909233179", "Proof|ID:1909233179-1"},
 CellID->1389502986],
Cell[1719305, 47982, 525, 13, 70, "ProofInfoBody",
 CellID->972044349]
}, Closed]],
Cell[CellGroupData[{
Cell[1719867, 48000, 11434, 272, 27, "ProofInfo",
 CellTags->{"Proof|ID:2125747942", "Proof|ID:2125747942-1"},
 CellID->673958282],
Cell[1731304, 48274, 524, 13, 70, "ProofInfoBody",
 CellID->71081292]
}, Closed]],
Cell[CellGroupData[{
Cell[1731865, 48292, 11435, 272, 27, "ProofInfo",
 CellTags->{"Proof|ID:1211430036", "Proof|ID:1211430036-1"},
 CellID->1679514424],
Cell[1743303, 48566, 524, 13, 70, "ProofInfoBody",
 CellID->321998331]
}, Closed]],
Cell[CellGroupData[{
Cell[1743864, 48584, 7690, 184, 27, "ProofInfo",
 CellTags->{"Proof|ID:2061027934", "Proof|ID:2061027934-1"},
 CellID->381202233],
Cell[1751557, 48770, 526, 13, 70, "ProofInfoBody",
 CellID->1850172725]
}, Closed]],
Cell[CellGroupData[{
Cell[1752120, 48788, 7690, 184, 27, "ProofInfo",
 CellTags->{"Proof|ID:1653917255", "Proof|ID:1653917255-1"},
 CellID->929231870],
Cell[1759813, 48974, 526, 13, 70, "ProofInfoBody",
 CellID->2027019073]
}, Closed]],
Cell[1760354, 48990, 1223, 30, 62, "Text",
 CellID->1006068998],
Cell[1761580, 49022, 47, 1, 12, "OpenEnvironment",
 CellID->695996923],
Cell[CellGroupData[{
Cell[1761652, 49027, 607, 12, 40, "EnvironmentHeader",
 CellID->750017357],
Cell[1762262, 49041, 1132, 27, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1690170957", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 1"},
 CellID->1690170957],
Cell[1763397, 49070, 926, 24, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:271680825", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 2"},
 CellID->271680825],
Cell[1764326, 49096, 3340, 100, 111, "FormalTextInputFormula",
 CellTags->{
  "ID:1572461042", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 3"},
 CellID->1572461042],
Cell[1767669, 49198, 1385, 41, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:2046865956", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 4"},
 CellID->2046865956],
Cell[1769057, 49241, 1244, 35, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:917844397", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate 5"},
 CellID->917844397],
Cell[1770304, 49278, 2181, 49, 53, "FormalTextInputFormula",
 CellTags->{
  "ID:810037327", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate"},
 CellID->810037327],
Cell[1772488, 49329, 1642, 50, 71, "FormalTextInputFormula",
 CellTags->{
  "ID:2111681775", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit bivariate I"},
 CellID->2111681775],
Cell[1774133, 49381, 65, 1, 24, "EndEnvironmentMarker",
 CellID->311506677]
}, Open  ]],
Cell[1774213, 49385, 49, 1, 16, "CloseEnvironment",
 CellID->1706489092],
Cell[CellGroupData[{
Cell[1774287, 49390, 4845, 122, 29, "ProofInfo",
 CellTags->{"Proof|ID:1690170957", "Proof|ID:1690170957-1"},
 CellID->2085296512],
Cell[1779135, 49514, 525, 13, 70, "ProofInfoBody",
 CellID->346853541]
}, Closed]],
Cell[CellGroupData[{
Cell[1779697, 49532, 4788, 120, 27, "ProofInfo",
 CellTags->{"Proof|ID:271680825", "Proof|ID:271680825-1"},
 CellID->1844198183],
Cell[1784488, 49654, 525, 13, 70, "ProofInfoBody",
 CellID->1602314431]
}, Closed]],
Cell[CellGroupData[{
Cell[1785050, 49672, 15570, 360, 27, "ProofInfo",
 CellTags->{"Proof|ID:1572461042", "Proof|ID:1572461042-1"},
 CellID->1050587621],
Cell[1800623, 50034, 483, 12, 70, "ProofInfoBody",
 CellID->966276044]
}, Closed]],
Cell[CellGroupData[{
Cell[1801143, 50051, 7087, 170, 27, "ProofInfo",
 CellTags->{"Proof|ID:2046865956", "Proof|ID:2046865956-1"},
 CellID->1535230662],
Cell[1808233, 50223, 525, 13, 70, "ProofInfoBody",
 CellID->155732320]
}, Closed]],
Cell[CellGroupData[{
Cell[1808795, 50241, 6249, 150, 27, "ProofInfo",
 CellTags->{"Proof|ID:917844397", "Proof|ID:917844397-1"},
 CellID->665088749],
Cell[1815047, 50393, 524, 13, 70, "ProofInfoBody",
 CellID->537828057]
}, Closed]],
Cell[CellGroupData[{
Cell[1815608, 50411, 5682, 138, 27, "ProofInfo",
 CellTags->{"Proof|ID:810037327", "Proof|ID:810037327-1"},
 CellID->1607548979],
Cell[1821293, 50551, 525, 13, 70, "ProofInfoBody",
 CellID->1719284110]
}, Closed]],
Cell[CellGroupData[{
Cell[1821855, 50569, 7475, 175, 27, "ProofInfo",
 CellTags->{"Proof|ID:2111681775", "Proof|ID:2111681775-1"},
 CellID->215040822],
Cell[1829333, 50746, 526, 13, 70, "ProofInfoBody",
 CellID->1969341009]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1829932, 50767, 289, 5, 43, "Section",
 CellTags->"sect::Theorems",
 CellID->1928916681],
Cell[1830224, 50774, 1266, 30, 82, "Text",
 CellID->1182788500],
Cell[1831493, 50806, 1032, 24, 64, "Text",
 CellID->976970424],
Cell[CellGroupData[{
Cell[1832550, 50834, 363, 9, 38, "Subsection",
 CellTags->"sect::M\[RawStar]",
 CellID->391063855],
Cell[1832916, 50845, 1775, 44, 139, "Text",
 CellID->457008407],
Cell[1834694, 50891, 297, 6, 33, "GlobalDeclaration",
 CellID->1760611200],
Cell[1834994, 50899, 1057, 34, 63, "Text",
 CellID->1013604180],
Cell[1836054, 50935, 48, 1, 12, "OpenEnvironment",
 CellID->1145076435],
Cell[CellGroupData[{
Cell[1836127, 50940, 504, 15, 40, "EnvironmentHeader",
 CellID->800620613],
Cell[1836634, 50957, 2387, 65, 47, "FormalTextInputFormula",
 CellTags->{
  "ID:2973275", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar] binary"},
 CellID->2973275],
Cell[1839024, 51024, 1541, 43, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:1258784171", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar] unary"},
 CellID->1258784171],
Cell[1840568, 51069, 64, 1, 24, "EndEnvironmentMarker",
 CellID->83476417]
}, Open  ]],
Cell[1840647, 51073, 49, 1, 16, "CloseEnvironment",
 CellID->2104356565],
Cell[1840699, 51076, 724, 17, 45, "Text",
 CellID->1285990268],
Cell[1841426, 51095, 298, 7, 30, "GlobalDeclaration",
 CellID->927538807],
Cell[1841727, 51104, 48, 1, 12, "OpenEnvironment",
 CellID->1868882626],
Cell[CellGroupData[{
Cell[1841800, 51109, 514, 15, 40, "EnvironmentHeader",
 CellID->2108232500],
Cell[1842317, 51126, 296, 8, 33, "GlobalDeclaration",
 CellID->1941261269],
Cell[1842616, 51136, 1393, 35, 47, "FormalTextInputFormula",
 CellTags->{
  "ID:1549582561", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "chainCrit holds for distinct indices"},
 CellID->1549582561],
Cell[1844012, 51173, 1470, 41, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:1966722449", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "74"},
 CellID->1966722449],
Cell[1845485, 51216, 1405, 40, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:164254989", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "75"},
 CellID->164254989],
Cell[1846893, 51258, 1018, 27, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:290364666", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M\[RawStar] (binary)"},
 CellID->290364666],
Cell[1847914, 51287, 974, 27, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1063863080", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "type of M\[RawStar] (unary)"},
 CellID->1063863080],
Cell[1848891, 51316, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1097467670]
}, Open  ]],
Cell[1848972, 51320, 49, 1, 16, "CloseEnvironment",
 CellID->1183566659],
Cell[CellGroupData[{
Cell[1849046, 51325, 6957, 172, 29, "ProofInfo",
 CellTags->{"Proof|ID:1549582561", "Proof|ID:1549582561-1"},
 CellID->1962722847],
Cell[1856006, 51499, 526, 13, 70, "ProofInfoBody",
 CellID->2031482966]
}, Closed]],
Cell[CellGroupData[{
Cell[1856569, 51517, 9170, 218, 27, "ProofInfo",
 CellTags->{"Proof|ID:1966722449", "Proof|ID:1966722449-1"},
 CellID->938794098],
Cell[1865742, 51737, 524, 13, 70, "ProofInfoBody",
 CellID->49039736]
}, Closed]],
Cell[CellGroupData[{
Cell[1866303, 51755, 8963, 214, 27, "ProofInfo",
 CellTags->{"Proof|ID:164254989", "Proof|ID:164254989-1"},
 CellID->895811688],
Cell[1875269, 51971, 524, 13, 70, "ProofInfoBody",
 CellID->774539108]
}, Closed]],
Cell[CellGroupData[{
Cell[1875830, 51989, 4836, 126, 27, "ProofInfo",
 CellTags->{"Proof|ID:290364666", "Proof|ID:290364666-1"},
 CellID->1493869216],
Cell[1880669, 52117, 525, 13, 70, "ProofInfoBody",
 CellID->1990230967]
}, Closed]],
Cell[CellGroupData[{
Cell[1881231, 52135, 3500, 96, 27, "ProofInfo",
 CellTags->{"Proof|ID:1063863080", "Proof|ID:1063863080-1"},
 CellID->391529240],
Cell[1884734, 52233, 525, 13, 70, "ProofInfoBody",
 CellID->690333880]
}, Closed]],
Cell[1885274, 52249, 3029, 83, 198, "Text",
 CellID->1583197414],
Cell[1888306, 52334, 599, 18, 24, "Text",
 CellID->809498243],
Cell[1888908, 52354, 47, 1, 12, "OpenEnvironment",
 CellID->378652997],
Cell[CellGroupData[{
Cell[1888980, 52359, 605, 20, 40, "EnvironmentHeader",
 CellID->1036416147],
Cell[1889588, 52381, 232, 6, 41, "GlobalDeclaration",
 CellID->255426543],
Cell[1889823, 52389, 344, 9, 33, "GlobalDeclaration",
 CellID->1342888001],
Cell[1890170, 52400, 1658, 42, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:1680577066", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M\[RawStar] (unary)"},
 CellID->1680577066],
Cell[1891831, 52444, 1729, 44, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:1866021817", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual] (unary)"},
 CellID->1866021817],
Cell[1893563, 52490, 2109, 58, 54, "FormalTextInputFormula",
 CellTags->{
  "ID:1272742036", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "\[LessEqual]M\[RawStar] (binary)"},
 CellID->1272742036],
Cell[1895675, 52550, 1913, 55, 54, "FormalTextInputFormula",
 CellTags->{
  "ID:462890005", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual] (binary)"},
 CellID->462890005],
Cell[1897591, 52607, 65, 1, 24, "EndEnvironmentMarker",
 CellID->982813600]
}, Open  ]],
Cell[1897671, 52611, 48, 1, 16, "CloseEnvironment",
 CellID->934500559],
Cell[CellGroupData[{
Cell[1897744, 52616, 6497, 164, 29, "ProofInfo",
 CellTags->{"Proof|ID:1680577066", "Proof|ID:1680577066-1"},
 CellID->631558205],
Cell[1904244, 52782, 525, 13, 70, "ProofInfoBody",
 CellID->503475645]
}, Closed]],
Cell[CellGroupData[{
Cell[1904806, 52800, 6486, 164, 27, "ProofInfo",
 CellTags->{"Proof|ID:1866021817", "Proof|ID:1866021817-1"},
 CellID->860652560],
Cell[1911295, 52966, 525, 13, 70, "ProofInfoBody",
 CellID->917867914]
}, Closed]],
Cell[CellGroupData[{
Cell[1911857, 52984, 10477, 248, 27, "ProofInfo",
 CellTags->{"Proof|ID:1272742036", "Proof|ID:1272742036-1"},
 CellID->135701756],
Cell[1922337, 53234, 526, 13, 70, "ProofInfoBody",
 CellID->1766556048]
}, Closed]],
Cell[CellGroupData[{
Cell[1922900, 53252, 10513, 250, 27, "ProofInfo",
 CellTags->{"Proof|ID:462890005", "Proof|ID:462890005-1"},
 CellID->829755125],
Cell[1933416, 53504, 523, 13, 70, "ProofInfoBody",
 CellID->421237545]
}, Closed]],
Cell[1933954, 53520, 1232, 36, 63, "Text",
 CellID->1211263146],
Cell[1935189, 53558, 90, 2, 12, "OpenEnvironment",
 CellID->1043926589],
Cell[CellGroupData[{
Cell[1935304, 53564, 971, 25, 40, "EnvironmentHeader",
 CellID->1274870949],
Cell[1936278, 53591, 873, 24, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:929561290", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual]M\[RawStar]"},
 CellID->929561290],
Cell[1937154, 53617, 872, 24, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:795795393", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[RawStar]\[LessEqual]M"},
 CellID->795795393],
Cell[1938029, 53643, 1234, 29, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:872843574", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M=M\[RawStar]"},
 CellID->872843574],
Cell[1939266, 53674, 107, 2, 24, "EndEnvironmentMarker",
 CellID->325406659]
}, Open  ]],
Cell[1939388, 53679, 90, 2, 16, "CloseEnvironment",
 CellID->692941230],
Cell[CellGroupData[{
Cell[1939503, 53685, 3661, 100, 29, "ProofInfo",
 CellTags->{"Proof|ID:929561290", "Proof|ID:929561290-1"},
 CellID->286309708],
Cell[1943167, 53787, 525, 13, 70, "ProofInfoBody",
 CellID->2072324147]
}, Closed]],
Cell[CellGroupData[{
Cell[1943729, 53805, 3662, 100, 27, "ProofInfo",
 CellTags->{"Proof|ID:795795393", "Proof|ID:795795393-1"},
 CellID->1883422493],
Cell[1947394, 53907, 525, 13, 70, "ProofInfoBody",
 CellID->1637224044]
}, Closed]],
Cell[CellGroupData[{
Cell[1947956, 53925, 3562, 98, 27, "ProofInfo",
 CellTags->{"Proof|ID:872843574", "Proof|ID:872843574-1"},
 CellID->1077588120],
Cell[1951521, 54025, 525, 13, 70, "ProofInfoBody",
 CellID->1552594980]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1952095, 54044, 229, 4, 30, "Subsection",
 CellTags->"sect::TheoremsGeneral",
 CellID->1847669776],
Cell[1952327, 54050, 624, 17, 45, "Text",
 CellID->2105593863],
Cell[1952954, 54069, 757, 16, 57, "GlobalDeclaration",
 CellTags->"???",
 CellID->560938714],
Cell[1953714, 54087, 1479, 36, 63, "Text",
 CellID->72362671],
Cell[1955196, 54125, 90, 2, 12, "OpenEnvironment",
 CellID->1437260198],
Cell[CellGroupData[{
Cell[1955311, 54131, 759, 19, 40, "EnvironmentHeader",
 CellID->852970621],
Cell[1956073, 54152, 1033, 29, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:642765392", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "76"},
 CellID->642765392],
Cell[1957109, 54183, 932, 24, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:592779258", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "maxdeg\[LessEqual]M"},
 CellID->592779258],
Cell[1958044, 54209, 107, 2, 24, "EndEnvironmentMarker",
 CellID->459797040]
}, Open  ]],
Cell[1958166, 54214, 89, 2, 16, "CloseEnvironment",
 CellID->52345895],
Cell[CellGroupData[{
Cell[1958280, 54220, 4620, 122, 29, "ProofInfo",
 CellTags->{"Proof|ID:642765392", "Proof|ID:642765392-1"},
 CellID->1075234758],
Cell[1962903, 54344, 525, 13, 70, "ProofInfoBody",
 CellID->1622776262]
}, Closed]],
Cell[CellGroupData[{
Cell[1963465, 54362, 3349, 92, 27, "ProofInfo",
 CellTags->{"Proof|ID:592779258", "Proof|ID:592779258-1"},
 CellID->582763895],
Cell[1966817, 54456, 547, 13, 70, "ProofInfoBody",
 CellID->1924730576]
}, Closed]],
Cell[1967379, 54472, 490, 16, 24, "Text",
 CellID->703221276],
Cell[1967872, 54490, 47, 1, 12, "OpenEnvironment",
 CellID->527748213],
Cell[CellGroupData[{
Cell[1967944, 54495, 604, 17, 40, "EnvironmentHeader",
 CellID->1751511295],
Cell[1968551, 54514, 296, 8, 33, "GlobalDeclaration",
 CellID->1051492688],
Cell[1968850, 54524, 890, 24, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:691015215", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "77"},
 CellID->691015215],
Cell[1969743, 54550, 1254, 33, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:402511110", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "78"},
 CellID->402511110],
Cell[1971000, 54585, 66, 1, 24, "EndEnvironmentMarker",
 CellID->2092750413]
}, Open  ]],
Cell[1971081, 54589, 48, 1, 16, "CloseEnvironment",
 CellID->938769538],
Cell[CellGroupData[{
Cell[1971154, 54594, 4446, 118, 29, "ProofInfo",
 CellTags->{"Proof|ID:691015215", "Proof|ID:691015215-1"},
 CellID->31238831],
Cell[1975603, 54714, 525, 13, 70, "ProofInfoBody",
 CellID->2057268967]
}, Closed]],
Cell[CellGroupData[{
Cell[1976165, 54732, 6108, 156, 27, "ProofInfo",
 CellTags->{"Proof|ID:402511110", "Proof|ID:402511110-1"},
 CellID->1203109870],
Cell[1982276, 54890, 525, 13, 70, "ProofInfoBody",
 CellID->1020066581]
}, Closed]],
Cell[1982816, 54906, 514, 12, 24, "Text",
 CellID->1406772823],
Cell[1983333, 54920, 48, 1, 12, "OpenEnvironment",
 CellID->1681684092],
Cell[CellGroupData[{
Cell[1983406, 54925, 315, 8, 40, "EnvironmentHeader",
 CellID->1926603590],
Cell[1983724, 54935, 973, 26, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:341822661", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "M\[LessEqual]L"},
 CellID->341822661],
Cell[1984700, 54963, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1589678083]
}, Open  ]],
Cell[1984781, 54967, 47, 1, 16, "CloseEnvironment",
 CellID->71356467],
Cell[CellGroupData[{
Cell[1984853, 54972, 3420, 94, 29, "ProofInfo",
 CellTags->{"Proof|ID:341822661", "Proof|ID:341822661-1"},
 CellID->1958581994],
Cell[1988276, 55068, 525, 13, 70, "ProofInfoBody",
 CellID->1851915245]
}, Closed]],
Cell[1988816, 55084, 346, 8, 44, "Text",
 CellID->2026828730]
}, Closed]],
Cell[CellGroupData[{
Cell[1989199, 55097, 246, 8, 30, "Subsection",
 CellTags->"sect::TheoremsSpecial",
 CellID->240164494],
Cell[1989448, 55107, 798, 21, 44, "Text",
 CellID->1113483777],
Cell[1990249, 55130, 182, 4, 21, "GlobalDeclaration",
 CellID->1550838377],
Cell[1990434, 55136, 37, 1, 24, "Text",
 CellID->1114476607],
Cell[1990474, 55139, 757, 16, 57, "GlobalDeclaration",
 CellTags->"???",
 CellID->299478056],
Cell[1991234, 55157, 37, 1, 24, "Text",
 CellID->2116435651],
Cell[1991274, 55160, 249, 5, 33, "GlobalDeclaration",
 CellID->1547442960],
Cell[CellGroupData[{
Cell[1991548, 55169, 271, 4, 28, "Subsubsection",
 CellID->679066213],
Cell[1991822, 55175, 1272, 29, 82, "Text",
 CellID->1198933695],
Cell[1993097, 55206, 47, 1, 12, "OpenEnvironment",
 CellID->140872528],
Cell[CellGroupData[{
Cell[1993169, 55211, 397, 9, 40, "EnvironmentHeader",
 CellID->260198431],
Cell[1993569, 55222, 227, 6, 32, "GlobalDeclaration",
 CellID->323697053],
Cell[1993799, 55230, 343, 6, 33, "GlobalDeclaration",
 CellID->1753587106],
Cell[1994145, 55238, 315, 7, 21, "GlobalDeclaration",
 CellID->1153504918],
Cell[1994463, 55247, 1421, 40, 56, "FormalTextInputFormula",
 CellTags->{
  "ID:245124153", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "79"},
 CellID->245124153],
Cell[1995887, 55289, 1437, 47, 63, "FormalTextInputFormula",
 CellTags->{
  "ID:952627996", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "80"},
 CellID->952627996],
Cell[1997327, 55338, 602, 17, 33, "GlobalDeclaration",
 CellID->301509033],
Cell[1997932, 55357, 966, 31, 81, "GlobalDeclaration",
 CellID->1728650251],
Cell[1998901, 55390, 1453, 31, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:2013760956", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "81"},
 CellID->2013760956],
Cell[2000357, 55423, 964, 28, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1011330545", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "82"},
 CellID->1011330545],
Cell[2001324, 55453, 1278, 37, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:718586393", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "83"},
 CellID->718586393],
Cell[2002605, 55492, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1994638750]
}, Open  ]],
Cell[2002686, 55496, 48, 1, 16, "CloseEnvironment",
 CellID->159419655],
Cell[CellGroupData[{
Cell[2002759, 55501, 8975, 216, 29, "ProofInfo",
 CellTags->{"Proof|ID:245124153", "Proof|ID:245124153-1"},
 CellID->1061920785],
Cell[2011737, 55719, 525, 13, 70, "ProofInfoBody",
 CellID->1202825855]
}, Closed]],
Cell[CellGroupData[{
Cell[2012299, 55737, 9879, 236, 27, "ProofInfo",
 CellTags->{"Proof|ID:952627996", "Proof|ID:952627996-1"},
 CellID->921465591],
Cell[2022181, 55975, 524, 13, 70, "ProofInfoBody",
 CellID->189186220]
}, Closed]],
Cell[CellGroupData[{
Cell[2022742, 55993, 15095, 350, 27, "ProofInfo",
 CellTags->{"Proof|ID:2013760956", "Proof|ID:2013760956-1"},
 CellID->605609267],
Cell[2037840, 56345, 524, 13, 70, "ProofInfoBody",
 CellID->47464659]
}, Closed]],
Cell[CellGroupData[{
Cell[2038401, 56363, 15728, 364, 27, "ProofInfo",
 CellTags->{"Proof|ID:1011330545", "Proof|ID:1011330545-1"},
 CellID->1676671824],
Cell[2054132, 56729, 526, 13, 70, "ProofInfoBody",
 CellID->1372119806]
}, Closed]],
Cell[CellGroupData[{
Cell[2054695, 56747, 12863, 300, 27, "ProofInfo",
 CellTags->{"Proof|ID:718586393", "Proof|ID:718586393-1"},
 CellID->1047095018],
Cell[2067561, 57049, 523, 13, 70, "ProofInfoBody",
 CellID->866926986]
}, Closed]],
Cell[2068099, 57065, 4303, 148, 204, "Text",
 CellID->1277801707],
Cell[2072405, 57215, 5148, 114, 285, "Text",
 CellID->696386456],
Cell[2077556, 57331, 513, 15, 24, "Text",
 CellID->981796601],
Cell[2078072, 57348, 47, 1, 12, "OpenEnvironment",
 CellID->435957245],
Cell[CellGroupData[{
Cell[2078144, 57353, 741, 19, 41, "EnvironmentHeader",
 CellID->2089484077],
Cell[2078888, 57374, 814, 24, 42, "GlobalDeclaration",
 CellID->2065858242],
Cell[2079705, 57400, 1342, 40, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:696507084", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "84"},
 CellID->696507084],
Cell[2081050, 57442, 895, 25, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:626970451", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "85"},
 CellID->626970451],
Cell[2081948, 57469, 1525, 36, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:789654176", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "86"},
 CellID->789654176],
Cell[2083476, 57507, 65, 1, 24, "EndEnvironmentMarker",
 CellID->672392106]
}, Open  ]],
Cell[2083556, 57511, 48, 1, 16, "CloseEnvironment",
 CellID->649471340],
Cell[CellGroupData[{
Cell[2083629, 57516, 11452, 270, 29, "ProofInfo",
 CellTags->{"Proof|ID:696507084", "Proof|ID:696507084-1"},
 CellID->785497368],
Cell[2095084, 57788, 525, 13, 70, "ProofInfoBody",
 CellID->1321524629]
}, Closed]],
Cell[CellGroupData[{
Cell[2095646, 57806, 8563, 207, 27, "ProofInfo",
 CellTags->{"Proof|ID:626970451", "Proof|ID:626970451-1"},
 CellID->1008962236],
Cell[2104212, 58015, 524, 13, 70, "ProofInfoBody",
 CellID->254095126]
}, Closed]],
Cell[CellGroupData[{
Cell[2104773, 58033, 8572, 207, 27, "ProofInfo",
 CellTags->{"Proof|ID:789654176", "Proof|ID:789654176-1"},
 CellID->543282043],
Cell[2113348, 58242, 525, 13, 70, "ProofInfoBody",
 CellID->1999610841]
}, Closed]],
Cell[2113888, 58258, 487, 9, 26, "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellID->995033752],
Cell[2114378, 58269, 345, 5, 24, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellID->8175233],
Cell[2114726, 58276, 508, 17, 25, "Item",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellID->1251264984],
Cell[2115237, 58295, 503, 15, 27, "Item",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellID->1241360375],
Cell[2115743, 58312, 1547, 42, 44, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellID->1860606636],
Cell[2117293, 58356, 1463, 39, 63, "Text",
 CellID->707373101],
Cell[2118759, 58397, 47, 1, 12, "OpenEnvironment",
 CellID->454483030],
Cell[CellGroupData[{
Cell[2118831, 58402, 655, 18, 41, "EnvironmentHeader",
 CellID->334142738],
Cell[2119489, 58422, 1706, 48, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1437590439", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "87"},
 CellID->1437590439],
Cell[2121198, 58472, 1572, 45, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:477242065", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "88"},
 CellID->477242065],
Cell[2122773, 58519, 65, 1, 24, "EndEnvironmentMarker",
 CellID->776728042]
}, Open  ]],
Cell[2122853, 58523, 49, 1, 16, "CloseEnvironment",
 CellID->1701216104],
Cell[CellGroupData[{
Cell[2122927, 58528, 7513, 184, 29, "ProofInfo",
 CellTags->{"Proof|ID:1437590439", "Proof|ID:1437590439-1"},
 CellID->773586293],
Cell[2130443, 58714, 524, 13, 70, "ProofInfoBody",
 CellID->140107252]
}, Closed]],
Cell[CellGroupData[{
Cell[2131004, 58732, 5334, 137, 27, "ProofInfo",
 CellTags->{"Proof|ID:477242065", "Proof|ID:477242065-1"},
 CellID->700215890],
Cell[2136341, 58871, 525, 13, 70, "ProofInfoBody",
 CellID->1979693258]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2136915, 58890, 213, 4, 22, "Subsubsection",
 CellTags->"sect::TheoremsFarExterior",
 CellID->1174010248],
Cell[2137131, 58896, 802, 18, 63, "Text",
 CellID->1038672145],
Cell[2137936, 58916, 687, 15, 62, "Text",
 CellID->167638010],
Cell[2138626, 58933, 870, 28, 25, "Text",
 CellID->42774035],
Cell[2139499, 58963, 47, 1, 12, "OpenEnvironment",
 CellID->370524994],
Cell[CellGroupData[{
Cell[2139571, 58968, 657, 16, 41, "EnvironmentHeader",
 CellID->1808105650],
Cell[2140231, 58986, 227, 6, 32, "GlobalDeclaration",
 CellID->616105124],
Cell[2140461, 58994, 228, 6, 21, "GlobalDeclaration",
 CellID->1179235710],
Cell[2140692, 59002, 296, 8, 33, "GlobalDeclaration",
 CellID->1322949089],
Cell[2140991, 59012, 793, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1041359623", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "89"},
 CellID->1041359623],
Cell[2141787, 59036, 797, 23, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:2130272336", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "90"},
 CellID->2130272336],
Cell[2142587, 59061, 872, 24, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1584641667", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "91"},
 CellID->1584641667],
Cell[2143462, 59087, 1139, 32, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1175076282", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "92"},
 CellID->1175076282],
Cell[2144604, 59121, 249, 7, 21, "GlobalDeclaration",
 CellID->1303947124],
Cell[2144856, 59130, 1215, 33, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:333980150", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "93"},
 CellID->333980150],
Cell[2146074, 59165, 863, 25, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1370308823", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "94"},
 CellID->1370308823],
Cell[2146940, 59192, 66, 1, 24, "EndEnvironmentMarker",
 CellID->1131751420]
}, Open  ]],
Cell[2147021, 59196, 49, 1, 16, "CloseEnvironment",
 CellID->1349563567],
Cell[CellGroupData[{
Cell[2147095, 59201, 6928, 174, 29, "ProofInfo",
 CellTags->{"Proof|ID:1041359623", "Proof|ID:1041359623-1"},
 CellID->650148495],
Cell[2154026, 59377, 525, 13, 70, "ProofInfoBody",
 CellID->114116961]
}, Closed]],
Cell[CellGroupData[{
Cell[2154588, 59395, 6928, 174, 27, "ProofInfo",
 CellTags->{"Proof|ID:2130272336", "Proof|ID:2130272336-1"},
 CellID->874306688],
Cell[2161519, 59571, 526, 13, 70, "ProofInfoBody",
 CellID->1181186212]
}, Closed]],
Cell[CellGroupData[{
Cell[2162082, 59589, 7636, 188, 27, "ProofInfo",
 CellTags->{"Proof|ID:1584641667", "Proof|ID:1584641667-1"},
 CellID->813414581],
Cell[2169721, 59779, 526, 13, 70, "ProofInfoBody",
 CellID->1535779146]
}, Closed]],
Cell[CellGroupData[{
Cell[2170284, 59797, 8672, 210, 27, "ProofInfo",
 CellTags->{"Proof|ID:1175076282", "Proof|ID:1175076282-1"},
 CellID->54321453],
Cell[2178959, 60009, 526, 13, 70, "ProofInfoBody",
 CellID->1473601407]
}, Closed]],
Cell[CellGroupData[{
Cell[2179522, 60027, 8526, 208, 27, "ProofInfo",
 CellTags->{"Proof|ID:333980150", "Proof|ID:333980150-1"},
 CellID->749915393],
Cell[2188051, 60237, 525, 13, 70, "ProofInfoBody",
 CellID->1548804265]
}, Closed]],
Cell[CellGroupData[{
Cell[2188613, 60255, 8440, 206, 27, "ProofInfo",
 CellTags->{"Proof|ID:1370308823", "Proof|ID:1370308823-1"},
 CellID->1775388908],
Cell[2197056, 60463, 526, 13, 70, "ProofInfoBody",
 CellID->1029312850]
}, Closed]],
Cell[2197597, 60479, 2848, 90, 169, "Text",
 CellID->1096841450],
Cell[2200448, 60571, 5490, 119, 285, "Text",
 CellID->100422464],
Cell[2205941, 60692, 1553, 46, 84, "Text",
 CellID->443594684],
Cell[2207497, 60740, 47, 1, 12, "OpenEnvironment",
 CellID->552391168],
Cell[CellGroupData[{
Cell[2207569, 60745, 659, 19, 41, "EnvironmentHeader",
 CellID->1097894802],
Cell[2208231, 60766, 228, 6, 32, "GlobalDeclaration",
 CellID->1460528271],
Cell[2208462, 60774, 371, 8, 21, "GlobalDeclaration",
 CellID->58178322],
Cell[2208836, 60784, 1216, 36, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1434774204", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "95"},
 CellID->1434774204],
Cell[2210055, 60822, 976, 30, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1867780171", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "96"},
 CellID->1867780171],
Cell[2211034, 60854, 270, 7, 30, "GlobalDeclaration",
 CellID->329161725],
Cell[2211307, 60863, 1048, 31, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:786584693", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "97"},
 CellID->786584693],
Cell[2212358, 60896, 1022, 31, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:183897269", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "98"},
 CellID->183897269],
Cell[2213383, 60929, 1403, 39, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1246401530", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "99"},
 CellID->1246401530],
Cell[2214789, 60970, 65, 1, 24, "EndEnvironmentMarker",
 CellID->146116537]
}, Open  ]],
Cell[2214869, 60974, 48, 1, 16, "CloseEnvironment",
 CellID->838044464],
Cell[CellGroupData[{
Cell[2214942, 60979, 8990, 220, 29, "ProofInfo",
 CellTags->{"Proof|ID:1434774204", "Proof|ID:1434774204-1"},
 CellID->4587845],
Cell[2223935, 61201, 524, 13, 70, "ProofInfoBody",
 CellID->704261409]
}, Closed]],
Cell[CellGroupData[{
Cell[2224496, 61219, 6493, 164, 27, "ProofInfo",
 CellTags->{"Proof|ID:1867780171", "Proof|ID:1867780171-1"},
 CellID->975302909],
Cell[2230992, 61385, 525, 13, 70, "ProofInfoBody",
 CellID->925555814]
}, Closed]],
Cell[CellGroupData[{
Cell[2231554, 61403, 7223, 180, 27, "ProofInfo",
 CellTags->{"Proof|ID:786584693", "Proof|ID:786584693-1"},
 CellID->1192023621],
Cell[2238780, 61585, 525, 13, 70, "ProofInfoBody",
 CellID->1973863468]
}, Closed]],
Cell[CellGroupData[{
Cell[2239342, 61603, 6819, 168, 27, "ProofInfo",
 CellTags->{"Proof|ID:183897269", "Proof|ID:183897269-1"},
 CellID->1230628713],
Cell[2246164, 61773, 481, 12, 70, "ProofInfoBody",
 CellID->90775516]
}, Closed]],
Cell[CellGroupData[{
Cell[2246682, 61790, 6833, 168, 27, "ProofInfo",
 CellTags->{"Proof|ID:1246401530", "Proof|ID:1246401530-1"},
 CellID->879169295],
Cell[2253518, 61960, 483, 12, 70, "ProofInfoBody",
 CellID->637478936]
}, Closed]],
Cell[2254016, 61975, 1877, 58, 104, "Text",
 CellID->1441855581],
Cell[2255896, 62035, 5971, 120, 285, "Text",
 CellID->482923784]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2261928, 62162, 149, 3, 43, "Section",
 CellTags->"sect::MainResults",
 CellID->356961332],
Cell[2262080, 62167, 907, 21, 62, "Text",
 CellID->52869115],
Cell[2262990, 62190, 259, 5, 24, "Text",
 CellID->1355349683],
Cell[CellGroupData[{
Cell[2263274, 62199, 523, 13, 38, "Subsection",
 CellTags->"sect::Invariant",
 CellID->1255092584],
Cell[2263800, 62214, 2093, 50, 83, "Text",
 CellID->1014400362],
Cell[2265896, 62266, 47, 1, 12, "OpenEnvironment",
 CellID->307314990],
Cell[CellGroupData[{
Cell[2265968, 62271, 744, 19, 40, "EnvironmentHeader",
 CellID->1542489228],
Cell[2266715, 62292, 181, 4, 21, "GlobalDeclaration",
 CellID->1656946773],
Cell[2266899, 62298, 755, 16, 57, "GlobalDeclaration",
 CellTags->"???",
 CellID->1004739293],
Cell[2267657, 62316, 892, 20, 61, "GlobalDeclaration",
 CellTags->"???",
 CellID->1633843438],
Cell[2268552, 62338, 1549, 43, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:2102726322", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "invariant"},
 CellID->2102726322],
Cell[2270104, 62383, 65, 1, 24, "EndEnvironmentMarker",
 CellID->643723384]
}, Open  ]],
Cell[2270184, 62387, 49, 1, 16, "CloseEnvironment",
 CellID->1917787441],
Cell[CellGroupData[{
Cell[2270258, 62392, 6517, 163, 29, "ProofInfo",
 CellTags->{"Proof|ID:2102726322", "Proof|ID:2102726322-1"},
 CellID->987040057],
Cell[2276778, 62557, 523, 13, 70, "ProofInfoBody",
 CellID->204517399]
}, Closed]],
Cell[2277316, 62573, 804, 17, 43, "Text",
 CellID->1493025671],
Cell[CellGroupData[{
Cell[2278145, 62594, 726, 22, 42, "Item",
 CellID->714527305],
Cell[2278874, 62618, 803, 17, 58, "Item",
 CellID->1061151323]
}, Open  ]],
Cell[2279692, 62638, 741, 22, 44, "Text",
 CellID->1082726349],
Cell[CellGroupData[{
Cell[2280458, 62664, 747, 21, 25, "Item",
 CellID->44989876],
Cell[2281208, 62687, 1383, 42, 78, "Item",
 CellID->496865089],
Cell[2282594, 62731, 583, 18, 42, "Item",
 CellID->1437698621]
}, Open  ]],
Cell[2283192, 62752, 564, 16, 44, "Text",
 CellID->1200874253]
}, Closed]],
Cell[CellGroupData[{
Cell[2283793, 62773, 557, 13, 30, "Subsection",
 CellTags->"sect::Bound",
 CellID->485945541],
Cell[2284353, 62788, 3163, 102, 180, "Text",
 CellID->1437968482],
Cell[2287519, 62892, 373, 7, 21, "GlobalDeclaration",
 CellID->746038455],
Cell[2287895, 62901, 37, 1, 24, "Text",
 CellID->1576662597],
Cell[2287935, 62904, 755, 16, 57, "GlobalDeclaration",
 CellTags->"???",
 CellID->954349703],
Cell[2288693, 62922, 1063, 33, 44, "Text",
 CellID->2127870379],
Cell[2289759, 62957, 47, 1, 12, "OpenEnvironment",
 CellID->247339406],
Cell[CellGroupData[{
Cell[2289831, 62962, 789, 24, 40, "EnvironmentHeader",
 CellID->1208461823],
Cell[2290623, 62988, 1120, 32, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:724101256", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "100"},
 CellID->724101256],
Cell[2291746, 63022, 65, 1, 24, "EndEnvironmentMarker",
 CellID->733284948]
}, Open  ]],
Cell[2291826, 63026, 48, 1, 16, "CloseEnvironment",
 CellID->498946658],
Cell[CellGroupData[{
Cell[2291899, 63031, 4269, 114, 29, "ProofInfo",
 CellTags->{"Proof|ID:724101256", "Proof|ID:724101256-1"},
 CellID->205519987],
Cell[2296171, 63147, 482, 12, 70, "ProofInfoBody",
 CellID->835900380]
}, Closed]],
Cell[2296668, 63162, 47, 1, 12, "OpenEnvironment",
 CellID->957582836],
Cell[CellGroupData[{
Cell[2296740, 63167, 756, 23, 40, "EnvironmentHeader",
 CellID->1193318642],
Cell[2297499, 63192, 1140, 31, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1990359651", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "bound"},
 CellID->1990359651],
Cell[2298642, 63225, 66, 1, 24, "EndEnvironmentMarker",
 CellID->2055477639]
}, Open  ]],
Cell[2298723, 63229, 48, 1, 16, "CloseEnvironment",
 CellID->292236524],
Cell[CellGroupData[{
Cell[2298796, 63234, 4185, 112, 29, "ProofInfo",
 CellTags->{"Proof|ID:1990359651", "Proof|ID:1990359651-1"},
 CellID->2132280415],
Cell[2302984, 63348, 526, 13, 70, "ProofInfoBody",
 CellID->1143064906]
}, Closed]],
Cell[2303525, 63364, 415, 8, 26, "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellID->618421757],
Cell[2303943, 63374, 1111, 36, 44, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10001.},
 CellID->1618954571],
Cell[2305057, 63412, 2296, 64, 102, "Text",
 CellID->1876059057]
}, Closed]],
Cell[CellGroupData[{
Cell[2307390, 63481, 316, 8, 30, "Subsection",
 CellTags->"sect::MainResult",
 CellID->1121720170],
Cell[2307709, 63491, 627, 18, 44, "Text",
 CellID->297111079],
Cell[2308339, 63511, 48, 1, 12, "OpenEnvironment",
 CellID->1239024718],
Cell[CellGroupData[{
Cell[2308412, 63516, 447, 10, 40, "EnvironmentHeader",
 CellID->1918271715],
Cell[2308862, 63528, 182, 4, 21, "GlobalDeclaration",
 CellID->1879526953],
Cell[2309047, 63534, 271, 6, 33, "GlobalDeclaration",
 CellID->592262441],
Cell[2309321, 63542, 1063, 29, 34, "GlobalDeclaration",
 CellID->1452306355],
Cell[2310387, 63573, 987, 27, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1891911403", 
   "Source:/home/amaletzk/Theorema Notebooks/Complexity/Complexity (no \
functor).nb", "main theorem"},
 CellID->1891911403],
Cell[2311377, 63602, 66, 1, 24, "EndEnvironmentMarker",
 CellID->2069358189]
}, Open  ]],
Cell[2311458, 63606, 48, 1, 16, "CloseEnvironment",
 CellID->116419994],
Cell[CellGroupData[{
Cell[2311531, 63611, 6867, 167, 29, "ProofInfo",
 CellTags->{"Proof|ID:1891911403", "Proof|ID:1891911403-1"},
 CellID->233183322],
Cell[2318401, 63780, 483, 12, 70, "ProofInfoBody",
 CellID->311475437]
}, Closed]],
Cell[2318899, 63795, 184, 4, 24, "Text",
 CellID->1678387744],
Cell[CellGroupData[{
Cell[2319108, 63803, 1169, 37, 61, "Item",
 CellID->149821961],
Cell[2320280, 63842, 2036, 70, 60, "Item",
 CellID->938510761]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2322377, 63919, 210, 4, 43, "Section",
 CellTags->"sect::Prover",
 CellID->1903011022],
Cell[2322590, 63925, 543, 12, 63, "Text",
 CellID->1608846916],
Cell[CellGroupData[{
Cell[2323158, 63941, 204, 3, 24, "Item",
 CellID->2142089810],
Cell[2323365, 63946, 135, 3, 24, "Item",
 CellID->565617232]
}, Open  ]],
Cell[2323515, 63952, 974, 16, 119, "Text",
 CellID->116836426],
Cell[2324492, 63970, 801, 13, 26, "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1090859055],
Cell[2325296, 63985, 1183, 36, 77, "Item",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->550862969],
Cell[2326482, 64023, 568, 10, 41, "Item",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1857320962],
Cell[2327053, 64035, 924, 25, 65, "Item",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->824107932],
Cell[CellGroupData[{
Cell[2328002, 64064, 115, 2, 38, "Subsection",
 CellID->78883476],
Cell[2328120, 64068, 232, 5, 24, "Text",
 CellID->2054152417],
Cell[CellGroupData[{
Cell[2328377, 64077, 579, 14, 42, "Item",
 CellID->1458172368],
Cell[2328959, 64093, 627, 18, 41, "Item",
 CellID->2016699663],
Cell[2329589, 64113, 413, 10, 41, "Item",
 CellID->1058238047],
Cell[2330005, 64125, 430, 8, 41, "Item",
 CellID->260116386]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2330484, 64139, 158, 3, 30, "Subsection",
 CellID->538287563],
Cell[CellGroupData[{
Cell[2330667, 64146, 255, 5, 24, "Item",
 CellID->608926717],
Cell[2330925, 64153, 257, 5, 24, "Item",
 CellID->917489538],
Cell[2331185, 64160, 245, 5, 24, "Item",
 CellID->527072755],
Cell[2331433, 64167, 515, 15, 24, "Item",
 CellID->86175489],
Cell[2331951, 64184, 392, 12, 24, "Item",
 CellID->1110939375],
Cell[2332346, 64198, 1053, 36, 42, "Item",
 CellID->333958307],
Cell[2333402, 64236, 766, 24, 41, "Item",
 CellID->225969300],
Cell[2334171, 64262, 892, 32, 24, "Item",
 CellID->5933658],
Cell[2335066, 64296, 418, 10, 24, "Item",
 CellID->832000632],
Cell[2335487, 64308, 809, 25, 42, "Item",
 CellID->413197361],
Cell[2336299, 64335, 536, 16, 24, "Item",
 CellID->1765219953],
Cell[2336838, 64353, 607, 18, 24, "Item",
 CellID->1448323563],
Cell[2337448, 64373, 973, 23, 58, "Item",
 CellID->1522961883]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2338470, 64402, 243, 4, 30, "Subsection",
 CellID->1059323634],
Cell[CellGroupData[{
Cell[2338738, 64410, 1010, 16, 41, "Item",
 CellID->1967607739],
Cell[2339751, 64428, 1031, 16, 24, "Item",
 CellID->1509344252],
Cell[2340785, 64446, 1369, 32, 41, "Item",
 CellID->624399819]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2342203, 64484, 314, 5, 30, "Subsection",
 CellID->1361386280],
Cell[CellGroupData[{
Cell[2342542, 64493, 4469, 145, 140, "Item",
 CellID->521411084],
Cell[2347014, 64640, 2238, 67, 61, "Item",
 CellID->132142128]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2349301, 64713, 416, 7, 30, "Subsection",
 CellTags->"sect::ProverTuples",
 CellID->1679395489],
Cell[CellGroupData[{
Cell[2349742, 64724, 197, 3, 28, "Subsubsection",
 CellID->417228170],
Cell[2349942, 64729, 5500, 167, 121, "Item",
 CellID->1268545251],
Cell[2355445, 64898, 841, 15, 43, "Text",
 CellID->219033739],
Cell[2356289, 64915, 3101, 75, 116, "Item",
 CellID->1468431481],
Cell[2359393, 64992, 444, 9, 26, "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1935127905],
Cell[2359840, 65003, 2026, 44, 145, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->1355311681]
}, Closed]],
Cell[CellGroupData[{
Cell[2361903, 65052, 231, 4, 22, "Subsubsection",
 CellID->1023287520],
Cell[CellGroupData[{
Cell[2362159, 65060, 2501, 61, 82, "Item",
 CellID->567474730],
Cell[2364663, 65123, 2238, 52, 64, "Item",
 CellID->2055491952],
Cell[2366904, 65177, 2541, 67, 84, "Item",
 CellID->1853217663]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2369494, 65250, 313, 5, 22, "Subsubsection",
 CellID->542161598],
Cell[CellGroupData[{
Cell[2369832, 65259, 4772, 128, 146, "Item",
 CellID->1872632954],
Cell[2374607, 65389, 3487, 105, 145, "Item",
 CellID->914117023]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2378155, 65501, 379, 6, 30, "Subsection",
 CellTags->"sect::ProverAddition",
 CellID->285731009],
Cell[2378537, 65509, 804, 16, 24, "Text",
 CellID->2061042824],
Cell[CellGroupData[{
Cell[2379366, 65529, 3308, 96, 89, "Item",
 CellID->600703031],
Cell[2382677, 65627, 2356, 58, 51, "Item",
 CellID->1067746268]
}, Open  ]],
Cell[2385048, 65688, 615, 20, 54, "Text",
 CellID->1062121198]
}, Closed]],
Cell[CellGroupData[{
Cell[2385700, 65713, 432, 7, 30, "Subsection",
 CellTags->"sect::ProverOrdering",
 CellID->915360441],
Cell[2386135, 65722, 999, 25, 25, "Text",
 CellID->1330669066],
Cell[CellGroupData[{
Cell[2387159, 65751, 1035, 25, 52, "Item",
 CellID->2002455362],
Cell[2388197, 65778, 2970, 88, 125, "Item",
 CellID->1349325019],
Cell[2391170, 65868, 2517, 67, 83, "Item",
 CellID->234874177],
Cell[2393690, 65937, 3532, 98, 126, "Item",
 CellID->1971664220],
Cell[2397225, 66037, 1914, 58, 70, "Item",
 CellID->1238552456]
}, Open  ]],
Cell[2399154, 66098, 731, 23, 70, "Text",
 CellID->547361213],
Cell[CellGroupData[{
Cell[2399910, 66125, 2777, 82, 70, "Item",
 CellID->567605550],
Cell[2402690, 66209, 1276, 20, 70, "Item",
 CellID->906766808]
}, Open  ]],
Cell[2403981, 66232, 429, 10, 70, "Text",
 CellID->1487453809],
Cell[2404413, 66244, 442, 9, 70, "Emphasized",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellID->907544784],
Cell[2404858, 66255, 3669, 104, 70, "Text",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 15001.},
 CellID->47814498]
}, Closed]],
Cell[CellGroupData[{
Cell[2408564, 66364, 404, 7, 30, "Subsection",
 CellTags->"sect::ProverMinMax",
 CellID->200612129],
Cell[2408971, 66373, 582, 15, 25, "Text",
 CellID->949427665],
Cell[CellGroupData[{
Cell[2409578, 66392, 3109, 105, 168, "Item",
 CellID->1140449286],
Cell[2412690, 66499, 510, 11, 24, "Item",
 CellID->1761540416]
}, Open  ]],
Cell[2413215, 66513, 768, 17, 43, "Text",
 CellID->10740590],
Cell[2413986, 66532, 4821, 120, 128, "Item",
 CellID->150731591],
Cell[2418810, 66654, 672, 21, 35, "Text",
 CellID->293040326],
Cell[CellGroupData[{
Cell[2419507, 66679, 2364, 50, 39, "Item",
 CellID->3623347],
Cell[2421874, 66731, 2363, 49, 39, "Item",
 CellID->1634809851],
Cell[2424240, 66782, 3657, 84, 65, "Item",
 CellID->785212987],
Cell[2427900, 66868, 3667, 87, 65, "Item",
 CellID->919021924]
}, Open  ]],
Cell[2431582, 66958, 682, 15, 24, "Text",
 CellID->1322323722]
}, Closed]],
Cell[CellGroupData[{
Cell[2432301, 66978, 385, 6, 30, "Subsection",
 CellTags->"sect::ProverQuantifiers",
 CellID->141176790],
Cell[CellGroupData[{
Cell[2432711, 66988, 265, 4, 28, "Subsubsection",
 CellID->592950640],
Cell[CellGroupData[{
Cell[2433001, 66996, 354, 7, 24, "Item",
 CellID->1506480535],
Cell[2433358, 67005, 356, 7, 24, "Item",
 CellID->1743222265],
Cell[2433717, 67014, 1211, 29, 53, "Item",
 CellID->1446434590],
Cell[2434931, 67045, 585, 10, 41, "Item",
 CellID->1820681756],
Cell[2435519, 67057, 565, 10, 24, "Item",
 CellID->860887451]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2436133, 67073, 193, 3, 22, "Subsubsection",
 CellID->942392630],
Cell[2436329, 67078, 742, 22, 47, "Text",
 CellID->2122823176],
Cell[CellGroupData[{
Cell[2437096, 67104, 4260, 120, 174, "Item",
 CellID->195555088],
Cell[2441359, 67226, 3528, 92, 140, "Item",
 CellID->1141312465]
}, Open  ]],
Cell[2444902, 67321, 550, 14, 70, "Text",
 CellID->1396162367],
Cell[CellGroupData[{
Cell[2445477, 67339, 5749, 171, 70, "Item",
 CellID->628443461],
Cell[2451229, 67512, 3259, 85, 70, "Item",
 CellID->1730837153]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2454549, 67604, 245, 4, 30, "Subsection",
 CellID->721887106],
Cell[CellGroupData[{
Cell[2454819, 67612, 1758, 50, 59, "Item",
 CellID->1647031003],
Cell[2456580, 67664, 1401, 31, 58, "Item",
 CellID->1971611069]
}, Open  ]],
Cell[2457996, 67698, 257, 5, 24, "Text",
 CellID->799082915]
}, Closed]],
Cell[CellGroupData[{
Cell[2458290, 67708, 329, 6, 30, "Subsection",
 CellTags->"sect::ProverRewriting",
 CellID->1387546506],
Cell[CellGroupData[{
Cell[2458644, 67718, 895, 26, 59, "Item",
 CellID->681646954],
Cell[2459542, 67746, 1163, 29, 42, "Item",
 CellID->1996689018],
Cell[2460708, 67777, 890, 16, 42, "Item",
 CellID->563991546],
Cell[2461601, 67795, 848, 16, 42, "Item",
 CellID->1703267404],
Cell[2462452, 67813, 759, 13, 24, "Item",
 CellID->389717320]
}, Open  ]],
Cell[CellGroupData[{
Cell[2463248, 67831, 119, 2, 28, "Subsubsection",
 CellID->1394287770],
Cell[2463370, 67835, 151, 2, 24, "Text",
 CellID->1647651918],
Cell[2463524, 67839, 1134, 27, 102, "Text",
 CellID->1285921573],
Cell[2464661, 67868, 1057, 23, 82, "Text",
 CellID->705235995]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2465779, 67898, 147, 3, 43, "Section",
 CellTags->"sect::Conclusion",
 CellID->1304797350],
Cell[2465929, 67903, 2009, 46, 158, "Text",
 CellID->162614402],
Cell[2467941, 67951, 1134, 28, 103, "Text",
 CellID->106646726],
Cell[2469078, 67981, 2499, 75, 150, "Text",
 CellID->215946755],
Cell[CellGroupData[{
Cell[2471602, 68060, 119, 2, 38, "Subsection",
 CellID->807851642],
Cell[2471724, 68064, 255, 6, 24, "Text",
 CellID->1748677645],
Cell[CellGroupData[{
Cell[2472004, 68074, 620, 15, 42, "Item",
 CellID->1950615972],
Cell[2472627, 68091, 610, 12, 42, "Item",
 CellID->532188132],
Cell[2473240, 68105, 647, 12, 42, "Item",
 CellID->83015438]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2473936, 68123, 125, 2, 30, "Subsection",
 CellID->1975960852],
Cell[2474064, 68127, 858, 18, 63, "Text",
 CellID->1518974484],
Cell[2474925, 68147, 303, 7, 24, "Text",
 CellID->1954625252]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2475277, 68160, 146, 3, 43, "Section",
 CellTags->"sect::References",
 CellID->664836793],
Cell[CellGroupData[{
Cell[2475448, 68167, 289, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellTags->"Buchberger1965",
 CellID->1095181227],
Cell[2475740, 68175, 735, 15, 83, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->279883113]
}, Open  ]],
Cell[CellGroupData[{
Cell[2476512, 68195, 340, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellTags->"Buchberger1979",
 CellID->1664980398],
Cell[2476855, 68204, 733, 17, 45, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->472274594]
}, Open  ]],
Cell[CellGroupData[{
Cell[2477625, 68226, 388, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10002.},
 CellTags->"Buchberger1979a",
 CellID->1739310831],
Cell[2478016, 68235, 655, 14, 44, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10002.},
 CellID->695030642]
}, Open  ]],
Cell[CellGroupData[{
Cell[2478708, 68254, 314, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10003.},
 CellTags->"Buchberger1983a",
 CellID->1873989285],
Cell[2479025, 68262, 620, 13, 45, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10003.},
 CellID->1667055178]
}, Open  ]],
Cell[CellGroupData[{
Cell[2479682, 68280, 364, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10004.},
 CellTags->"Buchberger1983b",
 CellID->212964405],
Cell[2480049, 68289, 453, 9, 43, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10004.},
 CellID->1076453871]
}, Open  ]],
Cell[CellGroupData[{
Cell[2480539, 68303, 287, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10005.},
 CellTags->"Buchberger1998",
 CellID->748631075],
Cell[2480829, 68311, 392, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10005.},
 CellID->998593826]
}, Open  ]],
Cell[CellGroupData[{
Cell[2481258, 68326, 290, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10006.},
 CellTags->"Buchberger2003",
 CellID->1719245896],
Cell[2481551, 68334, 380, 8, 43, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10006.},
 CellID->2121810688]
}, Open  ]],
Cell[CellGroupData[{
Cell[2481968, 68347, 290, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10007.},
 CellTags->"Buchberger2006",
 CellID->1062875853],
Cell[2482261, 68355, 445, 11, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10007.},
 CellID->1317313158]
}, Open  ]],
Cell[CellGroupData[{
Cell[2482743, 68371, 267, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10008.},
 CellTags->"Buchberger2014",
 CellID->473088647],
Cell[2483013, 68379, 406, 8, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10008.},
 CellID->1592513850]
}, Open  ]],
Cell[CellGroupData[{
Cell[2483456, 68392, 302, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10009.},
 CellTags->"Giese2007",
 CellID->723325741],
Cell[2483761, 68400, 457, 9, 43, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10009.},
 CellID->899421835]
}, Open  ]],
Cell[CellGroupData[{
Cell[2484255, 68414, 268, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10011.},
 CellTags->"Mayr-Meyer1982",
 CellID->1957558217],
Cell[2484526, 68422, 384, 9, 44, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10011.},
 CellID->1492494109]
}, Open  ]],
Cell[CellGroupData[{
Cell[2484947, 68436, 277, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10010.},
 CellTags->"Theorema",
 CellID->367095808],
Cell[2485227, 68444, 661, 16, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10010.},
 CellID->1720721919]
}, Open  ]],
Cell[CellGroupData[{
Cell[2485925, 68465, 291, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10011.},
 CellTags->"Windsteiger1999",
 CellID->470851853],
Cell[2486219, 68473, 611, 14, 45, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10011.},
 CellID->1151878676]
}, Open  ]],
Cell[CellGroupData[{
Cell[2486867, 68492, 283, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10012.},
 CellTags->"Winkler1984",
 CellID->805682494],
Cell[2487153, 68500, 560, 11, 45, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10012.},
 CellID->696375225]
}, Open  ]]
}, Closed]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
